/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Basic block structure represented as a cuboid.
 */


#include <config.h>

#include <iostream>
#include <cassert>
#include <cmath>
#include <cstring>
#include <typeinfo>

#if defined JWSCXX_HAVE_OPENGL
#include <GL/gl.h>
#include <GL/glu.h>
#elif defined JWSCXX_HAVE_OPENGL_FRAMEWORK
#include <OpenGL/OpenGL.h>
#endif

#include <inttypes.h>

#include <jwsc/base/limits.h>
#include <jwsc/vector/vector.h>
#include <jwsc/vector/vector_math.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/matrix/matrix_math.h>
#include <jwsc/prob/pdf.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/graphics/cuboid.h>
#include <jwsc++/graphics/polygon.h>

#include "model.h"
#include "component.h"
#include "object.h"
#include "block.h"


#define  SIZE                 50.0
#define  SIZE_MINN            10.0
#define  SIZE_MAXX            1000.0
#define  SIZE_MU              50.0
#define  SIZE_SIGMA           10.0
#define  SIZE_LAMBDA          100.0
#define  SIZE_ALPHA           0.1
#define  SIZE_BETA            10.0
#define  RATIO_SIZE_MU        1.0
#define  RATIO_SIZE_LAMBDA    1.0
#define  RATIO_SIZE_ALPHA     1.0
#define  RATIO_SIZE_BETA      0.01
#define  SIZE_DELTA_T         0.005
#define  SIZE_GRAD_ETA        1.0
#define  SIZE_SIGMA_DELTA_T   0.0001
#define  SIZE_SIGMA_GRAD_ETA  0.5
#define  RATIO_SIZE_DELTA_T   0.00005
#define  RATIO_SIZE_GRAD_ETA  0.005
#define  RATIO_SIZE_SIGMA_DELTA_T   0.0001
#define  RATIO_SIZE_SIGMA_GRAD_ETA  0.005

#define  POSN                 0.0
#define  POSN_MIN            -10000.0
#define  POSN_MAX             10000.0
#define  POSN_MU              0.0
#define  POSN_SIGMA           5.0
#define  POSN_LAMBDA          10.0
#define  POSN_ALPHA           0.5
#define  POSN_BETA            1.0
#define  POSN_DELTA_T         0.005
#define  POSN_GRAD_ETA        1.0
#define  POSN_SIGMA_DELTA_T   0.0001
#define  POSN_SIGMA_GRAD_ETA  0.5


/* -------------------------  Abstract_block::Context  ---------------------- */

Abstract_block::Context::Context() : Component::Context()
{
    typedef Model::Parameter::Context MPContext;
    typedef MPContext::Range          Range;
    typedef MPContext::Dynamics       Dynamics;
    typedef MPContext::Uniform        Uniform;
    typedef MPContext::Gaussian       Gaussian;
    typedef MPContext::Gaussian_gamma Gaussian_gamma;

    MPContext* c;

    for (uint32_t i = 0; i < 3; i++)
    {
        _size_c.add_context(MPContext());

        c = &(_size_c[ i ]);
        c->range(Range(SIZE_MINN, SIZE_MAXX));
        c->init_value(SIZE);
        c->likelihood(Gaussian(SIZE_MU, SIZE_SIGMA));
        c->prior(Gaussian_gamma(SIZE_MU, SIZE_LAMBDA, SIZE_ALPHA, SIZE_BETA));
        c->ratio_prior(Gaussian_gamma(RATIO_SIZE_MU, RATIO_SIZE_LAMBDA,
                    RATIO_SIZE_ALPHA, RATIO_SIZE_BETA));
        c->proposal(Gaussian(SIZE_MU, SIZE_SIGMA));
        c->gauss_proposal(Gaussian(SIZE_MU, SIZE_SIGMA));
        c->value_dynamics(Dynamics(SIZE_DELTA_T, SIZE_GRAD_ETA));
        c->mean_dynamics(Dynamics(SIZE_DELTA_T, SIZE_GRAD_ETA));
        c->stdev_dynamics(Dynamics(SIZE_SIGMA_DELTA_T, SIZE_SIGMA_GRAD_ETA));
        c->ratio_mean_dynamics(Dynamics(RATIO_SIZE_DELTA_T,
                    RATIO_SIZE_GRAD_ETA));
        c->ratio_stdev_dynamics(Dynamics(RATIO_SIZE_SIGMA_DELTA_T,
                    RATIO_SIZE_SIGMA_GRAD_ETA));
    }

    for (uint32_t i = 0; i < 2; i++)
    {
        _var_attach_pts_c.add_context(MPContext());

        c = &(_var_attach_pts_c[ i ]);
        c->range(Range(POSN_MIN, POSN_MAX));
        c->init_value(POSN);
        c->likelihood(Gaussian(POSN_MU, POSN_SIGMA));
        c->prior(Gaussian_gamma(POSN_MU, POSN_LAMBDA, POSN_ALPHA, POSN_BETA));
        c->proposal(Gaussian(POSN_MU, POSN_SIGMA));
        c->gauss_proposal(Gaussian(POSN_MU, POSN_SIGMA));
        c->value_dynamics(Dynamics(POSN_DELTA_T, POSN_GRAD_ETA));
        c->mean_dynamics(Dynamics(POSN_DELTA_T, POSN_GRAD_ETA));
        c->stdev_dynamics(Dynamics(POSN_SIGMA_DELTA_T, POSN_SIGMA_GRAD_ETA));
    }
}

Abstract_block::Context::Context(const Abstract_block::Context& ctx)
    : Component::Context(ctx)
{
    *this = ctx;
}

Abstract_block::Context::~Context()
{
}

Abstract_block::Context& Abstract_block::Context::operator=
(
    const Abstract_block::Context& ctx
)
{
    Component::Context::operator=(ctx);

    _size_c = ctx._size_c;
    _var_attach_pts_c = ctx._var_attach_pts_c;

    return *this;
}

const Component::Paramgroup::Context& Abstract_block::Context::size_c() const
{
    return _size_c;
}

void Abstract_block::Context::size_c(const Component::Paramgroup::Context& size)
{
    _size_c = size;
}

const Model::Parameter::Context& Abstract_block::Context::size_c(size_t i) const
{
    return _size_c[ i ];
}

void Abstract_block::Context::size_c
(
    size_t                           i, 
    const Model::Parameter::Context& size
)
{
    _size_c[ i ] = size;
}

const Component::Paramgroup::Context& 
Abstract_block::Context::var_attach_pts_c() const
{
    return _var_attach_pts_c;
}

void Abstract_block::Context::var_attach_pts_c
(
    const Component::Paramgroup::Context& p
)
{
    _var_attach_pts_c = p;
}

const Model::Parameter::Context& Abstract_block::Context::var_attach_pt_c
(
    size_t i
) 
const
{
    return _var_attach_pts_c[ i ];
}

void Abstract_block::Context::var_attach_pt_c
(
    size_t                           i, 
    const Model::Parameter::Context& p
)
{
    _var_attach_pts_c[ i ] = p;
}

/* ========================================================================== */






/* ---------------------------  Abstract_block  ----------------------------- */

/**
 * The block is centered at (0, 0, 0).
 *
 * @param  name  Name of this type of block subclass.
 * @param  ctx   Parameter context for this block.
 *
 * @throw  jwscxx::base::Arg_error  Dimensions must be > 0.
 */
Abstract_block::Abstract_block
(
    const char*                    name, 
    const Abstract_block::Context* ctx
) 
throw (jwscxx::base::Arg_error)
    : Component(name, 0, 0, 0, ctx),
      _size_p(&(ctx->size_c())),
      _attachments(6, Component::Attachment(0, false)),
      _cuboid(ctx->size_c(0).init_value(),
              ctx->size_c(1).init_value(),
              ctx->size_c(2).init_value(),
              -0.5*ctx->size_c(0).init_value(), 
              -0.5*ctx->size_c(1).init_value(),
              -0.5*ctx->size_c(2).init_value())
{
    size_t id = 3;

    _size_p.add_param(Parameter(id++, "Width", ctx->size_c(0).init_value(),
                &(ctx->size_c(0)), this));
    _size_p.add_param(Parameter(id++, "Height", ctx->size_c(1).init_value(),
                &(ctx->size_c(1)), this));
    _size_p.add_param(Parameter(id++,"Length", ctx->size_c(2).init_value(),
                &(ctx->size_c(2)), this));

    add_var_param(&(_size_p[0]));
    add_var_param(&(_size_p[1]));
    add_var_param(&(_size_p[2]));

    create_face_rotations();

    create_ctr_attach_pts(&id);
    add_ctr_attach_shift_deps();
    add_ctr_attach_scale_deps();

    create_var_attach_pts(&id, ctx);
    add_var_attach_shift_deps();
    add_var_attach_scale_deps();
    add_var_attach_conds();
}

/**
 * Used by block subclasses only.  The context @em ctx can safely be null,
 * assuming appropriate use thereafter.
 *
 * @param  name    Name of this type of block subclass.
 * @param  width   Width of the block (x).
 * @param  height  Height of the block (y).
 * @param  length  Length of the block (z).
 * @param  x       Center x position.
 * @param  y       Center y position.
 * @param  z       Center z position.
 * @param  ctx     Block context, can be null.
 */
Abstract_block::Abstract_block
(
    const char* name, 
    double      width, 
    double      height, 
    double      length, 
    double      x, 
    double      y, 
    double      z,
    const Abstract_block::Context* ctx
)
throw (jwscxx::base::Arg_error)
: Component(name, x, y, z, ctx),
  _size_p(0),
  _attachments(6, Component::Attachment(0, false)),
  _cuboid(width, height, length, x - 0.5*width, y - 0.5*height, z - 0.5*length)
{
    if (ctx)
    {
        _size_p.context(&(ctx->size_c()));
    }

    size_t id = 3;

    _size_p.add_param(Parameter(id++, "Width", width,
                (ctx) ? &(ctx->size_c(0)) : 0, this));
    _size_p.add_param(Parameter(id++, "Height", height,
                (ctx) ? &(ctx->size_c(1)) : 0, this));
    _size_p.add_param(Parameter(id++, "Length", length,
                (ctx) ? &(ctx->size_c(2)) : 0, this));

    add_var_param(&(_size_p[0]));
    add_var_param(&(_size_p[1]));
    add_var_param(&(_size_p[2]));

    create_face_rotations();

    create_ctr_attach_pts(&id);
    add_ctr_attach_shift_deps();
    add_ctr_attach_scale_deps();

    create_var_attach_pts(&id, ctx);
    add_var_attach_shift_deps();
    add_var_attach_scale_deps();
    add_var_attach_conds();
}

/** 
 * Does not re-initialize the dependents for this (new) block; the parameter
 * dependency pointers from @em b are copied over.
 *
 * @param  b  Block to copy into this one.
 */
Abstract_block::Abstract_block(const Abstract_block& b) 
    : Component(b), 
      _size_p(b._size_p),
      _ctr_attach_pts_p(b._ctr_attach_pts_p),
      _var_attach_pts_p(b._var_attach_pts_p),
      _attachments(b._attachments), 
      _cuboid(b._cuboid)
{
    add_var_param(&(_size_p[0]));
    add_var_param(&(_size_p[1]));
    add_var_param(&(_size_p[2]));

    for (size_t i = 0; i < 6; i++)
    {
        add_derived_param(&(_ctr_attach_pts_p[ i ][0]));
        add_derived_param(&(_ctr_attach_pts_p[ i ][1]));
        add_derived_param(&(_ctr_attach_pts_p[ i ][2]));
    }

    for (size_t i = 0; i < 6; i++)
    {
        add_derived_param(&(_var_attach_pts_p[ i ][0]));
        add_derived_param(&(_var_attach_pts_p[ i ][1]));
        add_derived_param(&(_var_attach_pts_p[ i ][2]));
    }

    enable_var_attach_pts();
    create_face_rotations();

    *this = b;
}

Abstract_block::~Abstract_block()
{
    free_face_rotations();
}

/** 
 * @param  b  Block to copy into this one.
 *
 * @return A reference to this block.
 */
Abstract_block& Abstract_block::operator=(const Abstract_block& b)
{
    Component::operator=(b);

    _attachments = b._attachments;
    _cuboid      = b._cuboid;

    _size_p = b._size_p;
    for (size_t i = 0; i < 3; i++)
    {
        _size_p[ i ].model(this);
    }

    _ctr_attach_pts_p = b._ctr_attach_pts_p;
    _var_attach_pts_p = b._var_attach_pts_p;
    for (size_t i = 0; i < 6; i++)
    {
        for (size_t j = 0; j < 3; j++)
        {
            _ctr_attach_pts_p[ i ][ j ].model(this);
            _var_attach_pts_p[ i ][ j ].model(this);
        }
    }

    return *this;
}

/** 
 * The file format is ASCII.
 *
 * @param  in  Input stream to read this block from.
 *
 * @throw  jwscxx::base::IO_error   Could not read from @em in.
 * @throw  jwscxx::base::Arg_error  Invalid arguments in file to read from.
 */
void Abstract_block::read(std::istream& in) 
    throw (jwscxx::base::IO_error, jwscxx::base::Arg_error)
{
    using std::ostringstream;
    using std::istringstream;
    using jwscxx::base::Arg_error;

    const char* type_name = typeid(*this).name();
    const char* field_value;

    if (!(field_value = read_field_value(in, "Type")))
    {
        throw Arg_error("Missing Type field");
    }
    if (strncmp(field_value, type_name, strlen(type_name)) != 0)
    {
        ostringstream ost;
        ost << "Tried to read a '" << field_value << "' as a '"
            << type_name << "'";
        throw Arg_error(ost.str());
    }

    for (size_t i = 0; i < 6; i++)
    {
        _attachments[ i ].read(in);
    }

    enable_var_attach_pts();

    Component::read(in);

    _cuboid.read(in);
}

/**
 * The file format is ASCII.
 *
 * @param  out  Output stream to write this block to.
 *
 * @throw  jwscxx::base::IO_error  Could not write to @em out.
 */
void Abstract_block::write(std::ostream& out) const 
    throw (jwscxx::base::IO_error)
{
    out << "Type: " << typeid(*this).name() << std::endl;

    for (size_t i = 0; i < 6; i++)
    {
        _attachments[ i ].write(out);
    }

    Component::write(out);

    _cuboid.write(out);
}

const Abstract_block::Context* Abstract_block::context() const
{
    return dynamic_cast<const Abstract_block::Context*>(Component::context());
}

/** 
 * We overload this function because the variable attachment points need to be
 * transformed into face coordinates.
 */
double Abstract_block::get_var_param_value(uint32_t d) const
{
    using namespace jwsc;

    Vector_d* v = 0;
    get_var_param_values(&v);
    assert(d < v->num_elts);
    double v_d = v->elts[ d ];
    free_vector_d(v);
    return v_d;
}

void Abstract_block::get_var_param_values(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    Model::get_var_param_values(p_out);
    Vector_d* p = *p_out;

    // Skip the size parameters and transform the variable attachment points
    // into face coords.
    d += 3;

    Vector_d* v = 0;
    for (size_t j = 0; j < 6; j++)
    {
        if (! _attachments[ j ].is_variable())
            continue;

        var_attach_pt_v(&v, j);

        if (object()) object()->transform_pt_from_world_to_object_coords(v);
        transform_pt_from_object_to_component_coords(v);
        transform_pt_from_block_to_face_coords(v, j);
        assert(fabs(v->elts[2]) <= 1.0e-6);

        p->elts[ d+0 ] = v->elts[0];
        p->elts[ d+1 ] = v->elts[1];
        d += 2;
    }
    free_vector_d(v);

    assert(d <= D);
}

/** 
 * The block must already be in Object coordinates, and for efficiency reasons,
 * the graphical representation is not updated, but a flag is set to indicate
 * that the update_graphical_rep function should be called when opportune.
 *
 * If a Parameter value could not be set, it is left unchanged and an exception
 * is thrown, after an attempt to set the other parameters is made.
 */
void Abstract_block::set_var_param_values_in_object_coords
(
    const jwsc::Vector_d* p
)
throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();
    double   delta;
    double   epsilon = 1.0e-8;
    double   e;

    assert(D == p->num_elts);

    std::ostringstream ost;

    for (size_t i = 0; i < 3; i++)
    {
        if (fabs(delta = p->elts[ d ] - size(i)) > 0)
        {
            e = (delta < 0) ? -epsilon : epsilon;
            try 
            { 
                _size_p[ i ].test_value_shift_deps(size(i) + delta + e); 
                _size_p[ i ].value_shift_deps(size(i) + delta);
            }
            catch (jwscxx::base::Arg_error e) 
            { ost << e.get_msg() << "  "; }
        }
        d++;
    }

    Vector_d* v_old = 0;
    Vector_d* v_new = 0;
    create_vector_d(&v_new, 4);

    for (size_t j = 0; j < 6; j++)
    {
        if (! _attachments[ j ].is_variable())
            continue;

        var_attach_pt_v(&v_old, j);

        v_new->elts[0] = p->elts[ d+0 ];
        v_new->elts[1] = p->elts[ d+1 ];
        v_new->elts[2] = 0;
        v_new->elts[3] = 1;

        transform_pt_from_face_to_block_coords(v_new, j);
        transform_pt_from_component_to_object_coords(v_new);

        for (size_t i = 0; i < 3; i++)
        {
            if (fabs(delta = v_new->elts[ i ] - v_old->elts[ i ]) > 0)
            {
                double v_i = v_new->elts[ i ];
                e = (delta < 0) ? -epsilon : epsilon;
                try
                {
                    _var_attach_pts_p[ j ][ i ].test_value_shift_deps(v_i + e);
                    _var_attach_pts_p[ j ][ i ].value_shift_deps(v_i);
                }
                catch (jwscxx::base::Arg_error e) 
                { ost << e.get_msg() << "  "; }
            }
        }

        d += 2;
    }

    free_vector_d(v_new);
    free_vector_d(v_old);

    assert(d <= D);

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

/**
 * The block must already be in Object coordinates, and for efficiency reasons,
 * the graphical representation is not updated, but a flag is set to indicate
 * that the update_graphical_rep function should be called when opportune.
 *
 * If a Parameter value could not be set, it is left unchanged and an exception
 * is thrown, after an attempt to set the other parameters is made.
 *
 * @param  p  Point to add to the Parameter values of this block.
 *
 * @throw  jwscxx::base::Arg_error  Invalid resulting parameter value.
 */
void Abstract_block::add_var_param_values_shift_deps(const jwsc::Vector_d* p)
    throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    uint32_t D = num_var_params();
    assert(D == p->num_elts);

    Vector_d* q = 0;
    get_var_param_values(&q);
    add_vectors_d(&q, p, q);

    try { set_var_param_values(q); }
    catch (jwscxx::base::Arg_error e)
    {
        free_vector_d(q);
        throw e;
    }

    free_vector_d(q);
}

/**
 * The block must already be in Object coordinates, and for efficiency reasons,
 * the graphical representation is not updated, but a flag is set to indicate
 * that the update_graphical_rep function should be called when opportune.
 *
 * If a Parameter value could not be set, it is left unchanged and an exception
 * is thrown, after an attempt to set the other parameters is made.
 *
 * @param  p  Point to add to the Parameter values of this block.
 *
 * @throw  jwscxx::base::Arg_error  Invalid resulting parameter value.
 */
void Abstract_block::add_var_param_values_scale_deps(const jwsc::Vector_d* p)
    throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();
    double   delta;
    double   epsilon = 1.0e-8;
    double   e;

    assert(D == p->num_elts);

    std::ostringstream ost;

    for (size_t i = 0; i < 3; i++)
    {
        if (fabs(delta = p->elts[ d ]) > 0)
        {
            e = (delta < 0) ? -epsilon : epsilon;
            try 
            { 
                _size_p[ i ].test_value_scale_deps(size(i) + delta + e); 
                _size_p[ i ].value_scale_deps(size(i) + delta); 
            }
            catch (jwscxx::base::Arg_error e) 
            { ost << e.get_msg() << "  "; }
        }
        d++;
    }

    Vector_d* v_old = 0;
    Vector_d* v_old_face = 0;
    Vector_d* v_new = 0;
    create_vector_d(&v_new, 4);

    for (size_t j = 0; j < 6; j++)
    {
        if (! _attachments[ j ].is_variable())
            continue;

        var_attach_pt_v(&v_old, j);
        var_attach_pt_v(&v_old_face, j);
        transform_pt_from_object_to_component_coords(v_old_face);
        transform_pt_from_block_to_face_coords(v_old_face, j);
        assert(fabs(v_old_face->elts[2]) <= 1.0e-6);

        v_new->elts[0] = v_old_face->elts[0] + p->elts[ d+0 ];
        v_new->elts[1] = v_old_face->elts[1] + p->elts[ d+1 ];
        v_new->elts[2] = 0;
        v_new->elts[3] = 1;

        transform_pt_from_face_to_block_coords(v_new, j);
        transform_pt_from_component_to_object_coords(v_new);

        for (size_t i = 0; i < 3; i++)
        {
            if (fabs(delta = v_new->elts[ i ] - v_old->elts[ i ]) > 0)
            {
                double v_i = v_new->elts[ i ];
                e = (delta < 0) ? -epsilon : epsilon;
                try
                {
                    _var_attach_pts_p[ j ][ i ].test_value_shift_deps(v_i + e);
                    _var_attach_pts_p[ j ][ i ].value_shift_deps(v_i);
                }
                catch (jwscxx::base::Arg_error e) 
                { ost << e.get_msg() << "  "; }
            }
        }

        d += 2;
    }

    free_vector_d(v_new);
    free_vector_d(v_old_face);
    free_vector_d(v_old);

    assert(d <= D);

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

void Abstract_block::set_var_attach_pt_values
(
    const jwsc::Vector_d* uv, 
    size_t                j
)
throw (jwscxx::base::Arg_error)
{
    uint32_t D = 2;
    assert(uv->num_elts == D);

    std::ostringstream ost;

    for (size_t i = 0; i < D; i++)
    {
        try 
        { 
            test_var_attach_pt_shift_deps(j, i, uv->elts[ i ]);
            var_attach_pt_shift_deps(j, i, uv->elts[ i ]);
        }
        catch (jwscxx::base::Arg_error e) 
        { ost << e.get_msg() << "  "; }
    }

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

void Abstract_block::set_var_attach_pt_stats
(
    const jwsc::Vector_d* means,
    const jwsc::Vector_d* stdevs,
    size_t                j
)
throw (jwscxx::base::Arg_error)
{
    uint32_t D = 2;
    assert(means->num_elts == D);
    assert(stdevs->num_elts == D);

    std::ostringstream ost;
    Paramgroup& v_p = _var_attach_pts_p[ j ];

    for (size_t xyz = 0; xyz < 3; xyz++)
    {
        size_t i = attach_pt_face_coord(j, xyz);
        if (i == 2) continue;

        double m = means->elts[ i ];
        double s = stdevs->elts[ i ];
        try 
        { 
            v_p[ xyz ].test_stats(m, s);
            v_p[ xyz ].stats(m, s);
        }
        catch (jwscxx::base::Arg_error e) 
        { ost << e.get_msg() << "  "; }
    }

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

void Abstract_block::print_var_param_values
(
    uint32_t      iteration, 
    std::ostream& out
)
const
{
    Component::print_var_param_values(iteration, out);

    for (size_t i = 0; i < 3; i++)
    {
        if (size_p(i)->num_conds() == 1)
        {
            double x = size_p(i)->value();
            double x_cond = size_p(i)->cond(0)->ptr()->value();
            double r = x / x_cond;

            out << iteration << ',' << name() << ',' << id()  << ','
                << "Ratio" << ',' << size_p(i)->name() << ','
                << r << std::endl;
        }
    }
}

void Abstract_block::copy_var_param_stats(const Abstract_block* b)
{
    using namespace jwsc;

    double m, s;
    for (uint32_t i = 0; i < 3; i++)
    {
        b->_size_p[ i ].stats(&m, &s);
        _size_p[ i ].stats(m, s);
    }

    for (uint32_t j = 0; j < 6; j++)
    {
        const Paramgroup& v_p = *(b->var_attach_pt_p(j));
        for (uint32_t i = 0; i < 3; i++)
        {
            v_p[ i ].stats(&m, &s);
            _var_attach_pts_p[ j ][ i ].stats(m, s);
        }
    }
}

/** @brief Scales the size parameter stats. */
void Abstract_block::scale_size_stats
(
    Abstract_block* b, 
    size_t          j
)
{
    typedef Parameter P;

    for (size_t i = 0; i < 3; i++)
    {
        if (i == j/2)
            continue;

        P* b1_s_i = &(this->_size_p[ i ]);
        P* b2_s_i = &(b->_size_p[ i ]);

        b1_s_i->mean(b2_s_i->value()*b1_s_i->mean());
        b1_s_i->stdev(b2_s_i->value()*b1_s_i->stdev());
    }
}

void Abstract_block::swap_var_attach_pts(const Abstract_block* b)
{
    using namespace jwsc;

    Vector_d* uv = 0;
    Vector_d* means = 0;
    Vector_d* stdevs = 0;

    create_vector_d(&means, 2);
    create_vector_d(&stdevs, 2);

    Object* obj = b->object();
    assert(obj);

    for (uint32_t i = 0; i < 6; i++)
    {
        if (! b->has_var_attach_pt(i))
            continue;

        if (i != 2 && i != 3)
            continue;

        // Swap the value.
        uint32_t j = swapped_face_index(i);
        b->var_attach_pt_v(&uv, i);
        obj->transform_pt_from_world_to_object_coords(uv);
        b->transform_pt_from_object_to_component_coords(uv);
        b->transform_pt_from_block_to_face_coords(uv, i);
        assert(fabs(uv->elts[2]) <= 1.0e-6);
        if (i == 2)
        {
            double x = uv->elts[0];
            uv->elts[0] = -uv->elts[1];
            uv->elts[1] = x;
        }
        else if (i == 3)
        {
            double x = uv->elts[0];
            uv->elts[0] = uv->elts[1];
            uv->elts[1] = -x;
        }
        transform_pt_from_face_to_block_coords(uv, j);
        transform_pt_from_component_to_object_coords(uv);
        var_attach_pt_v(uv, j);
    }

    free_vector_d(means);
    free_vector_d(stdevs);
    free_vector_d(uv);
}

double Abstract_block::world_coords_min_y() const
{
    return posn(1) - 0.5*size(1);
}

const Component::Paramgroup* Abstract_block::size_p() const
{
    return &_size_p;
}

const Component::Parameter* Abstract_block::size_p(size_t i) const
{
    return &(_size_p[ i ]);
}

/**
 * The resulting vector contains (indexed by) the width(0), height(1), and
 * length(2).
 */
void Abstract_block::size_v(jwsc::Vector_d** s_out) const
{
    using namespace jwsc;

    create_vector_d(s_out, 3);
    Vector_d* s = *s_out;

    s->elts[0] = _size_p[0].value();
    s->elts[1] = _size_p[1].value();
    s->elts[2] = _size_p[2].value();
}

double Abstract_block::size(size_t i) const
{
    return size_p(i)->value();
}

void Abstract_block::size_shift_deps(size_t i, double size)
{
    assert(i < 3);
    transform_from_world_to_object_coords();
    _size_p[ i ].value_shift_deps(size);
    update_object();
    transform_from_object_to_world_coords();
}

void Abstract_block::size_scale_deps(size_t i, double size)
{
    assert(i < 3);
    transform_from_world_to_object_coords();
    _size_p[ i ].value_scale_deps(size);
    update_object();
    transform_from_object_to_world_coords();
}

/**
 * @param  i       Size index.
 * @param  size    Size to test.
 *
 * @throw  jwscxx::base::Arg_error  Invalid size.
 */
void Abstract_block::test_size_shift_deps(size_t i, double size)
    throw (jwscxx::base::Arg_error)
{
    transform_from_world_to_object_coords();
    try 
    {
        _size_p[ i ].test_value_shift_deps(size); 
    }
    catch (jwscxx::base::Arg_error e)
    {
        transform_from_object_to_world_coords();
        throw e;
    }
    transform_from_object_to_world_coords();
}

/**
 * @param  i       Size index.
 * @param  size    Size to test.
 *
 * @throw  jwscxx::base::Arg_error  Invalid size.
 */
void Abstract_block::test_size_scale_deps(size_t i, double size)
    throw (jwscxx::base::Arg_error)
{
    transform_from_world_to_object_coords();
    try
    {
        _size_p[ i ].test_value_scale_deps(size);
    }
    catch (jwscxx::base::Arg_error e)
    {
        transform_from_object_to_world_coords();
        throw e;
    }
    transform_from_object_to_world_coords();
}

/**
 * The resulting attachment point is in homogeneous coordinates and the same
 * coordinate frame as the object.
 *
 * @param  a_out  The attachment point.
 * @param  i      Face to get the attachment point of.
 */
void Abstract_block::ctr_attach_pt_v(jwsc::Vector_d** a_out, size_t i) const
{
    using namespace jwsc;

    assert(i < 6);

    create_vector_d(a_out, 4);
    Vector_d* a = *a_out;

    a->elts[0] = _ctr_attach_pts_p[ i ][0].value();
    a->elts[1] = _ctr_attach_pts_p[ i ][1].value();
    a->elts[2] = _ctr_attach_pts_p[ i ][2].value();
    a->elts[3] = 1;
}

/**
 * The vector must be in homogeneous coordinates.
 *
 * @param  a  The attachment point.
 * @param  i  Face to set the attachment point of.
 */
void Abstract_block::ctr_attach_pt_v(const jwsc::Vector_d* a, size_t i)
{
    assert(i < 6);
    assert(a->num_elts == 4);
    assert(fabs(a->elts[3] - 1.0) <= 1.0e-6);

    _ctr_attach_pts_p[ i ][0].value(a->elts[0]);
    _ctr_attach_pts_p[ i ][1].value(a->elts[1]);
    _ctr_attach_pts_p[ i ][2].value(a->elts[2]);
}

const Component::Paramgroup* Abstract_block::var_attach_pt_p(size_t i) const
{
    return &_var_attach_pts_p[ i ];
}

/**
 * The position is set in face coordinates @em u or @em v. Select which one with
 * the parameter @em i.
 *
 * @param  j   The face containing the variable attachment point to set.
 * @param  i   The position coordinate of the attachment to set, either u=0 or
 *             v=1.
 * @param  uv  The position to set the attachment point to in u,v face coords.
 */
void Abstract_block::var_attach_pt_shift_deps(size_t j, size_t i, double uv)
{
    using namespace jwsc;

    assert(j < 6);
    assert(i < 2);

    Vector_d* v_old = 0;
    Vector_d* v_new = 0;

    transform_from_world_to_object_coords();

    var_attach_pt_v(&v_old, j);
    copy_vector_d(&v_new, v_old);

    transform_pt_from_object_to_component_coords(v_new);
    transform_pt_from_block_to_face_coords(v_new, j);

    v_new->elts[ i ] = uv;
    assert(fabs(v_new->elts[2]) <= 1.0e-6);

    transform_pt_from_face_to_block_coords(v_new, j);
    transform_pt_from_component_to_object_coords(v_new);

    for (size_t k = 0; k < 3; k++)
    {
        if (fabs(v_new->elts[ k ] - v_old->elts[ k ]) > 0)
        {
            _var_attach_pts_p[ j ][ k ].value_shift_deps(v_new->elts[ k ]);
        }
    }

    update_object();
    transform_from_object_to_world_coords();

    free_vector_d(v_new);
    free_vector_d(v_old);
}

/**
 * The position is set in face coordinates @em u or @em v. Select which one with
 * the parameter @em i.
 *
 * @param  j   The face containing the variable attachment point to set.
 * @param  i   The position coordinate of the attachment to set, either u=0 or
 *             v=1.
 * @param  uv  The position to set the attachment point to in u,v face coords.
 */
void Abstract_block::test_var_attach_pt_shift_deps
(
    size_t j, 
    size_t i, 
    double uv
)
throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    assert(j < 6);
    assert(i < 2);

    Vector_d* v_new = 0;
    Vector_d* v_old = 0;

    transform_from_world_to_object_coords();

    try 
    {
        var_attach_pt_v(&v_old, j);
        copy_vector_d(&v_new, v_old);

        transform_pt_from_object_to_component_coords(v_new);
        transform_pt_from_block_to_face_coords(v_new, j);

        v_new->elts[ i ] = uv;
        assert(fabs(v_new->elts[2]) <= 1.0e-6);

        transform_pt_from_face_to_block_coords(v_new, j);
        transform_pt_from_component_to_object_coords(v_new);

        for (size_t k = 0; k < 3; k++)
        {
            if (fabs(v_new->elts[ k ] - v_old->elts[ k ]) > 0)
            {
                _var_attach_pts_p[ j ][ i ].test_value_shift_deps(
                        v_new->elts[ k ]);
            }
        }
    }
    catch (jwscxx::base::Arg_error e)
    {
        transform_from_object_to_world_coords();
        free_vector_d(v_new);
        free_vector_d(v_old);
        throw e;
    }

    transform_from_object_to_world_coords();

    free_vector_d(v_new);
    free_vector_d(v_old);
}

/**
 * The resulting attachment point is in homogeneous coordinates and the same
 * coordinate frame as the object.
 *
 * @param  a_out  The attachment point.
 * @param  i      Face to get the attachment point of.
 */
void Abstract_block::var_attach_pt_v(jwsc::Vector_d** a_out, size_t i) const
{
    using namespace jwsc;

    assert(i < 6);

    create_vector_d(a_out, 4);
    Vector_d* a = *a_out;

    a->elts[0] = _var_attach_pts_p[ i ][0].value();
    a->elts[1] = _var_attach_pts_p[ i ][1].value();
    a->elts[2] = _var_attach_pts_p[ i ][2].value();
    a->elts[3] = 1;
}

/**
 * The vector must be in homogeneous coordinates.
 *
 * @param  a  The attachment point.
 * @param  i  Face to set the attachment point of.
 */
void Abstract_block::var_attach_pt_v(const jwsc::Vector_d* a, size_t i)
{
    assert(i < 6);
    assert(a->num_elts == 4);
    assert(fabs(a->elts[3] - 1.0) <= 1.0e-6);

    _var_attach_pts_p[ i ][0].value(a->elts[0]);
    _var_attach_pts_p[ i ][1].value(a->elts[1]);
    _var_attach_pts_p[ i ][2].value(a->elts[2]);
}

size_t Abstract_block::num_attachments() const
{
    size_t n = 0;

    for (size_t i = 0; i < 6; i++)
    {
        if (_attachments[ i ].ptr())
        {
            n++;
        }
    }

    return n;
}

size_t Abstract_block::num_ctr_attachments() const
{
    return num_attachments() - num_var_attachments();
}

size_t Abstract_block::num_var_attachments() const
{
    size_t n = 0;

    for (size_t i = 0; i < 6; i++)
    {
        if (_attachments[ i ].is_variable())
        {
            n++;
        }
    }

    return n;
}

size_t Abstract_block::attachment(const Component* c) const 
    throw (jwscxx::base::Arg_error)
{
    for (size_t i = 0; i < 6; i++)
    {
        if (c == attachment(i))
        {
            return i;
        }
    }

    throw jwscxx::base::Arg_error("Component is not an attachment");
}

/** May return 0, indicating there is no attachment at that point. */
Component* Abstract_block::attachment(size_t i) const
{
    assert(i < num_ctr_attach_pts());
    return _attachments[ i ].ptr();
}

void Abstract_block::attachment(size_t i, Component* c)
{
    assert(i < num_ctr_attach_pts());
    _attachments[ i ].ptr(c);
}

void Abstract_block::attach_shift_deps(uint32_t j, uint32_t i)
{
    typedef Component::Parameter P;

    assert(j < 6);
    assert(j % 2 == 0);

    P* s_i      = &(_size_p[ i ]);
    P* a_pos_ji = &(_ctr_attach_pts_p[ j+0 ][ i ]);
    P* a_neg_ji = &(_ctr_attach_pts_p[ j+1 ][ i ]);

    assert(s_i->has_shift_dep(P::Pos_half_dependent(a_pos_ji)));
    s_i->remove_shift_dep(P::Pos_half_dependent(a_pos_ji));
    s_i->remove_shift_dep(P::Neg_half_dependent(a_neg_ji));
    s_i->add_shift_dep(P::Pos_dependent(a_pos_ji));
    s_i->add_shift_dep(P::Zero_dependent(a_neg_ji));
}

void Abstract_block::attach_scale_deps(uint32_t j, uint32_t i)
{
    typedef Component::Parameter P;

    assert(j < 6);
    assert(j % 2 == 0);

    P* s_i      = &(_size_p[ i ]);
    P* a_pos_ji = &(_ctr_attach_pts_p[ j+0 ][ i ]);
    P* a_neg_ji = &(_ctr_attach_pts_p[ j+1 ][ i ]);

    assert(s_i->has_scale_dep(P::Fancy_pos_half_dependent(a_pos_ji)));
    s_i->remove_scale_dep(P::Fancy_pos_half_dependent(a_pos_ji));
    s_i->remove_scale_dep(P::Fancy_neg_half_dependent(a_neg_ji));
    s_i->add_scale_dep(P::Fancy_pos_dependent(a_pos_ji));
    s_i->add_scale_dep(P::Fancy_zero_dependent(a_neg_ji));
}           

void Abstract_block::detach_shift_deps(uint32_t j, uint32_t i)
{
    typedef Component::Parameter P;

    assert(j < 6);
    assert(j % 2 == 0);

    P* s_i      = &(_size_p[ i ]);
    P* a_pos_ji = &(_ctr_attach_pts_p[ j+0 ][ i ]);
    P* a_neg_ji = &(_ctr_attach_pts_p[ j+1 ][ i ]);
    
    assert(s_i->has_shift_dep(P::Pos_dependent(a_pos_ji)));
    s_i->add_shift_dep(P::Pos_half_dependent(a_pos_ji));
    s_i->add_shift_dep(P::Neg_half_dependent(a_neg_ji));
    s_i->remove_shift_dep(P::Pos_dependent(a_pos_ji));
    s_i->remove_shift_dep(P::Zero_dependent(a_neg_ji));
}

void Abstract_block::detach_scale_deps(uint32_t j, uint32_t i)
{
    typedef Component::Parameter P;

    assert(j < 6);
    assert(j % 2 == 0);

    P* s_i      = &(_size_p[ i ]);
    P* a_pos_ji = &(_ctr_attach_pts_p[ j+0 ][ i ]);
    P* a_neg_ji = &(_ctr_attach_pts_p[ j+1 ][ i ]);

    assert(s_i->has_scale_dep(P::Fancy_pos_dependent(a_pos_ji)));
    s_i->add_scale_dep(P::Fancy_pos_half_dependent(a_pos_ji));
    s_i->add_scale_dep(P::Fancy_neg_half_dependent(a_neg_ji));
    s_i->remove_scale_dep(P::Fancy_pos_dependent(a_pos_ji));
    s_i->remove_scale_dep(P::Fancy_zero_dependent(a_neg_ji));
}

size_t Abstract_block::num_ctr_attach_pts() const
{
    return _attachments.size();
}

/**
 * @return A uniform random sample from the index set.
 *
 * @throw  jwscxx::base::Arg_error  No indices in the set.
 */
uint32_t Abstract_block::sample_attach_pt_index
(
    const std::vector<uint32_t>& indices
)
throw (jwscxx::base::Arg_error)
{
    if (indices.size() == 0)
    {
        throw jwscxx::base::Arg_error("No selectable attachment points");
    }

    uint32_t i = (uint32_t)floor(jwsc::sample_uniform_pdf_d(0, indices.size()));
    if (i == indices.size())
    {
        i--;
    }

    return indices[ i ];
}

uint32_t Abstract_block::select_random_attach_pt() const
    throw (jwscxx::base::Arg_error)
{
    uint32_t i = (uint32_t)floor(jwsc::sample_uniform_pdf_d(0, 6));
    if (i == 6)
    {
        i--;
    }

    return i;
}

uint32_t Abstract_block::select_random_used_attach_pt() const
    throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < 6; i++)
    {
        if (attachment(i) != 0)
        {
            indices.push_back(i);
        }
    }

    return Abstract_block::sample_attach_pt_index(indices);
}

uint32_t Abstract_block::select_random_mergeable_attach_pt() const
    throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t j1 = 0; j1 < 6; j1++)
    {
        if (!attachment(j1))
            continue;

        Abstract_block* b;
        if (!(b = dynamic_cast<Abstract_block*>(attachment(j1))))
        {
            continue;
        }

        indices.push_back(j1);

        uint32_t j2 = (j1 % 2 == 0) ? j1+1 : j1-1;
        for (uint32_t k = 0; k < 6; k++)
        {
            if (j1 == k || j2 == k)
                continue;

            if (attachment(k) && b->attachment(k))
            {
                indices.pop_back();
                break;
            }
        }
    }

    return Abstract_block::sample_attach_pt_index(indices);
}

/** The block must be in object coordinates. */
void Abstract_block::assert_consistency()
{
    using namespace jwsc;

    for (size_t i = 0; i < 6; i++)
    {
        Component* c = _attachments[ i ].ptr();
        if (c)
        {
            Object::assert_attachment_consistency(this, c);
        }
    }

    Vector_d* p = 0;
    Vector_d* x_pos = 0;
    Vector_d* x_neg = 0;

    posn_v(&p);

    for (size_t i = 0; i < 3; i++)
    {
        size_t j = 2*i;
        double s = size(i);

        ctr_attach_pt_v(&x_pos, j+0);
        ctr_attach_pt_v(&x_neg, j+1);

        assert(fabs((x_pos->elts[ i ] - x_neg->elts[ i ]) - s) < 1.0e-2);
        assert(fabs((x_neg->elts[ i ] + 0.5*s) - p->elts[ i ]) < 1.0e-2);
        assert(fabs((x_pos->elts[ i ] - 0.5*s) - p->elts[ i ]) < 1.0e-2);

        var_attach_pt_v(&x_pos, j+0);
        var_attach_pt_v(&x_neg, j+1);

        assert(fabs((x_pos->elts[ i ] - x_neg->elts[ i ]) - s) < 1.0e-2);
        assert(fabs((x_neg->elts[ i ] + 0.5*s) - p->elts[ i ]) < 1.0e-2);
        assert(fabs((x_pos->elts[ i ] - 0.5*s) - p->elts[ i ]) < 1.0e-2);
    }

    free_vector_d(x_pos);
    free_vector_d(x_neg);
    free_vector_d(p);

    for (size_t i = 0; i < 3; i++)
    {
        assert(fabs(posn(i) - _cuboid.get_center()->elts[ i ]) < 1.0e-2);
    }

    double s_x, s_y, s_z;
    s_x = _cuboid.get_point(1)->elts[0] - _cuboid.get_point(0)->elts[0];
    s_y = _cuboid.get_point(2)->elts[1] - _cuboid.get_point(1)->elts[1];
    s_z = _cuboid.get_point(3)->elts[2] - _cuboid.get_point(2)->elts[2];
    assert(s_x > 0 && s_y > 0 && s_z > 0);
    assert(fabs(s_x - size(0)) < 1.0e-2);
    assert(fabs(s_y - size(1)) < 1.0e-2);
    assert(fabs(s_z - size(2)) < 1.0e-2);
}

void Abstract_block::ensure_consistency()
{
    using namespace jwsc;

    Vector_d* t = 0;
    Matrix_d* M = 0;

    for (size_t i = 0; i < 3; i++)
    {
        create_zero_vector_d(&t, 3);
        t->elts[ i ] = posn(i) - _cuboid.get_center()->elts[ i ];

        create_3d_homo_translation_matrix_1_d(&M, t);
        _cuboid.transform(M);

        assert(fabs(posn(i) - _cuboid.get_center()->elts[ i ]) < 1.0e-2);
    }

    free_matrix_d(M);
    free_vector_d(t);
}

uint32_t Abstract_block::swapped_face_index(uint32_t i) const
{
    switch (i)
    {
        case 0: return 5;
        case 1: return 4;
        case 2: return 2;
        case 3: return 3;
        case 4: return 0;
        case 5: return 1;
        default: abort();
    }
}

double Abstract_block::calc_var_param_log_likelihood() const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();
    double log_prob = 0;

    for (size_t i = 0; i < 3; i++)
    {
        double x = size_p(i)->value();
        if (size_p(i)->num_conds() == 1)
        {
            double x_cond = size_p(i)->cond(0)->ptr()->value();
            log_prob += size_p(i)->context()->calc_log_likelihood(x, x_cond);
        }
        else
        {
            log_prob += size_p(i)->context()->calc_log_likelihood(x);
        }
        d++;
    }

    Vector_d* v = 0;
    for (size_t j = 0; j < 6; j++)
    {
        if (! _attachments[ j ].is_variable())
            continue;

        var_attach_pt_v(&v, j);

        if (object()) object()->transform_pt_from_world_to_object_coords(v);
        transform_pt_from_object_to_component_coords(v);
        transform_pt_from_block_to_face_coords(v, j);

        const Paramgroup& v_p = _var_attach_pts_p[ j ];
        for (size_t xyz = 0; xyz < 3; xyz++)
        {
            size_t uv = attach_pt_face_coord(j, xyz);
            if (uv == 2) continue;

            double min = -0.5 * v_p[ xyz ].cond(0)->ptr()->value();
            double max =  0.5 * v_p[ xyz ].cond(0)->ptr()->value();

            log_prob += v_p[ xyz ].context()->calc_trunc_log_likelihood(
                    v->elts[ uv ], min, max);
        }

        d += 2;
    }
    free_vector_d(v);

    assert(d <= D);

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

double Abstract_block::calc_var_param_proposal_log_likelihood() const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();
    double log_prob = 0;
    double epsilon=1.0e-8;

    for (size_t i = 0; i < 3; i++)
    {
        double x = size_p(i)->value();
        if (size_p(i)->num_conds() == 1)
        {
            double x_cond = size_p(i)->cond(0)->ptr()->value();
            log_prob += size_p(i)->context()->calc_log_likelihood(x, x_cond);
        }
        else
        {
            log_prob += size_p(i)->context()->calc_log_likelihood(x);
        }
        d++;
    }

    Vector_d* v = 0;
    for (size_t j = 0; j < 6; j++)
    {
        if (! _attachments[ j ].is_variable())
            continue;

        var_attach_pt_v(&v, j);

        if (object()) object()->transform_pt_from_world_to_object_coords(v);
        transform_pt_from_object_to_component_coords(v);
        transform_pt_from_block_to_face_coords(v, j);

        const Paramgroup& v_p = _var_attach_pts_p[ j ];
        for (size_t xyz = 0; xyz < 3; xyz++)
        {
            size_t uv = attach_pt_face_coord(j, xyz);
            if (uv == 2) continue;

            // If v is outside [min+e,max-e], it means v was truncated, we don't
            // propose that.
            double min = -0.5 * v_p[ xyz ].cond(0)->ptr()->value() + epsilon;
            double max =  0.5 * v_p[ xyz ].cond(0)->ptr()->value() - epsilon;
            assert(min <= max);

            if (v->elts[ uv ] > min && v->elts[ uv ] < max)
            {
                log_prob += v_p[ xyz ].context()->calc_log_likelihood(
                        v->elts[ uv ]);
            }
        }

        d += 2;
    }
    free_vector_d(v);

    assert(d <= D);

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

void Abstract_block::gen_var_attach_pt_conditional_likelihood_sample
(
    size_t j
) 
{
    using namespace jwsc;

    Paramgroup& v_p = _var_attach_pts_p[ j ];

    for (size_t xyz = 0; xyz < 3; xyz++)
    {
        size_t uv = attach_pt_face_coord(j, xyz);
        if (uv == 2)
            continue;

        double min = -0.5 * v_p[ xyz ].cond(0)->ptr()->value();
        double max =  0.5 * v_p[ xyz ].cond(0)->ptr()->value();

        double x = v_p[ xyz ].context()->sample_trunc_likelihood(min, max);
        v_p[ xyz ].scratch(x);
    }
}

void Abstract_block::get_var_attach_pt_conditional_likelihood_sample
(
    jwsc::Vector_d** v_out, 
    size_t           j
) 
const
{
    using namespace jwsc;

    const Paramgroup& v_p = _var_attach_pts_p[ j ];

    create_vector_d(v_out, 2);
    Vector_d* v = *v_out;

    for (size_t xyz = 0; xyz < 3; xyz++)
    {
        size_t uv = attach_pt_face_coord(j, xyz);
        if (uv == 2)
            continue;

        v->elts[ uv ] = v_p[ xyz ].scratch();
    }
}

void Abstract_block::sample_var_attach_pt_likelihood
(
    jwsc::Vector_d** v_out, 
    size_t           j
) 
const
{
    using namespace jwsc;

    const Paramgroup& v_p = _var_attach_pts_p[ j ];

    create_vector_d(v_out, 2);
    Vector_d* v = *v_out;

    for (size_t xyz = 0; xyz < 3; xyz++)
    {
        size_t uv = attach_pt_face_coord(j, xyz);
        if (uv == 2)
            continue;

        double min = -0.5 * v_p[ xyz ].cond(0)->ptr()->value();
        double max =  0.5 * v_p[ xyz ].cond(0)->ptr()->value();

        v->elts[ uv ] = v_p[ xyz ].context()->sample_trunc_likelihood(min, max);
    }
}

void Abstract_block::gen_var_param_conditional_likelihood_sample()
{
    for (size_t i = 0; i < 3; i++)
    {
        double x;
        if (size_p(i)->num_conds() == 1)
        {
            double x_cond = size_p(i)->cond(0)->ptr()->scratch();
            x = size_p(i)->context()->sample_likelihood(x_cond);
        }
        else
        {
            x = size_p(i)->context()->sample_likelihood();
        }
        _size_p[ i ].scratch(x);
    }

    for (size_t j = 0; j < 6; j++)
    {
        if (! _attachments[ j ].is_variable())
            continue;

        gen_var_attach_pt_conditional_likelihood_sample(j);
    }
}

void Abstract_block::get_var_param_conditional_likelihood_sample
(
    jwsc::Vector_d** p_out
)
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    Vector_d* v = 0;

    create_vector_d(&v, 3);
    for (size_t i = 0; i < 3; i++)
    {
        v->elts[ i ] = size_p(i)->scratch();
    }
    copy_vector_section_into_vector_d(p, d, v, 0, v->num_elts);
    d += v->num_elts;

    for (size_t j = 0; j < 6; j++)
    {
        if (! _attachments[ j ].is_variable())
            continue;

        get_var_attach_pt_conditional_likelihood_sample(&v, j);
        copy_vector_section_into_vector_d(p, d, v, 0, v->num_elts);
        d += v->num_elts;
    }

    free_vector_d(v);

    assert(d <= D);
}

void Abstract_block::sample_var_param_likelihood(jwsc::Vector_d** p_out)
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    Vector_d* v = 0;

    create_vector_d(&v, 3);
    for (size_t i = 0; i < 3; i++)
    {
        if (size_p(i)->num_conds() == 1)
        {
            double x_cond = size_p(i)->cond(0)->ptr()->value();
            v->elts[ i ] = size_p(i)->context()->sample_likelihood(x_cond);
        }
        else
        {
            v->elts[ i ] = size_p(i)->context()->sample_likelihood();
        }
    }
    copy_vector_section_into_vector_d(p, d, v, 0, v->num_elts);
    d += v->num_elts;

    for (size_t j = 0; j < 6; j++)
    {
        if (! _attachments[ j ].is_variable())
            continue;

        sample_var_attach_pt_likelihood(&v, j);
        copy_vector_section_into_vector_d(p, d, v, 0, v->num_elts);
        d += v->num_elts;
    }

    free_vector_d(v);

    assert(d <= D);
}

double Abstract_block::calc_var_param_prior_log_prob() const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();
    double log_prob = 0;

    Vector_d* means = 0;
    Vector_d* stdevs = 0;

    Vector_d* min = 0;
    Vector_d* max = 0;

    if (size_p()->num_conds())
    {
        for (size_t i = 0; i < 3; i++)
        {
            double mean = size_p(i)->mean();
            double stdev = size_p(i)->stdev();
            log_prob += size_p(i)->context()->calc_prior_log_prob(mean, stdev);
        }
    }
    else
    {
        size_p()->means(&means);
        size_p()->stdevs(&stdevs);
        log_prob += size_p()->context()->calc_prior_log_prob(means, stdevs);
    }
    d += 3;

    create_vector_d(&means, 2);
    create_vector_d(&stdevs, 2);

    create_vector_d(&min, 2);
    create_vector_d(&max, 2);

    for (size_t j = 0; j < 6; j++)
    {
        if (! _attachments[ j ].is_variable())
            continue;

        const Paramgroup& v_p = _var_attach_pts_p[ j ];
        for (size_t xyz = 0; xyz < 3; xyz++)
        {
            size_t uv = attach_pt_face_coord(j, xyz);
            if (uv == 2) continue;

            means->elts[ uv ] = v_p[ xyz ].mean();
            stdevs->elts[ uv ] = v_p[ xyz ].stdev();

            min->elts[ uv ]  = -0.5 * v_p[ xyz ].cond(0)->ptr()->mean();
            min->elts[ uv ] += -6.0 * v_p[ xyz ].cond(0)->ptr()->stdev();
            max->elts[ uv ]  =  0.5 * v_p[ xyz ].cond(0)->ptr()->mean();
            max->elts[ uv ] +=  6.0 * v_p[ xyz ].cond(0)->ptr()->stdev();
        }

        // This is a bad idea.
        //log_prob += v_p.context()->calc_trunc_prior_log_prob(means, min, max,
        //        stdevs);
        log_prob += v_p.context()->calc_prior_log_prob(means, stdevs);

        d += 2;
    }

    free_vector_d(means);
    free_vector_d(stdevs);
    free_vector_d(min);
    free_vector_d(max);

    assert(d <= D);

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

void Abstract_block::sample_var_attach_pt_prior
(
    jwsc::Vector_d** means_out, 
    jwsc::Vector_d** stdevs_out, 
    size_t           j
)
const
{
    using namespace jwsc;

    const Paramgroup& v_p = _var_attach_pts_p[ j ];

    Vector_d* min = 0;
    Vector_d* max = 0;

    create_vector_d(&min, 2);
    create_vector_d(&max, 2);

    for (size_t xyz = 0; xyz < 3; xyz++)
    {
        size_t uv = attach_pt_face_coord(j, xyz);
        if (uv == 2)
            continue;

        min->elts[ uv ]  = -0.5 * v_p[ xyz ].cond(0)->ptr()->mean();
        min->elts[ uv ] += -6.0 * v_p[ xyz ].cond(0)->ptr()->stdev();
        max->elts[ uv ]  =  0.5 * v_p[ xyz ].cond(0)->ptr()->mean();
        max->elts[ uv ] +=  6.0 * v_p[ xyz ].cond(0)->ptr()->stdev();
    }

    v_p.context()->sample_trunc_prior(means_out, min, max, stdevs_out);

    free_vector_d(min);
    free_vector_d(max);
}

void Abstract_block::sample_var_param_prior
(
    jwsc::Vector_d** means_out,
    jwsc::Vector_d** stdevs_out
) 
const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    create_vector_d(means_out, D);
    create_vector_d(stdevs_out, D);

    Vector_d* means = *means_out;
    Vector_d* stdevs = *stdevs_out;
    Vector_d* m_q = 0;
    Vector_d* s_q = 0;

    if (size_p()->num_conds())
    {
        create_vector_d(&m_q, 3);
        create_vector_d(&s_q, 3);

        for (size_t i = 0; i < 3; i++)
        {
            double mean, stdev;
            size_p(i)->context()->sample_prior(&mean, &stdev);
            m_q->elts[ i ] = mean;
            s_q->elts[ i ] = stdev;
        }
    }
    else
    {
        size_p()->context()->sample_prior(&m_q, &s_q);
    }
    copy_vector_section_into_vector_d(means, d, m_q, 0, m_q->num_elts);
    copy_vector_section_into_vector_d(stdevs, d, s_q, 0, s_q->num_elts);
    d += m_q->num_elts;

    for (size_t j = 0; j < 6; j++)
    {
        if (! _attachments[ j ].is_variable())
            continue;

        sample_var_attach_pt_prior(&m_q, &s_q, j);
        copy_vector_section_into_vector_d(means, d, m_q, 0, m_q->num_elts);
        copy_vector_section_into_vector_d(stdevs, d, s_q, 0, s_q->num_elts);
        d += m_q->num_elts;
    }

    free_vector_d(m_q);
    free_vector_d(s_q);

    assert(d <= D);
}

void Abstract_block::sample_var_param_cat_proposal
(
    jwsc::Vector_d** means_out,
    jwsc::Vector_d** stdevs_out,
    const jwsc::Vector_d* means_in
) 
const
{
    using namespace jwsc;

    Model::sample_var_param_cat_proposal(means_out, stdevs_out, means_in);

    Vector_d* means = *means_out;
    Vector_d* stdevs = *stdevs_out;

    for (size_t i = 0; i < 3; i++)
    {
        double mean, stdev;
        double mean_in = means_in->elts[ i ];
        size_p(i)->context()->sample_prior(&mean, &stdev, mean_in);
        means->elts[ i ] = mean;
        stdevs->elts[ i ] = stdev;
    }
}

double Abstract_block::calc_var_param_proposal_log_prob() const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();
    double log_prob = 0;

    for (size_t i = 0; i < 3; i++)
    {
        double x = size_p(i)->value();
        log_prob += size_p(i)->context()->calc_proposal_log_prob(x);
        d++;
    }

    Vector_d* v = 0;
    for (size_t j = 0; j < 6; j++)
    {
        if (! _attachments[ j ].is_variable())
            continue;

        var_attach_pt_v(&v, j);

        if (object()) object()->transform_pt_from_world_to_object_coords(v);
        transform_pt_from_object_to_component_coords(v);
        transform_pt_from_block_to_face_coords(v, j);

        const Paramgroup& v_p = _var_attach_pts_p[ j ];
        for (size_t xyz = 0; xyz < 3; xyz++)
        {
            size_t uv = attach_pt_face_coord(j, xyz);
            if (uv == 2) continue;

            double min = -0.5 * v_p[ xyz ].cond(0)->ptr()->value();
            double max =  0.5 * v_p[ xyz ].cond(0)->ptr()->value();

            log_prob += v_p[ xyz ].context()->calc_trunc_proposal_log_prob(
                    v->elts[ uv ], min, max);
        }
        d += 2;
    }
    free_vector_d(v);

    assert(d <= D);

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

double Abstract_block::calc_var_param_gauss_proposal_log_prob() const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();
    double log_prob = 0;

    for (size_t i = 0; i < 3; i++)
    {
        double x = size_p(i)->value();
        log_prob += size_p(i)->context()->calc_gauss_proposal_log_prob(x);
        d++;
    }

    Vector_d* v = 0;
    for (size_t j = 0; j < 6; j++)
    {
        if (! _attachments[ j ].is_variable())
            continue;

        var_attach_pt_v(&v, j);

        if (object()) object()->transform_pt_from_world_to_object_coords(v);
        transform_pt_from_object_to_component_coords(v);
        transform_pt_from_block_to_face_coords(v, j);

        const Paramgroup& v_p = _var_attach_pts_p[ j ];
        for (size_t xyz = 0; xyz < 3; xyz++)
        {
            size_t uv = attach_pt_face_coord(j, xyz);
            if (uv == 2) continue;

            double min = -0.5 * v_p[ xyz ].cond(0)->ptr()->value();
            double max =  0.5 * v_p[ xyz ].cond(0)->ptr()->value();

            log_prob += v_p[ xyz ].context()->
                calc_trunc_gauss_proposal_log_prob(v->elts[ uv ], min, max);
        }
        d += 2;
    }
    free_vector_d(v);

    assert(d <= D);

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

double Abstract_block::calc_var_param_split_merge_proposal_log_prob
(
    const jwsc::Vector_d* p,
    const jwsc::Vector_d* means,
    uint32_t              j
)
const
{
    uint32_t d = 0;
    uint32_t D = num_var_params();
    double log_prob = 0;

    assert(D == means->num_elts);
    if (! _attachments[ j ].is_variable())
    {
        D += 2;
    }
    assert(D == p->num_elts);

    for (size_t i = 0; i < 3; i++)
    {
        double m = size_p(i)->context()->gauss_proposal().mu();
        double x = p->elts[ d ] + (m - means->elts[ d ]);
        log_prob += size_p(i)->context()->calc_gauss_proposal_log_prob(x);
        d++;
    }

    for (size_t k = 0; k < 6; k++)
    {
        if (_attachments[ k ].is_variable() && k != j)
        {
            d += 2;
        }
        else if (k == j)
        {
            const Paramgroup& v_p = _var_attach_pts_p[ j ];
            for (size_t xyz = 0; xyz < 3; xyz++)
            {
                size_t uv = attach_pt_face_coord(j, xyz);
                if (uv == 2)
                    continue;

                log_prob += v_p[ xyz ].context()->
                    calc_gauss_proposal_log_prob(p->elts[ d+uv ]);
            }
            d += 2;
        }
    }

    assert(d <= D);

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

double Abstract_block::calc_var_param_switch_proposal_log_prob
(
    const jwsc::Vector_d* p
)
const
{
    return 0;
}

void Abstract_block::sample_var_attach_pt_proposal
(
    jwsc::Vector_d** v_out, 
    size_t           j
) 
const
{
    using namespace jwsc;

    const Paramgroup& v_p = _var_attach_pts_p[ j ];

    create_vector_d(v_out, 2);
    Vector_d* v = *v_out;

    for (size_t xyz = 0; xyz < 3; xyz++)
    {
        size_t uv = attach_pt_face_coord(j, xyz);
        if (uv == 2)
            continue;

        double min = -0.5 * v_p[ xyz ].cond(0)->ptr()->value();
        double max =  0.5 * v_p[ xyz ].cond(0)->ptr()->value();

        v->elts[ uv ] = v_p[ xyz ].context()->sample_trunc_proposal(min, max);
    }
}

void Abstract_block::sample_var_attach_pt_gauss_proposal
(
    jwsc::Vector_d**      v_out, 
    size_t                j,
    const jwsc::Vector_d* means
) 
const
{
    using namespace jwsc;

    assert(means->num_elts == 2);

    const Paramgroup& v_p = _var_attach_pts_p[ j ];

    create_vector_d(v_out, 2);
    Vector_d* v = *v_out;

    for (size_t xyz = 0; xyz < 3; xyz++)
    {
        size_t uv = attach_pt_face_coord(j, xyz);
        if (uv == 2)
            continue;

        double min = -0.5 * v_p[ xyz ].cond(0)->ptr()->value();
        double max =  0.5 * v_p[ xyz ].cond(0)->ptr()->value();
        double mean = means->elts[ uv ];

        v->elts[ uv ] = v_p[ xyz ].context()->
            sample_trunc_gauss_proposal(mean, min, max);
    }
}

void Abstract_block::sample_var_param_proposal(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    Vector_d* v = 0;

    create_vector_d(&v, 3);
    for (size_t i = 0; i < 3; i++)
    {
        v->elts[ i ] = size_p(i)->context()->sample_proposal();
    }
    copy_vector_section_into_vector_d(p, d, v, 0, v->num_elts);
    d += v->num_elts;

    for (size_t j = 0; j < 6; j++)
    {
        if (! _attachments[ j ].is_variable())
            continue;

        sample_var_attach_pt_proposal(&v, j);
        copy_vector_section_into_vector_d(p, d, v, 0, v->num_elts);
        d += v->num_elts;
    }

    free_vector_d(v);

    assert(d <= D);
}

void Abstract_block::sample_var_param_gauss_proposal
(
    jwsc::Vector_d**      p_out,
    const jwsc::Vector_d* means
) 
const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    assert(D == means->num_elts);

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    Vector_d* v = 0;
    Vector_d* m = 0;

    create_vector_d(&v, 3);
    for (size_t i = 0; i < 3; i++)
    {
        double mean = means->elts[ i ];
        v->elts[ i ] = size_p(i)->context()->sample_gauss_proposal(mean);
    }
    copy_vector_section_into_vector_d(p, d, v, 0, v->num_elts);
    d += v->num_elts;

    for (size_t j = 0; j < 6; j++)
    {
        if (! _attachments[ j ].is_variable())
            continue;

        copy_vector_section_d(&m, means, d, 2);
        sample_var_attach_pt_gauss_proposal(&v, j, m);
        copy_vector_section_into_vector_d(p, d, v, 0, v->num_elts);
        d += v->num_elts;
    }

    free_vector_d(m);
    free_vector_d(v);

    assert(d <= D);
}

void Abstract_block::sample_var_param_split_merge_proposal
(
    jwsc::Vector_d**      p_out,
    const jwsc::Vector_d* means,
    uint32_t              j
) 
const
{
    using namespace jwsc;

    uint32_t m_d = 0;
    uint32_t p_d = 0;
    uint32_t D = num_var_params();

    assert(D == means->num_elts);
    if (! _attachments[ j ].is_variable())
    {
        D += 2;
    }

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    Vector_d* v = 0;
    Vector_d* m = 0;

    create_vector_d(&v, 3);
    for (size_t i = 0; i < 3; i++)
    {
        double mean = means->elts[ i ];
        v->elts[ i ] = size_p(i)->context()->sample_gauss_proposal(mean);
    }
    copy_vector_section_into_vector_d(p, p_d, v, 0, v->num_elts);
    p_d += v->num_elts;
    m_d += v->num_elts;

    for (size_t k = 0; k < 6; k++)
    {
        if (_attachments[ k ].is_variable() && k != j)
        {
            copy_vector_section_into_vector_d(p, p_d, means, m_d, 2);
            p_d += 2;
            m_d += 2;
        }
        else if (_attachments[ k ].is_variable() && k == j)
        {
            create_zero_vector_d(&m, 2);
            sample_var_attach_pt_gauss_proposal(&v, j, m);
            copy_vector_section_into_vector_d(p, p_d, v, 0, 2);
            p_d += 2;
            m_d += 2;
        }
        else if (k == j)
        {
            create_zero_vector_d(&m, 2);
            sample_var_attach_pt_gauss_proposal(&v, j, m);
            copy_vector_section_into_vector_d(p, p_d, v, 0, 2);
            p_d += 2;
        }
    }

    free_vector_d(m);
    free_vector_d(v);

    assert(p_d <= D);
}

void Abstract_block::sample_var_param_switch_proposal
(
    jwsc::Vector_d**      p_out,
    const jwsc::Vector_d* means
) 
const
{
    jwsc::copy_vector_section_d(p_out, means, 0, 3);
}

void Abstract_block::translate(double x, double y, double z)
{
    using namespace jwsc;

    Component::translate(x, y, z);

    Matrix_d* M = 0;
    Vector_d* a = 0;

    assert(create_3d_homo_translation_matrix_2_d(&M, x, y, z) == 0);

    for (size_t i = 0; i < 6; i++)
    {
        ctr_attach_pt_v(&a, i);
        multiply_matrix_and_vector_d(&a, M, a);
        ctr_attach_pt_v(a, i);

        var_attach_pt_v(&a, i);
        multiply_matrix_and_vector_d(&a, M, a);
        var_attach_pt_v(a, i);
    }

    _cuboid.transform(M);

    free_vector_d(a);
    free_matrix_d(M);
}

void Abstract_block::rotate(double phi, double x, double y, double z)
{
    using namespace jwsc;

    Component::rotate(phi, x, y, z);

    Matrix_d* M = 0;
    Vector_d* a = 0;

    assert(create_3d_homo_rotation_matrix_2_d(&M, phi, x, y, z) == 0);

    for (size_t i = 0; i < 6; i++)
    {
        ctr_attach_pt_v(&a, i);
        multiply_matrix_and_vector_d(&a, M, a);
        ctr_attach_pt_v(a, i);

        var_attach_pt_v(&a, i);
        multiply_matrix_and_vector_d(&a, M, a);
        var_attach_pt_v(a, i);
    }

    _cuboid.transform(M);

    free_vector_d(a);
    free_matrix_d(M);
}

/**
 * The point must be in homogeneous coodinates.
 *
 * @param  p  Point to transform from face to block coordinates.
 * @param  i  Face in this block to transform @em p from.
 */
void Abstract_block::transform_pt_from_face_to_block_coords
(
    jwsc::Vector_d* p, 
    size_t          i
) 
const
{
    using namespace jwsc;

    assert(p->num_elts == 4);

    Matrix_d* R_faces_t = 0;
    transpose_matrix_d(&R_faces_t, _R_faces[ i ]);
    multiply_matrix_and_vector_d(&p, R_faces_t, p);
    free_matrix_d(R_faces_t);

    switch (i)
    {
        case 0: p->elts[0] += 0.5*size(0);  break;
        case 1: p->elts[0] -= 0.5*size(0);  break;
        case 2: p->elts[1] += 0.5*size(1); break;
        case 3: p->elts[1] -= 0.5*size(1); break;
        case 4: p->elts[2] += 0.5*size(2); break;
        case 5: p->elts[2] -= 0.5*size(2); break;
        default: abort();
    }
}


/**
 * The point must be in homogeneous coodinates.
 *
 * @param  p  Point to transform from block to face coordinates.
 * @param  i  Face in this block to transform @em p to.
 */
void Abstract_block::transform_pt_from_block_to_face_coords
(
    jwsc::Vector_d* p,
    size_t          i
) 
const
{
    using namespace jwsc;

    switch (i)
    {
        case 0: p->elts[0] -= 0.5*size(0);  break;
        case 1: p->elts[0] += 0.5*size(0);  break;
        case 2: p->elts[1] -= 0.5*size(1); break;
        case 3: p->elts[1] += 0.5*size(1); break;
        case 4: p->elts[2] -= 0.5*size(2); break;
        case 5: p->elts[2] += 0.5*size(2); break;
        default: abort();
    }

    multiply_matrix_and_vector_d(&p, _R_faces[ i ], p);
}

bool Abstract_block::conditions(const Model* m) const
{
    if (Component::conditions(m))
        return true;

    for (uint32_t i = 0; i < 3; i++)
    {
        const Parameter& p = _size_p[ i ];
        for (uint32_t j = 0; j < p.num_conds(); j++)
        {
            if (m == p.cond(j)->ptr()->model())
            {
                return true;
            }
        }
    }

    for (uint32_t j = 0; j < 6; j++)
    {
        for (uint32_t i = 0; i < 3; i++)
        {
            const Parameter& p = _var_attach_pts_p[ j ][ i ];
            for (uint32_t k = 0; k < p.num_conds(); k++)
            {
                if (m == p.cond(k)->ptr()->model())
                {
                    return true;
                }
            }
        }
    }

    return false;
}

void Abstract_block::serialize(const std::vector<Model*>& models)
{
    Model::serialize(models);

    for (size_t i = 0; i < 6; i++)
    {
        _attachments[ i ].serialize(models);
    }
}

void Abstract_block::deserialize(const std::vector<Model*>& models)
{
    Model::deserialize(models);

    for (size_t i = 0; i < 6; i++)
    {
        _attachments[ i ].deserialize(models);
    }
}

void Abstract_block::update_pointers
(
    const std::vector<Model*>&       new_models,
    const std::vector<const Model*>& old_models
)
{
    Model::update_pointers(new_models, old_models);

    for (size_t i = 0; i < 6; i++)
    {
        _attachments[ i ].update_pointers(new_models, old_models);
    }
}

std::vector<const jwscxx::graphics::Polygon_d*> Abstract_block::faces() const
{
    std::vector<const jwscxx::graphics::Polygon_d*> faces(_cuboid.get_faces());

    return faces;
}

size_t Abstract_block::num_faces() const
{
    return _cuboid.num_faces();
}

uint32_t Abstract_block::set_sequential_face_ids(uint32_t id)
{
    try
    {
        for (size_t i = 0; i < _cuboid.num_faces(); i++)
        {
            _cuboid.set_face_id(i, id++);
        }
    }
    catch (jwscxx::base::Arg_error e)
    {
        // Should never happen.
        abort();
    }

    return id;
}

bool Abstract_block::graphical_rep_is_outdated() const
{
    return Model::graphical_rep_is_outdated();
}

void Abstract_block::graphical_rep_outdated()
{
    Model::graphical_rep_outdated();
}

/** The block must be in object coordinates. */
void Abstract_block::update_graphical_rep()
{
    using namespace jwsc;

    Model::update_graphical_rep();

    double t_x, t_y, t_z;
    double s_x, s_y, s_z;

    Matrix_d* M = 0;
    Matrix_d* T = 0;
    Matrix_d* S = 0;

    create_identity_matrix_d(&M, 4);

    t_x = _cuboid.get_center()->elts[0];
    t_y = _cuboid.get_center()->elts[1];
    t_z = _cuboid.get_center()->elts[2];

    create_3d_homo_translation_matrix_2_d(&T, -t_x, -t_y, -t_z);
    multiply_matrices_d(&M, T, M);

    _cuboid.transform(M);

    s_x = _cuboid.get_point(1)->elts[0] - _cuboid.get_point(0)->elts[0];
    s_y = _cuboid.get_point(2)->elts[1] - _cuboid.get_point(1)->elts[1];
    s_z = _cuboid.get_point(3)->elts[2] - _cuboid.get_point(2)->elts[2];
    assert(s_x > 0 && s_y > 0 && s_z > 0);

    s_x = size(0) / s_x;
    s_y = size(1) / s_y;
    s_z = size(2) / s_z;

    create_identity_matrix_d(&M, 4);
    create_3d_homo_scaling_matrix_2_d(&S, s_x, s_y, s_z);
    multiply_matrices_d(&M, S, M);

    t_x = posn(0);
    t_y = posn(1);
    t_z = posn(2);

    create_3d_homo_translation_matrix_2_d(&T, t_x, t_y, t_z);
    multiply_matrices_d(&M, T, M);

    _cuboid.transform(M);

    free_matrix_d(S);
    free_matrix_d(T);
    free_matrix_d(M);
}

void Abstract_block::wire_render_with_ids(const View* view) const
{
    wire_render_mesh_faces_with_ids(view, &_cuboid);
}

void Abstract_block::solid_render() const
{
    solid_render_mesh_faces(&_cuboid);

    // BEGIN DEBUG
    /*
    using namespace jwsc;

    for (size_t i = 0; i < 6; i++)
    {
        Vector_d* a = 0;
        ctr_attach_pt_v(&a, i);

        const Vector_d* n = _cuboid.get_face(i)->get_normal();

        double x1 = a->elts[0];
        double y1 = a->elts[1];
        double z1 = a->elts[2];
        double x2 = a->elts[0] + n->elts[0]*20;
        double y2 = a->elts[1] + n->elts[1]*20;
        double z2 = a->elts[2] + n->elts[2]*20;

        glBegin(GL_LINES);
        glVertex3d(x1, y1, z1);
        glVertex3d(x2, y2, z2);
        glEnd();
    }
    */
    // END DEBUG
}

void Abstract_block::create_face_rotations()
{
    using namespace jwsc;

    _R_faces.reserve(6);

    Matrix_d* r = 0;
    create_zero_matrix_d(&r, 4, 4);
    r->elts[0][0] =  0; r->elts[0][1] =  0; r->elts[0][2] = -1;
    r->elts[1][0] =  0; r->elts[1][1] =  1; r->elts[1][2] =  0;
    r->elts[2][0] =  1; r->elts[2][1] =  0; r->elts[2][2] =  0;
    r->elts[3][3] =  1;
    _R_faces.push_back(r);

    r = 0;
    create_zero_matrix_d(&r, 4, 4);
    r->elts[0][0] =  0; r->elts[0][1] =  0; r->elts[0][2] =  1;
    r->elts[1][0] =  0; r->elts[1][1] =  1; r->elts[1][2] =  0;
    r->elts[2][0] = -1; r->elts[2][1] =  0; r->elts[2][2] =  0;
    r->elts[3][3] =  1;
    _R_faces.push_back(r);

    r = 0;
    create_zero_matrix_d(&r, 4, 4);
    r->elts[0][0] =  1; r->elts[0][1] =  0; r->elts[0][2] =  0;
    r->elts[1][0] =  0; r->elts[1][1] =  0; r->elts[1][2] = -1;
    r->elts[2][0] =  0; r->elts[2][1] =  1; r->elts[2][2] =  0;
    r->elts[3][3] =  1;
    _R_faces.push_back(r);

    r = 0;
    create_zero_matrix_d(&r, 4, 4);
    r->elts[0][0] =  1; r->elts[0][1] =  0; r->elts[0][2] =  0;
    r->elts[1][0] =  0; r->elts[1][1] =  0; r->elts[1][2] =  1;
    r->elts[2][0] =  0; r->elts[2][1] = -1; r->elts[2][2] =  0;
    r->elts[3][3] =  1;
    _R_faces.push_back(r);

    r = 0;
    create_zero_matrix_d(&r, 4, 4);
    r->elts[0][0] =  1; r->elts[0][1] =  0; r->elts[0][2] =  0;
    r->elts[1][0] =  0; r->elts[1][1] =  1; r->elts[1][2] =  0;
    r->elts[2][0] =  0; r->elts[2][1] =  0; r->elts[2][2] =  1;
    r->elts[3][3] =  1;
    _R_faces.push_back(r);

    r = 0;
    create_zero_matrix_d(&r, 4, 4);
    r->elts[0][0] = -1; r->elts[0][1] =  0; r->elts[0][2] =  0;
    r->elts[1][0] =  0; r->elts[1][1] =  1; r->elts[1][2] =  0;
    r->elts[2][0] =  0; r->elts[2][1] =  0; r->elts[2][2] = -1;
    r->elts[3][3] =  1;
    _R_faces.push_back(r);
}

void Abstract_block::free_face_rotations()
{
    using namespace jwsc;

    for (size_t i = 0; i < 6; i++)
    {
        free_matrix_d(_R_faces[ i ]);
        _R_faces[ i ] = 0;
    }
}

void Abstract_block::create_ctr_attach_pts(size_t* id_in_out)
{
    using namespace jwsc;

    static const char* snames[3][6] = {
    {"Center Attachment W+ X", "Center Attachment W- X",
     "Center Attachment H+ X", "Center Attachment H- X",
     "Center Attachment L+ X", "Center Attachment L- X"},
    {"Center Attachment W+ Y", "Center Attachment W- Y",
     "Center Attachment H+ Y", "Center Attachment H- Y",
     "Center Attachment L+ Y", "Center Attachment L- Y"},
    {"Center Attachment W+ Z", "Center Attachment W- Z",
     "Center Attachment H+ Z", "Center Attachment H- Z",
     "Center Attachment L+ Z", "Center Attachment L- Z"}
    };

    Vector_d* p = 0;
    Matrix_d* M = 0;

    _ctr_attach_pts_p.reserve(6);

    size_t id = *id_in_out;

    for (size_t j = 0; j < 6; j++)
    {
        create_zero_vector_d(&p, 4);
        p->elts[3] = 1;

        multiply_matrix_and_vector_d(&p, _R_faces[ j ], p);

        switch (j)
        {
            case 0: p->elts[0] += 0.5*size(0); break;
            case 1: p->elts[0] -= 0.5*size(0); break;
            case 2: p->elts[1] += 0.5*size(1); break;
            case 3: p->elts[1] -= 0.5*size(1); break;
            case 4: p->elts[2] += 0.5*size(2); break;
            case 5: p->elts[2] -= 0.5*size(2); break;
            default: abort();
        }

        p->elts[0] += posn(0);
        p->elts[1] += posn(1);
        p->elts[2] += posn(2);

        Paramgroup grp(0);
        grp.add_param(Parameter(id++, snames[0][ j ], p->elts[0], 0, this));
        grp.add_param(Parameter(id++, snames[1][ j ], p->elts[1], 0, this));
        grp.add_param(Parameter(id++, snames[2][ j ], p->elts[2], 0, this));
        _ctr_attach_pts_p.push_back(grp);

        add_derived_param(&(_ctr_attach_pts_p[ j ][0]));
        add_derived_param(&(_ctr_attach_pts_p[ j ][1]));
        add_derived_param(&(_ctr_attach_pts_p[ j ][2]));
    }

    free_matrix_d(M);
    free_vector_d(p);

    *id_in_out = id;
}

void Abstract_block::create_var_attach_pts
(
    size_t*                        id_in_out, 
    const Abstract_block::Context* ctx
)
{
    using namespace jwsc;

    static const char* snames[3][6] = {
    {"Variable Attachment W+ X", "Variable Attachment W- X",
     "Variable Attachment H+ X", "Variable Attachment H- X",
     "Variable Attachment L+ X", "Variable Attachment L- X"},
    {"Variable Attachment W+ Y", "Variable Attachment W- Y",
     "Variable Attachment H+ Y", "Variable Attachment H- Y",
     "Variable Attachment L+ Y", "Variable Attachment L- Y"},
    {"Variable Attachment W+ Z", "Variable Attachment W- Z",
     "Variable Attachment H+ Z", "Variable Attachment H- Z",
     "Variable Attachment L+ Z", "Variable Attachment L- Z"}
    };

    Vector_d* p = 0;
    Matrix_d* M = 0;

    _var_attach_pts_p.reserve(6);

    size_t id = *id_in_out;

    for (size_t j = 0; j < 6; j++)
    {
        create_zero_vector_d(&p, 4);
        p->elts[3] = 1;

        multiply_matrix_and_vector_d(&p, _R_faces[ j ], p);

        switch (j)
        {
            case 0: p->elts[0] += 0.5*size(0); break;
            case 1: p->elts[0] -= 0.5*size(0); break;
            case 2: p->elts[1] += 0.5*size(1); break;
            case 3: p->elts[1] -= 0.5*size(1); break;
            case 4: p->elts[2] += 0.5*size(2); break;
            case 5: p->elts[2] -= 0.5*size(2); break;
            default: abort();
        }

        p->elts[0] += posn(0);
        p->elts[1] += posn(1);
        p->elts[2] += posn(2);

        Paramgroup grp((ctx) ? &(ctx->var_attach_pts_c()) : 0);
        const Model::Parameter::Context* ctx_p;
        for (size_t i = 0; i < 3; i++)
        {
            size_t uvw = attach_pt_face_coord(j, i);
            ctx_p = (ctx && uvw < 2) ? &(ctx->var_attach_pt_c(uvw)) : 0;

            grp.add_param(Parameter(id++, snames[ i ][ j ], p->elts[ i ], 
                        ctx_p, this));
        }
        _var_attach_pts_p.push_back(grp);

        add_derived_param(&(_var_attach_pts_p[ j ][0]));
        add_derived_param(&(_var_attach_pts_p[ j ][1]));
        add_derived_param(&(_var_attach_pts_p[ j ][2]));
    }

    free_matrix_d(M);
    free_vector_d(p);

    *id_in_out = id;
}

void Abstract_block::add_ctr_attach_shift_deps()
{
    typedef Parameter P;

    for (size_t i = 0; i < 3; i++)
    {
        size_t j = i*2;

        P* s       = &(_size_p[ i ]);
        P* p       = &(_posn_p[ i ]);
        P* a_pos_j = &(_ctr_attach_pts_p[ j+0 ][ i ]);
        P* a_neg_j = &(_ctr_attach_pts_p[ j+1 ][ i ]);

        s->add_shift_dep(P::Pos_half_dependent(a_pos_j));
        s->add_shift_dep(P::Neg_half_dependent(a_neg_j));

        a_pos_j->add_shift_dep(P::Pos_half_dependent(p));
        a_neg_j->add_shift_dep(P::Pos_half_dependent(p));

        a_pos_j->add_shift_dep(P::Zero_dependent(a_neg_j));
        a_neg_j->add_shift_dep(P::Zero_dependent(a_pos_j));

        for (size_t k = 0; k < 6; k += 2)
        {
            if (k == j) continue;

            P* a_pos_k = &(_ctr_attach_pts_p[ k+0 ][ i ]);
            P* a_neg_k = &(_ctr_attach_pts_p[ k+1 ][ i ]);

            a_pos_j->add_shift_dep(P::Pos_half_dependent(a_pos_k));
            a_pos_j->add_shift_dep(P::Pos_half_dependent(a_neg_k));
            a_neg_j->add_shift_dep(P::Pos_half_dependent(a_pos_k));
            a_neg_j->add_shift_dep(P::Pos_half_dependent(a_neg_k));
        }
    }
}

void Abstract_block::add_ctr_attach_scale_deps()
{
    typedef Parameter P;

    for (size_t i = 0; i < 3; i++)
    {
        size_t j = i*2;

        P* s       = &(_size_p[ i ]);
        P* p       = &(_posn_p[ i ]);
        P* a_pos_j = &(_ctr_attach_pts_p[ j+0 ][ i ]);
        P* a_neg_j = &(_ctr_attach_pts_p[ j+1 ][ i ]);

        a_pos_j->add_scale_dep(P::Pos_dependent(s));
        a_neg_j->add_scale_dep(P::Neg_dependent(s));

        s->add_scale_dep(P::Fancy_pos_half_dependent(a_pos_j));
        s->add_scale_dep(P::Fancy_neg_half_dependent(a_neg_j));

        a_pos_j->add_scale_dep(P::Pos_half_dependent(p));
        a_neg_j->add_scale_dep(P::Pos_half_dependent(p));

        a_pos_j->add_scale_dep(P::Zero_dependent(a_neg_j));
        a_neg_j->add_scale_dep(P::Zero_dependent(a_pos_j));

        for (size_t k = 0; k < 6; k += 2)
        {
            if (k == j) continue;

            P* a_pos_k = &(_ctr_attach_pts_p[ k+0 ][ i ]);
            P* a_neg_k = &(_ctr_attach_pts_p[ k+1 ][ i ]);

            a_pos_j->add_scale_dep(P::Pos_half_dependent(a_pos_k));
            a_pos_j->add_scale_dep(P::Pos_half_dependent(a_neg_k));
            a_neg_j->add_scale_dep(P::Pos_half_dependent(a_pos_k));
            a_neg_j->add_scale_dep(P::Pos_half_dependent(a_neg_k));
        }
    }
}

void Abstract_block::add_var_attach_shift_deps()
{
    typedef Parameter P;

    for (size_t j = 0; j < 6; j++)
    {
        for (size_t i = 0; i < 3; i++)
        {
            P* p     = &(_posn_p[ i ]);
            P* va_ji = &(_var_attach_pts_p[ j ][ i ]);
            P* ca_ji = &(_ctr_attach_pts_p[ j ][ i ]);

            ca_ji->add_shift_dep(P::Pos_dependent(va_ji));
            va_ji->add_shift_dep(P::Zero_dependent(p));

            for (size_t k = 0; k < 6; k += 2)
            {
                P* ca_pos_ki = &(_ctr_attach_pts_p[ k+0 ][ i ]);
                P* ca_neg_ki = &(_ctr_attach_pts_p[ k+1 ][ i ]);

                va_ji->add_shift_dep(P::Zero_dependent(ca_pos_ki));
                va_ji->add_shift_dep(P::Zero_dependent(ca_neg_ki));
            }
        }
    }
}

void Abstract_block::add_var_attach_scale_deps()
{
    typedef Parameter P;

    for (size_t j = 0; j < 6; j++)
    {
        for (size_t i = 0; i < 3; i++)
        {
            P* p     = &(_posn_p[ i ]);
            P* va_ji = &(_var_attach_pts_p[ j ][ i ]);
            P* ca_ji = &(_ctr_attach_pts_p[ j ][ i ]);

            if (i == j/2)
            {
                ca_ji->add_scale_dep(P::Pos_dependent(va_ji));
            }

            va_ji->add_scale_dep(P::Zero_dependent(p));

            for (size_t k = 0; k < 6; k += 2)
            {
                P* ca_pos_ki = &(_ctr_attach_pts_p[ k+0 ][ i ]);
                P* ca_neg_ki = &(_ctr_attach_pts_p[ k+1 ][ i ]);

                va_ji->add_scale_dep(P::Zero_dependent(ca_pos_ki));
                va_ji->add_scale_dep(P::Zero_dependent(ca_neg_ki));
            }
        }
    }
}

void Abstract_block::add_var_attach_conds()
{
    typedef Parameter P;

    for (size_t j = 0; j < 6; j++)
    {
        for (size_t i = 0; i < 2; i++)
        {
            size_t uv = attach_pt_face_coord(j, i);

            P* v = &(_var_attach_pts_p[ j ][ uv ]);
            P* s = &(_size_p[ uv ]);

            v->add_cond(P::Pointer(s));
        }
    }
}

void Abstract_block::enable_var_attach_pts()
{
    for (size_t j = 0; j < 6; j++)
    {
        if (_attachments[ j ].is_variable())
        {
            enable_var_attach_pt(j);
        }
    }
}

bool Abstract_block::has_var_attach_pt(size_t i) const
{
    return _attachments[ i ].is_variable();
}

size_t Abstract_block::attach_pt_face_coord(size_t face_j, size_t coord_i)
{
    assert(face_j < 6);
    assert(coord_i < 3);

    size_t x = 0;
    size_t y = 1;
    size_t z = 2;
    size_t u, v, w;

    switch (face_j)
    {
        case 0: case 1: u = z; v = y; w = x; break;
        case 2: case 3: u = x; v = z; w = y; break;
        case 4: case 5: u = x; v = y; w = z; break;
        default: abort();
    }

    switch (coord_i)
    {
        case 0: return u;
        case 1: return v;
        case 2: return w;
        default: abort();
    }
}

void Abstract_block::enable_var_attach_pt(size_t j)
{
    assert(j < 6);

    size_t x = 0;
    size_t y = 1;
    size_t z = 2;

    size_t u = attach_pt_face_coord(j, x);
    size_t v = attach_pt_face_coord(j, y);
    size_t w = attach_pt_face_coord(j, z);

    remove_derived_param(&(_var_attach_pts_p[ j ][ x ]));
    remove_derived_param(&(_var_attach_pts_p[ j ][ y ]));
    remove_derived_param(&(_var_attach_pts_p[ j ][ z ]));

    Parameter& u_p = _var_attach_pts_p[ j ][ u ];
    Parameter& v_p = _var_attach_pts_p[ j ][ v ];
    Parameter& w_p = _var_attach_pts_p[ j ][ w ];

    add_var_param(&u_p);
    add_var_param(&v_p);
    add_derived_param(&w_p);

    _attachments[ j ].variable(true);
}

void Abstract_block::disable_var_attach_pt(size_t j)
{
    using namespace jwsc;

    assert(j < 6);

    size_t x = 0;
    size_t y = 1;
    size_t z = 2;

    size_t u = attach_pt_face_coord(j, 0);
    size_t v = attach_pt_face_coord(j, 1);
    size_t w = attach_pt_face_coord(j, 2);

    remove_var_param(&(_var_attach_pts_p[ j ][ u ]));
    remove_var_param(&(_var_attach_pts_p[ j ][ v ]));
    remove_derived_param(&(_var_attach_pts_p[ j ][ w ]));

    Parameter& x_p = _var_attach_pts_p[ j ][ x ];
    Parameter& y_p = _var_attach_pts_p[ j ][ y ];
    Parameter& z_p = _var_attach_pts_p[ j ][ z ];

    add_derived_param(&x_p);
    add_derived_param(&y_p);
    add_derived_param(&z_p);

    _attachments[ j ].variable(false);
}

bool Abstract_block::is_mergeable(uint32_t i) const
{
    Abstract_block* b;
    if (!(b = dynamic_cast<Abstract_block*>(attachment(i))))
    {
        return false;
    }

    uint32_t j = (i % 2 == 0) ? i+1 : i-1;
    for (uint32_t k = 0; k < 6; k++)
    {
        if (i == k || j == k)
            continue;

        if (attachment(k) && b->attachment(k))
        {
            return false;
        }
    }

    return true;
}

/* ========================================================================== */







/* ----------------------------  Block::Context  ---------------------------- */

Block::Context::Context() : Abstract_block::Context()
{
}

Block::Context::Context(const Block::Context& ctx)
    : Abstract_block::Context(ctx)
{
    *this = ctx;
}

Block::Context& Block::Context::operator=(const Block::Context& ctx)
{
    Abstract_block::Context::operator=(ctx);
    return *this;
}

Block::Context* Block::Context::clone() const
{
    return new Block::Context(*this);
}

/* ========================================================================== */








/* ------------------------------  Block  ----------------------------------- */

/**
 * The block is centered at (0, 0, 0).
 *
 * @param  ctx  Parameter context for this block.
 *
 * @throw  jwscxx::base::Arg_error  Dimensions must be > 0.
 */
Block::Block(const Block::Context* ctx) throw (jwscxx::base::Arg_error)
    : Abstract_block("Block", ctx)
{
}

/**
 * The block is centered at (0, 0, 0).
 *
 * @param  w    Width.
 * @param  h    Height.
 * @param  l    Length.
 * @param  ctx  Parameter context for this block.
 *
 * @throw  jwscxx::base::Arg_error  Dimensions must be > 0.
 */
Block::Block(double w, double h, double l, const Block::Context* ctx) 
    throw (jwscxx::base::Arg_error)
    : Abstract_block("Block", w, h, l, 0, 0, 0, ctx)
{
}

/** 
 * Does not re-initialize the dependents for this (new) block; the parameter
 * dependency pointers from @em b are copied over.
 *
 * @param  b  Block to copy into this one.
 */
Block::Block(const Block& b) : Abstract_block(b)
{
    *this = b;
}

Block::~Block()
{
}

/** 
 * @param  b  Block to copy into this one.
 *
 * @return A reference to this block.
 */
Block& Block::operator=(const Block& b)
{
    Abstract_block::operator=(b);
    return *this;
}

/** @return A new copy of this block. */
Block* Block::clone() const
{
    return new Block(*this);
}

Block* Block::clone_and_swap() const throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    Block* b = new Block(size(2), size(1), size(0), context());
    b->copy_var_param_stats(this);
    b->swap_var_attach_pts(this);

    return b;
}

const Block::Context* Block::context() const
{
    return dynamic_cast<const Block::Context*>(Abstract_block::context());
}

uint32_t Block::attachable_block
(
    const std::vector<Component*>& components,
    bool                           height_only
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Block* b = dynamic_cast<Block*>(components[ i ]);
        if (b && height_only && 
                (b->attachment(2) == 0 || b->attachment(3) == 0))
        {
            indices.push_back(i);
        }
        else if (b && !height_only &&
                (b->num_ctr_attach_pts() - b->num_attachments()) > 0)
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Block::attachable_block_attach_pt
(
    const Block* b,
    bool         height_only
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t j = 0; j < 6; j++)
    {
        if (height_only && !(j == 2 || j == 3))
            continue;

        if (! b->attachment(j))
        {
            indices.push_back(j);
        }
    }

    return Abstract_block::sample_attach_pt_index(indices);
}

uint32_t Block::detachable_block
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (size_t i = 0; i < components.size(); i++)
    {
        Block* b = dynamic_cast<Block*>(components[ i ]);
        if (b && b->num_attachments() <= 1)
        {
            for (uint32_t j = 0; j < 6; j++)
            {
                if (dynamic_cast<Block*>(b->attachment(j)))
                {
                    indices.push_back(i);
                    break;
                }
            }
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Block::splitable_block_to_block
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        if (dynamic_cast<Block*>(components[ i ]))
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}
uint32_t Block::mergeable_block_from_block
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (size_t i = 0; i < components.size(); i++)
    {
        Block* b1 = dynamic_cast<Block*>(components[ i ]);
        if (b1)
        {
            for (uint32_t j = 0; j < 6; j++)
            {
                Block* b2 = dynamic_cast<Block*>(b1->attachment(j));
                if (b2 && b1->is_mergeable(j))
                {
                    indices.push_back(i);
                }
            }
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Block::mergeable_block_from_block_attach_pt(const Block* b1)
    throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t j = 0; j < 6; j++)
    {
        Block* b2 = dynamic_cast<Block*>(b1->attachment(j));
        if (b2 && b1->is_mergeable(j))
        {
            indices.push_back(j);
        }
    }

    return Abstract_block::sample_attach_pt_index(indices);
}

/* ========================================================================== */
