/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Basic block structure represented as a cuboid.
 */


#ifndef BLOCK_H
#define BLOCK_H


#include <config.h>

#include <inttypes.h>

#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/graphics/cuboid.h>
#include <jwsc++/graphics/polygon.h>

#include "model.h"
#include "component.h"


/** 
 * @class Abstract_block
 *
 * @brief Cuboid defined by a width, height and length at the origin or a
 * specified postion.
 *
 * The geometric representation of a block is positioned in Object
 * coordinates. It must be added to the Object to bring it into world
 * coordinates.
 */
class Abstract_block : public Component
{
    public:


/**
 * @class Abstract_block::Context
 *
 * @brief Context for block model parameters.
 */
class Context : public Component::Context
{
    public:

        /** @brief Constructs a new default Abstract_block::Context. */
        Context();

        /** @brief Constructs a new Abstract_block::Context from another. */
        Context(const Abstract_block::Context& ctx);

        /** @brief Deletes this block::Context. */
        virtual ~Context();

        /** @brief Copies a Abstract_block::Context into this one. */
        Abstract_block::Context& operator= (const Abstract_block::Context& ctx);

        /** @brief Clones this Context. */
        virtual Abstract_block::Context* clone() const = 0;

        const Component::Paramgroup::Context& size_c() const;

        void size_c(const Component::Paramgroup::Context& size);

        const Model::Parameter::Context& size_c(size_t i) const;

        void size_c(size_t i, const Model::Parameter::Context& size);

        const Component::Paramgroup::Context& var_attach_pts_c() const;

        void var_attach_pts_c(const Component::Paramgroup::Context& p);

        const Model::Parameter::Context& var_attach_pt_c(size_t i) const;

        void var_attach_pt_c(size_t i, const Model::Parameter::Context& p);


    protected:

        /** @brief Size of this block. */
        Component::Paramgroup::Context _size_c;

        /** @brief Variable block face attachment points. */
        Component::Paramgroup::Context _var_attach_pts_c;
};


    // Abstract_block
    public:

        /**
         * @name Standard
         *
         * Standard class members.
         * @{
         */

        /** @brief Constructs a default block at the origin. */
        Abstract_block
        (
            const char*                    name, 
            const Abstract_block::Context* ctx
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Constructs a specific block. */
        Abstract_block
        (
            const char* name,
            double      width, 
            double      height, 
            double      length, 
            double      x, 
            double      y, 
            double      z,
            const Abstract_block::Context* ctx
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Constructs a block by copying another. */
        Abstract_block(const Abstract_block& b);

        /** @brief Deletes this block. */
        virtual ~Abstract_block();

        /** @brief Copies a block into this one. */
        Abstract_block& operator= (const Abstract_block& b);

        /** @brief Clones this block. */
        virtual Abstract_block* clone() const = 0;

        /** @brief Creates a special copy of this block. */
        virtual Abstract_block* clone_and_swap() const 
            throw (jwscxx::base::Arg_error) = 0;

        /** @brief Reads this block from an input stream. */
        virtual void read(std::istream& in) throw (jwscxx::base::IO_error,
                jwscxx::base::Arg_error);

        /** @brief Writes this block to an output stream. */
        virtual void write(std::ostream& out) const 
            throw (jwscxx::base::IO_error);

        /** @brief Returns the model context. */
        virtual const Abstract_block::Context* context() const;

        /*@}*/


        /**
         * @name Variable Parameters
         *
         * Variable model parameter members.
         * @{
         */

        /** @brief Returns an instance parameter value. */
        virtual double get_var_param_value(uint32_t d) const;

        /** @brief Gets from the instance parameter values. */
        virtual void get_var_param_values(jwsc::Vector_d** p_out) const;

        /** @brief Adds to the instance parameter values. */
        virtual void add_var_param_values_shift_deps(const jwsc::Vector_d* p)
            throw (jwscxx::base::Arg_error);

        /** @brief Adds to the instance parameter values. */
        virtual void add_var_param_values_scale_deps(const jwsc::Vector_d* p)
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Sets the instance parameter values for a variable attachment
         * point.
         */
        void set_var_attach_pt_values
        (
            const jwsc::Vector_d* uv, 
            size_t                j
        )
        throw (jwscxx::base::Arg_error);

        /** 
         * @brief Sets the category parameter statistics over a variable
         * attachment point.
         */
        void set_var_attach_pt_stats
        (
            const jwsc::Vector_d* means,
            const jwsc::Vector_d* stdevs,
            size_t                j
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Prints variable parameter values. */
        virtual void print_var_param_values
        (
            uint32_t      iteration, 
            std::ostream& out
        )
        const;

        /*@}*/


        /**
         * @name posn
         *
         * Position parameters members.
         * @{
         */

        /** @brief Minimum y coordinate of this block. */
        double world_coords_min_y() const;

        /*@}*/


        /**
         * @name size
         *
         * Size parameter members.
         * @{
         */

        /** @brief Returns the size of this block. */
        const Component::Paramgroup* size_p() const;

        /** @brief Returns the size of this block. */
        const Component::Parameter* size_p(size_t i) const;

        /** @brief Returns the size of this block. */
        void size_v(jwsc::Vector_d** s_out) const;

        /** @brief Returns the size of this block. */
        double size(size_t i) const;

        /** @brief Sets the size of this block and shifts its dependents. */
        void size_shift_deps(size_t i, double size);

        /** @brief Sets the size of this block and scales its dependents. */
        void size_scale_deps(size_t i, double size);

        /** @brief Tests a size for this block and its shifted dependents. */
        void test_size_shift_deps(size_t i, double size)
            throw (jwscxx::base::Arg_error);

        /** @brief Tests a size for this block and its scaled dependents. */
        void test_size_scale_deps(size_t i, double size)
            throw (jwscxx::base::Arg_error);

        /** @brief Scales the size parameter stats. */
        void scale_size_stats
        (
            Abstract_block* b, 
            size_t          j
        );

        /*@}*/


        /**
         * @name attach_pt
         *
         * Face attachment points parameter members.
         * @{
         */

        /** @brief Gets a face center attachment point. */
        void ctr_attach_pt_v(jwsc::Vector_d** a_out, size_t i) const;

        /** @brief Returns the variable attachment point group of parameters. */
        const Component::Paramgroup* var_attach_pt_p(size_t i) const;

        /** 
         * @brief Sets the position of a variable attachment point and shifts
         * its dependents.
         */
        void var_attach_pt_shift_deps(size_t j, size_t i, double uv);

        /** 
         * @brief Test a position of a variable attachment point and shifts its
         * dependents.
         */
        void test_var_attach_pt_shift_deps(size_t j, size_t i, double uv)
            throw (jwscxx::base::Arg_error);

        /** @brief Gets a face variable attachment point. */
        void var_attach_pt_v(jwsc::Vector_d** a_out, size_t i) const;

        /** @brief Returns whether a variable attachment is enabled on face. */
        bool has_var_attach_pt(size_t i) const;

        /** 
         * @brief Enables a variable attachment point by placing it in the
         * appropriate parameter sets.
         */
        void enable_var_attach_pt(size_t j);

        /** 
         * @brief Disables a variable attachment point by removing it in the
         * appropriate parameter sets.
         */
        void disable_var_attach_pt(size_t j);

        /*@}*/


        /**
         * @name Attachment
         *
         * Component attachment members.
         * @{
         */

        /** @brief Returns the number of Components attached to this is one. */
        size_t num_attachments() const;

        /** 
         * @brief Returns the number of Components center-attached to this is
         * one.
         */
        size_t num_ctr_attachments() const;

        /** 
         * @brief Returns the number of Components variably-attached to this is
         * one.
         */
        size_t num_var_attachments() const;

        /** @brief Returns the index of an attachment. */
        size_t attachment(const Component* c) const 
            throw (jwscxx::base::Arg_error);

        /** @brief Returns the attachment at the specifed index. */
        Component* attachment(size_t i) const;

        /** @brief Sets the attachment at the specified index. */
        void attachment(size_t i, Component* c);

        /** 
         * @brief Returns the number of center attachment points on this
         * Component.
         */
        size_t num_ctr_attach_pts() const;

        static uint32_t sample_attach_pt_index
        (
            const std::vector<uint32_t>& indices
        )
        throw (jwscxx::base::Arg_error);

        uint32_t select_random_attach_pt() const
            throw (jwscxx::base::Arg_error);

        uint32_t select_random_used_attach_pt() const
            throw (jwscxx::base::Arg_error);

        uint32_t select_random_mergeable_attach_pt() const
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Asserts consitency among attachments and the graphical
         * representation.
         */
        virtual void assert_consistency();

        /** 
         * @brief Ensures consitency among attachments and the graphical
         * representation.
         */
        virtual void ensure_consistency();

        uint32_t swapped_face_index(uint32_t i) const;

        virtual bool is_mergeable(uint32_t i) const;

        /*@}*/


        /**
         * @name Likelihood
         *
         * Instance parameter likelihood members.
         * @{
         */

        /** @brief Calculates the log likelihood of the instance parameters. */
        virtual double calc_var_param_log_likelihood() const;

        /** @brief Calculates the log likelihood of the instance parameters. */
        virtual double calc_var_param_proposal_log_likelihood() const;

        /** 
         * @brief Generates variable attachment point instance parameter values
         * from the likelihood.
         */
        void sample_var_attach_pt_likelihood
        (
            jwsc::Vector_d** v_out,
            size_t           j
        ) 
        const;

        /** 
         * @brief Generates instance parameter values from the likelihood.
         */
        virtual void sample_var_param_likelihood(jwsc::Vector_d** p_out);

        void gen_var_attach_pt_conditional_likelihood_sample(size_t j);

        virtual void gen_var_param_conditional_likelihood_sample();

        void get_var_attach_pt_conditional_likelihood_sample
        (
            jwsc::Vector_d** v_out,
            size_t           j
        ) 
        const;

        virtual void get_var_param_conditional_likelihood_sample
        (
            jwsc::Vector_d** p_out
        );

        /*@}*/


        /**
         * @name Prior
         *
         * Category parameter prior probability members.
         * @{
         */

        /** @brief Calculates the log probability of this block. */
        virtual double calc_var_param_prior_log_prob() const;

        /** 
         * @brief Generates variable attachment point category parameter values
         * from the prior.
         */
        void sample_var_attach_pt_prior
        (
            jwsc::Vector_d** means_out, 
            jwsc::Vector_d** stdevs_out,
            size_t           j
        ) 
        const;

        /** @brief Generates a block sample point from the parameter prior. */
        virtual void sample_var_param_prior
        (
            jwsc::Vector_d** means_out,
            jwsc::Vector_d** stdevs_out
        ) 
        const;

        /*@}*/


        /**
         * @name Category Proposal
         *
         * Category proposal probability distribution members.
         * @{
         */

        void sample_var_param_cat_proposal
        (
            jwsc::Vector_d** means_out,
            jwsc::Vector_d** stdevs_out,
            const jwsc::Vector_d* means_in
        ) 
        const;

        /*@}*/


        /**
         * @name Proposal
         *
         * Proposal probability distribution members.
         * @{
         */

        /** 
         * @brief Calculates the proposal log probability of instance
         * parameters.
         */
        virtual double calc_var_param_proposal_log_prob() const;

        /** 
         * @brief Calculates the Gaussian proposal log probability of instance
         * parameters.
         */
        virtual double calc_var_param_gauss_proposal_log_prob() const;

        virtual double calc_var_param_split_merge_proposal_log_prob
        (
            const jwsc::Vector_d* q,
            const jwsc::Vector_d* means,
            uint32_t              j
        )
        const;

        double calc_var_param_switch_proposal_log_prob
        (
            const jwsc::Vector_d* q
        )
        const;

        /** 
         * @brief Generates variable attachment point instance parameter values
         * from the proposal.
         */
        void sample_var_attach_pt_proposal
        (
            jwsc::Vector_d** v_out, 
            size_t           j
        ) 
        const;

        /** 
         * @brief Generates variable attachment point instance parameter values
         * from the Gaussian proposal.
         */
        void sample_var_attach_pt_gauss_proposal
        (
            jwsc::Vector_d**      v_out, 
            size_t                j,
            const jwsc::Vector_d* means
        ) 
        const;

        /** 
         * @brief Generates instance parameter values from the proposal
         * distribution.
         */
        virtual void sample_var_param_proposal(jwsc::Vector_d** p_out) const;

        /**
         * @brief Generates instance parameter values from the Gaussian proposal
         * distribution.
         */
        virtual void sample_var_param_gauss_proposal
        (
            jwsc::Vector_d**      p_out,
            const jwsc::Vector_d* means
        ) 
        const;

        virtual void sample_var_param_split_merge_proposal
        (
            jwsc::Vector_d**      p_out,
            const jwsc::Vector_d* means,
            uint32_t              j
        ) 
        const;

        void sample_var_param_switch_proposal
        (
            jwsc::Vector_d**      p_out,
            const jwsc::Vector_d* means
        ) 
        const;

        /*@}*/


        /**
         * @name Transform
         *
         * Transformation of internal representation members.
         * @{
         */

        /** @brief Translates this Component. */
        virtual void translate(double x, double y, double z);

        /** @brief Rotates this Component about a vector. */
        virtual void rotate(double phi, double x, double y, double z);

        /** @brief Transforms a point from face to block coordinates. */
        void transform_pt_from_face_to_block_coords
        (
            jwsc::Vector_d* p, 
            size_t          i
        ) 
        const;

        /** @brief Transforms a point from block to face coordinates. */
        void transform_pt_from_block_to_face_coords
        (
            jwsc::Vector_d* p, 
            size_t          i
        ) 
        const;

        /*@}*/


        /**
         * @name Conditionals
         *
         * Conditional management members.
         * @{
         */

        /** @brief Returns whether this Component conditions on another. */
        virtual bool conditions(const Model* m) const;

        /*@}*/


        /**
         * @name Pointers
         *
         * Pointer management members.
         * @{
         */

        /** 
         * @brief Serializes Model and Parameter pointers in this block using
         * indices in the @em models collection.
         */
        void serialize(const std::vector<Model*>& models);

        /** 
         * @brief Deserializes Model and Parameter indices in this block using
         * the @em models collection of pointers.
         */
        void deserialize
        (
            const std::vector<Model*>& models
        );

        /** 
         * @brief Updates the Model and Parameter pointers in this block from
         * one set of models to another.
         */
        void update_pointers
        (
            const std::vector<Model*>&       new_models,
            const std::vector<const Model*>& old_models
        );

        /*@}*/


        /**
         * @name Graphical Representation
         *
         * Internal graphical representation members.
         * @{
         */

        /** 
         * @brief Returns the set of polygon faces representing the 3D structure
         * of this block.
         */
        virtual std::vector<const jwscxx::graphics::Polygon_d*> faces() const;

        /** @brief Returns the number of faces this block has. */
        virtual size_t num_faces() const;

        /** @brief Sets the sequential ids of a faces in this block. */
        virtual uint32_t set_sequential_face_ids(uint32_t id);

        /** 
         * @brief Returns whether the graphical representation is outdated and
         * needs updating.
         */
        bool graphical_rep_is_outdated() const;

        /** 
         * @brief Sets the graphical representation as outdated and needs
         * updating.
         */
        void graphical_rep_outdated();

        /** 
         * @brief Updates the graphical representation of this block with its
         * current parameter values.
         */
        virtual void update_graphical_rep();

        /*@}*/


        /**
         * @name Graphics
         *
         * Graphics rendering members.
         * @{
         */

        /** 
         * @brief Renders this block with the GL as a wire-frame using
         * polygon face ids as color.
         */
        virtual void wire_render_with_ids(const View* view) const;

        /** @brief Renders this block with the GL as a solid object. */
        virtual void solid_render() const;

        /*@}*/


    protected:

        /**
         * @name Parameters
         *
         * Generic model parameter members.
         * @{
         */

        /** @brief Sets the Parameter values from a point. */
        virtual void set_var_param_values_in_object_coords
        (
            const jwsc::Vector_d* p
        )
        throw (jwscxx::base::Arg_error);

        virtual void copy_var_param_stats(const Abstract_block* b);

        void swap_var_attach_pts(const Abstract_block* b);

        /*@}*/


        /**
         * @name attach_pt
         *
         * Face attachment points parameter members.
         * @{
         */

        /** @brief Sets a face center attachment point. */
        void ctr_attach_pt_v(const jwsc::Vector_d* a, size_t i);

        /** @brief Sets a face variable attachment point. */
        void var_attach_pt_v(const jwsc::Vector_d* a, size_t i);

        /** 
         * @brief Enables the variable attachment points by placing them in the
         * appropriate parameter sets.
         */
        void enable_var_attach_pts();

        /*@}*/


        /**
         * @name Attachment
         *
         * Component attachment members.
         * @{
         */

        /** @brief Adds internal shift dependencies induced by an attachment. */
        virtual void attach_shift_deps(uint32_t j, uint32_t i);

        /** @brief Adds internal scale dependencies induced by an attachment. */
        virtual void attach_scale_deps(uint32_t j, uint32_t i);

        /** 
         * @brief Removes internal shift dependencies induced by an detachment.
         */
        virtual void detach_shift_deps(uint32_t j, uint32_t i);

        /** 
         * @brief Removes internal scale dependencies induced by an detachment.
         */
        virtual void detach_scale_deps(uint32_t j, uint32_t i);

        /*@}*/


        static size_t attach_pt_face_coord(size_t face_j, size_t coord_i);


    private:

        /**
         * @name Representation
         *
         * Internal representation members.
         * @{
         */

        /** @brief Allocates and initializes the face rotation matrices. */
        void create_face_rotations();

        /** @brief Frees the face rotation matrices. */
        void free_face_rotations();

        /** @brief Allocates and initializes the face attachment points. */
        void create_ctr_attach_pts(size_t* id_in_out);

        /** @brief Allocates and initializes the face attachment points. */
        void create_var_attach_pts
        (
            size_t*                        id_in_out, 
            const Abstract_block::Context* ctx
        );

        /** 
         * @brief Adds the internal parameter shift dependencies for center
         * attachment points.
         */
        void add_ctr_attach_shift_deps();

        /** 
         * @brief Adds the internal parameter scale dependencies for center
         * attachment points.
         */
        void add_ctr_attach_scale_deps();

        /** 
         * @brief Adds the internal parameter shift dependencies for a variable
         * attachment points.
         */
        void add_var_attach_shift_deps();

        /** 
         * @brief Adds the internal parameter scale dependencies for a variable
         * attachment points.
         */
        void add_var_attach_scale_deps();

        /** 
         * @brief Adds the conditions for the variable attachment points.
         */
        void add_var_attach_conds();

        /*@}*/


    protected:

        /** 
         * @brief Size of this block. 
         *
         * Contains three parameters indexed by width=0, height=1, length=2.
         */
        Component::Paramgroup _size_p;

        /** 
         * @brief Block face attachment points.
         *
         * These attachments are fixed to the center of each face. The major
         * index is the block face; the minor is for x,y,z.
         */
        std::vector<Component::Paramgroup> _ctr_attach_pts_p;

        /** 
         * @brief Block face attachment points.
         *
         * These attachments vary their location about each face. The major
         * index is the block face; the minor is for x,y,z.
         */
        std::vector<Component::Paramgroup> _var_attach_pts_p;

        /**
         * @brief Models attached to this one.
         *
         * There are currently six potential attachments: one per size param of
         * this block. If nothing is attached to a side, the Pointer at the
         * corresponding index contains a null ptr.
         *
         * The indexing scheme is the same as that for faces of a Cuboid_d.
         */
        std::vector<Component::Attachment> _attachments;

        /** @brief Geometric representation of this block. */
        jwscxx::graphics::Cuboid_d _cuboid;


    private:

        /**
         * @name Representation
         *
         * Internal representation attributes.
         * @{
         */

        /**
         * @brief Block face rotation matrices.
         *
         * Rotates a point in face coordinates into component coordinates.
         */
        std::vector<jwsc::Matrix_d*> _R_faces;

        /*@}*/


    friend class Object;
};


/** 
 * @class Block
 *
 * @brief Cuboid defined by a width, height and length at the origin or a
 * specified postion.
 *
 * The geometric representation of a block is positioned in Object coordinates.
 * It must be added to the Object to bring it into world coordinates.
 */
class Block : public Abstract_block
{
    public:


/**
 * @class Block::Context
 *
 * @brief Context for block model parameters.
 */
class Context : public Abstract_block::Context
{
    public:

        /** @brief Constructs a new default Block::Context. */
        Context();

        /** @brief Constructs a new Block::Context from another. */
        Context(const Block::Context& ctx);

        /** @brief Copies a Block::Context into this one. */
        Block::Context& operator= (const Block::Context& ctx);

        /** @brief Clones this Context. */
        Block::Context* clone() const;
};


    // Block
    public:

        /**
         * @name Standard
         *
         * Standard class members.
         * @{
         */

        /** @brief Constructs a Block at the origin. */
        Block(const Block::Context* ctx) throw (jwscxx::base::Arg_error);

        /** @brief Constructs a Block at the origin. */
        Block(double w, double h, double l, const Block::Context* ctx) 
            throw (jwscxx::base::Arg_error);

        /** @brief Constructs a Block by copying another. */
        Block(const Block& c);

        /** @brief Deletes this Block. */
        virtual ~Block();

        /** @brief Copies a Block into this one. */
        Block& operator= (const Block& b);

        /** @brief Clones this block. */
        virtual Block* clone() const;

        /** @brief Creates a special copy of this block. */
        virtual Block* clone_and_swap() const throw (jwscxx::base::Arg_error);

        /** @brief Returns the model context. */
        const Block::Context* context() const;

        /*@}*/


        /**
         * @name Selection
         *
         * Random selection of block set dependent components and attach points.
         * @{
         */

        static uint32_t attachable_block
        (
            const std::vector<Component*>& components,
            bool                           height_only
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t attachable_block_attach_pt
        (
            const Block* b,
            bool         height_only
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t detachable_block
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t splitable_block_to_block
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t mergeable_block_from_block
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t mergeable_block_from_block_attach_pt
        (
            const Block* b1
        )
        throw (jwscxx::base::Arg_error);

        /*@}*/


    friend class Object;
};


#endif
