/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Set of block structures represented as cuboids.
 */


#include <config.h>

#include <iostream>
#include <typeinfo>
#include <cassert>
#include <cmath>
#include <cstring>

#if defined JWSCXX_HAVE_OPENGL
#include <GL/gl.h>
#include <GL/glu.h>
#elif defined JWSCXX_HAVE_OPENGL_FRAMEWORK
#include <OpenGL/OpenGL.h>
#endif

#include <inttypes.h>

#include <jwsc/base/limits.h>
#include <jwsc/vector/vector.h>
#include <jwsc/vector/vector_math.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/matrix/matrix_math.h>
#include <jwsc/prob/pdf.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/graphics/cuboid.h>
#include <jwsc++/graphics/polygon.h>

#include "model.h"
#include "component.h"
#include "object.h"
#include "block_set.h"


#define  SIZE                 15.0
#define  SIZE_MINN            10.0
#define  SIZE_MAXX            400.0
#define  SIZE_MU              15.0
#define  SIZE_SIGMA           10.0
#define  SIZE_LAMBDA          20.0
#define  SIZE_ALPHA           1.0
#define  SIZE_BETA            1.0
#define  RATIO_SIZE_MU        0.1
#define  RATIO_SIZE_LAMBDA    1.0
#define  RATIO_SIZE_ALPHA     1.0
#define  RATIO_SIZE_BETA      0.01
#define  SIZE_DELTA_T         0.005
#define  SIZE_GRAD_ETA        1.0
#define  SIZE_SIGMA_DELTA_T   0.0001
#define  SIZE_SIGMA_GRAD_ETA  0.5
#define  RATIO_SIZE_DELTA_T   0.00005
#define  RATIO_SIZE_GRAD_ETA  0.005
#define  RATIO_SIZE_SIGMA_DELTA_T   0.0001
#define  RATIO_SIZE_SIGMA_GRAD_ETA  0.005

#define  NUM_CROSS_MIN             0
#define  NUM_CROSS_MAX             10
#define  NUM_CROSS_MU              3
#define  NUM_CROSS_SIGMA           0.25
#define  NUM_CROSS_LAMBDA          1.0
#define  NUM_CROSS_ALPHA           1.0
#define  NUM_CROSS_BETA            0.01
#define  NUM_CROSS_DELTA_T         0.005
#define  NUM_CROSS_GRAD_ETA        1.0
#define  NUM_CROSS_SIGMA_DELTA_T   0.0001
#define  NUM_CROSS_SIGMA_GRAD_ETA  0.5

#define  CROSS_POSN_1                0.0
#define  CROSS_POSN_1_MIN            0.0
#define  CROSS_POSN_1_MAX            10000.0
#define  CROSS_POSN_1_MU             0.0
#define  CROSS_POSN_1_SIGMA          10.0
#define  RATIO_CROSS_POSN_1_MU       0.0
#define  RATIO_CROSS_POSN_1_LAMBDA   1.0
#define  RATIO_CROSS_POSN_1_ALPHA    1.0
#define  RATIO_CROSS_POSN_1_BETA     0.01
#define  CROSS_POSN_1_DELTA_T        0.005
#define  CROSS_POSN_1_GRAD_ETA       1.0
#define  RATIO_CROSS_POSN_1_DELTA_T  0.00005
#define  RATIO_CROSS_POSN_1_GRAD_ETA 0.005
#define  RATIO_CROSS_POSN_1_SIGMA_DELTA_T   0.0001
#define  RATIO_CROSS_POSN_1_SIGMA_GRAD_ETA  0.005

#define  CROSS_HEIGHT                10.0
#define  CROSS_HEIGHT_MIN            10.0
#define  CROSS_HEIGHT_MAX            400.0
#define  CROSS_HEIGHT_MU             10.0
#define  CROSS_HEIGHT_SIGMA          10.0
#define  RATIO_CROSS_HEIGHT_MU       0.1
#define  RATIO_CROSS_HEIGHT_LAMBDA   1.0
#define  RATIO_CROSS_HEIGHT_ALPHA    1.0
#define  RATIO_CROSS_HEIGHT_BETA     0.01
#define  CROSS_HEIGHT_DELTA_T        0.005
#define  CROSS_HEIGHT_GRAD_ETA       1.0
#define  RATIO_CROSS_HEIGHT_DELTA_T        0.00005
#define  RATIO_CROSS_HEIGHT_GRAD_ETA       0.005
#define  RATIO_CROSS_HEIGHT_SIGMA_DELTA_T  0.0001
#define  RATIO_CROSS_HEIGHT_SIGMA_GRAD_ETA 0.005

#define  CROSS_POSN_N                50.0
#define  CROSS_POSN_N_MIN            0.0
#define  CROSS_POSN_N_MAX            10000.0
#define  CROSS_POSN_N_MU             50.0
#define  CROSS_POSN_N_SIGMA          10.0
#define  RATIO_CROSS_POSN_N_MU       1.0
#define  RATIO_CROSS_POSN_N_LAMBDA   1.0
#define  RATIO_CROSS_POSN_N_ALPHA    1.0
#define  RATIO_CROSS_POSN_N_BETA     0.01
#define  CROSS_POSN_N_DELTA_T        0.005
#define  CROSS_POSN_N_GRAD_ETA       1.0
#define  RATIO_CROSS_POSN_N_DELTA_T  0.00005
#define  RATIO_CROSS_POSN_N_GRAD_ETA 0.005
#define  RATIO_CROSS_POSN_N_SIGMA_DELTA_T   0.0001
#define  RATIO_CROSS_POSN_N_SIGMA_GRAD_ETA  0.005


/* -----------------------------  Block_set::Context  ----------------------- */

Block_set::Context::Context() : Abstract_block::Context()
{
    typedef Model::Parameter::Context MPContext;
    typedef MPContext::Range          Range;
    typedef MPContext::Dynamics       Dynamics;
    typedef MPContext::Uniform        Uniform;
    typedef MPContext::Gaussian       Gaussian;
    typedef MPContext::Gaussian_gamma Gaussian_gamma;

    MPContext* c;

    for (uint32_t i = 0; i < 2; i++)
    {
        _inner_size_c.add_context(MPContext());

        c = &(_inner_size_c[ i ]);
        c->range(Range(SIZE_MINN, SIZE_MAXX));
        c->init_value(SIZE);
        c->likelihood(Gaussian(SIZE_MU, SIZE_SIGMA));
        c->prior(Gaussian_gamma(SIZE_MU, SIZE_LAMBDA, SIZE_ALPHA, SIZE_BETA));
        c->ratio_prior(Gaussian_gamma(RATIO_SIZE_MU, RATIO_SIZE_LAMBDA,
                    RATIO_SIZE_ALPHA, RATIO_SIZE_BETA));
        c->proposal(Gaussian(SIZE_MU, SIZE_SIGMA));
        c->gauss_proposal(Gaussian(SIZE_MU, SIZE_SIGMA));
        c->value_dynamics(Dynamics(SIZE_DELTA_T, SIZE_GRAD_ETA));
        c->mean_dynamics(Dynamics(SIZE_DELTA_T, SIZE_GRAD_ETA));
        c->stdev_dynamics(Dynamics(SIZE_SIGMA_DELTA_T, SIZE_SIGMA_GRAD_ETA));
        c->ratio_mean_dynamics(Dynamics(RATIO_SIZE_DELTA_T,
                    RATIO_SIZE_GRAD_ETA));
        c->ratio_stdev_dynamics(Dynamics(RATIO_SIZE_SIGMA_DELTA_T,
                    RATIO_SIZE_SIGMA_GRAD_ETA));
    }

    uint32_t c_i = 0;
    _cross_c.add_context(MPContext());
    c = &(_cross_c[ c_i++ ]);
    c->range(Range(NUM_CROSS_MIN, NUM_CROSS_MAX));
    c->init_value(NUM_CROSS_MIN);
    c->likelihood(Gaussian(NUM_CROSS_MU, NUM_CROSS_SIGMA));
    c->prior(Gaussian_gamma(NUM_CROSS_MU, NUM_CROSS_LAMBDA, NUM_CROSS_ALPHA,
                NUM_CROSS_BETA));
    c->proposal(Gaussian(NUM_CROSS_MU, NUM_CROSS_SIGMA));
    c->gauss_proposal(Gaussian(NUM_CROSS_MU, NUM_CROSS_SIGMA));
    c->value_dynamics(Dynamics(NUM_CROSS_DELTA_T, NUM_CROSS_GRAD_ETA));
    c->mean_dynamics(Dynamics(NUM_CROSS_DELTA_T, NUM_CROSS_GRAD_ETA));
    c->stdev_dynamics(Dynamics(NUM_CROSS_SIGMA_DELTA_T,
                NUM_CROSS_SIGMA_GRAD_ETA));

    _cross_c.add_context(MPContext());
    c = &(_cross_c[ c_i++ ]);
    c->range(Range(CROSS_POSN_1_MIN, CROSS_POSN_1_MAX));
    c->init_value(CROSS_POSN_1);
    c->likelihood(Gaussian(CROSS_POSN_1_MU, CROSS_POSN_1_SIGMA));
    c->ratio_prior(Gaussian_gamma(RATIO_CROSS_POSN_1_MU,
                RATIO_CROSS_POSN_1_LAMBDA, RATIO_CROSS_POSN_1_ALPHA,
                RATIO_CROSS_POSN_1_BETA));
    c->proposal(Gaussian(CROSS_POSN_1_MU, CROSS_POSN_1_SIGMA));
    c->gauss_proposal(Gaussian(CROSS_POSN_1_MU, CROSS_POSN_1_SIGMA));
    c->value_dynamics(Dynamics(CROSS_POSN_1_DELTA_T, CROSS_POSN_1_GRAD_ETA));
    c->ratio_mean_dynamics(Dynamics(RATIO_CROSS_POSN_1_DELTA_T,
                RATIO_CROSS_POSN_1_GRAD_ETA));
    c->ratio_stdev_dynamics(Dynamics(RATIO_CROSS_POSN_1_SIGMA_DELTA_T,
                RATIO_CROSS_POSN_1_SIGMA_GRAD_ETA));

    _cross_c.add_context(MPContext());
    c = &(_cross_c[ c_i++ ]);
    c->range(Range(CROSS_HEIGHT_MIN, CROSS_HEIGHT_MAX));
    c->init_value(CROSS_HEIGHT);
    c->likelihood(Gaussian(CROSS_HEIGHT_MU, CROSS_HEIGHT_SIGMA));
    c->ratio_prior(Gaussian_gamma(RATIO_CROSS_HEIGHT_MU,
                RATIO_CROSS_HEIGHT_LAMBDA, RATIO_CROSS_HEIGHT_ALPHA,
                RATIO_CROSS_HEIGHT_BETA));
    c->proposal(Gaussian(CROSS_HEIGHT_MU, CROSS_HEIGHT_SIGMA));
    c->gauss_proposal(Gaussian(CROSS_HEIGHT_MU, CROSS_HEIGHT_SIGMA));
    c->value_dynamics(Dynamics(CROSS_HEIGHT_DELTA_T, CROSS_HEIGHT_GRAD_ETA));
    c->ratio_mean_dynamics(Dynamics(RATIO_CROSS_HEIGHT_DELTA_T,
                RATIO_CROSS_HEIGHT_GRAD_ETA));
    c->ratio_stdev_dynamics(Dynamics(RATIO_CROSS_HEIGHT_SIGMA_DELTA_T,
                RATIO_CROSS_HEIGHT_SIGMA_GRAD_ETA));

    _cross_c.add_context(MPContext());
    c = &(_cross_c[ c_i++ ]);
    c->range(Range(CROSS_POSN_N_MIN, CROSS_POSN_N_MAX));
    c->init_value(CROSS_POSN_N);
    c->likelihood(Gaussian(CROSS_POSN_N_MU, CROSS_POSN_N_SIGMA));
    c->ratio_prior(Gaussian_gamma(RATIO_CROSS_POSN_N_MU,
                RATIO_CROSS_POSN_N_LAMBDA, RATIO_CROSS_POSN_N_ALPHA,
                RATIO_CROSS_POSN_N_BETA));
    c->proposal(Gaussian(CROSS_POSN_N_MU, CROSS_POSN_N_SIGMA));
    c->gauss_proposal(Gaussian(CROSS_POSN_N_MU, CROSS_POSN_N_SIGMA));
    c->value_dynamics(Dynamics(CROSS_POSN_N_DELTA_T, CROSS_POSN_N_GRAD_ETA));
    c->ratio_mean_dynamics(Dynamics(RATIO_CROSS_POSN_N_DELTA_T,
                RATIO_CROSS_POSN_N_GRAD_ETA));
    c->ratio_stdev_dynamics(Dynamics(RATIO_CROSS_POSN_N_SIGMA_DELTA_T,
                RATIO_CROSS_POSN_N_SIGMA_GRAD_ETA));
}

Block_set::Context::Context(const Block_set::Context& ctx)
    : Abstract_block::Context(ctx)
{
    *this = ctx;
}

Block_set::Context& Block_set::Context::operator=(const Block_set::Context& ctx)
{
    Abstract_block::Context::operator=(ctx);

    _inner_size_c = ctx._inner_size_c;
    _cross_c      = ctx._cross_c;

    return *this;
}

Block_set::Context* Block_set::Context::clone() const
{
    return new Block_set::Context(*this);
}

const Component::Paramgroup::Context& Block_set::Context::inner_size_c() const
{
    return _inner_size_c;
}

void Block_set::Context::inner_size_c
(
    const Component::Paramgroup::Context& size
)
{
    _inner_size_c = size;
}

const Model::Parameter::Context& Block_set::Context::inner_size_c(size_t i) 
    const
{
    return _inner_size_c[ i ];
}

void Block_set::Context::inner_size_c
(
    size_t                           i, 
    const Model::Parameter::Context& size
)
{
    _inner_size_c[ i ] = size;
}

const Component::Paramgroup::Context& Block_set::Context::cross_c() const
{
    return _cross_c;
}

void Block_set::Context::cross_c
(
    const Component::Paramgroup::Context& c
)
{
    _cross_c = c;
}

const Model::Parameter::Context& Block_set::Context::cross_c(size_t i) 
    const
{
    return _cross_c[ i ];
}

void Block_set::Context::cross_c
(
    size_t                           i, 
    const Model::Parameter::Context& c
)
{
    _cross_c[ i ] = c;
}

/* ========================================================================== */






/* -----------------------------  Block_set  -------------------------------- */

/**
 * The Block_set is centered at (0, 0, 0).
 *
 * @param  name  Name of the set of blocks.
 * @param  ctx   Parameter context for this Block_set.
 *
 * @throw  jwscxx::base::Arg_error  Dimensions must be > 0.
 */
Block_set::Block_set(const char* name, const Block_set::Context* ctx) 
    throw (jwscxx::base::Arg_error) 
    : Abstract_block(name, 
            ctx->size_c(0).init_value(), 
            ctx->size_c(1).init_value(),
            ctx->size_c(2).init_value(),
            0, 0, 0, ctx),
      _inner_size_p(&(ctx->inner_size_c())),
      _cross_p(&(ctx->cross_c()))
{
    size_t id = num_var_params() + 
                num_derived_params() +
                num_fixed_params();

    _cross_p.add_param(Parameter(id++, "Num Crosses", 
                0, &(ctx->cross_c(0)), this));
    _cross_p.add_param(Parameter(id++, "Cross 1 Position",
                ctx->cross_c(1).init_value(), &(ctx->cross_c(1)), this));
    _cross_p.add_param(Parameter(id++, "Cross Height", 
                ctx->cross_c(2).init_value(), &(ctx->cross_c(2)), this));
    _cross_p.add_param(Parameter(id++, "Cross N Position",
                ctx->cross_c(3).init_value(), &(ctx->cross_c(3)), this));

    add_derived_param(&_cross_p[0]);
    add_derived_param(&_cross_p[1]);
    add_derived_param(&_cross_p[2]);
    add_derived_param(&_cross_p[3]);
}

/**
 * The Block_set is centered at (0, 0, 0).
 *
 * @param  name    Name of the set of blocks.
 * @param  width   Width.
 * @param  height  Height.
 * @param  length  Length.
 * @param  ctx     Parameter context for this Block_set.
 *
 * @throw  jwscxx::base::Arg_error  Dimensions must be > 0.
 */
Block_set::Block_set
(
    const char* name, 
    double      width,
    double      height,
    double      length,
    const Block_set::Context* ctx
)
throw (jwscxx::base::Arg_error) 
    : Abstract_block(name, width, height, length, 0, 0, 0, ctx),
      _inner_size_p(&(ctx->inner_size_c())),
      _cross_p(&(ctx->cross_c()))
{
    size_t id = num_var_params() + 
                num_derived_params() +
                num_fixed_params();

    _cross_p.add_param(Parameter(id++, "Num Crosses", 
                0, &(ctx->cross_c(0)), this));
    _cross_p.add_param(Parameter(id++, "Cross 1 Position",
                ctx->cross_c(1).init_value(), &(ctx->cross_c(1)), this));
    _cross_p.add_param(Parameter(id++, "Cross Height", 
                ctx->cross_c(2).init_value(), &(ctx->cross_c(2)), this));
    _cross_p.add_param(Parameter(id++, "Cross N Position",
                ctx->cross_c(3).init_value(), &(ctx->cross_c(3)), this));

    add_derived_param(&_cross_p[0]);
    add_derived_param(&_cross_p[1]);
    add_derived_param(&_cross_p[2]);
    add_derived_param(&_cross_p[3]);
}

/**
 * The Block_set is centered at (0, 0, 0). If cross_N is one, the value of
 * cross_dN is ignored, as it is a derived parameter in that case.
 *
 * @param  name      Name of the set of blocks.
 * @param  width     Width.
 * @param  height    Height.
 * @param  length    Length.
 * @param  cross_N   Number of cross blocks.
 * @param  cross_d1  Distance from the top to the first cross block.
 * @param  cross_h   Height of each cross block.
 * @param  cross_dN  Distance from the top to the last cross block.
 * @param  ctx       Parameter context for this Block_set.
 *
 * @throw  jwscxx::base::Arg_error  Dimensions must be > 0.
 */
Block_set::Block_set
(
    const char* name, 
    double      width,
    double      height,
    double      length,
    double      cross_N,
    double      cross_d1,
    double      cross_h,
    double      cross_dN,
    const Block_set::Context* ctx
)
throw (jwscxx::base::Arg_error) 
    : Abstract_block(name, width, height, length, 0, 0, 0, ctx),
      _inner_size_p(&(ctx->inner_size_c())),
      _cross_p(&(ctx->cross_c()))
{
    size_t id = num_var_params() + 
                num_derived_params() +
                num_fixed_params();

    _cross_p.add_param(Parameter(id++, "Num Crosses", 0,
                (ctx) ? &(ctx->cross_c(0)) : 0, this));
    _cross_p.add_param(Parameter(id++, "Cross 1 Position", cross_d1,
                (ctx) ? &(ctx->cross_c(1)) : 0, this));
    _cross_p.add_param(Parameter(id++, "Cross Height", cross_h,
                (ctx) ? &(ctx->cross_c(2)) : 0, this));
    _cross_p.add_param(Parameter(id++, "Cross N Position", cross_dN,
                (ctx) ? &(ctx->cross_c(3)) : 0, this));

    add_derived_param(&_cross_p[0]);
    add_derived_param(&_cross_p[1]);
    add_derived_param(&_cross_p[2]);
    add_derived_param(&_cross_p[3]);
}

/** @param  b  Block_set to copy into this one. */
Block_set::Block_set(const Block_set& b) 
    : Abstract_block(b),
      _inner_size_p(b._inner_size_p),
      _cross_p(b._cross_p),
      _inner_cuboids(b._inner_cuboids),
      _cross_cuboids(b._cross_cuboids)
{
    add_derived_param(&_cross_p[0]);
    add_derived_param(&_cross_p[1]);
    add_derived_param(&_cross_p[2]);
    add_derived_param(&_cross_p[3]);

    uint32_t M = b.num_cross_blocks();

    if (M > 0)
    {
        for (uint32_t i = 0; i < 3; i++)
        {
            remove_derived_param(&(_cross_p[ i ]));
            add_var_param(&(_cross_p[ i ]));
        }
    }
    if (M > 1)
    {
        remove_derived_param(&(_cross_p[3]));
        add_var_param(&(_cross_p[3]));
    }

    *this = b;
}

Block_set::~Block_set()
{
}

/** 
 * @param  b  Block_set to copy into this one.
 *
 * @return A reference to this Block_set.
 */
Block_set& Block_set::operator=(const Block_set& b)
{
    Abstract_block::operator=(b);

    assert(_inner_cuboids.size() == b._inner_cuboids.size());
    for (size_t i = 0; i < b._inner_cuboids.size(); i++)
    {
        _inner_cuboids[ i ] = b._inner_cuboids[ i ];
    }
    _inner_size_p = b._inner_size_p;
    for (size_t i = 0; i < _inner_size_p.num_params(); i++)
    {
        _inner_size_p[ i ].model(this);
    }

    _cross_cuboids.clear();
    for (size_t i = 0; i < b._cross_cuboids.size(); i++)
    {
        _cross_cuboids.push_back(b._cross_cuboids[ i ]);
    }
    _cross_p = b._cross_p;
    for (size_t i = 0; i < _cross_p.num_params(); i++)
    {
        _cross_p[ i ].model(this);
    }

    return *this;
}

/** 
 * The file format is ASCII.
 *
 * @param  in  Input stream to read this Block_set from.
 *
 * @throw  jwscxx::base::IO_error   Could not read from @em in.
 * @throw  jwscxx::base::Arg_error  Invalid arguments in file to read from.
 */
void Block_set::read(std::istream& in) 
    throw (jwscxx::base::IO_error, jwscxx::base::Arg_error)
{
    using std::ostringstream;
    using std::istringstream;
    using jwscxx::base::Arg_error;

    const char* type_name = typeid(*this).name();
    const char* field_value;

    if (!(field_value = read_field_value(in, "Type")))
    {
        throw Arg_error("Missing Type field");
    }
    if (strncmp(field_value, type_name, strlen(type_name)) != 0)
    {
        ostringstream ost;
        ost << "Tried to read a '" << field_value << "' as a '"
            << type_name << "'";
        throw Arg_error(ost.str());
    }

    istringstream ist;
    uint32_t N;
    if (!(field_value = read_field_value(in, "Num Inner Blocks")))
    {
        throw Arg_error("Missing Num Inner Blocks field");
    }
    ist.str(field_value);
    ist >> N;
    if (ist.fail() || N != _inner_cuboids.size())
    {
        throw Arg_error("Invalid Num Inner Blocks");
    }
    ist.clear(std::ios_base::goodbit);

    for (size_t i = 0; i < _inner_cuboids.size(); i++)
    {
        _inner_cuboids[ i ].read(in);
    }

    if (!(field_value = read_field_value(in, "Num Cross Blocks")))
    {
        throw Arg_error("Missing Num Cross Blocks field");
    }
    ist.str(field_value);
    ist >> N;
    if (ist.fail())
    {
        throw Arg_error("Invalid Num Cross Blocks");
    }
    ist.clear(std::ios_base::goodbit);

    if (N > 0 && num_cross_blocks() == 0)
    {
        for (uint32_t i = 0; i < 3; i++)
        {
            remove_derived_param(&(_cross_p[ i ]));
            add_var_param(&(_cross_p[ i ]));
        }
    }
    if (N > 1 && num_cross_blocks() <= 1)
    {
        remove_derived_param(&(_cross_p[3]));
        add_var_param(&(_cross_p[3]));
    }

    if (!(field_value = read_field_value(in, "Num Cross Cuboids")))
    {
        throw Arg_error("Missing Num Cross Cuboids field");
    }
    ist.str(field_value);
    ist >> N;
    if (ist.fail())
    {
        throw Arg_error("Invalid Num Cross Cuboids");
    }
    ist.clear(std::ios_base::goodbit);

    _cross_cuboids.clear();
    for (size_t i = 0; i < N; i++)
    {
        _cross_cuboids.push_back(jwscxx::graphics::Cuboid_d(in));
    }

    Abstract_block::read(in);
}

/**
 * The file format is ASCII.
 *
 * @param  out  Output stream to write this Block_set to.
 *
 * @throw  jwscxx::base::IO_error  Could not write to @em out.
 */
void Block_set::write(std::ostream& out) const throw (jwscxx::base::IO_error)
{
    out << "Type: " << typeid(*this).name() << std::endl;

    out << "Num Inner Blocks: " << _inner_cuboids.size() << std::endl;
    for (size_t i = 0; i < _inner_cuboids.size(); i++)
    {
        _inner_cuboids[ i ].write(out);
    }

    out << "Num Cross Blocks: " << num_cross_blocks() << std::endl;
    out << "Num Cross Cuboids: " << _cross_cuboids.size() << std::endl;
    for (size_t i = 0; i < _cross_cuboids.size(); i++)
    {
        _cross_cuboids[ i ].write(out);
    }

    Abstract_block::write(out);
}

const Block_set::Context* Block_set::context() const
{
    return dynamic_cast<const Block_set::Context*>(Abstract_block::context());
}

uint32_t Block_set::num_inner_blocks() const
{
    return _inner_cuboids.size();
}

uint32_t Block_set::num_cross_blocks() const
{
    return cross(0);
}

bool Block_set::is_mergeable(uint32_t i) const
{
    if (!Abstract_block::is_mergeable(i))
        return false;

    if (num_cross_blocks())
        return false;

    Abstract_block* b;
    assert((b = dynamic_cast<Abstract_block*>(attachment(i))));

    uint32_t j = (i % 2 == 0) ? i+1 : i-1;
    for (uint32_t k = 0; k < 6; k++)
    {
        if (j == k)
            continue;

        assert(!dynamic_cast<Block_set*>(attachment(k)));

        if (dynamic_cast<Block_set*>(b->attachment(k)))
        {
            return false;
        }
    }

    return true;
}

/** 
 * The block set must already be in Object coordinates, and for efficiency
 * reasons, the graphical representation is not updated, but a flag is set to
 * indicate that the update_graphical_rep function should be called when
 * opportune.
 *
 * If a Parameter value could not be set, it is left unchanged and an exception
 * is thrown, after an attempt to set the other parameters is made.
 */
void Block_set::set_var_param_values_in_object_coords(const jwsc::Vector_d* p)
    throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    Abstract_block::set_var_param_values_in_object_coords(p);

    uint32_t N_i;
    uint32_t D = num_var_params();
    uint32_t d;
    double   delta;
    double   epsilon = 1.0e-8;
    double   e;

    std::ostringstream ost;

    N_i = _inner_size_p.num_params();
    d = var_param_inner_size_i(0);

    for (uint32_t i = 0; i < N_i; i++)
    {
        if (fabs(delta = p->elts[ d ] - inner_size(i)) > 0)
        {
            e = (delta < 0) ? -epsilon : epsilon;
            try 
            { 
                _inner_size_p[ i ].test_value_shift_deps(inner_size(i) + 
                        delta + e); 
                _inner_size_p[ i ].value_shift_deps(inner_size(i) + delta);
            }
            catch (jwscxx::base::Arg_error e) 
            { ost << e.get_msg() << "  "; }
        }
        d++;
    }

    assert(d <= D);
    N_i = num_var_param_cross_p();
    d = (N_i) ? var_param_cross_i(0) : 0;

    if (N_i > 0 && fabs(delta = floor(p->elts[ d ] - cross(0) + 0.5)) > 0 &&
            cross(0) > 1)
    {
        int n = delta;
        try 
        { 
            if (cross(0) + n <= 1)
            {
                throw jwscxx::base::Arg_error("Cross block count <= 1");
            }
            else if (n > 0)
            {
                test_add_cross_blocks(n);
                add_cross_blocks(n);
            }
            else if (n < 0)
            {
                test_remove_cross_blocks(-n);
                remove_cross_blocks(-n);
            }
        }
        catch (jwscxx::base::Arg_error e) 
        { ost << e.get_msg() << "  "; }
    }

    N_i = num_var_param_cross_p();
    d = (N_i) ? var_param_cross_i(1) : 0;

    for (uint32_t i = 1; i < N_i; i++)
    {
        if (fabs(delta = p->elts[ d ] - cross(i)) > 0)
        {
            e = (delta < 0) ? -epsilon : epsilon;
            try 
            { 
                _cross_p[ i ].test_value_shift_deps(cross(i) + delta + e); 
                _cross_p[ i ].value_shift_deps(cross(i) + delta);
            }
            catch (jwscxx::base::Arg_error e) 
            { ost << e.get_msg() << "  "; }
        }
        d++;
    }

    assert(d <= D);

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

/**
 * The block set must already be in Object coordinates, and for efficiency
 * reasons, the graphical representation is not updated, but a flag is set to
 * indicate that the update_graphical_rep function should be called when
 * opportune.
 *
 * If a Parameter value could not be set, it is left unchanged and an exception
 * is thrown, after an attempt to set the other parameters is made.
 *
 * @param  p  Point to add to the Parameter values of this block.
 *
 * @throw  jwscxx::base::Arg_error  Invalid resulting parameter value.
 */
void Block_set::add_var_param_values_scale_deps(const jwsc::Vector_d* p)
    throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    Abstract_block::add_var_param_values_scale_deps(p);

    uint32_t N_i;
    uint32_t D = num_var_params();
    uint32_t d;
    double   delta;
    double   epsilon = 1.0e-8;
    double   e;

    std::ostringstream ost;

    N_i = _inner_size_p.num_params();
    d = var_param_inner_size_i(0);

    for (uint32_t i = 0; i < N_i; i++)
    {
        if (fabs(delta = p->elts[ d ]) > 0)
        {
            e = (delta < 0) ? -epsilon : epsilon;
            try 
            { 
                _inner_size_p[ i ].test_value_scale_deps(inner_size(i) + 
                        delta + e); 
                _inner_size_p[ i ].value_scale_deps(inner_size(i) + delta);
            }
            catch (jwscxx::base::Arg_error e) 
            { ost << e.get_msg() << "  "; }
        }
        d++;
    }

    assert(d <= D);
    N_i = num_var_param_cross_p();
    d = (N_i) ? var_param_cross_i(0) : 0;

    if (N_i > 0 && fabs(delta = floor(p->elts[ d ] + 0.5)) > 0 && cross(0) > 1)
    {
        int n = delta;
        try 
        { 
            if (cross(0) + n <= 1)
            {
                throw jwscxx::base::Arg_error("Cross block count <= 1");
            }
            else if (n > 0)
            {
                test_add_cross_blocks(n);
                add_cross_blocks(n);
            }
            else if (n < 0)
            {
                test_remove_cross_blocks(-n);
                remove_cross_blocks(-n);
            }
        }
        catch (jwscxx::base::Arg_error e) 
        { ost << e.get_msg() << "  "; }
    }

    N_i = num_var_param_cross_p();
    d = (N_i) ? var_param_cross_i(1) : 0;

    for (uint32_t i = 1; i < N_i; i++)
    {
        if (fabs(delta = p->elts[ d ]) > 0)
        {
            e = (delta < 0) ? -epsilon : epsilon;
            try 
            { 
                _cross_p[ i ].test_value_scale_deps(cross(i) + delta + e); 
                _cross_p[ i ].value_scale_deps(cross(i) + delta);
            }
            catch (jwscxx::base::Arg_error e) 
            { ost << e.get_msg() << "  "; }
        }
        d++;
    }

    assert(d <= D);

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

void Block_set::print_var_param_values
(
    uint32_t      iteration, 
    std::ostream& out
)
const
{
    Abstract_block::print_var_param_values(iteration, out);

    double x, x_cond;
    double r;

    uint32_t N_i = _inner_size_p.num_params();
    for (uint32_t i = 0; i < N_i; i++)
    {
        x = inner_size_p(i)->value();
        x_cond = inner_size_p(i)->cond(0)->ptr()->value();
        r = x / x_cond;

        out << iteration << ',' << name() << ',' << id()  << ','
            << "Ratio" << ',' << inner_size_p(i)->name() << ','
            << r << std::endl;
    }

    N_i = num_var_param_cross_p();
    for (uint32_t i = 1; i < N_i; i++)
    {
        x = cross_p(i)->value();
        x_cond = cross_p(i)->cond(0)->ptr()->value();
        r = x / x_cond;

        out << iteration << ',' << name() << ',' << id()  << ','
            << "Ratio" << ',' << cross_p(i)->name() << ','
            << r << std::endl;
    }
}

void Block_set::copy_var_param_stats(const Abstract_block* b)
{
    using namespace jwsc;

    const Block_set* bs = dynamic_cast<const Block_set*>(b);
    assert(bs);

    Abstract_block::copy_var_param_stats(bs);

    double m, s;
    uint32_t N_i = _inner_size_p.num_params();
    assert(N_i == bs->_inner_size_p.num_params());
    for (size_t i = 0; i < N_i; i++)
    {
        bs->_inner_size_p[ i ].stats(&m, &s);
        _inner_size_p[ i ].stats(m, s);
    }

    N_i = _cross_p.num_params();
    assert(N_i == bs->_cross_p.num_params());
    for (size_t i = 0; i < N_i; i++)
    {
        bs->_cross_p[ i ].stats(&m, &s);
        _cross_p[ i ].stats(m, s);
    }
}

const Component::Paramgroup* Block_set::inner_size_p() const
{
    return &_inner_size_p;
}

const Component::Parameter* Block_set::inner_size_p(size_t i) const
{
    return &(_inner_size_p[ i ]);
}

void Block_set::inner_size_v(jwsc::Vector_d** s_out) const
{
    using namespace jwsc;

    size_t N_i = _inner_size_p.num_params();

    create_vector_d(s_out, N_i);
    Vector_d* s = *s_out;

    for (size_t i = 0; i < N_i; i++)
    {
        s->elts[ i ] = _inner_size_p[ i ].value();
    }
}

double Block_set::inner_size(size_t i) const
{
    return inner_size_p(i)->value();
}

void Block_set::inner_size_shift_deps(size_t i, double size)
{
    transform_from_world_to_object_coords();
    _inner_size_p[ i ].value_shift_deps(size);
    update_object();
    transform_from_object_to_world_coords();
}

void Block_set::inner_size_scale_deps(size_t i, double size)
{
    transform_from_world_to_object_coords();
    _inner_size_p[ i ].value_scale_deps(size);
    update_object();
    transform_from_object_to_world_coords();
}

/**
 * @param  i       Size index.
 * @param  size    Size to test.
 *
 * @throw  jwscxx::base::Arg_error  Invalid size.
 */
void Block_set::test_inner_size_shift_deps(size_t i, double size)
    throw (jwscxx::base::Arg_error)
{
    transform_from_world_to_object_coords();
    try 
    {
        _inner_size_p[ i ].test_value_shift_deps(size); 
    }
    catch (jwscxx::base::Arg_error e)
    {
        transform_from_object_to_world_coords();
        throw e;
    }
    transform_from_object_to_world_coords();
}

/**
 * @param  i       Size index.
 * @param  size    Size to test.
 *
 * @throw  jwscxx::base::Arg_error  Invalid size.
 */
void Block_set::test_inner_size_scale_deps(size_t i, double size)
    throw (jwscxx::base::Arg_error)
{
    transform_from_world_to_object_coords();
    try
    {
        _inner_size_p[ i ].test_value_scale_deps(size);
    }
    catch (jwscxx::base::Arg_error e)
    {
        transform_from_object_to_world_coords();
        throw e;
    }
    transform_from_object_to_world_coords();
}

uint32_t Block_set::var_param_inner_size_i(uint32_t j) const
{
    uint32_t i;

    if ((i = var_param(&(_inner_size_p[ j ]))) != num_var_params())
    {
        return i;
    }

    abort();
}

const Component::Paramgroup* Block_set::cross_p() const
{
    return &_cross_p;
}

const Component::Parameter* Block_set::cross_p(size_t i) const
{
    return &(_cross_p[ i ]);
}

double Block_set::cross(size_t i) const
{
    if (i == 0)
    {
        return floor(cross_p(i)->value() + 0.5);
    }
    return cross_p(i)->value();
}

uint32_t Block_set::var_param_cross_i(uint32_t j) const
{
    uint32_t i;

    if ((i = var_param(&(_cross_p[ j ]))) != num_var_params())
    {
        return i;
    }

    abort();
}

uint32_t Block_set::num_var_param_cross_p() const
{
    if (num_cross_blocks() == 0)
    {
        return 0;
    }
    else if (num_cross_blocks() == 1)
    {
        return 3;
    }
    return 4;
}

void Block_set::add_cross_blocks(uint32_t N)
{
    uint32_t M = num_cross_blocks();

    if (N == 0) 
        return;

    bool was_in_world_coords = false;
    if (object() && (was_in_world_coords = !(object()->in_object_coords())))
    {
        transform_from_world_to_object_coords();
    }

    if (M == 0)
    {
        for (uint32_t i = 0; i < 3; i++)
        {
            remove_derived_param(&(_cross_p[ i ]));
            add_var_param(&(_cross_p[ i ]));
        }
        if (N == 1)
        {
            try { _cross_p[3].value(cross(1) + cross(2)); }
            catch (jwscxx::base::Arg_error e) { abort(); }
        }
    }
    if ((M == 0 && N > 1) || M == 1)
    {
        remove_derived_param(&(_cross_p[3]));
        add_var_param(&(_cross_p[3]));
    }

    for (uint32_t n = 0; n < N; n++)
    {
        add_cross_block();
    }

    bool attach_deps = (attachment(3) || 
            ((object() && object()->lowest_component() == this)));

    if (attach_deps && M == 0)
    {
        attach_shift_deps_1();
        attach_scale_deps_1();
    }
    if (attach_deps && ((M == 0 && N > 1) || M == 1))
    {
        attach_shift_deps_2();
        attach_scale_deps_2();
    }

    if (was_in_world_coords)
    {
        transform_from_object_to_world_coords();
    }
}

void Block_set::test_add_cross_blocks(uint32_t N) 
    throw (jwscxx::base::Arg_error)
{
    _cross_p[0].test_value(num_cross_blocks()+N);
}

void Block_set::remove_cross_blocks(uint32_t N)
{
    uint32_t M = num_cross_blocks();
    assert(M >= N);

    if (N == 0) 
        return;

    bool was_in_world_coords = false;
    if (object() && (was_in_world_coords = !(object()->in_object_coords())))
    {
        transform_from_world_to_object_coords();
    }

    if ((M-N) == 0)
    {
        for (uint32_t i = 0; i < 3; i++)
        {
            remove_var_param(&(_cross_p[ i ]));
            add_derived_param(&(_cross_p[ i ]));
        }
    }
    if (((M-N) == 0 && N > 1) || (M-N) == 1)
    {
        remove_var_param(&(_cross_p[3]));
        add_derived_param(&(_cross_p[3]));

        try { _cross_p[3].value(cross(1) + cross(2)); }
        catch (jwscxx::base::Arg_error e) { abort(); }
    }

    bool detach_deps = (attachment(3) || 
            ((object() && object()->lowest_component() == this)));

    if (detach_deps && (M-N) == 0)
    {
        detach_shift_deps_1();
        detach_scale_deps_1();
    }
    if (detach_deps && (((M-N) == 0 && N > 1) || (M-N) == 1))
    {
        detach_shift_deps_2();
        detach_scale_deps_2();
    }

    for (uint32_t n = 0; n < N; n++)
    {
        remove_cross_block();
    }

    if (was_in_world_coords)
    {
        transform_from_object_to_world_coords();
    }
}

void Block_set::test_remove_cross_blocks(uint32_t N) 
    throw (jwscxx::base::Arg_error)
{
    _cross_p[0].test_value(num_cross_blocks()-N);
}

void Block_set::add_cross_block()
{
    typedef Parameter P;

    uint32_t N = num_cross_blocks();

    P* h       = &(_size_p[1]);
    P* d1      = &(_cross_p[1]);
    P* dh      = &(_cross_p[2]);
    P* dN      = &(_cross_p[3]);
    P* a_pos_j = &(_ctr_attach_pts_p[2][1]);

    if (N == 0)
    {
        h->add_shift_dep(P::Pos_half_dependent(d1));
        h->add_scale_dep(P::Fancy_pos_half_dependent(d1));
        a_pos_j->add_scale_dep(P::Pos_dependent(d1));
        d1->add_shift_dep(P::Pos_dependent(dN));
        d1->add_scale_dep(P::Pos_dependent(dN));

        d1->add_cond(P::Pointer(h));
        d1->has_ratio_prior(true);
        dh->add_cond(P::Pointer(h));
        dh->has_ratio_prior(true);
    }
    else if (N == 1)
    {
        h->add_shift_dep(P::Pos_half_dependent(dN));
        h->add_scale_dep(P::Fancy_pos_half_dependent(dN));
        a_pos_j->add_scale_dep(P::Pos_dependent(dN));
        d1->remove_shift_dep(P::Pos_dependent(dN));
        d1->remove_scale_dep(P::Pos_dependent(dN));

        dN->add_cond(P::Pointer(h));
        dN->has_ratio_prior(true);
    }
}

void Block_set::remove_cross_block()
{
    typedef Parameter P;

    uint32_t N = num_cross_blocks();
    assert(N > 0);

    try { _cross_p[0].value(N-1); }
    catch (jwscxx::base::Arg_error e) { abort(); }

    P* h       = &(_size_p[1]);
    P* d1      = &(_cross_p[1]);
    P* dh      = &(_cross_p[2]);
    P* dN      = &(_cross_p[3]);
    P* a_pos_j = &(_ctr_attach_pts_p[2][1]);

    if ((N-1) == 0)
    {
        h->remove_shift_dep(P::Pos_half_dependent(d1));
        h->remove_scale_dep(P::Fancy_pos_half_dependent(d1));
        a_pos_j->remove_scale_dep(P::Pos_dependent(d1));
        d1->remove_shift_dep(P::Pos_dependent(dN));
        d1->remove_scale_dep(P::Pos_dependent(dN));

        d1->remove_cond(P::Pointer(h));
        d1->has_ratio_prior(false);
        dh->remove_cond(P::Pointer(h));
        dh->has_ratio_prior(false);
    }
    else if ((N-1) == 1)
    {
        h->remove_shift_dep(P::Pos_half_dependent(dN));
        h->remove_scale_dep(P::Fancy_pos_half_dependent(dN));
        a_pos_j->remove_scale_dep(P::Pos_dependent(dN));
        d1->add_shift_dep(P::Pos_dependent(dN));
        d1->add_scale_dep(P::Pos_dependent(dN));

        dN->remove_cond(P::Pointer(h));
        dN->has_ratio_prior(false);
    }
}

void Block_set::attach_shift_deps(uint32_t j, uint32_t i)
{
    Abstract_block::attach_shift_deps(j, i);

    uint32_t N = num_cross_blocks();

    if (j == 2 && i == 1)
    {
        if (N > 0)
        {
            attach_shift_deps_1();
        }
        if (N > 1)
        {
            attach_shift_deps_2();
        }
    }
}

void Block_set::attach_shift_deps_1()
{
    typedef Component::Parameter P;

    P* h  = &(_size_p[1]);
    P* d1 = &(_cross_p[1]);

    h->remove_shift_dep(P::Pos_half_dependent(d1));
    h->add_shift_dep(P::Pos_dependent(d1));
}

void Block_set::attach_shift_deps_2()
{
    typedef Component::Parameter P;

    P* h  = &(_size_p[1]);
    P* dN = &(_cross_p[3]);

    h->remove_shift_dep(P::Pos_half_dependent(dN));
    h->add_shift_dep(P::Pos_dependent(dN));
}

void Block_set::attach_scale_deps(uint32_t j, uint32_t i)
{
    Abstract_block::attach_scale_deps(j, i);

    uint32_t N = num_cross_blocks();

    if (j == 2 && i == 1)
    {
        if (N > 0)
        {
            attach_scale_deps_1();
        }
        if (N > 1)
        {
            attach_scale_deps_2();
        }
    }
}           

void Block_set::attach_scale_deps_1()
{
    typedef Component::Parameter P;

    P* h  = &(_size_p[1]);
    P* d1 = &(_cross_p[1]);

    h->remove_scale_dep(P::Fancy_pos_half_dependent(d1));
    h->add_scale_dep(P::Fancy_pos_dependent(d1));
}

void Block_set::attach_scale_deps_2()
{
    typedef Component::Parameter P;

    P* h  = &(_size_p[1]);
    P* dN = &(_cross_p[3]);

    h->remove_scale_dep(P::Fancy_pos_half_dependent(dN));
    h->add_scale_dep(P::Fancy_pos_dependent(dN));
}

void Block_set::detach_shift_deps(uint32_t j, uint32_t i)
{
    Abstract_block::detach_shift_deps(j, i);

    uint32_t N = num_cross_blocks();

    if (j == 2 && i == 1)
    {
        if (N > 0)
        {
            detach_shift_deps_1();
        }
        if (N > 1)
        {
            detach_shift_deps_2();
        }
    }
}

void Block_set::detach_shift_deps_1()
{
    typedef Component::Parameter P;

    P* h  = &(_size_p[1]);
    P* d1 = &(_cross_p[1]);

    h->add_shift_dep(P::Pos_half_dependent(d1));
    h->remove_shift_dep(P::Pos_dependent(d1));
}

void Block_set::detach_shift_deps_2()
{
    typedef Component::Parameter P;

    P* h  = &(_size_p[1]);
    P* dN = &(_cross_p[3]);

    h->add_shift_dep(P::Pos_half_dependent(dN));
    h->remove_shift_dep(P::Pos_dependent(dN));
}

void Block_set::detach_scale_deps(uint32_t j, uint32_t i)
{
    Abstract_block::detach_scale_deps(j, i);

    uint32_t N = num_cross_blocks();

    if (j == 2 && i == 1)
    {
        if (N > 0)
        {
            detach_scale_deps_1();
        }
        if (N > 1)
        {
            detach_scale_deps_2();
        }
    }
}

void Block_set::detach_scale_deps_1()
{
    typedef Component::Parameter P;

    P* h  = &(_size_p[1]);
    P* d1 = &(_cross_p[1]);

    h->add_scale_dep(P::Fancy_pos_half_dependent(d1));
    h->remove_scale_dep(P::Fancy_pos_dependent(d1));
}

void Block_set::detach_scale_deps_2()
{
    typedef Component::Parameter P;

    P* h  = &(_size_p[1]);
    P* dN = &(_cross_p[3]);

    h->add_scale_dep(P::Fancy_pos_half_dependent(dN));
    h->remove_scale_dep(P::Fancy_pos_dependent(dN));
}

/** The Block_set must be in object coordinates. */
void Block_set::assert_consistency()
{
    using namespace jwsc;

    Abstract_block::assert_consistency();

    Vector_d* bp = 0;

    for (size_t i = 0; i < _inner_cuboids.size(); i++)
    {
        inner_block_posn_v(&bp, i);
        jwscxx::graphics::Cuboid_d& cuboid = _inner_cuboids[ i ];

        for (size_t j = 0; j < 3; j++)
        {
            assert(fabs(bp->elts[ j ] - cuboid.get_center()->elts[ j ]) 
                    < 1.0e-2);
        }
    }

    for (size_t i = 0; i < _cross_cuboids.size(); i++)
    {
        cross_block_posn_v(&bp, i);
        jwscxx::graphics::Cuboid_d& cuboid = _cross_cuboids[ i ];

        for (size_t j = 0; j < 3; j++)
        {
            assert(fabs(bp->elts[ j ] - cuboid.get_center()->elts[ j ]) 
                    < 1.0e-2);
        }
    }

    free_vector_d(bp);
}

void Block_set::ensure_consistency()
{
    using namespace jwsc;

    Abstract_block::ensure_consistency();

    Vector_d* bp = 0;
    Vector_d* t  = 0;
    Matrix_d* M  = 0;

    for (size_t i = 0; i < _inner_cuboids.size(); i++)
    {
        inner_block_posn_v(&bp, i);
        jwscxx::graphics::Cuboid_d& cuboid = _inner_cuboids[ i ];

        for (size_t j = 0; j < 3; j++)
        {
            create_zero_vector_d(&t, 3);
            t->elts[ j ] = bp->elts[ j ] - cuboid.get_center()->elts[ j ];

            create_3d_homo_translation_matrix_1_d(&M, t);
            cuboid.transform(M);

            assert(fabs(bp->elts[ j ] - cuboid.get_center()->elts[ j ]) 
                    < 1.0e-2);
        }
    }

    for (size_t i = 0; i < _cross_cuboids.size(); i++)
    {
        cross_block_posn_v(&bp, i);
        jwscxx::graphics::Cuboid_d& cuboid = _cross_cuboids[ i ];

        for (size_t j = 0; j < 3; j++)
        {
            create_zero_vector_d(&t, 3);
            t->elts[ j ] = bp->elts[ j ] - cuboid.get_center()->elts[ j ];

            create_3d_homo_translation_matrix_1_d(&M, t);
            cuboid.transform(M);

            assert(fabs(bp->elts[ j ] - cuboid.get_center()->elts[ j ]) 
                    < 1.0e-2);
        }
    }

    free_vector_d(bp);
    free_matrix_d(M);
    free_vector_d(t);
}

/**
 * A really small (negative) log probability is returned if the center of the
 * inner blocks is not within the @em b_j face of the block @em b.
 */
double Block_set::calc_inner_block_ctr_attach_pt_likelihood
(
    uint32_t              i,
    uint32_t              a_j,
    const Abstract_block* b, 
    uint32_t              b_j
) 
const
{
    using namespace jwsc;

    double log_prob = 0;

    Vector_d* v = 0;
    inner_block_ctr_attach_pt_v(&v, i, a_j);

    if (b->object()) b->object()->transform_pt_from_world_to_object_coords(v);
    b->transform_pt_from_object_to_component_coords(v);
    b->transform_pt_from_block_to_face_coords(v, b_j);

    const Paramgroup& a_p = _ctr_attach_pts_p[ a_j ];
    for (size_t xyz = 0; xyz < 3; xyz++)
    {
        size_t uv = attach_pt_face_coord(a_j, xyz);
        if (uv == 2) continue;

        assert(a_p[ xyz ].cond(0));
        double min = -0.5 * a_p[ xyz ].cond(0)->ptr()->value();
        double max =  0.5 * a_p[ xyz ].cond(0)->ptr()->value();
        double x   = v->elts[ uv ];

        if (x < min || x > max)
        {
            double y = (x < min) ? fabs(x - min) : fabs(x - max);
            y /= (max - min);
            log_prob += (y+1)*10*log(JWSC_MIN_LOG_ARG);
        }
    }
    free_vector_d(v);

    return log_prob;
}

/**
 * A really small (negative) log probability is returned if the space between
 * cross blocks is <= 0 or if the stacked cross blocks extend outside the block
 * set height.
 */
double Block_set::calc_cross_block_likelihood(uint32_t i) const
{
    using namespace jwsc;

    double log_prob = 0;

    uint32_t N = num_cross_blocks();
    assert(N > 0 && N > i);

    double h  = cross(2);
    double d1 = cross(1);
    double dN = cross(3);
    double s  = (N > 1) ? (dN - d1 - N*h) / (N - 1) : 0;

    if (s < 0)
    {
        log_prob += (fabs(s)+1)*10*log(JWSC_MIN_LOG_ARG);
    }

    double x = N*h + (N-1)*s;
    double x_cond = cross_p(1)->cond(0)->ptr()->value();
    double r = x / x_cond;

    if (r > 1)
    {
        log_prob += (fabs(x - x_cond)+1)*10*log(JWSC_MIN_LOG_ARG);
    }

    return log_prob;
}

double Block_set::calc_var_param_log_likelihood() const
{
    using namespace jwsc;

    double log_prob = 0;

    log_prob += Abstract_block::calc_var_param_log_likelihood();

    uint32_t D = num_var_params();
    uint32_t d = var_param_inner_size_i(0);
    uint32_t N_i = _inner_size_p.num_params();

    for (uint32_t i = 0; i < N_i; i++)
    {
        double x      = inner_size_p(i)->value();
        double x_cond = inner_size_p(i)->cond(0)->ptr()->value();
        double min    = 0;
        double max    = 0.5*x_cond;

        log_prob += inner_size_p(i)->context()->calc_trunc_log_likelihood(x,
                x_cond, min, max);
        d++;
    }

    assert(d <= D);

    if ((N_i = num_var_param_cross_p()))
    {
        log_prob += cross_p(0)->context()->calc_log_likelihood(cross(0));
        d = var_param_cross_i(1);

        for (uint32_t i = 1; i < N_i; i++)
        {
            double x      = cross_p(i)->value();
            double x_cond = cross_p(i)->cond(0)->ptr()->value();
            double min    = 0;
            double max    = x_cond;

            log_prob += cross_p(i)->context()->calc_trunc_log_likelihood(x,
                    x_cond, min, max);
            d++;
        }

        assert(d <= D);
    }

    for (size_t j = 0; j < 6; j++)
    {
        Abstract_block* b;
        if ((b = dynamic_cast<Abstract_block*>(attachment(j))))
        {
            size_t b_j = (j % 2 == 0) ? j+1 : j-1;
            for (uint32_t i = 0; i < _inner_cuboids.size(); i++)
            {
                log_prob += calc_inner_block_ctr_attach_pt_likelihood(i, j, 
                        b, b_j);
            }
        }
    }

    for (uint32_t i = 0; i < num_cross_blocks(); i++)
    {
        log_prob += calc_cross_block_likelihood(i);
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

/** 
 * Same as the log likelihood calculation, but doesn't include a penalty for
 * center points of inner blocks falling outside of attaching blocks.
 */
double Block_set::calc_var_param_proposal_log_likelihood() const
{
    using namespace jwsc;

    double log_prob = 0;
    double epsilon = 1.0e-8;

    log_prob += Abstract_block::calc_var_param_proposal_log_likelihood();

    uint32_t D   = num_var_params();
    uint32_t d   = var_param_inner_size_i(0);
    uint32_t N_i = _inner_size_p.num_params();

    for (uint32_t i = 0; i < N_i; i++)
    {
        double x      = inner_size_p(i)->value();
        double x_cond = inner_size_p(i)->cond(0)->ptr()->value();
        double min    = 0 + epsilon;
        double max    = 0.5*x_cond - epsilon;
        assert(min <= max);

        // If x is outside [min+e,max-e], it means x was truncated, we don't
        // propose that.
        if (x > min && x < max)
        {
            log_prob += inner_size_p(i)->context()->calc_log_likelihood(x, 
                    x_cond);
        }
        d++;
    }

    assert(d <= D);

    if ((N_i = num_var_param_cross_p()))
    {
        log_prob += cross_p(0)->context()->calc_log_likelihood(cross(0));
        d = var_param_cross_i(1);

        for (uint32_t i = 1; i < N_i; i++)
        {
            double x      = cross_p(i)->value();
            double x_cond = cross_p(i)->cond(0)->ptr()->value();
            double min    = 0 + epsilon;
            double max    = x_cond - epsilon;
            assert(min <= max);

            // If x is outside [min+e,max-e], it means x was truncated, we don't
            // propose that.
            if (x > min && x < max)
            {
                log_prob += cross_p(i)->context()->calc_log_likelihood(x,
                        x_cond);
            }
            d++;
        }

        assert(d <= D);
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

void Block_set::sample_var_param_likelihood(jwsc::Vector_d** p_out)
{
    using namespace jwsc;

    Abstract_block::sample_var_param_likelihood(p_out);
    Vector_d* p = *p_out;

    for (size_t j = 0; j < 6; j++)
    {
        Abstract_block* b;
        if ((b = dynamic_cast<Abstract_block*>(attachment(j))))
        {
            for (size_t i = 0; i < 3; i++)
            {
                if (i == j/2) continue;

                double min = 0;
                double max = b->size(i);
                if (size_p(i)->num_conds() == 1)
                {
                    double x_cond = size_p(i)->cond(0)->ptr()->value();
                    p->elts[ i ] = size_p(i)->context()->
                        sample_trunc_likelihood(x_cond, min, max);
                }
                else
                {
                    p->elts[ i ] = size_p(i)->context()->
                        sample_trunc_likelihood(min, max);
                }
            }
        }
    }

    uint32_t D   = num_var_params();
    uint32_t d   = var_param_inner_size_i(0);
    uint32_t N_i = _inner_size_p.num_params();

    for (uint32_t i = 0; i < N_i; i++)
    {
        double x_cond = inner_size_p(i)->cond(0)->ptr()->value();
        double min    = 0;
        double max    = 0.5*x_cond;

        p->elts[ d++ ] = inner_size_p(i)->context()->sample_trunc_likelihood(
                x_cond, min, max);
    }

    assert(d <= D);

    if ((N_i = num_var_param_cross_p()))
    {
        d = var_param_cross_i(0);
        p->elts[ d++ ] = cross_p(0)->context()->sample_likelihood();

        for (uint32_t i = 1; i < N_i; i++)
        {
            double x_cond = cross_p(i)->cond(0)->ptr()->value();
            double min    = 0;
            double max    = x_cond;

            p->elts[ d++ ] = cross_p(i)->context()->sample_trunc_likelihood(
                    x_cond, min, max);
        }

        assert(d <= D);
    }
}

void Block_set::gen_var_param_conditional_likelihood_sample()
{
    using namespace jwsc;

    Abstract_block::gen_var_param_conditional_likelihood_sample();

    uint32_t N_i = _inner_size_p.num_params();

    for (uint32_t i = 0; i < N_i; i++)
    {
        double x_cond = inner_size_p(i)->cond(0)->ptr()->scratch();
        double min    = 0;
        double max    = 0.5*x_cond;

        double x = inner_size_p(i)->context()->sample_trunc_likelihood(x_cond,
                min, max);
        _inner_size_p[ i ].scratch(x);
    }

    if ((N_i = num_var_param_cross_p()))
    {
        double x = cross_p(0)->context()->sample_likelihood();
        _cross_p[0].scratch(x);

        for (uint32_t i = 1; i < N_i; i++)
        {
            double x_cond = cross_p(i)->cond(0)->ptr()->scratch();
            double min    = 0;
            double max    = x_cond;

            double x = cross_p(i)->context()->sample_trunc_likelihood(x_cond,
                    min, max);
            _cross_p[ i ].scratch(x);
        }
    }
}

void Block_set::get_var_param_conditional_likelihood_sample
(
    jwsc::Vector_d** p_out
)
{
    using namespace jwsc;

    Abstract_block::get_var_param_conditional_likelihood_sample(p_out);
    Vector_d* p = *p_out;

    uint32_t D   = num_var_params();
    uint32_t d   = var_param_inner_size_i(0);
    uint32_t N_i = _inner_size_p.num_params();

    for (uint32_t i = 0; i < N_i; i++)
    {
        p->elts[ d++ ] = inner_size_p(i)->scratch();
    }

    assert(d <= D);

    if ((N_i = num_var_param_cross_p()))
    {
        d = var_param_cross_i(0);

        for (uint32_t i = 0; i < N_i; i++)
        {
            p->elts[ d++ ] = cross_p(i)->scratch();
        }

        assert(d <= D);
    }
}

double Block_set::calc_var_param_prior_log_prob() const
{
    using namespace jwsc;

    double   log_prob = 0;

    log_prob += Abstract_block::calc_var_param_prior_log_prob();

    uint32_t D   = num_var_params();
    uint32_t d   = var_param_inner_size_i(0);
    uint32_t N_i = _inner_size_p.num_params();

    for (uint32_t i = 0; i < N_i; i++)
    {
        double mean = inner_size_p(i)->mean();
        double stdev = inner_size_p(i)->stdev();
        log_prob += inner_size_p(i)->context()->calc_prior_log_prob(mean,stdev);
        d++;
    }

    assert(d <= D);

    if ((N_i = num_var_param_cross_p()))
    {
        double mean = cross_p(0)->mean();
        double stdev = cross_p(0)->stdev();
        log_prob += cross_p(0)->context()->calc_prior_log_prob(mean, stdev);
        d = var_param_cross_i(1);

        for (uint32_t i = 1; i < N_i; i++)
        {
            mean = cross_p(i)->mean();
            stdev = cross_p(i)->stdev();
            log_prob += cross_p(i)->context()->calc_prior_log_prob(mean, stdev);
            d++;
        }

        assert(d <= D);
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

void Block_set::sample_var_param_prior
(
    jwsc::Vector_d** means_out,
    jwsc::Vector_d** stdevs_out
) 
const
{
    using namespace jwsc;

    Abstract_block::sample_var_param_prior(means_out, stdevs_out);
    Vector_d* means = *means_out;
    Vector_d* stdevs = *stdevs_out;

    uint32_t D = num_var_params();
    uint32_t d = var_param_inner_size_i(0);
    uint32_t N_i = _inner_size_p.num_params();

    for (uint32_t i = 0; i < N_i; i++)
    {
        double mean, stdev;
        inner_size_p(i)->context()->sample_prior(&mean, &stdev);
        means->elts [ d   ] = mean;
        stdevs->elts[ d++ ] = stdev;
    }

    assert(d <= D);

    if ((N_i = num_var_param_cross_p()))
    {
        d = var_param_cross_i(0);

        for (uint32_t i = 0; i < N_i; i++)
        {
            double mean, stdev;
            cross_p(i)->context()->sample_prior(&mean, &stdev);
            means->elts [ d   ] = mean;
            stdevs->elts[ d++ ] = stdev;
        }

        assert(d <= D);
    }
}

void Block_set::sample_var_param_cat_proposal
(
    jwsc::Vector_d** means_out,
    jwsc::Vector_d** stdevs_out,
    const jwsc::Vector_d* means_in
) 
const
{
    using namespace jwsc;

    Abstract_block::sample_var_param_cat_proposal(means_out, stdevs_out, 
            means_in);
    Vector_d* means = *means_out;
    Vector_d* stdevs = *stdevs_out;

    uint32_t D   = num_var_params();
    uint32_t d   = var_param_inner_size_i(0);
    uint32_t N_i = _inner_size_p.num_params();

    for (uint32_t i = 0; i < N_i; i++)
    {
        double mean, stdev;
        double mean_in = means_in->elts[ d ];
        inner_size_p(i)->context()->sample_prior(&mean, &stdev, mean_in);
        means->elts [ d ] = mean;
        stdevs->elts[ d ] = stdev;
        d++;
    }

    assert(d <= D);

    if ((N_i = num_var_param_cross_p()))
    {
        d = var_param_cross_i(0);

        for (uint32_t i = 0; i < N_i; i++)
        {
            double mean, stdev;
            double mean_in = means_in->elts[ d ];
            cross_p(i)->context()->sample_prior(&mean, &stdev, mean_in);
            means->elts [ d ] = mean;
            stdevs->elts[ d ] = stdev;
            d++;
        }

        assert(d <= D);
    }
}

double Block_set::calc_var_param_proposal_log_prob() const
{
    using namespace jwsc;

    double log_prob = 0;

    log_prob += Abstract_block::calc_var_param_proposal_log_prob();

    uint32_t D   = num_var_params();
    uint32_t d   = var_param_inner_size_i(0);
    uint32_t N_i = _inner_size_p.num_params();

    for (uint32_t i = 0; i < N_i; i++)
    {
        double x  = inner_size_p(i)->value();
        log_prob += inner_size_p(i)->context()->calc_proposal_log_prob(x);
        d++;
    }

    assert(d <= D);

    if ((N_i = num_var_param_cross_p()))
    {
        d = var_param_cross_i(0);

        for (uint32_t i = 0; i < N_i; i++)
        {
            double x  = cross_p(i)->value();
            log_prob += cross_p(i)->context()->calc_proposal_log_prob(x);
            d++;
        }

        assert(d <= D);
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

double Block_set::calc_var_param_gauss_proposal_log_prob() const
{
    using namespace jwsc;

    double log_prob = 0;

    log_prob += Abstract_block::calc_var_param_gauss_proposal_log_prob();

    uint32_t D   = num_var_params();
    uint32_t d   = var_param_inner_size_i(0);
    uint32_t N_i = _inner_size_p.num_params();

    for (uint32_t i = 0; i < N_i; i++)
    {
        double x  = inner_size_p(i)->value();
        log_prob += inner_size_p(i)->context()->calc_gauss_proposal_log_prob(x);
        d++;
    }

    assert(d <= D);

    if ((N_i = num_var_param_cross_p()))
    {
        d = var_param_cross_i(0);

        for (uint32_t i = 0; i < N_i; i++)
        {
            double x  = cross_p(i)->value();
            log_prob += cross_p(i)->context()->calc_gauss_proposal_log_prob(x);
            d++;
        }

        assert(d <= D);
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

double Block_set::calc_var_param_split_merge_proposal_log_prob
(
    const jwsc::Vector_d* p,
    const jwsc::Vector_d* means,
    uint32_t              j
)
const
{
    using namespace jwsc;

    double log_prob = 0;

    log_prob += Abstract_block::calc_var_param_split_merge_proposal_log_prob(
            p, means, j);

    uint32_t N_i = _inner_size_p.num_params();
    uint32_t p_d = var_param_inner_size_i(0);
    uint32_t m_d = p_d;

    if (! _attachments[ j ].is_variable())
    {
        p_d += 2;
    }

    for (uint32_t i = 0; i < N_i; i++)
    {
        double m = inner_size_p(i)->context()->gauss_proposal().mu();
        double x = p->elts[ p_d++ ] + (m - means->elts[ m_d++ ]);
        log_prob += inner_size_p(i)->context()->calc_gauss_proposal_log_prob(x);
    }

    assert(p_d <= p->num_elts);
    assert(m_d <= means->num_elts);

    if ((N_i = num_var_param_cross_p()))
    {
        p_d = var_param_cross_i(0);
        m_d = p_d;

        if (! _attachments[ j ].is_variable())
        {
            p_d += 2;
        }

        for (uint32_t i = 0; i < N_i; i++)
        {
            double m = cross_p(i)->context()->gauss_proposal().mu();
            double x = p->elts[ p_d++ ] + (m - means->elts[ m_d++ ]);
            log_prob += cross_p(i)->context()->calc_gauss_proposal_log_prob(x);
        }

        assert(p_d <= p->num_elts);
        assert(m_d <= means->num_elts);
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

double Block_set::calc_var_param_switch_proposal_log_prob
(
    const jwsc::Vector_d* p
)
const
{
    using namespace jwsc;

    double log_prob = 0;

    log_prob += Abstract_block::calc_var_param_switch_proposal_log_prob(p);

    uint32_t N_i = _inner_size_p.num_params();
    uint32_t p_d = var_param_inner_size_i(0);

    for (uint32_t i = 0; i < N_i; i++)
    {
        double x = p->elts[ p_d++ ];
        double x_cond = inner_size_p(i)->cond(0)->ptr()->value();
        log_prob += inner_size_p(i)->context()->calc_log_likelihood(x, x_cond);
    }

    assert(p_d <= p->num_elts);

    if ((N_i = num_var_param_cross_p()))
    {
        p_d = var_param_cross_i(1);

        for (uint32_t i = 1; i < N_i; i++)
        {
            double x = p->elts[ p_d++ ];
            double x_cond = cross_p(i)->cond(0)->ptr()->value();
            log_prob += cross_p(i)->context()->calc_log_likelihood(x, x_cond);
        }

        assert(p_d <= p->num_elts);
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

double Block_set::calc_var_param_cross_proposal_log_prob
(
    const jwsc::Vector_d* p
)
const
{
    using namespace jwsc;

    double log_prob = 0;
    uint32_t N_i;

    if ((N_i = num_var_param_cross_p()))
    {
        log_prob += cross_p(0)->context()->calc_log_likelihood(cross(0));
        uint32_t p_d = var_param_cross_i(1);

        for (uint32_t i = 1; i < N_i; i++)
        {
            double x = p->elts[ p_d++ ];
            double x_cond = cross_p(i)->cond(0)->ptr()->value();
            log_prob += cross_p(i)->context()->calc_log_likelihood(x, x_cond);
        }

        assert(p_d <= p->num_elts);
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

void Block_set::sample_var_param_proposal(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    Abstract_block::sample_var_param_proposal(p_out);
    Vector_d* p = *p_out;

    uint32_t D   = num_var_params();
    uint32_t d   = var_param_inner_size_i(0);
    uint32_t N_i = _inner_size_p.num_params();

    for (uint32_t i = 0; i < N_i; i++)
    {
        p->elts[ d++ ] = inner_size_p(i)->context()->sample_proposal();
    }

    assert(d <= D);

    if ((N_i = num_var_param_cross_p()))
    {
        d = var_param_cross_i(0);

        for (uint32_t i = 0; i < N_i; i++)
        {
            p->elts[ d++ ] = inner_size_p(i)->context()->sample_proposal();
        }

        assert(d <= D);
    }
}

void Block_set::sample_var_param_gauss_proposal
(
    jwsc::Vector_d**      p_out,
    const jwsc::Vector_d* means
) 
const
{
    using namespace jwsc;

    uint32_t D = num_var_params();
    assert(D == means->num_elts);

    Abstract_block::sample_var_param_gauss_proposal(p_out, means);
    Vector_d* p = *p_out;

    uint32_t d = var_param_inner_size_i(0);
    uint32_t N_i = _inner_size_p.num_params();

    for (uint32_t i = 0; i < N_i; i++)
    {
        double m = means->elts[ d ];
        p->elts[ d++ ] = inner_size_p(i)->context()->sample_gauss_proposal(m);
    }

    if ((N_i = num_var_param_cross_p()))
    {
        d = var_param_cross_i(0);

        for (size_t i = 0; i < N_i; i++)
        {
            double m = means->elts[ d ];
            p->elts[ d++ ] = cross_p(i)->context()->sample_gauss_proposal(m);
        }
    }
}

void Block_set::sample_var_param_split_merge_proposal
(
    jwsc::Vector_d**      p_out,
    const jwsc::Vector_d* means,
    uint32_t              j
) 
const
{
    using namespace jwsc;

    Abstract_block::sample_var_param_split_merge_proposal(p_out, means, j);
    Vector_d* p = *p_out;

    uint32_t N_i = _inner_size_p.num_params();
    uint32_t p_d = var_param_inner_size_i(0);
    uint32_t m_d = p_d;

    if (! _attachments[ j ].is_variable())
    {
        p_d += 2;
    }

    for (size_t i = 0; i < N_i; i++)
    {
        double m = means->elts[ m_d++ ];
        p->elts[ p_d++ ] = inner_size_p(i)->context()->sample_gauss_proposal(m);
    }

    assert(p_d <= p->num_elts);
    assert(m_d <= means->num_elts);

    if ((N_i = num_var_param_cross_p()))
    {
        p_d = var_param_cross_i(0);
        m_d = p_d;

        if (! _attachments[ j ].is_variable())
        {
            p_d += 2;
        }

        for (size_t i = 0; i < N_i; i++)
        {
            double m = means->elts[ m_d++ ];
            p->elts[ p_d++ ] = cross_p(i)->context()->sample_gauss_proposal(m);
        }

        assert(p_d <= p->num_elts);
        assert(m_d <= means->num_elts);
    }
}

void Block_set::sample_var_param_switch_proposal
(
    jwsc::Vector_d**      p_out,
    const jwsc::Vector_d* means
) 
const
{
    using namespace jwsc;

    uint32_t D   = num_var_params();

    Vector_d* q = 0;
    Abstract_block::sample_var_param_switch_proposal(&q, means);

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    copy_vector_section_into_vector_d(p, 0, q, 0, 3);
    free_vector_d(q);

    uint32_t p_d = var_param_inner_size_i(0);
    uint32_t N_i = _inner_size_p.num_params();

    for (uint32_t i = 0; i < N_i; i++)
    {
        p->elts[ p_d++ ] = inner_size_p(i)->context()->sample_likelihood();
    }

    assert(p_d <= p->num_elts);

    if ((N_i = num_var_param_cross_p()))
    {
        p_d = var_param_cross_i(0);

        for (uint32_t i = 0; i < N_i; i++)
        {
            p->elts[ p_d++ ] = cross(i);
        }

        assert(p_d <= p->num_elts);
    }
}

void Block_set::sample_var_param_cross_proposal
(
    jwsc::Vector_d** p_out,
    bool             one_cross
) 
const
{
    using namespace jwsc;

    get_var_param_values(p_out);
    Vector_d* p = *p_out;

    uint32_t N_i;

    if ((N_i = num_var_param_cross_p()))
    {
        uint32_t p_d = var_param_cross_i(0);
        p->elts[ p_d ] = floor(cross_p(0)->context()->sample_likelihood()+0.5);
        if (one_cross)
        {
            p->elts[ p_d ] = 1;
        }
        else if (p->elts[ p_d ] <= 1)
        {
            p->elts[ p_d ] = 2;
        }

        double x_cond;
        double min, max;
        double d1, dN;

        x_cond = cross_p(1)->cond(0)->ptr()->value();
        min = 0;
        max = x_cond;
        d1 = p->elts[ p_d+1 ] = cross_p(1)->context()->
                        sample_trunc_likelihood(x_cond, min, max);
        dN = x_cond - d1;

        if (N_i > 3)
        {
            x_cond = cross_p(3)->cond(0)->ptr()->value();
            min = d1;
            max = x_cond;
            dN = p->elts[ p_d+3 ] = cross_p(3)->context()->
                sample_trunc_likelihood(x_cond, min, max);
        }

        x_cond = cross_p(2)->cond(0)->ptr()->value();
        min = cross_p(2)->context()->range().min();
        max = (dN - d1) / num_cross_blocks();
        p->elts[ p_d+2 ] = cross_p(2)->context()->
            sample_trunc_likelihood(x_cond, min, (max < min) ? min : max);

        assert(p_d <= p->num_elts);
    }
}

void Block_set::translate(double x, double y, double z)
{
    using namespace jwsc;

    Abstract_block::translate(x, y, z);

    Matrix_d* M = 0;

    assert(create_3d_homo_translation_matrix_2_d(&M, x, y, z) == 0);

    for (size_t i = 0; i < _inner_cuboids.size(); i++)
    {
        _inner_cuboids[ i ].transform(M);
    }

    for (size_t i = 0; i < _cross_cuboids.size(); i++)
    {
        _cross_cuboids[ i ].transform(M);
    }

    free_matrix_d(M);
}

void Block_set::rotate(double phi, double x, double y, double z)
{
    using namespace jwsc;

    Abstract_block::rotate(phi, x, y, z);

    Matrix_d* M = 0;

    assert(create_3d_homo_rotation_matrix_2_d(&M, phi, x, y, z) == 0);

    for (size_t i = 0; i < _inner_cuboids.size(); i++)
    {
        _inner_cuboids[ i ].transform(M);
    }

    for (size_t i = 0; i < _cross_cuboids.size(); i++)
    {
        _cross_cuboids[ i ].transform(M);
    }

    free_matrix_d(M);
}

bool Block_set::conditions(const Model* m) const
{
    if (Abstract_block::conditions(m))
        return true;

    uint32_t N_i = _inner_size_p.num_params();

    for (uint32_t i = 0; i < N_i; i++)
    {
        const Parameter& p = _inner_size_p[ i ];
        for (uint32_t j = 0; j < p.num_conds(); j++)
        {
            if (m == p.cond(j)->ptr()->model())
            {
                return true;
            }
        }
    }

    N_i = _cross_p.num_params();

    for (uint32_t i = 0; i < N_i; i++)
    {
        const Parameter& p = _cross_p[ i ];
        for (uint32_t j = 0; j < p.num_conds(); j++)
        {
            if (m == p.cond(j)->ptr()->model())
            {
                return true;
            }
        }
    }

    return false;
}

std::vector<const jwscxx::graphics::Polygon_d*> Block_set::faces() const
{
    using jwscxx::graphics::Polygon_d;

    std::vector<const Polygon_d*> faces;

    for (size_t i = 0; i < _inner_cuboids.size(); i++)
    {
        std::vector<const Polygon_d*> i_faces(_inner_cuboids[ i ].get_faces());
        faces.insert(faces.end(), i_faces.begin(), i_faces.end());
    }

    for (size_t i = 0; i < _cross_cuboids.size(); i++)
    {
        std::vector<const Polygon_d*> i_faces(_cross_cuboids[ i ].get_faces());
        faces.insert(faces.end(), i_faces.begin(), i_faces.end());
    }

    return faces;
}

/** The faces for the hull of this block set are not included. */
size_t Block_set::num_faces() const
{
    size_t N = 0;

    for (size_t i = 0; i < _inner_cuboids.size(); i++)
    {
        N += _inner_cuboids[ i ].num_faces();
    }

    for (size_t i = 0; i < _cross_cuboids.size(); i++)
    {
        N += _cross_cuboids[ i ].num_faces();
    }

    return N;
}

/** The faces for the hull of this block set are not included. */
uint32_t Block_set::set_sequential_face_ids(uint32_t id)
{
    try
    {
        for (size_t i = 0; i < _inner_cuboids.size(); i++)
        {
            for (size_t j = 0; j < _inner_cuboids[ i ].num_faces(); j++)
            {
                _inner_cuboids[ i ].set_face_id(j, id++);
            }
        }

        for (size_t i = 0; i < _cross_cuboids.size(); i++)
        {
            for (size_t j = 0; j < _cross_cuboids[ i ].num_faces(); j++)
            {
                _cross_cuboids[ i ].set_face_id(j, id++);
            }
        }
    }
    catch (jwscxx::base::Arg_error e)
    {
        // Should never happen.
        abort();
    }

    return id;
}

/** The block must be in object coordinates. */
void Block_set::update_graphical_rep()
{
    using namespace jwsc;

    Abstract_block::update_graphical_rep();

    double t_x, t_y, t_z;
    double s_x, s_y, s_z;

    Matrix_d* M = 0;
    Matrix_d* T = 0;
    Matrix_d* S = 0;
    Vector_d* p = 0;

    for (uint32_t i = 0; i < _inner_cuboids.size(); i++)
    {
        jwscxx::graphics::Cuboid_d& cuboid = _inner_cuboids[ i ];

        create_identity_matrix_d(&M, 4);

        t_x = cuboid.get_center()->elts[0];
        t_y = cuboid.get_center()->elts[1];
        t_z = cuboid.get_center()->elts[2];

        create_3d_homo_translation_matrix_2_d(&T, -t_x, -t_y, -t_z);
        multiply_matrices_d(&M, T, M);

        cuboid.transform(M);

        s_x = cuboid.get_point(1)->elts[0] - cuboid.get_point(0)->elts[0];
        s_y = cuboid.get_point(2)->elts[1] - cuboid.get_point(1)->elts[1];
        s_z = cuboid.get_point(3)->elts[2] - cuboid.get_point(2)->elts[2];
        assert(s_x > 0 && s_y > 0 && s_z > 0);

        s_x = inner_block_size(0) / s_x;
        s_y = inner_block_size(1) / s_y;
        s_z = inner_block_size(2) / s_z;

        create_identity_matrix_d(&M, 4);
        create_3d_homo_scaling_matrix_2_d(&S, s_x, s_y, s_z);
        multiply_matrices_d(&M, S, M);

        inner_block_posn_v(&p, i);
        t_x = p->elts[0];
        t_y = p->elts[1];
        t_z = p->elts[2];

        create_3d_homo_translation_matrix_2_d(&T, t_x, t_y, t_z);
        multiply_matrices_d(&M, T, M);

        cuboid.transform(M);
    }

    for (uint32_t i = 0; i < _cross_cuboids.size(); i++)
    {
        jwscxx::graphics::Cuboid_d& cuboid = _cross_cuboids[ i ];

        create_identity_matrix_d(&M, 4);

        t_x = cuboid.get_center()->elts[0];
        t_y = cuboid.get_center()->elts[1];
        t_z = cuboid.get_center()->elts[2];

        create_3d_homo_translation_matrix_2_d(&T, -t_x, -t_y, -t_z);
        multiply_matrices_d(&M, T, M);

        cuboid.transform(M);

        s_x = cuboid.get_point(1)->elts[0] - cuboid.get_point(0)->elts[0];
        s_y = cuboid.get_point(2)->elts[1] - cuboid.get_point(1)->elts[1];
        s_z = cuboid.get_point(3)->elts[2] - cuboid.get_point(2)->elts[2];
        assert(s_x > 0 && s_y > 0 && s_z > 0);

        s_x = cross_block_size(i, 0) / s_x;
        s_y = cross_block_size(i, 1) / s_y;
        s_z = cross_block_size(i, 2) / s_z;

        create_identity_matrix_d(&M, 4);
        create_3d_homo_scaling_matrix_2_d(&S, s_x, s_y, s_z);
        multiply_matrices_d(&M, S, M);

        cross_block_posn_v(&p, i);
        t_x = p->elts[0];
        t_y = p->elts[1];
        t_z = p->elts[2];

        create_3d_homo_translation_matrix_2_d(&T, t_x, t_y, t_z);
        multiply_matrices_d(&M, T, M);

        cuboid.transform(M);
    }

    free_matrix_d(S);
    free_matrix_d(T);
    free_matrix_d(M);
    free_vector_d(p);
}

/** The faces for the hull of this block set are not rendered. */
void Block_set::wire_render_with_ids(const View* view) const
{
    for (size_t i = 0; i < _inner_cuboids.size(); i++)
    {
        wire_render_mesh_faces_with_ids(view, &(_inner_cuboids[ i ]));
    }

    for (size_t i = 0; i < _cross_cuboids.size(); i++)
    {
        wire_render_mesh_faces_with_ids(view, &(_cross_cuboids[ i ]));
    }
}

/** The faces for the hull of this block set are not rendered. */
void Block_set::solid_render() const
{
    for (size_t i = 0; i < _inner_cuboids.size(); i++)
    {
        solid_render_mesh_faces(&(_inner_cuboids[ i ]));
    }

    for (size_t i = 0; i < _cross_cuboids.size(); i++)
    {
        solid_render_mesh_faces(&(_cross_cuboids[ i ]));
    }
}

/**
 * @param  i  Width, height, or length face of this Block_set.
 * @param  n  Parameter index of @em _inner_size_p to have depend on the
 *            inner size.
 */
void Block_set::add_inner_individual_deps(uint32_t i, uint32_t n)
{
    /*
    typedef Parameter P;
    assert(i < 3);
    uint32_t j = i*2;

    P* is = &(_inner_size_p[ n ]);
    P* bs = &(_size_p[ i ]);
    P* a_pos_j = &(_ctr_attach_pts_p[ j+0 ][ i ]);
    P* a_neg_j = &(_ctr_attach_pts_p[ j+1 ][ i ]);

    is->add_shift_dep(P::Pos_dependent(bs));
    is->add_scale_dep(P::Pos_dependent(bs));
    is->add_scale_dep(P::Fancy_pos_half_dependent(a_pos_j));
    is->add_scale_dep(P::Fancy_neg_half_dependent(a_neg_j));
    */
}

/**
 * @param  i  Width, height, or length face of this Block_set.
 * @param  n  Parameter index of @em _inner_size_p to have depend on the
 *            inner size.
 */
void Block_set::remove_inner_individual_deps(uint32_t i, uint32_t n)
{
    /*
    typedef Parameter P;
    assert(i < 3);
    uint32_t j = i*2;

    P* is = &(_inner_size_p[ n ]);
    P* bs = &(_size_p[ i ]);
    P* a_pos_j = &(_ctr_attach_pts_p[ j+0 ][ i ]);
    P* a_neg_j = &(_ctr_attach_pts_p[ j+1 ][ i ]);

    is->remove_shift_dep(P::Pos_dependent(bs));
    is->remove_scale_dep(P::Pos_dependent(bs));
    is->remove_scale_dep(P::Fancy_pos_half_dependent(a_pos_j));
    is->remove_scale_dep(P::Fancy_neg_half_dependent(a_neg_j));
    */
}

void Block_set::add_inner_ctr_attach_conds
(
    uint32_t               a_j, 
    Component::Paramgroup& size_p
)
{
    typedef Parameter P;

    for (size_t i = 0; i < 2; i++)
    {
        size_t uv = attach_pt_face_coord(a_j, i);

        P* a_p = &(_ctr_attach_pts_p[ a_j ][ uv ]);
        P* s   = &(size_p[ uv ]);

        a_p->add_cond(P::Pointer(s));
    }
}

void Block_set::remove_inner_ctr_attach_conds
(
    uint32_t               a_j, 
    Component::Paramgroup& size_p
)
{
    typedef Parameter P;

    for (size_t i = 0; i < 2; i++)
    {
        size_t uv = attach_pt_face_coord(a_j, i);

        P* a_p = &(_ctr_attach_pts_p[ a_j ][ uv ]);
        P* s   = &(size_p[ uv ]);

        a_p->remove_cond(P::Pointer(s));
    }
}

uint32_t Block_set::attachable_cross_blocks
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Block_set* b = dynamic_cast<Block_set*>(components[ i ]);
        if (b && b->num_cross_blocks() == 0)
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Block_set::detachable_cross_blocks
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Block_set* b = dynamic_cast<Block_set*>(components[ i ]);
        if (b && b->num_cross_blocks() > 0)
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Block_set::splitable_block_set_to_block
(
    const std::vector<Component*>& components,
    bool                           height_only
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Block* b = dynamic_cast<Block*>(components[ i ]);
        if (b)
        {
            for (uint32_t j = 0; j < 6; j++)
            {
                if (height_only && !(j == 2 || j == 3))
                    continue;

                if (!dynamic_cast<Block_set*>(b->attachment(j)))
                {
                    indices.push_back(i);
                    break;
                }
            }
        }

        /*
        if (b1 && (!dynamic_cast<Block_set*>(b1->attachment(2)) ||
                   !dynamic_cast<Block_set*>(b1->attachment(3))))
        {
            indices.push_back(i);

            for (uint32_t j = 0; j < 6; j++)
            {
                if (j != 2 && j != 3 && b1->attachment(j))
                {
                    indices.pop_back();
                    break;
                }
            }
        }
        */
    }

    return Object::sample_component_index(indices);
}

uint32_t Block_set::splitable_block_set_to_block_attach_pt
(
    const Block* b,
    bool         height_only
)
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    /*
    if (!dynamic_cast<Block_set*>(b->attachment(2)))
        indices.push_back(2);
    if (!dynamic_cast<Block_set*>(b->attachment(3)))
        indices.push_back(3);

    for (uint32_t j = 0; j < 6; j++)
    {
        if (j != 2 && j != 3 && b->attachment(j))
        {
            abort();
        }
    }
    */

    for (uint32_t j = 0; j < 6; j++)
    {
        if (height_only && !(j == 2 || j == 3))
            continue;

        if (!dynamic_cast<Block_set*>(b->attachment(j)))
        {
            indices.push_back(j);
            break;
        }
    }

    return Abstract_block::sample_attach_pt_index(indices);
}

/* ========================================================================== */








/* -------------------------------  Pair  ----------------------------------- */

/**
 * The Pair is centered at (0, 0, 0).
 *
 * @param  layout  Orientation of the inner blocks, along width or length.
 * @param  ctx     Parameter context for this Block_set.
 */
Pair::Pair(Pair::Layout layout, const Block_set::Context* ctx) 
    throw (jwscxx::base::Arg_error) 
    : Block_set("Pair", ctx)
{
    using jwscxx::graphics::Cuboid_d;

    static const double s[2] = {0.5,-0.5};

    size_t id = Abstract_block::num_var_params() + 
                Abstract_block::num_derived_params() +
                Abstract_block::num_fixed_params();

    _inner_size_p.add_param(Parameter(id++, "Inner Size", 
                ctx->inner_size_c(0).init_value(),
                &(ctx->inner_size_c(0)), this));

    add_var_param(&(_inner_size_p[0]));

    for (size_t i = 0; i < 2; i++)
    {
        double w, h, l;
        double x, y, z;

        w = (layout == WIDTH) ? inner_size(0) : size(0);
        h = size(1);
        l = (layout == LENGTH) ? inner_size(0) : size(2);

        x = posn(0) + ((layout == WIDTH) ? s[i]*(size(0) - w) : 0);
        y = posn(1);
        z = posn(2) + ((layout == LENGTH) ? s[i]*(size(2) - l) : 0);

        _inner_cuboids.push_back(Cuboid_d(w, h, l, x, y, z));
    }

    _layout = layout;

    add_inner_deps();
    add_inner_conds();
}

/**
 * The Pair is centered at (0, 0, 0).
 *
 * @param  width   Width.
 * @param  height  Height.
 * @param  length  Length.
 * @param  size_i  Inner size.
 * @param  layout  Orientation of the inner blocks, along width or length.
 * @param  ctx     Parameter context for this Block_set.
 */
Pair::Pair
(
    double       width,
    double       height,
    double       length,
    double       size_i,
    Pair::Layout layout, 
    const Block_set::Context* ctx
) 
throw (jwscxx::base::Arg_error) 
    : Block_set("Pair", width, height, length, ctx)
{
    using jwscxx::graphics::Cuboid_d;

    static const double s[2] = {0.5,-0.5};

    size_t id = Abstract_block::num_var_params() + 
                Abstract_block::num_derived_params() +
                Abstract_block::num_fixed_params();

    _inner_size_p.add_param(Parameter(id++, "Inner Size", size_i,
                &(ctx->inner_size_c(0)), this));

    add_var_param(&(_inner_size_p[0]));

    for (size_t i = 0; i < 2; i++)
    {
        double w, h, l;
        double x, y, z;

        w = (layout == WIDTH) ? inner_size(0) : size(0);
        h = size(1);
        l = (layout == LENGTH) ? inner_size(0) : size(2);

        x = posn(0) + ((layout == WIDTH) ? s[i]*(size(0) - w) : 0);
        y = posn(1);
        z = posn(2) + ((layout == LENGTH) ? s[i]*(size(2) - l) : 0);

        _inner_cuboids.push_back(Cuboid_d(w, h, l, x, y, z));
    }

    _layout = layout;

    add_inner_deps();
    add_inner_conds();
}

/**
 * The Pair is centered at (0, 0, 0).
 *
 * @param  width     Width.
 * @param  height    Height.
 * @param  length    Length.
 * @param  size_i    Inner size.
 * @param  cross_N   Number of cross blocks.
 * @param  cross_d1  Distance from the top to the first cross block.
 * @param  cross_h   Height of each cross block.
 * @param  cross_dN  Distance from the top to the last cross block.
 * @param  layout    Orientation of the inner blocks, along width or length.
 * @param  ctx       Parameter context for this Block_set.
 */
Pair::Pair
(
    double       width,
    double       height,
    double       length,
    double       size_i,
    double       cross_N,
    double       cross_d1,
    double       cross_h,
    double       cross_dN,
    Pair::Layout layout, 
    const Block_set::Context* ctx
) 
throw (jwscxx::base::Arg_error) 
    : Block_set("Pair", width, height, length, cross_N, cross_d1, cross_h,
            cross_dN, ctx)
{
    using jwscxx::graphics::Cuboid_d;

    static const double s[2] = {0.5,-0.5};

    size_t id = Abstract_block::num_var_params() + 
                Abstract_block::num_derived_params() +
                Abstract_block::num_fixed_params();

    _inner_size_p.add_param(Parameter(id++, "Inner Size", size_i,
                &(ctx->inner_size_c(0)), this));

    add_var_param(&(_inner_size_p[0]));

    for (size_t i = 0; i < 2; i++)
    {
        double w, h, l;
        double x, y, z;

        w = (layout == WIDTH) ? inner_size(0) : size(0);
        h = size(1);
        l = (layout == LENGTH) ? inner_size(0) : size(2);

        x = posn(0) + ((layout == WIDTH) ? s[i]*(size(0) - w) : 0);
        y = posn(1);
        z = posn(2) + ((layout == LENGTH) ? s[i]*(size(2) - l) : 0);

        _inner_cuboids.push_back(Cuboid_d(w, h, l, x, y, z));
    }

    _layout = layout;

    add_inner_deps();
    add_inner_conds();
    add_cross_blocks(cross_N);
}

/** @param  b  Pair of blocks to copy into this one. */
Pair::Pair(const Pair& b) : Block_set(b)
{
    add_var_param(&(_inner_size_p[0]));
    *this = b;
}

/** 
 * @param  b  Block_set to copy into this one.
 *
 * @return A reference to this Block_set.
 */
Pair& Pair::operator=(const Pair& b)
{
    Block_set::operator=(b);
    _layout = b._layout;

    return *this;
}

Pair* Pair::clone() const
{
    return new Pair(*this);
}

Pair* Pair::clone_and_swap() const throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    Layout l = (layout() == WIDTH) ? LENGTH : WIDTH;
    Pair* b = new Pair(size(2), size(1), size(0), inner_size(0), cross(0),
            cross(1), cross(2), cross(3), l, context());
    b->copy_var_param_stats(this);
    b->swap_var_attach_pts(this);

    return b;
}

/** 
 * The file format is ASCII.
 *
 * @param  in  Input stream to read this Pair from.
 *
 * @throw  jwscxx::base::IO_error   Could not read from @em in.
 * @throw  jwscxx::base::Arg_error  Invalid arguments in file to read from.
 */
void Pair::read(std::istream& in) 
    throw (jwscxx::base::IO_error, jwscxx::base::Arg_error)
{
    using std::ostringstream;
    using std::istringstream;
    using jwscxx::base::Arg_error;

    const char* type_name = typeid(*this).name();
    const char* field_value;

    if (!(field_value = read_field_value(in, "Type")))
    {
        throw Arg_error("Missing Type field");
    }
    if (strncmp(field_value, type_name, strlen(type_name)) != 0)
    {
        ostringstream ost;
        ost << "Tried to read a '" << field_value << "' as a '"
            << type_name << "'";
        throw Arg_error(ost.str());
    }

    Block_set::read(in);

    istringstream ist;
    size_t layout;
    if (!(field_value = read_field_value(in, "Pair Layout")))
    {
        throw Arg_error("Missing Pair Layout field");
    }
    ist.str(field_value);
    ist >> layout;
    _layout = static_cast<Pair::Layout>(layout);
    if (ist.fail())
    {
        throw Arg_error("Invalid Pair Layout");
    }
    ist.clear(std::ios_base::goodbit);
}

/**
 * The file format is ASCII.
 *
 * @param  out  Output stream to write this Pair to.
 *
 * @throw  jwscxx::base::IO_error  Could not write to @em out.
 */
void Pair::write(std::ostream& out) const throw (jwscxx::base::IO_error)
{
    out << "Type: " << typeid(*this).name() << std::endl;
    Block_set::write(out);
    out << "Pair Layout: " << layout() << std::endl;
}

double Pair::inner_block_size(uint32_t i) const
{
    double w = (layout() == WIDTH) ? inner_size(0) : size(0);
    double h = size(1);
    double l = (layout() == LENGTH) ? inner_size(0) : size(2);

    switch (i)
    {
        case 0:
            return w;
        case 1:
            return h;
        case 2:
            return l;
        default:
            abort();
    }
}

void Pair::inner_block_posn_v(jwsc::Vector_d** bp_out, uint32_t i) const
{
    using namespace jwsc;

    static const double s[2] = {0.5,-0.5};
    assert(i < 2);

    Vector_d* p = 0;
    posn_v(&p);
    if (object() && !(object()->in_object_coords())) 
    {
        object()->transform_pt_from_world_to_object_coords(p);
    }

    double w = (layout() == WIDTH) ? inner_size(0) : size(0);
    double l = (layout() == LENGTH) ? inner_size(0) : size(2);

    copy_vector_d(bp_out, p);
    Vector_d* bp = *bp_out;
    bp->elts[0] += ((layout() == WIDTH) ? s[i]*(size(0) - w) : 0);
    bp->elts[2] += ((layout() == LENGTH) ? s[i]*(size(2) - l) : 0);

    if (object() && !(object()->in_object_coords())) 
    {
        object()->transform_pt_from_object_to_world_coords(bp);
    }
    free_vector_d(p);
}

void Pair::inner_block_ctr_attach_pt_v
(
    jwsc::Vector_d** bp_out, 
    uint32_t         i, 
    uint32_t         j
)
const
{
    using namespace jwsc;

    static const double s[2] = {0.5,-0.5};
    assert(i < 2);

    Vector_d* p = 0;
    ctr_attach_pt_v(&p, j);
    if (object() && !(object()->in_object_coords())) 
    {
        object()->transform_pt_from_world_to_object_coords(p);
    }

    double w = (layout() == WIDTH) ? inner_size(0) : size(0);
    double l = (layout() == LENGTH) ? inner_size(0) : size(2);

    copy_vector_d(bp_out, p);
    Vector_d* bp = *bp_out;
    bp->elts[0] += ((layout() == WIDTH) ? s[i]*(size(0) - w) : 0);
    bp->elts[2] += ((layout() == LENGTH) ? s[i]*(size(2) - l) : 0);

    if (object() && !(object()->in_object_coords())) 
    {
        object()->transform_pt_from_object_to_world_coords(bp);
    }
    free_vector_d(p);
}

double Pair::cross_block_size(uint32_t i, uint32_t whl) const
{
    double s;

    // All cross blocks are the same for a pair.
    switch (whl)
    {
        case 0:
            s = size(0) - ((layout() == WIDTH) ? 2*inner_size(0) : 0);
            break;
        case 1:
            s = cross(2);
            break;
        case 2:
            s = size(2) - ((layout() == LENGTH) ? 2*inner_size(0) : 0);
            break;
        default:
            abort();
    }

    return (s < 1.0) ? 1.0 : s;
}

void Pair::cross_block_posn_v(jwsc::Vector_d** bp_out, uint32_t i) const
{
    using namespace jwsc;

    uint32_t N = num_cross_blocks();
    assert(N > 0 && N > i);

    double h  = cross(2);
    double d1 = cross(1);
    double dN = cross(3);
    double s  = (N > 1) ? (dN - d1 - N*h) / (N - 1) : 0;

    ctr_attach_pt_v(bp_out, 2);
    if (object() && !(object()->in_object_coords())) 
    {
        object()->transform_pt_from_world_to_object_coords(*bp_out);
    }

    Vector_d* bp = *bp_out;
    bp->elts[1] -= d1 + 0.5*h + i*h + i*s;

    if (object() && !(object()->in_object_coords())) 
    {
        object()->transform_pt_from_object_to_world_coords(bp);
    }
}

Pair::Layout Pair::layout() const
{
    return _layout;
}

void Pair::layout(Pair::Layout layout) throw (jwscxx::base::Arg_error)
{
    using jwscxx::graphics::Cuboid_d;

    static const double s[2] = {0.5,-0.5};
    assert(layout == WIDTH || layout == LENGTH);

    if (_layout == layout)
        return;

    remove_inner_deps();
    remove_inner_conds();
    _inner_cuboids.clear();

    transform_from_world_to_object_coords();

    for (size_t i = 0; i < 2; i++)
    {
        double w, h, l;
        double x, y, z;

        w = (layout == WIDTH) ? inner_size(0) : size(0);
        h = size(1);
        l = (layout == LENGTH) ? inner_size(0) : size(2);

        x = posn(0) + ((layout == WIDTH) ? s[i]*(size(0) - w) : 0);
        y = posn(1);
        z = posn(2) + ((layout == LENGTH) ? s[i]*(size(2) - l) : 0);

        _inner_cuboids.push_back(Cuboid_d(w, h, l, x, y, z));
    }

    _layout = layout;

    graphical_rep_outdated();
    update_object();
    transform_from_object_to_world_coords();

    add_inner_deps();
    add_inner_conds();
}

void Pair::add_cross_block()
{
    using namespace jwsc;
    using jwscxx::graphics::Cuboid_d;

    Block_set::add_cross_block();

    try { _cross_p[0].value(num_cross_blocks()+1); }
    catch (jwscxx::base::Arg_error e) { abort(); }

    uint32_t N = num_cross_blocks();
    double   w = cross_block_size(N-1, 0);
    double   h = cross_block_size(N-1, 1);
    double   l = cross_block_size(N-1, 2);

    _cross_cuboids.push_back(Cuboid_d(w, h, l, 0, 0, 0));

    double t_x, t_y, t_z;
    Matrix_d* T = 0;
    Vector_d* p = 0;

    for (uint32_t i = 0; i < N; i++)
    {
        Cuboid_d& cuboid = _cross_cuboids[ i ];

        cross_block_posn_v(&p, i);
        t_x = p->elts[0] - cuboid.get_center()->elts[0];
        t_y = p->elts[1] - cuboid.get_center()->elts[1];
        t_z = p->elts[2] - cuboid.get_center()->elts[2];

        create_3d_homo_translation_matrix_2_d(&T, t_x, t_y, t_z);
        cuboid.transform(T);
    }

    free_vector_d(p);
    free_matrix_d(T);
}

void Pair::remove_cross_block()
{
    using namespace jwsc;
    using jwscxx::graphics::Cuboid_d;

    Block_set::remove_cross_block();

    _cross_cuboids.pop_back();

    double t_x, t_y, t_z;
    Matrix_d* T = 0;
    Vector_d* p = 0;

    for (uint32_t i = 0; i < num_cross_blocks(); i++)
    {
        Cuboid_d& cuboid = _cross_cuboids[ i ];

        cross_block_posn_v(&p, i);
        t_x = p->elts[0] - cuboid.get_center()->elts[0];
        t_y = p->elts[1] - cuboid.get_center()->elts[1];
        t_z = p->elts[2] - cuboid.get_center()->elts[2];

        create_3d_homo_translation_matrix_2_d(&T, t_x, t_y, t_z);
        cuboid.transform(T);
    }

    free_vector_d(p);
    free_matrix_d(T);
}

/** The Block_set must be in object coordinates. */
void Pair::assert_consistency()
{
    using namespace jwsc;

    Block_set::assert_consistency();

    Vector_d* p = 0;
    Vector_d* bp = 0;

    posn_v(&p);

    for (size_t i = 0; i < _inner_cuboids.size(); i++)
    {
        inner_block_posn_v(&bp, i);

        double w = size(0);
        double l = size(2);
        double s_i = inner_size(0);

        double dw = fabs(p->elts[0] - bp->elts[0]);
        double dh = fabs(p->elts[1] - bp->elts[1]);
        double dl = fabs(p->elts[2] - bp->elts[2]);

        double s_x, s_y, s_z;
        jwscxx::graphics::Cuboid_d& cuboid = _inner_cuboids[ i ];
        s_x = cuboid.get_point(1)->elts[0] - cuboid.get_point(0)->elts[0];
        s_y = cuboid.get_point(2)->elts[1] - cuboid.get_point(1)->elts[1];
        s_z = cuboid.get_point(3)->elts[2] - cuboid.get_point(2)->elts[2];
        assert(s_x > 0 && s_y > 0 && s_z > 0);
        assert(fabs(s_y - size(1)) < 1.0e-2);

        if (layout() == WIDTH)
        {
            if (w < s_i)
                dw *= -1;
            assert(fabs(dw + 0.5*(s_i - w)) < 1.0e-2);
            assert(dh < 1.0e-2);
            assert(dl < 1.0e-2);

            assert(fabs(s_x - s_i) < 1.0e-2);
            assert(fabs(s_z - size(2)) < 1.0e-2);
        }
        else
        {
            if (l < s_i)
                dl *= -1;
            assert(dw < 1.0e-2);
            assert(dh < 1.0e-2);
            assert(fabs(dl + 0.5*(s_i - l)) < 1.0e-2);

            assert(fabs(s_x - size(0)) < 1.0e-2);
            assert(fabs(s_z - s_i) < 1.0e-2);
        }
    }

    free_vector_d(bp);
    free_vector_d(p);
}

void Pair::sample_var_param_switch_proposal
(
    jwsc::Vector_d**      p_out,
    const class Quartet*  bq
) 
const
{
    using namespace jwsc;

    uint32_t p_d = var_param_inner_size_i(0);

    jwsc::Vector_d* means = 0;
    bq->get_var_param_values(&means);
    Block_set::sample_var_param_switch_proposal(p_out, means);
    free_vector_d(means);

    assert((*p_out)->num_elts > p_d);

    if (layout() == WIDTH)
    {
        (*p_out)->elts[ p_d ] = bq->inner_size(0);
    }
    else
    {
        (*p_out)->elts[ p_d ] = bq->inner_size(1);
    }
}

Pair::Layout Pair::select_random_layout()
{
    if (jwsc::sample_uniform_pdf_d(0,1) < 0.5)
    {
        return WIDTH;
    }
    return LENGTH;
}

/** 
 * The shift and scale deps are the same between a Pair and its inner blocks.
 */
void Pair::add_inner_deps()
{
    if (layout() == WIDTH)
    {
        add_inner_individual_deps(0, 0);
    }
    else
    {
        assert(layout() == LENGTH);
        add_inner_individual_deps(2, 0);
    }
}

void Pair::add_inner_conds()
{
    typedef Parameter P;

    P* s_i = &(_inner_size_p[0]);
    P* s;
    
    if (layout() == WIDTH)
    {
        s = &(_size_p[0]);
        s_i->add_cond(P::Pointer(s));
    }
    else
    {
        assert(layout() == LENGTH);
        s = &(_size_p[2]);
        s_i->add_cond(P::Pointer(s));
    }

    s_i->has_ratio_prior(true);
}

/** 
 * The shift and scale deps are the same between a Pair and its inner blocks.
 */
void Pair::remove_inner_deps()
{
    if (layout() == WIDTH)
    {
        remove_inner_individual_deps(0, 0);
    }
    else
    {
        assert(layout() == LENGTH);
        remove_inner_individual_deps(2, 0);
    }
}

void Pair::remove_inner_conds()
{
    typedef Parameter P;

    P* s_i = &(_inner_size_p[0]);
    P* s;
    
    if (layout() == WIDTH)
    {
        s = &(_size_p[0]);
        s_i->remove_cond(P::Pointer(s));
    }
    else
    {
        assert(layout() == LENGTH);
        s = &(_size_p[2]);
        s_i->remove_cond(P::Pointer(s));
    }

    s_i->has_ratio_prior(false);
}

uint32_t Pair::attachable_block_to_pair
(
    const std::vector<Component*>& components,
    bool                           height_only
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Pair* b = dynamic_cast<Pair*>(components[ i ]);
        if (b && height_only && 
                (b->attachment(2) == 0 || b->attachment(3) == 0))
        {
            indices.push_back(i);
        }
        else if (b && !height_only &&
                (b->num_ctr_attach_pts() - b->num_attachments()) > 0)
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Pair::attachable_block_to_pair_attach_pt
(
    const Pair* b,
    bool        height_only
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t j = 0; j < 6; j++)
    {
        if (height_only && !(j == 2 || j == 3))
            continue;

        if (! b->attachment(j))
        {
            indices.push_back(j);
        }
    }

    return Abstract_block::sample_attach_pt_index(indices);
}

uint32_t Pair::detachable_pair_from_block
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Pair* b = dynamic_cast<Pair*>(components[ i ]);
        if (b && b->num_attachments() <= 1)
        {
            for (uint32_t j = 0; j < 6; j++)
            {
                if (dynamic_cast<Block*>(b->attachment(j)))
                {
                    indices.push_back(i);
                    break;
                }
            }
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Pair::detachable_block_from_pair
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Block* b = dynamic_cast<Block*>(components[ i ]);
        if (b && b->num_attachments() <= 1)
        {
            for (uint32_t j = 0; j < 6; j++)
            {
                if (dynamic_cast<Pair*>(b->attachment(j)))
                {
                    indices.push_back(i);
                    break;
                }
            }
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Pair::splitable_block_to_pair
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Pair* b1 = dynamic_cast<Pair*>(components[ i ]);
        if (b1)
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Pair::splitable_block_to_pair_attach_pt(const Pair* b)
    throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    indices.push_back(2);
    indices.push_back(3);

    return Abstract_block::sample_attach_pt_index(indices);
}

uint32_t Pair::mergeable_pair_from_block
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Pair* b1 = dynamic_cast<Pair*>(components[ i ]);
        if (b1)
        {
            for (uint32_t j = 0; j < 6; j++)
            {
                Block* b2 = dynamic_cast<Block*>(b1->attachment(j));
                if (b2 && b1->is_mergeable(j))
                {
                    indices.push_back(i);
                    break;
                }
            }
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Pair::mergeable_pair_from_block_attach_pt(const Pair* b1)
    throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t j = 0; j < 6; j++)
    {
        Block* b2 = dynamic_cast<Block*>(b1->attachment(j));
        if (b2 && b1->is_mergeable(j))
        {
            indices.push_back(j);
        }
    }

    return Abstract_block::sample_attach_pt_index(indices);
}

uint32_t Pair::mergeable_block_from_pair
(
    const std::vector<Component*>& components,
    bool                           height_only
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Block* b1 = dynamic_cast<Block*>(components[ i ]);
        if (b1)
        {
            for (uint32_t j = 0; j < 6; j++)
            {
                if (height_only && !(j == 2 || j == 3))
                    continue;

                Pair* b2 = dynamic_cast<Pair*>(b1->attachment(j));
                uint32_t k = (j % 2 == 0) ? j+1 : j-1;

                if (b2 && b2->is_mergeable(k))
                {
                    indices.push_back(i);
                    break;
                }
            }
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Pair::mergeable_block_from_pair_attach_pt
(
    const Block* b1,
    bool         height_only
)
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t j = 0; j < 6; j++)
    {
        if (height_only && !(j == 2 || j == 3))
            continue;

        Pair* b2 = dynamic_cast<Pair*>(b1->attachment(j));
        uint32_t k = (j % 2 == 0) ? j+1 : j-1;

        if (b2 && b2->is_mergeable(k))
        {
            indices.push_back(j);
        }
    }

    return Abstract_block::sample_attach_pt_index(indices);
}

uint32_t Pair::switchable_block_to_pair
(
    const std::vector<Component*>& components,
    bool                           height_only
) 
throw (jwscxx::base::Arg_error)
{
    typedef Abstract_block Ablock;
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Block* b1 = dynamic_cast<Block*>(components[ i ]);
        if (b1)
        {
            indices.push_back(i);
            for (uint32_t j = 0; j < 6; j++)
            {
                Ablock* b2 = dynamic_cast<Ablock*>(b1->attachment(j));
                if (b2 && height_only && !(j == 2 || j == 3))
                {
                    indices.pop_back();
                    break;
                }
                else if (b2 && dynamic_cast<Block_set*>(b1->attachment(j)))
                {
                    indices.pop_back();
                    break;
                }
            }
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Pair::switchable_pair_to_block
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Pair* b1 = dynamic_cast<Pair*>(components[ i ]);
        if (b1)
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Pair::switchable_pair_to_quartet
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Pair* b1 = dynamic_cast<Pair*>(components[ i ]);
        if (b1)
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}

/* ========================================================================== */








/* ------------------------------  Quartet  --------------------------------- */

/**
 * The Quartet is centered at (0, 0, 0).
 *
 * @param  ctx  Parameter context for this set of blocks..
 */
Quartet::Quartet(const Block_set::Context* ctx) throw (jwscxx::base::Arg_error) 
    : Block_set("Quartet", ctx)
{
    using jwscxx::graphics::Cuboid_d;

    static const double s[4][2] = {{-0.5,-0.5},{-0.5,0.5},{0.5,0.5},{0.5,-0.5}};

    size_t id = Abstract_block::num_var_params() + 
                Abstract_block::num_derived_params() +
                Abstract_block::num_fixed_params();

    _inner_size_p.add_param(Parameter(id++, "Inner Width", 
                ctx->inner_size_c(0).init_value(),
                &(ctx->inner_size_c(0)), this));
    _inner_size_p.add_param(Parameter(id++, "Inner Length", 
                ctx->inner_size_c(1).init_value(),
                &(ctx->inner_size_c(1)), this));

    add_var_param(&(_inner_size_p[0]));
    add_var_param(&(_inner_size_p[1]));

    for (size_t i = 0; i < 4; i++)
    {
        double w, h, l;
        double x, y, z;

        w = inner_size(0);
        h = size(1);
        l = inner_size(1);

        x = posn(0) + s[i][0]*(size(0) - w);
        y = posn(1);
        z = posn(2) + s[i][1]*(size(2) - l);

        _inner_cuboids.push_back(Cuboid_d(w, h, l, x, y, z));
    }

    add_inner_deps();
    add_inner_conds();
}

/**
 * The Quartet is centered at (0, 0, 0).
 *
 * @param  width     Width.
 * @param  height    Height.
 * @param  length    Length.
 * @param  width_i   Inner width.
 * @param  length_i  Inner length
 * @param  ctx       Parameter context for this Block_set.
 */
Quartet::Quartet
(
    double width,
    double height,
    double length,
    double width_i,
    double length_i,
    const Block_set::Context* ctx
) 
throw (jwscxx::base::Arg_error) 
    : Block_set("Quartet", width, height, length, ctx)
{
    using jwscxx::graphics::Cuboid_d;

    static const double s[4][2] = {{-0.5,-0.5},{-0.5,0.5},{0.5,0.5},{0.5,-0.5}};

    size_t id = Abstract_block::num_var_params() + 
                Abstract_block::num_derived_params() +
                Abstract_block::num_fixed_params();

    _inner_size_p.add_param(Parameter(id++, "Inner Width", width_i, 
                &(ctx->inner_size_c(0)), this));
    _inner_size_p.add_param(Parameter(id++, "Inner Length", length_i,
                &(ctx->inner_size_c(1)), this));

    add_var_param(&(_inner_size_p[0]));
    add_var_param(&(_inner_size_p[1]));

    for (size_t i = 0; i < 4; i++)
    {
        double w, h, l;
        double x, y, z;

        w = inner_size(0);
        h = size(1);
        l = inner_size(1);

        x = posn(0) + s[i][0]*(size(0) - w);
        y = posn(1);
        z = posn(2) + s[i][1]*(size(2) - l);

        _inner_cuboids.push_back(Cuboid_d(w, h, l, x, y, z));
    }

    add_inner_deps();
    add_inner_conds();
}

/**
 * The Quartet is centered at (0, 0, 0).
 *
 * @param  width     Width.
 * @param  height    Height.
 * @param  length    Length.
 * @param  width_i   Inner width.
 * @param  length_i  Inner length
 * @param  cross_N   Number of cross blocks.
 * @param  cross_d1  Distance from the top to the first cross block.
 * @param  cross_h   Height of each cross block.
 * @param  cross_dN  Distance from the top to the last cross block.
 * @param  ctx       Parameter context for this Block_set.
 */
Quartet::Quartet
(
    double   width,
    double   height,
    double   length,
    double   width_i,
    double   length_i,
    double   cross_N,
    double   cross_d1,
    double   cross_h,
    double   cross_dN,
    const Block_set::Context* ctx
) 
throw (jwscxx::base::Arg_error) 
    : Block_set("Quartet", width, height, length, cross_N, cross_d1, cross_h,
            cross_dN, ctx)
{
    using jwscxx::graphics::Cuboid_d;

    static const double s[4][2] = {{-0.5,-0.5},{-0.5,0.5},{0.5,0.5},{0.5,-0.5}};

    size_t id = Abstract_block::num_var_params() + 
                Abstract_block::num_derived_params() +
                Abstract_block::num_fixed_params();

    _inner_size_p.add_param(Parameter(id++, "Inner Width", width_i, 
                &(ctx->inner_size_c(0)), this));
    _inner_size_p.add_param(Parameter(id++, "Inner Length", length_i,
                &(ctx->inner_size_c(1)), this));

    add_var_param(&(_inner_size_p[0]));
    add_var_param(&(_inner_size_p[1]));

    for (size_t i = 0; i < 4; i++)
    {
        double w, h, l;
        double x, y, z;

        w = inner_size(0);
        h = size(1);
        l = inner_size(1);

        x = posn(0) + s[i][0]*(size(0) - w);
        y = posn(1);
        z = posn(2) + s[i][1]*(size(2) - l);

        _inner_cuboids.push_back(Cuboid_d(w, h, l, x, y, z));
    }

    add_inner_deps();
    add_inner_conds();
    add_cross_blocks(cross_N);
}

/** @param  b  Quartet to copy into this one. */
Quartet::Quartet(const Quartet& b) : Block_set(b)
{
    add_var_param(&(_inner_size_p[0]));
    add_var_param(&(_inner_size_p[1]));
    *this = b;
}

/** 
 * @param  b  Quartet to copy into this one.
 *
 * @return A reference to this Quartet.
 */
Quartet& Quartet::operator=(const Quartet& b)
{
    Block_set::operator=(b);
    return *this;
}

Quartet* Quartet::clone() const
{
    return new Quartet(*this);
}

Quartet* Quartet::clone_and_swap() const throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    Quartet* b = new Quartet(size(2), size(1), size(0), inner_size(1),
            inner_size(0), cross(0), cross(1), cross(2), cross(3), context());
    b->copy_var_param_stats(this);
    b->swap_var_attach_pts(this);

    return b;
}

/** 
 * The file format is ASCII.
 *
 * @param  in  Input stream to read this Quartet from.
 *
 * @throw  jwscxx::base::IO_error   Could not read from @em in.
 * @throw  jwscxx::base::Arg_error  Invalid arguments in file to read from.
 */
void Quartet::read(std::istream& in) 
    throw (jwscxx::base::IO_error, jwscxx::base::Arg_error)
{
    using std::ostringstream;
    using std::istringstream;
    using jwscxx::base::Arg_error;

    const char* type_name = typeid(*this).name();
    const char* field_value;

    if (!(field_value = read_field_value(in, "Type")))
    {
        throw Arg_error("Missing Type field");
    }
    if (strncmp(field_value, type_name, strlen(type_name)) != 0)
    {
        ostringstream ost;
        ost << "Tried to read a '" << field_value << "' as a '"
            << type_name << "'";
        throw Arg_error(ost.str());
    }

    Block_set::read(in);
}

/**
 * The file format is ASCII.
 *
 * @param  out  Output stream to write this Quartet to.
 *
 * @throw  jwscxx::base::IO_error  Could not write to @em out.
 */
void Quartet::write(std::ostream& out) const throw (jwscxx::base::IO_error)
{
    out << "Type: " << typeid(*this).name() << std::endl;
    Block_set::write(out);
}

double Quartet::inner_block_size(uint32_t i) const
{
    switch (i)
    {
        case 0:
            return inner_size(0);
        case 1:
            return size(1);
        case 2:
            return inner_size(1);
        default:
            abort();
    }
}

void Quartet::inner_block_posn_v(jwsc::Vector_d** bp_out, uint32_t i) const
{
    using namespace jwsc;

    static const double s[4][2] = {{-0.5,-0.5},{-0.5,0.5},{0.5,0.5},{0.5,-0.5}};
    assert(i < 4);

    Vector_d* p = 0;
    posn_v(&p);
    if (object() && !(object()->in_object_coords())) 
    {
        object()->transform_pt_from_world_to_object_coords(p);
    }

    copy_vector_d(bp_out, p);
    Vector_d* bp = *bp_out;
    bp->elts[0] += s[i][0]*(size(0) - inner_size(0));
    bp->elts[2] += s[i][1]*(size(2) - inner_size(1));

    if (object() && !(object()->in_object_coords())) 
    {
        object()->transform_pt_from_object_to_world_coords(bp);
    }
    free_vector_d(p);
}

void Quartet::inner_block_ctr_attach_pt_v
(
    jwsc::Vector_d** bp_out, 
    uint32_t         i, 
    uint32_t         j
)
const
{
    using namespace jwsc;

    static const double s[4][2] = {{-0.5,-0.5},{-0.5,0.5},{0.5,0.5},{0.5,-0.5}};
    assert(i < 4);

    Vector_d* p = 0;
    ctr_attach_pt_v(&p, j);
    if (object() && !(object()->in_object_coords())) 
    {
        object()->transform_pt_from_world_to_object_coords(p);
    }

    copy_vector_d(bp_out, p);
    Vector_d* bp = *bp_out;
    bp->elts[0] += s[i][0]*(size(0) - inner_size(0));
    bp->elts[2] += s[i][1]*(size(2) - inner_size(1));

    if (object() && !(object()->in_object_coords())) 
    {
        object()->transform_pt_from_object_to_world_coords(bp);
    }
    free_vector_d(p);
}

double Quartet::cross_block_size(uint32_t i, uint32_t whl) const
{
    uint32_t N = 4*num_cross_blocks();
    uint32_t j = i % 4;
    assert(N > 0 && N > i);

    // Cross pieces that span the width direction.
    bool cross_width = (j < 2);
    double s;

    switch (whl)
    {
        case 0:
            s = (cross_width) ? size(0) - 2*inner_size(0) : inner_size(0);
            break;
        case 1:
            s = cross(2);
            break;
        case 2:
            s = (cross_width) ? inner_size(1) : size(2) - 2*inner_size(1);
            break;
        default:
            abort();
    }

    return (s < 1.0) ? 1.0 : s;
}

void Quartet::cross_block_posn_v(jwsc::Vector_d** bp_out, uint32_t i) const
{
    using namespace jwsc;

    uint32_t M = num_cross_blocks();
    uint32_t N = 4*M;
    uint32_t j = i % 4;
    uint32_t k = i / 4;
    assert(N > 0 && N > i);

    // Cross pieces that span the width direction.
    bool cross_width = (j < 2);

    double h  = cross(2);
    double d1 = cross(1);
    double dM = cross(3);
    double s  = (M > 1) ? (dM - d1 - M*h) / (M - 1) : 0;

    ctr_attach_pt_v(bp_out, 2);
    if (object() && !(object()->in_object_coords())) 
    {
        object()->transform_pt_from_world_to_object_coords(*bp_out);
    }

    Vector_d* bp = *bp_out;
    bp->elts[0]  = (cross_width) ? 0 : 0.5*(size(0) - inner_size(0));
    bp->elts[1] -= d1 + 0.5*h + k*h + k*s;
    bp->elts[2]  = (cross_width) ? 0.5*(size(2) - inner_size(1)) : 0;

    if (j % 2)
    {
        bp->elts[0] *= -1;
        bp->elts[2] *= -1;
    }

    if (object() && !(object()->in_object_coords())) 
    {
        object()->transform_pt_from_object_to_world_coords(bp);
    }
}

void Quartet::add_cross_block()
{
    using namespace jwsc;
    using jwscxx::graphics::Cuboid_d;

    Block_set::add_cross_block();

    try { _cross_p[0].value(num_cross_blocks()+1); }
    catch (jwscxx::base::Arg_error e) { abort(); }

    uint32_t N = 4*num_cross_blocks();

    for (uint32_t i = 0; i < 4; i++)
    {
        double w = cross_block_size((N-4)+i, 0);
        double h = cross_block_size((N-4)+i, 1);
        double l = cross_block_size((N-4)+i, 2);
        _cross_cuboids.push_back(Cuboid_d(w, h, l, 0, 0, 0));
    }

    double t_x, t_y, t_z;
    Matrix_d* T = 0;
    Vector_d* p = 0;

    for (uint32_t i = 0; i < N; i++)
    {
        Cuboid_d& cuboid = _cross_cuboids[ i ];

        cross_block_posn_v(&p, i);
        t_x = p->elts[0] - cuboid.get_center()->elts[0];
        t_y = p->elts[1] - cuboid.get_center()->elts[1];
        t_z = p->elts[2] - cuboid.get_center()->elts[2];

        create_3d_homo_translation_matrix_2_d(&T, t_x, t_y, t_z);
        cuboid.transform(T);
    }

    free_vector_d(p);
    free_matrix_d(T);
}

void Quartet::remove_cross_block()
{
    using namespace jwsc;
    using jwscxx::graphics::Cuboid_d;

    Block_set::remove_cross_block();

    for (uint32_t i = 0; i < 4; i++)
    {
        _cross_cuboids.pop_back();
    }

    double t_x, t_y, t_z;
    Matrix_d* T = 0;
    Vector_d* p = 0;

    for (uint32_t i = 0; i < 4*num_cross_blocks(); i++)
    {
        Cuboid_d& cuboid = _cross_cuboids[ i ];

        cross_block_posn_v(&p, i);
        t_x = p->elts[0] - cuboid.get_center()->elts[0];
        t_y = p->elts[1] - cuboid.get_center()->elts[1];
        t_z = p->elts[2] - cuboid.get_center()->elts[2];

        create_3d_homo_translation_matrix_2_d(&T, t_x, t_y, t_z);
        cuboid.transform(T);
    }

    free_vector_d(p);
    free_matrix_d(T);
}

/** The Block_set must be in object coordinates. */
void Quartet::assert_consistency()
{
    using namespace jwsc;

    Block_set::assert_consistency();

    Vector_d* p = 0;
    Vector_d* bp = 0;

    posn_v(&p);

    for (size_t i = 0; i < _inner_cuboids.size(); i++)
    {
        inner_block_posn_v(&bp, i);

        double w = size(0);
        double l = size(2);

        double w_i = inner_size(0);
        double l_i = inner_size(1);

        double dw = fabs(p->elts[0] - bp->elts[0]);
        double dh = fabs(p->elts[1] - bp->elts[1]);
        double dl = fabs(p->elts[2] - bp->elts[2]);

        if (w < w_i)
        {
            dw *= -1;
        }
        if (l < l_i)
        {
            dl *= -1;
        }

        assert(fabs(dw + 0.5*(w_i - w)) < 1.0e-2);
        assert(dh < 1.0e-2);
        assert(fabs(dl + 0.5*(l_i - l)) < 1.0e-2);

        double s_x, s_y, s_z;
        jwscxx::graphics::Cuboid_d& cuboid = _inner_cuboids[ i ];
        s_x = cuboid.get_point(1)->elts[0] - cuboid.get_point(0)->elts[0];
        s_y = cuboid.get_point(2)->elts[1] - cuboid.get_point(1)->elts[1];
        s_z = cuboid.get_point(3)->elts[2] - cuboid.get_point(2)->elts[2];
        assert(s_x > 0 && s_y > 0 && s_z > 0);
        assert(fabs(s_x - inner_size(0)) < 1.0e-2);
        assert(fabs(s_y - size(1)) < 1.0e-2);
        assert(fabs(s_z - inner_size(1)) < 1.0e-2);
    }

    free_vector_d(bp);
    free_vector_d(p);
}

void Quartet::sample_var_param_switch_proposal
(
    jwsc::Vector_d**   p_out,
    const Pair*        bp
) 
const
{
    using namespace jwsc;

    uint32_t p_d = var_param_inner_size_i(0);

    jwsc::Vector_d* means = 0;
    bp->get_var_param_values(&means);
    Block_set::sample_var_param_switch_proposal(p_out, means);
    free_vector_d(means);

    assert((*p_out)->num_elts > p_d+1);

    if (bp->layout() == Pair::WIDTH)
    {
        (*p_out)->elts[ p_d+0 ] = bp->inner_size(0);
    }
    else
    {
        (*p_out)->elts[ p_d+1 ] = bp->inner_size(0);
    }
}

/** 
 * The shift and scale deps are the same between a Quartet and its inner blocks.
 */
void Quartet::add_inner_deps()
{
    add_inner_individual_deps(0, 0);
    add_inner_individual_deps(2, 1);
}

void Quartet::add_inner_conds()
{
    typedef Parameter P;

    P* w_i = &(_inner_size_p[0]);
    P* w   = &(_size_p[0]);

    P* l_i = &(_inner_size_p[1]);
    P* l   = &(_size_p[2]);
    
    w_i->add_cond(P::Pointer(w));
    w_i->has_ratio_prior(true);

    l_i->add_cond(P::Pointer(l));
    l_i->has_ratio_prior(true);
}

/** 
 * The shift and scale deps are the same between a Quartet and its inner blocks.
 */
void Quartet::remove_inner_deps()
{
    remove_inner_individual_deps(0, 0);
    remove_inner_individual_deps(2, 1);
}

void Quartet::remove_inner_conds()
{
    typedef Parameter P;

    P* w_i = &(_inner_size_p[0]);
    P* w   = &(_size_p[0]);

    P* l_i = &(_inner_size_p[1]);
    P* l   = &(_size_p[2]);
    
    w_i->remove_cond(P::Pointer(w));
    w_i->has_ratio_prior(false);

    l_i->remove_cond(P::Pointer(l));
    l_i->has_ratio_prior(false);
}

uint32_t Quartet::attachable_block_to_quartet
(
    const std::vector<Component*>& components,
    bool                           height_only
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Quartet* b = dynamic_cast<Quartet*>(components[ i ]);
        if (b && height_only && 
                (b->attachment(2) == 0 || b->attachment(3) == 0))
        {
            indices.push_back(i);
        }
        else if (b && !height_only &&
                (b->num_ctr_attach_pts() - b->num_attachments()) > 0)
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Quartet::attachable_block_to_quartet_attach_pt
(
    const Quartet* b,
    bool           height_only
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t j = 0; j < 6; j++)
    {
        if (height_only && !(j == 2 || j == 3))
            continue;

        if (! b->attachment(j))
        {
            indices.push_back(j);
        }
    }

    return Abstract_block::sample_attach_pt_index(indices);
}

uint32_t Quartet::detachable_quartet_from_block
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Quartet* b = dynamic_cast<Quartet*>(components[ i ]);
        if (b && b->num_attachments() <= 1)
        {
            for (uint32_t j = 0; j < 6; j++)
            {
                if (dynamic_cast<Block*>(b->attachment(j)))
                {
                    indices.push_back(i);
                    break;
                }
            }
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Quartet::detachable_block_from_quartet
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Block* b = dynamic_cast<Block*>(components[ i ]);
        if (b && b->num_attachments() <= 1)
        {
            for (uint32_t j = 0; j < 6; j++)
            {
                if (dynamic_cast<Quartet*>(b->attachment(j)))
                {
                    indices.push_back(i);
                    break;
                }
            }
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Quartet::splitable_block_to_quartet
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Quartet* b1 = dynamic_cast<Quartet*>(components[ i ]);
        if (b1)
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Quartet::splitable_block_to_quartet_attach_pt(const Quartet* b)
    throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    indices.push_back(2);
    indices.push_back(3);

    return Abstract_block::sample_attach_pt_index(indices);
}

uint32_t Quartet::mergeable_quartet_from_block
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Quartet* b1 = dynamic_cast<Quartet*>(components[ i ]);
        if (b1)
        {
            for (uint32_t j = 0; j < 6; j++)
            {
                Block* b2 = dynamic_cast<Block*>(b1->attachment(j));
                if (b2 && b1->is_mergeable(j))
                {
                    indices.push_back(i);
                    break;
                }
            }
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Quartet::mergeable_quartet_from_block_attach_pt(const Quartet* b1)
    throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t j = 0; j < 6; j++)
    {
        Block* b2 = dynamic_cast<Block*>(b1->attachment(j));
        if (b2 && b1->is_mergeable(j))
        {
            indices.push_back(j);
        }
    }

    return Abstract_block::sample_attach_pt_index(indices);
}

uint32_t Quartet::mergeable_block_from_quartet
(
    const std::vector<Component*>& components,
    bool                           height_only
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Block* b1 = dynamic_cast<Block*>(components[ i ]);
        if (b1)
        {
            for (uint32_t j = 0; j < 6; j++)
            {
                if (height_only && !(j == 2 || j == 3))
                    continue;

                Quartet* b2 = dynamic_cast<Quartet*>(b1->attachment(j));
                uint32_t k = (j % 2 == 0) ? j+1 : j-1;

                if (b2 && b2->is_mergeable(k))
                {
                    indices.push_back(i);
                    break;
                }
            }
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Quartet::mergeable_block_from_quartet_attach_pt
(
    const Block* b1,
    bool         height_only
)
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t j = 0; j < 6; j++)
    {
        if (height_only && !(j == 2 || j == 3))
            continue;

        Quartet* b2 = dynamic_cast<Quartet*>(b1->attachment(j));
        uint32_t k = (j % 2 == 0) ? j+1 : j-1;

        if (b2 && b2->is_mergeable(k))
        {
            indices.push_back(j);
        }
    }

    return Abstract_block::sample_attach_pt_index(indices);
}

uint32_t Quartet::switchable_block_to_quartet
(
    const std::vector<Component*>& components,
    bool                           height_only
) 
throw (jwscxx::base::Arg_error)
{
    typedef Abstract_block Ablock;
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Block* b1 = dynamic_cast<Block*>(components[ i ]);
        if (b1)
        {
            indices.push_back(i);
            for (uint32_t j = 0; j < 6; j++)
            {
                Ablock* b2 = dynamic_cast<Ablock*>(b1->attachment(j));
                if (b2 && height_only && !(j == 2 || j == 3))
                {
                    indices.pop_back();
                    break;
                }
                else if (b2 && dynamic_cast<Block_set*>(b1->attachment(j)))
                {
                    indices.pop_back();
                    break;
                }
            }
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Quartet::switchable_quartet_to_block
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Quartet* b1 = dynamic_cast<Quartet*>(components[ i ]);
        if (b1)
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}

uint32_t Quartet::switchable_quartet_to_pair
(
    const std::vector<Component*>& components
) 
throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < components.size(); i++)
    {
        Quartet* b1 = dynamic_cast<Quartet*>(components[ i ]);
        if (b1)
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}

/* ========================================================================== */
