/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Set of block structures represented as cuboids.
 */


#ifndef BLOCK_SET_H
#define BLOCK_SET_H


#include <config.h>

#include <inttypes.h>

#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/base/cloneable.h>
#include <jwsc++/graphics/cuboid.h>
#include <jwsc++/graphics/polygon.h>

#include "model.h"
#include "component.h"
#include "block.h"


/** 
 * @class Block_set
 *
 * @brief Set of symmetric blocks.
 */
class Block_set : public Abstract_block
{
    public:


/**
 * @class Block_set::Context
 *
 * @brief Context for Block_set model parameters.
 */
class Context : public Abstract_block::Context
{
    public:

        /** @brief Constructs a new default Block_set::Context. */
        Context();

        /** @brief Constructs a new Block_set::Context from another. */
        Context(const Block_set::Context& ctx);

        /** @brief Copies a Block_set::Context into this one. */
        Block_set::Context& operator= (const Block_set::Context& ctx);

        /** @brief Clones this Context. */
        Block_set::Context* clone() const;

        const Component::Paramgroup::Context& inner_size_c() const;

        void inner_size_c(const Component::Paramgroup::Context& size);

        const Model::Parameter::Context& inner_size_c(size_t i) const;

        void inner_size_c(size_t i, const Model::Parameter::Context& size);

        const Component::Paramgroup::Context& cross_c() const;

        void cross_c(const Component::Paramgroup::Context& c);

        const Model::Parameter::Context& cross_c(size_t i) const;

        void cross_c(size_t i, const Model::Parameter::Context& c);


    protected:

        /** 
         * @brief Inner size of blocks in this set.
         *
         * Contains two parameters indexed by width=0, length=1. The parameter
         * is a value in [0,1] and is the proportion of corresponding size of
         * the containing block's size.
         */
        Component::Paramgroup::Context _inner_size_c;

        /**
         * @brief Cross pieces between the inner blocks of this set.
         *
         * Contains four parameters (indexed by):
         *
         * - (0) Number of cross pieces.
         * - (1) Distance from block set top to center of first cross.
         * - (2) Height of each cross.
         * - (3) Distance from block set top to center of last cross.
         */
        Component::Paramgroup::Context _cross_c;
};


    // Block_set
    public:

        /**
         * @name Standard
         *
         * Standard class methods.
         * @{
         */

        /** @brief Constructs a default Block_set. */
        Block_set(const char* name, const Block_set::Context* ctx) 
            throw (jwscxx::base::Arg_error);

        /** @brief Constructs a specific Block_set. */
        Block_set
        (
            const char* name, 
            double      width, 
            double      height, 
            double      length, 
            const Block_set::Context* ctx
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Constructs a specific Block_set. */
        Block_set
        (
            const char* name, 
            double      width, 
            double      height, 
            double      length, 
            double      cross_N,
            double      cross_d1,
            double      cross_h,
            double      cross_dN,
            const Block_set::Context* ctx
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Constructs a Block_set by copying another. */
        Block_set(const Block_set& b);

        /** @brief Deletes this Block_set. */
        virtual ~Block_set();

        /** @brief Copies a Block_set into this one. */
        Block_set& operator= (const Block_set& b);

        /** @brief Clones this Block_set. */
        virtual Block_set* clone() const = 0;

        /** @brief Creates a special copy of this block. */
        virtual Block_set* clone_and_swap() const 
            throw (jwscxx::base::Arg_error) = 0;

        /** @brief Reads this Block_set from an input stream. */
        virtual void read(std::istream& in) throw (jwscxx::base::IO_error,
                jwscxx::base::Arg_error);

        /** @brief Writes this Block_set to an output stream. */
        virtual void write(std::ostream& out) const 
            throw (jwscxx::base::IO_error);

        /** @brief Returns the model context. */
        const Block_set::Context* context() const;

        uint32_t num_inner_blocks() const;

        virtual double inner_block_size(uint32_t i) const = 0;

        /** @brief Returns the position of an inner block. */
        virtual void inner_block_posn_v(jwsc::Vector_d** p_out, uint32_t i) 
            const = 0;

        /** @brief Returns the position of an inner block. */
        virtual void inner_block_ctr_attach_pt_v
        (
            jwsc::Vector_d** p_out, 
            uint32_t         i,
            uint32_t         j
        )
        const = 0;

        uint32_t num_cross_blocks() const;

        /** @brief Returns the position of a cross block. */
        virtual void cross_block_posn_v(jwsc::Vector_d** p_out, uint32_t i) 
            const = 0;

        bool is_mergeable(uint32_t i) const;

        /*@}*/


        /**
         * @name Variable Parameters
         *
         * Variable model parameter members.
         * @{
         */

        /** @brief Adds to the instance parameter values. */
        void add_var_param_values_scale_deps(const jwsc::Vector_d* p)
            throw (jwscxx::base::Arg_error);

        /** @brief Prints variable parameter values. */
        void print_var_param_values
        (
            uint32_t      iteration, 
            std::ostream& out
        )
        const;

        /*@}*/


        /**
         * @name inner_size
         *
         * Inner size parameter members.
         * @{
         */

        /** @brief Returns the inner size of this Block_set. */
        const Component::Paramgroup* inner_size_p() const;

        /** @brief Returns the inner size of this Block_set. */
        const Component::Parameter* inner_size_p(size_t i) const;

        /** @brief Returns the inner size of this Block_set. */
        void inner_size_v(jwsc::Vector_d** s_out) const;

        /** @brief Returns the inner size of this Block_set. */
        double inner_size(size_t i) const;

        /** 
         * @brief Sets the inner size of this Block_set and shifts its
         * dependents.
         */
        void inner_size_shift_deps(size_t i, double size);

        /** 
         * @brief Sets the inner size of this Block_set and scales its
         * dependents.
         */
        void inner_size_scale_deps(size_t i, double size);

        /** 
         * @brief Tests a inner size for this Block_set and its shifted
         * dependents.
         */
        void test_inner_size_shift_deps(size_t i, double size)
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Tests a inner size for this Block_set and its scaled
         * dependents.
         */
        void test_inner_size_scale_deps(size_t i, double size)
            throw (jwscxx::base::Arg_error);

        /**
         * @brief Returns the index of an inner size parameter in the parameter
         * vector created for this block set.
         */
        uint32_t var_param_inner_size_i(uint32_t j) const;

        /*@}*/


        /**
         * @name cross
         *
         * Cross block parameter members.
         * @{
         */

        /** @brief Returns the cross block paramgroup of this block set. */
        const Component::Paramgroup* cross_p() const;

        /** @brief Returns a cross block parameter of this block set. */
        const Component::Parameter* cross_p(size_t i) const;

        /** @brief Returns a cross block parameter value of this block set. */
        double cross(size_t i) const;

        /**
         * @brief Returns the index of a cross block parameter in the parameter
         * vector created for this block set.
         */
        uint32_t var_param_cross_i(uint32_t j) const;

        /** 
         * @brief Returns the number of cross block parameters currently
         * enabled as variable parameters.
         */
        uint32_t num_var_param_cross_p() const;

        /** @brief Adds cross blocks to this block set. */
        void add_cross_blocks(uint32_t N);

        /** @brief Tests adding cross blocks to this block set. */
        void test_add_cross_blocks(uint32_t N) throw (jwscxx::base::Arg_error);

        /** @brief Removes cross blocks from this block set. */
        void remove_cross_blocks(uint32_t N);

        /** @brief Tests removing cross blocks from this block set. */
        void test_remove_cross_blocks(uint32_t N) 
            throw (jwscxx::base::Arg_error);

        /*@}*/


        /**
         * @name Attachment
         *
         * Component attachment members.
         * @{
         */

        /** 
         * @brief Asserts consitency among attachments and the graphical
         * representation.
         */
        virtual void assert_consistency();

        /** 
         * @brief Ensures consitency among attachments and the graphical
         * representation.
         */
        void ensure_consistency();

        /*@}*/


        /**
         * @name Likelihood
         *
         * Instance parameter likelihood members.
         * @{
         */

        /** 
         * @brief Returns the log likelihood of a center point attached to
         * another block face.
         */
        double calc_inner_block_ctr_attach_pt_likelihood
        (
            uint32_t              i,
            uint32_t              a_j,
            const Abstract_block* b,
            uint32_t              b_j
        ) 
        const;

        /** @brief Returns the log likelihood of a cross block. */
        double calc_cross_block_likelihood(uint32_t i) const;

        /** @brief Calculates the log likelihood of the instance parameters. */
        double calc_var_param_log_likelihood() const;

        /** @brief Calculates the log likelihood of the instance parameters. */
        double calc_var_param_proposal_log_likelihood() const;

        /** 
         * @brief Generates instance parameter values from the likelihood.
         */
        void sample_var_param_likelihood(jwsc::Vector_d** p_out);

        void gen_var_param_conditional_likelihood_sample();

        void get_var_param_conditional_likelihood_sample
        (
            jwsc::Vector_d** p_out
        );

        /*@}*/


        /**
         * @name Prior
         *
         * Category parameter prior probability members.
         * @{
         */

        /** @brief Calculates the log probability. */
        double calc_var_param_prior_log_prob() const;

        /** @brief Generates a sample from the parameter prior. */
        void sample_var_param_prior
        (
            jwsc::Vector_d** means_out,
            jwsc::Vector_d** stdevs_out
        ) 
        const;

        /*@}*/


        /**
         * @name Category Proposal
         *
         * Category proposal probability distribution members.
         * @{
         */

        void sample_var_param_cat_proposal
        (
            jwsc::Vector_d** means_out,
            jwsc::Vector_d** stdevs_out,
            const jwsc::Vector_d* means_in
        ) 
        const;

        /*@}*/


        /**
         * @name Proposal
         *
         * Proposal probability distribution members.
         * @{
         */

        /** 
         * @brief Calculates the proposal log probability of instance
         * parameters.
         */
        double calc_var_param_proposal_log_prob() const;

        /** 
         * @brief Calculates the Gaussian proposal log probability of instance
         * parameters.
         */
        double calc_var_param_gauss_proposal_log_prob() const;

        double calc_var_param_split_merge_proposal_log_prob
        (
            const jwsc::Vector_d* q,
            const jwsc::Vector_d* means,
            uint32_t              j
        )
        const;

        double calc_var_param_switch_proposal_log_prob
        (
            const jwsc::Vector_d* q
        )
        const;

        double calc_var_param_cross_proposal_log_prob
        (
            const jwsc::Vector_d* q
        )
        const;

        /** 
         * @brief Generates instance parameter values from the proposal
         * distribution.
         */
        void sample_var_param_proposal(jwsc::Vector_d** p_out) const;

        /**
         * @brief Generates instance parameter values from the Gaussian proposal
         * distribution.
         */
        void sample_var_param_gauss_proposal
        (
            jwsc::Vector_d**      p_out,
            const jwsc::Vector_d* means
        ) 
        const;

        void sample_var_param_split_merge_proposal
        (
            jwsc::Vector_d**      p_out,
            const jwsc::Vector_d* means,
            uint32_t              j
        ) 
        const;

        virtual void sample_var_param_switch_proposal
        (
            jwsc::Vector_d**      p_out,
            const jwsc::Vector_d* means
        ) 
        const;

        void sample_var_param_cross_proposal
        (
            jwsc::Vector_d** p_out,
            bool             one_cross
        ) 
        const;

        /*@}*/


        /**
         * @name Transform
         *
         * Transformation of internal representation members.
         * @{
         */

        /** @brief Translates this Component. */
        void translate(double x, double y, double z);

        /** @brief Rotates this Component about a vector. */
        void rotate(double phi, double x, double y, double z);

        /*@}*/


        /**
         * @name Conditionals
         *
         * Conditional management members.
         * @{
         */

        /** @brief Returns whether this Component conditions on another. */
        bool conditions(const Model* m) const;

        /*@}*/


        /**
         * @name Graphical Representation
         *
         * Internal graphical representation members.
         * @{
         */

        /** 
         * @brief Returns the set of polygon faces representing the 3D structure
         * of this Block_set.
         */
        std::vector<const jwscxx::graphics::Polygon_d*> faces() const;

        /** @brief Returns the number of faces this Block_set has. */
        size_t num_faces() const;

        /** @brief Sets the sequential ids of a faces in this Block_set. */
        uint32_t set_sequential_face_ids(uint32_t id);

        /** 
         * @brief Updates the graphical representation of this block with its
         * current parameter values.
         */
        void update_graphical_rep();

        /*@}*/


        /**
         * @name Graphics
         *
         * Graphics rendering members.
         * @{
         */

        /** 
         * @brief Renders this block with the GL as a wire-frame using
         * polygon face ids as color.
         */
        void wire_render_with_ids(const View* view) const;

        /** @brief Renders this block with the GL as a solid object. */
        void solid_render() const;

        /*@}*/


        /**
         * @name Selection
         *
         * Random selection of block set dependent components and attach points.
         * @{
         */

        static uint32_t attachable_cross_blocks
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t detachable_cross_blocks
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t splitable_block_set_to_block
        (
            const std::vector<Component*>& components,
            bool                           height_only
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t splitable_block_set_to_block_attach_pt
        (
            const Block* b,
            bool         height_only
        )
        throw (jwscxx::base::Arg_error);

        /*@}*/


    protected:

        /**
         * @name Parameters
         *
         * Generic model parameter members.
         * @{
         */

        /** @brief Sets the Parameter values from a point. */
        void set_var_param_values_in_object_coords
        (
            const jwsc::Vector_d* p
        )
        throw (jwscxx::base::Arg_error);

        void copy_var_param_stats(const Abstract_block* b);

        /*@}*/


        /**
         * @name cross
         *
         * Cross block parameter members.
         * @{
         */

        /** @brief Adds a cross block to this block set. */
        virtual void add_cross_block();

        /** @brief Removes a cross block from this block set. */
        virtual void remove_cross_block();

        virtual double cross_block_size(uint32_t i, uint32_t whl) const = 0;

        /*@}*/


        /**
         * @name Attachment
         *
         * Component attachment members.
         * @{
         */

        /** @brief Adds internal shift dependencies induced by an attachment. */
        void attach_shift_deps(uint32_t j, uint32_t i);

        /** @brief Adds internal scale dependencies induced by an attachment. */
        void attach_scale_deps(uint32_t j, uint32_t i);

        /** 
         * @brief Removes internal shift dependencies induced by an detachment.
         */
        void detach_shift_deps(uint32_t j, uint32_t i);

        /** 
         * @brief Removes internal scale dependencies induced by an detachment.
         */
        void detach_scale_deps(uint32_t j, uint32_t i);

        /*@}*/


        /**
         * @name Representation
         *
         * Internal representation members.
         * @{
         */

        /** @brief Adds the internal block parameter dependencies. */
        virtual void add_inner_deps() = 0;

        /** @brief Adds the internal block parameter conditionals. */
        virtual void add_inner_conds() = 0;

        /** 
         * @brief Adds the internal block parameter dependencies for internal
         * blocks individually attached to a face in the block set.
         */
        void add_inner_individual_deps(uint32_t i, uint32_t n);

        /** @brief Removes the internal block parameter dependencies. */
        virtual void remove_inner_deps() = 0;

        /** @brief Removes the internal block parameter conditionals. */
        virtual void remove_inner_conds() = 0;

        /** 
         * @brief Removes the internal block parameter dependencies for internal
         * blocks individually attached to a face in the block set.
         */
        void remove_inner_individual_deps(uint32_t i, uint32_t n);

        /** 
         * @brief Adds conditionals between inner block center attachment
         * points and the size of an attaching block.
         */
        void add_inner_ctr_attach_conds
        (
            uint32_t               a_j, 
            Component::Paramgroup& size_p
        );

        /** 
         * @brief Removes conditionals from inner block center attachment
         * points and the size of an attaching block.
         */
        void remove_inner_ctr_attach_conds
        (
            uint32_t               a_j, 
            Component::Paramgroup& size_p
        );

        /*@}*/


    private:

        /**
         * @name Attachment
         *
         * Component attachment members.
         * @{
         */

        /** @brief Adds internal shift dependencies induced by an attachment. */
        void attach_shift_deps_1();

        /** @brief Adds internal shift dependencies induced by an attachment. */
        void attach_shift_deps_2();

        /** @brief Adds internal scale dependencies induced by an attachment. */
        void attach_scale_deps_1();

        /** @brief Adds internal scale dependencies induced by an attachment. */
        void attach_scale_deps_2();

        /** 
         * @brief Removes internal shift dependencies induced by an detachment.
         */
        void detach_shift_deps_1();

        /** 
         * @brief Removes internal shift dependencies induced by an detachment.
         */
        void detach_shift_deps_2();

        /** 
         * @brief Removes internal scale dependencies induced by an detachment.
         */
        void detach_scale_deps_1();

        /** 
         * @brief Removes internal scale dependencies induced by an detachment.
         */
        void detach_scale_deps_2();

        /*@}*/


    protected:

        /** 
         * @brief Inner size of blocks in this set.
         *
         * Contains two parameters indexed by width=0, length=1. The parameter
         * is a value in [0,1] and is the proportion of corresponding size of
         * the containing block's size.
         */
        Component::Paramgroup _inner_size_p;

        /**
         * @brief Cross pieces between the inner blocks of this set.
         *
         * Contains four parameters (indexed by):
         *
         * - (0) Number of cross pieces.
         * - (1) Distance from block set top to center of first cross.
         * - (2) Height of each cross.
         * - (3) Distance from block set top to center of last cross.
         */
        Component::Paramgroup _cross_p;

        /** @brief Geometric representation of inner blocks. */
        std::vector<jwscxx::graphics::Cuboid_d> _inner_cuboids;

        /** @brief Geometric representation of cross blocks. */
        std::vector<jwscxx::graphics::Cuboid_d> _cross_cuboids;


    friend class Object;
};



/** 
 * @class Pair
 *
 * @brief Pair of symmetric blocks.
 */
class Pair : public Block_set
{
    public:

        /** @brief Type of layout to orient the inner blocks with. */
        typedef enum
        {
            WIDTH=0, /*!< Inner blocks aligned along the width (x) direction. */
            LENGTH  /*!< Inner blocks aligned along the length (x) direction. */
        }
        Layout;


        /**
         * @name Standard
         *
         * Standard class methods.
         * @{
         */

        /** @brief Constructs a default Pair. */
        Pair(Pair::Layout layout, const Block_set::Context* ctx) 
            throw (jwscxx::base::Arg_error);

        /** @brief Constructs a specific Pair. */
        Pair
        (
            double       width,
            double       height,
            double       length,
            double       size_i,
            Pair::Layout layout, 
            const Block_set::Context* ctx
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Constructs a specific Pair. */
        Pair
        (
            double       width,
            double       height,
            double       length,
            double       size_i,
            double       cross_N,
            double       cross_d1,
            double       cross_h,
            double       cross_dN,
            Pair::Layout layout, 
            const Block_set::Context* ctx
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Constructs a Pair by copying another. */
        Pair(const Pair& b);

        /** @brief Copies a Pair into this one. */
        Pair& operator= (const Pair& b);

        /** @brief Clones this Pair. */
        Pair* clone() const;

        /** @brief Creates a special copy of this Pair. */
        Pair* clone_and_swap() const throw (jwscxx::base::Arg_error);

        /** @brief Reads this Pair from an input stream. */
        void read(std::istream& in) throw (jwscxx::base::IO_error,
                jwscxx::base::Arg_error);

        /** @brief Writes this Pair to an output stream. */
        void write(std::ostream& out) const throw (jwscxx::base::IO_error);

        double inner_block_size(uint32_t i) const;

        /** @brief Returns the position of an inner block. */
        void inner_block_posn_v(jwsc::Vector_d** p_out, uint32_t i) const;

        void inner_block_ctr_attach_pt_v
        (
            jwsc::Vector_d** p_out, 
            uint32_t         i, 
            uint32_t         j
        )
        const;

        void cross_block_posn_v(jwsc::Vector_d** bp_out, uint32_t i) const;

        /*@}*/


        /**
         * @name Layout
         *
         * Inner block orientation layout.
         * @{
         */

        /** @brief Returns the inner block layout. */
        Pair::Layout layout() const;

        /** @brief Sets the inner block layout. */
        void layout(Pair::Layout layout) throw (jwscxx::base::Arg_error);

        /*@}*/


        /**
         * @name Attachment
         *
         * Component attachment members.
         * @{
         */

        /** 
         * @brief Asserts consitency among attachments and the graphical
         * representation.
         */
        void assert_consistency();

        /*@}*/


        void sample_var_param_switch_proposal
        (
            jwsc::Vector_d**      p_out,
            const class Quartet*  bq
        ) 
        const;


        /** @brief Selects a random pair layout. */
        static Pair::Layout select_random_layout();


        /**
         * @name Selection
         *
         * Random selection of block set dependent components and attach points.
         * @{
         */

        static uint32_t attachable_block_to_pair
        (
            const std::vector<Component*>& components,
            bool                           height_only
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t attachable_block_to_pair_attach_pt
        (
            const Pair* b,
            bool        height_only
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t detachable_pair_from_block
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t detachable_block_from_pair
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t splitable_block_to_pair
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t splitable_block_to_pair_attach_pt
        (
            const Pair* b
        )
        throw (jwscxx::base::Arg_error);

        static uint32_t mergeable_pair_from_block
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t mergeable_pair_from_block_attach_pt
        (
            const Pair* b1
        )
        throw (jwscxx::base::Arg_error);

        static uint32_t mergeable_block_from_pair
        (
            const std::vector<Component*>& components,
            bool                           height_only
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t mergeable_block_from_pair_attach_pt
        (
            const Block* b,
            bool         height_only
        )
        throw (jwscxx::base::Arg_error);

        static uint32_t switchable_block_to_pair
        (
            const std::vector<Component*>& components,
            bool                           height_only
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t switchable_pair_to_block
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t switchable_pair_to_quartet
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        /*@}*/


    private:

        /**
         * @name cross
         *
         * Cross block parameter members.
         * @{
         */

        /** @brief Adds a cross block to this block set. */
        void add_cross_block();

        /** @brief Removes a cross block from this block set. */
        void remove_cross_block();

        double cross_block_size(uint32_t i, uint32_t whl) const;

        /*@}*/


        /**
         * @name Representation
         *
         * Internal representation members.
         * @{
         */

        /** @brief Adds the internal block parameter dependencies. */
        void add_inner_deps();

        /** @brief Adds the internal block parameter conditionals. */
        void add_inner_conds();

        /** @brief Removes the internal block parameter dependencies. */
        void remove_inner_deps();

        /** @brief Removes the internal block parameter conditionals. */
        void remove_inner_conds();

        /*@}*/


    private:

        /** @brief Inner block orientation layout. */
        Layout _layout;


    friend class Object;
};


/** 
 * @class Quartet
 *
 * @brief Four symmetric blocks.
 */
class Quartet : public Block_set
{
    public:

        /**
         * @name Standard
         *
         * Standard class methods.
         * @{
         */

        /** @brief Constructs a default Quartet. */
        Quartet(const Block_set::Context* ctx) 
            throw (jwscxx::base::Arg_error);

        /** @brief Constructs a specific Quartet. */
        Quartet
        (
            double width,
            double height,
            double length,
            double width_i,
            double length_i,
            const Block_set::Context* ctx
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Constructs a specific Quartet. */
        Quartet
        (
            double   width,
            double   height,
            double   length,
            double   width_i,
            double   length_i,
            double   cross_N,
            double   cross_d1,
            double   cross_h,
            double   cross_dN,
            const Block_set::Context* ctx
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Constructs a Quartet by copying another. */
        Quartet(const Quartet& b);

        /** @brief Copies a Quartet into this one. */
        Quartet& operator= (const Quartet& b);

        /** @brief Clones this Quartet. */
        Quartet* clone() const;

        /** @brief Creates a special copy of this Pair. */
        Quartet* clone_and_swap() const throw (jwscxx::base::Arg_error);

        /** @brief Reads this Quartet from an input stream. */
        void read(std::istream& in) throw (jwscxx::base::IO_error,
                jwscxx::base::Arg_error);

        /** @brief Writes this Quartet to an output stream. */
        void write(std::ostream& out) const throw (jwscxx::base::IO_error);

        double inner_block_size(uint32_t i) const;

        /** @brief Returns the position of an inner block. */
        void inner_block_posn_v(jwsc::Vector_d** p_out, uint32_t i) const;

        void inner_block_ctr_attach_pt_v
        (
            jwsc::Vector_d** p_out, 
            uint32_t         i, 
            uint32_t         j
        )
        const;

        void cross_block_posn_v(jwsc::Vector_d** bp_out, uint32_t i) const;

        /*@}*/


        /**
         * @name Attachment
         *
         * Component attachment members.
         * @{
         */

        /** 
         * @brief Asserts consitency among attachments and the graphical
         * representation.
         */
        void assert_consistency();

        /*@}*/


        void sample_var_param_switch_proposal
        (
            jwsc::Vector_d** p_out,
            const Pair*      bp
        ) 
        const;


        /**
         * @name Selection
         *
         * Random selection of block set dependent components and attach points.
         * @{
         */

        static uint32_t attachable_block_to_quartet
        (
            const std::vector<Component*>& components,
            bool                           height_only
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t attachable_block_to_quartet_attach_pt
        (
            const Quartet* b,
            bool           height_only
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t detachable_quartet_from_block
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t detachable_block_from_quartet
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t splitable_block_to_quartet
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t splitable_block_to_quartet_attach_pt
        (
            const Quartet* b
        )
        throw (jwscxx::base::Arg_error);

        static uint32_t mergeable_quartet_from_block
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t mergeable_quartet_from_block_attach_pt
        (
            const Quartet* b1
        )
        throw (jwscxx::base::Arg_error);

        static uint32_t mergeable_block_from_quartet
        (
            const std::vector<Component*>& components,
            bool                           height_only
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t mergeable_block_from_quartet_attach_pt
        (
            const Block* b,
            bool         height_only
        )
        throw (jwscxx::base::Arg_error);

        static uint32_t switchable_block_to_quartet
        (
            const std::vector<Component*>& components,
            bool                           height_only
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t switchable_quartet_to_block
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        static uint32_t switchable_quartet_to_pair
        (
            const std::vector<Component*>& components
        ) 
        throw (jwscxx::base::Arg_error);

        /*@}*/


    private:

        /**
         * @name cross
         *
         * Cross block parameter members.
         * @{
         */

        /** @brief Adds a cross block to this block set. */
        void add_cross_block();

        /** @brief Removes a cross block from this block set. */
        void remove_cross_block();

        double cross_block_size(uint32_t i, uint32_t whl) const;

        /*@}*/


        /**
         * @name Representation
         *
         * Internal representation members.
         * @{
         */

        /** @brief Adds the internal block parameter dependencies. */
        void add_inner_deps();

        /** @brief Adds the internal block parameter conditionals. */
        void add_inner_conds();

        /** @brief Removes the internal block parameter dependencies. */
        void remove_inner_deps();

        /** @brief Removes the internal block parameter conditionals. */
        void remove_inner_conds();

        /*@}*/


    friend class Object;
};


#endif
