/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Basic Object structure component.
 */


#include <config.h>

#include <iostream>
#include <istream>
#include <vector>
#include <typeinfo>
#include <algorithm>
#include <cassert>
#include <cmath>
#include <cstring>

#if defined JWSCXX_HAVE_OPENGL
#include <GL/gl.h>
#elif defined JWSCXX_HAVE_OPENGL_FRAMEWORK
#include <OpenGL/OpenGL.h>
#endif

#include <inttypes.h>

#include <jwsc/base/limits.h>
#include <jwsc/vector/vector.h>
#include <jwsc/vector/vector_math.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/matrix/matrix_math.h>
#include <jwsc/prob/pdf.h>
#include <jwsc/prob/mv_pdf.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/graphics/renderable.h>
#include <jwsc++/graphics/polygon.h>
#include <jwsc++/graphics/polymesh.h>

#include "model.h"
#include "view.h"
#include "object.h"
#include "component.h"


/* -------------------------  Component::Context  --------------------------- */

Component::Context::Context() : Model::Context()
{
}

Component::Context::Context(const Component::Context& ctx) 
    : Model::Context(ctx)
{
    *this = ctx;
}

Component::Context::~Context()
{
}

Component::Context& Component::Context::operator= (const Component::Context& ctx)
{
    Model::Context::operator=(ctx);

    return *this;
}

/* ========================================================================== */









/* -------------------  Component::Parameter::Dependent  -------------------- */

Component::Parameter::Dependent::Dependent(Component::Parameter* p)
    : Pointer(p)
{
}

Component::Parameter::Dependent::Dependent
(
    const Component::Parameter::Dependent& d
)
: Pointer(d)
{
}

Component::Parameter::Dependent::~Dependent()
{
}

Component::Parameter* Component::Parameter::Dependent::ptr() const
{
    Component::Parameter* p;
    assert(p = dynamic_cast<Component::Parameter*>(Pointer::ptr()));
    return p;
}

void Component::Parameter::Dependent::ptr(Component::Parameter* p)
{
    Pointer::ptr(p);
}

/** 
 * Adds the value of @em dx to this dependent parameter.
 *
 * @param  dx  Value to modify this Dependent with.
 */
void Component::Parameter::Dependent::modify_and_shift_deps(double dx)
{
    Parameter* p = ptr();
    p->value(p->value() + dx);
    p->model()->graphical_rep_outdated();

    // DEBUG
    assert(p->color() == 1);

    for (size_t i = 0; i < p->num_shift_deps(); i++)
    {
        p->shift_dep(i)->ptr()->inc_color();
    }

    for (size_t i = 0; i < p->num_shift_deps(); i++)
    {
        Dependent* d = p->shift_dep(i);
        Parameter* q = d->ptr();

        bool b = (dynamic_cast<Fancy_zero_dependent*>(d) || 
                  dynamic_cast<Fancy_pos_dependent*>(d) || 
                  dynamic_cast<Fancy_neg_dependent*>(d) ||
                  dynamic_cast<Fancy_pos_half_dependent*>(d) || 
                  dynamic_cast<Fancy_neg_half_dependent*>(d));

        if (!b && p->color() == q->color())
        {
            d->modify_and_shift_deps(dx);
        }
    }

    for (size_t i = 0; i < p->num_shift_deps(); i++)
    {
        p->shift_dep(i)->ptr()->dec_color();
    }
}

/** 
 * Adds the value of @em dx to this dependent parameter.
 *
 * @param  dx  Value to modify this Dependent with.
 */
void Component::Parameter::Dependent::modify_and_scale_deps(double dx)
{
    Parameter* p = ptr();
    p->value(p->value() + dx);
    p->model()->graphical_rep_outdated();

    // DEBUG
    assert(p->color() == 1);

    for (size_t i = 0; i < p->num_scale_deps(); i++)
    {
        p->scale_dep(i)->ptr()->inc_color();
    }

    for (size_t i = 0; i < p->num_scale_deps(); i++)
    {
        Dependent* d = p->scale_dep(i);
        Parameter* q = d->ptr();

        bool b = (dynamic_cast<Fancy_zero_dependent*>(d) || 
                  dynamic_cast<Fancy_pos_dependent*>(d) || 
                  dynamic_cast<Fancy_neg_dependent*>(d) ||
                  dynamic_cast<Fancy_pos_half_dependent*>(d) || 
                  dynamic_cast<Fancy_neg_half_dependent*>(d));

        if (!b && p->color() == q->color())
        {
            d->modify_and_scale_deps(dx);
        }
    }

    for (size_t i = 0; i < p->num_scale_deps(); i++)
    {
        p->scale_dep(i)->ptr()->dec_color();
    }
}

/** 
 * Tests adding the value of @em dx to this dependent parameter.
 *
 * @param  dx  Value to test modifying this Dependent with.
 */
void Component::Parameter::Dependent::test_modify_and_shift_deps(double dx)
const throw (jwscxx::base::Arg_error)
{
    std::ostringstream ost;

    Parameter* p = ptr();
    p->test_value(p->value() + dx);

    // DEBUG
    assert(p->color() == 1);

    for (size_t i = 0; i < p->num_shift_deps(); i++)
    {
        p->shift_dep(i)->ptr()->inc_color();
    }

    for (size_t i = 0; i < p->num_shift_deps(); i++)
    {
        Dependent* d = p->shift_dep(i);
        Parameter* q = d->ptr();

        bool b = (dynamic_cast<Fancy_zero_dependent*>(d) || 
                  dynamic_cast<Fancy_pos_dependent*>(d) || 
                  dynamic_cast<Fancy_neg_dependent*>(d) ||
                  dynamic_cast<Fancy_pos_half_dependent*>(d) || 
                  dynamic_cast<Fancy_neg_half_dependent*>(d));

        if (!b && p->color() == q->color())
        {
            try 
            { 
                d->test_modify_and_shift_deps(dx); 
            }
            catch (jwscxx::base::Arg_error e)
            {
                ost << e.get_msg() << "  ";
                break;
            }
        }
    }

    for (size_t i = 0; i < p->num_shift_deps(); i++)
    {
        p->shift_dep(i)->ptr()->dec_color();
    }

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

/** 
 * Tests adding the value of @em dx to this dependent parameter.
 *
 * @param  dx  Value to modify this Dependent with.
 */
void Component::Parameter::Dependent::test_modify_and_scale_deps(double dx)
const throw (jwscxx::base::Arg_error)
{
    std::ostringstream ost;

    Parameter* p = ptr();
    p->test_value(p->value() + dx);

    // DEBUG
    assert(p->color() == 1);

    for (size_t i = 0; i < p->num_scale_deps(); i++)
    {
        p->scale_dep(i)->ptr()->inc_color();
    }

    for (size_t i = 0; i < p->num_scale_deps(); i++)
    {
        Dependent* d = p->scale_dep(i);
        Parameter* q = d->ptr();

        bool b = (dynamic_cast<Fancy_zero_dependent*>(d) || 
                  dynamic_cast<Fancy_pos_dependent*>(d) || 
                  dynamic_cast<Fancy_neg_dependent*>(d) ||
                  dynamic_cast<Fancy_pos_half_dependent*>(d) || 
                  dynamic_cast<Fancy_neg_half_dependent*>(d));

        if (!b && p->color() == q->color())
        {
            try 
            { 
                d->test_modify_and_scale_deps(dx); 
            }
            catch (jwscxx::base::Arg_error e)
            {
                ost << e.get_msg() << "  ";
                break;
            }
        }
    }

    for (size_t i = 0; i < p->num_scale_deps(); i++)
    {
        p->scale_dep(i)->ptr()->dec_color();
    }

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

/* ========================================================================== */






/* ----------------  Component::Parameter::Zero_dependent  ------------------ */

Component::Parameter::Zero_dependent::Zero_dependent
(
    Component::Parameter* p
)
: Dependent(p)
{
}

Component::Parameter::Zero_dependent::Zero_dependent
(
    const Component::Parameter::Zero_dependent& d
)
: Dependent(d)
{
}

Component::Parameter::Zero_dependent::~Zero_dependent()
{
}

Component::Parameter::Zero_dependent* 
Component::Parameter::Zero_dependent::clone() const
{
    return new Zero_dependent(*this);
}

const char* Component::Parameter::Zero_dependent::type_name() const
{
    return typeid(*this).name();
}

void Component::Parameter::Zero_dependent::modify_and_shift_deps
(
    double dx
)
{
    //Dependent::modify_and_shift_deps(0.0);
}

void Component::Parameter::Zero_dependent::modify_and_scale_deps
(
    double dx
)
{
    //Dependent::modify_and_scale_deps(0.0);
}

void Component::Parameter::Zero_dependent::test_modify_and_shift_deps
(
    double dx
)
const throw (jwscxx::base::Arg_error)
{
    //Dependent::test_modify_and_shift_deps(0.0);
}

void Component::Parameter::Zero_dependent::test_modify_and_scale_deps
(
    double dx
)
const throw (jwscxx::base::Arg_error)
{
    //Dependent::test_modify_and_scale_deps(0.0);
}

/* ========================================================================== */






/* -----------------  Component::Parameter::Pos_dependent  ------------------ */

Component::Parameter::Pos_dependent::Pos_dependent(Component::Parameter* p)
    : Dependent(p)
{
}

Component::Parameter::Pos_dependent::Pos_dependent
(
    const Component::Parameter::Pos_dependent& d
)
: Dependent(d)
{
}

Component::Parameter::Pos_dependent::~Pos_dependent()
{
}

Component::Parameter::Pos_dependent* Component::Parameter::Pos_dependent::clone() 
    const
{
    return new Pos_dependent(*this);
}

const char* Component::Parameter::Pos_dependent::type_name() const
{
    return typeid(*this).name();
}

void Component::Parameter::Pos_dependent::modify_and_shift_deps(double dx)
{
    Dependent::modify_and_shift_deps(dx);
}

void Component::Parameter::Pos_dependent::modify_and_scale_deps(double dx)
{
    Dependent::modify_and_scale_deps(dx);
}

void Component::Parameter::Pos_dependent::test_modify_and_shift_deps
(
    double dx
)
const throw (jwscxx::base::Arg_error)
{
    Dependent::test_modify_and_shift_deps(dx);
}

void Component::Parameter::Pos_dependent::test_modify_and_scale_deps
(
    double dx
)
const throw (jwscxx::base::Arg_error)
{
    Dependent::test_modify_and_scale_deps(dx);
}

/* ========================================================================== */






/* --------------  Component::Parameter::Pos_half_dependent  ---------------- */

Component::Parameter::Pos_half_dependent::Pos_half_dependent
(
    Component::Parameter* p
)
: Pos_dependent(p)
{
}

Component::Parameter::Pos_half_dependent::Pos_half_dependent
(
    const Component::Parameter::Pos_half_dependent& d
)
: Pos_dependent(d)
{
}

Component::Parameter::Pos_half_dependent::~Pos_half_dependent()
{
}

Component::Parameter::Pos_half_dependent* 
Component::Parameter::Pos_half_dependent::clone() const
{
    return new Pos_half_dependent(*this);
}

const char* Component::Parameter::Pos_half_dependent::type_name() const
{
    return typeid(*this).name();
}

void Component::Parameter::Pos_half_dependent::modify_and_shift_deps(double dx)
{
    Pos_dependent::modify_and_shift_deps(0.5*dx);
}

void Component::Parameter::Pos_half_dependent::modify_and_scale_deps(double dx)
{
    Pos_dependent::modify_and_scale_deps(0.5*dx);
}

void Component::Parameter::Pos_half_dependent::test_modify_and_shift_deps
(
    double dx
)
const throw (jwscxx::base::Arg_error)
{
    Pos_dependent::test_modify_and_shift_deps(0.5*dx);
}

void Component::Parameter::Pos_half_dependent::test_modify_and_scale_deps
(
    double dx
)
const throw (jwscxx::base::Arg_error)
{
    Pos_dependent::test_modify_and_scale_deps(0.5*dx);
}

/* ========================================================================== */






/* -----------  Component::Parameter::Fancy_zero_dependent  ----------------- */

Component::Parameter::Fancy_zero_dependent::Fancy_zero_dependent
(
    Component::Parameter* p
)
: Zero_dependent(p)
{
}

Component::Parameter::Fancy_zero_dependent::Fancy_zero_dependent
(
    const Component::Parameter::Fancy_zero_dependent& d
)
: Zero_dependent(d)
{
}

Component::Parameter::Fancy_zero_dependent::~Fancy_zero_dependent()
{
}

Component::Parameter::Fancy_zero_dependent* 
Component::Parameter::Fancy_zero_dependent::clone() const
{
    return new Fancy_zero_dependent(*this);
}

const char* Component::Parameter::Fancy_zero_dependent::type_name() const
{
    return typeid(*this).name();
}

/* ========================================================================== */






/* -------------  Component::Parameter::Fancy_pos_dependent  ---------------- */

Component::Parameter::Fancy_pos_dependent::Fancy_pos_dependent
(
    Component::Parameter* p
)
: Pos_dependent(p)
{
}

Component::Parameter::Fancy_pos_dependent::Fancy_pos_dependent
(
    const Component::Parameter::Fancy_pos_dependent& d
)
: Pos_dependent(d)
{
}

Component::Parameter::Fancy_pos_dependent::~Fancy_pos_dependent()
{
}

Component::Parameter::Fancy_pos_dependent* 
Component::Parameter::Fancy_pos_dependent::clone() const
{
    return new Fancy_pos_dependent(*this);
}

const char* Component::Parameter::Fancy_pos_dependent::type_name() const
{
    return typeid(*this).name();
}

/* ========================================================================== */






/* -----------  Component::Parameter::Fancy_pos_half_dependent  ------------- */

Component::Parameter::Fancy_pos_half_dependent::Fancy_pos_half_dependent
(
    Component::Parameter* p
)
: Pos_half_dependent(p)
{
}

Component::Parameter::Fancy_pos_half_dependent::Fancy_pos_half_dependent
(
    const Component::Parameter::Fancy_pos_half_dependent& d
)
: Pos_half_dependent(d)
{
}

Component::Parameter::Fancy_pos_half_dependent::~Fancy_pos_half_dependent()
{
}

Component::Parameter::Fancy_pos_half_dependent* 
Component::Parameter::Fancy_pos_half_dependent::clone() const
{
    return new Fancy_pos_half_dependent(*this);
}

const char* Component::Parameter::Fancy_pos_half_dependent::type_name() const
{
    return typeid(*this).name();
}

/* ========================================================================== */






/* -----------------  Component::Parameter::Neg_dependent  ------------------ */

Component::Parameter::Neg_dependent::Neg_dependent(Component::Parameter* p)
    : Dependent(p)
{
}

Component::Parameter::Neg_dependent::Neg_dependent
(
    const Component::Parameter::Neg_dependent& d
)
: Dependent(d)
{
}

Component::Parameter::Neg_dependent::~Neg_dependent()
{
}

Component::Parameter::Neg_dependent* 
Component::Parameter::Neg_dependent::clone() 
    const
{
    return new Neg_dependent(*this);
}

const char* Component::Parameter::Neg_dependent::type_name() const
{
    return typeid(*this).name();
}

void Component::Parameter::Neg_dependent::modify_and_shift_deps(double dx)
{
    Dependent::modify_and_shift_deps(-dx);
}

void Component::Parameter::Neg_dependent::modify_and_scale_deps(double dx)
{
    Dependent::modify_and_scale_deps(-dx);
}

void Component::Parameter::Neg_dependent::test_modify_and_shift_deps
(
    double dx
)
const throw (jwscxx::base::Arg_error)
{
    Dependent::test_modify_and_shift_deps(-dx);
}

void Component::Parameter::Neg_dependent::test_modify_and_scale_deps
(
    double dx
)
const throw (jwscxx::base::Arg_error)
{
    Dependent::test_modify_and_scale_deps(-dx);
}

/* ========================================================================== */






/* --------------  Component::Parameter::Neg_half_dependent  ---------------- */

Component::Parameter::Neg_half_dependent::Neg_half_dependent
(
    Component::Parameter* p
)
: Neg_dependent(p)
{
}

Component::Parameter::Neg_half_dependent::Neg_half_dependent
(
    const Component::Parameter::Neg_half_dependent& d
)
: Neg_dependent(d)
{
}

Component::Parameter::Neg_half_dependent::~Neg_half_dependent()
{
}

Component::Parameter::Neg_half_dependent* 
Component::Parameter::Neg_half_dependent::clone() const
{
    return new Neg_half_dependent(*this);
}

const char* Component::Parameter::Neg_half_dependent::type_name() const
{
    return typeid(*this).name();
}

void Component::Parameter::Neg_half_dependent::modify_and_shift_deps
(
    double dx
)
{
    Neg_dependent::modify_and_shift_deps(0.5*dx);
}

void Component::Parameter::Neg_half_dependent::modify_and_scale_deps
(
    double dx
)
{
    Neg_dependent::modify_and_scale_deps(0.5*dx);
}

void Component::Parameter::Neg_half_dependent::test_modify_and_shift_deps
(
    double dx
)
const throw (jwscxx::base::Arg_error)
{
    Neg_dependent::test_modify_and_shift_deps(0.5*dx);
}

void Component::Parameter::Neg_half_dependent::test_modify_and_scale_deps
(
    double dx
)
const throw (jwscxx::base::Arg_error)
{
    Neg_dependent::test_modify_and_scale_deps(0.5*dx);
}

/* ========================================================================== */






/* -------------  Component::Parameter::Fancy_neg_dependent  ---------------- */

Component::Parameter::Fancy_neg_dependent::Fancy_neg_dependent
(
    Component::Parameter* p
)
: Neg_dependent(p)
{
}

Component::Parameter::Fancy_neg_dependent::Fancy_neg_dependent
(
    const Component::Parameter::Fancy_neg_dependent& d
)
: Neg_dependent(d)
{
}

Component::Parameter::Fancy_neg_dependent::~Fancy_neg_dependent()
{
}

Component::Parameter::Fancy_neg_dependent* 
Component::Parameter::Fancy_neg_dependent::clone() const
{
    return new Fancy_neg_dependent(*this);
}

const char* Component::Parameter::Fancy_neg_dependent::type_name() const
{
    return typeid(*this).name();
}

/* ========================================================================== */






/* ---------  Component::Parameter::Fancy_neg_half_dependent  --------------- */

Component::Parameter::Fancy_neg_half_dependent::Fancy_neg_half_dependent
(
    Component::Parameter* p
)
: Neg_half_dependent(p)
{
}

Component::Parameter::Fancy_neg_half_dependent::Fancy_neg_half_dependent
(
    const Component::Parameter::Fancy_neg_half_dependent& d
)
: Neg_half_dependent(d)
{
}

Component::Parameter::Fancy_neg_half_dependent::~Fancy_neg_half_dependent()
{
}

Component::Parameter::Fancy_neg_half_dependent* 
Component::Parameter::Fancy_neg_half_dependent::clone() const
{
    return new Fancy_neg_half_dependent(*this);
}

const char* Component::Parameter::Fancy_neg_half_dependent::type_name() const
{
    return typeid(*this).name();
}

/* ========================================================================== */






/* -------------------------  Component::Parameter  ------------------------- */

Component::Parameter::Parameter
(
    size_t                           id,
    const char*                      name, 
    double                           val, 
    const Model::Parameter::Context* ctx,
    Model*                           model
)
throw (jwscxx::base::Arg_error)
    : Model::Parameter(id, name, val, ctx, model)
{
    _color = 0;
}

Component::Parameter::Parameter(const Component::Parameter& p)
    : Model::Parameter(p)
{
    *this = p;
}

Component::Parameter::~Parameter()
{
    for (size_t i = 0; i < _shift_deps.size(); i++)
    {
        delete _shift_deps[ i ];
    }
    for (size_t i = 0; i < _scale_deps.size(); i++)
    {
        delete _scale_deps[ i ];
    }
    for (size_t i = 0; i < _conds.size(); i++)
    {
        delete _conds[ i ];
    }
}

/** 
 * @note Assignment copies the parameter pointers; this Parameter will have the
 * same dependecies and conditional as @em p.
 */
Component::Parameter& Component::Parameter::operator=
(
    const Component::Parameter& p
)
{
    Model::Parameter::operator=(p);

    for (size_t i = 0; i < _shift_deps.size(); i++)
    {
        delete _shift_deps[ i ];
    }
    for (size_t i = 0; i < _scale_deps.size(); i++)
    {
        delete _scale_deps[ i ];
    }
    for (size_t i = 0; i < _conds.size(); i++)
    {
        delete _conds[ i ];
    }

    _shift_deps = p._shift_deps;
    _scale_deps = p._scale_deps;
    _conds      = p._conds;
    _color      = p._color;

    for (size_t i = 0; i < _shift_deps.size(); i++)
    {
        _shift_deps[ i ] = p._shift_deps[ i ]->clone();
    }
    for (size_t i = 0; i < _scale_deps.size(); i++)
    {
        _scale_deps[ i ] = p._scale_deps[ i ]->clone();
    }
    for (size_t i = 0; i < _conds.size(); i++)
    {
        _conds[ i ] = p._conds[ i ]->clone();
    }

    return *this;
}

Component::Parameter* Component::Parameter::clone() const
{
    return new Parameter(*this);
}

/**
 * The file format is ASCII.
 *
 * @param  in  Input stream to read this Component::Parameter from.
 *
 * @throw  jwscxx::base::IO_error   Could not read from @em in.
 * @throw  jwscxx::base::Arg_error  Invalid arguments in file to read from.
 */
void Component::Parameter::read(std::istream& in) 
    throw (jwscxx::base::IO_error, jwscxx::base::Arg_error)
{
    Model::Parameter::read(in);

    read_dependents(in);
    read_conditionals(in);
}

void Component::Parameter::read_dependents(std::istream& in) 
    throw (jwscxx::base::IO_error, jwscxx::base::Arg_error)
{
    using std::ostringstream;
    using std::istringstream;
    using jwscxx::base::Arg_error;

    static const char* dnames[] = {" Shift Dependents", " Scale Dependents"};

    std::vector<Dependent*>* deps[2];

    deps[0] = &_shift_deps;
    deps[1] = &_scale_deps;

    for (uint32_t i = 0; i < 2; i++)
    {
        std::vector<Dependent*>* deps_i = deps[ i ];

        std::ostringstream field_name;
        field_name << "Num " << name() << dnames[ i ];
        const char* field_value;
        size_t N;

        if (!(field_value = read_field_value(in, field_name.str().c_str())))
        {
            ostringstream ost;
            ost << "Missing number: " << field_name.str();
            throw Arg_error(ost.str());
        }
        istringstream ist(field_value);
        ist >> N;
        if (ist.fail())
        {
            ostringstream ost;
            ost << field_name.str() << ": Invalid number";
            throw Arg_error(ost.str());
        }

        if (N == 0)
            continue;

        for (size_t n = 0; n < deps_i->size(); n++)
        {
            delete deps_i->at(n);
        }
        deps_i->clear();

        for (size_t n = 0; n < N; n++)
        {
            Dependent* d = 0;
            const char* type_name;

            if (!(type_name = read_field_value(in, "Type")))
            {
                throw Arg_error("Missing Type field");
            }

            if (strcmp(type_name, typeid(Zero_dependent).name()) == 0)
            {
                d = new Zero_dependent(0);
            }
            else if (strcmp(type_name, typeid(Pos_dependent).name()) == 0)
            {
                d = new Pos_dependent(0);
            }
            else if (strcmp(type_name, typeid(Pos_half_dependent).name()) == 0)
            {
                d = new Pos_half_dependent(0);
            }
            else if (strcmp(type_name, typeid(Fancy_zero_dependent).name()) 
                    == 0)
            {
                d = new Fancy_zero_dependent(0);
            }
            else if (strcmp(type_name, typeid(Fancy_pos_dependent).name()) == 0)
            {
                d = new Fancy_pos_dependent(0);
            }
            else if (strcmp(type_name, typeid(Fancy_pos_half_dependent).name()) 
                    == 0)
            {
                d = new Fancy_pos_half_dependent(0);
            }
            else if (strcmp(type_name, typeid(Neg_dependent).name()) == 0)
            {
                d = new Neg_dependent(0);
            }
            else if (strcmp(type_name, typeid(Neg_half_dependent).name()) == 0)
            {
                d = new Neg_half_dependent(0);
            }
            else if (strcmp(type_name, typeid(Fancy_neg_dependent).name()) == 0)
            {
                d = new Fancy_neg_dependent(0);
            }
            else if (strcmp(type_name, typeid(Fancy_neg_half_dependent).name()) 
                    == 0)
            {
                d = new Fancy_neg_half_dependent(0);
            }
            else
            {
                std::ostringstream ost;
                ost << "Unknown dependent type '" << type_name << "'";
                throw Arg_error(ost.str());
            }

            d->read(in);
            deps_i->push_back(d);
        }
    }
}

void Component::Parameter::read_conditionals(std::istream& in) 
    throw (jwscxx::base::IO_error, jwscxx::base::Arg_error)
{
    using std::ostringstream;
    using std::istringstream;
    using jwscxx::base::Arg_error;

    static const char* str = " Conditionals";

    ostringstream field_name;
    field_name << "Num " << name() << str;
    const char* field_value;
    size_t N;

    if (!(field_value = read_field_value(in, field_name.str().c_str())))
    {
        ostringstream ost;
        ost << "Missing number: " << field_name.str();
        throw Arg_error(ost.str());
    }
    istringstream ist(field_value);
    ist >> N;
    if (ist.fail())
    {
        ostringstream ost;
        ost << field_name.str() << ": Invalid number";
        throw Arg_error(ost.str());
    }
    ist.clear();

    if (N == 0) return;

    for (size_t n = 0; n < _conds.size(); n++)
    {
        delete _conds[ n ];
    }
    _conds.clear();

    for (size_t n = 0; n < N; n++)
    {
        Pointer* p = new Pointer(0);
        p->read(in);
        _conds.push_back(p);
    }
}

/**
 * The file format is ASCII.
 *
 * @param  out  Output stream to write this Component::Parameter to.
 *
 * @throw  jwscxx::base::IO_error  Could not write to @em out.
 */
void Component::Parameter::write(std::ostream& out) const
    throw (jwscxx::base::IO_error)
{
    Model::Parameter::write(out);

    write_dependents(out);
    write_conditionals(out);
}

void Component::Parameter::write_dependents(std::ostream& out) const
    throw (jwscxx::base::IO_error)
{
    static const char* dnames[] = {" Shift Dependents", " Scale Dependents"};

    const std::vector<Dependent*>* deps[2];

    deps[0] = &_shift_deps;
    deps[1] = &_scale_deps;

    for (uint32_t i = 0; i < 2; i++)
    {
        const std::vector<Dependent*>* deps_i = deps[ i ];

        size_t N = deps_i->size();
        out << "Num " << name() << dnames[ i ] << ": " << N << std::endl;

        for (size_t n = 0; n < N; n++)
        {
            out << "Type: " << deps_i->at(n)->type_name() << std::endl;
            deps_i->at(n)->write(out);
        }
    }
}

void Component::Parameter::write_conditionals(std::ostream& out) const
    throw (jwscxx::base::IO_error)
{
    static const char* str = " Conditionals";

    size_t N = _conds.size();
    out << "Num " << name() << str << ": " << N << std::endl;

    for (size_t n = 0; n < N; n++)
    {
        _conds[ n ]->write(out);
    }
}

/**
 * @param  value  New parameter value.
 *
 * @warning Aborts on invalid arguments.
 */
void Component::Parameter::value_shift_deps(double value)
{
    using namespace jwsc;

    // DEBUG
    assert(color() == 0);

    double dx = value - this->value();
    this->value(value);
    model()->graphical_rep_outdated();

    inc_color();
    for (size_t i = 0; i < num_shift_deps(); i++)
    {
        shift_dep(i)->ptr()->inc_color();
    }

    for (size_t i = 0; i < num_shift_deps(); i++)
    {
        shift_dep(i)->modify_and_shift_deps(dx);
    }

    for (size_t i = 0; i < num_shift_deps(); i++)
    {
        shift_dep(i)->ptr()->dec_color();
    }
    dec_color();

    // DEBUG
    assert(color() == 0);
}

/**
 * @param  value  New parameter value.
 *
 * @warning Aborts on invalid arguments.
 */
void Component::Parameter::value_scale_deps(double value)
{
    using namespace jwsc;

    // DEBUG
    assert(color() == 0);

    double dx = value - this->value();
    this->value(value);
    model()->graphical_rep_outdated();

    inc_color();
    for (size_t i = 0; i < num_scale_deps(); i++)
    {
        scale_dep(i)->ptr()->inc_color();
    }

    for (size_t i = 0; i < num_scale_deps(); i++)
    {
        scale_dep(i)->modify_and_scale_deps(dx);
    }

    for (size_t i = 0; i < num_scale_deps(); i++)
    {
        scale_dep(i)->ptr()->dec_color();
    }
    dec_color();

    // DEBUG
    assert(color() == 0);
}

void Component::Parameter::test_value_shift_deps(double value)
    throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    // DEBUG
    assert(color() == 0);

    std::ostringstream ost;

    double dx = value - this->value();
    test_value(value);

    inc_color();
    for (size_t i = 0; i < num_shift_deps(); i++)
    {
        shift_dep(i)->ptr()->inc_color();
    }

    for (size_t i = 0; i < num_shift_deps(); i++)
    {
        try
        {
            shift_dep(i)->test_modify_and_shift_deps(dx);
        }
        catch (jwscxx::base::Arg_error e)
        {
            ost << e.get_msg() << "  ";
        }
    }

    for (size_t i = 0; i < num_shift_deps(); i++)
    {
        shift_dep(i)->ptr()->dec_color();
    }
    dec_color();

    // DEBUG
    assert(color() == 0);

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

void Component::Parameter::test_value_scale_deps(double value)
    throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    // DEBUG
    assert(color() == 0);

    std::ostringstream ost;

    double dx = value - this->value();
    test_value(value);

    inc_color();
    for (size_t i = 0; i < num_scale_deps(); i++)
    {
        scale_dep(i)->ptr()->inc_color();
    }

    for (size_t i = 0; i < num_scale_deps(); i++)
    {
        try
        {
            scale_dep(i)->test_modify_and_scale_deps(dx);
        }
        catch (jwscxx::base::Arg_error e)
        {
            ost << e.get_msg() << "  ";
        }
    }

    for (size_t i = 0; i < num_scale_deps(); i++)
    {
        scale_dep(i)->ptr()->dec_color();
    }
    dec_color();

    // DEBUG
    assert(color() == 0);

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

size_t Component::Parameter::num_shift_deps() const
{
    return _shift_deps.size();
}

size_t Component::Parameter::num_scale_deps() const
{
    return _scale_deps.size();
}

Component::Parameter::Dependent* Component::Parameter::shift_dep(size_t i) const
{
    assert(num_shift_deps() > 0 && i < num_shift_deps());
    return _shift_deps[ i ];
}

Component::Parameter::Dependent* Component::Parameter::scale_dep(size_t i) const
{
    assert(num_scale_deps() > 0 && i < num_scale_deps());
    return _scale_deps[ i ];
}

void Component::Parameter::add_shift_dep
(
    const Component::Parameter::Dependent& d
)
{
    _shift_deps.push_back(d.clone());
}

void Component::Parameter::add_scale_dep
(
    const Component::Parameter::Dependent& d
)
{
    _scale_deps.push_back(d.clone());
}

void Component::Parameter::remove_shift_dep
(
    const Component::Parameter::Dependent& d
)
{
    std::vector<Dependent*>::iterator it;

    for (it = _shift_deps.begin(); it != _shift_deps.end(); it++)
    {
        Dependent* dd = *it;

        if (d.ptr() == dd->ptr() &&
            strcmp(d.type_name(), dd->type_name()) == 0)
        {
            _shift_deps.erase(it);
            delete dd;
            return;
        }
    }

    abort();
}

void Component::Parameter::remove_scale_dep
(
    const Component::Parameter::Dependent& d
)
{
    std::vector<Dependent*>::iterator it;

    for (it = _scale_deps.begin(); it != _scale_deps.end(); it++)
    {
        Dependent* dd = *it;

        if (d.ptr() == dd->ptr() &&
            strcmp(d.type_name(), dd->type_name()) == 0)
        {
            _scale_deps.erase(it);
            delete dd;
            return;
        }
    }

    abort();
}

bool Component::Parameter::has_shift_dep
(
    const Component::Parameter::Dependent& d
)
const
{
    std::vector<Dependent*>::const_iterator it;

    for (it = _shift_deps.begin(); it != _shift_deps.end(); it++)
    {
        const Dependent* dd = *it;

        if (d.ptr() == dd->ptr() &&
            strcmp(d.type_name(), dd->type_name()) == 0)
        {
            return true;
        }
    }

    return false;
}

bool Component::Parameter::has_scale_dep
(
    const Component::Parameter::Dependent& d
)
const
{
    std::vector<Dependent*>::const_iterator it;

    for (it = _scale_deps.begin(); it != _scale_deps.end(); it++)
    {
        const Dependent* dd = *it;

        if (d.ptr() == dd->ptr() &&
            strcmp(d.type_name(), dd->type_name()) == 0)
        {
            return true;
        }
    }

    return false;
}

size_t Component::Parameter::num_conds() const
{
    return _conds.size();
}

Model::Parameter::Pointer* Component::Parameter::cond(size_t i) const
{
    assert(i < num_conds());
    return _conds[ i ];
}

void Component::Parameter::add_cond
(
    const Model::Parameter::Pointer& p
)
{
    _conds.push_back(p.clone());
}

void Component::Parameter::remove_cond
(
    const Component::Parameter::Pointer& p
)
{
    std::vector<Pointer*>::iterator it;

    for (it = _conds.begin(); it != _conds.end(); it++)
    {
        Pointer* pp = *it;

        if (p.ptr() == pp->ptr() &&
            strcmp(p.type_name(), pp->type_name()) == 0)
        {
            _conds.erase(it);
            delete pp;
            return;
        }
    }

    abort();
}

bool Component::Parameter::has_cond
(
    const Component::Parameter::Pointer& p
)
const
{
    std::vector<Pointer*>::const_iterator it;

    for (it = _conds.begin(); it != _conds.end(); it++)
    {
        const Pointer* pp = *it;

        if (p.ptr() == pp->ptr() &&
            strcmp(p.type_name(), pp->type_name()) == 0)
        {
            return true;
        }
    }

    return false;
}

/**
 * @param  models  Collection of Model pointers to use for serializing.
 */
void Component::Parameter::serialize(const std::vector<Model*>& models)
{
    Model::Parameter::serialize(models);

    for (size_t i = 0; i < _shift_deps.size(); i++)
    {
        _shift_deps[ i ]->serialize(models);
    }
    for (size_t i = 0; i < _scale_deps.size(); i++)
    {
        _scale_deps[ i ]->serialize(models);
    }
    for (size_t i = 0; i < _conds.size(); i++)
    {
        _conds[ i ]->serialize(models);
    }
}

/**
 * @param  models  Collection of Model pointers to use for deserializing.
 */
void Component::Parameter::deserialize(const std::vector<Model*>& models)
{
    Model::Parameter::deserialize(models);

    for (size_t i = 0; i < _shift_deps.size(); i++)
    {
        _shift_deps[ i ]->deserialize(models);
    }
    for (size_t i = 0; i < _scale_deps.size(); i++)
    {
        _scale_deps[ i ]->deserialize(models);
    }
    for (size_t i = 0; i < _conds.size(); i++)
    {
        _conds[ i ]->deserialize(models);
    }
}

/**
 * @param  new_models  Collection of pointers to update this Parameter to.
 * @param  old_models  Collection of pointers to update this Parameter from.
 */
void Component::Parameter::update_pointers
(
    const std::vector<Model*>&       new_models,
    const std::vector<const Model*>& old_models
)
{
    Model::Parameter::update_pointers(new_models, old_models);

    for (size_t i = 0; i < _shift_deps.size(); i++)
    {
        _shift_deps[ i ]->update_pointers(new_models, old_models);
    }
    for (size_t i = 0; i < _scale_deps.size(); i++)
    {
        _scale_deps[ i ]->update_pointers(new_models, old_models);
    }
    for (size_t i = 0; i < _conds.size(); i++)
    {
        _conds[ i ]->update_pointers(new_models, old_models);
    }
}

uint32_t Component::Parameter::color() const
{
    return _color;
}

void Component::Parameter::inc_color()
{
    _color++;
}

/** Aborts if the color is already zero. */
void Component::Parameter::dec_color()
{
    assert(_color > 0);
    _color--;
}

/* ========================================================================== */








/* ---------------  Component::Paramgroup::Context::Density  ---------------- */

/**
 * If the value @em mu is outside the truncated region [mu_min,mu_max], the
 * density log of the density is a very small negative number; the pdf evaluates
 * extremely close to zero, which should not be attempted because precision will
 * not support it.
 *
 * @param  mu         Mean category parameters.
 * @param  mu_min     Minimum mean value.
 * @param  mu_max     Maximum mean value.
 * @param  sigma      Standard deviation of the category parameters.
 */
double Component::Paramgroup::Context::Density::trunc_log_pdf
(
    const jwsc::Vector_d* mu, 
    const jwsc::Vector_d* mu_min, 
    const jwsc::Vector_d* mu_max,
    const jwsc::Vector_d* sigma
) 
const
{
    using namespace jwsc;

    uint32_t D = mu->num_elts;

    assert(D == mu_min->num_elts);
    assert(mu_min->num_elts == mu_max->num_elts);

    for (uint32_t d = 0; d < D; d++)
    {
        double min = mu_min->elts[ d ];
        double max = mu_max->elts[ d ];
        double x   = mu->elts[ d ];
        assert(min <= max);

        double y = ((x < min) ? fabs(x - min) : fabs(x - max)) / (max - min);
        return (y+1)*10*log(JWSC_MIN_LOG_ARG);
    }

    return log_pdf(mu, sigma);
}

/* ========================================================================== */








/* ----------  Component::Paramgroup::Context::Iso_gauss_gamma  ------------- */

/**
 * @param  pi      Weight of each mixture component.
 * @param  mu      Gaussian means. Each mixture component is indexed by row.
 * @param  lambda  Isotropic variance scale for each mixture component.
 * @param  alpha   Gamma shape parameter.
 * @param  beta    Gamma scale parameter.
 *
 * @throw  jwscxx::base::Arg_error  Lambda, alpha, or beta too small.
 */
Component::Paramgroup::Context::Iso_gauss_gamma::Iso_gauss_gamma
(
    const jwsc::Vector_d* pi, 
    const jwsc::Matrix_d* mu, 
    double                lambda,
    double                alpha,
    double                beta
)
throw (jwscxx::base::Arg_error)
{
    assert(pi->num_elts == mu->num_rows);

    if (lambda < 1.0e-16)
    {
        throw jwscxx::base::Arg_error("Lambda too small");
    }
    if (alpha < 1.0e-16)
    {
        throw jwscxx::base::Arg_error("Alpha too small");
    }
    if (beta < 1.0e-16)
    {
        throw jwscxx::base::Arg_error("Beta too small");
    }

    _pi = 0;
    copy_vector_d(&_pi, pi);
    normalize_vector_sum_d(&_pi, _pi);

    _mu = 0;
    copy_matrix_d(&_mu, mu);

    _lambda = lambda;
    _alpha  = alpha;
    _beta   = beta;
}

/**
 * @param  d  Density to copy into this one.
 */
Component::Paramgroup::Context::Iso_gauss_gamma::Iso_gauss_gamma
(
    const Component::Paramgroup::Context::Iso_gauss_gamma& d
)
{
    _pi = 0;
    _mu = 0;

    *this = d;
}

Component::Paramgroup::Context::Iso_gauss_gamma::~Iso_gauss_gamma()
{
    free_vector_d(_pi);
    free_matrix_d(_mu);
}

Component::Paramgroup::Context::Iso_gauss_gamma* 
Component::Paramgroup::Context::Iso_gauss_gamma::clone() const
{
    return new Iso_gauss_gamma(*this);
}

Component::Paramgroup::Context::Iso_gauss_gamma& 
Component::Paramgroup::Context::Iso_gauss_gamma::operator= 
(
    const Component::Paramgroup::Context::Iso_gauss_gamma& ctx
)
{
    copy_vector_d(&_pi, ctx._pi);
    copy_matrix_d(&_mu, ctx._mu);

    _lambda = ctx._lambda;
    _alpha  = ctx._alpha;
    _beta   = ctx._beta;

    return *this;
}

/**
 * @param  mu     Mean to compute the density of.
 * @param  sigma  Standard deviation to compute the density of.
 */
double Component::Paramgroup::Context::Iso_gauss_gamma::pdf
(
    const jwsc::Vector_d* mu,
    const jwsc::Vector_d* sigma
)
const
{
    using namespace jwsc;

    uint32_t D = _mu->num_cols;
    uint32_t C = _mu->num_rows;

    assert(mu->num_elts == D);
    assert(sigma->num_elts == D);

    double p = 1.0;

    for (uint32_t d = 0; d < D; d++)
    {
        double sigma_sqr = sigma->elts[ d ]*sigma->elts[ d ];
        double g = 0;

        for (uint32_t c = 0; c < C; c++)
        {
            g += _pi->elts[ c ] * gaussian_pdf_d(_mu->elts[ c ][ d ], 
                    sqrt(_lambda*sigma_sqr), mu->elts[ d ]);
        }

        p *= g * gamma_pdf_d(_alpha, _beta, 1.0 / sigma_sqr);
    }

    return p;
}

/**
 * @param  mu     Mean to compute the density of.
 * @param  sigma  Standard deviation to compute the density of.
 */
double Component::Paramgroup::Context::Iso_gauss_gamma::log_pdf
(
    const jwsc::Vector_d* mu,
    const jwsc::Vector_d* sigma
)
const
{
    using namespace jwsc;

    uint32_t D = _mu->num_cols;
    uint32_t C = _mu->num_rows;

    assert(mu->num_elts == D);
    assert(sigma->num_elts == D);

    double log_p = 0.0;

    for (uint32_t d = 0; d < D; d++)
    {
        double sigma_sqr = sigma->elts[ d ]*sigma->elts[ d ];
        double g = 0;

        for (uint32_t c = 0; c < C; c++)
        {
            g += _pi->elts[ c ] * gaussian_pdf_d(_mu->elts[ c ][ d ], 
                    sqrt(_lambda*sigma_sqr), mu->elts[ d ]);
        }

        double log_g = log((g < JWSC_MIN_LOG_ARG) ? JWSC_MIN_LOG_ARG : g);

        log_p += log_g + log_gamma_pdf_d(_alpha, _beta, 1.0 / sigma_sqr);
    }

    return log_p;
}

/**
 * @param  mu_out     Sampled mean.
 * @param  mu_min     Minimum value the sample can fall in.
 * @param  mu_max     Maximum value the sample can fall in.
 * @param  sigma_out  Sampled standard deviation.
 */
void Component::Paramgroup::Context::Iso_gauss_gamma::sample
(
    jwsc::Vector_d**      mu_out, 
    const jwsc::Vector_d* mu_min,
    const jwsc::Vector_d* mu_max,
    jwsc::Vector_d**      sigma_out
)
const
{
    using namespace jwsc;

    uint32_t D = _mu->num_cols;
    uint32_t C = _mu->num_rows;
    uint32_t c;

    double u = sample_uniform_pdf_d(0, 1);
    double p = 0;

    double tau;
    double tau_min = 1.0e-16;
    double tau_max = 6*sqrt(_alpha)/_beta;

    double mu_0, sigma_0;
    double gauss_min, gauss_max;

    assert(mu_min->num_elts == D);
    assert(mu_min->num_elts == mu_max->num_elts);

    create_vector_d(mu_out, D);
    create_vector_d(sigma_out, D);

    Vector_d* mu = *mu_out;
    Vector_d* sigma = *sigma_out;

    for (c = 0; c < C; c++)
    {
        p += _pi->elts[ c ];

        if (u < p)
            break;
    }
    assert(c != C);

    for (uint32_t d = 0; d < D; d++)
    {
        tau = sample_gamma_pdf_d(_alpha, _beta, tau_min, tau_max);
        sigma->elts[ d ] = 1.0 / sqrt(tau);

        mu_0 = _mu->elts[ c ][ d ];
        sigma_0 = sqrt(_lambda / tau);
        gauss_min = mu_0 - 6*sigma_0;
        gauss_max = mu_0 + 6*sigma_0;

        assert(mu_min->elts[ d ] < mu_max->elts[ d ]);
        mu->elts[ d ] = mu_min->elts[ d ] - 1;
        while (mu->elts[ d ] < mu_min->elts[ d ] ||
               mu->elts[ d ] > mu_max->elts[ d ])
        {
            mu->elts[ d ] = sample_gaussian_pdf_d(mu_0, sigma_0, gauss_min,
                    gauss_max);
        }
    }
}

/* ========================================================================== */







/* ----------  Component::Paramgroup::Context::UV_gauss_gamma  -------------- */

/**
 * @param  lambda  Varaince scale for each mixture component.
 * @param  alpha   Gamma shape for each mixture component.
 * @param  beta    Gamma scale for each mixture component.
 *
 * @throw  jwscxx::base::Arg_error  Lambda, alpha, or beta too small.
 */
Component::Paramgroup::Context::UV_gauss_gamma::UV_gauss_gamma
(
    double lambda,
    double alpha,
    double beta
)
throw (jwscxx::base::Arg_error)
{
    if (lambda < 1.0e-16)
    {
        throw jwscxx::base::Arg_error("Lambda too small");
    }
    if (alpha < 1.0e-16)
    {
        throw jwscxx::base::Arg_error("Alpha too small");
    }
    if (beta < 1.0e-16)
    {
        throw jwscxx::base::Arg_error("Beta too small");
    }

    _lambda = lambda;
    _alpha  = alpha;
    _beta   = beta;
}

/**
 * @param  d  Density to copy into this one.
 */
Component::Paramgroup::Context::UV_gauss_gamma::UV_gauss_gamma
(
    const Component::Paramgroup::Context::UV_gauss_gamma& d
)
{
    *this = d;
}

Component::Paramgroup::Context::UV_gauss_gamma::~UV_gauss_gamma()
{
}

Component::Paramgroup::Context::UV_gauss_gamma* 
Component::Paramgroup::Context::UV_gauss_gamma::clone() const
{
    return new UV_gauss_gamma(*this);
}

Component::Paramgroup::Context::UV_gauss_gamma& 
Component::Paramgroup::Context::UV_gauss_gamma::operator= 
(
    const Component::Paramgroup::Context::UV_gauss_gamma& ctx
)
{
    _lambda = ctx._lambda;
    _alpha  = ctx._alpha;
    _beta   = ctx._beta;

    return *this;
}

/**
 * @param  mu     Mean to compute the density of.
 * @param  sigma  Standard deviation to compute the density of.
 */
double Component::Paramgroup::Context::UV_gauss_gamma::pdf
(
    const jwsc::Vector_d* mu,
    const jwsc::Vector_d* sigma
)
const
{
    using namespace jwsc;

    uint32_t D = 2;

    assert(mu->num_elts == D);
    assert(sigma->num_elts == D);

    double p = 1.0;

    for (uint32_t d = 0; d < D; d++)
    {
        double sigma_sqr = sigma->elts[ d ]*sigma->elts[ d ];
        double g = 0;

        g += 0.5 * gaussian_pdf_d(0, sqrt(_lambda*sigma_sqr), mu->elts[ d ]);
        g += 0.5 * 1.0;

        p *= g * gamma_pdf_d(_alpha, _beta, 1.0 / sigma_sqr);
    }

    return p;
}

/**
 * @param  mu     Mean to compute the density of.
 * @param  sigma  Standard deviation to compute the density of.
 */
double Component::Paramgroup::Context::UV_gauss_gamma::log_pdf
(
    const jwsc::Vector_d* mu,
    const jwsc::Vector_d* sigma
)
const
{
    using namespace jwsc;

    uint32_t D = 2;

    assert(mu->num_elts == D);
    assert(sigma->num_elts == D);

    double log_p = 0.0;

    for (size_t d = 0; d < D; d++)
    {
        double sigma_sqr = sigma->elts[ d ]*sigma->elts[ d ];
        double g = 0;

        g += 0.5 * gaussian_pdf_d(0, sqrt(_lambda*sigma_sqr), mu->elts[ d ]);
        g += 0.5 * 1.0;

        double log_g = log((g < JWSC_MIN_LOG_ARG) ? JWSC_MIN_LOG_ARG : g);

        log_p += log_g + log_gamma_pdf_d(_alpha, _beta, 1.0 / sigma_sqr);
    }

    return log_p;
}

/**
 * @param  mu_out     Sampled mean.
 * @param  mu_min     Minimum value the sample can fall in.
 * @param  mu_max     Maximum value the sample can fall in.
 * @param  sigma_out  Sampled standard deviation.
 */
void Component::Paramgroup::Context::UV_gauss_gamma::sample
(
    jwsc::Vector_d**      mu_out, 
    const jwsc::Vector_d* mu_min,
    const jwsc::Vector_d* mu_max,
    jwsc::Vector_d**      sigma_out
)
const
{
    using namespace jwsc;

    uint32_t D = 2;

    double tau;
    double tau_min = 1.0e-16;
    double tau_max = 6*sqrt(_alpha)/_beta;

    assert(mu_min->num_elts == D);
    assert(mu_min->num_elts == mu_max->num_elts);

    create_vector_d(mu_out, D);
    create_vector_d(sigma_out, D);

    Vector_d* mu = *mu_out;
    Vector_d* sigma = *sigma_out;

    double a = sample_uniform_pdf_d(0, 1);
    size_t d_u = (a < 0.5) ? 0 : 1;
    size_t d_g = (a < 0.5) ? 1 : 0;

    tau = sample_gamma_pdf_d(_alpha, _beta, tau_min, tau_max);
    sigma->elts[ d_u ] = 1.0 / sqrt(tau);
    mu->elts[ d_u ] = sample_uniform_pdf_d(mu_min->elts[ d_u ], 
            mu_max->elts[ d_u ]);

    tau = sample_gamma_pdf_d(_alpha, _beta, tau_min, tau_max);
    sigma->elts[ d_g ] = 1.0 / sqrt(tau);
    assert(mu_min->elts[ d_g ] < mu_max->elts[ d_g ]);
    mu->elts[ d_g ] = mu_min->elts[ d_g ] - 1;
    while (mu->elts[ d_g ] < mu_min->elts[ d_g ] ||
           mu->elts[ d_g ] > mu_max->elts[ d_g ])
    {
        mu->elts[ d_g ] = sample_gaussian_pdf_d(0, sqrt(_lambda/tau), 
                -6.0*sigma->elts[ d_g ], 6.0*sigma->elts[ d_g ]);
    }
}

/* ========================================================================== */







/* -------------------  Component::Paramgroup::Context  --------------------- */

/**
 * The default context density functions assume independent parameter densities
 * obtained from each Model::Parameter::Context.
 */
Component::Paramgroup::Context::Context()
{
    _prior = 0;
}

/**
 * @param  ctx  Group of parameter contexts to copy into this one.
 */
Component::Paramgroup::Context::Context
(
    const Component::Paramgroup::Context& ctx
)
{
    _prior = 0;

    *this = ctx;
}

Component::Paramgroup::Context::~Context()
{
    delete _prior;
}

/**
 * @param  ctx  Group of parameter contexts to assign to this one.
 */
Component::Paramgroup::Context& Component::Paramgroup::Context::operator= 
(
    const Component::Paramgroup::Context& ctx
)
{
    delete _prior;

    _prior = 0;

    if (ctx._prior)
    {
        _prior = ctx._prior->clone();
    }

    return *this;
}

/** Checks the bounds of the parameter group and aborts on invalid indices. */
const Model::Parameter::Context& 
Component::Paramgroup::Context::operator[](size_t i) const
{
    assert(i < _contexts.size());
    return _contexts[ i ];
}

/** Checks the bounds of the parameter group and aborts on invalid indices. */
Model::Parameter::Context& Component::Paramgroup::Context::operator[](size_t i)
{
    assert(i < _contexts.size());
    return _contexts[ i ];
}

void Component::Paramgroup::Context::add_context
(
    const Model::Parameter::Context& c
)
{
    _contexts.push_back(c);
}

const Component::Paramgroup::Context::Density& 
Component::Paramgroup::Context::prior() const
{
    return *_prior;
}

void Component::Paramgroup::Context::prior
(
    const Component::Paramgroup::Context::Density& prior
)
{
    delete _prior;
    _prior = prior.clone();
}

double Component::Paramgroup::Context::calc_prior_prob
(
    const jwsc::Vector_d* mu,
    const jwsc::Vector_d* sigma
) 
const
{
    uint32_t D = _contexts.size();

    assert(mu->num_elts == D);
    assert(sigma->num_elts == D);

    if (_prior)
    {
        return _prior->pdf(mu, sigma);
    }

    double p = 1.0;

    for (uint32_t d = 0; d < D; d++)
    {
        p *= _contexts[ d ].calc_prior_prob(mu->elts[ d ], sigma->elts[ d ]);
    }

    return p;
}

double Component::Paramgroup::Context::calc_prior_log_prob
(
    const jwsc::Vector_d* mu,
    const jwsc::Vector_d* sigma
) 
const
{
    uint32_t D = _contexts.size();

    assert(mu->num_elts == D);
    assert(sigma->num_elts == D);

    if (_prior)
    {
        return _prior->log_pdf(mu, sigma);
    }

    double p = 0.0;

    for (size_t d = 0; d < D; d++)
    {
        p += _contexts[ d ].calc_prior_log_prob(mu->elts[ d ], 
                sigma->elts[ d ]);
    }

    return p;
}

double Component::Paramgroup::Context::calc_trunc_prior_log_prob
(
    const jwsc::Vector_d* mu,
    const jwsc::Vector_d* mu_min,
    const jwsc::Vector_d* mu_max,
    const jwsc::Vector_d* sigma
) 
const
{
    uint32_t D = _contexts.size();

    assert(mu->num_elts == D);
    assert(mu->num_elts == mu_min->num_elts);
    assert(mu_min->num_elts == mu_max->num_elts);
    assert(sigma->num_elts == D);

    if (_prior)
    {
        return _prior->trunc_log_pdf(mu, mu_min, mu_max, sigma);
    }

    double p = 0.0;

    for (size_t d = 0; d < D; d++)
    {
        p += _contexts[ d ].calc_trunc_prior_log_prob(mu->elts[ d ],
                mu_min->elts[ d ], mu_max->elts[ d ], sigma->elts[ d ]);
    }

    return p;
}

void Component::Paramgroup::Context::sample_prior
(
    jwsc::Vector_d** mu_out,
    jwsc::Vector_d** sigma_out
) 
const
{
    using namespace jwsc;

    uint32_t D = _contexts.size();

    if (_prior)
    {
        Vector_d* mu_min = 0;
        Vector_d* mu_max = 0;

        create_vector_d(&mu_min, D);
        create_vector_d(&mu_max, D);

        for (uint32_t d = 0; d < D; d++)
        {
            mu_min->elts[ d ] = _contexts[ d ].range().min();
            mu_max->elts[ d ] = _contexts[ d ].range().max();
        }

        _prior->sample(mu_out, mu_min, mu_max, sigma_out); 

        free_vector_d(mu_min);
        free_vector_d(mu_max);

        return;
    }

    create_vector_d(mu_out, D);
    create_vector_d(sigma_out, D);

    Vector_d* mu = *mu_out;
    Vector_d* sigma = *sigma_out;

    for (uint32_t d = 0; d < D; d++)
    {
        _contexts[ d ].sample_prior(&(mu->elts[ d ]), &(sigma->elts[ d ]));
    }
}

void Component::Paramgroup::Context::sample_trunc_prior
(
    jwsc::Vector_d**      mu_out,
    const jwsc::Vector_d* mu_min,
    const jwsc::Vector_d* mu_max,
    jwsc::Vector_d**      sigma_out
)
const
{
    using namespace jwsc;

    uint32_t D = _contexts.size();

    if (_prior)
    {
        _prior->sample(mu_out, mu_min, mu_max, sigma_out); 
        return;
    }

    create_vector_d(mu_out, D);
    create_vector_d(sigma_out, D);

    Vector_d* mu = *mu_out;
    Vector_d* sigma = *sigma_out;

    for (uint32_t d = 0; d < D; d++)
    {
        _contexts[ d ].sample_prior(&(mu->elts[ d ]), &(sigma->elts[ d ]));
    }
}

/* ========================================================================== */







/* -----------------------  Component::Paramgroup  -------------------------- */

Component::Paramgroup::Paramgroup
(
    const Component::Paramgroup::Context* ctx
)
{
    _ctx = ctx;
}

Component::Paramgroup::Paramgroup(const Component::Paramgroup& g)
{
    *this = g;
}

Component::Paramgroup::~Paramgroup()
{
}

Component::Paramgroup& Component::Paramgroup::operator= 
(
    const Component::Paramgroup& g
)
{
    _ctx    = g._ctx;
    _params = g._params;

    return *this;
}

/** Checks the bounds of the parameter group and aborts on invalid indices. */
Component::Parameter& Component::Paramgroup::operator[](size_t i)
{
    assert(i < num_params());
    return _params[ i ];
}

/** Checks the bounds of the parameter group and aborts on invalid indices. */
const Component::Parameter& Component::Paramgroup::operator[](size_t i) const
{
    assert(i < num_params());
    return _params[ i ];
}

size_t Component::Paramgroup::num_params() const
{
    return _params.size();
}

void Component::Paramgroup::add_param(const Component::Parameter& p)
{
    _params.push_back(p);
}

void Component::Paramgroup::means(jwsc::Vector_d** means_out) const
{
    using namespace jwsc;

    size_t D = num_params();

    create_vector_d(means_out, D);
    Vector_d* means = *means_out;

    for (size_t d = 0; d < D; d++)
    {
        means->elts[ d ] = _params[ d ].mean();
    }
}

void Component::Paramgroup::means(const jwsc::Vector_d* means)
{
    using namespace jwsc;

    size_t D = num_params();
    assert(D == means->num_elts);

    for (size_t d = 0; d < D; d++)
    {
        _params[ d ].mean(means->elts[ d ]);
    }
}

void Component::Paramgroup::stdevs(jwsc::Vector_d** stdevs_out) const
{
    using namespace jwsc;

    size_t D = num_params();

    create_vector_d(stdevs_out, D);
    Vector_d* stdevs = *stdevs_out;

    for (size_t d = 0; d < D; d++)
    {
        stdevs->elts[ d ] = _params[ d ].stdev();
    }
}

void Component::Paramgroup::stdevs(const jwsc::Vector_d* stdevs)
{
    using namespace jwsc;

    size_t D = num_params();
    assert(D == stdevs->num_elts);

    for (size_t d = 0; d < D; d++)
    {
        _params[ d ].stdev(stdevs->elts[ d ]);
    }
}

size_t Component::Paramgroup::num_conds() const
{
    size_t D = num_params();
    size_t n = 0;

    for (size_t d = 0; d < D; d++)
    {
        n += _params[ d ].num_conds();
    }

    return n;
}

const Component::Paramgroup::Context* Component::Paramgroup::context() const
{
    return _ctx;
}

void Component::Paramgroup::context
(
    const Component::Paramgroup::Context* ctx
)
throw (jwscxx::base::Arg_error)
{
    _ctx = ctx;
}

/* ========================================================================== */






/* -------------------------  Component::Compare  --------------------------- */

bool Component::Compare::operator() 
(
    const Component* c_1,
    const Component* c_2
) 
const 
{ 
    return c_1->world_coords_min_y() < c_2->world_coords_min_y(); 
}

/* ========================================================================== */






/* ------------------------  Component::Attachment  ------------------------- */

Component::Attachment::Attachment(Component* c, bool variable)
    : Model::Pointer(c)
{
    _variable = variable;
}

Component::Attachment::Attachment
(
    const Component::Attachment& p
)
: Model::Pointer(p)
{
    _variable = p._variable;
}

Component::Attachment::~Attachment()
{
}

Component::Attachment* Component::Attachment::clone() const
{
    return new Attachment(*this);
}

void Component::Attachment::read(std::istream& in) 
    throw (jwscxx::base::IO_error, jwscxx::base::Arg_error)
{
    using std::ostringstream;
    using std::istringstream;
    using jwscxx::base::Arg_error;

    Model::Pointer::read(in);

    static const char* field_name = "Variable";
    const char* field_value;
    uint32_t x;

    if (!(field_value = read_field_value(in, field_name)))
    {
        ostringstream ost;
        ost << "Missing attachment info: " << field_name;
        throw Arg_error(ost.str());
    }
    istringstream ist(field_value);
    ist >> x;
    if (ist.fail())
    {
        ostringstream ost;
        ost << field_name << ": Invalid attachment info";
        throw Arg_error(ost.str());
    }
    _variable = (bool)x;
}

void Component::Attachment::write(std::ostream& out) const 
    throw (jwscxx::base::IO_error)
{
    Model::Pointer::write(out);

    static const char* field_name = "Variable";
    uint32_t x = (uint32_t)_variable;
    out << field_name << ": " << x << std::endl;
}

const char* Component::Attachment::type_name() const
{
    return typeid(*this).name();
}

Component* Component::Attachment::ptr() const
{
    if (Model::Pointer::ptr())
    {
        Component* c;
        assert(c = dynamic_cast<Component*>(Model::Pointer::ptr()));
        return c;
    }

    return 0;
}

void Component::Attachment::ptr(Component* c)
{
    Model::Pointer::ptr(c);
}

void Component::Attachment::variable(bool v)
{
    _variable = v;
}

bool Component::Attachment::is_variable() const
{
    return _variable;
}

/* ========================================================================== */






/* -----------------------------  Component  -------------------------------- */

Component::Component
(
    const char* name, 
    double      x, 
    double      y, 
    double      z,
    const Component::Context* ctx
)
throw (jwscxx::base::Arg_error)
: Model(name, ctx), _posn_p(0)
{
    size_t param_id = 0;

    _posn_p.add_param(Parameter(param_id++, "X Position", x, 0, this));
    _posn_p.add_param(Parameter(param_id++, "Y Position", y, 0, this));
    _posn_p.add_param(Parameter(param_id++, "Z Position", z, 0, this));

    add_derived_param(&(_posn_p[0]));
    add_derived_param(&(_posn_p[1]));
    add_derived_param(&(_posn_p[2]));

    _obj = 0;
    _id  = 0;
}

/** @param  c  Component to copy into this one. */
Component::Component(const Component& c) 
    : Model(c), _posn_p(c._posn_p)
{
    add_derived_param(&(_posn_p[0]));
    add_derived_param(&(_posn_p[1]));
    add_derived_param(&(_posn_p[2]));

    *this = c;
}

Component::~Component()
{
}

/** 
 * @param  c  Component to copy into this one.
 *
 * @return A reference to this Component.
 */
Component& Component::operator=(const Component& c)
{
    Model::operator=(c);

    _obj = c._obj;
    _id  = c._id;

    _posn_p = c._posn_p;
    for (size_t i = 0; i < 3; i++)
    {
        _posn_p[ i ].model(this);
    }

    return *this;
}

/** 
 * The file format is ASCII.
 *
 * @param  in  Input stream to read this Component from.
 *
 * @throw  jwscxx::base::IO_error   Could not read from @em in.
 * @throw  jwscxx::base::Arg_error  Invalid arguments in file to read from.
 */
void Component::read(std::istream& in) 
    throw (jwscxx::base::IO_error, jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;
    using std::istringstream;

    Model::read(in);

    istringstream ist;
    const char* field_value;

    if (!(field_value = read_field_value(in, "ID")))
    {
        throw Arg_error("Missing ID field");
    }
    ist.str(field_value);
    ist >> _id;
    if (ist.fail())
    {
        throw Arg_error("Invalid ID");
    }
    ist.clear(std::ios_base::goodbit);
}

/**
 * The file format is ASCII.
 *
 * @param  out  Output stream to write this Component to.
 *
 * @throw  jwscxx::base::IO_error  Could not write to @em out.
 */
void Component::write(std::ostream& out) const throw (jwscxx::base::IO_error)
{
    Model::write(out);

    static const char* field_name = "ID";
    out << field_name << ": " << id() << std::endl;
}

size_t Component::id() const
{
    return _id;
}

void Component::id(size_t id)
{
    _id = id;
}

Object* Component::object() const
{
    return _obj;
}

void Component::object(Object* obj)
{
    _obj = obj;
}

const Component::Context* Component::context() const
{
    return dynamic_cast<const Component::Context*>(Model::context());
}

void Component::set_var_param_values(const jwsc::Vector_d* p)
    throw (jwscxx::base::Arg_error)
{
    std::ostringstream ost;

    transform_from_world_to_object_coords();
    try 
    { 
        set_var_param_values_in_object_coords(p);
    }
    catch (jwscxx::base::Arg_error e) 
    { ost << e.get_msg(); }
    update_object();
    transform_from_object_to_world_coords();

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

void Component::print_var_param_values(uint32_t iteration, std::ostream& out)
    const
{
    for (uint32_t i = 0; i < num_var_params(); i++)
    {
        out << iteration << ',' << name() << ',' << id()  << ","
            << get_var_param_name(i) << ','
            << get_var_param_value(i) << std::endl;
    }
}

void Component::print_var_param_stats(uint32_t iteration, std::ostream& out)
    const
{
    double mean, stdev;

    for (uint32_t i = 0; i < num_var_params(); i++)
    {
        get_var_param_stats(&mean, &stdev, i);
        out  << iteration << ',' << name() << ',' << id() << ','
             << get_var_param_name(i) << ','
             << mean << ',' << stdev << std::endl;
    }
}

/** 
 * Should be called whenever the internal representation of this
 * Component (and its dependencies) changes. The components should be in object
 * coordinates.
 */
void Component::update_object()
{
    if (object())
    {
        object()->update_graphical_rep();
        object()->update_lowest_component();
    }
    else
    {
        update_graphical_rep();
    }
}

const Component::Parameter* Component::posn_p(size_t i) const
{
    assert(i < 3);
    return &(_posn_p[ i ]);
}

/**
 * The resulting vector is in homogeneous coordinates.
 */
void Component::posn_v(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    create_vector_d(p_out, 4);
    Vector_d* p = *p_out;

    p->elts[0] = _posn_p[0].value();
    p->elts[1] = _posn_p[1].value();
    p->elts[2] = _posn_p[2].value();
    p->elts[3] = 1.0;
}

/**
 * The vector must be in homogeneous coordinates.
 *
 * @param  p  Position of this Component.
 */
void Component::posn_v(const jwsc::Vector_d* p)
{
    assert(p->num_elts == 4);
    assert(fabs(p->elts[3] - 1.0) <= 1.0e-8);

    _posn_p[0].value(p->elts[0]);
    _posn_p[1].value(p->elts[1]);
    _posn_p[2].value(p->elts[2]);
}

double Component::posn(size_t i) const
{
    return posn_p(i)->value();
}

/**
 * The resulting attachment point is in homogeneous coordinates and the same
 * coordinate frame as the object.
 *
 * @param  a_out  The attachment point.
 * @param  i      Face to get the attachment point of.
 * @param  t      Attachment type.
 */
void Component::attach_pt_v
(
    jwsc::Vector_d** a_out, 
    size_t           i, 
    Attach_type      t
) 
const
{
    switch (t)
    {
        case CTR_ATTACH:
            ctr_attach_pt_v(a_out, i);
            break;
        case VAR_ATTACH:
            var_attach_pt_v(a_out, i);
            break;
        default:
            abort();
    }
}

/**
 * The vector must be in homogeneous coordinates.
 *
 * @param  a  The attachment point.
 * @param  i  Face to set the attachment point of.
 * @param  t  Attachment type.
 */
void Component::attach_pt_v(const jwsc::Vector_d* a, size_t i, Attach_type t)
{
    switch (t)
    {
        case CTR_ATTACH:
            ctr_attach_pt_v(a, i);
            break;
        case VAR_ATTACH:
            var_attach_pt_v(a, i);
            break;
        default:
            abort();
    }
}

/** Aborts if the ith attachment is not being used. */
Component::Attach_type Component::attach_pt_type(size_t i) const
{
    assert(attachment(i));

    if (has_var_attach_pt(i))
    {
        return VAR_ATTACH;
    }

    return CTR_ATTACH;
}

Component::Attach_type Component::select_random_attach_pt_type() const
{
    if (jwsc::sample_uniform_pdf_d(0,1) < 0.5)
    {
        return Component::CTR_ATTACH;
    }
    return Component::VAR_ATTACH;
}

bool Component::conditions(const Model* m) const
{
    for (uint32_t i = 0; i < 3; i++)
    {
        const Parameter& p = _posn_p[ i ];
        for (uint32_t j = 0; j < p.num_conds(); j++)
        {
            if (m == p.cond(j)->ptr()->model())
            {
                return true;
            }
        }
    }

    return false;
}

Component* Component::conditional()
{
    for (uint32_t j = 0; j < 6; j++)
    {
        Component* cc = attachment(j);
        if (cc && conditions(cc))
        {
            return cc;
        }
    }
    return 0;
}

void Component::translate(double x, double y, double z)
{
    using namespace jwsc;

    Matrix_d* M = 0;
    Vector_d* p = 0;

    assert(create_3d_homo_translation_matrix_2_d(&M, x, y, z) == 0);

    posn_v(&p);
    multiply_matrix_and_vector_d(&p, M, p);
    posn_v(p);

    free_vector_d(p);
    free_matrix_d(M);
}

void Component::rotate(double phi, double x, double y, double z)
{
    using namespace jwsc;

    Matrix_d* M = 0;
    Vector_d* p = 0;

    assert(create_3d_homo_rotation_matrix_2_d(&M, phi, x, y, z) == 0);

    posn_v(&p);
    multiply_matrix_and_vector_d(&p, M, p);
    posn_v(p);

    free_vector_d(p);
    free_matrix_d(M);
}

/**
 * The point must be in homogeneous coodinates.
 *
 * @param  p  Point to transform from component to object coordinates.
 */
void Component::transform_pt_from_component_to_object_coords(jwsc::Vector_d* p) 
    const
{
    using namespace jwsc;

    double t_x, t_y, t_z;

    if (!object() || object()->in_object_coords())
    {
        t_x = posn(0);
        t_y = posn(1);
        t_z = posn(2);
    }
    else
    {
        Vector_d* q = 0;
        posn_v(&q);
        object()->transform_pt_from_world_to_object_coords(q);
        t_x = q->elts[0];
        t_y = q->elts[1];
        t_z = q->elts[2];
        free_vector_d(q);
    }

    Matrix_d* M = 0;

    create_3d_homo_translation_matrix_2_d(&M, t_x, t_y, t_z);
    multiply_matrix_and_vector_d(&p, M, p);

    free_matrix_d(M);
}

/**
 * The point must be in homogeneous coodinates.
 *
 * @param  p  Point to transform from object to component coordinates.
 */
void Component::transform_pt_from_object_to_component_coords(jwsc::Vector_d* p) 
    const
{
    using namespace jwsc;

    double t_x, t_y, t_z;

    if (!object() || object()->in_object_coords())
    {
        t_x = posn(0);
        t_y = posn(1);
        t_z = posn(2);
    }
    else
    {
        Vector_d* q = 0;
        posn_v(&q);
        object()->transform_pt_from_world_to_object_coords(q);
        t_x = q->elts[0];
        t_y = q->elts[1];
        t_z = q->elts[2];
        free_vector_d(q);
    }

    Matrix_d* M = 0;

    create_3d_homo_translation_matrix_2_d(&M, -t_x, -t_y, -t_z);
    multiply_matrix_and_vector_d(&p, M, p);

    free_matrix_d(M);
}

void Component::transform_from_world_to_object_coords()
    throw (jwscxx::base::Arg_error)
{
    if (object()) 
    {
        object()->transform_from_world_to_object_coords();
    }
}

void Component::transform_from_object_to_world_coords()
    throw (jwscxx::base::Arg_error)
{
    if (object()) 
    {
        object()->transform_from_object_to_world_coords();
    }
}

void Component::wire_render() const
{
    std::vector<const jwscxx::graphics::Polygon_d*> faces = this->faces();

    for (uint32_t i = 0; i < faces.size(); i++)
    {
        faces[ i ]->wire_render();
    }
}

void Component::wire_occlude_render() const
{
    using jwscxx::graphics::Polygon_d;

    std::vector<const Polygon_d*> faces = this->faces();

    for (uint32_t i = 0; i < faces.size(); i++)
    {
        faces[ i ]->wire_occlude_render();
    }
}

void Component::wire_render_mesh_faces_with_ids
(
    const View*                         view,
    const jwscxx::graphics::Polymesh_d* mesh
) 
const
{
    using namespace jwsc;
    using jwscxx::graphics::Polygon_d;

    const Polygon_d* face_i;
    const Polygon_d* face_j;

    bool face_i_vis;
    bool face_j_vis;

    uint32_t silhouette_edge;
    uint32_t c;
    uint32_t i, j, k;

    for (i = 0; i < mesh->num_faces(); i++)
    {
        face_i = mesh->get_face(i);
        face_i_vis = face_i->is_visible(view);

        glBegin(GL_LINES);

        const Vector_d* p1 = face_i->get_point(3);
        const Vector_d* p2 = face_i->get_point(0);

        assert(face_i->get_id() <= 0x1FFFFFFF);

        j = mesh->adjacent_face(i, 3);
        face_j = mesh->get_face(j);
        face_j_vis = face_j->is_visible(view);
        silhouette_edge = ( face_i_vis && !face_j_vis) ||
                          (!face_i_vis &&  face_j_vis);

        c = face_i->get_id();
        c = (c << 2) | 0x03;
        c = (c << 1) | silhouette_edge;
        glColor4ub(c>>24&0xFF, c>>16&0xFF, c>>8&0xFF, c&0xFF);
        glVertex4d(p1->elts[0], p1->elts[1], p1->elts[2], p1->elts[3]);
        glVertex4d(p2->elts[0], p2->elts[1], p2->elts[2], p2->elts[3]);

        for (k = 0; k < 3; k++)
        {
            p1 = p2;
            p2 = face_i->get_point(k+1);

            j = mesh->adjacent_face(i, k);
            face_j = mesh->get_face(j);
            face_j_vis = face_j->is_visible(view);
            silhouette_edge = ( face_i_vis && !face_j_vis) ||
                              (!face_i_vis &&  face_j_vis);

            c = face_i->get_id();
            c = (c << 2) | k;
            c = (c << 1) | silhouette_edge;
            glColor4ub(c>>24&0xFF, c>>16&0xFF, c>>8&0xFF, c&0xFF);
            glVertex4d(p1->elts[0], p1->elts[1], p1->elts[2], p1->elts[3]);
            glVertex4d(p2->elts[0], p2->elts[1], p2->elts[2], p2->elts[3]);
        }

        glEnd();
    }
}

void Component::solid_render_mesh_faces
(
    const jwscxx::graphics::Polymesh_d* mesh
) 
const
{
    using namespace jwsc;
    using jwscxx::graphics::Polygon_d;

    static const double a = 1.0 / 3.0;

    const Polygon_d* face;
    const Polygon_d* adj_face;
    const Polygon_d* prev_adj_face;
    const Vector_d*  norm;
    const Vector_d*  adj_norm;
    const Vector_d*  prev_adj_norm;
    const Vector_d*  pt;

    double n_x, n_y, n_z;

    for (size_t i = 0; i < mesh->num_faces(); i++)
    {
        face = mesh->get_face(i);
        norm = face->get_normal();

        prev_adj_face = mesh->get_face(mesh->adjacent_face(i, 3));
        prev_adj_norm = prev_adj_face->get_normal();

        glBegin(GL_POLYGON);
        for (size_t j = 0; j < face->get_num_points(); j++)
        {
            adj_face = mesh->get_face(mesh->adjacent_face(i, j));
            adj_norm = adj_face->get_normal();

            n_x = prev_adj_norm->elts[0] + norm->elts[0] + adj_norm->elts[0];
            n_y = prev_adj_norm->elts[1] + norm->elts[1] + adj_norm->elts[1];
            n_z = prev_adj_norm->elts[2] + norm->elts[2] + adj_norm->elts[2];

            glNormal3d(a*n_x, a*n_y, a*n_z);

            pt = face->get_point(j);
            glVertex4d(pt->elts[0], pt->elts[1], pt->elts[2], pt->elts[3]);

            prev_adj_face = adj_face;
            prev_adj_norm = adj_norm;
        }
        glEnd();
    }
}

void Component::solid_render_with_ids(const View* view) const
{
    using namespace jwsc;
    using jwscxx::graphics::Polygon_d;

    std::vector<const Polygon_d*> faces = this->faces();

    for (uint32_t i = 0; i < faces.size(); i++)
    {
        const Polygon_d* face = faces[ i ];

        glBegin(GL_POLYGON);

        assert(face->get_id() <= 0x1FFFFFFF);

        uint32_t c = face->get_id();

        for (uint32_t j = 0; j < 4; j++)
        {
            const Vector_d* p = face->get_point(j);

            c = (c << 2) | j;
            c = (c << 1) | 0x0; // empty field.
            glColor4ub(c>>24&0xFF, c>>16&0xFF, c>>8&0xFF, c&0xFF);
            glVertex4d(p->elts[0], p->elts[1], p->elts[2], p->elts[3]);
        }

        glEnd();
    }
}

/* ========================================================================== */
