/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Basic Object structure component.
 */


#ifndef COMPONENT_H
#define COMPONENT_H


#include <config.h>

#include <iostream>
#include <vector>

#include <inttypes.h>

#include <jwsc/vector/vector.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/base/cloneable.h>
#include <jwsc++/graphics/renderable.h>
#include <jwsc++/graphics/polygon.h>
#include <jwsc++/graphics/polymesh.h>

#include "model.h"
#include "view.h"


class Object;


/** 
 * @brief Object structure component.
 */
class Component : public jwscxx::graphics::Wire_renderable, 
                  public jwscxx::graphics::Wire_occlude_renderable, 
                  public jwscxx::graphics::Solid_renderable,
                  public Model
{
    public:


/**
 * @class Component::Context
 *
 * @brief Context for Component model parameters.
 */
class Context : public Model::Context
{
    public:

        /** @brief Constructs a new default Component::Context. */
        Context();

        /** @brief Constructs a new Component::Context from another. */
        Context(const Component::Context& ctx);

        /** @brief Deletes this Component::Context. */
        virtual ~Context();

        /** @brief Copies a Component::Context into this one. */
        virtual Component::Context& operator= (const Component::Context& ctx);

        /** @brief Clones this Context. */
        virtual Component::Context* clone() const = 0;
};


/** 
 * @class Component::Parameter
 *
 * @brief Component parameter.
 */
class Parameter : public Model::Parameter
{
    public:


/**
 * @class Component::Parameter::Dependent
 *
 * @brief Component parameter dependent.
 *
 * Constains a pointer to a dependent parameter and member functions to update
 * its value.
 */
class Dependent : public Pointer
{
    public:

        Dependent(Component::Parameter* p);

        Dependent(const Component::Parameter::Dependent& d);

        virtual ~Dependent();

        virtual Component::Parameter::Dependent* clone() const = 0;

        virtual const char* type_name() const = 0;

        Component::Parameter* ptr() const;

        void ptr(Component::Parameter* p);

        /** 
         * @brief Modifies this dependent parameter by the specified value and
         * shifts its depedents.
         */
        virtual void modify_and_shift_deps(double dx);

        /** 
         * @brief Modifies this dependent by the specified value and scales its
         * dependents.
         */
        virtual void modify_and_scale_deps(double dx);

        /** 
         * @brief Tests modifying this dependent parameter by the specified
         * value and shifts its depedents.
         */
        virtual void test_modify_and_shift_deps(double dx) const 
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Tests modifying this dependent by the specified value and
         * scales its dependents.
         */
        virtual void test_modify_and_scale_deps(double dx) const 
            throw (jwscxx::base::Arg_error);
};


/**
 * @class Component::Parameter::Zero_dependent
 *
 * @brief Component parameter dependent.
 *
 * Constains a pointer to a dependent parameter and dummy member functions that
 * do not update its value.
 */
class Zero_dependent : public Dependent
{
    public:

        Zero_dependent(Component::Parameter* p);

        Zero_dependent(const Component::Parameter::Zero_dependent& dep);

        virtual ~Zero_dependent();

        virtual Component::Parameter::Zero_dependent* clone() const;

        virtual const char* type_name() const;

        /** 
         * @brief Modifies this dependent parameter by the specified value and
         * shifts its depedents.
         */
        void modify_and_shift_deps(double dx);

        /** 
         * @brief Modifies this dependent by the specified value and scales its
         * dependents.
         */
        void modify_and_scale_deps(double dx);

        /** 
         * @brief Tests modifying this dependent parameter by the specified
         * value and shifts its depedents.
         */
        void test_modify_and_shift_deps(double dx) const 
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Tests modifying this dependent by the specified value and
         * scales its dependents.
         */
        void test_modify_and_scale_deps(double dx) const 
            throw (jwscxx::base::Arg_error);
};


/**
 * @class Component::Parameter::Pos_dependent
 *
 * @brief Component parameter dependent.
 *
 * Constains a pointer to a dependent parameter and member functions to update
 * its value in a positive relationship; an update by a positive value has the
 * effect of a positive change of equal amount on this dependent.
 */
class Pos_dependent : public Dependent
{
    public:

        Pos_dependent(Component::Parameter* p);

        Pos_dependent(const Component::Parameter::Pos_dependent& dep);

        virtual ~Pos_dependent();

        virtual Component::Parameter::Pos_dependent* clone() const;

        virtual const char* type_name() const;

        /** 
         * @brief Modifies this dependent parameter by the specified value and
         * shifts its depedents.
         */
        virtual void modify_and_shift_deps(double dx);

        /** 
         * @brief Modifies this dependent by the specified value and scales its
         * dependents.
         */
        virtual void modify_and_scale_deps(double dx);

        /** 
         * @brief Tests modifying this dependent parameter by the specified
         * value and shifts its depedents.
         */
        virtual void test_modify_and_shift_deps(double dx) const 
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Tests modifying this dependent by the specified value and
         * scales its dependents.
         */
        virtual void test_modify_and_scale_deps(double dx) const 
            throw (jwscxx::base::Arg_error);
};


/**
 * @class Component::Parameter::Pos_half_dependent
 *
 * @brief Component parameter dependent.
 *
 * Constains a pointer to a dependent parameter and member functions to update
 * its value with half the magnitude; an update by a positve value @em dx has
 * the effect of a @em dx/2 change on this dependent.
 */
class Pos_half_dependent : public Pos_dependent
{
    public:

        Pos_half_dependent(Component::Parameter* p);

        Pos_half_dependent(const Component::Parameter::Pos_half_dependent& dep);

        virtual ~Pos_half_dependent();

        virtual Component::Parameter::Pos_half_dependent* clone() const;

        virtual const char* type_name() const;

        /** 
         * @brief Modifies this dependent parameter by the specified value and
         * shifts its depedents.
         */
        void modify_and_shift_deps(double dx);

        /** 
         * @brief Modifies this dependent by the specified value and scales its
         * dependents.
         */
        void modify_and_scale_deps(double dx);

        /** 
         * @brief Tests modifying this dependent parameter by the specified
         * value and shifts its depedents.
         */
        void test_modify_and_shift_deps(double dx) const 
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Tests modifying this dependent by the specified value and
         * scales its dependents.
         */
        void test_modify_and_scale_deps (double dx) const 
            throw (jwscxx::base::Arg_error);
};


/**
 * @class Component::Parameter::Fancy_zero_dependent
 *
 * @brief Component parameter dependent.
 */
class Fancy_zero_dependent : public Zero_dependent
{
    public:

        Fancy_zero_dependent(Component::Parameter* p);

        Fancy_zero_dependent
        (
            const Component::Parameter::Fancy_zero_dependent& dep
        );

        virtual ~Fancy_zero_dependent();

        virtual Component::Parameter::Fancy_zero_dependent* clone() const;

        virtual const char* type_name() const;
};


/**
 * @class Component::Parameter::Fancy_pos_dependent
 *
 * @brief Component parameter dependent.
 */
class Fancy_pos_dependent : public Pos_dependent
{
    public:

        Fancy_pos_dependent(Component::Parameter* p);

        Fancy_pos_dependent
        (
            const Component::Parameter::Fancy_pos_dependent& dep
        );

        virtual ~Fancy_pos_dependent();

        virtual Component::Parameter::Fancy_pos_dependent* clone() const;

        virtual const char* type_name() const;
};


/**
 * @class Component::Parameter::Fancy_pos_half_dependent
 *
 * @brief Component parameter dependent.
 */
class Fancy_pos_half_dependent : public Pos_half_dependent
{
    public:

        Fancy_pos_half_dependent(Component::Parameter* p);

        Fancy_pos_half_dependent
        (
            const Component::Parameter::Fancy_pos_half_dependent& dep
        );

        virtual ~Fancy_pos_half_dependent();

        virtual Component::Parameter::Fancy_pos_half_dependent* clone() const;

        virtual const char* type_name() const;
};


/**
 * @class Component::Parameter::Neg_dependent
 *
 * @brief Component parameter dependent.
 *
 * Constains a pointer to a dependent parameter and member functions to update
 * its value in a negative relationship; an update by a positve value has the
 * effect of an equatl but negative change on this dependent.
 */
class Neg_dependent : public Dependent
{
    public:

        Neg_dependent(Component::Parameter* p);

        Neg_dependent(const Component::Parameter::Neg_dependent& dep);

        virtual ~Neg_dependent();

        virtual Component::Parameter::Neg_dependent* clone() const;

        virtual const char* type_name() const;

        /** 
         * @brief Modifies this dependent parameter by the specified value and
         * shifts its depedents.
         */
        void modify_and_shift_deps(double dx);

        /** 
         * @brief Modifies this dependent by the specified value and scales its
         * dependents.
         */
        void modify_and_scale_deps(double dx);

        /** 
         * @brief Tests modifying this dependent parameter by the specified
         * value and shifts its depedents.
         */
        void test_modify_and_shift_deps(double dx) const 
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Tests modifying this dependent by the specified value and
         * scales its dependents.
         */
        void test_modify_and_scale_deps(double dx) const 
            throw (jwscxx::base::Arg_error);
};


/**
 * @class Component::Parameter::Neg_half_dependent
 *
 * @brief Component parameter dependent.
 *
 * Constains a pointer to a dependent parameter and member functions to update
 * its value with negative half the magnitude; an update by a positve value
 * @em dx has the effect of a @em -dx/2 change on this dependent.
 */
class Neg_half_dependent : public Neg_dependent
{
    public:

        Neg_half_dependent(Component::Parameter* p);

        Neg_half_dependent(const Component::Parameter::Neg_half_dependent& dep);

        virtual ~Neg_half_dependent();

        virtual Component::Parameter::Neg_half_dependent* clone() const;

        virtual const char* type_name() const;

        /** 
         * @brief Modifies this dependent parameter by the specified value and
         * shifts its depedents.
         */
        void modify_and_shift_deps(double dx);

        /** 
         * @brief Modifies this dependent by the specified value and scales its
         * dependents.
         */
        void modify_and_scale_deps(double dx);

        /** 
         * @brief Tests modifying this dependent parameter by the specified
         * value and shifts its depedents.
         */
        void test_modify_and_shift_deps(double dx) const 
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Tests modifying this dependent by the specified value and
         * scales its dependents.
         */
        void test_modify_and_scale_deps(double dx) const 
            throw (jwscxx::base::Arg_error);
};


/**
 * @class Component::Parameter::Fancy_neg_dependent
 *
 * @brief Component parameter dependent.
 */
class Fancy_neg_dependent : public Neg_dependent
{
    public:

        Fancy_neg_dependent(Component::Parameter* p);

        Fancy_neg_dependent
        (
            const Component::Parameter::Fancy_neg_dependent& dep
        );

        virtual ~Fancy_neg_dependent();

        virtual Component::Parameter::Fancy_neg_dependent* clone() const;

        virtual const char* type_name() const;
};


/**
 * @class Component::Parameter::Fancy_neg_half_dependent
 *
 * @brief Component parameter dependent.
 */
class Fancy_neg_half_dependent : public Neg_half_dependent
{
    public:

        Fancy_neg_half_dependent(Component::Parameter* p);

        Fancy_neg_half_dependent
        (
            const Component::Parameter::Fancy_neg_half_dependent& dep
        );

        virtual ~Fancy_neg_half_dependent();

        virtual Component::Parameter::Fancy_neg_half_dependent* clone() const;

        virtual const char* type_name() const;
};


    // Component::Parameter
    public:

        /**
         * @name Standard
         *
         * Standard class members.
         * @{
         */

        /** @brief Constructs a new parameter. */
        Parameter
        (
            size_t                           id,
            const char*                      name, 
            double                           value,
            const Model::Parameter::Context* ctx,
            Model*                           model
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Constructs a new parameter from another. */
        Parameter(const Component::Parameter& p);

        /** @brief Deletes this parameter. */
        virtual ~Parameter();

        /** @brief Performs a (deep) copy to this parameter. */
        virtual Component::Parameter& operator=(const Component::Parameter& p);

        /** @brief Clones this parameter. */
        Component::Parameter* clone() const;

        /** @brief Reads the value of this parameter. */
        void read(std::istream& in) 
            throw (jwscxx::base::IO_error, jwscxx::base::Arg_error);

        /** @brief Writes the value of this parameter. */
        void write(std::ostream& out) const
            throw (jwscxx::base::IO_error);

        /*@}*/


        /**
         * @name Value
         *
         * Parameter value members.
         * @{
         */

        void value_shift_deps(double value);

        void value_scale_deps(double value);

        void test_value_shift_deps(double value)
            throw (jwscxx::base::Arg_error);

        void test_value_scale_deps(double value)
            throw (jwscxx::base::Arg_error);

        /*@}*/


        /**
         * @name Dependents
         *
         * Dependent management members.
         * @{
         */

        size_t num_shift_deps() const;

        size_t num_scale_deps() const;

        Component::Parameter::Dependent* shift_dep(size_t i) const;

        Component::Parameter::Dependent* scale_dep(size_t i) const;

        void add_shift_dep(const Component::Parameter::Dependent& d);

        void add_scale_dep(const Component::Parameter::Dependent& d);

        void remove_shift_dep(const Component::Parameter::Dependent& d);

        void remove_scale_dep(const Component::Parameter::Dependent& d);

        bool has_shift_dep(const Component::Parameter::Dependent& d) const;

        bool has_scale_dep(const Component::Parameter::Dependent& d) const;

        /*@}*/


        /**
         * @name Conditionals
         *
         * Conditional management members.
         * @{
         */

        size_t num_conds() const;

        Model::Parameter::Pointer* cond(size_t i) const;

        void add_cond(const Model::Parameter::Pointer& p);

        void remove_cond(const Component::Parameter::Pointer& p);

        bool has_cond(const Component::Parameter::Pointer& p) const;

        /*@}*/


        /**
         * @name Pointers
         *
         * Internal Parameter pointer management members.
         * @{
         */

        /** 
         * @brief Serializes pointers in this Parameter to indices in the @em
         * models collection.
         */
        void serialize(const std::vector<Model*>& models);

        /** 
         * @brief Deserializes this Parameter from indices in the @em models
         * collection to pointers.
         */
        void deserialize(const std::vector<Model*>& models);

        /** 
         * @brief Updates the Parameter dependency pointers from one set of
         * models to another.
         */
        void update_pointers
        (
            const std::vector<Model*>&       old_models,
            const std::vector<const Model*>& new_models
        );

        /*@}*/


    private:

        /**
         * @name Color
         *
         * Flag for counting Parameter access by dependents.
         * @{
         */

        uint32_t color() const;

        /** @brief Increments the color of this Parameter. */
        void inc_color();

        /** @brief Decrements the color of this Parameter. */
        void dec_color();

        /*@}*/


        /**
         * @name Dependents
         *
         * Dependent management members.
         * @{
         */

        /** @brief Reads the dependents of this parameter. */
        void read_dependents(std::istream& in) 
            throw (jwscxx::base::IO_error, jwscxx::base::Arg_error);

        /** @brief Writes the dependents of this parameter. */
        void write_dependents(std::ostream& out) const
            throw (jwscxx::base::IO_error);

        /*@}*/


        /**
         * @name Conditionals
         *
         * Conditional management members.
         * @{
         */

        /** @brief Reads the conditionals of this parameter. */
        void read_conditionals(std::istream& in) 
            throw (jwscxx::base::IO_error, jwscxx::base::Arg_error);

        /** @brief Writes the conditionals of this parameter. */
        void write_conditionals(std::ostream& out) const
            throw (jwscxx::base::IO_error);

        /*@}*/


    private:

        /** @brief Parameters depending on this Parameter during a shift. */
        std::vector<Component::Parameter::Dependent*> _shift_deps;

        /** @brief Parameters depending on this Parameter during a scale. */
        std::vector<Component::Parameter::Dependent*> _scale_deps;

        /** 
         * @brief Parameter to condition the this Parameter on. 
         *
         * Current implementation of conditionals is limited to one (optional)
         * conditional for the mean value of this Parameter.
         */
        std::vector<Model::Parameter::Pointer*> _conds;

        /** 
         * @brief Flag for counting the number of times a parameter has been
         * accessed during dependency updates.
         *
         * Initialized to zero.
         */
        uint32_t _color;
};


/**
 * @class Component::Paramgroup
 *
 * @brief Group of Component::Parameter.
 *
 * Constains a collection of parameters that are logically and statistically
 * linked together. Prodides a density function over dependent category
 * parameters. Does not include instance parameters because they are all assumed
 * independent, conditioned on the category parameters.
 */
class Paramgroup
{
    public:


/**
 * @class Component::Paramgroup::Context
 *
 * @brief Component parameter group context.
 *
 * Constains context information for groups of parameters.
 */
class Context
{
    public:


/**
 * @class Component::Paramgroup::Context::Density
 *
 * @brief Density function over component category parameters.
 */
class Density : public jwscxx::base::Cloneable
{
    public:

        /** @brief Clones this Density. */
        virtual Component::Paramgroup::Context::Density* clone() const = 0;

        /** @brief Computes the probability density function. */
        virtual double pdf
        (
            const jwsc::Vector_d* mu, 
            const jwsc::Vector_d* sigma
        ) 
        const = 0;

        /** @brief Computes the log of the probability density function. */
        virtual double log_pdf
        (
            const jwsc::Vector_d* mu, 
            const jwsc::Vector_d* sigma
        ) 
        const = 0;

        /** @brief Evaluates a truncated version of the log pdf function. */
        double trunc_log_pdf
        (
            const jwsc::Vector_d* mu, 
            const jwsc::Vector_d* mu_min, 
            const jwsc::Vector_d* mu_max,
            const jwsc::Vector_d* sigma
        ) 
        const;

        /** @brief Draws a random sample from this Density. */
        virtual void sample
        (
            jwsc::Vector_d**      mu_out, 
            const jwsc::Vector_d* mu_min, 
            const jwsc::Vector_d* mu_max,
            jwsc::Vector_d**      sigma_out
        )
        const = 0;
};


/**
 * @class Component::Paramgroup::Context::Iso_gauss_gamma
 *
 * @brief Isotropic multivariate mixture of Gaussians and a Gamma density
 * function.
 *
 * Joint density function over independent category means and variances.  The
 * Isotropic Gaussian is over mean values. The Gamma is over a precision
 * (inverse variance).
 *
 * The variance scale (lambda) and Gamma is the same for each dimension of mu
 * and cluster.
 */
class Iso_gauss_gamma : public Density
{
    public:

        /** @brief Constructs a new isotropic Gaussian density. */
        Iso_gauss_gamma
        (
            const jwsc::Vector_d* pi,
            const jwsc::Matrix_d* mu,
            double                lambda,
            double                alpha,
            double                beta
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Constructs a new isotropic Gaussian density from another. */
        Iso_gauss_gamma
        (
            const Component::Paramgroup::Context::Iso_gauss_gamma& d
        );

        /** @brief Destroys an isotropic Gaussian density. */
        ~Iso_gauss_gamma();

        /** @brief Clones this density. */
        Component::Paramgroup::Context::Iso_gauss_gamma* clone() const;

        /** @brief Assigns one density to another. */
        Component::Paramgroup::Context::Iso_gauss_gamma& operator= 
        (
            const Component::Paramgroup::Context::Iso_gauss_gamma& ctx
        );

        /** @brief Computes the probability density function. */
        double pdf(const jwsc::Vector_d* mu, const jwsc::Vector_d* sigma) const;

        /** @brief Computes the log of the probability density function. */
        double log_pdf(const jwsc::Vector_d* mu, const jwsc::Vector_d* sigma) 
            const;

        /** @brief Draws a random sample from this density. */
        void sample
        (
            jwsc::Vector_d**      mu_out,
            const jwsc::Vector_d* mu_min,
            const jwsc::Vector_d* mu_max,
            jwsc::Vector_d**      sigma_out
        )
        const;


    private:

        /** @brief Gaussian mixture component weights. */
        jwsc::Vector_d* _pi;

        /** @brief Gaussian means for each mixture component. */
        jwsc::Matrix_d* _mu;

        /** @brief Gaussian variance scale shared by each mixture component. */
        double _lambda;

        /** @brief Gamma shape parameter over the shared isotropic precision. */
        double _alpha;

        /** @brief Gamma scale parameter over the shared isotropic precision. */
        double _beta;
};


/**
 * @class Component::Paramgroup::Context::UV_gauss_gamma
 *
 * @brief Mixture of two 1D Gaussians centered on the uv-axis of a component
 * face.
 *
 * Joint density function over vectors of independent means and variances.
 *
 * The UV Gaussian is over a mean position value on a component face. The Gamma
 * is over the precision (inverse variance) of that position.
 *
 * The mixture weights are the same for each axis.
 *
 * The variance scale (lambda) and Gamma is the same for each dimension of mu
 * and cluster.
 */
class UV_gauss_gamma : public Density
{
    public:

        /** @brief Constructs a new density. */
        UV_gauss_gamma(double lambda, double alpha, double beta) 
            throw (jwscxx::base::Arg_error);

        /** @brief Constructs a new density from another. */
        UV_gauss_gamma(const Component::Paramgroup::Context::UV_gauss_gamma& d);

        /** @brief Destroys the density. */
        ~UV_gauss_gamma();

        /** @brief Clones this density. */
        Component::Paramgroup::Context::UV_gauss_gamma* clone() const;

        /** @brief Assigns one density to another. */
        Component::Paramgroup::Context::UV_gauss_gamma& operator= 
        (
            const Component::Paramgroup::Context::UV_gauss_gamma& ctx
        );

        /** @brief Computes the probability density function. */
        double pdf(const jwsc::Vector_d* mu, const jwsc::Vector_d* sigma) const;

        /** @brief Computes the log of the probability density function. */
        double log_pdf(const jwsc::Vector_d* mu, const jwsc::Vector_d* sigma) 
            const;

        /** @brief Draws a random sample from this density. */
        void sample
        (
            jwsc::Vector_d**      mu_out, 
            const jwsc::Vector_d* mu_min,
            const jwsc::Vector_d* mu_max,
            jwsc::Vector_d**      sigma_out
        )
        const;


    private:

        /** @brief Variance scaling parameter for each 1D Gaussian. */
        double _lambda;

        /** @brief Gamma shape parameter for each 1D Gaussian. */
        double _alpha;

        /** @brief Gamma scale parameter for each 1D Gaussian. */
        double _beta;
};


    // Component::Paramgroup::Context
    public:

        /**
         * @name Standard
         *
         * Standard class members.
         * @{
         */

        /** 
         * @brief Constructs a new component parameter context that assumes
         * independent parameter densities.
         */
        Context();

        /** @brief Constructs a new component parameter context from another. */
        Context(const Component::Paramgroup::Context& ctx);

        /** @brief Deletes this context. */
        ~Context();

        /** 
         * @brief Assigns a group of component parameter context to this one.
         */
        Component::Paramgroup::Context& operator= 
        (
            const Component::Paramgroup::Context& ctx
        );

        /*@}*/


        /**
         * @name Parameter contexts
         *
         * Individual parameter context members.
         * @{
         */

        /** @brief Returns the context of a parameter in this Group. */
        const Model::Parameter::Context& operator[](size_t i) const;

        /** @brief Returns the context of a parameter in this Group. */
        Model::Parameter::Context& operator[](size_t i);

        /** @brief Adds a parameter to this Group. */
        void add_context(const Model::Parameter::Context& c);

        /*@}*/


        /**
         * @name Prior
         *
         * Category parameter group joint prior probability members.
         * @{
         */

        /** @brief Returns the prior density function. */
        const Component::Paramgroup::Context::Density& prior() const;

        /** @brief Sets the prior density function. */
        void prior(const Component::Paramgroup::Context::Density& prior);

        double calc_prior_prob
        (
            const jwsc::Vector_d* mu, 
            const jwsc::Vector_d* sigma
        ) 
        const;

        double calc_prior_log_prob
        (
            const jwsc::Vector_d* mu, 
            const jwsc::Vector_d* sigma
        ) 
        const;

        double calc_trunc_prior_log_prob
        (
            const jwsc::Vector_d* mu, 
            const jwsc::Vector_d* mu_min, 
            const jwsc::Vector_d* mu_max,
            const jwsc::Vector_d* sigma
        ) 
        const;

        void sample_prior(jwsc::Vector_d** mu_out, jwsc::Vector_d** sigma_out) 
            const;

        void sample_trunc_prior
        (
            jwsc::Vector_d**      mu_out,
            const jwsc::Vector_d* mu_min,
            const jwsc::Vector_d* mu_max,
            jwsc::Vector_d**      sigma_out
        ) 
        const;

        /*@}*/


    protected: 

        /** @brief Prior probability density over the category parameters. */
        Density* _prior;

        /** @brief Individual parameter contexts for the group. */
        std::vector<Model::Parameter::Context> _contexts;
};


    // Component::Paramgroup
    public:

        /**
         * @name Standard
         *
         * Standard class members.
         * @{
         */

        Paramgroup(const Component::Paramgroup::Context* ctx);

        Paramgroup(const Component::Paramgroup& g);

        ~Paramgroup();

        Component::Paramgroup& operator= 
        (
            const Component::Paramgroup& g
        );

        /*@}*/


        /**
         * @name Parameters
         *
         * Individual parameter access members.
         * @{
         */

        /** @brief Returns a parameter in this group. */
        Component::Parameter& operator[](size_t i);

        /** @brief Returns a parameter in this group. */
        const Component::Parameter& operator[](size_t i) const;

        /** @brief Returns the number of parameters in this group. */
        size_t num_params() const;

        /** @brief Adds a parameter to this group. */
        void add_param(const Component::Parameter& p);

        /*@}*/


        /**
         * @name Category Parameters
         *
         * Category parameter members.
         * @{
         */

        /** @brief Returns the mean category parameter. */
        void means(jwsc::Vector_d** means_out) const;

        /** @brief Sets the mean category parameter. */
        void means(const jwsc::Vector_d* means);

        /** @brief Returns the standard deviation category parameter. */
        void stdevs(jwsc::Vector_d** stdevs_out) const;

        /** @brief Sets the standard deviation category parameter. */
        void stdevs(const jwsc::Vector_d* stdevs);

        /** 
         * @brief Returns the number of conditionals of the parameters in this
         * group.
         */
        size_t num_conds() const;

        /*@}*/


        /**
         * @name Context
         *
         * Group parameter context members.
         * @{
         */

        const Component::Paramgroup::Context* context() const;

        void context(const Component::Paramgroup::Context* ctx)
            throw (jwscxx::base::Arg_error);

        /*@}*/


    protected:

        /** @brief Context of this Group of parameters. */
        const Component::Paramgroup::Context* _ctx;

        /** @brief The Parameter group. */
        std::vector<Component::Parameter> _params;
};


class Compare
{
    public:

        bool operator() (const Component* c_1, const Component* c_2) const;
};


/**
 * @class Component::Attachment
 *
 * @brief Component attachment pointer and other information.
 *
 * Constains a pointer to an attaching Component and member functions to
 * serialize and deserialize its value.
 */
class Attachment : public Model::Pointer
{
    public:

        Attachment(Component* c, bool variable);

        Attachment(const Component::Attachment& p);

        virtual ~Attachment();

        Component::Attachment* clone() const;

        void read(std::istream& in) 
            throw (jwscxx::base::IO_error, jwscxx::base::Arg_error);

        void write(std::ostream& out) const throw (jwscxx::base::IO_error);

        const char* type_name() const;

        Component* ptr() const;

        void ptr(Component* c);

        bool is_variable() const;

        void variable(bool v);


    protected:

        /** 
         * @brief Indicates whether the Component containing this Attachment 
         * has a varying position in an attachment.
         *
         * The non-varying attachment points are located at the center of their
         * faces.
         */
        bool _variable;
};


    // Component
    public:

        /**
         * @name Standard
         *
         * Standard class methods.
         * @{
         */

        /** @brief Constructs a Component positioned at a point. */
        Component
        (
            const char* name, 
            double      x, 
            double      y, 
            double      z,
            const Component::Context* ctx
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Constructs a Component by copying another. */
        Component(const Component& c);

        /** @brief Deletes this Component. */
        virtual ~Component();

        /** @brief Copies a Component into this one. */
        virtual Component& operator= (const Component& c);

        /** @brief Creates a copy of this Component. */
        virtual Component* clone() const = 0;

        /** @brief Creates a special copy of this Component. */
        virtual Component* clone_and_swap() const 
            throw (jwscxx::base::Arg_error) = 0;

        /** @brief Reads this Component from an input stream. */
        virtual void read(std::istream& in) throw (jwscxx::base::IO_error,
                jwscxx::base::Arg_error);

        /** @brief Writes this Component to an output stream. */
        virtual void write(std::ostream& out) const 
            throw (jwscxx::base::IO_error);

        /** 
         * @brief Returns the id of this Component.
         *
         * Unique within an object.
         */
        size_t id() const;

        /** @brief Returns the Object back pointer for this Component. */
        Object* object() const;

        /** @brief Sets the Object back pointer for this Component. */
        void object(Object* obj);

        /** @brief Returns the model context. */
        virtual const Component::Context* context() const;

        /*@}*/


        /**
         * @name Variable Parameters
         *
         * Variable model parameter members.
         * @{
         */

        /** @brief Gets the instance parameter values. */
        virtual void get_var_param_values(jwsc::Vector_d** p_out) const = 0;

        /** @brief Sets the instance parameter values. */
        void set_var_param_values(const jwsc::Vector_d* p)
            throw (jwscxx::base::Arg_error);

        /** @brief Adds to the instance parameter values. */
        virtual void add_var_param_values_shift_deps(const jwsc::Vector_d* p)
            throw (jwscxx::base::Arg_error) = 0;

        /** @brief Adds to the instance parameter values. */
        virtual void add_var_param_values_scale_deps(const jwsc::Vector_d* p)
            throw (jwscxx::base::Arg_error) = 0;

        /** @brief Prints variable parameter values. */
        virtual void print_var_param_values
        (
            uint32_t      iteration, 
            std::ostream& out
        )
        const;

        /** @brief Prints variable parameter statistics. */
        virtual void print_var_param_stats
        (
            uint32_t      iteration, 
            std::ostream& out
        )
        const;

        /*@}*/


        /**
         * @name posn
         *
         * Position parameters methods.
         * @{
         */

        /** @brief Returns a coord of the position of this Component. */
        const Component::Parameter* posn_p(size_t i) const;

        /** @brief Returns the position of this Component. */
        void posn_v(jwsc::Vector_d** p_out) const;

        /** @brief Returns coord of the position of this Component. */
        double posn(size_t i) const;

        /** 
         * @brief Asserts consitency among attachments and the graphical
         * representation.
         */
        virtual void assert_consistency() = 0;

        /** 
         * @brief Ensures consitency among attachments and the graphical
         * representation.
         */
        virtual void ensure_consistency() = 0;

        /** @brief Minimum y coordinate of this Component. */
        virtual double world_coords_min_y() const = 0;

        /*@}*/


        /** @brief Type of attachment to make on a face of this Component. */
        typedef enum
        {
            CTR_ATTACH=0,/*!< Attach to the center of this components face. */
            VAR_ATTACH   /*!< Attach to a varying point of this components. */
        }
        Attach_type;


        /**
         * @name attach_pt
         *
         * Face attachment points parameter members.
         * @{
         */

        /** @brief Gets a face attachment point. */
        void attach_pt_v(jwsc::Vector_d** a_out, size_t i, Attach_type t) const;

        /** @brief Gets the type of attachment at a face. */
        Component::Attach_type attach_pt_type(size_t i) const;

        /** @brief Selects a random type of attachment point. */
        Component::Attach_type select_random_attach_pt_type() const;

        /** @brief Gets a face center attachment point. */
        virtual void ctr_attach_pt_v(jwsc::Vector_d** a_out, size_t i) const =0;

        /** @brief Gets a face variable attachment point. */
        virtual void var_attach_pt_v(jwsc::Vector_d** a_out, size_t i) const =0;

        /** @brief Returns whether a variable attachment is enabled on face. */
        virtual bool has_var_attach_pt(size_t i) const = 0;

        /** 
         * @brief Enables a variable attachment point by placing it in the
         * appropriate parameter sets.
         */
        virtual void enable_var_attach_pt(size_t j) = 0;

        /** 
         * @brief Disables a variable attachment point by removing it in the
         * appropriate parameter sets.
         */
        virtual void disable_var_attach_pt(size_t j) = 0;

        /*@}*/


        /**
         * @name Attachment
         *
         * Component attachment members.
         * @{
         */

        /** @brief Returns the number of Components attached to this is one. */
        virtual size_t num_attachments() const = 0;

        /** @brief Returns the index of an attachment. */
        virtual size_t attachment(const Component* c) const 
            throw (jwscxx::base::Arg_error) = 0;

        /** @brief Returns the attachment at the specifed index. */
        virtual Component* attachment(size_t i) const = 0;

        /** @brief Sets the attachment at the specified index. */
        virtual void attachment(size_t i, Component* c) = 0;

        /** 
         * @brief Returns the number of center attachment points on this
         * Component.
         */
        virtual size_t num_ctr_attach_pts() const = 0;

        virtual uint32_t select_random_used_attach_pt() const
            throw (jwscxx::base::Arg_error) = 0;

        virtual uint32_t swapped_face_index(uint32_t i) const = 0;

        /*@}*/


        /**
         * @name Conditionals
         *
         * Conditional management members.
         * @{
         */

        /** @brief Returns whether this Component conditions on another. */
        virtual bool conditions(const Model* m) const;

        /** @brief Returns the component this one conditions on. */
        Component* conditional();

        /*@}*/


        /**
         * @name Transform
         *
         * Transformation of internal representation methods.
         * @{
         */

        /** @brief Translates this Component. */
        virtual void translate(double x, double y, double z);

        /** @brief Rotates this Component about a vector. */
        virtual void rotate(double phi, double x, double y, double z);

        /** 
         * @brief Transforms a point from this Component to object coordinates.
         */
        void transform_pt_from_component_to_object_coords(jwsc::Vector_d* p)
            const;

        /** 
         * @brief Transforms a point from object to this Component coordinates.
         */
        void transform_pt_from_object_to_component_coords(jwsc::Vector_d* p) 
            const;

        void transform_from_world_to_object_coords()
            throw (jwscxx::base::Arg_error);

        void transform_from_object_to_world_coords()
            throw (jwscxx::base::Arg_error);

        /*@}*/


        /**
         * @name Likelihood
         *
         * Instance parameter likelihood members.
         * @{
         */

        virtual void gen_var_param_conditional_likelihood_sample() = 0;

        virtual void get_var_param_conditional_likelihood_sample
        (
            jwsc::Vector_d** p_out
        ) = 0;

        /*@}*/


        /**
         * @name Pointers
         *
         * Pointer management members.
         * @{
         */

        /** 
         * @brief Serializes Model and Parameter pointers in this Component
         * using indices in the @em models collection.
         */
        virtual void serialize(const std::vector<Model*>& models) = 0;

        /** 
         * @brief Deserializes Model and Parameter indices in this Component
         * using the @em models collection of pointers.
         */
        virtual void deserialize
        (
            const std::vector<Model*>& models
        ) = 0;

        /** 
         * @brief Updates the Model and Parameter pointers in this Component
         * from one set of models to another.
         */
        virtual void update_pointers
        (
            const std::vector<Model*>&       new_models,
            const std::vector<const Model*>& old_models
        ) = 0;

        /*@}*/


        /**
         * @name Graphical Representation
         *
         * Internal graphical representation methods.
         * @{
         */

        /** 
         * @brief Returns the set of polygon faces representing the 3D structure
         * of this Component.
         */
        virtual std::vector<const jwscxx::graphics::Polygon_d*> faces() 
            const = 0;

        /** @brief Returns the number of faces this Component has. */
        virtual size_t num_faces() const = 0;

        /** @brief Sets the sequential ids of a faces in this Component. */
        virtual uint32_t set_sequential_face_ids(uint32_t id) = 0;

        /** 
         * @brief Returns whether the graphical representation is outdated and
         * needs updating.
         */
        virtual bool graphical_rep_is_outdated() const = 0;

        /** 
         * @brief Sets the graphical representation as outdated and needs
         * updating.
         */
        virtual void graphical_rep_outdated() = 0;

        /** 
         * @brief Updates the graphical representation of this Block with its
         * current parameter values.
         */
        virtual void update_graphical_rep() = 0;

        /*@}*/


        /**
         * @name Graphics
         *
         * Graphics rendering methods.
         * @{
         */

        /** @brief Renders this Component with the GL as a wire-frame. */
        void wire_render() const;

        /** 
         * @brief Renders this Component with the GL as a wire-frame using
         * polygon face ids as color.
         */
        virtual void wire_render_with_ids(const View* view) const = 0;

        /** 
         * @brief Renders this Component with the GL as an occluded wire-frame 
         * into the depth buffer, to hide unseen lines.
         */
        void wire_occlude_render() const;

        /** @brief Renders this Component with the GL as a solid object. */
        virtual void solid_render() const = 0;

        /** 
         * @brief Renders this Component with the GL as a wire-frame using
         * polygon face ids as color.
         */
        void solid_render_with_ids(const View* view) const;

        /*@}*/


    protected:

        /** @brief Updates the Object containing this Component. */
         void update_object();

        /**
         * @name attach_pt
         *
         * Face attachment points parameter members.
         * @{
         */

        /** @brief Sets a face center attachment point. */
        void attach_pt_v(const jwsc::Vector_d* a, size_t i, Attach_type t);

        /** @brief Sets a face center attachment point. */
        virtual void ctr_attach_pt_v(const jwsc::Vector_d* a, size_t i) = 0;

        /** @brief Sets a face variable attachment point. */
        virtual void var_attach_pt_v(const jwsc::Vector_d* a, size_t i) = 0;

        /*@}*/


        /**
         * @name Parameters
         *
         * Generic model parameter methods.
         * @{
         */

        /** @brief Sets the Parameter values from a point. */
        virtual void set_var_param_values_in_object_coords
        (
            const jwsc::Vector_d* p
        )
        throw (jwscxx::base::Arg_error) = 0;

        /*@}*/


        /**
         * @name Graphics
         *
         * Graphics rendering methods.
         * @{
         */

        /** 
         * @brief Renders this Component with the GL as a wire-frame using
         * polygon face ids as color.
         */
        void wire_render_mesh_faces_with_ids
        (
            const View*                         view,
            const jwscxx::graphics::Polymesh_d* mesh
        ) 
        const;

        /** @brief Renders this Component with the GL as a solid object. */
        void solid_render_mesh_faces
        (
            const jwscxx::graphics::Polymesh_d* mesh
        ) 
        const;

        /*@}*/


    private:

        /** 
         * @brief Sets the id of this Component.
         *
         * Must be within an object.
         */
        void id(size_t id);


        /**
         * @name posn
         *
         * Position parameters methods.
         * @{
         */

        /** @brief Sets the position of this Component. */
        void posn_v(const jwsc::Vector_d* p);

        /*@}*/


    protected:

        /** 
         * @brief Position of this Component.
         *
         * Contains three parameters indexed by x=0, y=1, z=2.
         */
        Component::Paramgroup _posn_p;

        /** 
         * @brief The Object containing this Component.
         *
         * Initially set to 0, until added to an Object.
         */
        Object* _obj;


    private:

        /** @brief Unique id of this Component within an object. */
        size_t _id;


    friend class Object;
    friend class Block_set;
};


#endif
