/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief MCMC molecular dynamics based sampler.
 */


#include <config.h>

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <list>
#include <cassert>

#include <inttypes.h>

#include <jwsc/math/constants.h>
#include <jwsc/vector/vector.h>
#include <jwsc/vector/vector_math.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/matrix/matrix_math.h>
#include <jwsc/prob/pdf.h>
#include <jwsc/stat/stat.h>
#include <jwsc/stat/dynamics.h>
#include <jwsc/image/image.h>
#include <jwsc/image/edge.h>

#include <jwsc++/base/exception.h>

#include "sampler.h"
#include "view.h"
#include "object.h"
#include "block.h"
#include "block_set.h"
#include "likelihood.h"
#include "offscreen.h"
#include "dynamics.h"


/* -------------------------  Dynamics  ------------------------------------- */

/**
 * @param  params      Parameters for the sampler.
 * @param  iterations  Number of iterations to run the sampler.
 *
 * @throw  jwscxx::base::Exception  An error occurred in the sampler.
 */
void Dynamics::run(Sampler::Parameters* params, uint32_t iterations)
    throw (jwscxx::base::Exception)
{
    Dynamics::Parameters* dynamics_params;
    assert((dynamics_params = dynamic_cast<Dynamics::Parameters*>(params)));

    run(dynamics_params, iterations);
}


/**
 * @param  params      Parameters for the sampler.
 * @param  iterations  Number of iterations to run the sampler.
 *
 * @throw  jwscxx::base::Exception  An error occurred in the sampler.
 */
void Dynamics::run(Dynamics::Parameters* params, uint32_t iterations)
    throw (jwscxx::base::Exception)
{
    typedef void* (*f)(void*);
    uint32_t N_t = params->num_threads();
    uint32_t N_i = params->num_instances();

    if (N_i < N_t)
    {
        N_t = N_i;
    }

    std::vector<Thread_manager> mgr(N_t);
    std::vector<pthread_t>      thr(N_t);
    std::vector<jwscxx::base::Exception*> ex(N_t, 0);

    for (uint32_t t = 0; t < N_t; t++)
    {
        mgr[ t ].thread(t);
        mgr[ t ].num_threads(N_t);
        mgr[ t ].sampler(this);
        mgr[ t ].params(params);
        mgr[ t ].iterations(iterations);
    }

    for (uint32_t t = 0; t < N_t; t++)
    {
        assert(pthread_create(&(thr[ t ]), 0, (f)run_thread_c, &(mgr[t])) == 0);
    }

    for (uint32_t t = 0; t < N_t; t++)
    {
        assert(pthread_join(thr[ t ], (void**)&(ex[ t ])) == 0);
        if (ex[ t ])
        {
            throw jwscxx::base::Exception(*(ex[ t ]));
        }
    }
}

jwscxx::base::Exception* Dynamics::run_thread_c
(
    Dynamics::Thread_manager* mgr
)
{
    try
    {
        assert(Offscreen::enabled());
        Offscreen::activate(mgr->thread());

        mgr->sampler()->run_thread(mgr);

        assert(Offscreen::enabled());
        Offscreen::deactivate(mgr->thread());
    }
    catch (jwscxx::base::Exception e)
    {
        return new jwscxx::base::Exception(e);
    }

    return 0;
}

/* ========================================================================== */







/* -----------------------  Dynamics::Instance  ----------------------------- */

/**
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 * @throw  jwscxx::base::IO_error   Could not open info file.
 */
Dynamics::Instance::Instance
(
    View*                       best_view,
    Object*                     best_obj,
    double                      best_img_ll,
    double                      best_potential,
    View*                       view,
    Object*                     obj,
    double                      img_ll,
    double                      potential,
    const jwsc::Matrix_f*       image_t,
    const jwsc::Edge_set_f*     edges,
    Likelihood::Correspondence* img_ll_corr,
    Likelihood::Cache*          img_ll_cache,
    jwsc::Matrix_u32*           img_ll_surf_assign,
    jwsc::Matrix_u32*           img_ll_edge_assign,
    jwsc::Matrix_u32*           model_mat,
    jwsc::Vector_d*             sample_mean_in,
    jwsc::Matrix_d*             sample_cov_in,
    const char*                 sample_info_fname,
    bool                        compress_sample_info,
    const char*                 sample_stats_info_fname
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
: Sampler::Instance(best_view, best_obj, best_img_ll, best_potential,
        view, obj, img_ll, potential, image_t, edges, img_ll_corr, img_ll_cache,
        img_ll_surf_assign, img_ll_edge_assign, model_mat, sample_mean_in,
        sample_cov_in, sample_info_fname, compress_sample_info,
        sample_stats_info_fname)
{
}


Dynamics::Instance::~Instance()
{
}

/* ========================================================================== */







/* ------------------------  Dynamics::Category  ---------------------------- */

/**
 * @param  rep_ins            Representative instance for the category.
 * @param  lp_potential       Log prior potential energy.
 * @param  potential          Potential energy of the likelihood and prior.
 * @param  sample_info_fname  File to write sample information to.
 *
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 * @throw  jwscxx::base::IO_error   Could not open info file.
 */
Dynamics::Category::Category
(
    const Instance* rep_ins,
    double          lp_potential,
    double          potential,
    const char*     sample_info_fname,
    bool            compress_sample_info
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
: Sampler::Category(rep_ins, lp_potential, potential, sample_info_fname,
        compress_sample_info)
{
}

Dynamics::Category::~Category()
{
}

/* -------------------------------------------------------------------------- */








/* -----------------------  Dynamics::Parameters  --------------------------- */

Dynamics::Parameters::Parameters
(
    const View::Context*      view_ctx,
    const Object::Context*    obj_ctx,
    const Block::Context*     block_ctx,
    const Block_set::Context* block_set_ctx,
    double                    img_ll_bg_prob,
    double                    img_ll_fg_prob,
    double                    img_ll_fg_noise_prob,
    double                    img_ll_fg_miss_prob,
    double                    img_ll_edge_noise_prob,
    double                    img_ll_edge_silho_miss_prob,
    double                    img_ll_edge_inner_miss_prob,
    uint32_t                  sample_stats_freq,
    uint32_t                  num_instances,
    uint32_t                  num_categories,
    uint32_t                  min_num_members,
    uint32_t                  num_threads
)
: Sampler::Parameters(view_ctx, obj_ctx, block_ctx, block_set_ctx,
        img_ll_bg_prob, img_ll_fg_prob, img_ll_fg_noise_prob,
        img_ll_fg_miss_prob, img_ll_edge_noise_prob,
        img_ll_edge_silho_miss_prob, img_ll_edge_inner_miss_prob,
        sample_stats_freq, num_instances, num_categories, min_num_members,
        num_threads)
{
}


Dynamics::Parameters::~Parameters()
{
}


/**
 * @param  q  Model parameter values.
 * @param  p  Model parameter momentum values.
 * @param  i  Index of the model instance to use.
 * @param  c  Index of the category.
 */
void Dynamics::Parameters::accept_var_param_values_dynamics
(
    const jwsc::Vector_d* q, 
    const jwsc::Vector_d* p,
    uint32_t              c,
    uint32_t              i
)
{
    using std::setw;
    using namespace jwsc;

    Vector_d* curr_q = 0;
    Vector_d* deltas = 0;

    get_var_param_values(&curr_q, c, i);
    subtract_vectors_d(&deltas, q, curr_q);

    add_var_param_values_scale_component_deps(deltas, c, i);

    update_ins_ll_kinetic(p, c, i);

    free_vector_d(curr_q);
    free_vector_d(deltas);

    Sampler::Instance* ins = instance(c, i);
    add_sample(ins);

    // DEBUG
    ins->obj->assert_component_consistency();
    ins->obj->ensure_component_consistency();

    /* This is here because we want to make sure an actual potential energy
     * value is computed instead of a cached value. but really it's pointless
     * to have a cache then -- dumb idea
    try 
    {
        Likelihood::calc_neg_log_likelihood(&(ins->potential),
                &(ins->img_ll), &(ins->model_mat), &(ins->img_ll_surf_assign),
                &(ins->img_ll_edge_assign), ins->image_t, ins->edges,
                ins->img_ll_corr, ins->view, ins->obj, img_ll_bg_prob(),
                img_ll_noise_prob(), img_ll_phantom_prob());
        ins->img_ll_cache->insert(ins->potential, ins->view, ins->obj);
    }
    catch (jwscxx::base::Arg_error e)
    {
        // Should never happen.
        abort();
    }
    */

    if (ins->ll_potential < ins->best_ll_potential)
    {
        *(ins->best_view)   = *(ins->view);
        *(ins->best_obj)    = *(ins->obj);
        ins->best_img_ll    = ins->img_ll;
        ins->best_ll_potential = ins->ll_potential;
    }

    ins->print_var_param_info();
}

/**
 * @param  q  Model parameter values.
 * @param  p  Model parameter momentum values.
 * @param  c  Index of the category.
 */
void Dynamics::Parameters::accept_var_param_stats_dynamics
(
    const jwsc::Vector_d* q, 
    const jwsc::Vector_d* p,
    uint32_t              c
)
{
    set_var_param_stats(q, c);
    update_cat_lp_kinetic(p, c);
    category(c)->print_var_param_info();
}

/**
 * @param  q  Model parameter values.
 * @param  p  Model parameter momentum values.
 */
void Dynamics::Parameters::accept_var_param_stats_dynamics
(
    const jwsc::Vector_d* q, 
    const jwsc::Vector_d* p
)
{
    using namespace jwsc;

    uint32_t  D  = 2*num_cat_var_params(0);
    uint32_t  d  = 0;
    Vector_d* qq = 0;
    Vector_d* pp = 0;

    for (uint32_t c = 0; c < num_categories(); c++)
    {
        uint32_t D_c = 2*num_var_params(c, 0);

        assert(copy_vector_section_d(&qq, q, d, D_c) == 0);
        assert(copy_vector_section_d(&pp, p, d, D_c) == 0);
        d += D_c;

        accept_var_param_stats_dynamics(qq, pp, c);
    }

    free_vector_d(qq);
    free_vector_d(pp);

    assert(d == D);
}

/* ========================================================================== */






/* ----------------------  Dynamics::Thread_manager  ------------------------ */

Dynamics::Thread_manager::Thread_manager(uint32_t thread)
    : Sampler::Thread_manager(thread)
{
}

Dynamics::Thread_manager::~Thread_manager() 
{ 
}

void Dynamics::Thread_manager::sampler(Sampler* sampler)
    throw (jwscxx::base::Arg_error)
{
    if (!(_sampler = dynamic_cast<Dynamics*>(sampler)))
    {
        throw jwscxx::base::Arg_error(
                "Dynamics thread manager requires a dynamics sampler");
    }
}

Dynamics* Dynamics::Thread_manager::sampler() const
{
    return dynamic_cast<Dynamics*>(_sampler);
}

void Dynamics::Thread_manager::params
(
    Sampler::Parameters* params
)
throw (jwscxx::base::Arg_error)
{
    if (!(_params = dynamic_cast<Dynamics::Parameters*>(params)))
    {
        throw jwscxx::base::Arg_error(
                "Dynamics thread manager requires dynamics sampler params");
    }
}

Dynamics::Parameters* Dynamics::Thread_manager::params() const
{
    return dynamic_cast<Dynamics::Parameters*>(_params);
}

/**
 * @param  p_out   Point to get the parameter stats and values into.
 * @param  mgr     Dynamics::Thread_manager to use.
 *
 * @return On success, 0 is returned. On error, a jwsc::Error is returned.
 */
void Dynamics::Thread_manager::get_var_param_stats_and_values_c
(
    jwsc::Vector_d**          p_out,
    Dynamics::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->toggle_params_lock();

    uint32_t t    = mgr->thread();
    uint32_t N_t  = mgr->num_threads();
    uint32_t N_c  = mgr->params()->num_categories();
    uint32_t N_i  = mgr->params()->num_instances();
    uint32_t D_tv = mgr->params()->num_cat_var_params(t);
    uint32_t D_ts = 2*D_tv;
    uint32_t d    = 0;
    uint32_t D    = (N_i % N_t > t) ? ((N_i / N_t) + 1)*D_tv : (N_i / N_t)*D_tv;

    if (t == 0)
    {
        D += D_ts;
    }

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    Vector_d* p_i = 0;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    if (t == 0)
    {
        mgr->params()->get_var_param_stats(&p_i);
        copy_vector_section_into_vector_d(p, d, p_i, 0, D_ts);
        d += D_ts;
    }
    mgr->desynchronize();

    for (uint32_t c = 0; c < N_c; c++)
    {
        for (uint32_t i = t; i < N_i+t; i += N_t)
        {
            if (i < N_i)
            {
                uint32_t D_i = mgr->params()->num_var_params(c, i);
                mgr->params()->get_var_param_values(&p_i, c, i);
                copy_vector_section_into_vector_d(p, d, p_i, 0, D_i);
                d += D_i;
            }
            mgr->toggle_params_lock();
        }
    }

    assert(get_num_unhandled_errors() == 0);
    assert(d == D);

    free_vector_d(p_i);
}

/**
 * @param  p    Point to set to the parameter statistics and values with.
 * @param  mgr  Dynamics::Thread_manager to use.
 *
 * @return On success, 0 is returned. On error, a jwsc::Error is returned.
 */
jwsc::Error* Dynamics::Thread_manager::set_var_param_stats_and_values_c
(
    const jwsc::Vector_d*     p,
    Dynamics::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->toggle_params_lock();

    uint32_t t    = mgr->thread();
    uint32_t N_t  = mgr->num_threads();
    uint32_t N_c  = mgr->params()->num_categories();
    uint32_t N_i  = mgr->params()->num_instances();
    uint32_t D_tv = mgr->params()->num_cat_var_params(t);
    uint32_t D_ts = 2*D_tv;
    uint32_t d    = 0;
    uint32_t D    = (N_i % N_t > t) ? ((N_i / N_t) + 1)*D_tv : (N_i / N_t)*D_tv;

    Vector_d* p_i = 0;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    if (t == 0)
    {
        copy_vector_section_d(&p_i, p, d, D_ts);
        mgr->params()->set_var_param_stats(p_i);

        d += D_tv;
        D += D_ts;
    }
    mgr->desynchronize();

    for (uint32_t c = 0; c < N_c; c++)
    {
        for (uint32_t i = t; i < N_i; i += N_t)
        {
            uint32_t D_i = mgr->params()->num_var_params(c, i);

            copy_vector_section_d(&p_i, p, d, D_i);
            mgr->params()->set_var_param_values(p_i, c, i);
            d += D_i;
        }
    }

    assert(d == D);
    free_vector_d(p_i);

    return 0;
}

/**
 * @param  E_out   Negative log likelihood and prior of the parameters in @em
 *                 params.
 * @param  mgr     Dynamics::Thread_manager to use.
 *
 * We assume the model instances are independent, both in the likelihood and
 * priors. So the energies across model instances is summed -- the potential
 * energy function is a negative log.
 *
 * @return On success, 0 is returned. On error, a jwsc::Error is returned.
 */
jwsc::Error* 
Dynamics::Thread_manager::get_var_param_stats_and_values_potential_c
(
    double*                   E_out, 
    Dynamics::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->toggle_params_lock();

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_c = mgr->params()->num_categories();
    uint32_t N_i = mgr->params()->num_instances();

    double E = 0;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    if (t == 0)
    {
        for (uint32_t c = 0; c < N_c; c++)
        {
            E += mgr->params()->cat_lp_potential(c);
        }
    }
    mgr->desynchronize();

    for (uint32_t c = 0; c < N_c; c++)
    {
        for (uint32_t i = t; i < N_i; i += N_t)
        {
            E += mgr->params()->ins_ll_potential(c, i);
        }
    }

    *E_out = E;

    return 0;
}

/**
 * @param  grad_out  Gradient of the negative log likelihood and prior of the
 *                   parameters in @em params
 * @param  mgr       Dynamics::Thread_manager to use.
 *
 * @return On success, 0 is returned. On error, a jwsc::Error is returned.
 */
jwsc::Error* 
Dynamics::Thread_manager::get_var_param_stats_and_values_grad_potential_c
(
    jwsc::Vector_d**          grad_out,
    Dynamics::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->toggle_params_lock();

    uint32_t t    = mgr->thread();
    uint32_t N_t  = mgr->num_threads();
    uint32_t N_c  = mgr->params()->num_categories();
    uint32_t N_i  = mgr->params()->num_instances();
    uint32_t D_tv = mgr->params()->num_cat_var_params(t);
    uint32_t D_ts = 2*D_tv;
    uint32_t d    = 0;
    uint32_t D    = (N_i % N_t > t) ? ((N_i / N_t) + 1)*D_tv : (N_i / N_t)*D_tv;

    if (t == 0)
    {
        D += D_ts;
    }

    create_vector_d(grad_out, D);
    Vector_d* grad = *grad_out;
    Vector_d* grad_i = 0;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    if (t == 0)
    {
        mgr->params()->grad_cat_potential(&grad_i);
        copy_vector_section_into_vector_d(grad, d, grad_i, 0, D_ts);
        d += D_ts;
    }
    mgr->desynchronize();

    for (uint32_t c = 0; c < N_c; c++)
    {
        for (uint32_t i = t; i < N_i+t; i += N_t)
        {
            if (i < N_i)
            {
                uint32_t D_i = mgr->params()->num_var_params(c, i);
                mgr->params()->grad_ins_potential(&grad_i, c, i);
                copy_vector_section_into_vector_d(grad, d, grad_i, 0, D_i);
                d += D_i;
            }
            mgr->toggle_params_lock();
        }
    }

    assert(get_num_unhandled_errors() == 0);
    assert(d == D);

    free_vector_d(grad_i);

    return 0;
}

/**
 * @param  delta_t_out  Stochastic dynamics delta_t values for the parameters.
 * @param  mgr          Dynamics::Thread_manager to use.
 */
void Dynamics::Thread_manager::get_var_param_stats_and_values_dynamics_delta_t_c
(
    jwsc::Vector_d**          delta_t_out, 
    Dynamics::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->toggle_params_lock();

    uint32_t t    = mgr->thread();
    uint32_t N_t  = mgr->num_threads();
    uint32_t N_c  = mgr->params()->num_categories();
    uint32_t N_i  = mgr->params()->num_instances();
    uint32_t D_tv = mgr->params()->num_cat_var_params(t);
    uint32_t D_ts = 2*D_tv;
    uint32_t d    = 0;
    uint32_t D    = (N_i % N_t > t) ? ((N_i / N_t) + 1)*D_tv : (N_i / N_t)*D_tv;

    if (t == 0)
    {
        D += D_ts;
    }

    create_vector_d(delta_t_out, D);
    Vector_d* delta_t = *delta_t_out;
    Vector_d* delta_t_i = 0;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    if (t == 0)
    {
        mgr->params()->get_var_param_stats_dynamics_delta_t(&delta_t_i);
        copy_vector_section_into_vector_d(delta_t, d, delta_t_i, 0, D_ts);
        d += D_ts;
    }
    mgr->desynchronize();

    for (uint32_t c = 0; c < N_c; c++)
    {
        for (uint32_t i = t; i < N_i; i += N_t)
        {
            uint32_t D_i = mgr->params()->num_var_params(c, i);

            mgr->params()->get_var_param_values_dynamics_delta_t(&delta_t_i, 
                    c, i);
            copy_vector_section_into_vector_d(delta_t, d, delta_t_i, 0, D_i);
            d += D_i;
        }
    }

    assert(get_num_unhandled_errors() == 0);
    assert(d == D);

    free_vector_d(delta_t_i);
}

/**
 * @param  q    Model parameter values.
 * @param  p    Model parameter momentum values.
 * @param  mgr  Dynamics::Thread_manager to use.
 *
 * @return On success, 0 is returned. On error, a jwsc::Error is returned.
 */
jwsc::Error* 
Dynamics::Thread_manager::accept_var_param_stats_and_values_dynamics_c
(
    const jwsc::Vector_d*     q,
    const jwsc::Vector_d*     p,
    Dynamics::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->toggle_params_lock();

    uint32_t t    = mgr->thread();
    uint32_t N_t  = mgr->num_threads();
    uint32_t N_c  = mgr->params()->num_categories();
    uint32_t N_i  = mgr->params()->num_instances();
    uint32_t D_tv = mgr->params()->num_cat_var_params(t);
    uint32_t D_ts = 2*D_tv;
    uint32_t d    = 0;
    uint32_t D    = (N_i % N_t > t) ? ((N_i / N_t) + 1)*D_tv : (N_i / N_t)*D_tv;

    Vector_d* q_i = 0;
    Vector_d* p_i = 0;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        for (uint32_t c = 0; c < N_c; c++)
        {
            for (uint32_t i = 0; i < N_i; i++)
            {
                mgr->params()->instance(c, i)->iteration++;
            }
        }
        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    if (t == 0)
    {
        copy_vector_section_d(&q_i, q, d, D_ts);
        copy_vector_section_d(&p_i, p, d, D_ts);
        mgr->params()->accept_var_param_stats_dynamics(q_i, p_i);

        d += D_ts;
        D += D_ts;
    }
    mgr->desynchronize();

    for (uint32_t c = 0; c < N_c; c++)
    {
        for (uint32_t i = t; i < N_i+t; i += N_t)
        {
            if (i < N_i)
            {
                uint32_t D_i = mgr->params()->num_var_params(c, i);
                copy_vector_section_d(&q_i, q, d, D_i);
                copy_vector_section_d(&p_i, p, d, D_i);
                mgr->params()->accept_var_param_values_dynamics(q_i, p_i, c, i);
                d += D_i;
            }
            mgr->toggle_params_lock();
        }
    }

    assert(get_num_unhandled_errors() == 0);
    assert(d == D);

    free_vector_d(q_i);
    free_vector_d(p_i);

    return 0;
}

/* ========================================================================== */








/* -----------------------  Stochastic_dynamics  ---------------------------- */

/**
 * @param  alpha  Proportion of momentum to conserve per iteration. Must be in
 *                [0,1]. (1-alpha*alpha) scales gaussian noise added to the
 *                momentum per iteration.
 * @param  kick   Frequency to kick the momenta of the simulated particle. 
 *
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 */
Stochastic_dynamics::Stochastic_dynamics(double alpha, uint32_t kick)
    throw (jwscxx::base::Arg_error)
{
    this->alpha(alpha);
    this->kick(kick);
}


Stochastic_dynamics::~Stochastic_dynamics()
{
}


double Stochastic_dynamics::alpha() const
{
    return _alpha;
}


void Stochastic_dynamics::alpha(double a) throw (jwscxx::base::Arg_error)
{
    if (a < 0 || a > 1)
    {
        throw jwscxx::base::Arg_error("Alpha not in [0,1]");
    }

    _alpha = a;
}


uint32_t Stochastic_dynamics::kick() const
{
    return _kick;
}


void Stochastic_dynamics::kick(uint32_t k) throw (jwscxx::base::Arg_error)
{
    _kick = k;
}


void Stochastic_dynamics::run_thread(Dynamics::Thread_manager* mgr) const 
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    typedef void   (*f1)(Vector_d**, const void*);
    typedef Error* (*f2)(Vector_d**, void*);
    typedef Error* (*f3)(const Vector_d*, const Vector_d*, void*);

    mgr->lock_sampler_params();

    Vector_d* delta_t = 0;
    mgr->get_var_param_stats_and_values_dynamics_delta_t_c(&delta_t, mgr);

    Error* err = stochastic_dynamics_2_d(mgr->iterations(), delta_t, alpha(), 
            kick(), mgr,
            (f1)(mgr->get_var_param_stats_and_values_c),
            (f2)(mgr->get_var_param_stats_and_values_grad_potential_c), 
            (f3)(mgr->accept_var_param_stats_and_values_dynamics_c));

    free_vector_d(delta_t);

    mgr->unlock_sampler_params();

    if (err)
    {
        throw jwscxx::base::Exception(err->msg);
    }
}

/* ========================================================================== */







/* -------------------  Stochastic_dynamics::Instance  ---------------------- */

/**
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 * @throw  jwscxx::base::IO_error   Could not open info file.
 */
Stochastic_dynamics::Instance::Instance
(
    View*                       best_view,
    Object*                     best_obj,
    double                      best_img_ll,
    double                      best_potential,
    View*                       view,
    Object*                     obj,
    double                      img_ll,
    double                      potential,
    const jwsc::Matrix_f*       image_t,
    const jwsc::Edge_set_f*     edges,
    Likelihood::Correspondence* img_ll_corr,
    Likelihood::Cache*          img_ll_cache,
    jwsc::Matrix_u32*           img_ll_surf_assign,
    jwsc::Matrix_u32*           img_ll_edge_assign,
    jwsc::Matrix_u32*           model_mat,
    jwsc::Vector_d*             sample_mean_in,
    jwsc::Matrix_d*             sample_cov_in,
    const char*                 sample_info_fname,
    bool                        compress_sample_info,
    const char*                 sample_stats_info_fname
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
: Dynamics::Instance(best_view, best_obj, best_img_ll, best_potential,
        view, obj, img_ll, potential, image_t, edges, img_ll_corr, img_ll_cache,
        img_ll_surf_assign, img_ll_edge_assign, model_mat, sample_mean_in,
        sample_cov_in, sample_info_fname, compress_sample_info,
        sample_stats_info_fname)
{
}


Stochastic_dynamics::Instance::~Instance()
{
}

/* ========================================================================== */








/* -------------------  Stochastic_dynamics::Category  ---------------------- */

/**
 * @param  rep_ins            Representative instance for the category.
 * @param  lp_potential       Log prior potential energy.
 * @param  potential          Potential energy of the likelihood and prior.
 * @param  sample_info_fname  File to write sample information to.
 *
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 * @throw  jwscxx::base::IO_error   Could not open info file.
 */
Stochastic_dynamics::Category::Category
(
    const Instance* rep_ins,
    double          lp_potential,
    double          potential,
    const char*     sample_info_fname,
    bool            compress_sample_info
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
: Dynamics::Category(rep_ins, lp_potential, potential, sample_info_fname,
        compress_sample_info)
{
}

Stochastic_dynamics::Category::~Category()
{
}

/* -------------------------------------------------------------------------- */








/* ----------------  Stochastic_dynamics::Parameters  ----------------------- */

Stochastic_dynamics::Parameters::Parameters
(
    const View::Context*      view_ctx,
    const Object::Context*    obj_ctx,
    const Block::Context*      block_ctx,
    const Block_set::Context* block_set_ctx,
    double                    img_ll_bg_prob,
    double                    img_ll_fg_prob,
    double                    img_ll_fg_noise_prob,
    double                    img_ll_fg_miss_prob,
    double                    img_ll_edge_noise_prob,
    double                    img_ll_edge_silho_miss_prob,
    double                    img_ll_edge_inner_miss_prob,
    uint32_t                  sample_stats_freq,
    uint32_t                  num_instances,
    uint32_t                  num_categories,
    uint32_t                  min_num_members,
    uint32_t                  num_threads
)
: Dynamics::Parameters(view_ctx, obj_ctx, block_ctx, block_set_ctx,
        img_ll_bg_prob, img_ll_fg_prob, img_ll_fg_noise_prob,
        img_ll_fg_miss_prob, img_ll_edge_noise_prob,
        img_ll_edge_silho_miss_prob, img_ll_edge_inner_miss_prob,
        sample_stats_freq, num_instances, num_categories, min_num_members,
        num_threads)
{
}


Stochastic_dynamics::Parameters::~Parameters()
{
}


/**
 * @param  i  Index of the Instance to get.
 * @param  c  Index of the category.
 *
 * @return A pointer to the indexed Instance.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Instance index.
 */
Stochastic_dynamics::Instance* Stochastic_dynamics::Parameters::instance
(
    uint32_t c,
    uint32_t i
) 
const throw (jwscxx::base::Arg_error)
{
    if (i >= num_instances())
    {
        throw jwscxx::base::Arg_error("Model set index too large");
    }

    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    Stochastic_dynamics::Instance* ins;
    assert((ins = dynamic_cast<Stochastic_dynamics::Instance*>(
                    _instances[ c ][ i ])));

    return ins;
}

/**
 * @param  c  Index of the category.
 *
 * @return A pointer to the indexed Category.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Category index.
 */
Stochastic_dynamics::Category* Stochastic_dynamics::Parameters::category
(
    uint32_t c
) 
const throw (jwscxx::base::Arg_error)
{
    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    Stochastic_dynamics::Category* cat;
    assert((cat = dynamic_cast<Stochastic_dynamics::Category*>(
                    _categories[ c ])));

    return cat;
}

/* ========================================================================== */








/* --------------------------  Hybrid_MC  ----------------------------------- */

/**
 * @param  alpha       Proportion of momentum to conserve per iteration. Must
 *                     be in [0,1]. (1-alpha*alpha) scales gaussian noise
 *                     added to the momentum per iteration.
 * @param  kick        Frequency to kick the momenta of the simulated particle. 
 * @param  leapfrog_i  Number of leapfrog iterations between each iteration of
 *                     the sampler.
 *
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 */
Hybrid_MC::Hybrid_MC(double alpha, uint32_t kick, uint32_t leapfrog_i)
    throw (jwscxx::base::Arg_error) : Stochastic_dynamics(alpha, kick)
{
    this->leapfrog_i(leapfrog_i);
}


Hybrid_MC::~Hybrid_MC()
{
}


uint32_t Hybrid_MC::leapfrog_i() const
{
    return _leapfrog_i;
}


void Hybrid_MC::leapfrog_i(uint32_t i)
{
    _leapfrog_i = i;
}


void Hybrid_MC::run_thread(Dynamics::Thread_manager* mgr) const 
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    typedef void   (*f1)(Vector_d**, const void*);
    typedef Error* (*f2)(const Vector_d*, void*);
    typedef Error* (*f3)(double*, void*);
    typedef Error* (*f4)(Vector_d**, void*);
    typedef Error* (*f5)(const Vector_d*, const Vector_d*, void*);

    Vector_d* delta_t = 0;
    mgr->get_var_param_stats_and_values_dynamics_delta_t_c(&delta_t, mgr);

    Error* err = hybrid_monte_carlo_2_d(mgr->iterations(), leapfrog_i(), 
            delta_t, alpha(), kick(), mgr,
            (f1)mgr->get_var_param_stats_and_values_c,
            (f2)mgr->set_var_param_stats_and_values_c,
            (f3)mgr->get_var_param_stats_and_values_potential_c,
            (f4)mgr->get_var_param_stats_and_values_grad_potential_c,
            (f5)mgr->accept_var_param_stats_and_values_dynamics_c,
            NULL);

    free_vector_d(delta_t);

    if (err)
    {
        throw jwscxx::base::Exception(err->msg);
    }
}

/* ========================================================================== */







/* -----------------------  Hybrid_MC::Instance  ---------------------------- */

/**
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 * @throw  jwscxx::base::IO_error   Could not open info file.
 */
Hybrid_MC::Instance::Instance
(
    View*                       best_view,
    Object*                     best_obj,
    double                      best_img_ll,
    double                      best_potential,
    View*                       view,
    Object*                     obj,
    double                      img_ll,
    double                      potential,
    const jwsc::Matrix_f*       image_t,
    const jwsc::Edge_set_f*     edges,
    Likelihood::Correspondence* img_ll_corr,
    Likelihood::Cache*          img_ll_cache,
    jwsc::Matrix_u32*           img_ll_surf_assign,
    jwsc::Matrix_u32*           img_ll_edge_assign,
    jwsc::Matrix_u32*           model_mat,
    jwsc::Vector_d*             sample_mean_in,
    jwsc::Matrix_d*             sample_cov_in,
    const char*                 sample_info_fname,
    bool                        compress_sample_info,
    const char*                 sample_stats_info_fname
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
: Stochastic_dynamics::Instance(best_view, best_obj, best_img_ll,
        best_potential, view, obj, img_ll, potential, image_t, edges,
        img_ll_corr, img_ll_cache, img_ll_surf_assign, img_ll_edge_assign,
        model_mat, sample_mean_in, sample_cov_in, sample_info_fname,
        compress_sample_info, sample_stats_info_fname)
{
}


Hybrid_MC::Instance::~Instance()
{
}

/* ========================================================================== */








/* -----------------------  Hybrid_MC::Category  ---------------------------- */

/**
 * @param  rep_ins            Representative instance for the category.
 * @param  lp_potential       Log prior potential energy.
 * @param  potential          Potential energy of the likelihood and prior.
 * @param  sample_info_fname  File to write sample information to.
 *
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 * @throw  jwscxx::base::IO_error   Could not open info file.
 */
Hybrid_MC::Category::Category
(
    const Instance* rep_ins,
    double          lp_potential,
    double          potential,
    const char*     sample_info_fname,
    bool            compress_sample_info
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
: Stochastic_dynamics::Category(rep_ins, lp_potential, potential, 
        sample_info_fname, compress_sample_info)
{
}

Hybrid_MC::Category::~Category()
{
}

/* -------------------------------------------------------------------------- */








/* --------------------  Hybrid_MC::Parameters  ----------------------------- */

Hybrid_MC::Parameters::Parameters
(
    const View::Context*      view_ctx,
    const Object::Context*    obj_ctx,
    const Block::Context*     block_ctx,
    const Block_set::Context* block_set_ctx,
    double                    img_ll_bg_prob,
    double                    img_ll_fg_prob,
    double                    img_ll_fg_noise_prob,
    double                    img_ll_fg_miss_prob,
    double                    img_ll_edge_noise_prob,
    double                    img_ll_edge_silho_miss_prob,
    double                    img_ll_edge_inner_miss_prob,
    uint32_t                  sample_stats_freq,
    uint32_t                  num_instances,
    uint32_t                  num_categories,
    uint32_t                  min_num_members,
    uint32_t                  num_threads
)
: Stochastic_dynamics::Parameters(view_ctx, obj_ctx, block_ctx, block_set_ctx,
        img_ll_bg_prob, img_ll_fg_prob, img_ll_fg_noise_prob,
        img_ll_fg_miss_prob, img_ll_edge_noise_prob,
        img_ll_edge_silho_miss_prob, img_ll_edge_inner_miss_prob,
        sample_stats_freq, num_instances, num_categories, min_num_members,
        num_threads)
{
}


Hybrid_MC::Parameters::~Parameters()
{
}


/**
 * @param  i  Index of the Instance to get.
 * @param  c  Index of the category.
 *
 * @return A pointer to the indexed Instance.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Instance index.
 */
Hybrid_MC::Instance* Hybrid_MC::Parameters::instance
(
    uint32_t c,
    uint32_t i
) 
const throw (jwscxx::base::Arg_error)
{
    if (i >= num_instances())
    {
        throw jwscxx::base::Arg_error("Model set index too large");
    }

    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    Hybrid_MC::Instance* ins;
    assert((ins = dynamic_cast<Hybrid_MC::Instance*>(_instances[ c ][ i ])));

    return ins;
}

/**
 * @param  c  Index of the category.
 *
 * @return A pointer to the indexed Category.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Category index.
 */
Hybrid_MC::Category* Hybrid_MC::Parameters::category(uint32_t c) const 
    throw (jwscxx::base::Arg_error)
{
    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    Hybrid_MC::Category* cat;
    assert((cat = dynamic_cast<Hybrid_MC::Category*>(_categories[ c ])));

    return cat;
}

/* ========================================================================== */








/* ---------------------------  Langevin  ----------------------------------- */

void Langevin::run_thread(Dynamics::Thread_manager* mgr) const 
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    typedef void   (*f1)(Vector_d**, const void*);
    typedef Error* (*f2)(Vector_d**, void*);
    typedef Error* (*f3)(const Vector_d*, void*);

    Vector_d* delta_t = 0;
    mgr->get_var_param_stats_and_values_dynamics_delta_t_c(&delta_t, mgr);

    Error* err = langevin_2_d(mgr->iterations(), delta_t, mgr,
            (f1)(mgr->get_var_param_stats_and_values_c),
            (f2)(mgr->get_var_param_stats_and_values_grad_potential_c), 
            (f3)(mgr->accept_var_param_stats_and_values_dynamics_c));

    free_vector_d(delta_t);

    if (err)
    {
        throw jwscxx::base::Exception(err->msg);
    }
}

/* ========================================================================== */







/* ------------------------  Langevin::Instance  ---------------------------- */

/**
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 * @throw  jwscxx::base::IO_error   Could not open info file.
 */
Langevin::Instance::Instance
(
    View*                       best_view,
    Object*                     best_obj,
    double                      best_img_ll,
    double                      best_potential,
    View*                       view,
    Object*                     obj,
    double                      img_ll,
    double                      potential,
    const jwsc::Matrix_f*       image_t,
    const jwsc::Edge_set_f*     edges,
    Likelihood::Correspondence* img_ll_corr,
    Likelihood::Cache*          img_ll_cache,
    jwsc::Matrix_u32*           img_ll_surf_assign,
    jwsc::Matrix_u32*           img_ll_edge_assign,
    jwsc::Matrix_u32*           model_mat,
    jwsc::Vector_d*             sample_mean_in,
    jwsc::Matrix_d*             sample_cov_in,
    const char*                 sample_info_fname,
    bool                        compress_sample_info,
    const char*                 sample_stats_info_fname
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
: Dynamics::Instance(best_view, best_obj, best_img_ll, best_potential,
        view, obj, img_ll, potential, image_t, edges, img_ll_corr, img_ll_cache,
        img_ll_surf_assign, img_ll_edge_assign, model_mat, sample_mean_in,
        sample_cov_in, sample_info_fname, compress_sample_info,
        sample_stats_info_fname)
{
}


Langevin::Instance::~Instance()
{
}

/* ========================================================================== */








/* ------------------------  Langevin::Category  ---------------------------- */

/**
 * @param  rep_ins            Representative instance for the category.
 * @param  lp_potential       Log prior potential energy.
 * @param  potential          Potential energy of the likelihood and prior.
 * @param  sample_info_fname  File to write sample information to.
 *
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 * @throw  jwscxx::base::IO_error   Could not open info file.
 */
Langevin::Category::Category
(
    const Instance* rep_ins,
    double          lp_potential,
    double          potential,
    const char*     sample_info_fname,
    bool            compress_sample_info
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
: Dynamics::Category(rep_ins, lp_potential, potential, sample_info_fname,
        compress_sample_info)
{
}

Langevin::Category::~Category()
{
}

/* -------------------------------------------------------------------------- */








/* -----------------------  Langevin::Parameters  --------------------------- */

Langevin::Parameters::Parameters
(
    const View::Context*      view_ctx,
    const Object::Context*    obj_ctx,
    const Block::Context*     block_ctx,
    const Block_set::Context* block_set_ctx,
    double                    img_ll_bg_prob,
    double                    img_ll_fg_prob,
    double                    img_ll_fg_noise_prob,
    double                    img_ll_fg_miss_prob,
    double                    img_ll_edge_noise_prob,
    double                    img_ll_edge_silho_miss_prob,
    double                    img_ll_edge_inner_miss_prob,
    uint32_t                  sample_stats_freq,
    uint32_t                  num_instances,
    uint32_t                  num_categories,
    uint32_t                  min_num_members,
    uint32_t                  num_threads
)
: Dynamics::Parameters(view_ctx, obj_ctx, block_ctx, block_set_ctx,
        img_ll_bg_prob, img_ll_fg_prob, img_ll_fg_noise_prob,
        img_ll_fg_miss_prob, img_ll_edge_noise_prob,
        img_ll_edge_silho_miss_prob, img_ll_edge_inner_miss_prob,
        sample_stats_freq, num_instances, num_categories, min_num_members,
        num_threads)
{
}


Langevin::Parameters::~Parameters()
{
}


/**
 * @param  i  Index of the Instance to get.
 * @param  c  Index of the category.
 *
 * @return A pointer to the indexed Instance.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Instance index.
 */
Langevin::Instance* Langevin::Parameters::instance(uint32_t c, uint32_t i)
    const throw (jwscxx::base::Arg_error)
{
    if (i >= num_instances())
    {
        throw jwscxx::base::Arg_error("Model set index too large");
    }

    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    Langevin::Instance* ins;
    assert((ins = dynamic_cast<Langevin::Instance*>(_instances[ c ][ i ])));

    return ins;
}

/**
 * @param  c  Index of the category.
 *
 * @return A pointer to the indexed Category.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Category index.
 */
Langevin::Category* Langevin::Parameters::category(uint32_t c) const 
    throw (jwscxx::base::Arg_error)
{
    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    Langevin::Category* cat;
    assert((cat = dynamic_cast<Langevin::Category*>(_categories[ c ])));

    return cat;
}

/* ========================================================================== */








/* --------------------------  Hyperdynamics  ------------------------------- */

/**
 * @param  gamma   Step size for gradient descent for each parameter. Very
 *                 small positive value.
 * @param  eta     Difference parameter for derivative numerical
 *                 approximations. Small positive value.
 * @param  height  Bias scaling of the Voter's function (h). Height of the
 *                 smallest transition area is good starting value.
 * @param  dist    Distance scale from one minima to another in terms of
 *                 2*pi*d.
 *
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 */
Hyperdynamics::Hyperdynamics
(
    double gamma,
    double eta,
    double height,
    double dist
)
throw (jwscxx::base::Arg_error) 
{
    this->gamma(gamma);
    this->eta(eta);
    this->height(height);
    this->dist(dist);
}


Hyperdynamics::~Hyperdynamics()
{
}


double Hyperdynamics::gamma() const
{
    return _gamma;
}


void Hyperdynamics::gamma(double gamma) throw (jwscxx::base::Arg_error)
{
    if (gamma <= 0)
    {
        throw jwscxx::base::Arg_error("Gamma too small");
    }
    _gamma = gamma;
}


double Hyperdynamics::eta() const
{
    return _eta;
}


void Hyperdynamics::eta(double eta) throw (jwscxx::base::Arg_error)
{
    if (eta <= 0)
    {
        throw jwscxx::base::Arg_error("Eta too small");
    }
    _eta = eta;
}


double Hyperdynamics::height() const
{
    return _height;
}


void Hyperdynamics::height(double h)
{
    _height = h;
}


double Hyperdynamics::dist() const
{
    return _dist;
}


void Hyperdynamics::dist(double d)
{
    _dist = d;
}


void Hyperdynamics::run_thread(Hyperdynamics::Thread_manager* mgr) const 
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    typedef void   (*f1)(Vector_d**, const void*);
    typedef Error* (*f2)(const Vector_d*, void*);
    typedef Error* (*f3)(double*, void*);
    typedef Error* (*f4)(Vector_d**, void*);
    typedef Error* (*f5)(const Vector_d*, void*, double);

    Vector_d* delta_t = 0;
    mgr->get_var_param_stats_and_values_dynamics_delta_t_c(&delta_t, mgr);

    Error* err = langevin_hyperdynamics_2_d(mgr->iterations(), delta_t, gamma(),
            eta(), height(), dist(), mgr,
            (f1)mgr->get_var_param_stats_and_values_c,
            (f2)mgr->set_var_param_stats_and_values_c,
            (f3)mgr->get_var_param_stats_and_values_potential_c,
            (f4)mgr->get_var_param_stats_and_values_grad_potential_c,
            (f5)mgr->accept_bias_var_param_values_c);

    free_vector_d(delta_t);

    if (err)
    {
        throw jwscxx::base::Exception(err->msg);
    }
}

/* ========================================================================== */







/* -------------------  Hyperdynamics::Thread_manager  ---------------------- */

Hyperdynamics::Thread_manager::Thread_manager(uint32_t thread)
    : Dynamics::Thread_manager(thread)
{
}


Hyperdynamics::Thread_manager::~Thread_manager() 
{ 
}


void Hyperdynamics::Thread_manager::sampler(Sampler* sampler)
    throw (jwscxx::base::Arg_error)
{
    if (!(_sampler = dynamic_cast<Hyperdynamics*>(sampler)))
    {
        throw jwscxx::base::Arg_error(
               "Hyperdynamics thread manager requires a hyperdynamics sampler");
    }
}

Hyperdynamics* Hyperdynamics::Thread_manager::sampler() const
{
    return dynamic_cast<Hyperdynamics*>(_sampler);
}

void Hyperdynamics::Thread_manager::params
(
    Sampler::Parameters* params
)
throw (jwscxx::base::Arg_error)
{
    if (!(_params = dynamic_cast<Hyperdynamics::Parameters*>(params)))
    {
        throw jwscxx::base::Arg_error(
          "Hyperdynamics thread manager requires hyperdynamics sampler params");
    }
}

Hyperdynamics::Parameters* Hyperdynamics::Thread_manager::params() const
{
    return dynamic_cast<Hyperdynamics::Parameters*>(_params);
}

/**
 * @param  p     Point to accept and update this Parameters with.
 * @param  mgr   Sampler thread manager.
 * @param  bias  Energy bias applied at the point.
 *
 * @return On success, 0 is returned. On error, a jwsc::Error is returned.
 */
jwsc::Error* Hyperdynamics::Thread_manager::accept_bias_var_param_values_c
(
    const jwsc::Vector_d*          p, 
    Hyperdynamics::Thread_manager* mgr,
    double                         bias
)
{
    using namespace jwsc;

    mgr->toggle_params_lock();

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_c = mgr->params()->num_categories();
    uint32_t N_i = mgr->params()->num_instances();
    uint32_t D_t = mgr->params()->num_cat_var_params(t);
    uint32_t d   = 0;
    uint32_t D   = (N_i % N_t > t) ? ((N_i / N_t) + 1)*D_t : (N_i / N_t)*D_t;

    Vector_d* p_i = 0;

    for (uint32_t c = 0; c < N_c; c++)
    {
        for (uint32_t i = t; i < N_i; i += N_t)
        {
            uint32_t D_i = mgr->params()->num_var_params(c, i);

            copy_vector_section_d(&p_i, p, d, D_i);
            mgr->params()->accept_bias_var_param_values(p_i, bias, c, i);
            d += D_t;
        }
    }

    assert(d == D);
    free_vector_d(p_i);

    return 0;
}

/* ========================================================================== */





/* ---------------------  Hyperdynamics::Instance  -------------------------- */

/**
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 * @throw  jwscxx::base::IO_error   Could not open info file.
 */
Hyperdynamics::Instance::Instance
(
    View*                       best_view,
    Object*                     best_obj,
    double                      best_img_ll,
    double                      best_potential,
    View*                       view,
    Object*                     obj,
    double                      img_ll,
    double                      potential,
    const jwsc::Matrix_f*       image_t,
    const jwsc::Edge_set_f*     edges,
    Likelihood::Correspondence* img_ll_corr,
    Likelihood::Cache*          img_ll_cache,
    jwsc::Matrix_u32*           img_ll_surf_assign,
    jwsc::Matrix_u32*           img_ll_edge_assign,
    jwsc::Matrix_u32*           model_mat,
    jwsc::Vector_d*             sample_mean_in,
    jwsc::Matrix_d*             sample_cov_in,
    const char*                 sample_info_fname,
    bool                        compress_sample_info,
    const char*                 sample_stats_info_fname
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
: Langevin::Instance(best_view, best_obj, best_img_ll, best_potential,
        view, obj, img_ll, potential, image_t, edges, img_ll_corr, img_ll_cache,
        img_ll_surf_assign, img_ll_edge_assign, model_mat, sample_mean_in,
        sample_cov_in, sample_info_fname, compress_sample_info,
        sample_stats_info_fname)
{
}

Hyperdynamics::Instance::~Instance()
{
}

/* ========================================================================== */








/* --------------------  Hyperdynamics::Category  --------------------------- */

/**
 * @param  rep_ins            Representative instance for the category.
 * @param  lp_potential       Log prior potential energy.
 * @param  potential          Potential energy of the likelihood and prior.
 * @param  sample_info_fname  File to write sample information to.
 *
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 * @throw  jwscxx::base::IO_error   Could not open info file.
 */
Hyperdynamics::Category::Category
(
    const Instance* rep_ins,
    double          lp_potential,
    double          potential,
    const char*     sample_info_fname,
    bool            compress_sample_info
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
: Langevin::Category(rep_ins, lp_potential, potential, sample_info_fname,
        compress_sample_info)
{
}

Hyperdynamics::Category::~Category()
{
}

/* -------------------------------------------------------------------------- */








/* --------------------  Hyperdynamics::Parameters  ------------------------- */

Hyperdynamics::Parameters::Parameters
(
    const View::Context*      view_ctx,
    const Object::Context*    obj_ctx,
    const Block::Context*     block_ctx,
    const Block_set::Context* block_set_ctx,
    double                    img_ll_bg_prob,
    double                    img_ll_fg_prob,
    double                    img_ll_fg_noise_prob,
    double                    img_ll_fg_miss_prob,
    double                    img_ll_edge_noise_prob,
    double                    img_ll_edge_silho_miss_prob,
    double                    img_ll_edge_inner_miss_prob,
    uint32_t                  sample_stats_freq,
    uint32_t                  num_instances,
    uint32_t                  num_categories,
    uint32_t                  min_num_members,
    uint32_t                  num_threads
)
: Langevin::Parameters(view_ctx, obj_ctx, block_ctx, block_set_ctx,
        img_ll_bg_prob, img_ll_fg_prob, img_ll_fg_noise_prob,
        img_ll_fg_miss_prob, img_ll_edge_noise_prob,
        img_ll_edge_silho_miss_prob, img_ll_edge_inner_miss_prob,
        sample_stats_freq, num_instances, num_categories, min_num_members,
        num_threads)
{
    _bias = 0;
    _best_bias = 0;
}


Hyperdynamics::Parameters::~Parameters()
{
}


/**
 * @param  i  Index of the Instance to get.
 * @param  c  Index of the category.
 *
 * @return A pointer to the indexed Instance.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Instance index.
 */
Hyperdynamics::Instance* Hyperdynamics::Parameters::instance
(
    uint32_t c,
    uint32_t i
) 
const throw (jwscxx::base::Arg_error)
{
    if (i >= num_instances())
    {
        throw jwscxx::base::Arg_error("Model set index too large");
    }

    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    Hyperdynamics::Instance* ins;
    assert((ins = dynamic_cast<Hyperdynamics::Instance*>(
                    _instances[ c ][ i ])));

    return ins;
}

/**
 * @param  c  Index of the category.
 *
 * @return A pointer to the indexed Category.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Category index.
 */
Hyperdynamics::Category* Hyperdynamics::Parameters::category(uint32_t c) 
    const throw (jwscxx::base::Arg_error)
{
    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    Hyperdynamics::Category* cat;
    assert((cat = dynamic_cast<Hyperdynamics::Category*>(_categories[ c ])));

    return cat;
}

double Hyperdynamics::Parameters::bias() const
{
    return _bias;
}


double Hyperdynamics::Parameters::best_bias() const
{
    return _best_bias;
}


/**
 * @param  q     Point to accept and update this Parameters with.
 * @param  bias  Energy bias applied at the point.
 * @param  i     Index of the model instance to use.
 * @param  c     Index of the category.
 */
void Hyperdynamics::Parameters::accept_bias_var_param_values
(
    const jwsc::Vector_d* q, 
    double                bias,
    uint32_t              c,
    uint32_t              i
)
{
    using std::setw;
    using namespace jwsc;

    Vector_d* curr_q = 0;
    Vector_d* deltas = 0;

    get_var_param_values(&curr_q, c, i);
    subtract_vectors_d(&deltas, q, curr_q);

    add_var_param_values_scale_component_deps(deltas, c, i);

    free_vector_d(curr_q);
    free_vector_d(deltas);

    _bias = bias;

    Instance* ins = instance(c, i);
    add_sample(ins);

    try 
    {
        Likelihood::calc_neg_log_likelihood(&(ins->ll_potential),
                &(ins->img_ll), &(ins->model_mat), &(ins->img_ll_surf_assign),
                &(ins->img_ll_edge_assign), ins->image_t, ins->edges,
                ins->img_ll_corr, ins->view, ins->obj, img_ll_bg_prob(),
                img_ll_fg_prob(), img_ll_fg_noise_prob(), img_ll_fg_miss_prob(),
                img_ll_edge_noise_prob(), img_ll_edge_silho_miss_prob(),
                img_ll_edge_inner_miss_prob());
        ins->img_ll_cache->insert(ins->ll_potential, ins->view, ins->obj);
    }
    catch (jwscxx::base::Arg_error e)
    {
        // Should never happen.
        abort();
    }

    // TODO
    //if ((potential() + bias) < (best_potential() + best_bias()))
    if (ins->ll_potential < ins->best_ll_potential)
    {
        _best_bias = _bias;

        *(ins->best_view)   = *(ins->view);
        *(ins->best_obj)    = *(ins->obj);
        ins->best_img_ll    = ins->img_ll;
        ins->best_ll_potential = ins->ll_potential;
    }

    ins->iteration++;
    ins->print_var_param_info();
}

/* ========================================================================== */
