/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief MCMC molecular dynamics based sampler.
 */


#ifndef SAMPLER_DYNAMICS_H
#define SAMPLER_DYNAMICS_H


#include <config.h>

#include <vector>

#include <inttypes.h>

#include <jwsc/vector/vector.h>

#include <jwsc++/base/exception.h>

#include "view.h"
#include "object.h"
#include "block.h"
#include "block_set.h"
#include "sampler.h"


/**
 * @class Dynamics
 *
 * MCMC sampler for finding the most likely set of model parameters under
 * some detected edge points.
 */
class Dynamics : public Sampler
{
    public:


/**
 * @class Dynamics::Instance
 *
 * @brief Instance of models being inferred by the Sampler from data.
 *
 * This class is more or less a C structure with a helpful allocator and
 * deallocator. It's contained by the Sampler::Parameters class and used in a
 * vector of Instance.
 */
class Instance : public Sampler::Instance
{
    public:

        /** @brief Constructs a new Instance for the Sampler. */
        Instance
        (
            View*                       best_view,
            Object*                     best_obj,
            double                      best_img_ll,
            double                      best_potential,
            View*                       view,
            Object*                     obj,
            double                      img_ll,
            double                      potential,
            const jwsc::Matrix_f*       image_t,
            const jwsc::Edge_set_f*     edges,
            Likelihood::Correspondence* img_ll_corr,
            Likelihood::Cache*          img_ll_cache,
            jwsc::Matrix_u32*           img_ll_surf_assign,
            jwsc::Matrix_u32*           img_ll_edge_assign,
            jwsc::Matrix_u32*           model_mat,
            jwsc::Vector_d*             sample_mean_in,
            jwsc::Matrix_d*             sample_cov_in,
            const char*                 sample_info_fname,
            bool                        compress_sample_info,
            const char*                 sample_stats_info_fname
        )
        throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);


        /** @brief Deletes a Instance. */
        virtual ~Instance();
};


/**
 * @class Dynamics::Category
 *
 * @brief Category of objects being inferred by the Sampler from data.
 *
 * This class is more or less a C structure with a helpful allocator and
 * deallocator. It's contained by the Sampler::Parameters class and used in a
 * vector of subcategories.
 */
class Category : public Sampler::Category
{
    public:

        /** @brief Constructs a new Category for the Sampler. */
        Category
        (
            const Instance* rep_ins,
            double          lp_potential,
            double          potential,
            const char*     sample_info_fname,
            bool            compress_sample_info
        )
        throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);

        /** @brief Deletes this Instance. */
        virtual ~Category();
};


/**
 * @class Dynamics::Parameters
 *
 * Parameters for a stochastic dynamics sampler.
 */
class Parameters : public Sampler::Parameters
{
    public:

        /** @brief Constructs a Dynamics::Parameters. */
        Parameters
        (
            const View::Context*      view_ctx,
            const Object::Context*    obj_ctx,
            const Block::Context*     block_ctx,
            const Block_set::Context* block_set_ctx,
            double                    img_ll_bg_prob,
            double                    img_ll_fg_prob,
            double                    img_ll_fg_noise_prob,
            double                    img_ll_fg_miss_prob,
            double                    img_ll_edge_noise_prob,
            double                    img_ll_edge_silho_miss_prob,
            double                    img_ll_edge_inner_miss_prob,
            uint32_t                  sample_stats_freq,
            uint32_t                  num_instances,
            uint32_t                  num_categories,
            uint32_t                  min_num_members,
            uint32_t                  num_threads
        );


        /** @brief Deletes a Sampler_move_parameters. */
        virtual ~Parameters();


        /** 
         * @brief Returns a Instance from this set of sampler Parameters.
         */
        virtual Dynamics::Instance* instance(uint32_t c, uint32_t i)
            const throw (jwscxx::base::Arg_error) = 0;


        /** 
         * @brief Returns a category from this set of sampler Parameters.
         */
        virtual Dynamics::Category* category(uint32_t c) const
            throw (jwscxx::base::Arg_error) = 0;


        /** @brief Accepts a point as a model sample. */
        void accept_var_param_values_dynamics
        (
            const jwsc::Vector_d* q, 
            const jwsc::Vector_d* p, 
            uint32_t              c,
            uint32_t              i
        );

        /** @brief Accepts a point as a model sample. */
        void accept_var_param_stats_dynamics
        (
            const jwsc::Vector_d* q, 
            const jwsc::Vector_d* p, 
            uint32_t              c
        );

        /** @brief Accepts a point as a model sample. */
        void accept_var_param_stats_dynamics
        (
            const jwsc::Vector_d* q, 
            const jwsc::Vector_d* p
        );
};


/**
 * @class Dynamics::Thread_manager
 *
 * Provides the thread synchronization to execute code by only one thread while
 * the others (busy) wait for it to finish. When any thread leaves the
 * synchronization block, the code is guarenteed to have been complete executed.
 *
 * The template for a synchronization block is as follows
 *
 * @code
 * synchronize();
 *
 * if (last_thread())
 * {
 *     // Code to be executed synchronously by only one thread goes here.
 *
 *     release();
 * }
 * else
 * {
 *     wait();
 * }
 *
 * if (thread() == t)
 * {
 *     // Code to be executed by the t^th thread goes here.
 * }
 *
 * // Code to be executed by all threads goes here.
 *
 * desynchronize();
 * @endcode
 */
class Thread_manager : public Sampler::Thread_manager
{
    public:

        /** @brief Constructs a thread manager. */
        Thread_manager(uint32_t thread=0);

        /** @brief Deletes a thread manager. */
        virtual ~Thread_manager();

        /** @brief Sets the Dynamics sampler to use. */
        virtual void sampler(Sampler* sampler) throw (jwscxx::base::Arg_error);

        /** @brief Returns the Dynamics sampler to use. */
        virtual Dynamics* sampler() const;

        /** @brief Sets the Dynamics::Parameters to use. */
        virtual void params(Sampler::Parameters* params)
            throw (jwscxx::base::Arg_error);

        /** @brief Returns the Dynamics::Parameters to use. */
        virtual Dynamics::Parameters* params() const;

        /** 
         * @brief Gets a vector of parameter statistics and valuess from the
         * parameter structure.
         *
         * Static C interface to thread-safe
         * Sampler::Parameters::get_var_param_stats() and
         * Sampler::Parameters::get_var_param_values().
         */
        static void get_var_param_stats_and_values_c
        (
            jwsc::Vector_d**          p_out, 
            Dynamics::Thread_manager* mgr
        );

        /** 
         * @brief Sets the parameter structure from a vector of parameters.
         *
         * Static C interface to thread-safe
         * Sampler::Parameters::set_var_param_stats() and
         * Sampler::Parameters::set_var_param_values().
         */
        static jwsc::Error* set_var_param_stats_and_values_c
        (
            const jwsc::Vector_d*     p, 
            Dynamics::Thread_manager* mgr
        );

        /** 
         * @brief Returns the potential energy of the model in the parameters.
         *
         * Static C interface to a thread-safe 
         * Sampler::Parameters::cat_potential() and
         * Sampler::Parameters::ins_potential().
         */
        static jwsc::Error* get_var_param_stats_and_values_potential_c
        (
            double*                   E_out, 
            Dynamics::Thread_manager* mgr
        );

        /** 
         * @brief Returns the gradient of the potential energy function for
         * the model in the parameters.
         *
         * Static C interface to a thread-safe 
         * Sampler::Parameters::grad_potential().
         */
        static jwsc::Error* get_var_param_stats_and_values_grad_potential_c
        (
            jwsc::Vector_d**          grad_out, 
            Dynamics::Thread_manager* mgr
        );

        /** 
         * @brief Returns a vector of stochastic dynamics delta_t from the
         * parameter structure. 
         *
         * Static C interface to a thread-safe 
         * Sampler::Parameters::get_var_param_dynamics_stats_delta_t() and
         * Sampler::Parameters::get_var_param_dynamics_values_delta_t().
         */
        static void get_var_param_stats_and_values_dynamics_delta_t_c
        (
            jwsc::Vector_d**          delta_t_out, 
            Dynamics::Thread_manager* mgr
        );

        /**
         * @brief Accepts a model point.
         *
         * Static C interface to a thread-safe
         * Dynamics::Parameters::accept_var_param_dynamics_stats() and
         * Dynamics::Parameters::accept_var_param_dynamics_values().
         */
        static jwsc::Error* accept_var_param_stats_and_values_dynamics_c
        (
            const jwsc::Vector_d*     q,
            const jwsc::Vector_d*     p,
            Dynamics::Thread_manager* mgr
        );
};


    // Dynamics
    public:

        /** @brief Runs the iterative MCMC sampler. */
        void run(Dynamics::Parameters* params, uint32_t iterations)
            throw (jwscxx::base::Exception);


    protected:

        /** @brief Runs the iterative MCMC sampler. */
        void run(Sampler::Parameters* params, uint32_t iterations)
            throw (jwscxx::base::Exception);

        virtual void run_thread(Dynamics::Thread_manager* mgr) const
            throw (jwscxx::base::Exception) = 0;


        static jwscxx::base::Exception* run_thread_c
        (
            Dynamics::Thread_manager* mgr
        );
};


/**
 * @class Stochastic_dynamics
 *
 * MCMC sampler for finding the most likely set of model parameters under
 * some detected edge points.
 */
class Stochastic_dynamics : public Dynamics
{
    public:


/**
 * @class Stochastic_dynamics::Instance
 *
 * @brief Instance of models being inferred by the Sampler from data.
 *
 * This class is more or less a C structure with a helpful allocator and
 * deallocator. It's contained by the Sampler::Parameters class and used in a
 * vector of Instance.
 */
class Instance : public Dynamics::Instance
{
    public:

        /** @brief Constructs a new Instance for the Sampler. */
        Instance
        (
            View*                       best_view,
            Object*                     best_obj,
            double                      best_img_ll,
            double                      best_potential,
            View*                       view,
            Object*                     obj,
            double                      img_ll,
            double                      potential,
            const jwsc::Matrix_f*       image_t,
            const jwsc::Edge_set_f*     edges,
            Likelihood::Correspondence* img_ll_corr,
            Likelihood::Cache*          img_ll_cache,
            jwsc::Matrix_u32*           img_ll_surf_assign,
            jwsc::Matrix_u32*           img_ll_edge_assign,
            jwsc::Matrix_u32*           model_mat,
            jwsc::Vector_d*             sample_mean_in,
            jwsc::Matrix_d*             sample_cov_in,
            const char*                 sample_info_fname,
            bool                        compress_sample_info,
            const char*                 sample_stats_info_fname
        )
        throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);


        /** @brief Deletes a Instance. */
        virtual ~Instance();
};


/**
 * @class Stochastic_dynamics::Category
 *
 * @brief Category of objects being inferred by the Sampler from data.
 *
 * This class is more or less a C structure with a helpful allocator and
 * deallocator. It's contained by the Sampler::Parameters class and used in a
 * vector of subcategories.
 */
class Category : public Dynamics::Category
{
    public:

        /** @brief Constructs a new Category for the Sampler. */
        Category
        (
            const Instance* rep_ins,
            double          lp_potential,
            double          potential,
            const char*     sample_info_fname,
            bool            compress_sample_info
        )
        throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);

        /** @brief Deletes this Instance. */
        virtual ~Category();
};


/**
 * @class Stochastic_dynamics::Parameters
 *
 * Parameters for a sampler.
 */
class Parameters : public Dynamics::Parameters
{
    public:

        /** @brief Constructs the sampler Parameters. */
        Parameters
        (
            const View::Context*      view_ctx,
            const Object::Context*    obj_ctx,
            const Block::Context*     block_ctx,
            const Block_set::Context* block_set_ctx,
            double                    img_ll_bg_prob,
            double                    img_ll_fg_prob,
            double                    img_ll_fg_noise_prob,
            double                    img_ll_fg_miss_prob,
            double                    img_ll_edge_noise_prob,
            double                    img_ll_edge_silho_miss_prob,
            double                    img_ll_edge_inner_miss_prob,
            uint32_t                  sample_stats_freq,
            uint32_t                  num_instances,
            uint32_t                  num_categories,
            uint32_t                  min_num_members,
            uint32_t                  num_threads
        );


        /** @brief Deletes a Sampler_move_parameters. */
        virtual ~Parameters();


        /** 
         * @brief Returns a Instance from this set of sampler Parameters.
         */
        virtual Stochastic_dynamics::Instance* instance
        (
            uint32_t c, 
            uint32_t i
        ) 
        const throw (jwscxx::base::Arg_error);


        /** 
         * @brief Returns a category from this set of sampler Parameters.
         */
        virtual Stochastic_dynamics::Category* category(uint32_t c) const
            throw (jwscxx::base::Arg_error);
};


    // Stochastic_dynamics
    public:

        /** @brief Deletes an MCMC Sampler and all its moves. */
        Stochastic_dynamics(double alpha, uint32_t kick) 
            throw (jwscxx::base::Arg_error);


        /** @brief Deletes an MCMC Sampler and all its moves. */
        virtual ~Stochastic_dynamics();


        /** Returns the proportion of momentum to conserve per iteration. */
        double alpha() const;


        /** Sets the proportion of momentum to conserve per iteration. */
        void alpha(double a) throw (jwscxx::base::Arg_error);


        /** Returns the frequency to kick the particle and reset its momenta. */
        uint32_t kick() const;


        /** Sets the frequency to kick the particle and reset its momenta. */
        void kick(uint32_t k) throw (jwscxx::base::Arg_error);


        /** @brief Runs the iterative MCMC sampler. */
        virtual void run_thread(Dynamics::Thread_manager* mgr) const 
            throw (jwscxx::base::Exception);


    protected:

        /** 
         * @brief Proportion of momentum to conserve per iteration.
         *
         * Must be in [0,1].
         *
         * (1-alpha*alpha) scales gaussian noise added to the momentum per
         * iteration.
         */
        double _alpha;

        /**
         * @brief Frequncy (in iterations) to kick the momenta variables of the
         * particle and reset them to a non-stochastic value.
         *
         * Causes the partical to jump out of a local minima.
         */
        uint32_t _kick;
};


/**
 * @class Hybrid_MC
 *
 * MCMC sampler for finding the most likely set of model parameters under
 * some detected edge points.
 */
class Hybrid_MC : public Stochastic_dynamics
{
    public:


/**
 * @class Hybrid_MC::Instance
 *
 * @brief Instance of models being inferred by the Sampler from data.
 *
 * This class is more or less a C structure with a helpful allocator and
 * deallocator. It's contained by the Sampler::Parameters class and used in a
 * vector of Instance.
 */
class Instance : public Stochastic_dynamics::Instance
{
    public:

        /** @brief Constructs a new Instance for the Sampler. */
        Instance
        (
            View*                       best_view,
            Object*                     best_obj,
            double                      best_img_ll,
            double                      best_potential,
            View*                       view,
            Object*                     obj,
            double                      img_ll,
            double                      potential,
            const jwsc::Matrix_f*       image_t,
            const jwsc::Edge_set_f*     edges,
            Likelihood::Correspondence* img_ll_corr,
            Likelihood::Cache*          img_ll_cache,
            jwsc::Matrix_u32*           img_ll_surf_assign,
            jwsc::Matrix_u32*           img_ll_edge_assign,
            jwsc::Matrix_u32*           model_mat,
            jwsc::Vector_d*             sample_mean_in,
            jwsc::Matrix_d*             sample_cov_in,
            const char*                 sample_info_fname,
            bool                        compress_sample_info,
            const char*                 sample_stats_info_fname
        )
        throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);


        /** @brief Deletes a Instance. */
        virtual ~Instance();
};


/**
 * @class Hybrid_MC::Category
 *
 * @brief Category of objects being inferred by the Sampler from data.
 *
 * This class is more or less a C structure with a helpful allocator and
 * deallocator. It's contained by the Sampler::Parameters class and used in a
 * vector of subcategories.
 */
class Category : public Stochastic_dynamics::Category
{
    public:

        /** @brief Constructs a new Category for the Sampler. */
        Category
        (
            const Instance* rep_ins,
            double          lp_potential,
            double          potential,
            const char*     sample_info_fname,
            bool            compress_sample_info
        )
        throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);

        /** @brief Deletes this Instance. */
        virtual ~Category();
};


/**
 * @class Hybrid_MC::Parameters
 *
 * Parameters for a sampler.
 */
class Parameters : public Stochastic_dynamics::Parameters
{
    public:

        /** @brief Constructs the sampler Parameters. */
        Parameters
        (
            const View::Context*      view_ctx,
            const Object::Context*    obj_ctx,
            const Block::Context*     block_ctx,
            const Block_set::Context* block_set_ctx,
            double                    img_ll_bg_prob,
            double                    img_ll_fg_prob,
            double                    img_ll_fg_noise_prob,
            double                    img_ll_fg_miss_prob,
            double                    img_ll_edge_noise_prob,
            double                    img_ll_edge_silho_miss_prob,
            double                    img_ll_edge_inner_miss_prob,
            uint32_t                  sample_stats_freq,
            uint32_t                  num_instances,
            uint32_t                  num_categories,
            uint32_t                  min_num_members,
            uint32_t                  num_threads
        );


        /** @brief Deletes a Parameters. */
        virtual ~Parameters();


        /** 
         * @brief Returns a Instance from this set of sampler Parameters.
         */
        Hybrid_MC::Instance* instance(uint32_t c, uint32_t i) const
            throw (jwscxx::base::Arg_error);


        /** 
         * @brief Returns a category from this set of sampler Parameters.
         */
        Hybrid_MC::Category* category(uint32_t c) const
            throw (jwscxx::base::Arg_error);
};


    // Hybrid_MC
    public:

        /** @brief Constructs a Hybrid MCMC Sampler. */
        Hybrid_MC(double alpha, uint32_t kick, uint32_t leapfrog_i) 
            throw (jwscxx::base::Arg_error);


        /** @brief Deletes an MCMC Sampler and all its moves. */
        virtual ~Hybrid_MC();


        /** 
         * Returns the number of leapfrog iterations between each iteration
         * of the sampler.
         */
        uint32_t leapfrog_i() const;


        /** 
         * Sets the number of leapfrog iterations between each iteration 
         * of the sampler.
         */
        void leapfrog_i(uint32_t i);


        /** @brief Runs the iterative MCMC sampler. */
        void run_thread(Dynamics::Thread_manager* mgr) const
            throw (jwscxx::base::Exception);


    protected:

        /** 
         * @brief Number of leapfrog iterations between each iteration of the
         * sampler.
         */
        uint32_t _leapfrog_i;
};


/**
 * @class Langevin
 *
 * MCMC sampler for finding the most likely set of model parameters under
 * some detected edge points.
 */
class Langevin : public Dynamics
{
    public:


/**
 * @class Langevin::Instance
 *
 * @brief Instance of models being inferred by the Sampler from data.
 *
 * This class is more or less a C structure with a helpful allocator and
 * deallocator. It's contained by the Sampler::Parameters class and used in a
 * vector of Instance.
 */
class Instance : public Dynamics::Instance
{
    public:

        /** @brief Constructs a new Instance for the Sampler. */
        Instance
        (
            View*                       best_view,
            Object*                     best_obj,
            double                      best_img_ll,
            double                      best_potential,
            View*                       view,
            Object*                     obj,
            double                      img_ll,
            double                      potential,
            const jwsc::Matrix_f*       image_t,
            const jwsc::Edge_set_f*     edges,
            Likelihood::Correspondence* img_ll_corr,
            Likelihood::Cache*          img_ll_cache,
            jwsc::Matrix_u32*           img_ll_surf_assign,
            jwsc::Matrix_u32*           img_ll_edge_assign,
            jwsc::Matrix_u32*           model_mat,
            jwsc::Vector_d*             sample_mean_in,
            jwsc::Matrix_d*             sample_cov_in,
            const char*                 sample_info_fname,
            bool                        compress_sample_info,
            const char*                 sample_stats_info_fname
        )
        throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);


        /** @brief Deletes a Instance. */
        virtual ~Instance();
};


/**
 * @class Langevin::Category
 *
 * @brief Category of objects being inferred by the Sampler from data.
 *
 * This class is more or less a C structure with a helpful allocator and
 * deallocator. It's contained by the Sampler::Parameters class and used in a
 * vector of subcategories.
 */
class Category : public Dynamics::Category
{
    public:

        /** @brief Constructs a new Category for the Sampler. */
        Category
        (
            const Instance* rep_ins,
            double          lp_potential,
            double          potential,
            const char*     sample_info_fname,
            bool            compress_sample_info
        )
        throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);

        /** @brief Deletes this Instance. */
        virtual ~Category();
};


/**
 * @class Langevin::Parameters
 *
 * Parameters for a sampler.
 */
class Parameters : public Dynamics::Parameters
{
    public:

        /** @brief Constructs the sampler Parameters. */
        Parameters
        (
            const View::Context*      view_ctx,
            const Object::Context*    obj_ctx,
            const Block::Context*     block_ctx,
            const Block_set::Context* block_set_ctx,
            double                    img_ll_bg_prob,
            double                    img_ll_fg_prob,
            double                    img_ll_fg_noise_prob,
            double                    img_ll_fg_miss_prob,
            double                    img_ll_edge_noise_prob,
            double                    img_ll_edge_silho_miss_prob,
            double                    img_ll_edge_inner_miss_prob,
            uint32_t                  sample_stats_freq,
            uint32_t                  num_instances,
            uint32_t                  num_categories,
            uint32_t                  min_num_members,
            uint32_t                  num_threads
        );


        /** @brief Deletes a Sampler_move_parameters. */
        virtual ~Parameters();


        /** 
         * @brief Returns a Instance from this set of sampler Parameters.
         */
        virtual Langevin::Instance* instance(uint32_t c, uint32_t i)
            const throw (jwscxx::base::Arg_error);


        /** 
         * @brief Returns a category from this set of sampler Parameters.
         */
        virtual Langevin::Category* category(uint32_t c) const
            throw (jwscxx::base::Arg_error);
};


    // Langevin
    public:

        /** @brief Runs the iterative MCMC sampler. */
        virtual void run_thread(Dynamics::Thread_manager* mgr) const 
            throw (jwscxx::base::Exception);
};


/**
 * @class Hyperdynamics
 *
 * MCMC sampler for finding the most likely set of model parameters under
 * some detected edge points.
 */
class Hyperdynamics : public Langevin
{
    public:


/**
 * @class Hyperdynamics::Instance
 *
 * @brief Instance of models being inferred by the Sampler from data.
 *
 * This class is more or less a C structure with a helpful allocator and
 * deallocator. It's contained by the Sampler::Parameters class and used in a
 * vector of Instance.
 */
class Instance : public Langevin::Instance
{
    public:

        /** @brief Constructs a new Instance for the Sampler. */
        Instance
        (
            View*                       best_view,
            Object*                     best_obj,
            double                      best_img_ll,
            double                      best_potential,
            View*                       view,
            Object*                     obj,
            double                      img_ll,
            double                      potential,
            const jwsc::Matrix_f*       image_t,
            const jwsc::Edge_set_f*     edges,
            Likelihood::Correspondence* img_ll_corr,
            Likelihood::Cache*          img_ll_cache,
            jwsc::Matrix_u32*           img_ll_surf_assign,
            jwsc::Matrix_u32*           img_ll_edge_assign,
            jwsc::Matrix_u32*           model_mat,
            jwsc::Vector_d*             sample_mean_in,
            jwsc::Matrix_d*             sample_cov_in,
            const char*                 sample_info_fname,
            bool                        compress_sample_info,
            const char*                 sample_stats_info_fname
        )
        throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);


        /** @brief Deletes a Instance. */
        virtual ~Instance();
};


/**
 * @class Hyperdynamics::Category
 *
 * @brief Category of objects being inferred by the Sampler from data.
 *
 * This class is more or less a C structure with a helpful allocator and
 * deallocator. It's contained by the Sampler::Parameters class and used in a
 * vector of subcategories.
 */
class Category : public Langevin::Category
{
    public:

        /** @brief Constructs a new Category for the Sampler. */
        Category
        (
            const Instance* rep_ins,
            double          lp_potential,
            double          potential,
            const char*     sample_info_fname,
            bool            compress_sample_info
        )
        throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);

        /** @brief Deletes this Instance. */
        virtual ~Category();
};


/**
 * @class Hyperdynamics::Parameters
 *
 * Parameters for a sampler.
 */
class Parameters : public Langevin::Parameters
{
    public:

        /** @brief Constructs the sampler Parameters. */
        Parameters
        (
            const View::Context*      view_ctx,
            const Object::Context*    obj_ctx,
            const Block::Context*     block_ctx,
            const Block_set::Context* block_set_ctx,
            double                    img_ll_bg_prob,
            double                    img_ll_fg_prob,
            double                    img_ll_fg_noise_prob,
            double                    img_ll_fg_miss_prob,
            double                    img_ll_edge_noise_prob,
            double                    img_ll_edge_silho_miss_prob,
            double                    img_ll_edge_inner_miss_prob,
            uint32_t                  sample_stats_freq,
            uint32_t                  num_instances,
            uint32_t                  num_categories,
            uint32_t                  min_num_members,
            uint32_t                  num_threads
        );


        /** @brief Deletes a Sampler_move_parameters. */
        virtual ~Parameters();


        /** 
         * @brief Returns a Instance from this set of sampler Parameters.
         */
        Hyperdynamics::Instance* instance(uint32_t c, uint32_t i)
            const throw (jwscxx::base::Arg_error);


        /** 
         * @brief Returns a category from this set of sampler Parameters.
         */
        Hyperdynamics::Category* category(uint32_t c) const
            throw (jwscxx::base::Arg_error);


        /** @brief Returns the energy bias of the model in the parameters. */
        double bias() const;


        /** 
         * @brief Returns the best energy bias of the model in the
         * parameters.
         */
        double best_bias() const;


        void accept_bias_var_param_values
        (
            const jwsc::Vector_d* q, 
            double                bias,
            uint32_t              c,
            uint32_t              i
        );


    protected:

        /** @brief Bias for the Instance cummulative energy. */
        double _bias;

        /** @brief Best bias for the Instance cummulative energy. */
        double _best_bias;
};


    protected:


/**
 * @class Hyperdynamics::Thread_manager
 *
 * Provides the thread synchronization to execute code by only one thread while
 * the others (busy) wait for it to finish. When any thread leaves the
 * synchronization block, the code is guarenteed to have been complete executed.
 *
 * The template for a synchronization block is as follows
 *
 * @code
 * synchronize();
 *
 * if (last_thread())
 * {
 *     // Code to be executed synchronously by only one thread goes here.
 *
 *     release();
 * }
 * else
 * {
 *     wait();
 * }
 *
 * if (thread() == t)
 * {
 *     // Code to be executed by the t^th thread goes here.
 * }
 *
 * // Code to be executed by all threads goes here.
 *
 * desynchronize();
 * @endcode
 */
class Thread_manager : public Dynamics::Thread_manager
{
    public:

        /** @brief Constructs a Dynamics::Parameters. */
        Thread_manager(uint32_t thread=0);

        /** @brief Deletes a Sampler_move_parameters. */
        virtual ~Thread_manager();

        /** @brief Sets the Dynamics sampler to use. */
        void sampler(Sampler* sampler) throw (jwscxx::base::Arg_error);

        /** @brief Returns the Dynamics sampler to use. */
        Hyperdynamics* sampler() const;

        /** @brief Sets the Dynamics::Parameters to use. */
        void params(Sampler::Parameters* params)
            throw (jwscxx::base::Arg_error);

        /** @brief Returns the Dynamics::Parameters to use. */
        Hyperdynamics::Parameters* params() const;

        /** 
         * @brief Accepts a point as a model sample.
         *
         * Static C inferface to
         * Dynamics::Parameters::accept_var_param_values().
         */
        static jwsc::Error* accept_bias_var_param_values_c
        (
            const jwsc::Vector_d*          p, 
            Hyperdynamics::Thread_manager* mgr,
            double                         bias
        );
};


    // Hyperdynamics
    public:

        /** @brief Constructs an MCMC Sampler. */
        Hyperdynamics(double gamma, double eta, double height, double dist) 
            throw (jwscxx::base::Arg_error);


        /** @brief Deletes an MCMC Sampler and all its moves. */
        virtual ~Hyperdynamics();


        /** 
         * @brief Returns the step size for gradient descent for each
         * parameter.
         */
        double gamma() const;


        /** 
         * @brief Sets the step size for gradient descent for each parameter.
         */
        void gamma(double gamma) throw (jwscxx::base::Arg_error);


        /**
         * @brief Returns the difference parameter for derivative numerical
         * approximations.
         */
        double eta() const;


        /**
         * @brief Sets the difference parameter for derivative numerical
         * approximations.
         */
        void eta(double eta) throw (jwscxx::base::Arg_error);


        /** @brief Returns the bias scaling of Voter's function. */
        double height() const;


        /** 
         * @brief Sets the bias scaling of Voter's function.
         *
         * Height of the smallest transition area is a good starting value.
         */
        void height(double h);


        /** 
         * @brief Returns the distance scale from one minimum to another in
         * terms of 2*pi*d.
         */
        double dist() const;


        /** 
         * @brief Sets the distance scale from one minimum to another in terms
         * of 2*pi*d.
         */
        void dist(double d);


        /** @brief Runs the iterative MCMC sampler. */
        virtual void run_thread(Hyperdynamics::Thread_manager* mgr) const
            throw (jwscxx::base::Exception);


    protected:

        /** 
         * @brief Step size for gradient descent for each parameter. Very
         * small positive value.
         */
        double _gamma;

        /** 
         * @brief Difference parameter for derivative numerical
         * approximations. Small positive value.
         */
        double _eta;

        /** 
         * @brief Bias scaling of the Voter's function (h). Height of the
         * smallest transition area is good starting value.
         */
        double _height;

        /** 
         * @brief Distance scale from one minima to another in terms of
         * 2*pi*d.
         */
        double _dist;
};


#endif
