/* 
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Processes model category parameters.
 */


#include <config.h>

#include <iostream>
#include <cstdlib>
#include <cassert>
#include <list>

#include <inttypes.h>

#if defined FURNITURE_HAVE_OPENGL
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <X11/Xlib.h>
#elif defined FURNITURE_HAVE_OPENGL_FRAMEWORK
#include <OpenGL/OpenGL.h>
#endif

#if defined FURNITURE_HAVE_GLUT_FRAMEWORK
#include <GLUT/glut.h>
#elif defined FURNITURE_HAVE_GLUT
#include <GL/glut.h>
#endif

#include <jwsc/base/error.h>
#include <jwsc/image/image.h>
#include <jwsc/image/image_io.h>

#include <jwsc++/base/option.h>
#include <jwsc++/base/exception.h>

#include "offscreen.h"
#include "render.h"
#include "options.h"


/** @brief Prints the usage of the program. */
void print_usage()
{
    std::cerr << "usage: furniture-cat OPTIONS\n";
    std::cerr << "where OPTIONS is one or more of the following:\n";
}

void sample_category
(
    View**              view_out,
    std::list<Object*>& objs_out,
    View*               cat_view, 
    Object*             cat_obj
)
{
    using namespace jwsc;

    assert(*view_out == 0);
    assert(objs_out.size() == 0);

    Vector_d* p = 0;

    *view_out = cat_view->clone();
    View* view = *view_out;

    cat_view->sample_var_param_likelihood(&p);
    //create_zero_vector_d(&p, 3);
    //p->elts[0] = 0.3;
    p->elts[1] = 1000;
    //p->elts[2] = 0.7;
    try { view->set_var_param_values(p); }
    catch (jwscxx::base::Arg_error e) { }
    view->print_var_param_values(0, std::cout);

    cat_obj->sample_var_param_likelihood(&p);
    double azimuth = p->elts[0];
    double posn_x  = p->elts[1];
    double posn_y  = p->elts[2];
    double posn_z  = p->elts[3];

    for (uint32_t n = 0; n < opts.num_cat_samples; n++)
    {
        Object* obj = cat_obj->clone();
        cat_obj->sample_var_param_likelihood(&p);
        p->elts[0] = azimuth;
        p->elts[1] = posn_x;
        p->elts[2] = posn_y;
        p->elts[3] = posn_z;
        try { obj->set_var_param_values(p); }
        catch (jwscxx::base::Arg_error e) 
        { std::cerr << e.get_msg() << std::endl; }

        objs_out.push_back(obj);
    }

    free_vector_d(p);
}


void render_sample
(
    jwsc::Image_f* sample_img, 
    const View*    view, 
    const Object*  obj
)
{
    using namespace jwsc;

    glEnable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);
    glClearColor(0, 0, 0, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    view->set_gl_projection();
    view->set_gl_modelview();

    //glColor3f(0.02, 0.02, 0.02);
    //glColor3f(0.015, 0.015, 0.015);
    glColor3f(0.01, 0.01, 0.01);
    //glColor3f(0.005, 0.005, 0.005);
    Render::solid_object(view, obj);

    glFlush();

    Image_f* img = 0;
    View::capture_gl_view(&img);

    for (uint32_t row = 0; row < img->num_rows; row++)
    {
        for (uint32_t col = 0; col < img->num_cols; col++)
        {
            sample_img->pxls[ row ][ col ].r -= img->pxls[ row ][ col ].r;
            sample_img->pxls[ row ][ col ].g -= img->pxls[ row ][ col ].g;
            sample_img->pxls[ row ][ col ].b -= img->pxls[ row ][ col ].b;
        }
    }
}


void render_samples(const View* view, const std::list<Object*>& objs)
{
    using namespace jwsc;

    Image_f* img = 0;
    list<Object*>::const_iterator it;

    glClearColor(1, 1, 1, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    View::capture_gl_view(&img);

    for (it = objs.begin(); it != objs.end(); it++)
    {
        render_sample(img, view, *it);
    }

    write_image_f(img, opts.cat_sample_img_out_fname);
    free_image_f(img);
}


/** @brief Main function for the @em furniture program. */
int main(int argc, const char** argv)
{
    using std::list;

    int rstatus = EXIT_SUCCESS;

    View* cat_view = 0;
    Object* cat_obj = 0;
    View* view = 0;
    list<Object*> objs;
    list<Object*>::iterator it;

    try
    {
        process_furniture_cat_options(argc, argv);

        Offscreen::enable();
        read_category(&cat_view, &cat_obj);
        cat_view->print_var_param_stats(0, std::cout);
        cat_obj->print_var_param_stats(0, std::cout);
        sample_category(&view, objs, cat_view, cat_obj);
        render_samples(view, objs);
        Offscreen::disable();
    }
    catch (jwscxx::base::Exception e)
    {
        cerr << "furniture-cat: ";
        e.print();
        rstatus = EXIT_FAILURE;
    }

    for (it = objs.begin(); it != objs.end(); it++)
    {
        delete *it;
    }
    delete view;
    delete cat_view;
    delete cat_obj;
    clean_up_furniture_cat();

    if (jwsc::get_num_unhandled_errors() > 0)
    {
        std::cerr << "furniture: Unhandled errors: " 
                  << jwsc::get_num_unhandled_errors();
        rstatus = EXIT_FAILURE;
    }

    return rstatus;
}
