/* 
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @mainpage Furniture: Learning 3D Object Models from Images
 *
 *
 * @section abstract Abstract
 *
 * We develop an approach to learn stochastic geometric models of object
 * categories from single view images. We focus here on models expressible as a
 * spatially contiguous assemblage of blocks. Model topologies are learned
 * across groups of images, and one or more such topologies is linked to an
 * object category (e.g.  chairs). Fitting learned topologies to an image can be
 * used to identify the object class, as well as detail its geometry. The latter
 * goes beyond labeling objects, as it provides the geometric structure of
 * particular instances.
 * 
 * We learn the models using joint statistical inference over structure
 * parameters, camera parameters, and instance parameters. These produce an
 * image likelihood through a statistical imaging model. For inference we use
 * trans-dimensional sampling to explore topology hypotheses, and alternate
 * between Metropolis-Hastings and stochastic dynamics to explore instance
 * parameters at an effective rate. During inference we maintain convergence to
 * the target distribution. In practice we are often able to find a good optimum
 * in the samples. 
 * 
 * We experiment with data sets consisting of images of standard furniture such
 * as tables and chairs. We demonstrate that we can learn models that are able
 * to infer category and geometry in this kind of data.
 *
 *
 * @section implement-notes Implementation notes
 *
 * Dealing with position parameter ranges is a tricky issue. We need to specify
 * limits on them to produce samples from a distribution. But we don't really
 * care if a position extends beyond the range. The solution dealing with this
 * implemented here is currently less than satisfactory: we abort when a
 * position parameter range is violated and expect the user to increase the
 * range size in the input configuration file. There must be a better solution,
 * just don't have time right now to worry about it.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Fits 3D furniture models to an image with a random view.
 *
 * Uses an MCMC sampler to fit the models.
 */


#include <config.h>

#include <iostream>
#include <cstdlib>

#include <inttypes.h>

#if defined FURNITURE_HAVE_OPENMPI
#include <mpi.h>
#endif

#include <jwsc/base/error.h>

#include <jwsc++/base/option.h>
#include <jwsc++/base/exception.h>

#include "sampler.h"
#include "onscreen.h"
#include "offscreen.h"
#include "options.h"


/** @brief Prints the usage of the program. */
void print_usage()
{
    std::cerr << "usage: furniture-learn OPTIONS <img-001 ... img-NNN | img-%%04d>\n";
    std::cerr << "where OPTIONS is one or more of the following:\n";
}


/** @brief Main function for the @em furniture program. */
int main(int argc, char** argv)
{
    int rstatus = EXIT_SUCCESS;

#if defined FURNITURE_HAVE_OPENMPI
    MPI_Init(&argc, &argv);
#endif

    Sampler* sampler = 0;
    Sampler::Parameters* params = 0;

    try
    {
        process_furniture_options(argc, (const char**) argv);

        if (opts.interactive)
        {
            Onscreen::enable(argc, (const char**)argv, "Furniture");
            create_sampler_and_params(&sampler, &params);
            Onscreen::start(sampler, params);
        }
        else
        {
            Offscreen::enable();
            create_sampler_and_params(&sampler, &params);
            sampler->run(params, opts.iterations);
            write_results(params);
            Offscreen::disable();
        }
    }
    catch (jwscxx::base::Exception e)
    {
        cerr << "furniture-learn: ";
        e.print();
        rstatus = EXIT_FAILURE;
    }

    delete sampler;
    delete params;

    clean_up_furniture();

    if (jwsc::get_num_unhandled_errors() > 0)
    {
        std::cerr << "furniture-learn: Unhandled errors: " 
                  << jwsc::get_num_unhandled_errors();
        rstatus = EXIT_FAILURE;
    }

#if defined FURNITURE_HAVE_OPENMPI
    MPI_Finalize();
#endif

    return rstatus;
}
