/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Functions to compute the likelihood of models given observed data.
 */


#include <config.h>

#include <cmath>
#include <iostream>
#include <iomanip>
#include <list>
#include <vector>
#include <algorithm>
#include <cmath>
#include <cassert>
#include <stdexcept>

#include <inttypes.h>
#include <pthread.h>

#if defined FURNITURE_HAVE_OPENGL
#include <GL/gl.h>
#include <GL/glu.h>
#elif defined FURNITURE_HAVE_OPENGL_FRAMEWORK
#include <OpenGL/OpenGL.h>
#endif

#include <jwsc/base/bits.h>
#include <jwsc/base/limits.h>
#include <jwsc/math/constants.h>
#include <jwsc/prob/pdf.h>
#include <jwsc/prob/pmf.h>
#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/matrix/matrix_math.h>
#include <jwsc/image/edge.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/graphics/polygon.h>

#include "view.h"
#include "object.h"
#include "likelihood.h"


/* ------------  Likelihood::Correspondence::Point::Compare  ---------------- */

bool Likelihood::Correspondence::Point::Compare::operator() 
(
    const Point* p1, 
    const Point* p2
) 
const 
{ 
    return p1->norm_dist < p2->norm_dist; 
}

/* -------------------------------------------------------------------------- */




/* ------------------  Likelihood::Correspondence::Point  ------------------- */

Likelihood::Correspondence::Point::Point
(
    const jwsc::Edge_point_f* e_pt, 
    uint32_t                  e_pt_num,
    uint32_t                  m_pt_row,
    uint32_t                  m_pt_col,
    uint32_t                  num_angles,
    double                    angle_sigma,
    double                    dist_sigma
)
: angles(num_angles, 0), gauss_angles(num_angles, 1),
  norm_dists(num_angles, 0), gauss_norm_dists(num_angles, 1)
{
    using namespace jwsc;

    this->e_pt       = e_pt; 
    this->e_pt_num   = e_pt_num;
    this->m_pt_row   = m_pt_row;
    this->m_pt_col   = m_pt_col;

    double v_x = (double)m_pt_col - (double)e_pt->col;
    double v_y = (double)m_pt_row - (double)e_pt->row;
    grad_dist = sqrt(v_x*v_x + v_y*v_y);

    if (grad_dist > 1.0e-16)
    {
        v_x /= grad_dist;
        v_y /= grad_dist;
    }

    assert(num_angles > 0);

    dtheta = JWSC_PI_2 / num_angles;
    double theta  = 0.5*dtheta;

    for (uint32_t i = 0; i < angles.size(); i++)
    {
        angles[ i ] = theta;

        double cos_theta = cos(theta);
        double sin_theta = sin(theta);
        double dcol = (double)e_pt->col - (double)m_pt_col;
        double drow = (double)e_pt->row - (double)m_pt_row;

        norm_dists[ i ] = fabs(
                (v_x*cos_theta - v_y*sin_theta)*dcol +
                (v_x*sin_theta + v_y*cos_theta)*drow);

        gauss_angles[ i ] = gaussian_pdf_d(0, angle_sigma, angles[ i ]);

        gauss_norm_dists[ i ] = gaussian_pdf_d(0, dist_sigma, norm_dists[ i ]);

        theta += dtheta;
    }
}

bool Likelihood::Correspondence::Point::operator< (const Point& p) const 
{ 
    return grad_dist < p.grad_dist; 
}

/* -------------------------------------------------------------------------- */






/* --------------------  Likelihood::Correspondence  ------------------------ */

/** @brief Mutex for OpenGL access. */
//static pthread_mutex_t gl_mux = PTHREAD_MUTEX_INITIALIZER;

Likelihood::Correspondence::Correspondence
(
    const jwsc::Edge_set_f* data_edges,
    uint32_t                num_rows,
    uint32_t                num_cols,
    uint32_t                num_angles,
    double                  angle_sigma,
    double                  dist_sigma,
    double                  max_dist
)
: m_pts_corr(num_rows, std::vector< std::list<Point> >(num_cols)),
  e_pts_corr(data_edges->total_num_pts)
{
    uint32_t i, j;
    uint32_t n = 0;

    this->max_dist = max_dist;

    for (i = 0; i < data_edges->num_edges; i++)
    {
        for (j = 0; j < data_edges->num_pts[ i ]; j++)
        {
            add_edge_pts_along_line(&(data_edges->pts[ i ][ j ]), n++, num_rows,
                    num_cols, num_angles, angle_sigma, dist_sigma, max_dist);
        }
    }

    for (i = 0; i < m_pts_corr.size(); i++)
    {
        for (j = 0; j < m_pts_corr[ i ].size(); j++)
        {
            m_pts_corr[ i ][ j ].sort();
        }
    }

    for (i = 0; i < e_pts_corr.size(); i++)
    {
        // A good beginning reserve value would be the number of projected 
        // model lines on average
        e_pts_corr[ i ].reserve(15);
    }
}


/**
 * The returned structure is indexed by edge point, then correspondence point
 * containing the model point info.
 */
const std::vector< std::vector<const Likelihood::Correspondence::Point*> >* 
Likelihood::Correspondence::generate_for_model
(
    View*                   view,
    Object*                 obj,
    const jwsc::Matrix_u32* model_mat
)
{
    using std::vector;
    using std::list;
    using jwscxx::graphics::Polygon_d;
    using namespace jwsc;

    assert(m_pts_corr.size() == (model_mat->num_rows));
    assert(m_pts_corr[0].size() == (model_mat->num_cols));

    for (uint32_t i = 0; i < e_pts_corr.size(); i++)
    {
        // Over time the vectors will have a nice reserved size.
        e_pts_corr[ i ].clear();
    }

    const vector<Polygon_d>& model_faces = obj->project();

    for (uint32_t i = 0; i < m_pts_corr.size(); i++)
    {
        for (uint32_t j = 0; j < m_pts_corr[ i ].size(); j++)
        {
            uint32_t id = model_mat->elts[ i ][ j ];

            if (id > 0 && !m_pts_corr[ i ][ j ].empty())
            {
                // Only use the closest edge point for now. But this is
                // potentially an area to experiment with.
                //
                // After a lot of experimentation, this restriction seems to
                // make sense; our model points can only generate one edge point
                // apiece. The most logical edge point it generated is the
                // closest one (in terms of grad_dist).
                Point* c_pt = &(m_pts_corr[ i ][ j ].front());
                //std::list<Point>::iterator it = m_pts_corr[ i ][ j ].begin();
                //for (; it != m_pts_corr[ i ][ j ].end(); it++)
                //{
                    //Point* c_pt = &(*it);

                // Not used here, in the likelihood function for edges only.
                //uint32_t silhouette_edge = id & 0x01;

                uint32_t edge_id = (id >> 1) & 0x03;
                assert(edge_id >= 0 && edge_id <= 3);

                uint32_t face_id = id >> 3;
                assert(face_id >= 1);

                const Polygon_d* face = &(model_faces[ face_id - 1 ]);
                const Vector_d*  p1   = face->get_point(edge_id);
                const Vector_d*  p2   = face->get_point((edge_id+1) % 4);

                // We need to somehow handle that the y coords are inverse for
                // image and model spaces. That's why we have a negative in the
                // y below.
                double dcol =   c_pt->e_pt->dcol;
                double drow =   c_pt->e_pt->drow;
                double x    = - (p1->elts[1] - p2->elts[1]);
                double y    = - (p1->elts[0] - p2->elts[0]);
                double m    = 1.0 / sqrt(x*x + y*y);
                double dp   = fabs(dcol*x*m + drow*y*m);

                if (dp >= 1.0) dp = 1.0 - 1.0e-16;

                uint32_t angle_index = (uint32_t)floor(acos(dp) / c_pt->dtheta);
                assert(angle_index < c_pt->angles.size());

                c_pt->angle           = c_pt->angles.at(angle_index);
                c_pt->gauss_angle     = c_pt->gauss_angles.at(angle_index);
                c_pt->norm_dist       = c_pt->norm_dists.at(angle_index);
                c_pt->gauss_norm_dist = c_pt->gauss_norm_dists.at(angle_index);

                e_pts_corr.at(c_pt->e_pt_num).push_back(c_pt);
                //}
            }
            else if (id > 0)
            {
                // There were no edge points whose gradient vector crosses paths
                // with this model point.
            }
        }
    }

    // Sort each correspondence point vector by distance.
    for (uint32_t i = 0; i < e_pts_corr.size(); i++)
    {
        std::sort(e_pts_corr[ i ].begin(), e_pts_corr[ i ].end(), 
                Point::Compare());
    }

    return &e_pts_corr;
}

double Likelihood::Correspondence::get_max_dist() const 
{ 
    return max_dist;
}

/** 
 * Uses Bresenham's line drawing algorithm to trace out the path in the matrix
 * along the gradent vector of the edge point.
 */
void Likelihood::Correspondence::add_edge_pts_along_line
(
    const jwsc::Edge_point_f* e_pt,
    uint32_t                  e_pt_num,
    uint32_t                  num_rows,
    uint32_t                  num_cols,
    uint32_t                  num_angles,
    double                    angle_sigma,
    double                    dist_sigma,
    double                    max_dist
)
{
    const static float sqrt_2 = 1.414213562373095f;

    uint32_t x_0 = e_pt->col;
    uint32_t y_0 = e_pt->row;

    m_pts_corr[ y_0 ][ x_0 ].push_back(Point(e_pt, e_pt_num, y_0,
                x_0, num_angles, angle_sigma, dist_sigma));

    float dx = e_pt->dcol;
    float dy = e_pt->drow;

    bool steep = fabs(dy) > fabs(dx);

    if (steep)
    {
        float dt = dy;
        dy = dx;
        dx = dt;

        uint32_t t_0 = y_0;
        y_0 = x_0;
        x_0 = t_0;
    }

    // assure dx >= 0 (left to right)
    if (dx < 0)
    {
        dx *= -1;
        dy *= -1;
    }

    // now we have positive dx (x-step) and neg or pos dy (y-step)

    float error = 0;
    float derr  = fabs(dy) / dx;

    float dist_r = 0;
    float dist_l = 0;

    // oh well, we'll never have too big anyway
    int32_t row;
    int32_t col;
    int32_t x = 0;
    int32_t y = 0;
    bool done_r = false;
    bool done_l = false;

    while (!done_r || !done_l)
    {
        x++;
        error += derr;
        if (error >= 0.5)
        {
            y = (dy > 0) ? y + 1 : y - 1;
            error -= 1.0;

            dist_r += sqrt_2;
            dist_l += sqrt_2;
        }
        else
        {
            dist_r += 1;
            dist_l += 1;
        }

        row = (steep) ? (x_0 + x) : (y_0 + y);
        col = (steep) ? (y_0 + y) : (x_0 + x);
        done_r = row < 0 || row >= (int32_t)num_rows || 
                 col < 0 || col >= (int32_t)num_cols || 
                 dist_r > max_dist;

        if (!done_r)
        {
            m_pts_corr[ row ][ col ].push_back(Point(e_pt, e_pt_num, row, 
                        col, num_angles, angle_sigma, dist_sigma));
        }

        row = (steep) ? (x_0 - x) : (y_0 - y);
        col = (steep) ? (y_0 - y) : (x_0 - x);
        done_l = row < 0 || row >= (int32_t)num_rows || 
                 col < 0 || col >= (int32_t)num_cols ||
                 dist_l > max_dist;

        if (!done_l)
        {
            m_pts_corr[ row ][ col ].push_back(Point(e_pt, e_pt_num, row, 
                        col, num_angles, angle_sigma, dist_sigma));
        }
    }
}

/* -------------------------------------------------------------------------- */






/* --------------------  Likelihood::Cache  --------------------------------- */

Likelihood::Cache::Cache(uint32_t max_size, uint32_t k, double t) 
{ 
    this->max_size = max_size; 
    this->size = 0; 
    this->pt_size = 0;
    this->k = k; 
    this->t = t; 
    num_hits = 0; 
    num_misses = 0; 
    q = 0; 
    q_1 = 0; 
    q_2 = 0; 
}

Likelihood::Cache::~Cache()
{ 
    free_vector_d(q);
    free_vector_d(q_1);
    free_vector_d(q_2);

    std::list<jwsc::Vector_d*>::iterator it;
    for (it = cache.begin(); it != cache.end(); it++)
    {
        jwsc::free_vector_d(*it);
    }
}

bool Likelihood::Cache::query
(
    double*       potential_out,
    const View*   view,
    const Object* obj
)
throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;
    using std::list;

    // Create the query point from view and object.
    uint32_t D_1 = view->num_var_params();
    uint32_t D_2 = obj->num_var_params();
    uint32_t D   = D_1 + D_2;

    if (size > 0 && pt_size != D)
    {
        throw jwscxx::base::Arg_error("Cache query point has invalid size");
    }

    create_vector_d(&q, D);

    view->get_var_param_values(&q_1);
    assert(copy_vector_section_into_vector_d(q, 0, q_1, 0, D_1) == 0);

    obj->get_var_param_values(&q_2);
    assert(copy_vector_section_into_vector_d(q, D_1, q_2, 0, D_2) == 0);

    // Get the k closest points in the cache.
    uint32_t k_closest_size = 0;
    k_closest.clear();

    std::list<Vector_d*>::iterator it;
    for (it = cache.begin(); it != cache.end(); it++)
    {
        Vector_d* p = *it;

        p->elts[D+1] = 0;
        for (uint32_t d = 0; d < D; d++)
        {
            p->elts[D+1] += (q->elts[d] - p->elts[d])*(q->elts[d] - p->elts[d]);
        }
        if (k_closest_size == 0)
        {
            k_closest.push_back(p);
            k_closest_size++;
        }
        else
        {
            std::list<Vector_d*>::iterator itt;
            for (itt = k_closest.begin(); itt != k_closest.end(); itt++)
            {
                Vector_d* pk = *itt;

                if (p->elts[D+1] < pk->elts[D+1])
                {
                    itt = k_closest.insert(itt, p);
                    if (k_closest_size == k)
                    {
                        k_closest.pop_back();
                    }
                    break;
                }
            }

            if (k_closest_size < k && itt == k_closest.end())
            {
                k_closest.push_back(p);
                k_closest_size++;
            }
        }
    }


    // Test if the query point q has a cache hit by the k closest values.
    bool found = false;

    if (k_closest.begin() != k_closest.end())
    {
        double dist = 0;;
        found = true;

        for (it = k_closest.begin(); it != k_closest.end(); it++)
        {
            dist += t - (*it)->elts[D+1];

            if ((*it)->elts[D+1] > t)
            {
                found = false;
                break;
            }
        }

        double potential = 0;

        if (found)
        {
            num_hits++;

            if (dist < 1.0e-12)
            {
                potential = k_closest.front()->elts[D];
            }
            else
            {
                // This is an average in log probability space, probably not the
                // best idea.
                for (it = k_closest.begin(); it != k_closest.end(); it++)
                {
                    double w = (t - (*it)->elts[D+1]) / dist;
                    potential += w * (*it)->elts[D];
                }
            }

            //fprintf(stderr, "cache hit: %e\n", potential);

            *potential_out = potential;
        }
        else
        {
            num_misses++;
        }
    }
    else
    {
        num_misses++;
    }

    return found;
}


void Likelihood::Cache::insert
(
    double                      potential,
    const View*                 view,
    const Object*               obj
)
throw (jwscxx::base::Arg_error)
{
    uint32_t D_1 = view->num_var_params();
    uint32_t D_2 = obj->num_var_params();
    uint32_t D   = D_1 + D_2;

    jwsc::Vector_d* p = 0;

    if (max_size == 0)
        return;

    if (size == 0)
    {
        pt_size = D;
    }
    else if (pt_size != D)
    {
        throw jwscxx::base::Arg_error(
                "Inserting invalid sized point into cache");
    }

    if (size == max_size)
    {
        p = cache.back();
        cache.pop_back();
    }
    else
    {
        jwsc::create_vector_d(&p, D+2);
        size++;
    }

    view->get_var_param_values(&q_1);
    assert(copy_vector_section_into_vector_d(p, 0, q_1, 0, D_1) == 0);

    obj->get_var_param_values(&q_2);
    assert(copy_vector_section_into_vector_d(p, D_1, q_2, 0, D_2) == 0);

    p->elts[D]   = potential;
    p->elts[D+1] = 0;

    cache.push_front(p);
}


void Likelihood::Cache::clear()
{ 
    std::list<jwsc::Vector_d*>::iterator it;
    for (it = cache.begin(); it != cache.end(); it++)
    {
        jwsc::free_vector_d(*it);
    }

    cache.clear();
    k_closest.clear();

    size = 0; 
    pt_size = 0;
    num_hits = 0; 
    num_misses = 0; 
}

uint32_t Likelihood::Cache::get_num_hits() const 
{
    return num_hits;
}

uint32_t Likelihood::Cache::get_num_misses() const 
{
    return num_misses;
}

uint32_t Likelihood::Cache::get_cache_size() const 
{
    return size;
}

uint32_t Likelihood::Cache::get_cache_pt_size() const 
{
    return pt_size;
}

float Likelihood::Cache::get_percent_hits() const 
{ 
    return (num_hits+num_misses > 0) 
        ? (float)num_hits / (num_hits+num_misses) : 0; 
}

/* -------------------------------------------------------------------------- */





/* ---------------------------  Likelihood  --------------------------------- */

/**
 * @param  model_mat_out  Matrix containing the rendered model scene. 
 * @param  view           View to render under.
 * @param  obj            Object model to render.
 */
void Likelihood::render_wire_model_scene
(
    jwsc::Matrix_u32** model_mat_out, 
    const View*        view, 
    const Object*      obj
)
{
    using namespace jwsc;

    glClearColor(0, 0, 0, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    view->set_gl_projection();
    view->set_gl_modelview();

    obj->wire_occlude_render(); 
    obj->wire_render_with_ids(view); 

    glFlush();

    uint32_t num_cols = (uint32_t)view->get_gl_viewport_width();
    uint32_t num_rows = (uint32_t)view->get_gl_viewport_height();
    uint32_t row, col;
    uint32_t bot, top;

    create_matrix_u32(model_mat_out, num_rows, num_cols);
    Matrix_u32* model_mat = *model_mat_out;

    glReadPixels(0, 0, num_cols, num_rows, GL_RGBA, GL_UNSIGNED_BYTE, 
            *(model_mat->elts));

    // Swap rows (and bytes if little endian).
    for (row = 0; row < num_rows/2; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            top = model_mat->elts[ row ][ col ];
            bot = model_mat->elts[ num_rows-row-1 ][ col ];

            if (top > 0 || bot > 0)
            {
#ifndef JWSC_BIGENDIAN
                bswap_u32(&(top));
                bswap_u32(&(bot));
#endif
                model_mat->elts[ row ][ col ] = bot;
                model_mat->elts[ num_rows-row-1 ][ col ] = top;
            }
        }
    }

#ifndef JWSC_BIGENDIAN
    if (num_rows % 2 > 0)
    {
        row = num_rows/2;
        for (col = 0; col < num_cols; col++)
        {
            if (model_mat->elts[ row ][ col ] > 0)
            {
                bswap_u32(&(model_mat->elts[ row ][ col ]));
            }
        }
    }
#endif
}

/**
 * @param  model_mat_out  Matrix containing the rendered model scene. 
 * @param  view           View to render under.
 * @param  obj            Object model to render.
 */
void Likelihood::render_solid_model_scene
(
    jwsc::Matrix_u32** model_mat_out, 
    const View*        view, 
    const Object*      obj
)
{
    using namespace jwsc;

    glClearColor(0, 0, 0, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    view->set_gl_projection();
    view->set_gl_modelview();

    obj->solid_render_with_ids(view); 

    glFlush();

    uint32_t num_cols = (uint32_t)view->get_gl_viewport_width();
    uint32_t num_rows = (uint32_t)view->get_gl_viewport_height();
    uint32_t row, col;
    uint32_t bot, top;

    create_matrix_u32(model_mat_out, num_rows, num_cols);
    Matrix_u32* model_mat = *model_mat_out;

    glReadPixels(0, 0, num_cols, num_rows, GL_RGBA, GL_UNSIGNED_BYTE, 
            *(model_mat->elts));

    // Swap rows (and bytes if little endian).
    for (row = 0; row < num_rows/2; row++)
    {
        for (col = 0; col < num_cols; col++)
        {
            top = model_mat->elts[ row ][ col ];
            bot = model_mat->elts[ num_rows-row-1 ][ col ];

            if (top > 0 || bot > 0)
            {
#ifndef JWSC_BIGENDIAN
                bswap_u32(&(top));
                bswap_u32(&(bot));
#endif
                model_mat->elts[ row ][ col ] = bot;
                model_mat->elts[ num_rows-row-1 ][ col ] = top;
            }
        }
    }

#ifndef JWSC_BIGENDIAN
    if (num_rows % 2 > 0)
    {
        row = num_rows/2;
        for (col = 0; col < num_cols; col++)
        {
            if (model_mat->elts[ row ][ col ] > 0)
            {
                bswap_u32(&(model_mat->elts[ row ][ col ]));
            }
        }
    }
#endif
}

double Likelihood::clipped_point_penalty
(
    const View*           view, 
    const jwsc::Vector_d* p
)
{
    using namespace jwsc;

    GLdouble glM[16] = {0};

    glGetDoublev(GL_MODELVIEW_MATRIX, glM);

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glMultMatrixd(glM);
    glGetDoublev(GL_PROJECTION_MATRIX, glM);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);

    Matrix_d* M = 0;
    create_matrix_d(&M, 4, 4);
    for (size_t row = 0; row < 4; row++)
    {
        for (size_t col = 0; col < 4; col++)
        {
            M->elts[ row ][ col ] = glM[ col*4 + row  ];
        }
    }

    double width  = view->get_gl_viewport_width();
    double height = view->get_gl_viewport_height();

    Vector_d* uv = 0;
    multiply_matrix_and_vector_d(&uv, M, p);

    double w = 1.0 / uv->elts[ 3 ];
    double u = 0.5*width* (uv->elts[ 0 ]*w + 1.0);
    double v = 0.5*height*(uv->elts[ 1 ]*w + 1.0);
    free_matrix_d(M);
    free_vector_d(uv);

    double e_u = 0;
    double e_v = 0;
    if (u < 0 || u > width)
    {
        e_u = (u < 0) ? fabs(u) : u - width;
    }
    if (v < 0 || v > height)
    {
        e_v = (v < 0) ? fabs(v) : v - height;
    }
    double penalty = 0;
    penalty += e_u*10*log(JWSC_MIN_LOG_ARG);
    penalty += e_v*10*log(JWSC_MIN_LOG_ARG);

    return penalty;
}

double Likelihood::clipped_object_penalty
(
    const View*   view, 
    const Object* obj
)
{
    using namespace jwsc;

    Vector_d* p = 0;
    double penalty = 0;

    for (uint32_t i = 0; i < obj->num_components(); i++)
    {
        Component* c = obj->component(i);
        for (uint32_t j = 0; j < c->num_ctr_attach_pts(); j++)
        {
            c->ctr_attach_pt_v(&p, j);
            penalty += clipped_point_penalty(view, p);
        }

        Block_set* bs;
        if ((bs = dynamic_cast<Block_set*>(c)))
        {
            for (uint32_t k = 0; k < bs->num_inner_blocks(); k++)
            {
                for (uint32_t jj = 0; jj < bs->num_ctr_attach_pts(); jj++)
                {
                    bs->inner_block_ctr_attach_pt_v(&p, k, jj);
                    penalty += clipped_point_penalty(view, p);
                }
            }
        }
    }

    free_vector_d(p);

    return penalty;
}

/** 
 * @param  image_ll_out     Image log likelihood.
 * @param  wire_model_mat_out  Rendered wire-frame model scene.
 * @param  surf_ll_assign_out  Surface model assignment type for each pixel.
 * @param  edge_ll_assign_out  Edge model assignment type for each pixel.
 * @param  image_t          Thresholded input image.
 * @param  edges            Detected input image edges.
 * @param  corr             Data to model point correspondence.
 * @param  view             View model to use.
 * @param  obj              Object model to use.
 * @param  bg_prob          Background pixel probability.
 * @param  surf_prob        Surface point pixel probability.
 * @param  surf_noise_prob  Noisy surface point pixel probability.
 * @param  surf_miss_prob   Missing surrface point pixel probabiliity.
 * @param  edge_noise_prob  Noisy edge point pixel probability.
 * @param  edge_silho_miss_prob  Missing edge point pixel probabiliity.
 * @param  edge_inner_miss_prob  Missing edge point pixel probabiliity.
 *
 * The log likelihood pixel assignment matrix for the surface point detections
 * (@em *surf_ll_assign_out) has the following interpretation:
 *
 * @code
 *    0 - Background
 *    1 - Surface
 *    2 - Noisy surface
 *    3 - Missing surface
 * @endcode
 *
 * The log likelihood pixel assignment matrix for the edge detections (@em
 * *edge_ll_assign_out) has the following interpretation:
 *
 * @code
 *    0 - Background
 *    1 - Edge
 *    2 - Noisy edge
 *    3 - Missing silhouette edge
 *    4 - Missing inner edge
 * @endcode
 */
void Likelihood::calc_image_log_likelihood
(
    double*                 image_ll_out,
    jwsc::Matrix_u32**      wire_model_mat_out,
    jwsc::Matrix_u32**      surf_ll_assign_out,
    jwsc::Matrix_u32**      edge_ll_assign_out,
    const jwsc::Matrix_f*   image_t,
    const jwsc::Edge_set_f* edges,
    Correspondence*         corr,
    View*                   view,
    Object*                 obj,
    double                  bg_prob,
    double                  surf_prob,
    double                  surf_noise_prob,
    double                  surf_miss_prob,
    double                  edge_noise_prob,
    double                  edge_silho_miss_prob,
    double                  edge_inner_miss_prob
) 
throw (jwscxx::base::Arg_error)
{
    using std::vector;
    using namespace jwsc;

    if (bg_prob < JWSC_MIN_LOG_ARG) bg_prob = JWSC_MIN_LOG_ARG;
    if (surf_prob < JWSC_MIN_LOG_ARG) surf_prob = JWSC_MIN_LOG_ARG;
    if (surf_noise_prob < JWSC_MIN_LOG_ARG) surf_noise_prob = JWSC_MIN_LOG_ARG;
    if (edge_noise_prob < JWSC_MIN_LOG_ARG) edge_noise_prob = JWSC_MIN_LOG_ARG;
    if (surf_miss_prob < JWSC_MIN_LOG_ARG) surf_miss_prob = JWSC_MIN_LOG_ARG;
    if (edge_silho_miss_prob < JWSC_MIN_LOG_ARG) edge_silho_miss_prob 
        = JWSC_MIN_LOG_ARG;
    if (edge_inner_miss_prob < JWSC_MIN_LOG_ARG) edge_inner_miss_prob 
        = JWSC_MIN_LOG_ARG;

    render_wire_model_scene(wire_model_mat_out, view, obj);
    const Matrix_u32* wire_model_mat = *wire_model_mat_out;

    create_zero_matrix_u32(edge_ll_assign_out, wire_model_mat->num_rows,
            wire_model_mat->num_cols);
    Matrix_u32* edge_ll_assign = *edge_ll_assign_out;

    Matrix_u32* solid_model_mat = 0;
    render_solid_model_scene(&solid_model_mat, view, obj);

    create_zero_matrix_u32(surf_ll_assign_out, solid_model_mat->num_rows,
            solid_model_mat->num_cols);
    Matrix_u32* surf_ll_assign = *surf_ll_assign_out;

    assert(image_t->num_rows == solid_model_mat->num_rows);
    assert(image_t->num_cols == solid_model_mat->num_cols);

    const vector< vector<const Correspondence::Point*> >* corr_pts;
    corr_pts = corr->generate_for_model(view, obj, wire_model_mat);

    double ll = 0;

    uint8_t silhouette;

    uint32_t num_edge_bg_pts = wire_model_mat->num_rows * 
                               wire_model_mat->num_cols;
    uint32_t num_edge_pts = 0;
    uint32_t num_edge_noise_pts = 0;
    uint32_t num_edge_miss_pts = 0;
    uint32_t num_edge_silho_miss_pts = 0;
    uint32_t num_edge_inner_miss_pts = 0;

    for (uint32_t i = 0; i < corr_pts->size(); i++)
    {
        const Edge_point_f* e_pt = &((*(edges->pts))[ i ]);

        if ((*corr_pts)[ i ].size() > 0)
        {
            const Correspondence::Point* c_pt = (*corr_pts)[ i ][0];
            assert(c_pt->e_pt->row == e_pt->row);
            assert(c_pt->e_pt->col == e_pt->col);

            uint32_t k = (*corr_pts)[ i ].size();
            double   s = 0;

            uint32_t closest_c_pt = 0;
            double min_dist = (*corr_pts)[ i ][ 0 ]->norm_dist;

            for (uint32_t j = 0; j < k; j++)
            {
                c_pt = (*corr_pts)[ i ][ j ];

                s += (1.0/k) * c_pt->gauss_angle * c_pt->gauss_norm_dist;

                if (c_pt->norm_dist < min_dist)
                {
                    closest_c_pt = j;
                    min_dist = c_pt->norm_dist;
                }
            }

            ll += log((s > JWSC_MIN_LOG_ARG) ? s : JWSC_MIN_LOG_ARG);

            if (k > 1)
            {
                num_edge_miss_pts += k-1;

                for (uint32_t j = 0; j < k; j++)
                {
                    if (j == closest_c_pt) 
                        continue;

                    c_pt = (*corr_pts)[ i ][ j ];

                    uint32_t row = c_pt->m_pt_row;
                    uint32_t col = c_pt->m_pt_col;

                    silhouette = wire_model_mat->elts[ row ][ col ] & 0x01;
                    if (silhouette)
                    {
                        num_edge_silho_miss_pts++;
                        edge_ll_assign->elts[ row ][ col ] = 3;
                    }
                    else
                    {
                        num_edge_inner_miss_pts++;
                        edge_ll_assign->elts[ row ][ col ] = 4;
                    }
                }
            }

            for (uint32_t j = 0; j < k; j++)
            {
                c_pt = (*corr_pts)[ i ][ j ];
                wire_model_mat->elts[ c_pt->m_pt_row ][ c_pt->m_pt_col ] = 0;
            }

            num_edge_pts++;
            edge_ll_assign->elts[ e_pt->row ][ e_pt->col ] = 1;
        }
        else
        {
            num_edge_noise_pts++;
            edge_ll_assign->elts[ e_pt->row ][ e_pt->col ] = 2;
        }
    }

    for (uint32_t row = 0; row < wire_model_mat->num_rows; row++)
    {
        for (uint32_t col = 0; col < wire_model_mat->num_cols; col++)
        {
            if (wire_model_mat->elts[ row ][ col ] > 0)
            {
                num_edge_miss_pts++;

                silhouette = wire_model_mat->elts[ row ][ col ] & 0x01;
                if (silhouette)
                {
                    num_edge_silho_miss_pts++;
                    edge_ll_assign->elts[ row ][ col ] = 3;
                }
                else
                {
                    num_edge_inner_miss_pts++;
                    edge_ll_assign->elts[ row ][ col ] = 4;
                }
            } 
        }
    }

    num_edge_bg_pts -= num_edge_pts + num_edge_noise_pts + num_edge_miss_pts;

    uint32_t N = num_edge_bg_pts + num_edge_pts + num_edge_miss_pts + 
        num_edge_noise_pts;
    assert(N == wire_model_mat->num_rows*wire_model_mat->num_cols);
    assert(num_edge_miss_pts == 
            (num_edge_silho_miss_pts + num_edge_inner_miss_pts));

    ll += num_edge_bg_pts * log(bg_prob);
    ll += num_edge_noise_pts * log(edge_noise_prob);
    ll += num_edge_silho_miss_pts * log(edge_silho_miss_prob);
    ll += num_edge_inner_miss_pts * log(edge_inner_miss_prob);


    uint32_t num_surf_bg_pts = 0;
    uint32_t num_surf_pts = 0;
    uint32_t num_surf_noise_pts = 0;
    uint32_t num_surf_miss_pts = 0;


    for (uint32_t row = 0; row < solid_model_mat->num_rows; row++)
    {
        for (uint32_t col = 0; col < solid_model_mat->num_cols; col++)
        {
            bool image_t_bg = (image_t->elts[ row ][ col ] > 0);

            if (!image_t_bg && solid_model_mat->elts[ row ][ col ] > 0)
            {
                num_surf_pts++;
                surf_ll_assign->elts[ row ][ col ] = 1;
            }
            else if (image_t_bg && solid_model_mat->elts[ row ][ col ] > 0)
            {
                num_surf_miss_pts++;
                surf_ll_assign->elts[ row ][ col ] = 3;
            }
            else if (!image_t_bg && solid_model_mat->elts[ row ][ col ] == 0)
            {
                num_surf_noise_pts++;
                surf_ll_assign->elts[ row ][ col ] = 2;
            }
            else
            {
                num_surf_bg_pts++;
                surf_ll_assign->elts[ row ][ col ] = 0;
            }
        }
    }

    N = num_surf_bg_pts + num_surf_pts + num_surf_miss_pts + 
        num_surf_noise_pts;
    assert(N == solid_model_mat->num_rows*solid_model_mat->num_cols);

    ll += num_surf_bg_pts * log(bg_prob);
    ll += num_surf_pts * log(surf_prob);
    ll += num_surf_noise_pts * log(surf_noise_prob);
    ll += num_surf_miss_pts * log(surf_miss_prob);

    ll += clipped_object_penalty(view, obj);

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(ll) && !std::isnan(ll));
#endif

    *image_ll_out = ll;

    free_matrix_u32(solid_model_mat);
}

/** 
 * This is a potential energy function.
 *
 * @param  potential_out    Negative log likelihood of the image and models.
 * @param  image_ll_out     Image log likelihood.
 * @param  model_mat_out    Rendered model scene.
 * @param  surf_ll_assign_out  Surface model assignment type for each pixel.
 * @param  edge_ll_assign_out  Edge model assignment type for each pixel.
 * @param  image_t          Thresholded input image.
 * @param  edges            Detected input image edges.
 * @param  corr             Data to model point correspondence.
 * @param  view             View model to use.
 * @param  obj              Object model to use.
 * @param  bg_prob          Background pixel probability.
 * @param  surf_prob        Surface point pixel probability.
 * @param  surf_noise_prob  Noisy surface point pixel probability.
 * @param  surf_miss_prob   Missing surrface point pixel probabiliity.
 * @param  edge_noise_prob  Noisy edge point pixel probability.
 * @param  edge_silho_miss_prob  Missing edge point pixel probabiliity.
 * @param  edge_inner_miss_prob  Missing edge point pixel probabiliity.
 */
void Likelihood::calc_neg_log_likelihood
(
    double*                 potential_out,
    double*                 image_ll_out,
    jwsc::Matrix_u32**      model_mat_out,
    jwsc::Matrix_u32**      surf_ll_assign_out,
    jwsc::Matrix_u32**      edge_ll_assign_out,
    const jwsc::Matrix_f*   image_t,
    const jwsc::Edge_set_f* edges,
    Correspondence*         corr,
    View*                   view,
    Object*                 obj,
    double                  bg_prob,
    double                  surf_prob,
    double                  surf_noise_prob,
    double                  surf_miss_prob,
    double                  edge_noise_prob,
    double                  edge_silho_miss_prob,
    double                  edge_inner_miss_prob
) 
throw (jwscxx::base::Arg_error)
{
    calc_image_log_likelihood(image_ll_out, model_mat_out, surf_ll_assign_out,
            edge_ll_assign_out, image_t, edges, corr, view, obj, bg_prob,
            surf_prob, surf_noise_prob, surf_miss_prob, edge_noise_prob,
            edge_silho_miss_prob, edge_inner_miss_prob);

    double view_ll = view->calc_var_param_log_likelihood();
    double obj_ll  = obj->calc_var_param_log_likelihood();

    *potential_out = - (*image_ll_out + view_ll + obj_ll);
}

/** 
 * This is a potential energy function.
 *
 * @param  potential_out    Negative log likelihood and prior.
 * @param  image_ll_out     Image log likelihood.
 * @param  model_mat_out    Rendered model scene.
 * @param  surf_ll_assign_out  Surface model assignment type for each pixel.
 * @param  edge_ll_assign_out  Edge model assignment type for each pixel.
 * @param  image_t          Thresholded input image.
 * @param  edges            Detected input image edges.
 * @param  corr             Data to model point correspondence.
 * @param  view             View model to use.
 * @param  obj              Object model to use.
 * @param  bg_prob          Background pixel probability.
 * @param  surf_prob        Surface point pixel probability.
 * @param  surf_noise_prob  Noisy surface point pixel probability.
 * @param  surf_miss_prob   Missing surrface point pixel probabiliity.
 * @param  edge_noise_prob  Noisy edge point pixel probability.
 * @param  edge_silho_miss_prob  Missing edge point pixel probabiliity.
 * @param  edge_inner_miss_prob  Missing edge point pixel probabiliity.
 */
void Likelihood::calc_neg_log_likelihood_and_prior
(
    double*                 potential_out,
    double*                 image_ll_out,
    jwsc::Matrix_u32**      model_mat_out,
    jwsc::Matrix_u32**      surf_ll_assign_out,
    jwsc::Matrix_u32**      edge_ll_assign_out,
    const jwsc::Matrix_f*   image_t,
    const jwsc::Edge_set_f* edges,
    Correspondence*         corr,
    View*                   view,
    Object*                 obj,
    double                  bg_prob,
    double                  surf_prob,
    double                  surf_noise_prob,
    double                  surf_miss_prob,
    double                  edge_noise_prob,
    double                  edge_silho_miss_prob,
    double                  edge_inner_miss_prob
) 
throw (jwscxx::base::Arg_error)
{
    calc_neg_log_likelihood(potential_out, image_ll_out, model_mat_out,
            surf_ll_assign_out, edge_ll_assign_out, image_t, edges, corr, view,
            obj, bg_prob, surf_prob, surf_noise_prob, surf_miss_prob,
            edge_noise_prob, edge_silho_miss_prob, edge_inner_miss_prob);

    double view_lp = view->calc_var_param_prior_log_prob();
    double obj_lp  = obj->calc_var_param_prior_log_prob();

    *potential_out = *potential_out - view_lp - obj_lp;
}

/** 
 * This is essentially the gradient of a potential energy function.
 *
 * The parameter @em potential is the potential energy of one set of instance
 * parameters (@em view and @em obj) and the category parameters.
 */
void Likelihood::calc_grad_neg_log_likelihood_and_prior_wrt_ins_params
(
    jwsc::Vector_d**        grad_out,
    jwsc::Matrix_u32**      model_mat_out,
    jwsc::Matrix_u32**      surf_ll_assign_out,
    jwsc::Matrix_u32**      edge_ll_assign_out,
    double                  potential,
    const jwsc::Matrix_f*   image_t,
    const jwsc::Edge_set_f* edges,
    Correspondence*         corr,
    Cache*                  cache,
    View*                   view,
    Object*                 obj,
    double                  bg_prob,
    double                  surf_prob,
    double                  surf_noise_prob,
    double                  surf_miss_prob,
    double                  edge_noise_prob,
    double                  edge_silho_miss_prob,
    double                  edge_inner_miss_prob,
    bool                    two_point_estimate
) 
throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    uint32_t D_1 = view->num_var_params();
    uint32_t D_2 = obj->num_var_params();
    uint32_t D   = D_1 + D_2;

    create_vector_d(grad_out, D);
    Vector_d* grad = *grad_out;

    View   d_view(*view);
    Object d_obj(*obj);

    uint32_t i = 0;
    double   ll_1, ll_2;
    double   E_1, E_2;
    bool     cache_hit;

    Vector_d* p   = 0;
    Vector_d* eta = 0;

    view->get_var_param_values_dynamics_grad_eta(&eta);

    double E = potential;

    for (uint32_t d = 0; d < D_1; d++)
    {
        view->get_var_param_values(&p);

        d_view = *view;
        p->elts[ d ] += eta->elts[ d ];

        try { d_view.set_var_param_values(p); }
        catch (jwscxx::base::Arg_error e) { }

        try {
            if (!(cache_hit = cache->query(&E_1, &d_view, obj)))
            {
                calc_neg_log_likelihood_and_prior(&E_1, &ll_1, model_mat_out,
                        surf_ll_assign_out, edge_ll_assign_out, image_t, edges,
                        corr, &d_view, obj, bg_prob, surf_prob, surf_noise_prob,
                        surf_miss_prob, edge_noise_prob, edge_silho_miss_prob,
                        edge_inner_miss_prob);
                cache->insert(E_1, &d_view, obj);
            }
        } catch (jwscxx::base::Arg_error e) { abort(); }

        if (two_point_estimate) 
        {
            d_view = *view;
            p->elts[ d ] -= 2 * eta->elts[ d ];

            try { d_view.set_var_param_values(p); }
            catch (jwscxx::base::Arg_error e) { }

            try {
                if (!(cache_hit = cache->query(&E_2, &d_view, obj)))
                {
                    calc_neg_log_likelihood_and_prior(&E_2, &ll_2,
                            model_mat_out, surf_ll_assign_out,
                            edge_ll_assign_out, image_t, edges, corr, &d_view,
                            obj, bg_prob, surf_prob, surf_noise_prob,
                            surf_miss_prob, edge_noise_prob, 
                            edge_silho_miss_prob, edge_inner_miss_prob);
                    cache->insert(E_2, &d_view, obj);
                }
            } catch (jwscxx::base::Arg_error e) { abort(); }

            grad->elts[ i ] = (E_1 - E_2) / (2 * eta->elts[ d ]);
        } 
        else 
        {
            grad->elts[ i ] = (E_1 - E) / eta->elts[ d ];
        }

        i++;
    }

    obj->get_var_param_values_dynamics_grad_eta(&eta);

    for (uint32_t d = 0; d < D_2; d++)
    {
        obj->get_var_param_values(&p);

        d_obj = *obj;
        p->elts[ d ] += eta->elts[ d ];

        try { d_obj.set_var_param_values(p); }
        catch (jwscxx::base::Arg_error e) { }

        try {
            if (!(cache_hit = cache->query(&E_1, view, &d_obj)))
            {
                calc_neg_log_likelihood_and_prior(&E_1, &ll_1, model_mat_out,
                        surf_ll_assign_out, edge_ll_assign_out, image_t, edges,
                        corr, view, &d_obj, bg_prob, surf_prob, surf_noise_prob,
                        surf_miss_prob, edge_noise_prob, edge_silho_miss_prob,
                        edge_inner_miss_prob);
                cache->insert(E_1, view, &d_obj);
            }
        } catch (jwscxx::base::Arg_error e) { abort(); }

        if (two_point_estimate) 
        {
            d_obj = *obj;
            p->elts[ d ] -= 2 * eta->elts[ d ];

            try { d_obj.set_var_param_values(p); }
            catch (jwscxx::base::Arg_error e) { }

            try {
                if (!(cache_hit = cache->query(&E_2, view, &d_obj)))
                {
                    calc_neg_log_likelihood_and_prior(&E_2, &ll_2,
                            model_mat_out, surf_ll_assign_out,
                            edge_ll_assign_out, image_t, edges, corr, view,
                            &d_obj, bg_prob, surf_prob, surf_noise_prob,
                            surf_miss_prob, edge_noise_prob,
                            edge_silho_miss_prob, edge_inner_miss_prob);
                    cache->insert(E_2, view, &d_obj);
                }
            } catch (jwscxx::base::Arg_error e) { abort(); }

            grad->elts[ i ] = (E_1 - E_2) / (2 * eta->elts[ d ]);
        } 
        else 
        {
            grad->elts[ i ] = (E_1 - E) / eta->elts[ d ];
        }

        i++;
    }

    assert(grad->num_elts == i);

    free_vector_d(p);
    free_vector_d(eta);
}

void Likelihood::calc_grad_neg_log_likelihood_and_prior_wrt_cat_params
(
    jwsc::Vector_d**                  grad_out,
    double                            lp_potential,
    const View*                       cat_view,
    const Object*                     cat_obj,
    const std::vector<const View*>&   ins_views,
    const std::vector<const Object*>& ins_objs,
    bool                              two_point_estimate
) 
throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    assert(ins_views.size() == ins_objs.size());

    uint32_t D_1 = cat_view->num_var_params();
    uint32_t D_2 = cat_obj->num_var_params();
    uint32_t D   = D_1 + D_2;
    uint32_t N_i = ins_views.size();

    create_vector_d(grad_out, 2*D);
    Vector_d* grad = *grad_out;

    View   d_view(*cat_view);
    Object d_obj(*cat_obj);

    uint32_t g = 0;
    double   view_lp, view_ll;
    double   obj_lp, obj_ll;
    double   E_1, E_2;
    double   E = lp_potential;

    Vector_d* p[2]   = {0};
    Vector_d* eta[2] = {0};

    cat_view->get_var_param_stats_dynamics_grad_eta(&(eta[0]), &(eta[1]));

    for (uint32_t s = 0; s < 2; s++)
    {
        for (uint32_t d = 0; d < D_1; d++)
        {
            cat_view->get_var_param_stats(&(p[0]), &(p[1]));

            d_view = *cat_view;
            p[ s ]->elts[ d ] += eta[ s ]->elts[ d ];

            try { d_view.set_var_param_stats(p[0], p[1]); }
            catch (jwscxx::base::Arg_error e) { }

            view_lp = d_view.calc_var_param_prior_log_prob();
            obj_lp  = cat_obj->calc_var_param_prior_log_prob();
            E_1     = - (view_lp + obj_lp);

            for (uint32_t i = 0; i < N_i; i++)
            {
                d_view = *(ins_views[ i ]);

                try { d_view.set_var_param_stats(p[0], p[1]); }
                catch (jwscxx::base::Arg_error e) { }

                view_ll = d_view.calc_var_param_log_likelihood();
                obj_ll  = ins_objs[ i ]->calc_var_param_log_likelihood();
                E_1    += - (view_ll + obj_ll);
            }

            if (two_point_estimate) 
            {
                d_view = *cat_view;
                p[ s ]->elts[ d ] -= 2 * eta[ s ]->elts[ d ];

                try { d_view.set_var_param_stats(p[0], p[1]); }
                catch (jwscxx::base::Arg_error e) { }

                view_lp = d_view.calc_var_param_prior_log_prob();
                obj_lp  = cat_obj->calc_var_param_prior_log_prob();
                E_2     = - (view_lp + obj_lp);

                for (uint32_t i = 0; i < N_i; i++)
                {
                    d_view = *(ins_views[ i ]);

                    try { d_view.set_var_param_stats(p[0], p[1]); }
                    catch (jwscxx::base::Arg_error e) { }

                    view_ll = d_view.calc_var_param_log_likelihood();
                    obj_ll  = ins_objs[ i ]->calc_var_param_log_likelihood();
                    E_2    += - (view_ll + obj_ll);
                }

                grad->elts[ g ] = (E_1 - E_2) / (2 * eta[ s ]->elts[ d ]);
            } 
            else 
            {
                grad->elts[ g ] = (E_1 - E) / eta[ s ]->elts[ d ];
            }

            g++;
        }
    }

    cat_obj->get_var_param_stats_dynamics_grad_eta(&(eta[0]), &(eta[1]));

    for (uint32_t s = 0; s < 2; s++)
    {
        for (uint32_t d = 0; d < D_2; d++)
        {
            cat_obj->get_var_param_stats(&(p[0]), &(p[1]));

            d_obj = *cat_obj;
            p[ s ]->elts[ d ] += eta[ s ]->elts[ d ];

            try { d_obj.set_var_param_stats(p[0], p[1]); }
            catch (jwscxx::base::Arg_error e) { }

            view_lp = cat_view->calc_var_param_prior_log_prob();
            obj_lp  = d_obj.calc_var_param_prior_log_prob();
            E_1     = - (view_lp + obj_lp);

            for (uint32_t i = 0; i < N_i; i++)
            {
                d_obj = *(ins_objs[ i ]);

                try { d_obj.set_var_param_stats(p[0], p[1]); }
                catch (jwscxx::base::Arg_error e) { }

                view_ll = ins_views[ i ]->calc_var_param_log_likelihood();
                obj_ll  = d_obj.calc_var_param_log_likelihood();
                E_1    += - (view_ll + obj_ll);
            }

            if (two_point_estimate) 
            {
                d_obj = *cat_obj;
                p[ s ]->elts[ d ] -= 2 * eta[ s ]->elts[ d ];

                try { d_obj.set_var_param_stats(p[0], p[1]); }
                catch (jwscxx::base::Arg_error e) { }

                view_lp = cat_view->calc_var_param_prior_log_prob();
                obj_lp  = d_obj.calc_var_param_prior_log_prob();
                E_2     = - (view_lp + obj_lp);

                for (uint32_t i = 0; i < N_i; i++)
                {
                    d_obj = *(ins_objs[ i ]);

                    try { d_obj.set_var_param_stats(p[0], p[1]); }
                    catch (jwscxx::base::Arg_error e) { }

                    view_ll = ins_views[ i ]->calc_var_param_log_likelihood();
                    obj_ll  = d_obj.calc_var_param_log_likelihood();
                    E_2    += - (view_ll + obj_ll);
                }

                grad->elts[ g ] = (E_1 - E_2) / (2 * eta[ s ]->elts[ d ]);
            } 
            else 
            {
                grad->elts[ g ] = (E_1 - E) / eta[ s ]->elts[ d ];
            }

            g++;
        }
    }

    assert(grad->num_elts == g);

    free_vector_d(p[0]);
    free_vector_d(p[1]);
    free_vector_d(eta[0]);
    free_vector_d(eta[1]);
}

/* -------------------------------------------------------------------------- */
