/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Functions to compute the likelihood of models given observed data.
 */


#ifndef LIKELIHOOD_H
#define LIKELIHOOD_H


#include <config.h>

#include <list>
#include <vector>

#include <jwsc/base/time.h>
#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/image/edge.h>

#include <jwsc++/base/exception.h>

#include "view.h"
#include "object.h"


namespace Likelihood 
{
    /**
     * @class Correspondence
     *
     * Contains pointers to data edge point per cell.
     */
    class Correspondence
    {
        public:

            /**
             * @brief Creates a new correspondence between edge points and
             * model points.
             */
            Correspondence
            (
                const jwsc::Edge_set_f* data_edges,
                uint32_t                num_rows,
                uint32_t                num_cols,
                uint32_t                num_angles,
                double                  angle_sigma,
                double                  dist_sigma,
                double                  max_dist
            );


            /**
             * @class Point
             *
             * Correspondence point between a model point and edge point.
             */
            class Point
            {
                public:
            
                    Point
                    (
                        const jwsc::Edge_point_f* e_pt, 
                        uint32_t                  e_pt_num,
                        uint32_t                  m_pt_row,
                        uint32_t                  m_pt_col,
                        uint32_t                  num_angles,
                        double                    angle_sigma,
                        double                    dist_sigma
                    );

                    class Compare
                    {
                        public:

                            bool operator() (const Point* p1, const Point* p2) 
                                const;
                    };

            
                    bool operator< (const Point& p) const;

                    /** @brief Edge point detected in the input image. */
                    const jwsc::Edge_point_f* e_pt;
            
                    /** 
                     * @brief Index number of the edge point detected,
                     * starting with 0.
                     */
                    uint32_t e_pt_num;
            
                    /** 
                     * @brief Row number for the model point corresponding to
                     * the edge point.
                     */
                    uint32_t m_pt_row;
            
                    /** 
                     * @brief Column number for the model point corresponding
                     * to the edge point.
                     */
                    uint32_t m_pt_col;

                    /** 
                     * @brief Difference in theta between discretized angles.
                     */
                    double dtheta;

                    /**
                     * @brief Discretized angles between model point
                     * normals and an edge point's gradient.
                     *
                     * Because of symmetry, ranges 0 to PI/2.
                     */
                    std::vector<float> angles;

                    /**
                     * @brief Angle between model point normal and an edge
                     * point's gradient.
                     *
                     * Because of symmetry, ranges 0 to PI/2.
                     */
                    float angle;

                    /**
                     * @brief Gaussian likelihood of the discretized angles 
                     * angle.
                     */
                    std::vector<double> gauss_angles;

                    /** @brief Gaussian likelihood of the angle. */
                    double gauss_angle;
            
                    /** 
                     * @brief Euclidean distance from the model line to the
                     * edge point along the normal direction defined by the
                     * discretized angles.
                     */
                    std::vector<float> norm_dists;
            
                    /** 
                     * @brief Euclidean distance from the model line to the
                     * edge point along the normal direction defined by the
                     * discretized angles.
                     */
                    float norm_dist;
            
                    /** 
                     * @brief Gaussian likelihood of the Euclidean normal
                     * distance.
                     */
                    std::vector<double> gauss_norm_dists;
            
                    /** 
                     * @brief Gaussian likelihood of the Euclidean normal
                     * distance.
                     */
                    double gauss_norm_dist;

                    /** 
                     * @brief Euclidean distance from the edge point to the
                     * model point along the gradient direction.
                     */
                    float grad_dist;
            };


            /**
             * @brief Generates a correspondence set between each edge point
             * and the model points.
             */
            const std::vector< std::vector<const Point*> >* generate_for_model
            (
                View*                   view,
                Object*                 obj,
                const jwsc::Matrix_u32* model_mat
            );

            double get_max_dist() const;

        private:

            /** 
             * @brief Adds edge points to the matrix of correspondences along
             * the direction of the point's gradient.
             */
            void add_edge_pts_along_line
            (
                const jwsc::Edge_point_f* e_pt,
                uint32_t                  e_pt_num,
                uint32_t                  num_rows,
                uint32_t                  num_cols,
                uint32_t                  num_angles,
                double                    angle_sigma,
                double                    dist_sigma,
                double                    max_dist
            );

            /** 
             * @brief Matrix of correspondences beween a model point and edge 
             * points.
             *
             * The matrix is indexed according to the image coordinates of each
             * model point. Computed only once when the constructor for
             * Correspondence is called. This is essentially the
             * pre-processing data-structure.
             */
            std::vector< std::vector< std::list<Point> > > m_pts_corr;

            /**
             * @brief Vector of correspondences between an edge points and
             * model points.
             *
             * The vector is indexed by edge point number, then the vector of
             * correspondence points gives all the model points matching it.
             * This is computed for each rendering of a model, but storage is
             * re-used each time.
             */
            std::vector< std::vector<const Point*> > e_pts_corr;

            double max_dist;
    };


    /**
     * @class Cache
     *
     * Log likelihood cache.
     */
    class Cache
    {
        public:

            Cache(uint32_t max_size, uint32_t k, double t);

            ~Cache();
            
            /** 
             * @brief Queries a log likelihood cache for a (interpolated)
             * value.
             */
            bool query
            (
                double*       potential_out,
                const View*   view,
                const Object* obj
            )
            throw (jwscxx::base::Arg_error);

            /** @brief Inserts a log likelihood value into the cache. */
            void insert
            (
                double        potential,
                const View*   view,
                const Object* obj
            )
            throw (jwscxx::base::Arg_error);

            /** @brief Clears the cache. */
            void clear();

            uint32_t get_num_hits() const;

            uint32_t get_num_misses() const;

            uint32_t get_cache_size() const;

            uint32_t get_cache_pt_size() const;

            float get_percent_hits() const;

        private:

            uint32_t max_size;
            uint32_t size;

            /** 
             * @brief Size (dimension) of the points in this Cache. 
             *
             * Set when the first point is inserted.
             */
            uint32_t pt_size;

            /** 
             * @brief Number of cache elements needed with t distance to have
             * a hit.
             */
            uint32_t k;

            /** 
             * @brief Distance a set of k elements have to be from a query to
             * be considered a cache hit.
             */
            double t;

            uint32_t num_hits;
            uint32_t num_misses;
            jwsc::Vector_d* q;
            jwsc::Vector_d* q_1;
            jwsc::Vector_d* q_2;
            std::list<jwsc::Vector_d*> cache;
            std::list<jwsc::Vector_d*> k_closest;
    };


    /**
     * @brief Renders an Object under a View using the GL and captures
     * the image.
     */
    void render_wire_model_scene
    (
        jwsc::Matrix_u32** model_mat_out, 
        const View*        view, 
        const Object*       obj
    );


    /**
     * @brief Renders an Object under a View using the GL and captures
     * the image.
     */
    void render_solid_model_scene
    (
        jwsc::Matrix_u32** model_mat_out, 
        const View*        view, 
        const Object*       obj
    );


    double clipped_point_penalty
    (
        const View*           view, 
        const jwsc::Vector_d* p
    );
    
    
    double clipped_object_penalty
    (
        const View*   view, 
        const Object* obj
    );


    /**
     * @brief Calculates the image log likelihood for an Object under a View.
     */
    void calc_image_log_likelihood
    (
        double*                 image_ll_out,
        jwsc::Matrix_u32**      model_mat_out,
        jwsc::Matrix_u32**      surf_ll_assign_out,
        jwsc::Matrix_u32**      edge_ll_assign_out,
        const jwsc::Matrix_f*   image_t,
        const jwsc::Edge_set_f* edges,
        Correspondence*         corr,
        View*                   view,
        Object*                 obj,
        double                  bg_prob,
        double                  surf_prob,
        double                  surf_noise_prob,
        double                  surf_miss_prob,
        double                  edge_noise_prob,
        double                  edge_silho_miss_prob,
        double                  edge_inner_miss_prob
    ) 
    throw (jwscxx::base::Arg_error);


    /**
     * @brief Calculates the negative log likelihood for an Object
     * under a View.
     */
    void calc_neg_log_likelihood
    (
        double*                 potential_out,
        double*                 image_ll_out,
        jwsc::Matrix_u32**      model_mat_out,
        jwsc::Matrix_u32**      surf_ll_assign_out,
        jwsc::Matrix_u32**      edge_ll_assign_out,
        const jwsc::Matrix_f*   image_t,
        const jwsc::Edge_set_f* edges,
        Correspondence*         corr,
        View*                   view,
        Object*                 obj,
        double                  bg_prob,
        double                  surf_prob,
        double                  surf_noise_prob,
        double                  surf_miss_prob,
        double                  edge_noise_prob,
        double                  edge_silho_miss_prob,
        double                  edge_inner_miss_prob
    ) 
    throw (jwscxx::base::Arg_error);


    /**
     * @brief Calculates the negative log likelihood and prior for an Object
     * under a View.
     */
    void calc_neg_log_likelihood_and_prior
    (
        double*                 potential_out,
        double*                 image_ll_out,
        jwsc::Matrix_u32**      model_mat_out,
        jwsc::Matrix_u32**      surf_ll_assign_out,
        jwsc::Matrix_u32**      edge_ll_assign_out,
        const jwsc::Matrix_f*   image_t,
        const jwsc::Edge_set_f* edges,
        Correspondence*         corr,
        View*                   view,
        Object*                 obj,
        double                  bg_prob,
        double                  surf_prob,
        double                  surf_noise_prob,
        double                  surf_miss_prob,
        double                  edge_noise_prob,
        double                  edge_silho_miss_prob,
        double                  edge_inner_miss_prob
    ) 
    throw (jwscxx::base::Arg_error);


    /**
     * @brief Calculates the gradient of the negative log likelihood and prior
     * of an Object under a View with respect to the instance parameters.
     */
    void calc_grad_neg_log_likelihood_and_prior_wrt_ins_params
    (
        jwsc::Vector_d**        grad_out,
        jwsc::Matrix_u32**      model_mat_out,
        jwsc::Matrix_u32**      surf_ll_assign_out,
        jwsc::Matrix_u32**      edge_ll_assign_out,
        double                  potential,
        const jwsc::Matrix_f*   image_t,
        const jwsc::Edge_set_f* edges,
        Correspondence*         corr,
        Cache*                  cache,
        View*                   view,
        Object*                 obj,
        double                  bg_prob,
        double                  surf_prob,
        double                  surf_noise_prob,
        double                  surf_miss_prob,
        double                  edge_noise_prob,
        double                  edge_silho_miss_prob,
        double                  edge_inner_miss_prob,
        bool                    two_point_estimate
    ) 
    throw (jwscxx::base::Arg_error);


    /**
     * @brief Calculates the gradient of the negative log prior of an Object
     * under a View with respect to the category parameters.
     */
    void calc_grad_neg_log_likelihood_and_prior_wrt_cat_params
    (
        jwsc::Vector_d**                  grad_out,
        double                            lp_potential,
        const View*                       view,
        const Object*                     obj,
        const std::vector<const View*>&   ins_views,
        const std::vector<const Object*>& ins_objs,
        bool                              two_point_estimate
    ) 
    throw (jwscxx::base::Arg_error);
}


#endif
