/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief MCMC Metropolis-Hastings sampler.
 */


#ifndef METROPOLIS_H
#define METROPOLIS_H


#include <config.h>

#include <list>
#include <vector>
#include <string>

#include <inttypes.h>
#include <pthread.h>

#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/image/edge.h>

#include <jwsc++/base/exception.h>

#include "view.h"
#include "object.h"
#include "block.h"
#include "block_set.h"
#include "likelihood.h"
#include "sampler.h"
#include "dynamics.h"


/**
 * @class Metropolis
 *
 * MCMC sampler for finding the most likely set of model parameters under
 * some detected edge points.
 */
class Metropolis : public Sampler
{
    public:


/**
 * @class Metropolis::Instance
 *
 * @brief Instance of models being inferred by the Sampler from data.
 *
 * This class is more or less a C structure with a helpful allocator and
 * deallocator. It's contained by the Sampler::Parameters class and used in a
 * vector of Instance.
 */
class Instance : public Sampler::Instance
{
    public:

        /**
         * @name Standard
         *
         * Standard class members.
         * @{
         */

        /** @brief Constructs a new Instance for the Sampler. */
        Instance
        (
            View*                       best_view,
            Object*                     best_obj,
            double                      best_img_ll,
            double                      best_potential,
            View*                       view,
            Object*                     obj,
            double                      img_ll,
            double                      potential,
            const jwsc::Matrix_f*       image_t,
            const jwsc::Edge_set_f*     edges,
            Likelihood::Correspondence* img_ll_corr,
            Likelihood::Cache*          img_ll_cache,
            jwsc::Matrix_u32*           img_ll_surf_assign,
            jwsc::Matrix_u32*           img_ll_edge_assign,
            jwsc::Matrix_u32*           model_mat,
            jwsc::Vector_d*             sample_mean_in,
            jwsc::Matrix_d*             sample_cov_in,
            const char*                 sample_info_fname,
            bool                        compress_sample_info,
            const char*                 sample_stats_info_fname
        )
        throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);

        /** @brief Deletes this Instance. */
        virtual ~Instance();

        /*@}*/


        /**
         * @name Variable Parameters
         *
         * Variable model parameter members.
         * @{
         */

        /** @brief Gets a vector of proposed model parameters. */
        void get_var_param_proposal_values(jwsc::Vector_d** p_out) const;

        /** 
         * @brief Adds a vector to the proposed model parameters and scales
         * dependent components' parameters.
         */
        void add_var_param_proposal_values_scale_component_deps
        (
            const jwsc::Vector_d* p
        );

        /** @brief Gets a vector of proposed model parameters. */
        void get_var_param_proposal_stats(jwsc::Vector_d** p_out) const;

        /** @brief Gets a vector of proposed model parameters. */
        void set_var_param_proposal_stats(const jwsc::Vector_d* p);

        /** @brief Prints the variable parameter proposal information. */
        void print_var_param_proposal_info
        (
            const char* move_name,
            double      membership
        );

        /** @brief Prints the variable parameter proposal information. */
        void print_var_param_proposal_info
        (
            const char*   move_name, 
            double        membership,
            std::ostream& out
        );

        /*@}*/


        /** @brief Proposed View. */
        View* prop_view;

        /** @brief Proposed Object. */
        Object* prop_obj;

        /** @brief Potential energy of the proposed model likelihoods. */
        double prop_ll_potential;

        /** @brief Kinetic energy of the proposed model likelihoods. */
        double prop_ll_kinetic;

        /** @brief Log likelihood of the proposed models. */
        double prop_img_ll;

        /** @brief Whether the last sampler proposal was accepted. */
        bool accepted;
};


/**
 * @class Metropolis::Category
 *
 * @brief Category of objects being inferred by the Sampler from data.
 *
 * This class is more or less a C structure with a helpful allocator and
 * deallocator. It's contained by the Sampler::Parameters class and used in a
 * vector of subcategories.
 */
class Category : public Sampler::Category
{
    public:

        /**
         * @name Standard
         *
         * Standard class members.
         * @{
         */

        /** @brief Constructs a new Category for the Sampler. */
        Category
        (
            const Metropolis::Instance* rep_ins,
            double                      lp_potential,
            double                      potential,
            const char*                 sample_info_fname,
            bool                        compress_sample_info
        )
        throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);

        /** @brief Deletes this Instance. */
        virtual ~Category();

        /*@}*/


        /**
         * @name Variable Parameters
         *
         * Variable model parameter members.
         * @{
         */

        /** @brief Prints variable parameter information. */
        void print_var_param_proposal_info(const char* move_name);

        /** @brief Prints the variable parameter proposal information. */
        void print_var_param_proposal_info
        (
            const char*   move_name,
            std::ostream& out
        );

        /*@}*/


        bool accepted() const;


        /** @brief Most recently generated View sample. */
        const View* prop_view;

        /** @brief Most recently generated Object sample. */
        const Object* prop_obj;

        /** @brief Potential energy of the model priors. */
        double prop_lp_potential;

        /** @brief Kinetic energy of the model priors. */
        double prop_lp_kinetic;

        /** @brief Potential energy of the models. */
        double prop_potential;

        /** @brief Kinetic energy of the models. */
        double prop_kinetic;


    private:

        /** @brief Whether the last proposal was accepted. */
        const bool* _accepted;
};


class Move;


/**
 * @class Parameters
 */
class Parameters : public Sampler::Parameters
{
    public:

        /**
         * @name Standard
         *
         * Standard class members.
         * @{
         */

        /** @brief Constructs a Metropolis::Parameters. */
        Parameters
        (
            const View::Context*      view_ctx,
            const Object::Context*    obj_ctx,
            const Block::Context*     block_ctx,
            const Block_set::Context* block_set_ctx,
            double                    img_ll_bg_prob,
            double                    img_ll_fg_prob,
            double                    img_ll_fg_noise_prob,
            double                    img_ll_fg_miss_prob,
            double                    img_ll_edge_noise_prob,
            double                    img_ll_edge_silho_miss_prob,
            double                    img_ll_edge_inner_miss_prob,
            uint32_t                  sample_stats_freq,
            uint32_t                  num_instances,
            uint32_t                  num_categories,
            uint32_t                  min_num_members,
            uint32_t                  num_threads
        );

        /** @brief Deletes a Sampler_move_parameters. */
        virtual ~Parameters();

        /*@}*/


        /**
         * @name Model Instance
         *
         * Model instance members.
         * @{
         */

        /** 
         * @brief Returns a Instance from this set of sampler Parameters.
         */
        Metropolis::Instance* instance(uint32_t c, uint32_t i) const
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Updates the model category parameter statistics using the
         * model instances.
         */
        void update_instance_proposal_stats();

        /*@}*/


        /**
         * @name Object category
         *
         * Object category members.
         * @{
         */

        /** 
         * @brief Returns a category from this set of sampler Parameters.
         */
        Metropolis::Category* category(uint32_t c) const
            throw (jwscxx::base::Arg_error);

        /*@}*/


        /**
         * @name Variable Parameters
         *
         * Variable model parameter members.
         * @{
         */

        /** @brief Gets a vector of proposed model parameters. */
        void get_var_param_proposal_values
        (
            jwsc::Vector_d** p_out, 
            uint32_t         c, 
            uint32_t         i
        ) 
        const;

        /** 
         * @brief Adds a vector to the proposed model parameters and scales
         * dependent components' parameters.
         */
        void add_var_param_proposal_values_scale_component_deps
        (
            const jwsc::Vector_d* p,
            uint32_t              c,
            uint32_t              i
        );

        /** @brief Sets a vector of proposed variable parameter values. */
        void set_var_param_proposal_values_dynamics
        (
            const jwsc::Vector_d* q,
            const jwsc::Vector_d* p,
            uint32_t              c,
            uint32_t              i
        );

        /** @brief Gets a vector of proposed model parameters. */
        void get_var_param_proposal_stats
        (
            jwsc::Vector_d** p_out, 
            uint32_t         c
        ) 
        const;

        /** @brief Gets a vector of proposed model parameters. */
        void set_var_param_proposal_stats
        (
            const jwsc::Vector_d* p, 
            uint32_t              c
        );

        void set_var_param_proposal_stats_dynamics
        (
            const jwsc::Vector_d* q,
            const jwsc::Vector_d* p,
            uint32_t              c
        );

        /** @brief Prints the variable parameter proposal information. */
        void print_var_param_proposal_info
        (
            const std::vector<Metropolis::Move*>& moves
        );

        /*@}*/


        /**
         * @name Energy
         *
         * Energy function members.
         * @{
         */

        /** @brief Gets the gradient of the potential energy function. */
        void proposal_grad_ins_potential
        (
            jwsc::Vector_d** grad_out, 
            uint32_t         c, 
            uint32_t         i
        )
        const;

        /** @brief Gets the gradient of the potential energy function. */
        void proposal_grad_cat_potential
        (
            jwsc::Vector_d** grad_out, 
            uint32_t         c
        )
        const;

        /*@}*/


        /**
         * @name Proposal
         *
         * Proposal management members.
         * @{
         */

        /** 
         * @brief Initializes the proposed models in a Category with a
         * copy of the most recently accepted models.
         */
        void init_cat_proposal(uint32_t c);

        /** 
         * @brief Initializes the proposed models in a Instance with a
         * copy of the most recently accepted models.
         */
        void init_ins_proposal(uint32_t c, uint32_t i);

        /** 
         * @brief Copies the proposed models of a Category into the most
         * recently accepted models.
         */
        void accept_cat_proposal(uint32_t c);

        /** 
         * @brief Copies the proposed models of a Instance into the most
         * recently accepted models.
         */
        void accept_ins_proposal(uint32_t c, uint32_t i);

        /** 
         * @brief Updates values that depend on the proposed models in a
         * Category.
         */
        void update_cat_proposal_dependents(uint32_t c);

        /** 
         * @brief Updates values that depend on the proposed models in a
         * Category.
         */
        void update_ins_proposal_cat_dependents(uint32_t c, uint32_t i);

        /** 
         * @brief Updates values that depend on the proposed models in an
         * Instance.
         */
        void update_ins_proposal_dependents(uint32_t c, uint32_t i);

        /*@}*/


    protected:

        /** @brief Updates the kinetic energy of the model parameters. */
        void update_proposal_ins_ll_kinetic
        (
            const jwsc::Vector_d* p,
            uint32_t              c,
            uint32_t              i
        );

        /** @brief Updates the kinetic energy of the model parameters. */
        void update_proposal_cat_lp_kinetic
        (
            const jwsc::Vector_d* p,
            uint32_t              c
        );
};


class Move;


/**
 * @class Metropolis::Thread_manager
 *
 * Provides the thread synchronization to execute code by only one thread while
 * the others (busy) wait for it to finish. When any thread leaves the
 * synchronization block, the code is guarenteed to have been complete executed.
 *
 * The template for a synchronization block is as follows
 *
 * @code
 * synchronize();
 *
 * if (last_thread())
 * {
 *     // Code to be executed synchronously by only one thread goes here.
 *
 *     release();
 * }
 * else
 * {
 *     wait();
 * }
 *
 * if (thread() == t)
 * {
 *     // Code to be executed by the t^th thread goes here.
 * }
 *
 * // Code to be executed by all threads goes here.
 *
 * desynchronize();
 * @endcode
 */
class Thread_manager : public Sampler::Thread_manager
{
    public:

        /** @brief Constructs a Metropolis::Parameters. */
        Thread_manager(uint32_t thread=0);

        /** @brief Deletes a Sampler_move_parameters. */
        virtual ~Thread_manager();

        /** @brief Sets the category currently used by the sampler. */
        void category(uint32_t c);

        /** @brief Returns the category currently used by the sampler. */
        uint32_t category() const;

        /** @brief Sets the Metropolis sampler to use. */
        void sampler(Sampler* sampler) throw (jwscxx::base::Arg_error);

        /** @brief Returns the Metropolis sampler to use. */
        Metropolis* sampler() const;

        /** @brief Sets the Metropolis::Parameters to use. */
        void params(Sampler::Parameters* params) 
            throw (jwscxx::base::Arg_error);

        /** @brief Returns the Metropolis::Parameters to use. */
        Metropolis::Parameters* params() const;

        /** @brief Sets the Metropolis::Move to use for a category. */
        void move(Metropolis::Move* move, uint32_t c);

        /** @brief Returns the Metropolis::Move to use for a category. */
        Metropolis::Move* move(uint32_t c);

        /** @brief Returns the moves used by a run of the sampler. */
        const std::vector<Metropolis::Move*>& moves();

        /** @brief Selects a sampler move for all threads to run. */
        void select_sampler_moves() throw (jwscxx::base::Arg_error);

        /** @brief Assigns category membership. */
        void assign_min_potential_members();

        /** @brief Prints proposal information. */
        void print_var_param_proposal_info();

        /** 
         * @brief Gets a vector of parameters from the models.
         *
         * Static C interface to
         * Sampler::Parameters::get_var_param_proposal_values().
         */
        static void get_var_param_proposal_values_c
        (
            jwsc::Vector_d**                  p_out,
            const Metropolis::Thread_manager* mgr
        );

        /** 
         * @brief Gets a vector of parameters from the models.
         *
         * Static C interface to
         * Sampler::Parameters::get_var_param_proposal_values().
         */
        static void get_var_param_proposal_stats_c
        (
            jwsc::Vector_d**                  p_out,
            const Metropolis::Thread_manager* mgr
        );

        /** @brief Gets a vector of proposed model parameters. */
        static void get_var_param_proposal_stats_and_values_c
        (
            jwsc::Vector_d**            p_out,
            Metropolis::Thread_manager* mgr
        );

        /** 
         * @brief Returns the gradient of the potential energy function for
         * the proposed models
         *
         * Static C interface to
         * Metropolis::Parameters::
         * get_var_param_proposal_values_grad_potential().
         */
        static jwsc::Error* get_var_param_proposal_values_grad_ins_potential_c
        (
            jwsc::Vector_d**            grad_out, 
            Metropolis::Thread_manager* mgr
        );

        /** 
         * @brief Returns the gradient of the potential energy function for
         * the proposed models
         *
         * Static C interface to
         * Metropolis::Parameters::
         * get_var_param_proposal_stats_grad_potential().
         */
        static jwsc::Error* get_var_param_proposal_stats_grad_cat_potential_c
        (
            jwsc::Vector_d**            grad_out, 
            Metropolis::Thread_manager* mgr
        );

        /** 
         * @brief Returns the gradient of the potential energy function for
         * the model in the parameters.
         *
         * Static C interface to a thread-safe 
         * Sampler::Parameters::grad_potential().
         */
        static jwsc::Error* get_var_param_proposal_stats_and_values_grad_potential_c
        (
            jwsc::Vector_d**            grad_out, 
            Metropolis::Thread_manager* mgr
        );

        /** 
         * @brief Sets the gradient of the potential energy function for
         * the proposed models
         *
         * Static C interface to
         * Metropolis::Parameters::set_var_param_proposal_grad_potential().
         */
        static jwsc::Error* set_var_param_proposal_values_dynamics_c
        (
            const jwsc::Vector_d*       q,
            const jwsc::Vector_d*       p,
            Metropolis::Thread_manager* mgr
        );

        /** 
         * @brief Sets the gradient of the potential energy function for
         * the proposed models
         *
         * Static C interface to
         * Metropolis::Parameters::set_var_param_proposal_grad_potential().
         */
        static jwsc::Error* set_var_param_proposal_stats_dynamics_c
        (
            const jwsc::Vector_d*       q,
            const jwsc::Vector_d*       p,
            Metropolis::Thread_manager* mgr
        );

        /**
         * @brief Accepts a model point.
         */
        static jwsc::Error* set_var_param_proposal_stats_and_values_dynamics_c
        (
            const jwsc::Vector_d*       q,
            const jwsc::Vector_d*       p,
            Metropolis::Thread_manager* mgr
        );

        /** 
         * @brief Returns the stochastic dynamics parameters for models
         *
         * Static C interface to
         * Metropolis::Parameters::get_var_param_values_dynamics_delta_t().
         */
        static void get_var_param_values_dynamics_delta_t_c
        (
            jwsc::Vector_d**            delta_t_out, 
            Metropolis::Thread_manager* mgr
        );

        /** 
         * @brief Returns the stochastic dynamics parameters for models
         *
         * Static C interface to
         * Metropolis::Parameters::get_var_param_stats_dynamics_delta_t().
         */
        static void get_var_param_stats_dynamics_delta_t_c
        (
            jwsc::Vector_d**            delta_t_out, 
            Metropolis::Thread_manager* mgr
        );

        /** 
         * @brief Returns a vector of stochastic dynamics delta_t from the
         * parameter structure. 
         *
         * Static C interface to a thread-safe 
         * Sampler::Parameters::get_var_param_dynamics_stats_delta_t() and
         * Sampler::Parameters::get_var_param_dynamics_values_delta_t().
         */
        static void get_var_param_stats_and_values_dynamics_delta_t_c
        (
            jwsc::Vector_d**            delta_t_out, 
            Metropolis::Thread_manager* mgr
        );


    private:

        /** @brief Category index. */
        uint32_t _category;

        /** 
         * @brief Sampler move to run at the next iteration for each category.
         */
        static std::vector<Metropolis::Move*> _moves;
};


/**
 * @class Move
 *
 * MCMC Metropolis-Hastings move.
 */
class Move : public jwscxx::base::Cloneable
{
    public:

        /** @brief Constructs a Move. */
        Move(const char* name, double p) throw (jwscxx::base::Arg_error);

        virtual ~Move();

        /** @brief Clones a Move. */
        virtual Metropolis::Move* clone() const = 0;

        /** @brief Runs this move and either accepts or rejects. */
        virtual void run(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Returns the name of this move. */
        const char* name() const;

        /** @brief Returns the probability of this move. */
        double prob() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        virtual bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const = 0;

        /** @brief Clips a parameter value against its context range. */
        void clip_param_val(double* x, const Model::Parameter::Context& ctx) 
            const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        virtual void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the move in a managed thread. */
        virtual void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception) = 0;

        /** @brief Accepts or rejects the move in a managed thread. */
        virtual void accept_or_reject(Metropolis::Thread_manager* mgr) = 0;

        /** @brief Debugging. */
        void assert_params(Metropolis::Thread_manager* mgr);


    private:

        /** @brief Sets the probability of this move. */
        void prob(double prob) throw (jwscxx::base::Arg_error);


    protected:

        /** @brief Single word name of this move. */
        std::string _name;

        /** @brief Probability of this move. */
        double _prob;
};


/**
 * @class Metropolis::Hybrid_dynamics
 *
 * @brief Stochastic dynamics interface for Metropolis moves.
 */
class Hybrid_dynamics
{
    public:


/**
 * @class Metropolis::Hybrid_dynamics::Stochastic
 *
 * MCMC sampler for finding the most likely set of model parameters under
 * some detected edge points.
 */
class Stochastic : public Stochastic_dynamics
{
    public:

        /** @brief Deletes an MCMC Sampler and all its moves. */
        Stochastic(uint32_t iterations, double alpha, uint32_t kick) 
            throw (jwscxx::base::Arg_error);

        /** Returns the number of iterations to run. */
        uint32_t iterations() const;

        /** Sets the the number of iterations to run. */
        void iterations(uint32_t i) throw (jwscxx::base::Arg_error);

        /** @brief Runs the iterative MCMC sampler. */
        void run_thread(Metropolis::Thread_manager* mgr) const 
            throw (jwscxx::base::Exception);

    private:

        uint32_t _iterations;
};


    // Hybrid_dynamics
    public:

        /** @brief Constructs a Gauss_move. */
        Hybrid_dynamics
        (
            double      alpha,
            uint32_t    kick,
            uint32_t    leapfrog_i
        ) 
        throw (jwscxx::base::Arg_error);

        /** Returns the proportion of momentum to conserve per iteration. */
        double alpha() const;

        /** Sets the proportion of momentum to conserve per iteration. */
        void alpha(double a) throw (jwscxx::base::Arg_error);

        /** Returns the frequency to kick the particle and reset its momenta. */
        uint32_t kick() const;

        /** Sets the frequency to kick the particle and reset its momenta. */
        void kick(uint32_t k) throw (jwscxx::base::Arg_error);

        /** 
         * Returns the number of leapfrog iterations between each iteration
         * of the sampler.
         */
        uint32_t leapfrog_i() const;

        /** 
         * Sets the number of leapfrog iterations between each iteration 
         * of the sampler.
         */
        void leapfrog_i(uint32_t i);


    protected:

        /** 
         * @brief Proportion of momentum to conserve per iteration.
         *
         * Must be in [0,1].
         *
         * (1-alpha*alpha) scales gaussian noise added to the momentum per
         * iteration.
         */
        double _alpha;

        /**
         * @brief Frequncy (in iterations) to kick the momenta variables of the
         * particle and reset them to a non-stochastic value.
         *
         * Causes the partical to jump out of a local minima.
         */
        uint32_t _kick;

        /** 
         * @brief Number of leapfrog iterations between each iteration of the
         * sampler.
         */
        uint32_t _leapfrog_i;
};


/**
 * @class Metropolis::Diffusion
 *
 * MCMC Sampler move that modifies the value of a model parameter.
 */
class Diffusion : public Move
{
    public:

        /** @brief Constructs a Diffusion. */
        Diffusion(const char* name, double prob) 
            throw (jwscxx::base::Arg_error);

        /** @brief Deletes a Diffusion. */
        virtual ~Diffusion();
};


/**
 * @class Metropolis::Diffusion_ins
 *
 * MCMC Sampler move that modifies the value of model instance parameters.
 */
class Diffusion_ins : public Diffusion
{
    public:

        /** @brief Constructs a Diffusion_ins. */
        Diffusion_ins(const char* name, double prob) 
            throw (jwscxx::base::Arg_error);

        /** @brief Deletes a Diffusion_ins. */
        virtual ~Diffusion_ins();


    protected:

        /** @brief Initializes the move in a managed thread. */
        virtual void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the move in a managed thread. */
        virtual void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception) = 0;

        /** @brief Accepts or rejects the move in a managed thread. */
        void accept_or_reject(Metropolis::Thread_manager* mgr);

        /** @brief Returns the most recent acceptance log probability. */
        double log_alpha(size_t i) const;

        /** @brief Sets the most recent acceptance log probability. */
        void log_alpha(double log_alpha, size_t i);

        /** 
         * @brief Returns whether the most recent move should be rejected
         * outright.
         */
        double reject(size_t i) const;

        /** 
         * @brief Sets whether the most recent move should be rejected outright.
         */
        void reject(bool reject, size_t i);


    protected:

        /** 
         * @brief Acceptance log probability of the most recent run.
         *
         * Index is by instance.
         */
        std::vector<double> _log_alpha;

        /** 
         * @brief Whether the most recent move run should be rejected outright.
         *
         * Index is by instance.
         */
        std::vector<bool> _reject;
};


/**
 * @class Metropolis::Swap
 *
 * @brief Proposes a swapping the width and length of the object model.
 */
class Swap : public Diffusion_ins
{
    public:

        /** @brief Constructs a sampler move. */
        Swap(const char* name, double prob=1.0) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Swap* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Runs the move in a managed thread. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);
};


/**
 * @class Metropolis::Gauss
 *
 * @brief Proposes a Gaussian move across all parameters in furniture.
 */
class Gauss : public Diffusion_ins
{
    public:

        /** @brief Constructs a sampler move. */
        Gauss(const char* name, double prob=1.0) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Gauss* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Runs the move in a managed thread. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);
};


/**
 * @class Metropolis::MV_gauss_shift
 *
 * @brief Proposes a multivariate Gaussian move across all parameters in
 * furniture. 
 *
 * The dependencies of components are shifted in position.
 */
class MV_gauss_shift : public Diffusion_ins
{
    public:

        /** @brief Constructs a MV_gauss_shift. */
        MV_gauss_shift(const char* name, double prob=1.0) 
            throw (jwscxx::base::Arg_error);

        /** @brief Deletes a move. */
        virtual ~MV_gauss_shift();

        /** @brief Clones a Move. */
        virtual Metropolis::MV_gauss_shift* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Runs the move in a managed thread. */
        virtual void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);
};


/**
 * @class Metropolis::View_gauss
 *
 * @brief Proposes a multivariate Gaussian move across view parameters in
 * furniture. 
 */
class View_gauss : public MV_gauss_shift
{
    public:

        /** @brief Constructs a move. */
        View_gauss(const char* name, double prob=1.0) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::View_gauss* clone() const;


    protected:

        /** @brief Runs the move in a managed thread. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);
};


/**
 * @class Metropolis::Object_gauss_shift
 *
 * @brief Proposes a multivariate Gaussian move across view parameters in
 * furniture. 
 *
 * The dependencies of components are shifted in position.
 */
class Object_gauss_shift : public MV_gauss_shift
{
    public:

        /** @brief Constructs a move. */
        Object_gauss_shift(const char* name, double prob=1.0) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Object_gauss_shift* clone() const;


    protected:

        /** @brief Runs the move in a managed thread. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);
};


/**
 * @class Metropolis::Component_gauss_shift
 *
 * @brief Proposes a multivariate Gaussian move across view parameters in
 * furniture. 
 *
 * The dependencies of components are shifted in position.
 */
class Component_gauss_shift : public MV_gauss_shift
{
    public:

        /** @brief Constructs a move. */
        Component_gauss_shift(const char* name, double prob=1.0) 
            throw (jwscxx::base::Arg_error);

        /** @brief Deletes a move. */
        virtual ~Component_gauss_shift();

        /** @brief Clones a Move. */
        virtual Metropolis::Component_gauss_shift* clone() const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        virtual void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the move in a managed thread. */
        virtual void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Returns the index of the component to change. */
        uint32_t comp_i();

        /** 
         * @brief Returns the number of object var params before the ith
         * component selected for change.
         */
        uint32_t comp_n();


    protected:

        /** @brief The index of the component to change. */
        uint32_t _comp_i;

        /** 
         * @brief The number of variable parameters in the object before the
         * ith component selected for change.
         */
        uint32_t _comp_n;
};


/**
 * @class Metropolis::MV_gauss_scale
 *
 * @brief Proposes a multivariate Gaussian move across all parameters in
 * furniture. 
 *
 * The dependencies of components are scaled in position.
 */
class MV_gauss_scale : public Diffusion_ins
{
    public:

        /** @brief Constructs a MV_gauss_scale. */
        MV_gauss_scale(const char* name, double prob=1.0) 
            throw (jwscxx::base::Arg_error);

        /** @brief Deletes a move. */
        virtual ~MV_gauss_scale();

        /** @brief Clones a Move. */
        virtual Metropolis::MV_gauss_scale* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Runs the move in a managed thread. */
        virtual void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);
};


/**
 * @class Metropolis::Object_gauss_scale
 *
 * @brief Proposes a multivariate Gaussian move across view parameters in
 * furniture. 
 *
 * The dependencies of components are scaleed in position.
 */
class Object_gauss_scale : public MV_gauss_scale
{
    public:

        /** @brief Constructs a move. */
        Object_gauss_scale(const char* name, double prob=1.0) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Object_gauss_scale* clone() const;


    protected:

        /** @brief Runs the move in a managed thread. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);
};


/**
 * @class Metropolis::Component_gauss_scale
 *
 * @brief Proposes a multivariate Gaussian move across view parameters in
 * furniture. 
 *
 * The dependencies of components are scaleed in position.
 */
class Component_gauss_scale : public MV_gauss_scale
{
    public:

        /** @brief Constructs a move. */
        Component_gauss_scale(const char* name, double prob=1.0) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Component_gauss_scale* clone() const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the move in a managed thread. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Returns the index of the component to change. */
        uint32_t comp_i();

        /** 
         * @brief Returns the number of object var params before the ith
         * component selected for change.
         */
        uint32_t comp_n();


    protected:

        /** @brief The index of the component to change. */
        uint32_t _comp_i;

        /** 
         * @brief The number of variable parameters in the object before the
         * ith component selected for change.
         */
        uint32_t _comp_n;
};


/**
 * @class Metropolis::Cross
 *
 * @brief Proposes changing the number and size of cross blocks.
 */
class Cross : public Component_gauss_shift
{
    public:

        /** @brief Constructs a Cross. */
        Cross(const char* name, double prob=1.0) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Cross::Cross* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the move in a managed thread. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);
};


/**
 * @class Metropolis::Recognize
 *
 * @brief Proposes a multivariate Gaussian move across all parameters in
 * furniture. 
 *
 * The dependencies of components are scaled in position.
 */
class Recognize : public Diffusion_ins
{
    public:

        /** @brief Constructs a Recognize. */
        Recognize(const char* name, double prob=1.0) 
            throw (jwscxx::base::Arg_error);

        /** @brief Deletes a move. */
        virtual ~Recognize();

        /** @brief Clones a Move. */
        Metropolis::Recognize* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Runs the move in a managed thread. */
        virtual void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);
};


/**
 * @class Metropolis::Dynamics_ins
 *
 * MCMC Sampler move that modifies the value of model category parameters.
 */
class Dynamics_ins : public Diffusion_ins,
                     public Hybrid_dynamics
{
    public:

        /** @brief Constructs a move. */
        Dynamics_ins
        (
            const char* name, 
            double      alpha, 
            uint32_t    kick, 
            uint32_t    leapfrog_i, 
            double      prob
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Clones a move. */
        Metropolis::Dynamics_ins* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Runs the move in a managed thread. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);
};


/**
 * @class Metropolis::Diffusion_cat
 *
 * MCMC Sampler move that modifies the value of model category parameters.
 */
class Diffusion_cat : public Diffusion
{
    public:

        /** @brief Constructs a Diffusion_cat. */
        Diffusion_cat(const char* name, double prob) 
            throw (jwscxx::base::Arg_error);

        /** @brief Deletes a Diffusion_cat. */
        virtual ~Diffusion_cat();

        /** @brief Clones a Move. */
        virtual Metropolis::Diffusion_cat* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        virtual bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        virtual void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the move in a managed thread. */
        virtual void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Accepts or rejects the move in a managed thread. */
        void accept_or_reject(Metropolis::Thread_manager* mgr);

        /** @brief Returns the most recent acceptance log probability. */
        double log_alpha() const;

        /** @brief Sets the most recent acceptance log probability. */
        void log_alpha(double log_alpha);

        /** 
         * @brief Returns whether the most recent move should be rejected
         * outright.
         */
        double reject() const;

        /** 
         * @brief Sets whether the most recent move should be rejected outright.
         */
        void reject(bool reject);


    protected:

        /** @brief Acceptance log probability of the most recent run. */
        double _log_alpha;

        /** 
         * @brief Whether the most recent move run should be rejected outright.
         */
        bool _reject;
};


/**
 * @class Metropolis::View_cat
 *
 * MCMC Sampler move that modifies the value of view category parameters.
 */
class View_cat : public Diffusion_cat
{
    public:

        /** @brief Constructs a move. */
        View_cat(const char* name, double prob) throw (jwscxx::base::Arg_error);

        /** @brief Clones a move. */
        Metropolis::View_cat* clone() const;


    protected:

        /** @brief Runs the move in a managed thread. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);
};


/**
 * @class Metropolis::Object_cat
 *
 * MCMC Sampler move that modifies the value of object category parameters.
 */
class Object_cat : public Diffusion_cat
{
    public:

        /** @brief Constructs a move. */
        Object_cat(const char* name, double prob) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a move. */
        Metropolis::Object_cat* clone() const;


    protected:

        /** @brief Runs the move in a managed thread. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);
};


/**
 * @class Metropolis::Component_cat
 *
 * MCMC Sampler move that modifies the value of component category parameters.
 */
class Component_cat : public Diffusion_cat
{
    public:

        /** @brief Constructs a move. */
        Component_cat(const char* name, double prob) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a move. */
        Metropolis::Component_cat* clone() const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the move in a managed thread. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Returns the index of the component to change. */
        uint32_t comp_i();


    protected:

        /** @brief The index of the component to change. */
        uint32_t _comp_i;
};


/**
 * @class Metropolis::Dynamics_cat
 *
 * MCMC Sampler move that modifies the value of model category parameters.
 */
class Dynamics_cat : public Diffusion_cat, 
                     public Hybrid_dynamics
{
    public:

        /** @brief Constructs a move. */
        Dynamics_cat
        (
            const char* name, 
            double      alpha, 
            uint32_t    kick, 
            uint32_t    leapfrog_i, 
            double      prob
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Clones a move. */
        Metropolis::Dynamics_cat* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Runs the move in a managed thread. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);


    private:

        /** Stochastic dynamics sampler shared by each thread. */
        Stochastic _dynamics;
};


/**
 * @class Metropolis::Jump
 *
 * An MCMC sampler move that modifies the number model parameter.
 */
class Jump : public Move
{
    public:

        /** @brief Constructs a Jump. */
        Jump(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Deletes a Jump. */
        virtual ~Jump();

        /** @brief Runs this move and either accepts or rejects. */
        void run(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Returns the probability of this move. */
        double pair_prob() const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        virtual void init(Metropolis::Thread_manager* mgr);

        /** @brief Accepts or rejects the move in a managed thread. */
        void accept_or_reject(Metropolis::Thread_manager* mgr);

        /** @brief Updates the category acceptance log probabilities. */
        virtual void update_cat_log_alpha
        (
            Metropolis::Thread_manager* mgr
        ) = 0;

        /** 
         * @brief Returns the most recent instance parameter acceptance log
         * probability.
         */
        double ins_log_alpha(size_t i) const;

        /** 
         * @brief Sets the most recent instance parameter acceptance log
         * probability.
         */
        void ins_log_alpha(double log_alpha, size_t i);

        /** 
         * @brief Returns the most recent category parameter acceptance log
         * probability.
         */
        double cat_log_alpha() const;

        /** 
         * @brief Sets the most recent category parameter acceptance log
         * probability.
         */
        void cat_log_alpha(double log_alpha);

        /** 
         * @brief Returns whether the most recent move should be rejected
         * outright.
         */
        double reject() const;

        /** 
         * @brief Sets whether the most recent move should be rejected outright.
         */
        void reject(bool reject);

        /** @brief Returns the number of object components in a category. */
        uint32_t num_obj_components() const;

        /** 
         * @brief Returns the sum of a subset of the largest acceptance log
         * probabilities across instances.
         */
        virtual double sum_n_best_instance_log_alphas
        (
            Metropolis::Thread_manager* mgr
        );

        /** @brief Variable parameter means for the view. */
        const jwsc::Vector_d* view_means() const;

        /** @brief Variable parameter stdevs for the view. */
        const jwsc::Vector_d* view_stdevs() const;

        /** @brief Variable parameter means for the object. */
        const jwsc::Vector_d* obj_means() const;

        /** @brief Variable parameter stdevs for the object. */
        const jwsc::Vector_d* obj_stdevs() const;


    private:

        /** @brief Sets the paired move probability of this move. */
        void pair_prob(double prob) throw (jwscxx::base::Arg_error);


    protected:

        /** 
         * @brief Instance parameter acceptance log probability of the most
         * recent run.
         *
         * Index is by instance.
         */
        std::vector<double> _ins_log_alpha;

        /** 
         * @brief Category parameter acceptance log probability of the most
         * recent run.
         */
        double _cat_log_alpha;

        /** 
         * @brief Whether the most recent move run should be rejected outright.
         */
        bool _reject;

        /** 
         * @brief The number of object components in a category.
         */
        uint32_t _num_obj_components;

        /** @brief Probability of the paired, balanced move of this move. */
        double _pair_prob;

        /** @brief Variable parameter means for the view. */
        jwsc::Vector_d* _view_means;

        /** @brief Variable parameter stdevs for the view. */
        jwsc::Vector_d* _view_stdevs;

        /** @brief Variable parameter means for the object. */
        jwsc::Vector_d* _obj_means;

        /** @brief Variable parameter stdevs for the object. */
        jwsc::Vector_d* _obj_stdevs;
};


/**
 * @class Metropolis::Birth
 *
 * @brief Proposes a adding a component to the Object in Metropolis::Parameters.
 */
class Birth : public Jump
{
    public:

        /** @brief Constructs a birth move. */
        Birth(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Deletes a birth move. */
        virtual ~Birth();


    protected:

        /** @brief Initializes the move in a managed thread. */
        virtual void init(Metropolis::Thread_manager* mgr);

        /** @brief Returns the index of the block to attach to. */
        uint32_t attach_i();

        /** @brief Returns the block face attachment point index to use. */
        uint32_t attach_pt_i();

        /** @brief Returns the type of attachment point to use. */
        Component::Attach_type attach_pt_t();

        /** @brief Variable parameter means for the birthed component. */
        const jwsc::Vector_d* means() const;

        /** @brief Variable parameter stdevs for the birthed component. */
        const jwsc::Vector_d* stdevs() const;

        /** @brief Variable parameter means for the birthed component. */
        const jwsc::Vector_d* pt_means() const;

        /** @brief Variable parameter stdevs for the birthed component. */
        const jwsc::Vector_d* pt_stdevs() const;

        void sample_var_param_stats
        (
            Metropolis::Thread_manager* mgr,
            uint32_t                    i,
            const Abstract_block*       b1, 
            const Abstract_block*       b2
        );


    protected:

        /** @brief The index of the block to attach to. */
        uint32_t _attach_i;

        /** @brief The block face attachment point index to use. */
        uint32_t _attach_pt_i;

        /** @brief The type of attachment point to use. */
        Component::Attach_type _attach_pt_t;

        /** @brief Variable parameter means for the birthed component. */
        jwsc::Vector_d* _means;

        /** @brief Variable parameter stdevs for the birthed component. */
        jwsc::Vector_d* _stdevs;

        /** @brief Variable parameter means for the birthed component. */
        jwsc::Vector_d* _pt_means;

        /** @brief Variable parameter stdevs for the birthed component. */
        jwsc::Vector_d* _pt_stdevs;
};


/**
 * @class Metropolis::Block_birth
 *
 * @brief Proposes a adding a block to the object in Metropolis::Parameters.
 */
class Block_birth : public Birth
{
    public:

        /** @brief Constructs a birth move. */
        Block_birth(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Block_birth* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the birth jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);
};


/**
 * @class Metropolis::Block_to_block_birth
 *
 * @brief Proposes a adding a block to another in Metropolis::Parameters.
 */
class Block_to_block_birth : public Birth
{
    public:

        /** @brief Constructs a birth move. */
        Block_to_block_birth(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Block_to_block_birth* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the birth jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);
};


/**
 * @class Metropolis::Pair_birth
 *
 * @brief Proposes a adding a block pair to the object in
 * Metropolis::Parameters.
 */
class Pair_birth : public Birth
{
    public:

        /** @brief Constructs a birth move. */
        Pair_birth(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Pair_birth* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the birth jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);

        /** @brief Returns the type of pair layout use. */
        Pair::Layout layout();


    protected:

        /** @brief The type of pair layout to use. */
        Pair::Layout _layout;
};


/**
 * @class Metropolis::Pair_to_block_birth
 *
 * @brief Proposes a adding a block pair to a block in Metropolis::Parameters.
 */
class Pair_to_block_birth : public Birth
{
    public:

        /** @brief Constructs a birth move. */
        Pair_to_block_birth
        (
             const char* name, 
             double      p, 
             double      pair_p,
             bool        height_only
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Pair_to_block_birth* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the birth jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);

        /** @brief Returns the type of pair layout use. */
        Pair::Layout layout();


    private:

        /** @brief The type of pair layout to use. */
        Pair::Layout _layout;

        bool _height_only;
};


/**
 * @class Metropolis::Block_to_pair_birth
 *
 * @brief Proposes a adding a block pair to a block in Metropolis::Parameters.
 */
class Block_to_pair_birth : public Birth
{
    public:

        /** @brief Constructs a birth move. */
        Block_to_pair_birth
        (
            const char* name, 
            double      p, 
            double      pair_p, 
            bool        height_only
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Block_to_pair_birth* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the birth jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);


    private:

        bool _height_only;
};


/**
 * @class Metropolis::Quartet_birth
 *
 * @brief Proposes a adding a block pair to the object in
 * Metropolis::Parameters.
 */
class Quartet_birth : public Birth
{
    public:

        /** @brief Constructs a birth move. */
        Quartet_birth(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Quartet_birth* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the birth jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);
};


/**
 * @class Metropolis::Quartet_to_block_birth
 *
 * @brief Proposes a adding a block quartet to a block in
 * Metropolis::Parameters.
 */
class Quartet_to_block_birth : public Birth
{
    public:

        /** @brief Constructs a birth move. */
        Quartet_to_block_birth
        (
            const char* name, 
            double      p, 
            double      pair_p,
            bool        height_only
        ) 
        throw (jwscxx::base::Arg_error);


        /** @brief Clones a Move. */
        Metropolis::Quartet_to_block_birth* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the birth jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);


    private:

        bool _height_only;
};


/**
 * @class Metropolis::Block_to_quartet_birth
 *
 * @brief Proposes a adding a block quartet to a block in
 * Metropolis::Parameters.
 */
class Block_to_quartet_birth : public Birth
{
    public:

        /** @brief Constructs a birth move. */
        Block_to_quartet_birth
        (
            const char* name, 
            double      p, 
            double      pair_p,
            bool        height_only
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Clones a Move. */
        Metropolis::Block_to_quartet_birth* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the birth jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);


    private:

        bool _height_only;
};


/**
 * @class Metropolis::Cross_birth
 *
 * @brief Proposes a adding a set of cross blocks to a block set.
 */
class Cross_birth : public Birth
{
    public:

        /** @brief Constructs a birth move. */
        Cross_birth
        (
            const char* name, 
            double      p, 
            double      pair_p
        )
        throw (jwscxx::base::Arg_error);


        /** @brief Clones a Move. */
        Metropolis::Cross_birth* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        void sample_var_param_stats
        (
            Metropolis::Thread_manager* mgr,
            uint32_t                    i,
            const Block_set*            b
        );

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the birth jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);

        bool one_cross();


    private:

        bool _one_cross;
};


/**
 * @class Metropolis::Death
 *
 * @brief Proposes a removing a component from the Object in
 * Metropolis::Parameters.
 */
class Death : public Jump
{
    public:

        /** @brief Constructs a death move. */
        Death(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Deletes a death move. */
        virtual ~Death();


    protected:

        /** @brief Initializes the move in a managed thread. */
        virtual void init(Metropolis::Thread_manager* mgr);

        /** @brief Returns the index of the block to detach. */
        uint32_t detach_i();

        /** @brief Returns the block face attach point index to detach from. */
        uint32_t detach_pt_i();

        /** 
         * @brief Returns the sum of a subset of the largest acceptance log
         * probabilities across instances.
         */
        double sum_n_best_instance_log_alphas
        (
            Metropolis::Thread_manager* mgr
        );


    protected:

        /** @brief The index of the block to detach. */
        uint32_t _detach_i;

        /** @brief The block face attachment point index to detach from. */
        uint32_t _detach_pt_i;
};


/**
 * @class Metropolis::Block_death
 *
 * @brief Proposes a removing a block from the object in Metropolis::Parameters.
 */
class Block_death : public Death
{
    public:

        /** @brief Constructs a death move. */
        Block_death(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Block_death* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);
};


/**
 * @class Metropolis::Block_from_block_death
 *
 * @brief Proposes a removing a block from another in Metropolis::Parameters.
 */
class Block_from_block_death : public Death
{
    public:

        /** @brief Constructs a death move. */
        Block_from_block_death(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Block_from_block_death* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);
};


/**
 * @class Metropolis::Pair_death
 *
 * @brief Proposes a removing a block pair from the object in
 * Metropolis::Parameters.
 */
class Pair_death : public Death
{
    public:

        /** @brief Constructs a death move. */
        Pair_death(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Pair_death* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);

        /** @brief Returns the type of pair layout use. */
        Pair::Layout layout();


    protected:

        /** @brief The type of pair layout to use. */
        Pair::Layout _layout;
};


/**
 * @class Metropolis::Pair_from_block_death
 *
 * @brief Proposes a removing a block pair from a block in
 * Metropolis::Parameters.
 */
class Pair_from_block_death : public Death
{
    public:

        /** @brief Constructs a death move. */
        Pair_from_block_death(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Pair_from_block_death* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);

        /** @brief Returns the type of pair layout use. */
        Pair::Layout layout();


    protected:

        /** @brief The type of pair layout to use. */
        Pair::Layout _layout;
};


/**
 * @class Metropolis::Block_from_pair_death
 *
 * @brief Proposes a removing a block pair from a block in
 * Metropolis::Parameters.
 */
class Block_from_pair_death : public Death
{
    public:

        /** @brief Constructs a death move. */
        Block_from_pair_death(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Block_from_pair_death* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);
};


/**
 * @class Metropolis::Quartet_death
 *
 * @brief Proposes a removing a block quartet from the object in
 * Metropolis::Parameters.
 */
class Quartet_death : public Death
{
    public:

        /** @brief Constructs a death move. */
        Quartet_death(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Quartet_death* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);
};


/**
 * @class Metropolis::Quartet_from_block_death
 *
 * @brief Proposes a removing a block quartet from a block in
 * Metropolis::Parameters.
 */
class Quartet_from_block_death : public Death
{
    public:

        /** @brief Constructs a death move. */
        Quartet_from_block_death(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Quartet_from_block_death* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);
};


/**
 * @class Metropolis::Block_from_quartet_death
 *
 * @brief Proposes a removing a block quartet from a block in
 * Metropolis::Parameters.
 */
class Block_from_quartet_death : public Death
{
    public:

        /** @brief Constructs a death move. */
        Block_from_quartet_death(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Block_from_quartet_death* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);
};


/**
 * @class Metropolis::Cross_death
 *
 * @brief Proposes a removing a set of cross blocks from a block set.
 */
class Cross_death : public Death
{
    public:

        /** @brief Constructs a death move. */
        Cross_death(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Cross_death* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);

        bool one_cross();


    private:

        bool _one_cross;
};


/**
 * @class Metropolis::Split
 *
 * @brief Proposes a splitting a component into another in the Object in
 * Metropolis::Parameters.
 */
class Split : public Jump
{
    public:

        /** @brief Constructs a split move. */
        Split(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Deletes a split move. */
        virtual ~Split();


    protected:

        /** @brief Initializes the move in a managed thread. */
        virtual void init(Metropolis::Thread_manager* mgr);

        /** @brief Returns the index of the block to split. */
        uint32_t split_i();

        /** @brief Returns the block face attachment point index to use. */
        uint32_t split_pt_1_i();

        /** @brief Returns the block face attachment point index to use. */
        uint32_t split_pt_2_i();

        /** @brief Returns the type of attachment point to use. */
        Component::Attach_type split_pt_1_t();

        /** @brief Returns the type of attachment point to use. */
        Component::Attach_type split_pt_2_t();

        double split_1_r();

        double split_2_r();

        const std::vector<uint32_t>& split_1_a();

        const std::vector<uint32_t>& split_2_a();

        /** @brief Returns the index of the block to use for re-positioning. */
        uint32_t c_posn_i();

        /** @brief Variable parameter means for the birthed component. */
        const jwsc::Vector_d* means_1() const;

        /** @brief Variable parameter stdevs for the birthed component. */
        const jwsc::Vector_d* stdevs_1() const;

        /** @brief Variable parameter means for the birthed component. */
        const jwsc::Vector_d* means_2() const;

        /** @brief Variable parameter stdevs for the birthed component. */
        const jwsc::Vector_d* stdevs_2() const;

        /** @brief Variable parameter means for the birthed component. */
        const jwsc::Vector_d* pt_means() const;

        /** @brief Variable parameter stdevs for the birthed component. */
        const jwsc::Vector_d* pt_stdevs() const;

        void sample_var_param_stats
        (
            Metropolis::Thread_manager* mgr,
            uint32_t                    i,
            const jwsc::Vector_d*       bs_m,
            const jwsc::Vector_d*       bs_s,
            const Abstract_block*       b1, 
            const Abstract_block*       b2
        );


    protected:

        /** @brief The index of the block to split. */
        uint32_t _split_i;

        /** @brief The block face attachment point index to use. */
        uint32_t _split_pt_1_i;

        /** @brief The block face attachment point index to use. */
        uint32_t _split_pt_2_i;

        /** @brief The type of attachment point to use. */
        Component::Attach_type _split_pt_1_t;

        /** @brief The type of attachment point to use. */
        Component::Attach_type _split_pt_2_t;

        double _split_1_r;

        double _split_2_r;

        std::vector<uint32_t> _split_1_a;

        std::vector<uint32_t> _split_2_a;

        /** @brief The index of the block to use for re-positioning. */
        uint32_t _c_posn_i;

        /** @brief Variable parameter means for the birthed component. */
        jwsc::Vector_d* _means_1;

        /** @brief Variable parameter stdevs for the birthed component. */
        jwsc::Vector_d* _stdevs_1;

        /** @brief Variable parameter means for the birthed component. */
        jwsc::Vector_d* _means_2;

        /** @brief Variable parameter stdevs for the birthed component. */
        jwsc::Vector_d* _stdevs_2;

        /** @brief Variable parameter means for the birthed component. */
        jwsc::Vector_d* _pt_means;

        /** @brief Variable parameter stdevs for the birthed component. */
        jwsc::Vector_d* _pt_stdevs;
};


/**
 * @class Metropolis::Block_to_block_split
 *
 * @brief Proposes a splitting a block into another in Metropolis::Parameters.
 */
class Block_to_block_split : public Split
{
    public:

        /** @brief Constructs a split move. */
        Block_to_block_split(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Block_to_block_split* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the split jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);
};


/**
 * @class Metropolis::Pair_to_block_split
 *
 * @brief Proposes a splitting a block into another in Metropolis::Parameters.
 */
class Pair_to_block_split : public Split
{
    public:

        /** @brief Constructs a split move. */
        Pair_to_block_split
        (
            const char* name, 
            double      p, 
            double      pair_p,
            bool        height_only
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Pair_to_block_split* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the split jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);

        /** @brief Returns the type of pair layout use. */
        Pair::Layout layout();


    private:

        /** @brief The type of pair layout to use. */
        Pair::Layout _layout;

        bool _height_only;
};


/**
 * @class Metropolis::Block_to_pair_split
 *
 * @brief Proposes a splitting a block into another in Metropolis::Parameters.
 */
class Block_to_pair_split : public Split
{
    public:

        /** @brief Constructs a split move. */
        Block_to_pair_split(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Block_to_pair_split* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the split jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);
};


/**
 * @class Metropolis::Quartet_to_block_split
 *
 * @brief Proposes a splitting a block into another in Metropolis::Parameters.
 */
class Quartet_to_block_split : public Split
{
    public:

        /** @brief Constructs a split move. */
        Quartet_to_block_split
        (
            const char* name, 
            double      p, 
            double      pair_p,
            bool        height_only
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Quartet_to_block_split* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the split jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);


    private:

        bool _height_only;
};


/**
 * @class Metropolis::Block_to_quartet_split
 *
 * @brief Proposes a splitting a block into another in Metropolis::Parameters.
 */
class Block_to_quartet_split : public Split
{
    public:

        /** @brief Constructs a split move. */
        Block_to_quartet_split(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Block_to_quartet_split* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs the split jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);
};


/**
 * @class Metropolis::Merge
 *
 * @brief Proposes a merging a component with another from the Object in
 * Metropolis::Parameters.
 */
class Merge : public Jump
{
    public:

        /** @brief Constructs a merge move. */
        Merge(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Deletes a merge move. */
        virtual ~Merge();


    protected:

        /** @brief Initializes the move in a managed thread. */
        virtual void init(Metropolis::Thread_manager* mgr);

        /** @brief Returns the index of the block to merge. */
        uint32_t merge_1_i();

        /** @brief Returns the index of the block to merge. */
        uint32_t merge_2_i();

        /** @brief Returns the block face attach point index to merge across. */
        uint32_t merge_pt_1_i();

        /** @brief Returns the block face attach point index to merge across. */
        uint32_t merge_pt_2_i();

        /** @brief Returns the index of the block to use for re-positioning. */
        uint32_t c_posn_i();

        /** @brief Variable parameter means for the birthed component. */
        const jwsc::Vector_d* means() const;

        /** @brief Variable parameter stdevs for the birthed component. */
        const jwsc::Vector_d* stdevs() const;

        void sample_var_param_stats
        (
            Metropolis::Thread_manager* mgr,
            uint32_t                    i,
            const Abstract_block*       bm,
            const Abstract_block*       b1,
            const Abstract_block*       b2
        );

        /** 
         * @brief Returns the sum of a subset of the largest acceptance log
         * probabilities across instances.
         */
        double sum_n_best_instance_log_alphas
        (
            Metropolis::Thread_manager* mgr
        );


    protected:

        /** @brief The index of the first block to merge. */
        uint32_t _merge_1_i;

        /** @brief The index of the second block to merge. */
        uint32_t _merge_2_i;

        /** @brief The block face attachment point index to merge across. */
        uint32_t _merge_pt_1_i;

        /** @brief The block face attachment point index to merge across. */
        uint32_t _merge_pt_2_i;

        /** @brief The index of the block to use for re-positioning. */
        uint32_t _c_posn_i;

        /** @brief Variable parameter means for the birthed component. */
        jwsc::Vector_d* _means;

        /** @brief Variable parameter stdevs for the birthed component. */
        jwsc::Vector_d* _stdevs;
};


/**
 * @class Metropolis::Block_from_block_merge
 *
 * @brief Proposes a merging a block into another in Metropolis::Parameters.
 */
class Block_from_block_merge : public Merge
{
    public:

        /** @brief Constructs a merge move. */
        Block_from_block_merge(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Block_from_block_merge* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);
};


/**
 * @class Metropolis::Pair_from_block_merge
 *
 * @brief Proposes a merging a block into another in Metropolis::Parameters.
 */
class Pair_from_block_merge : public Merge
{
    public:

        /** @brief Constructs a merge move. */
        Pair_from_block_merge(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Pair_from_block_merge* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);
};


/**
 * @class Metropolis::Block_from_pair_merge
 *
 * @brief Proposes a merging a block into another in Metropolis::Parameters.
 */
class Block_from_pair_merge : public Merge
{
    public:

        /** @brief Constructs a merge move. */
        Block_from_pair_merge
        (
            const char* name, 
            double      p, 
            double      pair_p,
            bool        height_only
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Block_from_pair_merge* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);


    private:

        bool _height_only;
};


/**
 * @class Metropolis::Quartet_from_block_merge
 *
 * @brief Proposes a merging a block into another in Metropolis::Parameters.
 */
class Quartet_from_block_merge : public Merge
{
    public:

        /** @brief Constructs a merge move. */
        Quartet_from_block_merge(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Quartet_from_block_merge* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);
};


/**
 * @class Metropolis::Block_from_quartet_merge
 *
 * @brief Proposes a merging a block into another in Metropolis::Parameters.
 */
class Block_from_quartet_merge : public Merge
{
    public:

        /** @brief Constructs a merge move. */
        Block_from_quartet_merge
        (
            const char* name, 
            double      p, 
            double      pair_p,
            bool        height_only
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Block_from_quartet_merge* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);


    private:

        bool _height_only;
};


/**
 * @class Metropolis::Switch
 *
 * @brief Proposes a merging a component with another from the Object in
 * Metropolis::Parameters.
 */
class Switch : public Jump
{
    public:

        /** @brief Constructs a merge move. */
        Switch(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Deletes a merge move. */
        virtual ~Switch();


    protected:

        /** @brief Initializes the move in a managed thread. */
        virtual void init(Metropolis::Thread_manager* mgr);

        /** @brief Updates the category acceptance log probabilities. */
        void update_cat_log_alpha(Metropolis::Thread_manager* mgr);

        /** @brief Returns the index of the block to switch. */
        uint32_t switch_i();

        /** @brief Variable parameter means for the switched component. */
        const jwsc::Vector_d* means() const;

        /** @brief Variable parameter stdevs for the switched component. */
        const jwsc::Vector_d* stdevs() const;

        void sample_var_param_stats
        (
            Metropolis::Thread_manager* mgr,
            uint32_t                    i,
            const Abstract_block*       b1,
            const Abstract_block*       b2
        );


    protected:

        /** @brief The index of the first block to merge. */
        uint32_t _switch_i;

        /** @brief Variable parameter means for the switched component. */
        jwsc::Vector_d* _means;

        /** @brief Variable parameter stdevs for the switched component. */
        jwsc::Vector_d* _stdevs;
};


/**
 * @class Metropolis::Block_to_pair_switch
 *
 * @brief Proposes a switching a block into a Pair.
 */
class Block_to_pair_switch : public Switch
{
    public:

        /** @brief Constructs a merge move. */
        Block_to_pair_switch
        (
            const char* name, 
            double      p, 
            double      pair_p,
            bool        height_only
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Block_to_pair_switch* clone() const;

        Pair::Layout layout();

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);


    private:

        /** @brief The type of pair layout to use. */
        Pair::Layout _layout;

        bool _height_only;
};


/**
 * @class Metropolis::Pair_to_block_switch
 *
 * @brief Proposes a switching a block into a Pair.
 */
class Pair_to_block_switch : public Switch
{
    public:

        /** @brief Constructs a merge move. */
        Pair_to_block_switch(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Pair_to_block_switch* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);
};


/**
 * @class Metropolis::Block_to_quartet_switch
 *
 * @brief Proposes a switching a block into a Quartet.
 */
class Block_to_quartet_switch : public Switch
{
    public:

        /** @brief Constructs a merge move. */
        Block_to_quartet_switch
        (
            const char* name, 
            double      p, 
            double      pair_p,
            bool        height_only
        ) 
        throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Block_to_quartet_switch* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);


    private:

        bool _height_only;
};


/**
 * @class Metropolis::Quartet_to_block_switch
 *
 * @brief Proposes a switching a block into a Quartet.
 */
class Quartet_to_block_switch : public Switch
{
    public:

        /** @brief Constructs a merge move. */
        Quartet_to_block_switch(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Quartet_to_block_switch* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);
};


/**
 * @class Metropolis::Pair_to_quartet_switch
 *
 * @brief Proposes a switching a Pair into a Quartet.
 */
class Pair_to_quartet_switch : public Switch
{
    public:

        /** @brief Constructs a merge move. */
        Pair_to_quartet_switch(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Pair_to_quartet_switch* clone() const;

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        void sample_var_param_stats
        (
            Metropolis::Thread_manager* mgr,
            uint32_t                    i,
            const Pair*                 b1,
            const Quartet*              b2
        );
};


/**
 * @class Metropolis::Quartet_to_pair_switch
 *
 * @brief Proposes a switching a Pair into a Quartet.
 */
class Quartet_to_pair_switch : public Switch
{
    public:

        /** @brief Constructs a merge move. */
        Quartet_to_pair_switch(const char* name, double p, double pair_p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Clones a Move. */
        Metropolis::Quartet_to_pair_switch* clone() const;

        Pair::Layout layout();

        /** 
         * @brief Returns whether a set of object components is supported by
         * this move.
         */
        bool has_supported_components
        (
            const std::vector<Component*>& components
        ) 
        const;


    protected:

        /** @brief Initializes the move in a managed thread. */
        void init(Metropolis::Thread_manager* mgr);

        /** @brief Runs death jump move. */
        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        void sample_var_param_stats
        (
            Metropolis::Thread_manager* mgr,
            uint32_t                    i,
            const Quartet*              b1,
            const Pair*                 b2
        );


    protected:

        /** @brief The type of pair layout to use. */
        Pair::Layout _layout;
};


/**
 * @class Metropolis::Jump_pair
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Jump_pair
{
    public:

        /** @brief Constructs a jump pair. */
        Jump_pair
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Deletes a jump pair. */
        virtual ~Jump_pair();

        /** @brief Creates the first jump move in the pair. */
        virtual const Jump& first() const = 0;

        /** @brief Creates the first jump move in the pair. */
        virtual const Jump& second() const = 0;


    protected:

        /** @brief Single word name of the 1st balanced jump move. */
        std::string _name_1;

        /** @brief Probability of the 1st balanced jump move in a Sampler. */
        double _prob_1;

        /** @brief Single word name of the 2nd balanced jump move. */
        std::string _name_2;

        /** @brief Probability of the 2nd balanced jump move in a Sampler. */
        double _prob_2;
};


/**
 * @class Metropolis::Birth_death
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Birth_death : public Jump_pair
{
    public:

        /** @brief Constructs a jump pair. */
        Birth_death
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Deletes a jump pair. */
        virtual ~Birth_death();

        /** @brief Creates the first jump move in the pair. */
        virtual const Birth& first() const = 0;

        /** @brief Creates the first jump move in the pair. */
        virtual const Death& second() const = 0;
};


/**
 * @class Metropolis::Block_birth_death
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Block_birth_death : public Birth_death
{
    public:

        /** @brief Constructs a jump pair. */
        Block_birth_death
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_birth& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_death& second() const;


    private:

        Block_birth _first;

        Block_death _second;
};


/**
 * @class Metropolis::Block_block_birth_death
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Block_block_birth_death : public Birth_death
{
    public:

        /** @brief Constructs a jump pair. */
        Block_block_birth_death
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_to_block_birth& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_from_block_death& second() const;


    private:

        Block_to_block_birth _first;

        Block_from_block_death _second;
};


/**
 * @class Metropolis::Pair_birth_death
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Pair_birth_death : public Birth_death
{
    public:

        /** @brief Constructs a jump pair. */
        Pair_birth_death
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Pair_birth& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Pair_death& second() const;


    private:

        Pair_birth _first;

        Pair_death _second;
};


/**
 * @class Metropolis::Pair_block_birth_death
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Pair_block_birth_death : public Birth_death
{
    public:

        /** @brief Constructs a jump pair. */
        Pair_block_birth_death
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Pair_to_block_birth& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Pair_from_block_death& second() const;


    private:

        Pair_to_block_birth _first;

        Pair_from_block_death _second;
};


/**
 * @class Metropolis::Block_pair_birth_death
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Block_pair_birth_death : public Birth_death
{
    public:

        /** @brief Constructs a jump pair. */
        Block_pair_birth_death
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_to_pair_birth& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_from_pair_death& second() const;


    private:

        Block_to_pair_birth _first;

        Block_from_pair_death _second;
};


/**
 * @class Metropolis::Quartet_birth_death
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Quartet_birth_death : public Birth_death
{
    public:

        /** @brief Constructs a jump pair. */
        Quartet_birth_death
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Quartet_birth& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Quartet_death& second() const;


    private:

        Quartet_birth _first;

        Quartet_death _second;
};


/**
 * @class Metropolis::Quartet_block_birth_death
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Quartet_block_birth_death : public Birth_death
{
    public:

        /** @brief Constructs a jump pair. */
        Quartet_block_birth_death
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Quartet_to_block_birth& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Quartet_from_block_death& second() const;


    private:

        Quartet_to_block_birth _first;

        Quartet_from_block_death _second;
};


/**
 * @class Metropolis::Block_quartet_birth_death
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Block_quartet_birth_death : public Birth_death
{
    public:

        /** @brief Constructs a jump pair. */
        Block_quartet_birth_death
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_to_quartet_birth& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_from_quartet_death& second() const;


    private:

        Block_to_quartet_birth _first;

        Block_from_quartet_death _second;
};


/**
 * @class Metropolis::Cross_birth_death
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Cross_birth_death : public Birth_death
{
    public:

        /** @brief Constructs a jump pair. */
        Cross_birth_death
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Cross_birth& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Cross_death& second() const;


    private:

        Cross_birth _first;

        Cross_death _second;
};


/**
 * @class Metropolis::Split_merge
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Split_merge : public Jump_pair
{
    public:

        /** @brief Constructs a jump pair. */
        Split_merge
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Deletes a jump pair. */
        virtual ~Split_merge();

        /** @brief Creates the first jump move in the pair. */
        virtual const Split& first() const = 0;

        /** @brief Creates the first jump move in the pair. */
        virtual const Merge& second() const = 0;
};


/**
 * @class Metropolis::Block_block_split_merge
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Block_block_split_merge : public Split_merge
{
    public:

        /** @brief Constructs a jump pair. */
        Block_block_split_merge
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_to_block_split& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_from_block_merge& second() const;


    private:

        Block_to_block_split _first;

        Block_from_block_merge _second;
};


/**
 * @class Metropolis::Pair_block_split_merge
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Pair_block_split_merge : public Split_merge
{
    public:

        /** @brief Constructs a jump pair. */
        Pair_block_split_merge
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Pair_to_block_split& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Pair_from_block_merge& second() const;


    private:

        Pair_to_block_split _first;

        Pair_from_block_merge _second;
};


/**
 * @class Metropolis::Block_pair_split_merge
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Block_pair_split_merge : public Split_merge
{
    public:

        /** @brief Constructs a jump pair. */
        Block_pair_split_merge
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_to_pair_split& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_from_pair_merge& second() const;


    private:

        Block_to_pair_split _first;

        Block_from_pair_merge _second;
};


/**
 * @class Metropolis::Quartet_block_split_merge
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Quartet_block_split_merge : public Split_merge
{
    public:

        /** @brief Constructs a jump pair. */
        Quartet_block_split_merge
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Quartet_to_block_split& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Quartet_from_block_merge& second() const;


    private:

        Quartet_to_block_split _first;

        Quartet_from_block_merge _second;
};


/**
 * @class Metropolis::Block_quartet_split_merge
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Block_quartet_split_merge : public Split_merge
{
    public:

        /** @brief Constructs a jump pair. */
        Block_quartet_split_merge
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_to_quartet_split& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_from_quartet_merge& second() const;


    private:

        Block_to_quartet_split _first;

        Block_from_quartet_merge _second;
};


/**
 * @class Metropolis::Switch_pair
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Switch_pair : public Jump_pair
{
    public:

        /** @brief Constructs a jump pair. */
        Switch_pair
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Deletes a jump pair. */
        virtual ~Switch_pair();

        /** @brief Creates the first jump move in the pair. */
        virtual const Switch& first() const = 0;

        /** @brief Creates the first jump move in the pair. */
        virtual const Switch& second() const = 0;
};


/**
 * @class Metropolis::Block_pair_switch
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Block_pair_switch : public Switch_pair
{
    public:

        /** @brief Constructs a jump pair. */
        Block_pair_switch
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_to_pair_switch& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Pair_to_block_switch& second() const;


    private:

        Block_to_pair_switch _first;

        Pair_to_block_switch _second;
};


/**
 * @class Metropolis::Block_quartet_switch
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Block_quartet_switch : public Switch_pair
{
    public:

        /** @brief Constructs a jump pair. */
        Block_quartet_switch
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Block_to_quartet_switch& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Quartet_to_block_switch& second() const;


    private:

        Block_to_quartet_switch _first;

        Quartet_to_block_switch _second;
};


/**
 * @class Metropolis::Pair_quartet_switch
 *
 * Pair of MCMC sampler jump moves that modify the number model parameter.
 */
class Pair_quartet_switch : public Switch_pair
{
    public:

        /** @brief Constructs a jump pair. */
        Pair_quartet_switch
        (
            const char* name_1, 
            const char* name_2,
            double      p_1, 
            double      p_2
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Pair_to_quartet_switch& first() const;

        /** @brief Creates the first jump move in the pair. */
        const Metropolis::Quartet_to_pair_switch& second() const;


    private:

        Pair_to_quartet_switch _first;

        Quartet_to_pair_switch _second;
};


    // Metropolis
    public:

        /** @brief Constructs an MCMC sampler. */
        Metropolis
        (
            double   diff_prob,
            double   jump_prob,
            uint32_t jump_set_size
        );

        /** @brief Deletes an MCMC Sampler and all its moves. */
        virtual ~Metropolis();

        /** 
         * @brief Returns the number of model sets used in computing the
         * accceptance probability of a jump move.
         */
        uint32_t jump_set_size() const;

        /** @brief Returns the sampler moves. */
        std::list<Metropolis::Move*> moves();

        /** @brief Returns the sampler diffusion moves. */
        std::list<Metropolis::Move*> diffusion_moves();

        /** @brief Returns the sampler jump moves. */
        std::list<Metropolis::Move*> jump_moves();

        /** @brief Adds a sampler diffusion move. */
        void add_diffusion(const Metropolis::Diffusion& move);

        /** @brief Adds a sampler jump move. */
        void add_jump(const Metropolis::Jump_pair& move);

        /** @brief Diffusion move probability. */
        double diffusion_prob() const;

        /** @brief Jump move probability. */
        double jump_prob() const;

        /** @brief Runs the iterative MCMC sampler. */
        void run(Metropolis::Parameters* params, uint32_t iterations)
            throw (jwscxx::base::Exception);


    protected:

        /** @brief Runs the iterative MCMC sampler. */
        void run(Sampler::Parameters* params, uint32_t iterations)
            throw (jwscxx::base::Exception);

        void run_thread(Metropolis::Thread_manager* mgr)
            throw (jwscxx::base::Exception);

        static jwscxx::base::Exception* run_thread_c
        (
            Metropolis::Thread_manager* mgr
        );

        /** @brief List of diffusion moves to use in the sampler. */
        std::list<Metropolis::Move*> _diff_moves;

        /** @brief List of jump moves to use in the sampler. */
        std::list<Metropolis::Move*> _jump_moves;

        /** @brief Probability of a diffusion move. */
        double _diff_prob;

        /** @brief Probability of a jump move. */
        double _jump_prob;

        /** 
         * @brief Number of model instances used in computing the acceptance
         * probability for a jump move.
         */
        uint32_t _jump_set_size;
};


#endif
