/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Parametric generative model.
 *
 * Specialed for inference with MCMC sampling techniques.
 */


#include <config.h>

#include <cstdlib>
#include <cmath>
#include <cassert>
#include <cstring>
#include <iostream>
#include <sstream>
#include <vector>
#include <typeinfo>
#include <algorithm>

#include <jwsc/base/limits.h>
#include <jwsc/prob/pdf.h>
#include <jwsc/vector/vector.h>
#include <jwsc/vector/vector_math.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/base/cloneable.h>
#include <jwsc++/base/readable.h>
#include <jwsc++/base/writeable.h>

#include "model.h"


#define  TRUNC_TRY_LIMIT  100


/* -----------------------------  Model::Context  --------------------------- */

Model::Context::Context()
{
}

Model::Context::Context(const Model::Context& ctx)
{
    *this = ctx;
}

Model::Context::~Context()
{
}

Model::Context& Model::Context::operator=(const Model::Context& ctx)
{
    return *this;
}

/* ========================================================================== */








/* -------------------  Model::Parameter::Context::Range  ------------------- */

Model::Parameter::Context::Range::Range(double min, double max) 
    throw (jwscxx::base::Arg_error) 
{ 
    set(min, max);
}

Model::Parameter::Context::Range* 
Model::Parameter::Context::Range::clone() const
{
    return new Range(_min, _max);
}

double Model::Parameter::Context::Range::min() const 
{ 
    return _min; 
}

double Model::Parameter::Context::Range::max() const 
{ 
    return _max; 
}

void Model::Parameter::Context::Range::set(double min, double max) 
    throw (jwscxx::base::Arg_error)
{
    if (min > max)
    {
        throw jwscxx::base::Arg_error("Minimum larger than maximum");
    }
    _min = min;
    _max = max;
}

void Model::Parameter::Context::Range::test(double value) const 
    throw (jwscxx::base::Arg_error)
{
    if (_min > value || _max < value)
    {
        std::ostringstream ost;
        ost << value << " outside range [" << _min << "," << _max << "]";
        throw jwscxx::base::Arg_error(ost.str());
    }
}

/* ========================================================================== */








/* ------------------  Model::Parameter::Context::Density  ------------------ */

/**
 * If the value @em x is outside the truncated region [a,b], the density log of
 * the density is a very small negative number; the pdf evaluates extremely
 * close to zero, which should not be attempted because precision will not
 * support it.
 *
 * @param  x    Value to evaluate under the pdf.
 * @param  min  Minimum value of the truncated pdf.
 * @param  max  Maximum value of the truncated pdf.
 */
double Model::Parameter::Context::Density::trunc_log_pdf
(
    double x, 
    double min, 
    double max
) 
const
{
    if (x < min || x > max)
    {
        double y = ((x < min) ? fabs(x - min) : fabs(x - max)) / (max - min);
        return (y+1)*10*log(JWSC_MIN_LOG_ARG);
    }

    return log_pdf(x);
}

/* ========================================================================== */







/* ------------------  Model::Parameter::Context::Uniform  ------------------ */

Model::Parameter::Context::Uniform* 
Model::Parameter::Context::Uniform::clone() const
{
    return new Uniform();
}

double Model::Parameter::Context::Uniform::pdf(double x) const
{ 
    return 1;
}

double Model::Parameter::Context::Uniform::log_pdf(double x) const
{ 
    return 0;
}

double Model::Parameter::Context::Uniform::sample
(
    const Model::Parameter::Context::Range* range
)
const
{ 
    double min = range->min() + 1.0e-6;
    double max = range->max() - 1.0e-6;

    if (max <= min)
    {
        return 0.5*(range->min() + range->max());
    }

    return jwsc::sample_uniform_pdf_d(min, max);
}

/* ========================================================================== */






/* ------------------  Model::Parameter::Context::Gaussian  ----------------- */

Model::Parameter::Context::Gaussian::Gaussian(double mu, double sigma) 
    throw (jwscxx::base::Arg_error) 
{ 
    set(mu, sigma);
}

Model::Parameter::Context::Gaussian* 
Model::Parameter::Context::Gaussian::clone() const
{
    return new Gaussian(_mu, _sigma);
}

double Model::Parameter::Context::Gaussian::mu() const 
{ 
    return _mu; 
}

double Model::Parameter::Context::Gaussian::sigma() const 
{ 
    return _sigma; 
}

void Model::Parameter::Context::Gaussian::set(double mu, double sigma)
    throw (jwscxx::base::Arg_error)
{
    if (sigma < 1.0e-8)
    {
        throw jwscxx::base::Arg_error("Sigma too small");
    }
    _mu = mu; 
    _sigma = sigma;
}

double Model::Parameter::Context::Gaussian::pdf(double x) const
{ 
    return jwsc::gaussian_pdf_d(_mu, _sigma, x); 
}

double Model::Parameter::Context::Gaussian::pdf
(
    double x, 
    double mu, 
    double sigma
) 
const
{ 
    return jwsc::gaussian_pdf_d(mu, sigma, x); 
}

double Model::Parameter::Context::Gaussian::log_pdf(double x) const
{ 
    return jwsc::log_gaussian_pdf_d(_mu, _sigma, x); 
}

double Model::Parameter::Context::Gaussian::log_pdf
(
    double x, 
    double mu,
    double sigma
) 
const
{ 
    return jwsc::log_gaussian_pdf_d(mu, sigma, x); 
}

/**
 * If the value @em x is outside the truncated region [a,b], the density log of
 * the density is a very small negative number; the pdf evaluates extremely
 * close to zero, which should not be attempted because precision will not
 * support it.
 *
 * @param  x    Value to evaluate under the pdf.
 * @param  min  Minimum value of the truncated pdf.
 * @param  max  Maximum value of the truncated pdf.
 */
double Model::Parameter::Context::Gaussian::trunc_log_pdf
(
    double x, 
    double min, 
    double max
) 
const
{
    if (x < min || x > max)
    {
        double y = ((x < min) ? fabs(x - min) : fabs(x - max)) / (max - min);
        return (y+1)*10*log(JWSC_MIN_LOG_ARG);
    }

    return log_pdf(x);
}

/**
 * If the value @em x is outside the truncated region [a,b], the density log of
 * the density is a very small negative number; the pdf evaluates extremely
 * close to zero, which should not be attempted because precision will not
 * support it.
 *
 * @param  x      Value to evaluate under the pdf.
 * @param  mu     Gaussian mean.
 * @param  sigma  Gaussian standard deviation.
 * @param  min    Minimum value of the truncated pdf.
 * @param  max    Maximum value of the truncated pdf.
 */
double Model::Parameter::Context::Gaussian::trunc_log_pdf
(
    double x, 
    double mu,
    double sigma,
    double min, 
    double max
) 
const
{
    if (x < min || x > max)
    {
        double y = ((x < min) ? fabs(x - min) : fabs(x - max)) / (max - min);
        return (y+1)*10*log(JWSC_MIN_LOG_ARG);
    }

    return log_pdf(x, mu, sigma);
}

double Model::Parameter::Context::Gaussian::sample
(
    const Model::Parameter::Context::Range* range
) 
const
{ 
    return sample(_mu, _sigma, range);
}

double Model::Parameter::Context::Gaussian::sample
(
    double                                  mu, 
    double                                  sigma,
    const Model::Parameter::Context::Range* range
)
const
{ 
    double min = range->min() + 1.0e-6;
    double max = range->max() - 1.0e-6;

    if (max <= min)
    {
        return 0.5*(range->min() + range->max());
    }

    uint32_t N = TRUNC_TRY_LIMIT;
    double x = min - 1;

    while ((N > 0) && (x < min || x > max))
    {
        x = mu + jwsc::sample_gaussian_pdf_d(0, sigma, -6*sigma, 6*sigma);
        N--;
    }

    if (x < min)
    {
        return min;
    }
    if (x > max)
    {
        return max;
    }

    return x;
}

/* ========================================================================== */






/* --------------  Model::Parameter::Context::Circular_gaussian  ------------ */

Model::Parameter::Context::Circular_gaussian::Circular_gaussian
(
    double mu, 
    double sigma
)
throw (jwscxx::base::Arg_error) 
    : Gaussian(mu, sigma)
{ 
}

double Model::Parameter::Context::Circular_gaussian::sample
(
    const Model::Parameter::Context::Range* range
) 
const
{ 
    return sample(_mu, range);
}

double Model::Parameter::Context::Circular_gaussian::sample
(
    double                                  mu,
    const Model::Parameter::Context::Range* range
)
const
{ 
    double rmin = range->min();
    double rmax = range->max();
    double d    = rmax - rmin;
    double min  = rmin + 1.0e-6;
    double max  = rmax - 1.0e-6;

    if (max <= min)
    {
        return 0.5*(rmin + rmax);
    }

    uint32_t N = TRUNC_TRY_LIMIT;
    double x = min - 1;

    while ((N > 0) && (x < min || x > max))
    {
        x = mu + jwsc::sample_gaussian_pdf_d(0, _sigma, -6*_sigma, 6*_sigma);
        if (x > rmax && (x - d) >= rmin)
        {
            x -= d;
        }
        else if (x < rmin && (x + d) <= rmax)
        {
            x += d;
        }
        N--;
    }

    if (x < min)
    {
        return min;
    }
    if (x > max)
    {
        return max;
    }

    return x;
}

/* ========================================================================== */







/* ---------------  Model::Parameter::Context::Gaussian_gamma  -------------- */

Model::Parameter::Context::Gaussian_gamma::Gaussian_gamma
(
    double mu, 
    double lambda,
    double alpha,
    double beta
) 
throw (jwscxx::base::Arg_error) 
{ 
    set(mu, lambda, alpha, beta);
}

Model::Parameter::Context::Gaussian_gamma* 
Model::Parameter::Context::Gaussian_gamma::clone() const
{
    return new Gaussian_gamma(_mu, _lambda, _alpha, _beta);
}

double Model::Parameter::Context::Gaussian_gamma::mu() const 
{ 
    return _mu; 
}

double Model::Parameter::Context::Gaussian_gamma::lambda() const 
{ 
    return _lambda; 
}

double Model::Parameter::Context::Gaussian_gamma::alpha() const 
{ 
    return _alpha; 
}

double Model::Parameter::Context::Gaussian_gamma::beta() const 
{ 
    return _beta; 
}

void Model::Parameter::Context::Gaussian_gamma::set
(
    double mu, 
    double lambda,
    double alpha,
    double beta
)
throw (jwscxx::base::Arg_error)
{
    if (lambda < 1.0e-8)
    {
        throw jwscxx::base::Arg_error("Lambda too small");
    }
    if (alpha < 1.0e-8)
    {
        throw jwscxx::base::Arg_error("Alpha too small");
    }
    if (beta < 1.0e-8)
    {
        throw jwscxx::base::Arg_error("Beta too small");
    }

    _mu     = mu; 
    _lambda = lambda;
    _alpha  = alpha;
    _beta   = beta;
}

double Model::Parameter::Context::Gaussian_gamma::pdf(double mu) const
{
    return pdf(mu, 1);
}

double Model::Parameter::Context::Gaussian_gamma::pdf(double mu, double sigma) 
    const
{
    using namespace jwsc;

    return gaussian_pdf_d(_mu, sqrt(_lambda)*sigma, mu) * 
        gamma_pdf_d(_alpha, _beta, 1.0/(sigma*sigma));
}

double Model::Parameter::Context::Gaussian_gamma::log_pdf(double mu) const
{ 
    return log_pdf(mu, 1);
}

double Model::Parameter::Context::Gaussian_gamma::log_pdf
(
    double mu, 
    double sigma
)
const
{ 
    using namespace jwsc;

    return log_gaussian_pdf_d(_mu, sqrt(_lambda)*sigma, mu) + 
        log_gamma_pdf_d(_alpha, _beta, 1.0/(sigma*sigma));
}

/**
 * If the value @em x is outside the truncated region [a,b], the density log of
 * the density is a very small negative number; the pdf evaluates extremely
 * close to zero, which should not be attempted because precision will not
 * support it.
 *
 * @param  mu      Mean to evaluate under the pdf.
 * @param  mu_min  Minimum value of the truncated pdf.
 * @param  mu_max  Maximum value of the truncated pdf.
 * @param  sigma   Standard deviation to evaluate under the pdf.
 */
double Model::Parameter::Context::Gaussian_gamma::trunc_log_pdf
(
    double mu,
    double mu_min, 
    double mu_max,
    double sigma
) 
const
{
    if (mu < mu_min || mu > mu_max)
    {
        double y = (mu < mu_min) ? fabs(mu - mu_min) : fabs(mu - mu_max);
        y /= (mu_max - mu_min);
        return (y+1)*10*log(JWSC_MIN_LOG_ARG);
    }

    return log_pdf(mu, sigma);
}

double Model::Parameter::Context::Gaussian_gamma::sample
(
    const Model::Parameter::Context::Range* range
)
const
{ 
    double mu;
    Gaussian_gamma::sample(&mu, range);
    return mu;
}

void Model::Parameter::Context::Gaussian_gamma::sample
(
    double*                                 mu_out,
    const Model::Parameter::Context::Range* mu_range
)
const
{ 
    double sigma;
    sample(mu_out, mu_range, &sigma);
}

void Model::Parameter::Context::Gaussian_gamma::sample
(
    double*                                 mu_out,
    const Model::Parameter::Context::Range* mu_range,
    double*                                 sigma_out
)
const
{ 
    sample(mu_out, _mu, mu_range, sigma_out);
}

void Model::Parameter::Context::Gaussian_gamma::sample
(
    double*                                 mu_out,
    double                                  mu_in,
    const Model::Parameter::Context::Range* mu_range,
    double*                                 sigma_out
)
const
{ 
    using namespace jwsc;

    uint32_t N;
    double mu;
    double mu_min = mu_range->min() + 1.0e-6;
    double mu_max = mu_range->max() - 1.0e-6;
    double tau;
    double tau_min = 1.0e-8;
    double tau_max = 6.0*sqrt(_alpha)/_beta;

    if (tau_max <= tau_min)
    {
        tau = 0.5*(tau_min + tau_max);
    }
    else
    {
        N = TRUNC_TRY_LIMIT;
        tau = tau_min - 1;
        while ((N > 0) && (tau < tau_min || tau > tau_max))
        {
            tau = sample_gamma_pdf_d(_alpha, _beta, tau_min, tau_max);
            N--;
        }
        if (tau < tau_min)
        {
            tau = tau_min;
        }
        if (tau > tau_max)
        {
            tau = tau_max;
        }
    }

    if (mu_max <= mu_min)
    {
        mu = 0.5*(mu_range->min() + mu_range->max());
    }
    else
    {
        N = TRUNC_TRY_LIMIT;
        mu = mu_min - 1;
        while ((N > 0) && (mu < mu_min || mu > mu_max))
        {
            mu = mu_in + sample_gaussian_pdf_d(0, sqrt(_lambda/tau),
                    -6*_lambda/tau, 6*_lambda/tau);
            N--;
        }
        if (mu < mu_min)
        {
            mu = mu_min;
        }
        if (mu > mu_max)
        {
            mu = mu_max;
        }
    }

    *mu_out = mu;
    *sigma_out = sqrt(1.0/tau);
}

/* ========================================================================== */






/* -------------------  Model::Parameter::Context::Dynamics  ---------------- */

Model::Parameter::Context::Dynamics::Dynamics(double delta_t, double grad_eta) 
    throw (jwscxx::base::Arg_error) 
{ 
    set(delta_t, grad_eta);
}

Model::Parameter::Context::Dynamics* 
Model::Parameter::Context::Dynamics::clone() const
{
    return new Dynamics(_delta_t, _grad_eta);
}

double Model::Parameter::Context::Dynamics::delta_t() const 
{ 
    return _delta_t; 
}

double Model::Parameter::Context::Dynamics::grad_eta() const 
{ 
    return _grad_eta; 
}

void Model::Parameter::Context::Dynamics::set(double delta_t, double grad_eta) 
    throw (jwscxx::base::Arg_error)
{
    if (delta_t <= 0)
    {
        throw jwscxx::base::Arg_error("delta-t not positive");
    }
    if (grad_eta <= 0)
    {
        throw jwscxx::base::Arg_error("grad-eta not positive");
    }
    _delta_t  = delta_t;
    _grad_eta = grad_eta;
}

/* ========================================================================== */







/* -----------------------  Model::Parameter::Context  ---------------------- */

Model::Parameter::Context::Context()
{
    _range          = new Range();
    _likelihood     = new Gaussian();
    _prior          = new Gaussian_gamma();
    _ratio_prior    = new Gaussian_gamma();
    _proposal       = new Uniform();
    _gauss_proposal = new Gaussian();
    _value_dynamics = new Dynamics();
    _mean_dynamics  = new Dynamics();
    _stdev_dynamics = new Dynamics();
    _ratio_mean_dynamics  = new Dynamics();
    _ratio_stdev_dynamics = new Dynamics();
    _has_ratio_prior      = false;
    
    _init_value = _range->min();
}

Model::Parameter::Context::Context(const Model::Parameter::Context& ctx)
{
    _range          = 0;
    _likelihood     = 0;
    _prior          = 0;
    _ratio_prior    = 0;
    _proposal       = 0;
    _gauss_proposal = 0;
    _value_dynamics = 0;
    _mean_dynamics  = 0;
    _stdev_dynamics = 0;
    _ratio_mean_dynamics  = 0;
    _ratio_stdev_dynamics = 0;
    _has_ratio_prior      = false;

    *this = ctx;
}

Model::Parameter::Context::~Context()
{
    delete _range;
    delete _likelihood;
    delete _prior;
    delete _ratio_prior;
    delete _proposal;
    delete _gauss_proposal;
    delete _value_dynamics;
    delete _mean_dynamics;
    delete _stdev_dynamics;
    delete _ratio_mean_dynamics;
    delete _ratio_stdev_dynamics;
}

/** 
 * Performs a deep copy of this Model::Parameter::Context.
 *
 * @param  ctx  Model::Parameter::Context to copy into this one.
 *
 * @return A reference to this Model::Parameter::Context.
 */
Model::Parameter::Context& 
Model::Parameter::Context::operator=(const Model::Parameter::Context& ctx)
{
    delete _range;
    delete _likelihood;
    delete _prior;
    delete _ratio_prior;
    delete _proposal;
    delete _gauss_proposal;
    delete _value_dynamics;
    delete _mean_dynamics;
    delete _stdev_dynamics;
    delete _ratio_mean_dynamics;
    delete _ratio_stdev_dynamics;

    _init_value     = ctx._init_value;
    _range          = ctx._range->clone();
    _likelihood     = ctx._likelihood->clone();
    _prior          = ctx._prior->clone();
    _ratio_prior    = ctx._ratio_prior->clone();
    _proposal       = ctx._proposal->clone();
    _gauss_proposal = ctx._gauss_proposal->clone();
    _value_dynamics = ctx._value_dynamics->clone();
    _mean_dynamics  = ctx._mean_dynamics->clone();
    _stdev_dynamics = ctx._stdev_dynamics->clone();
    _ratio_mean_dynamics  = ctx._ratio_mean_dynamics->clone();
    _ratio_stdev_dynamics = ctx._ratio_stdev_dynamics->clone();
    _has_ratio_prior      = ctx._has_ratio_prior;

    return *this;
}

/** @return A new copy of this context. */
Model::Parameter::Context* Model::Parameter::Context::clone() const
{
    return new Context(*this);
}

double Model::Parameter::Context::init_value() const 
{ 
    return _init_value; 
}

/** Clips the _init_value against the new range. */
void Model::Parameter::Context::init_value(double value)
{
    if (value < _range->min())
    {
        value = _range->min();
    }
    if (value > _range->max())
    {
        value = _range->max();
    }

    _init_value = value;
}

const Model::Parameter::Context::Range&
Model::Parameter::Context::range() const
{ 
    return *_range; 
}

/** Clips the _init_value against the new range. */
void Model::Parameter::Context::range
(
    const Model::Parameter::Context::Range& range
)
{
    delete _range;
    _range = range.clone();

    if (_init_value < range.min())
    {
        _init_value = range.min();
    }
    if (_init_value > range.max())
    {
        _init_value = range.max();
    }
}

void Model::Parameter::Context::test_range(double x) const 
    throw (jwscxx::base::Arg_error)
{ 
    _range->test(x); 
}

const Model::Parameter::Context::Gaussian&
Model::Parameter::Context::likelihood() const 
{ 
    return *_likelihood; 
}

void Model::Parameter::Context::likelihood
(
    const Model::Parameter::Context::Gaussian& likelihood
)
{ 
    delete _likelihood;
    _likelihood = likelihood.clone();
}

double Model::Parameter::Context::calc_likelihood(double x) const
{ 
    return _likelihood->pdf(x); 
}

double Model::Parameter::Context::calc_likelihood(double x, double s) const
{ 
    double mu = s * _likelihood->mu();
    double sigma = s * _likelihood->sigma();
    return _likelihood->pdf(x, mu, sigma);
}

double Model::Parameter::Context::calc_log_likelihood(double x) const
{ 
    return _likelihood->log_pdf(x); 
}

double Model::Parameter::Context::calc_log_likelihood(double x, double s) const
{ 
    double mu = s * _likelihood->mu();
    double sigma = s * _likelihood->sigma();
    return _likelihood->log_pdf(x, mu, sigma); 
}

/** 
 * If the value is outside the range, the probability is practically zero.
 */
double Model::Parameter::Context::calc_trunc_log_likelihood
(
    double x,
    double min,
    double max
) 
const
{ 
    return _likelihood->trunc_log_pdf(x, min, max);
}

/** 
 * If the value is outside the range, the probability is practically zero.
 */
double Model::Parameter::Context::calc_trunc_log_likelihood
(
    double x,
    double s,
    double min,
    double max
) 
const
{ 
    double mu = s * _likelihood->mu();
    double sigma = s * _likelihood->sigma();
    return _likelihood->trunc_log_pdf(x, mu, sigma, min, max);
}

double Model::Parameter::Context::sample_likelihood() const
{ 
    return _likelihood->sample(_range); 
}

double Model::Parameter::Context::sample_likelihood(double s) const
{ 
    double mu = s * _likelihood->mu();
    double sigma = s * _likelihood->sigma();
    return _likelihood->sample(mu, sigma, _range); 
}

double Model::Parameter::Context::sample_trunc_likelihood
(
    double min, 
    double max
) 
const
{ 
    assert(min <= max);

    if (min == max)
        return min;

    uint32_t N = TRUNC_TRY_LIMIT;
    double x = min - 1;

    while ((N > 0) && (x < min || x > max))
    {
        x = sample_likelihood();
        N--;
    }

    if (x < min)
    {
        return min;
    }
    if (x > max)
    {
        return max;
    }

    return x;
}

double Model::Parameter::Context::sample_trunc_likelihood
(
    double s,
    double min, 
    double max
) 
const
{ 
    assert(min <= max);

    if (min == max)
        return min;

    uint32_t N = TRUNC_TRY_LIMIT;
    double x = min - 1;

    while ((N > 0) && (x < min || x > max))
    {
        x = sample_likelihood(s);
        N--;
    }

    if (x < min)
    {
        return min;
    }
    if (x > max)
    {
        return max;
    }

    return x;
}

const Model::Parameter::Context::Gaussian_gamma&
Model::Parameter::Context::prior() const 
{ 
    if (has_ratio_prior())
    {
        return *_ratio_prior;
    }

    return *_prior; 
}

void Model::Parameter::Context::prior
(
    const Model::Parameter::Context::Gaussian_gamma& prior
)
{ 
    delete _prior;
    _prior = prior.clone();
}

double Model::Parameter::Context::calc_prior_prob(double mean, double stdev) 
    const
{ 
    if (has_ratio_prior())
    {
        return _ratio_prior->pdf(mean, stdev);
    }

    return _prior->pdf(mean, stdev); 
}

double Model::Parameter::Context::calc_prior_log_prob(double mean, double stdev)
    const
{ 
    if (has_ratio_prior())
    {
        return _ratio_prior->log_pdf(mean, stdev);
    }

    return _prior->log_pdf(mean, stdev); 
}

/** 
 * If the mean is outside the range, the probability is practically zero.
 */
double Model::Parameter::Context::calc_trunc_prior_log_prob
(
    double mean,
    double mean_min,
    double mean_max,
    double stdev
) 
const
{ 
    if (has_ratio_prior())
    {
        return _ratio_prior->trunc_log_pdf(mean, mean_min, mean_max, stdev);
    }

    return _prior->trunc_log_pdf(mean, mean_min, mean_max, stdev);
}

void Model::Parameter::Context::sample_prior
(
    double* mean_out, 
    double* stdev_out
)
const
{ 
    if (has_ratio_prior())
    {
        sample_ratio_prior(mean_out, stdev_out);
    }
    else
    {
        _prior->sample(mean_out, _range, stdev_out); 
    }
}

void Model::Parameter::Context::sample_prior
(
    double* mean_out, 
    double* stdev_out,
    double  mean_in
)
const
{ 
    if (has_ratio_prior())
    {
        sample_ratio_prior(mean_out, stdev_out, mean_in);
    }
    else
    {
        _prior->sample(mean_out, mean_in, _range, stdev_out); 
    }
}

bool Model::Parameter::Context::has_ratio_prior() const
{
    return _has_ratio_prior;
}

void Model::Parameter::Context::has_ratio_prior(bool b)
{
    _has_ratio_prior = b;
}

void Model::Parameter::Context::ratio_prior
(
    const Model::Parameter::Context::Gaussian_gamma& prior
)
{ 
    delete _ratio_prior;
    _ratio_prior = prior.clone();
}

void Model::Parameter::Context::sample_ratio_prior
(
    double* mean_out, 
    double* stdev_out
)
const
{ 
    Range r(min_ratio, max_ratio);
    _ratio_prior->sample(mean_out, &r, stdev_out); 
}

void Model::Parameter::Context::sample_ratio_prior
(
    double* mean_out, 
    double* stdev_out,
    double  mean_in
)
const
{ 
    Range r(min_ratio, max_ratio);
    _ratio_prior->sample(mean_out, mean_in, &r, stdev_out); 
}

const Model::Parameter::Context::Density&
Model::Parameter::Context::proposal() const 
{ 
    return *_proposal; 
}

void Model::Parameter::Context::proposal
(
    const Model::Parameter::Context::Density& proposal
)
{ 
    delete _proposal;
    _proposal = proposal.clone();
}

double Model::Parameter::Context::calc_proposal_prob(double x) const
{ 
    return _proposal->pdf(x); 
}

double Model::Parameter::Context::calc_proposal_log_prob(double x) const
{ 
    return _proposal->log_pdf(x); 
}

double Model::Parameter::Context::calc_trunc_proposal_log_prob
(
    double x,
    double min,
    double max
) 
const
{ 
    return _proposal->trunc_log_pdf(x, min, max); 
}

double Model::Parameter::Context::sample_proposal() const
{ 
    return _proposal->sample(_range); 
}

double Model::Parameter::Context::sample_trunc_proposal
(
    double min, 
    double max
) 
const
{ 
    assert(min <= max);

    if (min == max)
        return min;

    uint32_t N = TRUNC_TRY_LIMIT;
    double x = min - 1;

    while ((N > 0) && (x < min || x > max))
    {
        x = sample_proposal();
        N--;
    }

    if (x < min)
    {
        return min;
    }
    if (x > max)
    {
        return max;
    }

    return x;
}

const Model::Parameter::Context::Gaussian&
Model::Parameter::Context::gauss_proposal() const 
{ 
    return *_gauss_proposal; 
}

void Model::Parameter::Context::gauss_proposal
(
    const Model::Parameter::Context::Gaussian& gauss_proposal
)
{ 
    delete _gauss_proposal;
    _gauss_proposal = gauss_proposal.clone();
}

double Model::Parameter::Context::calc_gauss_proposal_prob(double x) const
{ 
    return _gauss_proposal->pdf(x); 
}

double Model::Parameter::Context::calc_gauss_proposal_log_prob(double x) const
{ 
    return _gauss_proposal->log_pdf(x); 
}

double Model::Parameter::Context::calc_trunc_gauss_proposal_log_prob
(
    double x,
    double min,
    double max
) 
const
{ 
    return _gauss_proposal->trunc_log_pdf(x, min, max); 
}

double Model::Parameter::Context::sample_gauss_proposal() const
{ 
    return _gauss_proposal->sample(_range); 
}

double Model::Parameter::Context::sample_gauss_proposal(double mean) const
{ 
    return _gauss_proposal->sample(mean, _gauss_proposal->sigma(), _range); 
}

double Model::Parameter::Context::sample_trunc_gauss_proposal
(
    double min, 
    double max
) 
const
{ 
    assert(min <= max);

    if (min == max)
        return min;

    uint32_t N = TRUNC_TRY_LIMIT;
    double x = min - 1;

    while ((N > 0) && (x < min || x > max))
    {
        x = sample_gauss_proposal();
        N--;
    }

    if (x < min)
    {
        return min;
    }
    if (x > max)
    {
        return max;
    }

    return x;
}

double Model::Parameter::Context::sample_trunc_gauss_proposal
(
    double mean,
    double min, 
    double max
) 
const
{ 
    assert(min <= max);

    if (min == max)
        return min;

    uint32_t N = TRUNC_TRY_LIMIT;
    double x = min - 1;

    while ((N > 0) && (x < min || x > max))
    {
        x = sample_gauss_proposal(mean);
        N--;
    }

    if (x < min)
    {
        return min;
    }
    if (x > max)
    {
        return max;
    }

    return x;
}

const Model::Parameter::Context::Dynamics&
Model::Parameter::Context::value_dynamics() const 
{ 
    return *_value_dynamics;
}

void Model::Parameter::Context::value_dynamics
(
    const Model::Parameter::Context::Dynamics& dynamics
)
{ 
    delete _value_dynamics;
    _value_dynamics = dynamics.clone();
}

const Model::Parameter::Context::Dynamics&
Model::Parameter::Context::mean_dynamics() const 
{ 
    if (has_ratio_prior())
    {
        return *_ratio_mean_dynamics;
    }

    return *_mean_dynamics;
}

void Model::Parameter::Context::mean_dynamics
(
    const Model::Parameter::Context::Dynamics& dynamics
)
{ 
    delete _mean_dynamics;
    _mean_dynamics = dynamics.clone();
}

const Model::Parameter::Context::Dynamics&
Model::Parameter::Context::stdev_dynamics() const 
{ 
    if (has_ratio_prior())
    {
        return *_ratio_stdev_dynamics;
    }

    return *_stdev_dynamics;
}

void Model::Parameter::Context::stdev_dynamics
(
    const Model::Parameter::Context::Dynamics& dynamics
)
{ 
    delete _stdev_dynamics;
    _stdev_dynamics = dynamics.clone();
}

void Model::Parameter::Context::ratio_mean_dynamics
(
    const Model::Parameter::Context::Dynamics& dynamics
)
{ 
    delete _ratio_mean_dynamics;
    _ratio_mean_dynamics = dynamics.clone();
}

void Model::Parameter::Context::ratio_stdev_dynamics
(
    const Model::Parameter::Context::Dynamics& dynamics
)
{ 
    delete _ratio_stdev_dynamics;
    _ratio_stdev_dynamics = dynamics.clone();
}

const double Model::Parameter::Context::min_stdev = 1.0e-2;
const double Model::Parameter::Context::min_ratio = 1.0e-2;
const double Model::Parameter::Context::max_ratio = 5.0;

/* ========================================================================== */






/* ----------------------  Model::Parameter::Pointer  ----------------------- */

Model::Parameter::Pointer::Pointer(Model::Parameter* p)
{
    ptr(p);
}

Model::Parameter::Pointer::Pointer(const Model::Parameter::Pointer& p)
{
    ptr(p._ptr);
}

Model::Parameter::Pointer::~Pointer()
{
}

Model::Parameter::Pointer* Model::Parameter::Pointer::clone() const
{
    return new Pointer(*this);
}

const char* Model::Parameter::Pointer::type_name() const
{
    return typeid(*this).name();
}

void Model::Parameter::Pointer::read(std::istream& in) 
    throw (jwscxx::base::IO_error, jwscxx::base::Arg_error)
{
    using std::ostringstream;
    using std::istringstream;
    using jwscxx::base::Arg_error;

    static const char* field_name = "Parameter";
    const char* field_value;
    size_t x;

    if (!(field_value = read_field_value(in, field_name)))
    {
        ostringstream ost;
        ost << "Missing serialized parameter: " << field_name;
        throw Arg_error(ost.str());
    }
    istringstream ist(field_value);
    ist >> x;
    if (ist.fail())
    {
        ostringstream ost;
        ost << field_name << ": Invalid serialized parameter";
        throw Arg_error(ost.str());
    }
    ptr((Parameter*)x);
}

void Model::Parameter::Pointer::write(std::ostream& out) const 
    throw (jwscxx::base::IO_error)
{
    static const char* field_name = "Parameter";
    size_t x = (size_t)_ptr;
    out << field_name << ": " << x << std::endl;
}

Model::Parameter* Model::Parameter::Pointer::ptr() const
{
    return _ptr;
}

void Model::Parameter::Pointer::ptr(Model::Parameter* p)
{
    _ptr = p;
}

/**
 * On function call, the internal pointer should be to a parameter. On function
 * return, contains a serialized index into the parameters of the all the @em
 * models.
 *
 * @param  models  Collection of Model pointers to use for serializing.
 */
void Model::Parameter::Pointer::serialize(const std::vector<Model*>& models)
{
    assert(ptr());

    size_t k = 0;

    for (size_t i = 0; i < models.size(); i++)
    {
        const Model* m = models[ i ];
        size_t j;

        j = m->fixed_param(ptr());
        if (j != m->num_fixed_params())
        {
            ptr((Model::Parameter*)(k + j + 1));
            return;
        }
        k += j;

        j = m->var_param(ptr());
        if (j != m->Model::num_var_params())
        {
            ptr((Model::Parameter*)(k + j + 1));
            return;
        }
        k += j;

        j = m->derived_param(ptr());
        if (j != m->num_derived_params())
        {
            ptr((Model::Parameter*)(k + j + 1));
            return;
        }
        k += j;
    }

    abort();
}

/**
 * On function call, the internal pointer should contain a serialized index
 * to a parameter.  On function return, it is a pointer to an indexed parameter
 * in the Model collection.
 *
 * @param  models  Collection of Model pointers to use for deserializing.
 */
void Model::Parameter::Pointer::deserialize(const std::vector<Model*>& models)
{
    assert(ptr());

    size_t k = (size_t)(ptr()) - 1;

    for (size_t i = 0; i < models.size(); i++)
    {
        const Model* m = models[ i ];
        size_t j;
        
        if (k < (j = m->num_fixed_params()))
        {
            ptr(m->fixed_param(k));
            return;
        }
        k -= j;
        
        if (k < (j = m->Model::num_var_params()))
        {
            ptr(m->var_param(k));
            return;
        }
        k -= j;
        
        if (k < (j = m->num_derived_params()))
        {
            ptr(m->derived_param(k));
            return;
        }
        k -= j;
    }

    abort();
}

/**
 * On function call, the internal pointer should be to a parameter in the @em
 * old_modes set. On function return, it is updated to the same parameter, but
 * in the @em new_models set.
 *
 * The model sets must be sorted according to the model pointers in 
 * @em old_models.
 *
 * @param  new_models  Collection of Models to use for updating to.
 * @param  old_models  Collection of Models to use for updating from. Both sets
 *                     must be sorted according to this one.
 */
void Model::Parameter::Pointer::update_pointers
(
    const std::vector<Model*>&       new_models,
    const std::vector<const Model*>& old_models
)
{
    assert(ptr());
    assert(old_models.size() == new_models.size());

    for (size_t i = 0; i < old_models.size(); i++)
    {
        const Model* m = old_models[ i ];
        size_t j;

        if (m == ptr()->model())
        {
            j = m->fixed_param(ptr());
            if (j != m->num_fixed_params())
            {
                ptr(new_models[ i ]->fixed_param(j));
                return;
            }

            j = m->var_param(ptr());
            if (j != m->Model::num_var_params())
            {
                ptr(new_models[ i ]->var_param(j));
                return;
            }

            j = m->derived_param(ptr());
            if (j != m->num_derived_params())
            {
                ptr(new_models[ i ]->derived_param(j));
                return;
            }
        }
    }

    for (size_t i = 0; i < new_models.size(); i++)
    {
        const Model* m = new_models[ i ];

        if (m == ptr()->model())
        {
            return;
        }
    }

    abort();
}

/* ========================================================================== */






/* -----------------------  Model::Parameter::Compare  ---------------------- */

bool Model::Parameter::Compare::operator() 
(
    const Model::Parameter* p1,
    const Model::Parameter* p2
) 
const 
{ 
    return p1->id() < p2->id(); 
}

/* ========================================================================== */






/* --------------------------  Model::Parameter  ---------------------------- */

Model::Parameter::Parameter
(
    size_t                           id,
    const char*                      name,
    double                           value,
    const Model::Parameter::Context* ctx,
    Model*                           model
)
throw (jwscxx::base::Arg_error)
{
    _id    = id;
    _name  = name;
    _ctx   = (ctx) ? ctx->clone() : 0;
    _model = model;

    double mean = (ctx) ? ctx->likelihood().mu() : 0;
    double stdev = (ctx) ? ctx->likelihood().sigma() : 0;

    test_value(value);
    test_stats(mean, stdev);

    this->value(value);
    scratch(0);
    stats(mean, stdev);

    has_ratio_prior(false);
}

Model::Parameter::Parameter(const Model::Parameter& p)
{
    _ctx = 0;

    *this = p;
}

Model::Parameter::~Parameter()
{
    delete _ctx;
}

/** 
 * @note Assignment copies the parameter pointers; this Parameter will have the
 * same model as @em p.
 */
Model::Parameter& Model::Parameter::operator=(const Model::Parameter& p)
{
    delete _ctx;

    _id      = p._id;
    _name    = p._name;
    _value   = p._value;
    _scratch = p._scratch;
    _mean    = p._mean;
    _stdev   = p._stdev;
    _ctx     = (p._ctx) ? p._ctx->clone() : 0;
    _model   = p._model;
    _has_ratio_prior = p._has_ratio_prior;

    return *this;
}

Model::Parameter* Model::Parameter::clone() const
{
    return new Parameter(*this);
}

/**
 * The file format is ASCII.
 *
 * @param  in  Input stream to read this Model::Parameter from.
 *
 * @throw  jwscxx::base::IO_error   Could not read from @em in.
 * @throw  jwscxx::base::Arg_error  Invalid arguments in file to read from.
 */
void Model::Parameter::read(std::istream& in) 
    throw (jwscxx::base::IO_error, jwscxx::base::Arg_error)
{
    using std::ostringstream;
    using std::istringstream;
    using std::vector;
    using jwscxx::base::Arg_error;

    const char* field_value;
    double v, m, s;
    uint32_t b;

    ostringstream ost;

    if (!(field_value = read_field_value(in, _name)))
    {
        ost << "Missing parameter: " << _name;
        throw Arg_error(ost.str());
    }
    istringstream ist(field_value);
    ist >> v >> m >> s >> b;
    if (ist.fail())
    {
        ost << _name << ": Invalid value";
        throw Arg_error(ost.str());
    }
    ist.clear(std::ios_base::goodbit);

    try
    {
        value(v);
        mean(m);
        stdev(s);
        has_ratio_prior(b);
    }
    catch (Arg_error e)
    {
        ost << _name << ": " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/**
 * The file format is ASCII.
 *
 * @param  out  Output stream to write this Model::Parameter to.
 *
 * @throw  jwscxx::base::IO_error  Could not write to @em out.
 */
void Model::Parameter::write(std::ostream& out) const
    throw (jwscxx::base::IO_error)
{
    out.precision(16);
    out.setf(std::ios::fixed,std::ios::floatfield);
    out << _name << ": " << _value << " " 
        << _mean << " " << _stdev << " " 
        << _has_ratio_prior << std::endl;
}

size_t Model::Parameter::id() const
{
    return _id;
}

const char* Model::Parameter::name() const
{
    return _name;
}

double Model::Parameter::value() const
{
    return _value;
}

/**
 * @param  value  New parameter value.
 *
 * @warning Aborts on invalid arguments.
 */
void Model::Parameter::value(double value)
{
    try { test_value(value); }
    catch (jwscxx::base::Arg_error e)
    { 
        std::cerr << __FILE__ << ":" << __LINE__ << " "
                  << e.get_msg() << std::endl;
        abort(); 
    }

    if (_ctx)
    {
        double min = _ctx->range().min() + 1.0e-6;
        double max = _ctx->range().max() - 1.0e-6;

        if (max <= min)
        {
            _value = 0.5*(_ctx->range().min() + _ctx->range().max());
        }
        else if ((max - value) < 0)
        {
            _value = max;
        }
        else if ((value - min) < 0)
        {
            _value = min;
        }
        else
        {
            _value = value;
        }

        try { test_value(_value); }
        catch (jwscxx::base::Arg_error e)
        { 
            std::cerr << __FILE__ << ":" << __LINE__ << " "
                      << e.get_msg() << std::endl;
            abort(); 
        }
    }
    else
    {
        _value = value;
    }
}

void Model::Parameter::test_value(double value) const
    throw (jwscxx::base::Arg_error)
{
    if (_ctx)
    {
        try
        {
            _ctx->test_range(value);
        }
        catch (jwscxx::base::Arg_error e)
        {
            std::ostringstream ost;
            ost << _name << ": " << e.get_msg();
            throw jwscxx::base::Arg_error(ost.str());
        }
    }

}

double Model::Parameter::scratch() const
{
    return _scratch;
}

void Model::Parameter::scratch(double value)
{
    _scratch = value;
}

double Model::Parameter::mean() const
{
    return _mean;
}

/**
 * @param  mean  New parameter mean.
 *
 * @warning Aborts on invalid arguments.
 */
void Model::Parameter::mean(double mean)
{
    stats(mean, stdev());
}

void Model::Parameter::test_mean(double mean) const
    throw (jwscxx::base::Arg_error)
{
    test_stats(mean, stdev());
}

double Model::Parameter::stdev() const
{
    return _stdev;
}

/**
 * @param  stdev  New parameter stdev.
 *
 * @warning Aborts on invalid arguments.
 */
void Model::Parameter::stdev(double stdev)
{
    stats(mean(), stdev);
}

void Model::Parameter::test_stdev(double stdev) const
    throw (jwscxx::base::Arg_error)
{
    test_stats(mean(), stdev);
}

void Model::Parameter::stats(double* mean_out, double* stdev_out) const
{
    *mean_out = _mean;
    *stdev_out = _stdev;
}

/**
 * @param  mean   New parameter mean.
 * @param  stdev  New parameter standard deviation..
 *
 * @warning Aborts on invalid arguments.
 */
void Model::Parameter::stats(double mean, double stdev)
{
    try { test_stats(mean, stdev); }
    catch (jwscxx::base::Arg_error e) { abort(); }

    if (_ctx)
    {
        double min_stdev = _ctx->min_stdev + 1.0e-6;

        _mean = mean;
        _stdev = ((stdev - min_stdev) < 0) ?  min_stdev : stdev;

        try { test_stats(_mean, _stdev); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        Context::Gaussian gauss = _ctx->likelihood();
        gauss.set(_mean, _stdev);
        _ctx->likelihood(gauss);
    }
    else
    {
        _mean = mean;
        _stdev = stdev;
    }
}

void Model::Parameter::test_stats(double mean, double stdev) const
    throw (jwscxx::base::Arg_error)
{
    if (_ctx)
    {
        if (stdev < _ctx->min_stdev)
        {
            std::ostringstream ost;
            ost << _name << ": Standard deviation too small";
            throw jwscxx::base::Arg_error(ost.str());
        }
    }

}

bool Model::Parameter::has_ratio_prior() const
{
    return _has_ratio_prior;
}

void Model::Parameter::has_ratio_prior(bool b)
{
    _has_ratio_prior = b;
    if (_ctx)
    {
        _ctx->has_ratio_prior(b);
    }
}

const Model::Parameter::Context* Model::Parameter::context() const
{
    return _ctx;
}

void Model::Parameter::context
(
    const Model::Parameter::Context* ctx, 
    bool                             clip
)
throw (jwscxx::base::Arg_error)
{
    if (!ctx)
    {
        delete _ctx;
        _ctx = 0;
        return;
    }

    if (clip)
    {
        if (_value < ctx->range().min())
        {
            _value = ctx->range().min();
        }
        if (_value > ctx->range().max())
        {
            _value = ctx->range().max();
        }
    }
    else
    {
        ctx->test_range(_value);
    }

    delete _ctx;
    _ctx = ctx->clone();
}

Model* Model::Parameter::model() const
{
    return _model;
}

void Model::Parameter::model(Model* m)
{
    _model = m;
}

/**
 * @param  models  Collection of Model pointers to use for serializing.
 */
void Model::Parameter::serialize(const std::vector<Model*>& models)
{
}

/**
 * @param  models  Collection of Model pointers to use for deserializing.
 */
void Model::Parameter::deserialize(const std::vector<Model*>& models
)
{
}

/**
 * @param  new_models  Collection of pointers to update this Parameter to.
 * @param  old_models  Collection of pointers to update this Parameter from.
 */
void Model::Parameter::update_pointers
(
    const std::vector<Model*>&       new_models,
    const std::vector<const Model*>& old_models
)
{
}

/* ========================================================================== */







/* ----------------------------  Model::Pointer  ---------------------------- */

Model::Pointer::Pointer(Model* m)
{
    ptr(m);
}

Model::Pointer::Pointer(const Model::Pointer& p)
{
    ptr(p._ptr);
}

Model::Pointer::~Pointer()
{
}

Model::Pointer* Model::Pointer::clone() const
{
    return new Pointer(*this);
}

const char* Model::Pointer::type_name() const
{
    return typeid(*this).name();
}

void Model::Pointer::read(std::istream& in) 
    throw (jwscxx::base::IO_error, jwscxx::base::Arg_error)
{
    using std::ostringstream;
    using std::istringstream;
    using jwscxx::base::Arg_error;

    static const char* field_name = "Model";
    const char* field_value;
    size_t x;

    if (!(field_value = read_field_value(in, field_name)))
    {
        ostringstream ost;
        ost << "Missing serialized model: " << field_name;
        throw Arg_error(ost.str());
    }
    istringstream ist(field_value);
    ist >> x;
    if (ist.fail())
    {
        ostringstream ost;
        ost << field_name << ": Invalid serialized model";
        throw Arg_error(ost.str());
    }
    ptr((Model*)x);
}

void Model::Pointer::write(std::ostream& out) const 
    throw (jwscxx::base::IO_error)
{
    static const char* field_name = "Model";
    size_t x = (size_t)_ptr;
    out << field_name << ": " << x << std::endl;
}

Model* Model::Pointer::ptr() const
{
    return _ptr;
}

void Model::Pointer::ptr(Model* m)
{
    _ptr = m;
}

/**
 * On function call, the internal pointer should be to a model. On function
 * return, contains a serialized index into the @em models.
 *
 * @param  models  Collection of Model pointers to use for serializing.
 */
void Model::Pointer::serialize(const std::vector<Model*>& models)
{
    if (ptr() == 0)
        return;

    for (size_t i = 0; i < models.size(); i++)
    {
        assert(sizeof(i) >= 4 && i <= 0xFFFFFFFF);

        if (ptr() == models[ i ])
        {
            ptr((Model*)(i + 1));
            return;
        }
    }

    abort();
}

/**
 * On function call, the internal pointer should contain a serialized index to a
 * model.  On function return, it is a pointer to an index in the model
 * collection.
 *
 * @param  models  Collection of Model pointers to use for deserializing.
 */
void Model::Pointer::deserialize(const std::vector<Model*>& models)
{
    if (ptr() == 0)
        return;

    size_t j = (size_t)(ptr()) - 1;

    for (size_t i = 0; i < models.size(); i++)
    {
        if (j == i)
        {
            ptr(models[ j ]);
            return;
        }
    }

    abort();
}

/**
 * On function call, the internal pointer should be to a model in the @em
 * old_modes set. On function return, it is updated to the same model, but
 * in the @em new_models set.
 *
 * @param  new_models  Collection of Models to use for updating to.
 * @param  old_models  Collection of Models to use for updating from.
 */
void Model::Pointer::update_pointers
(
    const std::vector<Model*>&       new_models,
    const std::vector<const Model*>& old_models
)
{
    if (ptr() == 0)
        return;

    assert(old_models.size() == new_models.size());

    for (size_t i = 0; i < old_models.size(); i++)
    {
        if (ptr() == old_models[ i ])
        {
            ptr(new_models[ i ]);
            return;
        }
    }

    abort();
}

/* ========================================================================== */








/* --------------------------------  Model  --------------------------------- */

Model::Model(const char* name, const Model::Context* ctx)
{
    _name = name;
    _ctx  = ctx;
    _graphical_rep_outdated = false;
}

Model::Model(const Model& m)
{
    *this = m;
}

Model::~Model()
{
}

/** 
 * @param  m  Model to copy into this one.
 *
 * @return A reference to this Model.
 */
Model& Model::operator= (const Model& m)
{
    _name = m._name;
    _ctx  = m._ctx;
    _graphical_rep_outdated = m._graphical_rep_outdated;

    return *this;
}

/**
 * The file format is ASCII.
 *
 * @param  in  Input stream to read this Model from.
 *
 * @throw  jwscxx::base::IO_error   Could not read from @em in.
 * @throw  jwscxx::base::Arg_error  Invalid arguments in file to read from.
 */
void Model::read(std::istream& in) 
    throw (jwscxx::base::IO_error, jwscxx::base::Arg_error)
{
    using std::ostringstream;
    using std::istringstream;
    using jwscxx::base::Arg_error;

    const char* type_name = typeid(*this).name();
    const char* field_value;

    if (!(field_value = read_field_value(in, "Type")))
    {
        throw Arg_error("Missing Type field");
    }
    if (strncmp(field_value, type_name, strlen(type_name)) != 0)
    {
        ostringstream ost;
        ost << "Tried to read a '" << field_value << "' as a '"
            << type_name << "'";
        throw Arg_error(ost.str());
    }

    for (size_t i = 0; i < _fixed_params.size(); i++)
    {
        _fixed_params[ i ]->read(in);
    }

    for (size_t i = 0; i < _var_params.size(); i++)
    {
        _var_params[ i ]->read(in);
    }

    for (size_t i = 0; i < _derived_params.size(); i++)
    {
        _derived_params[ i ]->read(in);
    }
}

/**
 * The file format is ASCII.
 *
 * @param  out  Output stream to write this Model to.
 *
 * @throw  jwscxx::base::IO_error  Could not write to @em out.
 */
void Model::write(std::ostream& out) const 
    throw (jwscxx::base::IO_error)
{
    out << "Type: " << typeid(*this).name() << std::endl;

    for (size_t i = 0; i < _fixed_params.size(); i++)
    {
        _fixed_params[ i ]->write(out);
    }

    for (size_t i = 0; i < _var_params.size(); i++)
    {
        _var_params[ i ]->write(out);
    }

    for (size_t i = 0; i < _derived_params.size(); i++)
    {
        _derived_params[ i ]->write(out);
    }
}

const Model::Context* Model::context() const
{
    return _ctx;
}

const char* Model::name() const
{
    return _name;
}

/**
 * @param  i  Index of the parameter to get.
 *
 * @return A pointer to the indexed parameter.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Parameter index.
 */
Model::Parameter* Model::fixed_param(uint32_t i) const 
    throw (jwscxx::base::Arg_error)
{
    if (i >= num_fixed_params())
    {
        throw jwscxx::base::Arg_error("Parameter index too large");
    }

    return _fixed_params[ i ];
}

size_t Model::fixed_param(const Model::Parameter* p) const
    throw (jwscxx::base::Arg_error)
{
    std::map<const Parameter*, size_t>::const_iterator it;
    it = _fixed_params_index.find(p);
    if (it == _fixed_params_index.end())
    {
        return num_fixed_params();
    }
    return (*it).second;
}

void Model::add_fixed_param(Model::Parameter* p)
{
    Parameter::Compare cmpr;
    std::vector<Parameter*>::iterator j_it;
    std::map<const Parameter*, size_t>::iterator k_it;
    
    j_it = lower_bound(_fixed_params.begin(), _fixed_params.end(), p, cmpr);
    size_t i = j_it - _fixed_params.begin();

    for (; j_it != _fixed_params.end(); j_it++)
    {
        assert(*j_it != p);
        k_it = _fixed_params_index.find(*j_it);
        assert(k_it != _fixed_params_index.end());
        (*k_it).second++;
    }

    _fixed_params_index[ p ] = i;
    _fixed_params.insert(_fixed_params.begin() + i, p);
}

Model::Parameter* Model::remove_fixed_param(Model::Parameter* p)
{
    size_t i = fixed_param(p);

    std::vector<Parameter*>::iterator j_it;
    std::map<const Parameter*, size_t>::iterator k_it;

    j_it = _fixed_params.begin() + i;
    k_it = _fixed_params_index.find(*j_it);
    assert(k_it != _fixed_params_index.end());
    _fixed_params_index.erase(k_it);

    for (j_it += 1; j_it != _fixed_params.end(); j_it++)
    {
        k_it = _fixed_params_index.find(*j_it);
        assert(k_it != _fixed_params_index.end());
        (*k_it).second--;
    }

    j_it = _fixed_params.begin() + i;
    _fixed_params.erase(j_it);

    return p;
}

/**
 * @param  i  Index of the parameter to get.
 *
 * @return A pointer to the indexed parameter.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Parameter index.
 */
const Model::Parameter* Model::const_var_param(uint32_t i) const 
    throw (jwscxx::base::Arg_error)
{
    return var_param(i);
}

/**
 * @param  i  Index of the parameter to get.
 *
 * @return A pointer to the indexed parameter.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Parameter index.
 */
Model::Parameter* Model::var_param(uint32_t i) const 
    throw (jwscxx::base::Arg_error)
{
    if (i >= Model::num_var_params())
    {
        throw jwscxx::base::Arg_error("Parameter index too large");
    }

    return _var_params[ i ];
}

/** 
 * Uses an associative map to provide efficient look-up. If @em p is not a
 * variable parameter, num_var_params() is returned.
 */
size_t Model::var_param(const Model::Parameter* p) const
    throw (jwscxx::base::Arg_error)
{
    std::map<const Parameter*, size_t>::const_iterator it;
    it = _var_params_index.find(p);
    if (it == _var_params_index.end())
    {
        return Model::num_var_params();
    }
    return (*it).second;
}

void Model::add_var_param(Model::Parameter* p)
{
    Parameter::Compare cmpr;
    std::vector<Parameter*>::iterator j_it;
    std::map<const Parameter*, size_t>::iterator k_it;
    
    j_it = lower_bound(_var_params.begin(), _var_params.end(), p, cmpr);
    size_t i = j_it - _var_params.begin();

    for (; j_it != _var_params.end(); j_it++)
    {
        assert(*j_it != p);
        k_it = _var_params_index.find(*j_it);
        assert(k_it != _var_params_index.end());
        (*k_it).second++;
    }

    _var_params_index[ p ] = i;
    _var_params.insert(_var_params.begin() + i, p);
}

Model::Parameter* Model::remove_var_param(Model::Parameter* p)
{
    size_t i = var_param(p);

    std::vector<Parameter*>::iterator j_it;
    std::map<const Parameter*, size_t>::iterator k_it;

    j_it = _var_params.begin() + i;
    k_it = _var_params_index.find(*j_it);
    assert(k_it != _var_params_index.end());
    _var_params_index.erase(k_it);

    for (j_it += 1; j_it != _var_params.end(); j_it++)
    {
        k_it = _var_params_index.find(*j_it);
        assert(k_it != _var_params_index.end());
        (*k_it).second--;
    }

    j_it = _var_params.begin() + i;
    _var_params.erase(j_it);

    return p;
}

/**
 * @param  i  Index of the parameter to get.
 *
 * @return A pointer to the indexed parameter.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Parameter index.
 */
Model::Parameter* Model::derived_param(uint32_t i) const 
    throw (jwscxx::base::Arg_error)
{
    if (i >= num_derived_params())
    {
        throw jwscxx::base::Arg_error("Parameter index too large");
    }

    return _derived_params[ i ];
}

size_t Model::derived_param(const Model::Parameter* p) const
    throw (jwscxx::base::Arg_error)
{
    std::map<const Parameter*, size_t>::const_iterator it;
    it = _derived_params_index.find(p);
    if (it == _derived_params_index.end())
    {
        return num_derived_params();
    }
    return (*it).second;
}

void Model::add_derived_param(Model::Parameter* p)
{
    Parameter::Compare cmpr;
    std::vector<Parameter*>::iterator j_it;
    std::map<const Parameter*, size_t>::iterator k_it;
    
    j_it = lower_bound(_derived_params.begin(), _derived_params.end(), p, cmpr);
    size_t i = j_it - _derived_params.begin();

    for (; j_it != _derived_params.end(); j_it++)
    {
        assert(*j_it != p);
        k_it = _derived_params_index.find(*j_it);
        assert(k_it != _derived_params_index.end());
        (*k_it).second++;
    }

    _derived_params_index[ p ] = i;
    _derived_params.insert(_derived_params.begin() + i, p);
}

Model::Parameter* Model::remove_derived_param(Model::Parameter* p)
{
    size_t i = derived_param(p);

    std::vector<Parameter*>::iterator j_it;
    std::map<const Parameter*, size_t>::iterator k_it;

    j_it = _derived_params.begin() + i;
    k_it = _derived_params_index.find(*j_it);
    assert(k_it != _derived_params_index.end());
    _derived_params_index.erase(k_it);

    for (j_it += 1; j_it != _derived_params.end(); j_it++)
    {
        k_it = _derived_params_index.find(*j_it);
        assert(k_it != _derived_params_index.end());
        (*k_it).second--;
    }

    j_it = _derived_params.begin() + i;
    _derived_params.erase(j_it);

    return p;
}

uint32_t Model::num_fixed_params() const
{
    return _fixed_params.size();
}

uint32_t Model::num_var_params() const
{
    return _var_params.size();
}

uint32_t Model::num_derived_params() const
{
    return _derived_params.size();
}

const char* Model::get_var_param_name(uint32_t d) const
{
    uint32_t D = _var_params.size();
    assert(d < D);

    std::vector<const char*> names = get_var_param_names();
    return names[ d ];
}

std::vector<const char*> Model::get_var_param_names() const
{
    uint32_t D = _var_params.size();
    std::vector<const char*> names;

    for (uint32_t d = 0; d < D; d++)
    {
        names.push_back(_var_params[ d ]->name());
    }

    return names;
}

double Model::get_var_param_value(uint32_t d) const
{
    using namespace jwsc;

    uint32_t D = _var_params.size();
    assert(d < D);

    return _var_params[ d ]->value();
}

void Model::get_var_param_values(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    uint32_t D = _var_params.size();

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;

    for (uint32_t d = 0; d < D; d++)
    {
        p->elts[ d ] = _var_params[ d ]->value();
    }
}

/**
 * If a Parameter value could not be set, it is left unchanged and an exception
 * is thrown, after an attempt to set the other parameters is made.
 *
 * @param  p  New parameter values.
 */
void Model::set_var_param_values(const jwsc::Vector_d* p) 
    throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    uint32_t D = _var_params.size();
    double delta;

    assert(D == p->num_elts);

    std::ostringstream ost;

    for (uint32_t d = 0; d < D; d++)
    {
        if (fabs(delta = p->elts[ d ] - _var_params[ d ]->value()) > 0)
        {
            try 
            { 
                double x = _var_params[ d ]->value() + delta;
                _var_params[ d ]->test_value(x);
                _var_params[ d ]->value(x);
            }
            catch (jwscxx::base::Arg_error e) 
            { ost << e.get_msg() << "  "; }
        }
    }

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

/**
 * If a Parameter value could not be set, it is left unchanged and an exception
 * is thrown, after an attempt to set the other parameters is made.
 *
 * @param  p  Point to add to the Parameter values.
 *
 * @throw  jwscxx::base::Arg_error  Invalid resulting parameter value.
 */
void Model::add_var_param_values(const jwsc::Vector_d* p) 
    throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    uint32_t D = _var_params.size();

    assert(D == p->num_elts);

    Vector_d* q = 0;
    get_var_param_values(&q);
    add_vectors_d(&q, p, q);

    try
    {
        set_var_param_values(q);
    }
    catch (jwscxx::base::Arg_error e)
    {
        free_vector_d(q);
        throw e;
    }

    free_vector_d(q);
}

void Model::get_var_param_values_dynamics_grad_eta(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    uint32_t D = _var_params.size();

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;

    for (uint32_t d = 0; d < D; d++)
    {
        p->elts[ d ] = _var_params[ d ]->context()->value_dynamics().grad_eta();
    }
}

void Model::get_var_param_values_dynamics_delta_t(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    uint32_t D = _var_params.size();

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;

    for (uint32_t d = 0; d < D; d++)
    {
        p->elts[ d ] = _var_params[ d ]->context()->value_dynamics().delta_t();
    }
}

void Model::print_var_param_values(uint32_t iteration, std::ostream& out) const
{
    for (uint32_t i = 0; i < Model::num_var_params(); i++)
    {
        out << iteration << ',' << name() << ','
            << Model::get_var_param_name(i) << ','
            << Model::get_var_param_value(i) << std::endl;
    }
}

void Model::get_var_param_stats
(
    double*  mean_out, 
    double*  stdev_out, 
    uint32_t d
) 
const
{
    using namespace jwsc;

    uint32_t D = _var_params.size();
    assert(d < D);

    _var_params[ d ]->stats(mean_out, stdev_out);
}

void Model::get_var_param_stats
(
    jwsc::Vector_d** means_out,
    jwsc::Vector_d** stdevs_out
) 
const
{
    using namespace jwsc;

    uint32_t D = _var_params.size();

    create_vector_d(means_out, D);
    create_vector_d(stdevs_out, D);

    Vector_d* means = *means_out;
    Vector_d* stdevs = *stdevs_out;

    for (uint32_t d = 0; d < D; d++)
    {
        means->elts[ d ] = _var_params[ d ]->mean();
        stdevs->elts[ d ] = _var_params[ d ]->stdev();
    }
}

void Model::set_var_param_stats
(
    const jwsc::Vector_d* means,
    const jwsc::Vector_d* stdevs
)
throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    uint32_t D = _var_params.size();

    assert(D <= means->num_elts);
    assert(D <= stdevs->num_elts);

    std::ostringstream ost;

    for (uint32_t d = 0; d < D; d++)
    {
        try 
        { 
            double m = means->elts[ d ];
            double s = stdevs->elts[ d ];
            _var_params[ d ]->test_stats(m, s);
            _var_params[ d ]->stats(m, s);
        }
        catch (jwscxx::base::Arg_error e) 
        { ost << e.get_msg() << "  "; }
    }

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

void Model::get_var_param_stats_dynamics_grad_eta
(
    jwsc::Vector_d** means_out,
    jwsc::Vector_d** stdevs_out
) 
const
{
    using namespace jwsc;

    uint32_t D = _var_params.size();

    create_vector_d(means_out, D);
    create_vector_d(stdevs_out, D);

    Vector_d* means = *means_out;
    Vector_d* stdevs = *stdevs_out;

    for (uint32_t d = 0; d < D; d++)
    {
        double mean, stdev;
        mean = _var_params[ d ]->context()->mean_dynamics().grad_eta();
        stdev = _var_params[ d ]->context()->stdev_dynamics().grad_eta();
        means->elts[ d ] = mean;
        stdevs->elts[ d ] = stdev;
    }
}

void Model::get_var_param_stats_dynamics_delta_t
(
    jwsc::Vector_d** means_out,
    jwsc::Vector_d** stdevs_out
) const
{
    using namespace jwsc;

    uint32_t D = _var_params.size();

    create_vector_d(means_out, D);
    create_vector_d(stdevs_out, D);

    Vector_d* means = *means_out;
    Vector_d* stdevs = *stdevs_out;

    for (uint32_t d = 0; d < D; d++)
    {
        double mean, stdev;
        mean = _var_params[ d ]->context()->mean_dynamics().delta_t();
        stdev = _var_params[ d ]->context()->stdev_dynamics().delta_t();
        means->elts[ d ] = mean;
        stdevs->elts[ d ] = stdev;
    }
}

void Model::print_var_param_stats(uint32_t iteration, std::ostream& out) const
{
    double mean, stdev;

    for (uint32_t i = 0; i < Model::num_var_params(); i++)
    {
        Model::get_var_param_stats(&mean, &stdev, i);
        out  << iteration << ',' << name() << ','
             << Model::get_var_param_name(i) << ','
             << mean << ',' << stdev << std::endl;
    }
}

double Model::calc_var_param_log_likelihood() const
{
    double log_prob = 0;

    for (uint32_t i = 0; i < _var_params.size(); i++)
    {
        double x = _var_params[ i ]->value();
        log_prob += _var_params[ i ]->context()->calc_log_likelihood(x);
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

void Model::sample_var_param_likelihood(jwsc::Vector_d** p_out)
{
    using namespace jwsc;

    uint32_t D = _var_params.size();

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;

    for (uint32_t d = 0; d < D; d++)
    {
        p->elts[ d ] = _var_params[ d ]->context()->sample_likelihood();
    }
}

double Model::calc_var_param_prior_log_prob() const
{
    double mean, stdev;
    double log_prob = 0;

    for (uint32_t i = 0; i < _var_params.size(); i++)
    {
        Parameter* param = _var_params[ i ];
        mean = param->mean();
        stdev = param->stdev();
        log_prob += param->context()->calc_prior_log_prob(mean, stdev);
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

void Model::sample_var_param_prior
(
    jwsc::Vector_d** means_out,
    jwsc::Vector_d** stdevs_out
) 
const
{
    using namespace jwsc;

    uint32_t D = _var_params.size();
    double mu, sigma;

    create_vector_d(means_out, D);
    create_vector_d(stdevs_out, D);

    Vector_d* means = *means_out;
    Vector_d* stdevs = *stdevs_out;

    for (uint32_t d = 0; d < D; d++)
    {
        _var_params[ d ]->context()->sample_prior(&mu, &sigma);
        means->elts[ d ] = mu;
        stdevs->elts[ d ] = sigma;
    }
}

void Model::sample_var_param_cat_proposal
(
    jwsc::Vector_d** means_out,
    jwsc::Vector_d** stdevs_out,
    const jwsc::Vector_d* means_in
) 
const
{
    using namespace jwsc;

    uint32_t D = _var_params.size();
    double mu, sigma;
    double mean_in;

    create_vector_d(means_out, D);
    create_vector_d(stdevs_out, D);

    Vector_d* means = *means_out;
    Vector_d* stdevs = *stdevs_out;

    for (uint32_t d = 0; d < D; d++)
    {
        mean_in = means_in->elts[ d ];
        _var_params[ d ]->context()->sample_prior(&mu, &sigma, mean_in);
        means->elts[ d ] = mu;
        stdevs->elts[ d ] = sigma;
    }
}

double Model::calc_var_param_proposal_log_prob() const
{
    double log_prob = 0;

    for (uint32_t i = 0; i < _var_params.size(); i++)
    {
        Parameter* param = _var_params[ i ];
        double x = param->value();
        log_prob += param->context()->calc_proposal_log_prob(x);
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

double Model::calc_var_param_gauss_proposal_log_prob() const
{
    double log_prob = 0;

    for (uint32_t i = 0; i < _var_params.size(); i++)
    {
        Parameter* param = _var_params[ i ];
        double x = param->value();
        log_prob += param->context()->calc_gauss_proposal_log_prob(x);
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

void Model::sample_var_param_proposal(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    uint32_t D = _var_params.size();

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;

    for (uint32_t d = 0; d < D; d++)
    {
        p->elts[ d ] = _var_params[ d ]->context()->sample_proposal();
    }
}

/** Uses zero Gaussian means. */
void Model::sample_var_param_gauss_proposal(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    uint32_t D = _var_params.size();
    Vector_d* means = 0;

    create_zero_vector_d(&means, D);
    sample_var_param_gauss_proposal(p_out, means);
    free_vector_d(means);
}

void Model::sample_var_param_gauss_proposal
(
    jwsc::Vector_d**      p_out,
    const jwsc::Vector_d* means
) 
const
{
    using namespace jwsc;

    uint32_t D = _var_params.size();
    assert(D == means->num_elts);

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;

    for (uint32_t d = 0; d < D; d++)
    {
        Parameter* param = _var_params[ d ];
        double mu_d = means->elts[ d ];
        p->elts[ d ] = param->context()->sample_gauss_proposal(mu_d);
    }
}

void Model::get_var_param_gauss_proposal_sigma(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    uint32_t D = _var_params.size();

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;

    for (uint32_t d = 0; d < D; d++)
    {
        Parameter* param = _var_params[ d ];
        p->elts[ d ] = param->context()->gauss_proposal().sigma();
    }
}

void Model::serialize(const std::vector<Model*>& models)
{
    for (size_t i = 0; i < _fixed_params.size(); i++)
    {
        _fixed_params[ i ]->serialize(models);
    }

    for (size_t i = 0; i < _var_params.size(); i++)
    {
        _var_params[ i ]->serialize(models);
    }

    for (size_t i = 0; i < _derived_params.size(); i++)
    {
        _derived_params[ i ]->serialize(models);
    }
}

void Model::deserialize(const std::vector<Model*>& models)
{
    for (size_t i = 0; i < _fixed_params.size(); i++)
    {
        _fixed_params[ i ]->deserialize(models);
    }

    for (size_t i = 0; i < _var_params.size(); i++)
    {
        _var_params[ i ]->deserialize(models);
    }

    for (size_t i = 0; i < _derived_params.size(); i++)
    {
        _derived_params[ i ]->deserialize(models);
    }
}

void Model::update_pointers
(
    const std::vector<Model*>&       new_models,
    const std::vector<const Model*>& old_models
)
{
    for (size_t i = 0; i < _fixed_params.size(); i++)
    {
        _fixed_params[ i ]->update_pointers(new_models, old_models);
    }

    for (size_t i = 0; i < _var_params.size(); i++)
    {
        _var_params[ i ]->update_pointers(new_models, old_models);
    }

    for (size_t i = 0; i < _derived_params.size(); i++)
    {
        _derived_params[ i ]->update_pointers(new_models, old_models);
    }
}

bool Model::graphical_rep_is_outdated() const
{
    return _graphical_rep_outdated;
}

void Model::graphical_rep_outdated()
{
    _graphical_rep_outdated = true;
}

void Model::update_graphical_rep()
{
    _graphical_rep_outdated = false;
}

/* ========================================================================== */
