/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Parametric generative model.
 *
 * Specialed for inference with MCMC sampling techniques.
 */


#ifndef MODEL_H
#define MODEL_H


#include <config.h>

#include <iostream>
#include <sstream>
#include <vector>
#include <map>

#include <jwsc/vector/vector.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/base/cloneable.h>
#include <jwsc++/base/readable.h>
#include <jwsc++/base/writeable.h>


/** 
 * @class Model
 *
 * @brief Parametric generative model.
 */
class Model : public jwscxx::base::Cloneable, 
              public jwscxx::base::Writeable, 
              public jwscxx::base::Readable
{
    public:


/** 
 * @class Model::Context
 *
 * @brief Model context for all parameters.
 */
class Context : public jwscxx::base::Cloneable
{
    public:

        Context();

        Context(const Context& m);

        virtual ~Context();

        virtual Context& operator= (const Context& ctx);

        virtual Context* clone() const = 0;
};


/** 
 * @class Model::Parameter
 *
 * @brief Model parameter containing both instance and category information.
 *
 * The category parameters are shared across multiple instances.
 */
class Parameter : public jwscxx::base::Cloneable, 
                  public jwscxx::base::Writeable, 
                  public jwscxx::base::Readable
{
    public:


/**
 * @class Model::Parameter::Context
 *
 * @brief Model parameter context.
 *
 * Constains context information such as parameter range, prior, proposal, etc.
 */
class Context : public jwscxx::base::Cloneable
{
    public:


/** 
 * @class Model::Parameter::Context::Range
 *
 * @brief Range of a model parameter.
 */
class Range : public jwscxx::base::Cloneable
{
    public:

        /** @brief Constructs a new parameter Range. */
        Range(double min=0, double max=0) throw (jwscxx::base::Arg_error);

        /** @brief Clones this Range. */
        Model::Parameter::Context::Range* clone() const;

        double min() const;

        double max() const;

        /** @brief Sets extents of this Range. */
        void set(double min, double max) throw (jwscxx::base::Arg_error);

        /** @brief Tests a parameter value against this Range. */
        void test(double value) const throw (jwscxx::base::Arg_error);

    private:

        double _min;

        double _max;
};


/**
 * @class Model::Parameter::Context::Density
 *
 * @brief Density function over a model parameter.
 */
class Density : public jwscxx::base::Cloneable
{
    public:

        virtual Model::Parameter::Context::Density* clone() const = 0;

        virtual double pdf(double x) const = 0;

        virtual double log_pdf(double x) const = 0;

        /** @brief Evaluates a truncated version of the log pdf function. */
        virtual double trunc_log_pdf(double x, double min, double max) const;

        virtual double sample(const Model::Parameter::Context::Range* range) 
            const = 0;
};


/**
 * @class Model::Parameter::Context::Uniform
 *
 * @brief Uniform density function over a model parameter.
 */
class Uniform : public Density
{
    public:

        Model::Parameter::Context::Uniform* clone() const;

        double pdf(double x) const;

        double log_pdf(double x) const;

        double sample(const Model::Parameter::Context::Range* range) const;
};


/**
 * @class Model::Parameter::Context::Gaussian
 *
 * @brief Gaussian density function over a model parameter.
 */
class Gaussian : public Density
{
    public:

        Gaussian(double mu=0, double sigma=1) throw (jwscxx::base::Arg_error);

        Model::Parameter::Context::Gaussian* clone() const;

        double mu() const;

        double sigma() const;

        void set(double mu, double sigma) throw (jwscxx::base::Arg_error);

        double pdf(double x) const;

        double pdf(double x, double mu, double sigma) const;

        double log_pdf(double x) const;

        double log_pdf(double x, double mu, double sigma) const;

        double trunc_log_pdf(double x, double min, double max) const;

        double trunc_log_pdf
        (
            double x, 
            double mu, 
            double sigma, 
            double min, 
            double max
        ) 
        const;

        virtual double sample(const Model::Parameter::Context::Range* range) 
            const;

        virtual double sample
        (
            double                                  mu,
            double                                  sigma,
            const Model::Parameter::Context::Range* range
        ) 
        const;

    protected:

        double _mu;

        double _sigma;
};


/**
 * @class Model::Parameter::Context::Circular_gaussian
 *
 * @brief Gaussian density function over a model parameter.
 */
class Circular_gaussian : public Gaussian
{
    public:

        Circular_gaussian(double mu=0, double sigma=1) 
            throw (jwscxx::base::Arg_error);

        double sample(const Model::Parameter::Context::Range* range) const;

        double sample
        (
            double                                  mu, 
            const Model::Parameter::Context::Range* range
        ) 
        const;
};


/**
 * @class Model::Parameter::Context::Gaussian_gamma
 *
 * @brief Gaussian-Gamma joint density function over a pair of mean and variance
 * parameters.
 */
class Gaussian_gamma : public Density
{
    public:

        Gaussian_gamma
        (
            double mu=0, 
            double lambda=1, 
            double alpha=1, 
            double beta=1
        ) 
        throw (jwscxx::base::Arg_error);

        Model::Parameter::Context::Gaussian_gamma* clone() const;

        double mu() const;

        double lambda() const;

        double alpha() const;

        double beta() const;

        void set(double mu, double lambda, double alpha, double beta) 
            throw (jwscxx::base::Arg_error);

        double pdf(double mu) const;

        double pdf(double mu, double sigma) const;

        double log_pdf(double mu) const;

        double log_pdf(double mu, double sigma) const;

        /** @brief Evaluates a truncated version of the log pdf function. */
        double trunc_log_pdf
        (
            double mu,
            double mu_min, 
            double mu_max,
            double sigma
        )
        const;

        double sample(const Model::Parameter::Context::Range* range) const;

        void sample
        (
            double*                                 mu_out,
            const Model::Parameter::Context::Range* mu_range
        )
        const;

        void sample
        (
            double*                                 mu_out,
            const Model::Parameter::Context::Range* mu_range,
            double*                                 sigma_out
        )
        const;

        void sample
        (
            double*                                 mu_out,
            double                                  mu_in,
            const Model::Parameter::Context::Range* mu_range,
            double*                                 sigma_out
        )
        const;

    protected:

        /** @brief Mean value. */
        double _mu;

        /** @brief Variance scale factor in the Gaussian. */
        double _lambda;

        /** @brief Gamma shape parameter. */
        double _alpha;

        /** @brief Gamma scale parameter. */
        double _beta;
};


/** 
 * @class Model::Parameter::Context::Dynamics
 *
 * @brief Stochastic dynamics sampling parameters.
 */
class Dynamics : public jwscxx::base::Cloneable
{
    public:

        Dynamics(double delta_t=1, double grad_eta=1) 
            throw (jwscxx::base::Arg_error);

        Model::Parameter::Context::Dynamics* clone() const;

        double delta_t() const;

        double grad_eta() const;

        void set(double delta_t, double grad_eta) 
            throw (jwscxx::base::Arg_error);

    private:

        double _delta_t;

        double _grad_eta;
};


    // Model::Parameter::Context
    public:

        /**
         * @name Standard
         *
         * Standard class members.
         * @{
         */

        /** 
         * @brief Constructs a new model parameter context with default 
         * values.
         */
        Context();

        /** @brief Constructs a new model parameter context from another. */
        Context(const Context& ctx);

        /** @brief Deletes this context. */
        ~Context();

        /** @brief Assigns one context to another. */
        Model::Parameter::Context& operator= 
        (
            const Model::Parameter::Context& ctx
        );

        /** @brief Copies this context. */
        Model::Parameter::Context* clone() const;

        /*@}*/


        /**
         * @name Value
         *
         * Parameter value members.
         * @{
         */

        /** @brief Returns the parameter initial value. */
        double init_value() const;

        /** @brief Sets the parameter initial value. */
        void init_value(double value);

        /*@}*/


        /**
         * @name Range
         *
         * Parameter range members.
         * @{
         */

        /** @brief Returns the parameter range. */
        const Model::Parameter::Context::Range& range() const;

        /** @brief Sets the parameter range. */
        void range(const Model::Parameter::Context::Range& range);

        /** @brief Tests a parameter value against its range. */
        void test_range(double x) const throw (jwscxx::base::Arg_error);

        /*@}*/


        /**
         * @name Likelihood
         *
         * Instance parameter likelihood members.
         * @{
         */

        /** @brief Returns the likelihood of an instance parameter. */
        const Model::Parameter::Context::Gaussian& likelihood() const;

        /** @brief Sets the likelihood of an instance parameter. */
        void likelihood(const Model::Parameter::Context::Gaussian& likelihood);

        /** 
         * @brief Calculates the Gaussian likelihood of an instance parameter.
         */
        double calc_likelihood(double x) const;

        /** 
         * @brief Calculates the mean-scaled Gaussian likelihood of an instance
         * parameter.
         */
        double calc_likelihood(double x, double s) const;

        /** 
         * @brief Calculates the Gaussian log likelihood of an instance
         * parameter.
         */
        double calc_log_likelihood(double x) const;

        /** 
         * @brief Calculates the mean-scaled Gaussian log likelihood of an
         * instance parameter.
         */
        double calc_log_likelihood(double x, double s) const;

        /** 
         * @brief Calculates the truncated Gaussian log probability of an
         * instance parameter.
         */
        double calc_trunc_log_likelihood(double x, double min, double max) 
            const;

        /** 
         * @brief Calculates the mean-scaled truncated Gaussian log probability
         * of an instance parameter.
         */
        double calc_trunc_log_likelihood
        (
            double x, 
            double s, 
            double min, 
            double max
        ) 
        const;

        /** @brief Samples the Gaussian likelihood of an instance parameter. */
        double sample_likelihood() const;

        /** 
         * @brief Samples the mean-scaled Gaussian likelihood of an instance
         * parameter.
         */
        double sample_likelihood(double s) const;

        /** 
         * @brief Samples the truncated Gaussian likelihood of an instance 
         * parameter.
         */
        double sample_trunc_likelihood(double min, double max) const;

        /** 
         * @brief Samples the mean-scaled truncated Gaussian likelihood of an
         * instance parameter.
         */
        double sample_trunc_likelihood(double s, double min, double max) const;

        /*@}*/


        /**
         * @name Prior
         *
         * Category parameter prior probability members.
         * @{
         */

        /** @brief Returns the prior of a category parameter. */
        const Model::Parameter::Context::Gaussian_gamma& prior() const;

        /** @brief Sets the prior of a category parameter. */
        void prior(const Model::Parameter::Context::Gaussian_gamma& prior);

        /** @brief Sets the prior of a category parameter. */
        void ratio_prior
        (
            const Model::Parameter::Context::Gaussian_gamma& prior
        );

        /** 
         * @brief Calculates the Gaussian-Gamma probability of a category
         * parameter.
         */
        double calc_prior_prob(double mean, double stdev) const;

        /** 
         * @brief Calculates the Gaussian-Gamma log probability of a category
         * parameter.
         */
        double calc_prior_log_prob(double mean, double stdev) const;

        /** 
         * @brief Calculates the Gaussian-Gamma log probability of a category
         * parameter.
         */
        double calc_trunc_prior_log_prob
        (
            double mean,
            double mean_min, 
            double mean_max,
            double stdev
        ) 
        const;

        /** @brief Generates a category parameter sample. */
        void sample_prior(double* mean_out, double* stdev_out) const;

        /** @brief Generates a category parameter sample. */
        void sample_prior
        (
            double* mean_out, 
            double* stdev_out, 
            double  mean_in
        ) 
        const;

        /** @brief Returns whether the parameter uses a ratio prior. */
        bool has_ratio_prior() const;

        /** @brief Sets whether the parameter uses a ratio prior. */
        void has_ratio_prior(bool b);

        /*@}*/


        /**
         * @name Proposal
         *
         * Proposal probability distribution members.
         * @{
         */

        /** @brief Returns the proposal of an instance parameter. */
        const Model::Parameter::Context::Density& proposal() const;

        /** @brief Sets the proposal of an instance parameter. */
        void proposal
        (
            const Model::Parameter::Context::Density& proposal
        );

        /** 
         * @brief Calculates the proposal probability of an instance parameter.
         */
        double calc_proposal_prob(double x) const;

        /** 
         * @brief Calculates the proposal log probability of an instance
         * parameter.
         */
        double calc_proposal_log_prob(double x) const;

        /** 
         * @brief Calculates the proposal log probability of an instance
         * parameter.
         */
        double calc_trunc_proposal_log_prob
        (
            double x, 
            double min, 
            double max
        ) 
        const;

        /** @brief Samples the proposal of an instance parameter. */
        double sample_proposal() const;

        /** @brief Samples the proposal of an instance parameter. */
        double sample_trunc_proposal(double min, double max) const;

        /*@}*/


        /**
         * @name Gaussian Proposal
         *
         * Gaussian proposal probability distribution members.
         * @{
         */

        /** @brief Returns the Gaussian proposal of an instance parameter. */
        const Model::Parameter::Context::Gaussian& gauss_proposal() const;

        /** @brief Sets the Gaussian proposal of an instance parameter. */
        void gauss_proposal
        (
            const Model::Parameter::Context::Gaussian& proposal
        );

        /** 
         * @brief Calculates the Gaussian proposal probability of an instance
         * parameter.
         */
        double calc_gauss_proposal_prob(double x) const;

        /** 
         * @brief Calculates the Gaussian proposal log probability of an
         * instance parameter.
         */
        double calc_gauss_proposal_log_prob(double x) const;

        /** 
         * @brief Calculates the Gaussian proposal log probability of an
         * instance parameter.
         */
        double calc_trunc_gauss_proposal_log_prob
        (
            double x, 
            double min, 
            double max
        ) 
        const;

        /** @brief Samples the Gaussian proposal of an instance parameter. */
        double sample_gauss_proposal() const;

        /** 
         * @brief Samples the Gaussian proposal of an instance parameter given
         * its mean.
         */
        double sample_gauss_proposal(double mean) const;

        /** @brief Samples the Gaussian proposal of an instance parameter. */
        double sample_trunc_gauss_proposal(double min, double max) const;

        /** 
         * @brief Samples the Gaussian proposal of an instance parameter given
         * its mean.
         */
        double sample_trunc_gauss_proposal(double mean, double min, double max) 
            const;

        /*@}*/


        /**
         * @name Dynamics
         *
         * Stachastic dynamics parameter members.
         * @{
         */

        /** 
         * @brief Returns the stochastic dynamics parameters for the instance
         * parameter values.
         */
        const Model::Parameter::Context::Dynamics& value_dynamics() const;

        /** 
         * @brief Sets the stochastic dynamics parameters for the instance
         * parameter values.
         */
        void value_dynamics
        (
            const Model::Parameter::Context::Dynamics& dynamics
        );

        /** 
         * @brief Returns the stochastic dynamics parameters for the mean
         * category parameters.
         */
        const Model::Parameter::Context::Dynamics& mean_dynamics() const;

        /** 
         * @brief Sets the stochastic dynamics parameters for the mean category
         * parameters.
         */
        void mean_dynamics
        (
            const Model::Parameter::Context::Dynamics& dynamics
        );

        /** 
         * @brief Returns the stochastic dynamics parameters for the standard
         * deviation category parameters.
         */
        const Model::Parameter::Context::Dynamics& stdev_dynamics() const;

        /** 
         * @brief Sets the stochastic dynamics parameters for the standard
         * deviation category parameters.
         */
        void stdev_dynamics
        (
            const Model::Parameter::Context::Dynamics& dynamics
        );

        /** 
         * @brief Sets the stochastic dynamics parameters for the mean category
         * parameters.
         */
        void ratio_mean_dynamics
        (
            const Model::Parameter::Context::Dynamics& dynamics
        );

        /** 
         * @brief Sets the stochastic dynamics parameters for the standard
         * deviation category parameters.
         */
        void ratio_stdev_dynamics
        (
            const Model::Parameter::Context::Dynamics& dynamics
        );

        /*@}*/

        /**
         * @brief Minimum allowable standard deviation.
         *
         * Anything less than this causes precision problems.
         *
         * @todo Should be a run-time parameter.
         */
        static const double min_stdev;

        /**
         * @brief Minimum allowable parameter ratio value.
         *
         * @todo Should be a run-time parameter.
         */
        static const double min_ratio;

        /**
         * @brief Maximum allowable parameter ratio value.
         *
         * @todo Should be a run-time parameter.
         */
        static const double max_ratio;


    private:

        /**
         * @name Ratio Prior
         *
         * Ratio category parameter prior probability members.
         * @{
         */

        /** @brief Generates a category parameter sample. */
        void sample_ratio_prior(double* mean_out, double* stdev_out) const;

        /** @brief Generates a category parameter sample. */
        void sample_ratio_prior
        (
            double* mean_out, 
            double* stdev_out, 
            double  mean_in
        ) 
        const;

        /*@}*/


    private: 

        /** @brief Initial instance parameter value. */
        double _init_value;

        /** @brief Range over the instance and mean category parameters. */
        Range* _range;

        /** 
         * @brief Likelihood function over an instance parameter.
         *
         * The mean and standard deviation of the Gaussian over a parameter
         * instance are category-level parameters shared across a set of
         * instances. A copy of these category parameters are kept in this and
         * other similar instance contexts.
         */
        Gaussian* _likelihood;

        /** 
         * @brief Prior probability density over the category parameters in the
         * likelihood.
         *
         * The category parameters in the likelihood comprise a Gaussian mean
         * and standard deviation.
         */
        Gaussian_gamma* _prior;

        /** 
         * @brief Prior probability density over the category parameters in the
         * likelihood.
         *
         * The category parameters in the likelihood comprise a Gaussian mean
         * and standard deviation.
         *
         * The mean value in the likelihood is assumed to be a ratio of instance
         * parameters.
         */
        Gaussian_gamma* _ratio_prior;

        /** 
         * @brief Proposal probability density over the instance parameter
         * values.
         *
         * An arbitrary density function.
         */
        Density* _proposal;

        /** 
         * @brief Proposal probability density over the category parameter
         * values.
         */
        Gaussian* _gauss_proposal;

        /** 
         * @brief Stochastic dynamics sampling parameters for the instance
         * parameter values.
         */
        Dynamics* _value_dynamics;

        /** 
         * @brief Stochastic dynamics sampling parameters for the mean category
         * parameters.
         */
        Dynamics* _mean_dynamics;

        /** 
         * @brief Stochastic dynamics sampling parameters for the standard
         * deviation category parameters.
         */
        Dynamics* _stdev_dynamics;

        /** 
         * @brief Stochastic dynamics sampling parameters for the mean category
         * parameters.
         */
        Dynamics* _ratio_mean_dynamics;

        /** 
         * @brief Stochastic dynamics sampling parameters for the standard
         * deviation category parameters.
         */
        Dynamics* _ratio_stdev_dynamics;

        /** 
         * @brief Whether the parameter uses a ratio prior. 
         *
         * Initially set to false.
         */
        bool _has_ratio_prior;
};


/**
 * @class Model::Parameter::Pointer
 *
 * @brief Pointer to a Model::Parameter.
 *
 * Constains a pointer to a model parameter and member functions to serialize
 * and deserialize its value.
 */
class Pointer : public jwscxx::base::Readable,
                public jwscxx::base::Writeable,
                public jwscxx::base::Cloneable
{
    public:

        Pointer(Model::Parameter* p);

        Pointer(const Model::Parameter::Pointer& p);

        virtual ~Pointer();

        virtual Model::Parameter::Pointer* clone() const;

        void read(std::istream& in) 
            throw (jwscxx::base::IO_error, jwscxx::base::Arg_error);

        void write(std::ostream& out) const throw (jwscxx::base::IO_error);

        virtual const char* type_name() const;

        Model::Parameter* ptr() const;

        void ptr(Model::Parameter* p);

        /** 
         * @brief Replaces the Parameter pointer address with its index in a
         * collection of models.
         */
        void serialize(const std::vector<Model*>& models);

        /** 
         * @brief Replaces a Parameter pointer index in a collection of models
         * with its address.
         */
        void deserialize(const std::vector<Model*>& models);

        /** 
         * @brief Updates the Parameter pointer addresses from one set of models
         * to another.
         */
        void update_pointers
        (
            const std::vector<Model*>&       new_models,
            const std::vector<const Model*>& old_models
        );


    protected:

        /** @brief The Parameter pointer. */
        Model::Parameter* _ptr;
};


class Compare
{
    public:

        bool operator() (const Model::Parameter* p1, const Model::Parameter* p2)
            const;
};


    // Model::Parameter
    public:

        /**
         * @name Standard
         *
         * Standard class members.
         * @{
         */

        /** @brief Constructs a new parameter. */
        Parameter
        (
            size_t                           id,
            const char*                      name, 
            double                           value, 
            const Model::Parameter::Context* ctx,
            Model*                           model
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Constructs a new parameter from another. */
        Parameter(const Model::Parameter& p);

        /** @brief Deletes this parameter. */
        virtual ~Parameter();

        /** @brief Performs a (deep) copy to this parameter. */
        virtual Model::Parameter& operator=(const Model::Parameter& p);

        /** @brief Clones this parameter. */
        virtual Model::Parameter* clone() const;

        /** @brief Reads the value of this parameter. */
        virtual void read(std::istream& in) 
            throw (jwscxx::base::IO_error, jwscxx::base::Arg_error);

        /** @brief Writes the value of this parameter. */
        virtual void write(std::ostream& out) const
            throw (jwscxx::base::IO_error);

        /** @brief Returns the id of this parameter. */
        size_t id() const;

        /** @brief Returns the name of this parameter. */
        const char* name() const;

        /*@}*/


        /**
         * @name Value
         *
         * Instance parameter value members.
         * @{
         */

        /** @brief Returns the parameter value. */
        double value() const;

        /** @brief Sets the parameter value. */
        void value(double value);

        /** @brief Tests a parameter value against its range. */
        void test_value(double value) const throw (jwscxx::base::Arg_error);

        /** @brief Returns the scratch value stored. */
        double scratch() const;

        /** @brief Sets the scratch value. */
        void scratch(double value);

        /*@}*/


        /**
         * @name Stats
         *
         * Category parameter members.
         * @{
         */

        /** @brief Returns the parameter mean. */
        double mean() const;

        /** @brief Sets the parameter mean. */
        void mean(double mean);

        /** @brief Tests a parameter mean against its range. */
        void test_mean(double mean) const throw (jwscxx::base::Arg_error);

        /** @brief Returns the parameter standard deviation. */
        double stdev() const;

        /** @brief Sets the parameter standard deviation. */
        void stdev(double stdev);

        /** @brief Tests a parameter standard deviation against its range. */
        void test_stdev(double stdev) const throw (jwscxx::base::Arg_error);

        /** @brief Returns the parameter statistics. */
        void stats(double* mean_out, double* stdev_out) const;

        /** @brief Sets the parameter statistics. */
        void stats(double mean, double stdev);

        /** @brief Tests a parameter statistics against its range. */
        void test_stats(double mean, double stdev) const 
            throw (jwscxx::base::Arg_error);

        /*@}*/


        /**
         * @name Ratio Prior
         *
         * Ratio category parameter prior probability members.
         * @{
         */

        /** @brief Returns whether the parameter uses a ratio prior. */
        bool has_ratio_prior() const;

        /** @brief Sets whether the parameter uses a ratio prior. */
        void has_ratio_prior(bool b);

        /*@}*/


        /**
         * @name Context
         *
         * Parameter context members.
         * @{
         */

        /** @brief Returns the parameter context. */
        const Model::Parameter::Context* context() const;

        /** @brief Sets the parameter context. */
        void context(const Model::Parameter::Context* ctx, bool clip=false)
            throw (jwscxx::base::Arg_error);

        /*@}*/


        /**
         * @name Model
         *
         * Parameter model members.
         * @{
         */

        /** @brief Returns the model containing this parameter. */
        Model* model() const;

        /** @brief Sets the model containing thsi parameter. */
        void model(Model* m);

        /*@}*/


        /**
         * @name Pointers
         *
         * Internal Parameter pointer management members.
         * @{
         */

        /** 
         * @brief Serializes pointers in this Parameter to indices in the @em
         * models collection.
         */
        virtual void serialize(const std::vector<Model*>& models);

        /** 
         * @brief Deserializes this Parameter from indices in the @em models
         * collection to pointers.
         */
        virtual void deserialize(const std::vector<Model*>& models);

        /** 
         * @brief Updates the Parameter dependency pointers from one set of
         * models to another.
         */
        virtual void update_pointers
        (
            const std::vector<Model*>&       old_models,
            const std::vector<const Model*>& new_models
        );

        /*@}*/


    private:

        /** @brief Unique id (within a model) of this Parameter. */
        size_t _id;

        /** @brief Short name of this Parameter. */
        const char* _name;

        /** @brief Instance parameter value. */
        double _value;

        /** @brief Category parameter mean over the instance parameter value. */
        double _mean;

        /** 
         * @brief Category parameter standard deviation over the instance
         * parameter value.
         */
        double _stdev;

        /** 
         * @brief Whether the parameter uses a ratio prior. 
         *
         * Initially set to false.
         */
        bool _has_ratio_prior;

        /** 
         * @brief Context of this Parameter.
         *
         * Each instance parameter has a copy of the same context, which
         * contains the category parameters.
         */
        Model::Parameter::Context* _ctx;

        /** @brief The Model containing this Parameter. */
        Model* _model;

        /** @brief Scratch value storage. */
        double _scratch;
};


/**
 * @class Model::Pointer
 *
 * @brief Pointer to a Model.
 *
 * Constains a pointer to a model parameter and member functions to serialize
 * and deserialize its value.
 */
class Pointer : public jwscxx::base::Readable,
                public jwscxx::base::Writeable,
                public jwscxx::base::Cloneable
{
    public:

        Pointer(Model* m);

        Pointer(const Model::Pointer& p);

        virtual ~Pointer();

        virtual Model::Pointer* clone() const;

        virtual void read(std::istream& in) 
            throw (jwscxx::base::IO_error, jwscxx::base::Arg_error);

        virtual void write(std::ostream& out) const 
            throw (jwscxx::base::IO_error);

        virtual const char* type_name() const;

        // Tried this as virtual and caused many problems because we were
        // comparing Component pointers to Model pointers, which may not have
        // the same address in C++.
        Model* ptr() const;

        void ptr(Model* m);

        /** 
         * @brief Replaces the Parameter pointer address with its index in a
         * collection of models.
         */
        void serialize(const std::vector<Model*>& models);

        /** 
         * @brief Replaces a Parameter pointer index in a collection of models
         * with its address.
         */
        void deserialize(const std::vector<Model*>& models);

        /** 
         * @brief Updates the Parameter pointer addresses from one set of models
         * to another.
         */
        void update_pointers
        (
            const std::vector<Model*>&       new_models,
            const std::vector<const Model*>& old_models
        );


    protected:

        /** @brief The Parameter pointer. */
        Model* _ptr;
};


    // Model
    public:

        /**
         * @name Standard
         *
         * Standard class members.
         * @{
         */

        /** @brief Constructs a new model. */
        Model(const char* name, const Model::Context* ctx);

        /** @brief Constructs a new model from another. */
        Model(const Model& m);

        /** @brief Deletes this model. */
        virtual ~Model();

        /** @brief Assigns a model to another. */
        virtual Model& operator= (const Model& m);

        /** @brief Copies a model. */
        virtual Model* clone() const = 0;

        /** @brief Reads a model from an input stream. */
        virtual void read(std::istream& in) 
            throw (jwscxx::base::IO_error, jwscxx::base::Arg_error);

        /** @brief Writes a model to an output stream. */
        virtual void write(std::ostream& out) const
            throw (jwscxx::base::IO_error);

        /** @brief Returns the model context. */
        virtual const Model::Context* context() const;

        /** @brief Returns the model name. */
        const char* name() const;

        /*@}*/


        /**
         * @name Variable Parameters
         *
         * Variable model parameter members.
         * @{
         */

        /** @brief Returns an indexed Parameter. */
        const Model::Parameter* const_var_param(uint32_t i) const 
            throw (jwscxx::base::Arg_error);

        /** @brief Returns the number of variable parameters. */
        virtual uint32_t num_var_params() const;

        /** @brief Returns an instance parameter name. */
        virtual const char* get_var_param_name(uint32_t d) const;

        /** @brief Returns an instance parameter name. */
        virtual std::vector<const char*> get_var_param_names() const;

        /** @brief Returns an instance parameter value. */
        virtual double get_var_param_value(uint32_t d) const;

        /** @brief Returns the instance parameter values. */
        virtual void get_var_param_values(jwsc::Vector_d** p_out) const;

        /** @brief Sets the instance parameter values. */
        virtual void set_var_param_values(const jwsc::Vector_d* p) 
            throw (jwscxx::base::Arg_error);

        /** @brief Adds to the instance parameter values. */
        virtual void add_var_param_values(const jwsc::Vector_d* p)
            throw (jwscxx::base::Arg_error);

        /** @brief Returns the parameter dynamics gradient steps. */
        virtual void get_var_param_values_dynamics_grad_eta
        (
            jwsc::Vector_d** p_out
        )
        const;

        /** @brief Returns the parameter dynamics gradient steps. */
        virtual void get_var_param_values_dynamics_delta_t
        (
            jwsc::Vector_d** p_out
        )
        const;

        /** @brief Prints variable parameter values. */
        virtual void print_var_param_values
        (
            uint32_t      iteration, 
            std::ostream& out
        )
        const;

        /** 
         * @brief Returns the category parameter statistics over an instance
         * parameter value.
         */
        virtual void get_var_param_stats
        (
            double*  mean_out,
            double*  stdev_out,
            uint32_t d
        ) 
        const;

        /** 
         * @brief Returns the category parameter statistics over the instance
         * parameter values.
         */
        virtual void get_var_param_stats
        (
            jwsc::Vector_d** means_out,
            jwsc::Vector_d** stdevs_out
        ) 
        const;

        /** 
         * @brief Sets the category parameter statistics over the instance
         * parameter values.
         */
        virtual void set_var_param_stats
        (
            const jwsc::Vector_d* means,
            const jwsc::Vector_d* stdevs
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Returns the parameter dynamics gradient steps. */
        virtual void get_var_param_stats_dynamics_grad_eta
        (
            jwsc::Vector_d** means_out,
            jwsc::Vector_d** stdevs_out
        )
        const;

        /** @brief Returns the parameter dynamics gradient steps. */
        virtual void get_var_param_stats_dynamics_delta_t
        (
            jwsc::Vector_d** means_out,
            jwsc::Vector_d** stdevs_out
        )
        const;

        /** @brief Prints variable parameter statistics. */
        virtual void print_var_param_stats
        (
            uint32_t      iteration, 
            std::ostream& out
        )
        const;

        /*@}*/


        /**
         * @name Likelihood
         *
         * Instance parameter likelihood members.
         * @{
         */

        /** 
         * @brief Calculates the log likelihood of the instance parameters.
         */
        virtual double calc_var_param_log_likelihood() const;

        /** 
         * @brief Generates instance parameter values from the likelihood.
         */
        virtual void sample_var_param_likelihood(jwsc::Vector_d** p_out);

        /*@}*/


        /**
         * @name Prior
         *
         * Category parameter prior probability members.
         * @{
         */

        /** 
         * @brief Calculates the log prior probability of category parameters.
         */
        virtual double calc_var_param_prior_log_prob() const;

        /** @brief Generates category parameters from the prior. */
        virtual void sample_var_param_prior
        (
            jwsc::Vector_d** means_out,
            jwsc::Vector_d** stdevs_out
        ) 
        const;

        /*@}*/


        /**
         * @name Category Proposal
         *
         * Category proposal probability distribution members.
         * @{
         */

        virtual void sample_var_param_cat_proposal
        (
            jwsc::Vector_d** means_out,
            jwsc::Vector_d** stdevs_out,
            const jwsc::Vector_d* means_in
        ) 
        const;

        /*@}*/


        /**
         * @name Proposal
         *
         * Proposal probability distribution members.
         * @{
         */

        /** 
         * @brief Calculates the proposal log probability of instance
         * parameters.
         */
        virtual double calc_var_param_proposal_log_prob() const;

        /** 
         * @brief Calculates the Gaussian proposal log probability of instance
         * parameters.
         */
        virtual double calc_var_param_gauss_proposal_log_prob() const;

        /** 
         * @brief Generates instance parameter values from the proposal
         * distribution.
         */
        virtual void sample_var_param_proposal(jwsc::Vector_d** p_out) const;

        /** 
         * @brief Generates instance parameter values from the Gaussian proposal
         * distribution.
         */
        void sample_var_param_gauss_proposal(jwsc::Vector_d** p_out) const;

        /** 
         * @brief Generates instance parameter values from the Gaussian proposal
         * distribution.
         */
        virtual void sample_var_param_gauss_proposal
        (
            jwsc::Vector_d**      p_out,
            const jwsc::Vector_d* means
        ) 
        const;

        /** 
         * @brief Returns the instance parameter Gaussian proposal sigmas as a
         * point.
         */
        virtual void get_var_param_gauss_proposal_sigma
        (
            jwsc::Vector_d** p_out
        )
        const;

        /*@}*/


        /**
         * @name Pointers
         *
         * Pointer management members.
         * @{
         */

        /** 
         * @brief Serializes Model and Parameter pointers in this Model using
         * indices in the @em models collection.
         */
        virtual void serialize(const std::vector<Model*>& models);

        /** 
         * @brief Deserializes Model and Parameter indices in this Model using
         * the @em models collection of pointers.
         */
        virtual void deserialize
        (
            const std::vector<Model*>& models
        );

        /** 
         * @brief Updates the Model and Parameter pointers in this Model from
         * one set of models to another.
         */
        virtual void update_pointers
        (
            const std::vector<Model*>&       new_models,
            const std::vector<const Model*>& old_models
        );

        /*@}*/


        /**
         * @name Graphical Representation
         *
         * Internal graphical representation methods.
         * @{
         */

        /** 
         * @brief Returns whether the graphical representation is outdated and
         * needs updating.
         *
         * This facility enables grouping multiple updates into one call at
         * opportune moments.
         */
        virtual bool graphical_rep_is_outdated() const;

        /** 
         * @brief Sets the graphical representation as outdated and needs
         * updating.
         *
         * This facility enables grouping multiple updates into one call at
         * opportune moments.
         */
        virtual void graphical_rep_outdated();

        /** 
         * @brief Updates the graphical representation of this Model with its
         * current parameter values.
         *
         * This is a callback member function used by the Model::Parameter when
         * a value changes.
         */
        virtual void update_graphical_rep();

        /*@}*/


    protected:

        /**
         * @name Fixed Parameters
         *
         * Fixed model parameter members.
         * @{
         */

        /** @brief Returns the number of fixed Parameter. */
        uint32_t num_fixed_params() const;

        /** @brief Returns an indexed Parameter. */
        Model::Parameter* fixed_param(uint32_t i) const 
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Returns Parameter index.
         *
         * Uses an associative map to provide efficient look-up. If @em p is not
         * a fixed parameter, num_fixed_params() is returned.
         */
        size_t fixed_param(const Model::Parameter* p) const
            throw (jwscxx::base::Arg_error);

        void add_fixed_param(Model::Parameter* p);

         Model::Parameter* remove_fixed_param(Model::Parameter* p);

        /*@}*/


        /**
         * @name Variable Parameters
         *
         * Variable model parameter members.
         * @{
         */

        /** @brief Returns an indexed Parameter. */
        Model::Parameter* var_param(uint32_t i) const 
            throw (jwscxx::base::Arg_error);

        /** @brief Returns Parameter index. */
        size_t var_param(const Model::Parameter* p) const
            throw (jwscxx::base::Arg_error);

        /** @brief Adds a variable Parameter to this Model. */
        void add_var_param(Model::Parameter* p);

        Model::Parameter* remove_var_param(Model::Parameter* p);

        /*@}*/


        /**
         * @name Derived Parameters
         *
         * Derived model parameter members.
         * @{
         */

        /** @brief Returns the number of derived Parameter. */
        uint32_t num_derived_params() const;

        /** @brief Returns an indexed Parameter. */
        Model::Parameter* derived_param(uint32_t i) const 
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Returns Parameter index.
         *
         * Uses an associative map to provide efficient look-up. If @em p is not
         * a derived parameter, num_derived_params() is returned.
         */
        size_t derived_param(const Model::Parameter* p) const
            throw (jwscxx::base::Arg_error);

        void add_derived_param(Model::Parameter* p);

        Model::Parameter* remove_derived_param(Model::Parameter* p);

        /*@}*/


    private:

        /** @brief Model name. */
        const char* _name;

        /**
         * @name Fixed Parameters
         *
         * Fixed model parameter members.
         * @{
         */

        /** 
         * @brief Vector of fixed Model parameters.
         *
         * Set of Model::Parameter pointers fixed at program initialization.
         */
        std::vector<Model::Parameter*> _fixed_params;

        /** @brief Map of parameter pointers to indices. */
        std::map<const Model::Parameter*, size_t> _fixed_params_index;

        /*@}*/


        /**
         * @name Variable Parameters
         *
         * Variable model parameter members.
         * @{
         */

        /** 
         * @brief Vector of changable Model parameters.
         *
         * Set of Model::Parameter pointers directly inferred from data during
         * learning and recognition.
         */
        std::vector<Model::Parameter*> _var_params;

        /** @brief Map of parameter pointers to indices. */
        std::map<const Model::Parameter*, size_t> _var_params_index;

        /*@}*/


        /**
         * @name Derived Parameters
         *
         * Derived model parameter members.
         * @{
         */

        /** 
         * @brief Vector of derived Model parameters.
         *
         * Set of Model::Parameter pointers indirectly inferred from data during
         * learning and recognition. They are derived from the _var_params.
         */
        std::vector<Model::Parameter*> _derived_params;

        /** @brief Map of parameter pointers to indices. */
        std::map<const Model::Parameter*, size_t> _derived_params_index;

        /*@}*/

        /** 
         * @brief Indicates whether the graphical representation needs
         * updating.
         */
        bool _graphical_rep_outdated;

        /** @brief Model context. */
        const Context* _ctx;
};


#endif
