/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Generic 3D object model.
 *
 * Currently composed of a collection of 3D Component.
 */


#include <config.h>

#include <typeinfo>
#include <stdexcept>
#include <iostream>
#include <sstream>
#include <vector>
#include <algorithm>
#include <cmath>
#include <cstring>
#include <cassert>

#if defined JWSCXX_HAVE_OPENGL
#include <GL/gl.h>
#include <GL/glu.h>
#elif defined JWSCXX_HAVE_OPENGL_FRAMEWORK
#include <OpenGL/OpenGL.h>
#endif

#include <jwsc/base/limits.h>
#include <jwsc/math/constants.h>
#include <jwsc/prob/pmf.h>
#include <jwsc/prob/pdf.h>
#include <jwsc/vector/vector.h>
#include <jwsc/vector/vector_math.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/matrix/matrix_math.h>
#include <jwsc/image/image.h>
#include <jwsc/image/image_io.h>
#include <jwsc/image/edge.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/graphics/polygon.h>

#include "model.h"
#include "view.h"
#include "component.h"
#include "block.h"
#include "block_set.h"
#include "object.h"


#define  AZIMUTH            0.0
#define  AZIMUTH_MIN       -JWSC_PI
#define  AZIMUTH_MAX        JWSC_PI
#define  AZIMUTH_MU         0.0
#define  AZIMUTH_SIGMA      0.1
#define  AZIMUTH_LAMBDA     9.0
#define  AZIMUTH_ALPHA      1.0
#define  AZIMUTH_BETA       0.25
#define  AZIMUTH_DELTA_T    0.00005
#define  AZIMUTH_GRAD_ETA   0.005
#define  AZIMUTH_SIGMA_DELTA_T    0.0001
#define  AZIMUTH_SIGMA_GRAD_ETA   0.005

#define  POSN               0.0
#define  POSN_MIN          -10000
#define  POSN_MAX           10000
#define  POSN_MU            0.0
#define  POSN_SIGMA         10.0
#define  POSN_LAMBDA        100.0
#define  POSN_ALPHA         0.1
#define  POSN_BETA          10.0
#define  POSN_DELTA_T       0.005
#define  POSN_GRAD_ETA      1.0
#define  POSN_SIGMA_DELTA_T   0.0001
#define  POSN_SIGMA_GRAD_ETA  0.5


/* ---------------------------  Object::Context  ---------------------------- */

Object::Context::Context() 
    : Model::Context(), 
      _posn_c(3, Model::Parameter::Context())
{
    typedef Model::Parameter::Context MPContext;
    typedef MPContext::Range          Range;
    typedef MPContext::Dynamics       Dynamics;
    typedef MPContext::Uniform        Uniform;
    typedef MPContext::Gaussian       Gaussian;
    typedef MPContext::Gaussian       Circular_gaussian;
    typedef MPContext::Gaussian_gamma Gaussian_gamma;

    MPContext* c = &_azimuth_c;
    c->range(Range(AZIMUTH_MIN, AZIMUTH_MAX));
    c->init_value(AZIMUTH);
    c->likelihood(Circular_gaussian(AZIMUTH_MU, AZIMUTH_SIGMA));
    c->prior(Gaussian_gamma(AZIMUTH_MU, AZIMUTH_LAMBDA, AZIMUTH_ALPHA,
                AZIMUTH_BETA));
    c->proposal(Uniform());
    c->gauss_proposal(Circular_gaussian(AZIMUTH_MU, AZIMUTH_SIGMA));
    c->value_dynamics(Dynamics(AZIMUTH_DELTA_T, AZIMUTH_GRAD_ETA));
    c->mean_dynamics(Dynamics(AZIMUTH_DELTA_T, AZIMUTH_GRAD_ETA));
    c->stdev_dynamics(Dynamics(AZIMUTH_SIGMA_DELTA_T, AZIMUTH_SIGMA_GRAD_ETA));

    for (uint32_t i = 0; i < 3; i++)
    {
        c = &(_posn_c[ i ]);
        c->range(Range(POSN_MIN, POSN_MAX));
        c->init_value(POSN);
        c->likelihood(Gaussian(POSN_MU, POSN_SIGMA));
        c->prior(Gaussian_gamma(POSN_MU, POSN_LAMBDA, POSN_ALPHA, POSN_BETA));
        c->proposal(Gaussian(POSN_MU, POSN_SIGMA));
        c->gauss_proposal(Gaussian(POSN_MU, POSN_SIGMA));
        c->value_dynamics(Dynamics(POSN_DELTA_T, POSN_GRAD_ETA));
        c->mean_dynamics(Dynamics(POSN_DELTA_T, POSN_GRAD_ETA));
        c->stdev_dynamics(Dynamics(POSN_SIGMA_DELTA_T, POSN_SIGMA_GRAD_ETA));
    }
}

Object::Context::Context(const Object::Context& ctx) 
    : Model::Context(ctx)
{
    *this = ctx;
}

Object::Context::~Context()
{
}

Object::Context& Object::Context::operator= (const Object::Context& ctx)
{
    Model::Context::operator=(ctx);

    _azimuth_c = ctx._azimuth_c;
    _posn_c    = ctx._posn_c;

    return *this;
}

Object::Context* Object::Context::clone() const
{
    return new Object::Context(*this);
}

const Model::Parameter::Context& Object::Context::azimuth_c() const
{
    return _azimuth_c;
}

void Object::Context::azimuth_c(const Model::Parameter::Context& azimuth)
{
    _azimuth_c = azimuth;
}

const Model::Parameter::Context& Object::Context::posn_c(size_t i) const
{
    assert(i < 3);
    return _posn_c[ i ];
}

void Object::Context::posn_c(size_t i, const Model::Parameter::Context& posn)
{
    assert(i < 3);
    _posn_c[ i ] = posn;
}

/* ========================================================================== */








/* -------------------------------  Object  --------------------------------- */

/**
 * Its rotation angle is about a vector eminating from the projection of the
 * position on the x,z-plane and parallel to the y-axis.
 */
Object::Object
(
    const Object::Context*    obj_ctx, 
    const Block::Context*     block_ctx,
    const Block_set::Context* block_set_ctx
) 
    : Model("Object", obj_ctx),
      _azimuth_p(0, "Azimuth", 0, &(obj_ctx->azimuth_c()), this),
      _lowest_component(0, false),
      _components(),
      _projection()
{
    size_t param_id = 1;

    _posn_p.reserve(3);
    _posn_p.push_back(Component::Parameter(param_id++, "X Position",
                obj_ctx->posn_c(0).init_value(), &(obj_ctx->posn_c(0)), this));
    _posn_p.push_back(Component::Parameter(param_id++, "Y Position",
                obj_ctx->posn_c(1).init_value(), &(obj_ctx->posn_c(1)), this));
    _posn_p.push_back(Component::Parameter(param_id++, "Z Position",
                obj_ctx->posn_c(2).init_value(), &(obj_ctx->posn_c(2)), this));

    add_var_param(&_azimuth_p);
    add_var_param(&(_posn_p[0]));
    add_var_param(&(_posn_p[1]));
    add_var_param(&(_posn_p[2]));

    _block_ctx     = block_ctx;
    _block_set_ctx = block_set_ctx;

    _obj_coords_posn    = 0;
    _obj_coords_azimuth = 0;
    _obj_coords         = false;

    azimuth(obj_ctx->azimuth_c().init_value());

    _next_component_id = 1;
}

/**
 * Its rotation angle is about a vector eminating from the projection of the
 * position on the x,z-plane and parallel to the y-axis.
 */
Object::Object
(
    double                    azimuth,
    double                    x,
    double                    y,
    double                    z,
    const Object::Context*    obj_ctx, 
    const Block::Context*     block_ctx,
    const Block_set::Context* block_set_ctx
) 
    : Model("Object", obj_ctx),
      _azimuth_p(0, "Azimuth", 0, &(obj_ctx->azimuth_c()), this),
      _lowest_component(0, false),
      _components(),
      _projection()
{
    size_t param_id = 1;

    _posn_p.reserve(3);
    _posn_p.push_back(Component::Parameter(param_id++, "X Position", x,
                &(obj_ctx->posn_c(0)), this));
    _posn_p.push_back(Component::Parameter(param_id++, "Y Position", y,
                &(obj_ctx->posn_c(1)), this));
    _posn_p.push_back(Component::Parameter(param_id++, "Z Position", z,
                &(obj_ctx->posn_c(2)), this));

    add_var_param(&_azimuth_p);
    add_var_param(&(_posn_p[0]));
    add_var_param(&(_posn_p[1]));
    add_var_param(&(_posn_p[2]));

    _block_ctx     = block_ctx;
    _block_set_ctx = block_set_ctx;

    _obj_coords_posn    = 0;
    _obj_coords_azimuth = 0;
    _obj_coords         = false;

    this->azimuth(azimuth);

    _next_component_id = 1;
}

/** @param  obj  Object to copy into this one. */
Object::Object(const Object& obj) 
    : Model(obj), 
      _azimuth_p(obj._azimuth_p),
      _posn_p(obj._posn_p),
      _lowest_component(obj._lowest_component),
      _components(),
      _projection(obj._projection)
{
    add_var_param(&_azimuth_p);
    add_var_param(&(_posn_p[0]));
    add_var_param(&(_posn_p[1]));
    add_var_param(&(_posn_p[2]));

    _obj_coords_posn = 0;

    *this = obj;
}

Object::~Object() 
{ 
    for (size_t i = 0; i < _components.size(); i++)
    {
        delete _components[ i ];
    }

    free_vector_d(_obj_coords_posn);
}

/** 
 * Performs a deep copy of this Object.
 *
 * @param  obj  Object to copy into this one.
 *
 * @return A reference to this Object.
 */
Object& Object::operator=(const Object& obj)
{
    Model::operator=(obj);

    _lowest_component   = obj._lowest_component;
    _projection         = obj._projection;
    _block_ctx          = obj._block_ctx;
    _block_set_ctx      = obj._block_set_ctx;
    _obj_coords_azimuth = obj._obj_coords_azimuth;
    _obj_coords         = obj._obj_coords;
    _next_component_id  = obj._next_component_id;
    _used_component_ids = obj._used_component_ids;

    copy_vector_d(&_obj_coords_posn, obj._obj_coords_posn);

    _azimuth_p = obj._azimuth_p;
    _azimuth_p.model(this);

    _posn_p = obj._posn_p;
    for (size_t i = 0; i < 3; i++)
    {
        _posn_p[ i ].model(this);
    }

    for (size_t i = 0; i < num_components(); i++)
    {
        delete component(i);
    }

    _components.clear();
    _components.reserve(obj.num_components());

    std::vector<Model*> new_models(1, this);
    std::vector<const Model*> old_models(1, &obj);

    for (size_t i = 0; i < obj.num_components(); i++)
    {
        _components.push_back(obj.component(i)->clone());
        component(i)->object(this);

        new_models.push_back(component(i));
        old_models.push_back(obj.component(i));
    }

    for (size_t i = 0; i < new_models.size(); i++)
    {
        new_models[ i ]->update_pointers(new_models, old_models);
    }

    return *this;
}

/** @return A new copy of this Object. */
Object* Object::clone() const
{
    return new Object(*this);
}

/**
 * The file format is ASCII.
 *
 * @param  in  Input stream to read this Object from.
 *
 * @throw  jwscxx::base::IO_error   Could not read from @em in.
 * @throw  jwscxx::base::Arg_error  Invalid arguments in file to read from.
 */
void Object::read(std::istream& in) 
    throw (jwscxx::base::IO_error, jwscxx::base::Arg_error)
{
    using jwscxx::base::Readable;
    using jwscxx::base::Arg_error;
    using std::istringstream;

    assert(!in_object_coords());

    Model::read(in);

    istringstream ist;

    const char* field_value;
    uint32_t N;

    if (!(field_value = read_field_value(in, "Num Components")))
    {
        throw Arg_error("Missing Num Components field");
    }
    ist.str(field_value);
    ist >> N;
    if (ist.fail())
    {
        throw Arg_error("Invalid Num Components");
    }
    ist.clear(std::ios_base::goodbit);

    if (!(field_value = read_field_value(in, "Next Component ID")))
    {
        throw Arg_error("Missing Next Component ID field");
    }
    ist.str(field_value);
    ist >> _next_component_id;
    if (ist.fail())
    {
        throw Arg_error("Invalid next component ID");
    }
    ist.clear(std::ios_base::goodbit);

    _lowest_component.read(in);

    for (size_t i = 0; i < num_components(); i++)
    {
        delete _components[ i ];
    }
    _components.clear();

    for (size_t n = 0; n < N; n++)
    {
        std::streampos in_posn = in.tellg();

        Component* c = 0;
        const char* type_name;

        if (!(type_name = Readable::read_field_value(in, "Type")))
        {
            throw Arg_error("Missing Type field");
        }

        in.seekg(in_posn);

        if (strcmp(type_name, typeid(Block).name()) == 0)
        {
            c = new Block(_block_ctx);
        }
        else if (strcmp(type_name, typeid(Quartet).name()) == 0)
        {
            c = new Quartet(_block_set_ctx);
        }
        else if (strcmp(type_name, typeid(Pair).name()) == 0)
        {
            c = new Pair(Pair::WIDTH, _block_set_ctx);
        }
        else
        {
            std::ostringstream ost;
            ost << "Unknown component type '" << type_name << "'";
            throw Arg_error(ost.str());
        }

        c->read(in);
        c->object(this);
        _components.push_back(c);
    }

    std::vector<Model*> models(1, this);
    for (size_t i = 0; i < num_components(); i++)
    {
        models.push_back(component(i));
    }

    for (size_t i = 0; i < models.size(); i++)
    {
        models[ i ]->deserialize(models);
    }

    _used_component_ids.clear();
    for (size_t i = 0; i < num_components(); i++)
    {
        _used_component_ids.push_back(_components[ i ]->id());
    }
    sort(_used_component_ids.begin(), _used_component_ids.end());
}

/**
 * The file format is ASCII.
 *
 * @param  out  Output stream to write this Object to.
 *
 * @throw  jwscxx::base::IO_error  Could not write to @em out.
 */
void Object::write(std::ostream& out) const throw (jwscxx::base::IO_error)
{
    assert(!in_object_coords());

    Object* obj = this->clone();
    std::vector<Model*> models(1, obj);

    for (size_t i = 0; i < obj->num_components(); i++)
    {
        models.push_back(obj->component(i));
    }

    obj->serialize(models);
    for (size_t i = 1; i < models.size(); i++)
    {
        Model* m = models[ i ];
        m->serialize(models);
    }

    obj->Model::write(out);
    out << "Num Components: " << obj->num_components() << std::endl;
    out << "Next Component ID: " << obj->_next_component_id << std::endl;
    obj->_lowest_component.write(out);

    for (size_t i = 0; i < obj->num_components(); i++)
    {
        obj->component(i)->write(out);;
    }

    delete obj;
}

const Object::Context* Object::context() const
{
    return dynamic_cast<const Object::Context*>(Model::context());
}

uint32_t Object::num_var_params() const
{
    uint32_t D = Model::num_var_params();

    for (size_t i = 0; i < _components.size(); i++)
    {
        D += _components[ i ]->num_var_params();
    }

    return D;
}

const char* Object::get_var_param_name(uint32_t d) const
{
    assert(d < num_var_params());
    std::vector<const char*> names = get_var_param_names();
    return names[ d ];
}

std::vector<const char*> Object::get_var_param_names() const
{
    std::vector<const char*> names = Model::get_var_param_names();

    for (size_t i = 0; i < _components.size(); i++)
    {
        std::vector<const char*> c_names = 
            _components[ i ]->get_var_param_names();

        names.insert(names.end(), c_names.begin(), c_names.end());
    }

    return names;
}

double Object::get_var_param_value(uint32_t d) const
{
    using namespace jwsc;

    assert(d < num_var_params());
    uint32_t D = 0;

    D = Model::num_var_params();
    if (d < D)
    {
        return var_param(d)->value();
    }
    d -= D;

    for (size_t i = 0; i < _components.size(); i++)
    {
        D = _components[ i ]->num_var_params();
        if (d < D)
        {
            return _components[ i ]->get_var_param_value(d);
        }
        d -= D;
    }

    abort();
}

void Object::get_var_param_values(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    Vector_d* q = 0;

    Model::get_var_param_values(&q);
    assert(copy_vector_section_into_vector_d(p, d, q, 0, q->num_elts) == 0);
    d += q->num_elts;

    for (size_t i = 0; i < _components.size(); i++)
    {
        _components[ i ]->get_var_param_values(&q);
        assert(copy_vector_section_into_vector_d(p, d, q, 0, q->num_elts) == 0);
        d += q->num_elts;
    }

    free_vector_d(q);

    assert(d == D);
}

/** 
 * If a Parameter value could not be set, it is left unchanged and an exception
 * is thrown, after an attempt to set the other parameters is made.
 */
void Object::set_var_param_values(const jwsc::Vector_d* p)
    throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    uint32_t  d = 0;
    uint32_t  d_posn = 1;
    uint32_t  D = num_var_params();
    double    delta;
    double    epsilon = 1.0e-8;
    double    e;
    Vector_d* p_posn = 0;

    assert(D == p->num_elts);

    std::ostringstream ost;

    create_init_vector_d(&p_posn, 4, 1);
    p_posn->elts[0] = p->elts[ d_posn+0 ];
    p_posn->elts[1] = p->elts[ d_posn+1 ];
    p_posn->elts[2] = p->elts[ d_posn+2 ];
    transform_pt_from_world_to_object_coords(p_posn);

    transform_from_world_to_object_coords();

    if (fabs(delta = p->elts[ d ] - azimuth()) > 0)
    {
        e = (delta < 0) ? -epsilon : epsilon;
        try 
        { 
            test_azimuth_in_obj_coords(azimuth() + delta + e); 
            azimuth_in_obj_coords(azimuth() + delta);
        }
        catch (jwscxx::base::Arg_error e) 
        { ost << e.get_msg() << "  "; }
    }
    d++;

    // See note below.
    d += 3;

    Vector_d* q = 0;
    for (size_t i = 0; i < _components.size(); i++)
    {
        uint32_t D_q = _components[ i ]->num_var_params();

        copy_vector_section_d(&q, p, d, D_q);
        d += q->num_elts;

        try { _components[ i ]->set_var_param_values_in_object_coords(q); }
        catch (jwscxx::base::Arg_error e)
        { ost << e.get_msg(); }
    }
    free_vector_d(q);

    assert(d == D);

    // After all the shifting from variable attachment points and size changes,
    // reposition the object to its intended place. Note that this is for shift
    // only.
    for (size_t i = 0; i < 3; i++)
    {
        if (fabs(delta = p_posn->elts[ i ] - posn(i)) > 0)
        {
            e = (delta < 0) ? -epsilon : epsilon;
            try 
            { 
                _posn_p[ i ].test_value_shift_deps(p_posn->elts[ i ] + e);
                _posn_p[ i ].value_shift_deps(p_posn->elts[ i ]);
            }
            catch (jwscxx::base::Arg_error e) 
            { ost << e.get_msg() << "  "; }
        }
    }

    update_graphical_rep();
    update_lowest_component();
    transform_from_object_to_world_coords();

    free_vector_d(p_posn);

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

/**
 * If a Parameter value could not be set, it is left unchanged and an exception
 * is thrown, after an attempt to set the other parameters is made.
 *
 * @param  p  Point to add to the Parameter values of this Object.
 *
 * @throw  jwscxx::base::Arg_error  Invalid resulting parameter value.
 */
void Object::add_var_param_values(const jwsc::Vector_d* p) 
    throw (jwscxx::base::Arg_error)
{
    add_var_param_values_shift_component_deps(p);
}

/**
 * If a Parameter value could not be set, it is left unchanged and an exception
 * is thrown, after an attempt to set the other parameters is made.
 *
 * @param  p  Point to add to the Parameter values of this Object.
 *
 * @throw  jwscxx::base::Arg_error  Invalid resulting parameter value.
 */
void Object::add_var_param_values_shift_component_deps(const jwsc::Vector_d* p) 
    throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    uint32_t D = num_var_params();
    assert(D == p->num_elts);

    Vector_d* q = 0;
    get_var_param_values(&q);
    add_vectors_d(&q, p, q);

    try { set_var_param_values(q); }
    catch (jwscxx::base::Arg_error e)
    {
        free_vector_d(q);
        throw e;
    }

    free_vector_d(q);
}

/**
 * If a Parameter value could not be set, it is left unchanged and an exception
 * is thrown, after an attempt to set the other parameters is made.
 *
 * @param  p  Point to add to the Parameter values of this Object.
 *
 * @throw  jwscxx::base::Arg_error  Invalid resulting parameter value.
 */
void Object::add_var_param_values_scale_component_deps(const jwsc::Vector_d* p) 
    throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    uint32_t  d = 0;
    uint32_t  d_posn = 1;
    uint32_t  D = num_var_params();
    double    delta;
    double    epsilon = 1.0e-8;
    double    e;
    Vector_d* p_posn = 0;
    Matrix_d* M = 0;

    assert(D == p->num_elts);

    std::ostringstream ost;

    create_init_vector_d(&p_posn, 4, 1);
    p_posn->elts[0] = p->elts[ d_posn+0 ];
    p_posn->elts[1] = p->elts[ d_posn+1 ];
    p_posn->elts[2] = p->elts[ d_posn+2 ];
    create_3d_homo_rotation_matrix_2_d(&M, -azimuth(), 0, 1, 0);
    multiply_matrix_and_vector_d(&p_posn, M, p_posn);

    transform_from_world_to_object_coords();

    if (fabs(delta = p->elts[ d ]) > 0)
    {
        e = (delta < 0) ? -epsilon : epsilon;
        try 
        { 
            test_azimuth_in_obj_coords(azimuth() + delta + e); 
            azimuth_in_obj_coords(azimuth() + delta);
        }
        catch (jwscxx::base::Arg_error e) 
        { ost << e.get_msg() << "  "; }
    }
    d++;

    for (size_t i = 0; i < 3; i++)
    {
        if (fabs(delta = p_posn->elts[ i ]) > 0)
        {
            e = (delta < 0) ? -epsilon : epsilon;
            try 
            { 
                _posn_p[ i ].test_value_scale_deps(posn(i) + delta + e);
                _posn_p[ i ].value_scale_deps(posn(i) + delta);
            }
            catch (jwscxx::base::Arg_error e) 
            { ost << e.get_msg() << "  "; }
        }
        d++;
    }

    Vector_d* q = 0;
    for (size_t i = 0; i < _components.size(); i++)
    {
        uint32_t D_q = _components[ i ]->num_var_params();

        copy_vector_section_d(&q, p, d, D_q);
        d += q->num_elts;

        try { _components[ i ]->add_var_param_values_scale_deps(q); }
        catch (jwscxx::base::Arg_error e)
        { ost << e.get_msg(); }
    }
    free_vector_d(q);

    assert(d == D);

    update_graphical_rep();
    update_lowest_component();
    transform_from_object_to_world_coords();

    free_vector_d(p_posn);
    free_matrix_d(M);

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

void Object::get_var_param_values_dynamics_grad_eta(jwsc::Vector_d** p_out) 
    const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    Vector_d* q = 0;

    Model::get_var_param_values_dynamics_grad_eta(&q);
    assert(copy_vector_section_into_vector_d(p, d, q, 0, q->num_elts) == 0);
    d += q->num_elts;

    for (size_t i = 0; i < _components.size(); i++)
    {
        _components[ i ]->get_var_param_values_dynamics_grad_eta(&q);
        assert(copy_vector_section_into_vector_d(p, d, q, 0, q->num_elts) == 0);
        d += q->num_elts;
    }

    free_vector_d(q);

    assert(d == D);
}

void Object::get_var_param_values_dynamics_delta_t(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    Vector_d* q = 0;

    Model::get_var_param_values_dynamics_delta_t(&q);
    assert(copy_vector_section_into_vector_d(p, d, q, 0, q->num_elts) == 0);
    d += q->num_elts;

    for (size_t i = 0; i < _components.size(); i++)
    {
        _components[ i ]->get_var_param_values_dynamics_delta_t(&q);
        assert(copy_vector_section_into_vector_d(p, d, q, 0, q->num_elts) == 0);
        d += q->num_elts;
    }

    free_vector_d(q);

    assert(d == D);
}

void Object::print_var_param_values(uint32_t iteration, std::ostream& out) const
{
    Model::print_var_param_values(iteration, out);

    for (uint32_t i = 0; i < num_components(); i++)
    {
        component(i)->print_var_param_values(iteration, out);
    }
}

void Object::get_var_param_stats
(
    double*  mean_out, 
    double*  stdev_out, 
    uint32_t d
) 
const
{
    using namespace jwsc;

    assert(d < num_var_params());
    uint32_t D = 0;

    D = Model::num_var_params();
    if (d < D)
    {
        var_param(d)->stats(mean_out, stdev_out);
        return;
    }
    d -= D;

    for (size_t i = 0; i < _components.size(); i++)
    {
        D = _components[ i ]->num_var_params();
        if (d < D)
        {
            _components[ i ]->get_var_param_stats(mean_out, stdev_out, d);
            return;
        }
        d -= D;
    }

    abort();
}

void Object::get_var_param_stats
(
    jwsc::Vector_d** means_out,
    jwsc::Vector_d** stdevs_out
) 
const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    create_vector_d(means_out, D);
    create_vector_d(stdevs_out, D);
    Vector_d* means = *means_out;
    Vector_d* stdevs = *stdevs_out;

    Vector_d* m_q = 0;
    Vector_d* s_q = 0;

    Model::get_var_param_stats(&m_q, &s_q);
    copy_vector_section_into_vector_d(means, d, m_q, 0, m_q->num_elts);
    copy_vector_section_into_vector_d(stdevs, d, s_q, 0, s_q->num_elts);
    d += m_q->num_elts;

    for (size_t i = 0; i < _components.size(); i++)
    {
        _components[ i ]->get_var_param_stats(&m_q, &s_q);
        copy_vector_section_into_vector_d(means, d, m_q, 0, m_q->num_elts);
        copy_vector_section_into_vector_d(stdevs, d, s_q, 0, s_q->num_elts);
        d += m_q->num_elts;
    }

    free_vector_d(m_q);
    free_vector_d(s_q);

    assert(d == D);
}

void Object::set_var_param_stats
(
    const jwsc::Vector_d* means,
    const jwsc::Vector_d* stdevs
)
throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    assert(D <= means->num_elts);
    assert(D <= stdevs->num_elts);

    Vector_d* m_q = 0;
    Vector_d* s_q = 0;

    copy_vector_section_d(&m_q, means, d, Model::num_var_params());
    copy_vector_section_d(&s_q, stdevs, d, Model::num_var_params());
    Model::set_var_param_stats(m_q, s_q);
    d += m_q->num_elts;

    for (size_t i = 0; i < _components.size(); i++)
    {
        uint32_t D_q = _components[ i ]->num_var_params();

        copy_vector_section_d(&m_q, means, d, D_q);
        copy_vector_section_d(&s_q, stdevs, d, D_q);
        d += m_q->num_elts;

        _components[ i ]->set_var_param_stats(m_q, s_q);
    }

    free_vector_d(m_q);
    free_vector_d(s_q);

    assert(d <= D);
}

void Object::get_var_param_stats_dynamics_grad_eta
(
    jwsc::Vector_d** means_out,
    jwsc::Vector_d** stdevs_out
) 
const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    create_vector_d(means_out, D);
    create_vector_d(stdevs_out, D);

    Vector_d* means = *means_out;
    Vector_d* stdevs = *stdevs_out;

    Vector_d* m_q = 0;
    Vector_d* s_q = 0;

    Model::get_var_param_stats_dynamics_grad_eta(&m_q, &s_q);
    copy_vector_section_into_vector_d(means, d, m_q, 0, m_q->num_elts);
    copy_vector_section_into_vector_d(stdevs, d, s_q, 0, s_q->num_elts);
    d += m_q->num_elts;

    for (size_t i = 0; i < _components.size(); i++)
    {
        _components[ i ]->get_var_param_stats_dynamics_grad_eta(&m_q, &s_q);
        copy_vector_section_into_vector_d(means, d, m_q, 0, m_q->num_elts);
        copy_vector_section_into_vector_d(stdevs, d, s_q, 0, s_q->num_elts);
        d += m_q->num_elts;
    }

    free_vector_d(m_q);
    free_vector_d(s_q);

    assert(d == D);
}

void Object::get_var_param_stats_dynamics_delta_t
(
    jwsc::Vector_d** means_out,
    jwsc::Vector_d** stdevs_out
) 
const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    create_vector_d(means_out, D);
    create_vector_d(stdevs_out, D);

    Vector_d* means = *means_out;
    Vector_d* stdevs = *stdevs_out;

    Vector_d* m_q = 0;
    Vector_d* s_q = 0;

    Model::get_var_param_stats_dynamics_delta_t(&m_q, &s_q);
    copy_vector_section_into_vector_d(means, d, m_q, 0, m_q->num_elts);
    copy_vector_section_into_vector_d(stdevs, d, s_q, 0, s_q->num_elts);
    d += m_q->num_elts;

    for (size_t i = 0; i < _components.size(); i++)
    {
        _components[ i ]->get_var_param_stats_dynamics_delta_t(&m_q, &s_q);
        copy_vector_section_into_vector_d(means, d, m_q, 0, m_q->num_elts);
        copy_vector_section_into_vector_d(stdevs, d, s_q, 0, s_q->num_elts);
        d += m_q->num_elts;
    }

    free_vector_d(m_q);
    free_vector_d(s_q);

    assert(d == D);
}

void Object::print_var_param_stats(uint32_t iteration, std::ostream& out) const
{
    Model::print_var_param_stats(iteration, out);

    for (uint32_t i = 0; i < num_components(); i++)
    {
        component(i)->print_var_param_stats(iteration, out);
    }
}

const Model::Parameter* Object::azimuth_p() const
{
    return &_azimuth_p;
}

double Object::azimuth() const
{
    return _azimuth_p.value();
}

/**
 * @param  azimuth  Rotation angle about the local y-axis of this Object. In
 *                  radians.
 *
 * @warning Aborts on invalid arguments.
 */
void Object::azimuth(double azimuth)
{
    transform_from_world_to_object_coords();
    azimuth_in_obj_coords(azimuth);
    update_graphical_rep();
    transform_from_object_to_world_coords();
}

void Object::azimuth_in_obj_coords(double azimuth)
{
    double min = azimuth_p()->context()->range().min();
    double max = azimuth_p()->context()->range().max();
    if (azimuth > max && (azimuth - JWSC_2_PI) >= min)
    {
        azimuth -= JWSC_2_PI;
    }
    else if (azimuth < min && (azimuth + JWSC_2_PI) <= max)
    {
        azimuth += JWSC_2_PI;
    }

    _azimuth_p.value(azimuth);
}

/**
 * @param  azimuth  Rotation angle about the local y-axis of this Object. In
 *                  radians.
 *
 * @throw  jwscxx::base::Arg_error  Invalid azimuth angle.
 */
void Object::test_azimuth(double azimuth) const throw (jwscxx::base::Arg_error)
{
    test_azimuth_in_obj_coords(azimuth);
}

/**
 * @param  azimuth  Rotation angle about the local y-axis of this Object. In
 *                  radians.
 *
 * @throw  jwscxx::base::Arg_error  Invalid azimuth angle.
 */
void Object::test_azimuth_in_obj_coords(double azimuth) const 
    throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    double min = azimuth_p()->context()->range().min();
    double max = azimuth_p()->context()->range().max();
    if (azimuth > max && (azimuth - JWSC_2_PI) >= min)
    {
        azimuth -= JWSC_2_PI;
    }
    else if (azimuth < min && (azimuth + JWSC_2_PI) <= max)
    {
        azimuth += JWSC_2_PI;
    }

    azimuth_p()->test_value(azimuth);
}

const Component::Parameter* Object::posn_p(size_t i) const
{
    assert(i < 3);
    return &(_posn_p[ i ]);
}

double Object::posn(size_t i) const
{
    return posn_p(i)->value();
}

/**
 * The resulting vector is in homogeneous coordinates.
 */
void Object::posn_v(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    create_vector_d(p_out, 4);
    Vector_d* p = *p_out;

    p->elts[0] = _posn_p[0].value();
    p->elts[1] = _posn_p[1].value();
    p->elts[2] = _posn_p[2].value();
    p->elts[3] = 1.0;
}

/**
 * The vector must be in homogeneous coordinates.
 *
 * @param  p  Position of this Object.
 */
void Object::posn_v(const jwsc::Vector_d* p)
{
    assert(p->num_elts == 4);
    if (fabs(p->elts[3] - 1.0) > 1.0e-6)
    {
        std::cerr << __FILE__ << ":" << __LINE__ << " "
                  << p->elts[3] << std::endl;
        abort();
    }

    _posn_p[0].value(p->elts[0]);
    _posn_p[1].value(p->elts[1]);
    _posn_p[2].value(p->elts[2]);
}

void Object::position_object_below_component(const Component* c)
{
    using namespace jwsc;

    const size_t y = 1;
    const size_t neg_h = 2*y + 1;

    Vector_d* a_neg_h = 0;
    const Abstract_block* b;

    if ((b = dynamic_cast<const Abstract_block*>(c)))
    {
        b->ctr_attach_pt_v(&a_neg_h, neg_h);
    }
    else
    {
        abort(); // Unknown component type.
    }

    posn_v(a_neg_h);

    free_vector_d(a_neg_h);
}

void Object::position_attachment
(
    const Object*          obj, 
    Component*             c, 
    Component::Attach_type t
)
{
    Abstract_block* b;

    if ((b = dynamic_cast<Abstract_block*>(c)))
    {
        position_attachment(obj, b, t);
    }
    else
    {
        abort(); // Unknown component type.
    }
}

void Object::position_attachment
(
    const Object*          obj, 
    Abstract_block*        b, 
    Component::Attach_type t
)
{
    using namespace jwsc;

    const size_t y = 1;
    const size_t neg_h = 2*y + 1;

    Vector_d* blk_a_neg_hxyz = 0;
    b->attach_pt_v(&blk_a_neg_hxyz, neg_h, t);

    double t_x = blk_a_neg_hxyz->elts[0];
    double t_y = blk_a_neg_hxyz->elts[1];
    double t_z = blk_a_neg_hxyz->elts[2];

    b->translate(-t_x, -t_y, -t_z);
    b->rotate(obj->azimuth(), 0, 1, 0);
    b->attach_pt_v(&blk_a_neg_hxyz, neg_h, t);

    t_x = obj->posn(0) - blk_a_neg_hxyz->elts[0];
    t_y = obj->posn(1) - blk_a_neg_hxyz->elts[1];
    t_z = obj->posn(2) - blk_a_neg_hxyz->elts[2];

    b->translate(t_x, t_y, t_z);

    free_vector_d(blk_a_neg_hxyz);
}

void Object::position_attachment
(
    const Component*       c1, 
    size_t                 i1,
    Component::Attach_type t1,
    Component*             c2,
    size_t                 i2,
    Component::Attach_type t2
)
{
    const Abstract_block* b1;
    Abstract_block* b2;

    if ((b1 = dynamic_cast<const Abstract_block*>(c1)) &&
        (b2 = dynamic_cast<Abstract_block*>(c2)))
    {
        position_attachment(b1, i1, t1, b2, i2, t2);
    }
    else
    {
        abort(); // Unknown component type.
    }
}

void Object::position_attachment
(
    const Abstract_block*  b1, 
    size_t                 i1, 
    Component::Attach_type t1,
    Abstract_block*        b2, 
    size_t                 i2,
    Component::Attach_type t2
)
{
    using namespace jwsc;

    Object* obj = b1->object();
    assert(obj);

    Vector_d* b1_a = 0;
    Vector_d* b2_a = 0;
    b1->attach_pt_v(&b1_a, i1, t1);
    b2->attach_pt_v(&b2_a, i2, t2);

    double t_x = b2_a->elts[0];
    double t_y = b2_a->elts[1];
    double t_z = b2_a->elts[2];

    b2->translate(-t_x, -t_y, -t_z);
    b2->rotate(obj->azimuth(), 0, 1, 0);
    b2->attach_pt_v(&b2_a, i2, t2);

    t_x = b1_a->elts[0] - b2_a->elts[0];
    t_y = b1_a->elts[1] - b2_a->elts[1];
    t_z = b1_a->elts[2] - b2_a->elts[2];

    b2->translate(t_x, t_y, t_z);

    free_vector_d(b1_a);
    free_vector_d(b2_a);
}

void Object::position_detachment(Component* c)
{
    using namespace jwsc;

    Object* obj = c->object();
    assert(obj);

    Vector_d* p = 0;
    c->posn_v(&p);

    double t_x = p->elts[0];
    double t_y = p->elts[1];
    double t_z = p->elts[2];

    c->translate(-t_x, -t_y, -t_z);
    c->rotate(-obj->azimuth(), 0, 1, 0);

    free_vector_d(p);
}

/**
 * The position must be in object coordinates.
 *
 * @param  p  Point to set the object position with.
 *
 * @warning Aborts on invalid arguments.
 */
void Object::posn_shift_deps_v(const jwsc::Vector_d* p)
{
    assert(p->num_elts >= 3);
    transform_from_world_to_object_coords();
    for (uint32_t i = 0; i < 3; i++)
    {
        _posn_p[ i ].value_shift_deps(p->elts[ i ]);
    }
    update_graphical_rep();
    transform_from_object_to_world_coords();
}

/** This object and the point @em p are expected to be in world coordinates. */
void Object::add_posn_shift_deps_v(const jwsc::Vector_d* p)
{
    using namespace jwsc;

    Vector_d* p1 = 0;
    Vector_d* p2 = 0;
    posn_v(&p2);

    add_vectors_d(&p1, p, p2);
    transform_pt_from_world_to_object_coords(p1);
    posn_shift_deps_v(p1);

    free_vector_d(p1);
    free_vector_d(p2);
}

/**
 * The position must be in object coordinates.
 *
 * @param  i     Index of the coordinate to set.
 * @param  posn  Position coord of this Object.
 *
 * @warning Aborts on invalid arguments.
 */
void Object::posn_shift_deps(size_t i, double posn)
{
    assert(i < 3);
    transform_from_world_to_object_coords();
    _posn_p[ i ].value_shift_deps(posn);
    update_graphical_rep();
    transform_from_object_to_world_coords();
}

/**
 * The position must be in object coordinates.
 *
 * @param  i     Index of the coordinate to set.
 * @param  posn  Position coord of this Object.
 *
 * @warning Aborts on invalid arguments.
 */
void Object::posn_scale_deps(size_t i, double posn)
{
    assert(i < 3);
    transform_from_world_to_object_coords();
    _posn_p[ i ].value_scale_deps(posn);
    update_graphical_rep();
    transform_from_object_to_world_coords();
}

/**
 * The position must be in object coordinates.
 *
 * @param  i     Index of the coordinate to set.
 * @param  posn  Position coord of this Object.
 *
 * @throw  jwscxx::base::Arg_error  Invalid position.
 */
void Object::test_posn_shift_deps(size_t i, double posn)
    throw (jwscxx::base::Arg_error)
{
    transform_from_world_to_object_coords();
    try 
    {
        //posn_p(i)->test_value_shift_deps(posn);
        _posn_p[ i ].test_value_shift_deps(posn);
    }
    catch (jwscxx::base::Arg_error e)
    {
        transform_from_object_to_world_coords();
        throw e;
    }
    transform_from_object_to_world_coords();
}

/**
 * The position must be in object coordinates.
 *
 * @param  i     Index of the coordinate to set.
 * @param  posn  Position coord of this Object.
 *
 * @throw  jwscxx::base::Arg_error  Invalid position.
 */
void Object::test_posn_scale_deps(size_t i, double posn)
    throw (jwscxx::base::Arg_error)
{
    transform_from_world_to_object_coords();
    try 
    {
        //posn_p(i)->test_value_scale_deps(posn);
        _posn_p[ i ].test_value_scale_deps(posn);
    }
    catch (jwscxx::base::Arg_error e)
    {
        transform_from_object_to_world_coords();
        throw e;
    }
    transform_from_object_to_world_coords();
}

void Object::assert_component_consistency()
{
    if (num_components() == 0)
        return;

    bool was_in_world_coords;

    if ((was_in_world_coords = !in_object_coords()))
    {
        transform_from_world_to_object_coords();
    }

    assert_attachment_consistency(this, lowest_component());

    for (size_t i = 0; i < num_components(); i++)
    {
        _components[ i ]->assert_consistency();
    }

    if (was_in_world_coords)
    {
        transform_from_object_to_world_coords();
    }
}

void Object::ensure_component_consistency()
{
    if (num_components() == 0)
        return;

    for (size_t i = 0; i < num_components(); i++)
    {
        _components[ i ]->ensure_consistency();
    }
}

/**
 * The point must be in homogeneous coodinates.
 *
 * @param  p  Point to transform from object to world coordinates.
 */
void Object::transform_pt_from_object_to_world_coords(jwsc::Vector_d* p) const
{
    using namespace jwsc;

    double t_x, t_y, t_z;
    double phi;

    if (in_object_coords())
    {
        t_x = _obj_coords_posn->elts[0];
        t_y = _obj_coords_posn->elts[1];
        t_z = _obj_coords_posn->elts[2];
        phi = _obj_coords_azimuth;
    }
    else
    {
        t_x = posn(0);
        t_y = posn(1);
        t_z = posn(2);
        phi = azimuth();
    }

    Matrix_d* M = 0;

    create_3d_homo_rotation_matrix_2_d(&M, phi, 0, 1, 0);
    multiply_matrix_and_vector_d(&p, M, p);

    create_3d_homo_translation_matrix_2_d(&M, t_x, t_y, t_z);
    multiply_matrix_and_vector_d(&p, M, p);

    free_matrix_d(M);
}

/**
 * The point must be in homogeneous coodinates.
 *
 * @param  p  Point to transform from world to object coordinates.
 */
void Object::transform_pt_from_world_to_object_coords(jwsc::Vector_d* p) const
{
    using namespace jwsc;

    double t_x, t_y, t_z;
    double phi;

    if (in_object_coords())
    {
        t_x = _obj_coords_posn->elts[0];
        t_y = _obj_coords_posn->elts[1];
        t_z = _obj_coords_posn->elts[2];
        phi = _obj_coords_azimuth;
    }
    else
    {
        t_x = posn(0);
        t_y = posn(1);
        t_z = posn(2);
        phi = azimuth();
    }

    Matrix_d* M = 0;

    create_3d_homo_translation_matrix_2_d(&M, -t_x, -t_y, -t_z);
    multiply_matrix_and_vector_d(&p, M, p);

    create_3d_homo_rotation_matrix_2_d(&M, -phi, 0, 1, 0);
    multiply_matrix_and_vector_d(&p, M, p);

    free_matrix_d(M);
}

void Object::transform_from_world_to_object_coords()
{
    using namespace jwsc;

    assert(!in_object_coords());

    posn_v(&_obj_coords_posn);
    _obj_coords_azimuth = azimuth();

    double t_x = _obj_coords_posn->elts[0];
    double t_y = _obj_coords_posn->elts[1];
    double t_z = _obj_coords_posn->elts[2];
    double phi = _obj_coords_azimuth;

    for (size_t i = 0; i < _components.size(); i++)
    {
        _components[ i ]->translate(-t_x, -t_y, -t_z);
        _components[ i ]->rotate(-phi, 0, 1, 0);
    }

    Vector_d* p = 0;
    create_zero_vector_d(&p, 4);
    p->elts[3] = 1;
    posn_v(p);
    free_vector_d(p);

    _obj_coords = true;
}

void Object::transform_from_object_to_world_coords()
{
    using namespace jwsc;

    assert(in_object_coords());

    Vector_d* p_obj = 0;
    Vector_d* p_world = 0;
    Matrix_d* M = 0;

    posn_v(&p_obj);
    create_3d_homo_rotation_matrix_2_d(&M, _obj_coords_azimuth, 0, 1, 0);
    multiply_matrix_and_vector_d(&p_world, M, p_obj);
    free_matrix_d(M);

    p_world->elts[0] += _obj_coords_posn->elts[0];
    p_world->elts[1] += _obj_coords_posn->elts[1];
    p_world->elts[2] += _obj_coords_posn->elts[2];
    posn_v(p_world);

    for (size_t i = 0; i < _components.size(); i++)
    {
        double t_x = p_obj->elts[0];
        double t_y = p_obj->elts[1];
        double t_z = p_obj->elts[2];

        _components[ i ]->translate(-t_x, -t_y, -t_z);
        _components[ i ]->rotate(azimuth(), 0, 1, 0);

        t_x = p_world->elts[0];
        t_y = p_world->elts[1];
        t_z = p_world->elts[2];

        _components[ i ]->translate(t_x, t_y, t_z);
    }

    free_vector_d(p_world);
    free_vector_d(p_obj);

    _obj_coords = false;
}

bool Object::in_object_coords() const
{
    return _obj_coords;
}

void Object::test_swap_object_coords() throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    Vector_d* means = 0;
    Vector_d* stdevs = 0;

    Object* obj = new Object(azimuth(), posn(0), posn(1), posn(2), context(),
            _block_ctx, _block_set_ctx);
    Model::get_var_param_stats(&means, &stdevs);
    obj->Model::set_var_param_stats(means, stdevs);

    free_vector_d(means);
    free_vector_d(stdevs);

    try
    {
        Component* c_i = lowest_component();
        Component* c_j = c_i->clone_and_swap();
        obj->attach(c_j, Component::CTR_ATTACH);
        swap_component_coords(c_j, c_i, 3); 
        obj->azimuth(obj->azimuth() + JWSC_PI_2);
    }
    catch (jwscxx::base::Arg_error e) { abort(); }

    delete obj;
}

void Object::swap_object_coords(bool pos_swap)
{
    using namespace jwsc;

    Vector_d* means = 0;
    Vector_d* stdevs = 0;

    Object* obj = new Object(azimuth(), posn(0), posn(1), posn(2), context(),
            _block_ctx, _block_set_ctx);
    Model::get_var_param_stats(&means, &stdevs);
    obj->Model::set_var_param_stats(means, stdevs);

    free_vector_d(means);
    free_vector_d(stdevs);

    try
    {
        Component* c_i = lowest_component();
        Component* c_j = c_i->clone_and_swap();
        obj->attach(c_j, Component::CTR_ATTACH);
        c_j->id(c_i->id());
        swap_component_coords(c_j, c_i, 3); 
        obj->azimuth(obj->azimuth() + ((pos_swap) ? 1.0 : -1.0)*JWSC_PI_2);
    }
    catch (jwscxx::base::Arg_error e) { abort(); }

    obj->_next_component_id  = _next_component_id;
    obj->_used_component_ids = _used_component_ids;

    uint32_t N = num_components();
    for (uint32_t n1 = 0; n1 < N; n1++)
    {
        if (component(n1)->id() == obj->component(n1)->id())
            continue;

        assert(n1+1 < N);
        for (uint32_t n2 = n1+1; n2 < N; n2++)
        {
            if (component(n1)->id() == obj->component(n2)->id())
            {
                Component* c = obj->_components[ n1 ];
                obj->_components[ n1 ] = obj->_components[ n2 ];
                obj->_components[ n2 ] = c;
                break;
            }
        }
        assert(component(n1)->id() == obj->component(n1)->id());
    }

    *this = *obj;
    delete obj;
}

/**
 * @param  c_j1    Copy of @em c_i1 to put the copied attachments into.
 * @param  c_i1    Component to copy the attachment from.
 * @param  i1_ign  Attachment index of c_i1 to ignore.
 */
void Object::swap_component_coords
(
    Component* c_j1,
    Component* c_i1,
    uint32_t   i1_ign
)
throw (jwscxx::base::Arg_error)
{
    typedef Component::Attach_type Attach_t;

    Object* obj = c_j1->object();

    for (uint32_t i1 = 0; i1 < c_i1->num_ctr_attach_pts(); i1++)
    {
        Component* c_i2 = c_i1->attachment(i1);

        if (!c_i2 || i1 == i1_ign)
            continue;

        uint32_t i2 = c_i2->attachment(c_i1);
        Attach_t t1 = c_i1->attach_pt_type(i1);
        Attach_t t2 = c_i2->attach_pt_type(i2);
        uint32_t j1 = i1;
        uint32_t j2 = i2;

        Component* c_j2 = c_i2->clone_and_swap();
        obj->attach(c_j1, j1, t1, c_j2, j2, t2);

        c_j2->id(c_i2->id());

        if (c_i1->conditions(c_i2))
        {
            swap_conds(c_j1, j1, c_j2, j2);
        }

        swap_component_coords(c_j2, c_i2, i2);
    }
}

void Object::swap_conds(Component* c1, size_t j1, Component* c2, size_t j2)
{
    Abstract_block* b1;
    Abstract_block* b2;

    if ((b1 = dynamic_cast<Abstract_block*>(c1)) &&
        (b2 = dynamic_cast<Abstract_block*>(c2)))
    {
        detach_conds(b1, j1, b2, j2);
        attach_conds(b2, j2, b1, j1);
    }
    else
    {
        abort(); // Unknown component type.
    }
}

double Object::calc_var_param_log_likelihood() const
{
    double log_prob = Model::calc_var_param_log_likelihood();

    log_prob += jwsc::log_geometric_pmf_d(1.0-1.0e-16, num_components());
    //log_prob += jwsc::log_geometric_pmf_d(1.0-1.0e-1, num_components());

    for (size_t i = 0; i < _components.size(); i++)
    {
        log_prob += _components[ i ]->calc_var_param_log_likelihood();
    }

    for (size_t i = 0; i < _components.size(); i++)
    {
        Component* c1 = _components[ i ];
        for (size_t j = i+1; j < _components.size(); j++)
        {
            Component* c2 = _components[ j ];
            double y = intersection_ratio(c1, c2);
            if(y > 1.0e-2)
            {
                log_prob += (y+1)*100*log(JWSC_MIN_LOG_ARG);
            }
        }
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

void Object::sample_var_param_likelihood(jwsc::Vector_d** p_out)
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    Vector_d* q = 0;

    Model::sample_var_param_likelihood(&q);
    assert(copy_vector_section_into_vector_d(p, d, q, 0, q->num_elts) == 0);
    d += q->num_elts;

    // There is only one component with no conditionals!!
    Component* root;
    for (size_t i = 0; i < _components.size(); i++)
    {
        if (_components[ i ]->conditional() == 0)
        {
            root = _components[ i ];
            break;
        }
    }

    uint32_t n = gen_var_param_conditional_likelihood_sample(root, 0);
    assert(n == _components.size());

    for (size_t i = 0; i < _components.size(); i++)
    {
        _components[ i ]->get_var_param_conditional_likelihood_sample(&q);
        assert(copy_vector_section_into_vector_d(p, d, q, 0, q->num_elts) == 0);
        d += q->num_elts;
    }

    free_vector_d(q);

    assert(d == D);
}

uint32_t Object::gen_var_param_conditional_likelihood_sample
(
    Component* c, 
    uint32_t   n
)
{
    c->gen_var_param_conditional_likelihood_sample();
    n++;

    for (uint32_t j = 0; j < 6; j++)
    {
        Component* cc = c->attachment(j);
        if (cc && cc->conditions(c))
        {
            n = gen_var_param_conditional_likelihood_sample(cc, n);
        }
    }

    return n;
}

double Object::calc_var_param_prior_log_prob() const
{
    double log_prob = Model::calc_var_param_prior_log_prob();

    for (size_t i = 0; i < _components.size(); i++)
    {
        log_prob += _components[ i ]->calc_var_param_prior_log_prob();
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

void Object::sample_var_param_prior
(
    jwsc::Vector_d** means_out,
    jwsc::Vector_d** stdevs_out
) 
const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    create_vector_d(means_out, D);
    create_vector_d(stdevs_out, D);

    Vector_d* m = *means_out;
    Vector_d* s = *stdevs_out;
    Vector_d* m_q = 0;
    Vector_d* s_q = 0;

    Model::sample_var_param_prior(&m_q, &s_q);
    copy_vector_section_into_vector_d(m, d, m_q, 0, m_q->num_elts);
    copy_vector_section_into_vector_d(s, d, s_q, 0, s_q->num_elts);
    d += m_q->num_elts;

    for (size_t i = 0; i < _components.size(); i++)
    {
        _components[ i ]->sample_var_param_prior(&m_q, &s_q);
        copy_vector_section_into_vector_d(m, d, m_q, 0, m_q->num_elts);
        copy_vector_section_into_vector_d(s, d, s_q, 0, s_q->num_elts);
        d += m_q->num_elts;
    }

    free_vector_d(m_q);
    free_vector_d(s_q);

    assert(d == D);
}

void Object::sample_var_param_cat_proposal
(
    jwsc::Vector_d** means_out,
    jwsc::Vector_d** stdevs_out,
    const jwsc::Vector_d* means_in
) 
const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    create_vector_d(means_out, D);
    create_vector_d(stdevs_out, D);

    Vector_d* m = *means_out;
    Vector_d* s = *stdevs_out;
    Vector_d* m_q = 0;
    Vector_d* s_q = 0;
    Vector_d* w   = 0;

    copy_vector_section_d(&w, means_in, d, Model::num_var_params());
    Model::sample_var_param_cat_proposal(&m_q, &s_q, w);
    copy_vector_section_into_vector_d(m, d, m_q, 0, m_q->num_elts);
    copy_vector_section_into_vector_d(s, d, s_q, 0, s_q->num_elts);
    d += m_q->num_elts;

    for (size_t i = 0; i < _components.size(); i++)
    {
        copy_vector_section_d(&w, means_in, d, 
                _components[ i ]->num_var_params());
        _components[ i ]->sample_var_param_cat_proposal(&m_q, &s_q, w);
        copy_vector_section_into_vector_d(m, d, m_q, 0, m_q->num_elts);
        copy_vector_section_into_vector_d(s, d, s_q, 0, s_q->num_elts);
        d += m_q->num_elts;
    }

    free_vector_d(m_q);
    free_vector_d(s_q);
    free_vector_d(w);

    assert(d == D);
}

double Object::calc_var_param_proposal_log_prob() const
{
    double log_prob = Model::calc_var_param_proposal_log_prob();

    for (size_t i = 0; i < _components.size(); i++)
    {
        log_prob += _components[ i ]->calc_var_param_proposal_log_prob();
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

double Object::calc_var_param_gauss_proposal_log_prob() const
{
    double log_prob = Model::calc_var_param_gauss_proposal_log_prob();

    for (size_t i = 0; i < _components.size(); i++)
    {
        log_prob += _components[ i ]->calc_var_param_gauss_proposal_log_prob();
    }

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(log_prob) && !std::isnan(log_prob));
#endif

    return log_prob;
}

void Object::sample_var_param_proposal(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    Vector_d* q = 0;

    Model::sample_var_param_proposal(&q);
    assert(copy_vector_section_into_vector_d(p, d, q, 0, q->num_elts) == 0);
    d += q->num_elts;

    for (size_t i = 0; i < _components.size(); i++)
    {
        _components[ i ]->sample_var_param_proposal(&q);
        assert(copy_vector_section_into_vector_d(p, d, q, 0, q->num_elts) == 0);
        d += q->num_elts;
    }

    free_vector_d(q);

    assert(d == D);
}

void Object::sample_var_param_gauss_proposal
(
    jwsc::Vector_d**      p_out,
    const jwsc::Vector_d* means
) 
const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    assert(D == means->num_elts);

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    Vector_d* q = 0;
    Vector_d* w = 0;

    copy_vector_section_d(&w, means, d, Model::num_var_params());
    Model::sample_var_param_gauss_proposal(&q, w);
    assert(copy_vector_section_into_vector_d(p, d, q, 0, q->num_elts) == 0);
    d += q->num_elts;

    for (size_t i = 0; i < _components.size(); i++)
    {
        copy_vector_section_d(&w, means, d, _components[ i ]->num_var_params());
        _components[ i ]->sample_var_param_gauss_proposal(&q, w);
        copy_vector_section_into_vector_d(p, d, q, 0, q->num_elts);
        d += q->num_elts;
    }

    free_vector_d(w);
    free_vector_d(q);

    assert(d == D);
}

void Object::get_var_param_gauss_proposal_sigma(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    Vector_d* q = 0;

    Model::get_var_param_gauss_proposal_sigma(&q);
    assert(copy_vector_section_into_vector_d(p, d, q, 0, q->num_elts) == 0);
    d += q->num_elts;

    for (size_t i = 0; i < _components.size(); i++)
    {
        _components[ i ]->get_var_param_gauss_proposal_sigma(&q);
        assert(copy_vector_section_into_vector_d(p, d, q, 0, q->num_elts) == 0);
        d += q->num_elts;
    }

    free_vector_d(q);

    assert(d == D);
}

void Object::serialize(const std::vector<Model*>& models)
{
    Model::serialize(models);
    _lowest_component.serialize(models);
}

void Object::deserialize(const std::vector<Model*>& models)
{
    Model::deserialize(models);
    _lowest_component.deserialize(models);
}

void Object::update_pointers
(
    const std::vector<Model*>&       new_models,
    const std::vector<const Model*>& old_models
)
{
    Model::update_pointers(new_models, old_models);
    _lowest_component.update_pointers(new_models, old_models);
}

/** The object must be in object coordinates. */
void Object::update_graphical_rep()
{
    Model::update_graphical_rep();

    for (size_t i = 0; i < num_components(); i++)
    {
        Component* c = component(i);
        if (c->graphical_rep_is_outdated())
        {
            c->update_graphical_rep();
        }
    }
}

/** 
 * Assumes the GL projection and modelview matrices are already setup for the
 * current View (Camera).
 *
 * Does not set color or material effects.
 */
void Object::render() const 
{ 
    wire_occlude_render(); 
    wire_render(); 
}

void Object::render_with_ids(const View* view) const 
{ 
    wire_occlude_render(); 
    wire_render_with_ids(view); 
}

/** 
 * Assumes the GL projection and modelview matrices are already setup for the
 * current View (Camera).
 *
 * Does not set color or material effects. 
 */
void Object::wire_render() const
{
    for (size_t i = 0; i < _components.size(); i++)
    {
        _components[ i ]->wire_render();
    }
}

/**
 * Assumes the GL projection and modelview matrices are already setup for the
 * current View (Camera).
 */
void Object::wire_render_with_ids(const View* view) const
{
    uint32_t id = 1;

    for (size_t i = 0; i < _components.size(); i++)
    {
        id = _components[ i ]->set_sequential_face_ids(id);
        _components[ i ]->wire_render_with_ids(view);
    }
}

/** 
 * Assumes the GL projection and modelview matrices are already setup for the
 * current View (Camera).
 *
 * Does not render into the color buffer, only the depth buffer.
 *
 * After calling this method, call wire_render() to render this Object with
 * hidden lines occluded.
 */
void Object::wire_occlude_render() const
{
    for (size_t i = 0; i < _components.size(); i++)
    {
        _components[ i ]->wire_occlude_render();
    }
}

/** 
 * Assumes the GL projection and modelview matrices are already setup for the
 * current View (Camera).
 *
 * Does not set color or material effects.
 */
void Object::solid_render() const
{
    for (size_t i = 0; i < _components.size(); i++)
    {
        _components[ i ]->solid_render();
    }
}

/**
 * Assumes the GL projection and modelview matrices are already setup for the
 * current View (Camera).
 */
void Object::solid_render_with_ids(const View* view) const
{
    uint32_t id = 1;

    for (size_t i = 0; i < _components.size(); i++)
    {
        id = _components[ i ]->set_sequential_face_ids(id);
        _components[ i ]->solid_render_with_ids(view);
    }
}

/**
 * Assumes the GL projection and modelview matrices are already setup for the
 * current View (Camera).
 *
 * @return A collection of projected components.
 */
const std::vector<jwscxx::graphics::Polygon_d>& Object::project()
{
    using std::vector;
    using jwscxx::graphics::Polygon_d;

    _projection.clear();

    for (uint32_t i = 0; i < _components.size(); i++)
    {
        vector<const Polygon_d*> faces = _components[ i ]->faces();

        for (uint32_t j = 0; j < faces.size(); j++)
        {
            _projection.push_back(*(faces[ j ]));
        }
    }

    for (uint32_t i = 0; i < _projection.size(); i++)
    {
        _projection[ i ].set_id(i+1);
        _projection[ i ].project();
    }

    return _projection;
}

uint32_t Object::num_components() const
{
    return _components.size();
}

Component* Object::component(size_t i) const
    throw (jwscxx::base::Arg_error)
{
    if (i >= num_components())
    {
        throw jwscxx::base::Arg_error("Object component index too large");
    }

    return _components[ i ];
}

size_t Object::component(const Component* c) const
    throw (jwscxx::base::Arg_error)
{
    std::vector<Component*>::const_iterator c_it;
    
    c_it = find(_components.begin(), _components.end(), c);
    if (c_it == _components.end())
    {
        throw jwscxx::base::Arg_error("Component not found in object");
    }

    return c_it - _components.begin();
}

const std::vector<Component*>& Object::components() const
{
    return _components;
}

/** 
 * The function detaches the object position from the currently set base
 * component, repositions the object and reattaches it to the lowest component.
 *
 * If there are no components, the function does nothing. UNLESS, the base
 * component is set. In which case it is detached and un-set.
 *
 * If there are components, and the base is not the lowest component already, it
 * is detached and re-attached as the lowest component.
 */
void Object::update_lowest_component()
{
    using namespace jwsc;

    std::vector<Component*>::iterator c_it;
    Component::Compare cmpr;
    
    c_it = min_element(_components.begin(), _components.end(), cmpr);

    if (num_components() && (*c_it == lowest_component()))
        return;

    if (lowest_component())
    {
        detach_deps(this, lowest_component());
        lowest_component(0);
    }

    if (num_components())
    {
        position_object_below_component(*c_it);
        attach_deps(this, *c_it);
        lowest_component(*c_it);
    }
}

void Object::swap_component_indices(uint32_t i, uint32_t j)
{
    assert(i < num_components());
    assert(j < num_components());

    if (i == j)
        return;

    Component* c = _components[ i ];
    _components[ i ] = _components[ j ];
    _components[ j ] = c;
}

/** Increments (or loops) the next next component id. */
size_t Object::next_component_id(bool* looped)
{
    static const uint32_t MAX = 0xFF;
    assert(sizeof(size_t) >= sizeof(uint32_t));

    size_t id = _next_component_id;

    if (id == MAX)
    {
        _next_component_id = 1;
        *looped = true;
    }
    else
    {
        _next_component_id++;
        *looped = false;
    }

    return id;
}

/** 
 * Increments (or loops) the next next component ID.
 *
 * Aborts if all IDs are in use.
 */
size_t Object::next_unused_component_id()
{
    size_t   id = 0;
    size_t   used_id = 0;
    bool     looped;
    uint32_t num_looped = 0;
    std::vector<size_t>::iterator it;

    while (id == used_id)
    {
        id = next_component_id(&looped);
        it = lower_bound(_used_component_ids.begin(),
                _used_component_ids.end(), id);
        used_id = (it == _used_component_ids.end()) ? 0 : *it;

        if (looped)
        {
            num_looped++;
            assert(num_looped < 2);
        }
    }

    _used_component_ids.insert(it, id);

    return id;
}

void Object::remove_used_component_id(size_t id)
{
    std::vector<size_t>::iterator id_it;

    id_it = lower_bound(_used_component_ids.begin(),
            _used_component_ids.end(), id);
    assert(id_it != _used_component_ids.end());

    _used_component_ids.erase(id_it);
}

/** If there is no base component, null is returned. */
Component* Object::lowest_component() const
{
    return _lowest_component.ptr();
}

void Object::lowest_component(Component* c)
{
    _lowest_component.ptr(c);
}

void Object::add_component(Component* c)
{
    _components.push_back(c);
    c->object(this);
    c->id(next_unused_component_id());
    update_lowest_component();
}

void Object::remove_component(Component* c)
{
    std::vector<Component*>::iterator c_it;
    c_it = find(_components.begin(), _components.end(), c);
    assert(c_it != _components.end());

    _components.erase(c_it);
    remove_used_component_id(c->id());
    c->id(0);
    c->object(0);
    update_lowest_component();
}

double Object::intersection_ratio
(
    const Component* c1, 
    const Component* c2
) 
const
{
    const Abstract_block* b1;
    const Abstract_block* b2;

    double r = 0;

    if ((b1 = dynamic_cast<const Abstract_block*>(c1)) &&
        (b2 = dynamic_cast<const Abstract_block*>(c2)))
    {
        r = intersection_ratio(b1, b2);
    }
    else
    {
        abort(); // Unknown component type.
    }

    return r;
}

/** The object must be in object coordinates. */
double Object::intersection_ratio
(
    const Abstract_block* b1, 
    const Abstract_block* b2
) 
const
{
    using namespace jwsc;

    double v  = 1;
    double v1 = b1->size(0)*b1->size(1)*b1->size(2);
    double v2 = b2->size(0)*b2->size(1)*b2->size(2);
    double v_max = (v1 > v2) ? v1 : v2;

    Vector_d* p1_pos = 0;
    Vector_d* p1_neg = 0;
    Vector_d* p2_pos = 0;
    Vector_d* p2_neg = 0;

    for (uint32_t i = 0; i < 3; i++)
    {
        uint32_t j = i*2;

        b1->ctr_attach_pt_v(&p1_pos, j+0);
        b1->ctr_attach_pt_v(&p1_neg, j+1);
        transform_pt_from_world_to_object_coords(p1_pos);
        transform_pt_from_world_to_object_coords(p1_neg);
        double x1_pos = p1_pos->elts[ i ];
        double x1_neg = p1_neg->elts[ i ];

        b2->ctr_attach_pt_v(&p2_pos, j+0);
        b2->ctr_attach_pt_v(&p2_neg, j+1);
        transform_pt_from_world_to_object_coords(p2_pos);
        transform_pt_from_world_to_object_coords(p2_neg);
        double x2_pos = p2_pos->elts[ i ];
        double x2_neg = p2_neg->elts[ i ];

        if (x1_neg >= x2_pos || x2_neg >= x1_pos)
        {
            free_vector_d(p1_pos);
            free_vector_d(p1_neg);
            free_vector_d(p2_pos);
            free_vector_d(p2_neg);
            return 0;
        }

        if (x1_neg >= x2_neg && x1_pos <= x2_pos)
        {
            v *= (x1_pos - x1_neg);
        }
        else if (x2_neg >= x1_neg && x2_pos <= x1_pos)
        {
            v *= (x2_pos - x2_neg);
        }
        else if (x1_neg >= x2_neg)
        {
            v *= (x2_pos - x1_neg);
        }
        else
        {
            assert(x2_neg >= x1_neg);
            v *= (x1_pos - x2_neg);
        }
    }

    free_vector_d(p1_pos);
    free_vector_d(p1_neg);
    free_vector_d(p2_pos);
    free_vector_d(p2_neg);

    return v / v_max;
}

/**
 * @return A uniform random sample from the index set.
 *
 * @throw  jwscxx::base::Arg_error  No indices in the set.
 */
uint32_t Object::sample_component_index(const std::vector<uint32_t>& indices)
    throw (jwscxx::base::Arg_error)
{
    if (indices.size() == 0)
    {
        throw jwscxx::base::Arg_error("No components");
    }

    size_t i = (size_t)floor(jwsc::sample_uniform_pdf_d(0, indices.size()));
    if (i == indices.size())
    {
        i--;
    }

    return indices[ i ];
}

/**
 * @return An random component index in the set.
 *
 * @throw  jwscxx::base::Arg_error  No components in the set.
 */
uint32_t Object::select_random_component() const 
    throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < num_components(); i++)
    {
        if (_components[ i ])
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}

/**
 * @return An random block index in the set.
 *
 * @throw  jwscxx::base::Arg_error  No block components in the set.
 */
uint32_t Object::select_random_block() const 
    throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < num_components(); i++)
    {
        if (dynamic_cast<Block*>(_components[ i ]))
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}

/**
 * @return An random block pair index in the set.
 *
 * @throw  jwscxx::base::Arg_error  No block components in the set.
 */
uint32_t Object::select_random_pair() const 
    throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < num_components(); i++)
    {
        if (dynamic_cast<Pair*>(_components[ i ]))
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}

/**
 * @return An random block quartet index in the set.
 *
 * @throw  jwscxx::base::Arg_error  No block components in the set.
 */
uint32_t Object::select_random_quartet() const 
    throw (jwscxx::base::Arg_error)
{
    std::vector<uint32_t> indices;

    for (uint32_t i = 0; i < num_components(); i++)
    {
        if (dynamic_cast<Quartet*>(_components[ i ]))
        {
            indices.push_back(i);
        }
    }

    return Object::sample_component_index(indices);
}

/** 
 * The Component @em c is attached to this Object and Parameter dependencies are
 * established. The component is translated and rotated into position with this
 * Object.
 *
 * There must not be any other components in the Object already.
 */
void Object::attach(Component* c, Component::Attach_type t)
{
    assert(num_components() == 0);
    assert(lowest_component() == 0);

    lowest_component(c);
    position_attachment(this, c, t);
    attach_deps(this, c);
    add_component(c);
}

/** 
 * Attaches Component @em c2 to Component @em c1 and adds @em c2 to this
 * Object. Parameter dependencies between the two components are established.
 * The attaching component is translated and rotated into position.
 *
 * The component @em c1 must already be in this Object.
 */
void Object::attach
(
    Component*             c1, 
    size_t                 i1, 
    Component::Attach_type t1, 
    Component*             c2, 
    size_t                 i2, 
    Component::Attach_type t2
) 
{
    try { component(c1); } 
    catch (jwscxx::base::Arg_error e) { abort(); }

    position_attachment(c1, i1, t1, c2, i2, t2);
    attach_deps_and_conds(c1, i1, t1, c2, i2, t2);
    add_component(c2);
}

/** 
 * Detaches Component @em c from this Object.
 *
 * The component cannot have another attachment for this operation.
 */
Component* Object::detach(Component* c)
{
    assert(num_components() == 1);
    assert(c->num_attachments() == 0);
    try { component(c); } 
    catch (jwscxx::base::Arg_error e) { abort(); }

    lowest_component(0);
    detach_deps(this, c);
    position_detachment(c);
    remove_component(c);

    return c;
}

/*
 * Detaches Component @em c2 from Component @em c1 and removes @em c2 from
 * this Object. Parameter dependencies between the two components are
 * removed.
 *
 * Both components must be in this Object.
 */
Component* Object::detach
(
    Component* c1, 
    size_t     i1, 
    Component* c2, 
    size_t     i2
) 
{
    try 
    { 
        component(c1);
        component(c2); 
    }
    catch (jwscxx::base::Arg_error e) { abort(); }
    assert(c2->num_attachments() == 1);

    detach_deps_and_conds(c1, i1, c2, i2);
    position_detachment(c2);
    remove_component(c2);

    return c2;
}

void Object::attach_deps(Object* obj, Component* c) 
{
    Abstract_block* b;

    if ((b = dynamic_cast<Abstract_block*>(c)))
    {
        attach_shift_deps(obj, b);
        attach_scale_deps(obj, b);
    }
    else
    {
        abort(); // Unknown component type.
    }
}

void Object::attach_shift_deps(Object* obj, Abstract_block* b) 
{
    typedef Component::Parameter P;

    const size_t y = 1;
    const size_t pos_h = 2*y + 0;
    const size_t neg_h = 2*y + 1;

    // It should never be the case we attach the object to a block that has a
    // neg_h attachment already.
    assert(b->attachment(neg_h) == 0);

    b->attach_shift_deps(pos_h, y);

    for (size_t j = 0; j < 3; j++)
    {
        P* obj_p_j      = &(obj->_posn_p[ j ]);
        P* blk_p_j      = &(b->_posn_p[ j ]);
        P* blk_a_neg_hj = &(b->_ctr_attach_pts_p[ neg_h ][ j ]);

        obj_p_j->add_shift_dep(P::Pos_dependent(blk_p_j));
        blk_a_neg_hj->add_shift_dep(P::Pos_dependent(obj_p_j));

        for (size_t k = 0; k < 6; k += 2)
        {
            P* blk_a_pos_kj = &(b->_ctr_attach_pts_p[ k+0 ][ j ]);
            P* blk_a_neg_kj = &(b->_ctr_attach_pts_p[ k+1 ][ j ]);

            obj_p_j->add_shift_dep(P::Pos_dependent(blk_a_pos_kj));
            obj_p_j->add_shift_dep(P::Pos_dependent(blk_a_neg_kj));
        }
    }
}

void Object::attach_scale_deps(Object* obj, Abstract_block* b) 
{
    typedef Component::Parameter P;

    const size_t y = 1;
    const size_t pos_h = 2*y + 0;
    const size_t neg_h = 2*y + 1;

    // It should never be the case we attach the object to a block that has a
    // neg_h attachment already.
    assert(b->attachment(neg_h) == 0);

    b->attach_scale_deps(pos_h, y);

    for (size_t j = 0; j < 3; j++)
    {
        P* obj_p_j      = &(obj->_posn_p[ j ]);
        P* blk_p_j      = &(b->_posn_p[ j ]);
        P* blk_a_neg_hj = &(b->_ctr_attach_pts_p[ neg_h ][ j ]);

        if (j == y) 
        {
            obj_p_j->add_scale_dep(P::Pos_dependent(blk_a_neg_hj));
            blk_a_neg_hj->add_scale_dep(P::Pos_dependent(obj_p_j));
        }
        else
        {
            obj_p_j->add_scale_dep(P::Pos_dependent(blk_p_j));
            blk_a_neg_hj->add_scale_dep(P::Pos_dependent(obj_p_j));

            for (size_t k = 0; k < 6; k += 2)
            {
                P* blk_a_pos_kj = &(b->_ctr_attach_pts_p[ k+0 ][ j ]);
                P* blk_a_neg_kj = &(b->_ctr_attach_pts_p[ k+1 ][ j ]);

                obj_p_j->add_scale_dep(P::Pos_dependent(blk_a_pos_kj));
                obj_p_j->add_scale_dep(P::Pos_dependent(blk_a_neg_kj));
            }
        }
    }
}

/** Attaches Component @em c2 to Component @em c1. */
void Object::attach_deps_and_conds
(
    Component*             c1, 
    size_t                 i1,
    Component::Attach_type t1,
    Component*             c2, 
    size_t                 i2,
    Component::Attach_type t2
)
{
    Abstract_block* b1;
    Abstract_block* b2;

    assert(c1->attachment(i1) == 0);
    assert(c2->attachment(i2) == 0);

    if ((b1 = dynamic_cast<Abstract_block*>(c1)) &&
        (b2 = dynamic_cast<Abstract_block*>(c2)))
    {
        attach_deps(b1, i1, t1, b2, i2, t2);
        attach_conds(b1, i1, b2, i2);
    }
    else
    {
        abort(); // Unknown component type.
    }

    c1->attachment(i1, c2);
    c2->attachment(i2, c1);
}

/** Attaches block @em b2 to block @em b1. */
void Object::attach_deps
(
    Abstract_block*        b1, 
    size_t                 j1, 
    Component::Attach_type t1,
    Abstract_block*        b2, 
    size_t                 j2,
    Component::Attach_type t2
) 
{
    typedef Component::Parameter P;
    typedef Component::Attach_type Attach_t;

    // If b1 is not +w (j=0), +h (j=2), +l (j=4), swap it with b2. That way we
    // are always attaching a positive side of b1 to a negative side of b2.
    // There should not be any attachments already.
    if (j1 % 2)
    {
        Abstract_block* b = b1; b1 = b2; b2 = b;
        size_t   j = j1; j1 = j2; j2 = j;
        Attach_t t = t1; t1 = t2; t2 = t;
    }

    assert(j1 % 2 == 0);
    assert(j1 == j2-1);

    size_t i = j1 / 2;
    size_t j = j1;

    P* b2_s_i      = &(b2->_size_p[ i ]);
    P* b2_a_pos_ji = &(b2->_ctr_attach_pts_p[ j+0 ][ i ]);

    // If b2 has not already had its neg attach deps removed (by the object for
    // example), do it now.
    if (b2_s_i->has_shift_dep(P::Pos_half_dependent(b2_a_pos_ji)))
    {
        b2->attach_shift_deps(j, i);
    }
    if (b2_s_i->has_scale_dep(P::Fancy_pos_half_dependent(b2_a_pos_ji)))
    {
        b2->attach_scale_deps(j, i);
    }

    if (t1 == Component::CTR_ATTACH && t2 == Component::CTR_ATTACH)
    {
        ctr_ctr_attach_shift_deps(b1, j1, b2, j2);
        ctr_ctr_attach_scale_deps(b1, j1, b2, j2);
    }
    else if (t1 == Component::VAR_ATTACH && t2 == Component::CTR_ATTACH)
    {
        var_ctr_attach_shift_deps(b1, j1, b2, j2);
        var_ctr_attach_scale_deps(b1, j1, b2, j2);
        b1->enable_var_attach_pt(j1);
    }
    else if (t1 == Component::CTR_ATTACH && t2 == Component::VAR_ATTACH)
    {
        ctr_var_attach_shift_deps(b1, j1, b2, j2);
        ctr_var_attach_scale_deps(b1, j1, b2, j2);
        b2->enable_var_attach_pt(j2);
    }
    else
    {
        abort();
    }
}

/** Attaches block @em b2 to block @em b1. */
void Object::attach_conds
(
    Abstract_block*  b1, 
    size_t           j1, 
    Abstract_block*  b2, 
    size_t           j2
) 
{
    typedef Component::Parameter P;

    assert(j1/2 == j2/2);

    for (size_t i = 0; i < 3; i++)
    {
        if (i == j1/2)
            continue;

        P* b1_s_i = &(b1->_size_p[ i ]);
        P* b2_s_i = &(b2->_size_p[ i ]);

        b2_s_i->add_cond(P::Pointer(b1_s_i));
        b2_s_i->has_ratio_prior(true);
    }

    Block_set* bset;
    if ((bset = dynamic_cast<Block_set*>(b1)))
    {
        bset->add_inner_ctr_attach_conds(j1, b2->_size_p);
    }
    else if ((bset = dynamic_cast<Block_set*>(b2)))
    {
        bset->add_inner_ctr_attach_conds(j2, b1->_size_p);
    }
}

/** Attaches block @em b2 to block @em b1. */
void Object::ctr_ctr_attach_shift_deps
(
    Abstract_block* b1, 
    size_t          j1, 
    Abstract_block* b2, 
    size_t          j2
) 
{
    typedef Component::Parameter P;

    assert(j1 % 2 == 0);
    assert(j1 == j2-1);

    for (size_t i = 0; i < 3; i++)
    {
        P* b1_a_pos_ji = &(b1->_ctr_attach_pts_p[ j1 ][ i ]);
        P* b2_a_neg_ji = &(b2->_ctr_attach_pts_p[ j2 ][ i ]);

        P* b1_p_i = &(b1->_posn_p[ i ]);
        P* b2_p_i = &(b2->_posn_p[ i ]);

        b1_a_pos_ji->add_shift_dep(P::Pos_dependent(b2_p_i));
        b2_a_neg_ji->add_shift_dep(P::Pos_dependent(b1_p_i));

        for (size_t k = 0; k < 6; k += 2)
        {
            P* b1_a_pos_ki = &(b1->_ctr_attach_pts_p[ k+0 ][ i ]);
            P* b1_a_neg_ki = &(b1->_ctr_attach_pts_p[ k+1 ][ i ]);

            P* b2_a_pos_ki = &(b2->_ctr_attach_pts_p[ k+0 ][ i ]);
            P* b2_a_neg_ki = &(b2->_ctr_attach_pts_p[ k+1 ][ i ]);

            b1_a_pos_ji->add_shift_dep(P::Pos_dependent(b2_a_pos_ki));
            b1_a_pos_ji->add_shift_dep(P::Pos_dependent(b2_a_neg_ki));
            b2_a_neg_ji->add_shift_dep(P::Pos_dependent(b1_a_pos_ki));
            b2_a_neg_ji->add_shift_dep(P::Pos_dependent(b1_a_neg_ki));
        }
    }
}

/** 
 * Attaches block @em b2 with a center attachment point to block @em b1 with a
 * variable attachment point. The @em b1 attachment point must be on the
 * positive ordinate and @em b2 the negative.
 */
void Object::var_ctr_attach_shift_deps
(
    Abstract_block* b1, 
    size_t          j1, 
    Abstract_block* b2, 
    size_t          j2
) 
{
    typedef Component::Parameter P;

    assert(j1 % 2 == 0);
    assert(j1 == j2-1);

    for (size_t i = 0; i < 3; i++)
    {
        P* b1_a_pos_ji = 0;
        P* b2_a_neg_ji = &(b2->_ctr_attach_pts_p[ j2 ][ i ]);

        P* b1_p_i = &(b1->_posn_p[ i ]);
        P* b2_p_i = &(b2->_posn_p[ i ]);

        if (i == j1/2)
        {
            b1_a_pos_ji = &(b1->_ctr_attach_pts_p[ j1 ][ i ]);
        }
        else
        {
            b1_a_pos_ji = &(b1->_var_attach_pts_p[ j1 ][ i ]);

        }

        b1_a_pos_ji->add_shift_dep(P::Pos_dependent(b2_p_i));
        b2_a_neg_ji->add_shift_dep(P::Pos_dependent(b1_p_i));

        for (size_t k = 0; k < 6; k += 2)
        {
            P* b1_a_pos_ki = &(b1->_ctr_attach_pts_p[ k+0 ][ i ]);
            P* b1_a_neg_ki = &(b1->_ctr_attach_pts_p[ k+1 ][ i ]);

            P* b2_a_pos_ki = &(b2->_ctr_attach_pts_p[ k+0 ][ i ]);
            P* b2_a_neg_ki = &(b2->_ctr_attach_pts_p[ k+1 ][ i ]);

            b1_a_pos_ji->add_shift_dep(P::Pos_dependent(b2_a_pos_ki));
            b1_a_pos_ji->add_shift_dep(P::Pos_dependent(b2_a_neg_ki));
            b2_a_neg_ji->add_shift_dep(P::Pos_dependent(b1_a_pos_ki));
            b2_a_neg_ji->add_shift_dep(P::Pos_dependent(b1_a_neg_ki));
        }
    }
}

/** Attaches block @em b2 to block @em b1. */
void Object::ctr_var_attach_shift_deps
(
    Abstract_block* b1, 
    size_t          j1, 
    Abstract_block* b2, 
    size_t          j2
) 
{
    typedef Component::Parameter P;

    assert(j1 % 2 == 0);
    assert(j1 == j2-1);

    for (size_t i = 0; i < 3; i++)
    {
        P* b1_a_pos_ji = &(b1->_ctr_attach_pts_p[ j1 ][ i ]);
        P* b2_a_neg_ji = 0;

        P* b1_p_i = &(b1->_posn_p[ i ]);
        P* b2_p_i = &(b2->_posn_p[ i ]);

        if (i == j1/2)
        {
            b2_a_neg_ji = &(b2->_ctr_attach_pts_p[ j2 ][ i ]);
        }
        else
        {
            b2_a_neg_ji = &(b2->_var_attach_pts_p[ j2 ][ i ]);
        }

        b1_a_pos_ji->add_shift_dep(P::Pos_dependent(b2_p_i));
        b2_a_neg_ji->add_shift_dep(P::Pos_dependent(b1_p_i));

        for (size_t k = 0; k < 6; k += 2)
        {
            P* b1_a_pos_ki = &(b1->_ctr_attach_pts_p[ k+0 ][ i ]);
            P* b1_a_neg_ki = &(b1->_ctr_attach_pts_p[ k+1 ][ i ]);

            P* b2_a_pos_ki = &(b2->_ctr_attach_pts_p[ k+0 ][ i ]);
            P* b2_a_neg_ki = &(b2->_ctr_attach_pts_p[ k+1 ][ i ]);

            b1_a_pos_ji->add_shift_dep(P::Pos_dependent(b2_a_pos_ki));
            b1_a_pos_ji->add_shift_dep(P::Pos_dependent(b2_a_neg_ki));
            b2_a_neg_ji->add_shift_dep(P::Pos_dependent(b1_a_pos_ki));
            b2_a_neg_ji->add_shift_dep(P::Pos_dependent(b1_a_neg_ki));
        }
    }
}

/** Attaches block @em b2 to block @em b1. */
void Object::ctr_ctr_attach_scale_deps
(
    Abstract_block* b1, 
    size_t          j1, 
    Abstract_block* b2, 
    size_t          j2
) 
{
    typedef Component::Parameter P;

    assert(j1 % 2 == 0);
    assert(j1 == j2-1);

    for (size_t i = 0; i < 3; i++)
    {
        size_t j = i*2;

        P* b1_a_pos_ji = &(b1->_ctr_attach_pts_p[ j1 ][ i ]);
        P* b2_a_neg_ji = &(b2->_ctr_attach_pts_p[ j2 ][ i ]);

        P* b1_p_i = &(b1->_posn_p[ i ]);
        P* b2_p_i = &(b2->_posn_p[ i ]);

        if (j == j1)
        {
            b1_a_pos_ji->add_scale_dep(P::Pos_dependent(b2_a_neg_ji));
            b2_a_neg_ji->add_scale_dep(P::Pos_dependent(b1_a_pos_ji));
        }
        else
        {
            b1_a_pos_ji->add_scale_dep(P::Pos_dependent(b2_p_i));
            b2_a_neg_ji->add_scale_dep(P::Pos_dependent(b1_p_i));

            for (size_t k = 0; k < 6; k += 2)
            {
                P* b1_a_pos_ki = &(b1->_ctr_attach_pts_p[ k+0 ][ i ]);
                P* b1_a_neg_ki = &(b1->_ctr_attach_pts_p[ k+1 ][ i ]);

                P* b2_a_pos_ki = &(b2->_ctr_attach_pts_p[ k+0 ][ i ]);
                P* b2_a_neg_ki = &(b2->_ctr_attach_pts_p[ k+1 ][ i ]);

                b1_a_pos_ji->add_scale_dep(P::Pos_dependent(b2_a_pos_ki));
                b1_a_pos_ji->add_scale_dep(P::Pos_dependent(b2_a_neg_ki));
                b2_a_neg_ji->add_scale_dep(P::Pos_dependent(b1_a_pos_ki));
                b2_a_neg_ji->add_scale_dep(P::Pos_dependent(b1_a_neg_ki));
            }
        }
    }
}

/** 
 * Attaches block @em b2 with a center attachment point to block @em b1 with a
 * variable attachment point. The @em b1 attachment point must be on the
 * positive ordinate and @em b2 the negative.
 */
void Object::var_ctr_attach_scale_deps
(
    Abstract_block* b1, 
    size_t          j1, 
    Abstract_block* b2, 
    size_t          j2
) 
{
    typedef Component::Parameter P;

    assert(j1 % 2 == 0);
    assert(j1 == j2-1);

    for (size_t i = 0; i < 3; i++)
    {
        size_t j = i*2;

        P* b1_a_pos_ji = 0;
        P* b2_a_neg_ji = &(b2->_ctr_attach_pts_p[ j2 ][ i ]);

        P* b2_p_i = &(b2->_posn_p[ i ]);

        if (j == j1)
        {
            b1_a_pos_ji = &(b1->_ctr_attach_pts_p[ j1 ][ i ]);

            b1_a_pos_ji->add_scale_dep(P::Pos_dependent(b2_a_neg_ji));
            b2_a_neg_ji->add_scale_dep(P::Pos_dependent(b1_a_pos_ji));
        }
        else
        {
            b1_a_pos_ji = &(b1->_var_attach_pts_p[ j1 ][ i ]);

            b1_a_pos_ji->add_scale_dep(P::Pos_dependent(b2_p_i));
            b2_a_neg_ji->add_scale_dep(P::Pos_dependent(b1_a_pos_ji));

            for (size_t k = 0; k < 6; k += 2)
            {
                P* b2_a_pos_ki = &(b2->_ctr_attach_pts_p[ k+0 ][ i ]);
                P* b2_a_neg_ki = &(b2->_ctr_attach_pts_p[ k+1 ][ i ]);

                b1_a_pos_ji->add_scale_dep(P::Pos_dependent(b2_a_pos_ki));
                b1_a_pos_ji->add_scale_dep(P::Pos_dependent(b2_a_neg_ki));
            }
        }
    }
}

/** Attaches block @em b2 to block @em b1. */
void Object::ctr_var_attach_scale_deps
(
    Abstract_block* b1, 
    size_t          j1, 
    Abstract_block* b2, 
    size_t          j2
) 
{
    typedef Component::Parameter P;

    assert(j1 % 2 == 0);
    assert(j1 == j2-1);

    for (size_t i = 0; i < 3; i++)
    {
        size_t j = i*2;

        P* b1_a_pos_ji = &(b1->_ctr_attach_pts_p[ j1 ][ i ]);
        P* b2_a_neg_ji = 0;

        P* b1_p_i = &(b1->_posn_p[ i ]);

        if (j == j1)
        {
            b2_a_neg_ji = &(b2->_ctr_attach_pts_p[ j2 ][ i ]);

            b1_a_pos_ji->add_scale_dep(P::Pos_dependent(b2_a_neg_ji));
            b2_a_neg_ji->add_scale_dep(P::Pos_dependent(b1_a_pos_ji));
        }
        else
        {
            b2_a_neg_ji = &(b2->_var_attach_pts_p[ j2 ][ i ]);

            b2_a_neg_ji->add_scale_dep(P::Pos_dependent(b1_p_i));
            b1_a_pos_ji->add_scale_dep(P::Pos_dependent(b2_a_neg_ji));

            for (size_t k = 0; k < 6; k += 2)
            {
                P* b1_a_pos_ki = &(b2->_ctr_attach_pts_p[ k+0 ][ i ]);
                P* b1_a_neg_ki = &(b2->_ctr_attach_pts_p[ k+1 ][ i ]);

                b2_a_neg_ji->add_scale_dep(P::Pos_dependent(b1_a_pos_ki));
                b2_a_neg_ji->add_scale_dep(P::Pos_dependent(b1_a_neg_ki));
            }
        }
    }
}

void Object::detach_deps(Object* obj, Component* c) 
{
    Abstract_block* b;

    if ((b = dynamic_cast<Abstract_block*>(c)))
    {
        detach_shift_deps(obj, b);
        detach_scale_deps(obj, b);
    }
    else
    {
        abort(); // Unknown component type.
    }
}

void Object::detach_shift_deps(Object* obj, Abstract_block* b)
{
    typedef Component::Parameter P;

    const size_t y = 1;
    const size_t pos_h = 2*y + 0;
    const size_t neg_h = 2*y + 1;

    if (!b->attachment(neg_h))
    {
        b->detach_shift_deps(pos_h, y);
    }

    for (size_t j = 0; j < 3; j++)
    {
        P* obj_p_j      = &(obj->_posn_p[ j ]);
        P* blk_p_j      = &(b->_posn_p[ j ]);
        P* blk_a_neg_hj = &(b->_ctr_attach_pts_p[ neg_h ][ j ]);

        obj_p_j->remove_shift_dep(P::Pos_dependent(blk_p_j));
        blk_a_neg_hj->remove_shift_dep(P::Pos_dependent(obj_p_j));

        for (size_t k = 0; k < 6; k += 2)
        {
            P* blk_a_pos_kj = &(b->_ctr_attach_pts_p[ k+0 ][ j ]);
            P* blk_a_neg_kj = &(b->_ctr_attach_pts_p[ k+1 ][ j ]);

            obj_p_j->remove_shift_dep(P::Pos_dependent(blk_a_pos_kj));
            obj_p_j->remove_shift_dep(P::Pos_dependent(blk_a_neg_kj));
        }
    }
}

void Object::detach_scale_deps(Object* obj, Abstract_block* b)
{
    typedef Component::Parameter P;

    const size_t y = 1;
    const size_t pos_h = 2*y + 0;
    const size_t neg_h = 2*y + 1;

    P* blk_a_neg_hy = &(b->_ctr_attach_pts_p[ neg_h ][ y ]);

    if (!b->attachment(neg_h))
    {
        b->detach_scale_deps(pos_h, y);
    }

    for (size_t j = 0; j < 3; j++)
    {
        P* obj_p_j      = &(obj->_posn_p[ j ]);
        P* blk_p_j      = &(b->_posn_p[ j ]);
        P* blk_a_neg_hj = &(b->_ctr_attach_pts_p[ neg_h ][ j ]);

        if (j == y) 
        {
            obj_p_j->remove_scale_dep(P::Pos_dependent(blk_a_neg_hy));
            blk_a_neg_hj->remove_scale_dep(P::Pos_dependent(obj_p_j));
        }
        else
        {
            obj_p_j->remove_scale_dep(P::Pos_dependent(blk_p_j));
            blk_a_neg_hj->remove_scale_dep(P::Pos_dependent(obj_p_j));

            for (size_t k = 0; k < 6; k += 2)
            {
                P* blk_a_pos_kj = &(b->_ctr_attach_pts_p[ k+0 ][ j ]);
                P* blk_a_neg_kj = &(b->_ctr_attach_pts_p[ k+1 ][ j ]);

                obj_p_j->remove_scale_dep(P::Pos_dependent(blk_a_pos_kj));
                obj_p_j->remove_scale_dep(P::Pos_dependent(blk_a_neg_kj));
            }
        }
    }
}

/** Detaches Component @em c2 from Component @em c1. */
void Object::detach_deps_and_conds
(
    Component* c1, 
    size_t     i1, 
    Component* c2, 
    size_t     i2
)
{
    Abstract_block* b1;
    Abstract_block* b2;

    assert(c1->attachment(i1) == (Component*)c2);
    assert(c2->attachment(i2) == (Component*)c1);

    if ((b1 = dynamic_cast<Abstract_block*>(c1)) &&
        (b2 = dynamic_cast<Abstract_block*>(c2)))
    {
        detach_deps(b1, i1, b2, i2);
        detach_conds(b1, i1, b2, i2);
    }
    else
    {
        abort(); // Unknown component type.
    }

    c1->attachment(i1, 0);
    c2->attachment(i2, 0);
}

/** Attaches block @em b2 to block @em b1. */
void Object::detach_deps
(
    Abstract_block* b1, 
    size_t          j1, 
    Abstract_block* b2, 
    size_t          j2
)
{
    typedef Component::Attach_type Attach_t;

    Attach_t t1 = b1->attach_pt_type(j1);
    Attach_t t2 = b2->attach_pt_type(j2);

    // If b1 is not +w (j=0), +h (j=2), +l (j=4), swap it with b2. That way we
    // are always attaching a positive side of b1 to a negative side of b2.
    // There should not be any attachments already.
    if (j1 % 2)
    {
        Abstract_block* b = b1; b1 = b2; b2 = b;
        size_t   j = j1; j1 = j2; j2 = j;
        Attach_t t = t1; t1 = t2; t2 = t;
    }

    if (t1 == Component::CTR_ATTACH && t2 == Component::CTR_ATTACH)
    {
        ctr_ctr_detach_shift_deps(b1, j1, b2, j2);
        ctr_ctr_detach_scale_deps(b1, j1, b2, j2);
    }
    else if (t1 == Component::VAR_ATTACH && t2 == Component::CTR_ATTACH)
    {
        var_ctr_detach_shift_deps(b1, j1, b2, j2);
        var_ctr_detach_scale_deps(b1, j1, b2, j2);
        b1->disable_var_attach_pt(j1);
    }
    else if (t1 == Component::CTR_ATTACH && t2 == Component::VAR_ATTACH)
    {
        ctr_var_detach_shift_deps(b1, j1, b2, j2);
        ctr_var_detach_scale_deps(b1, j1, b2, j2);
        b2->disable_var_attach_pt(j2);
    }
    else
    {
        abort();
    }
}

void Object::detach_conds
(
    Abstract_block* b1, 
    size_t          j1, 
    Abstract_block* b2, 
    size_t          j2
) 
{
    typedef Component::Parameter P;

    assert(j1/2 == j2/2);

    for (size_t i = 0; i < 3; i++)
    {
        if (i == j1/2)
            continue;

        P* b1_s_i = &(b1->_size_p[ i ]);
        P* b2_s_i = &(b2->_size_p[ i ]);

        if (b1_s_i->has_cond(P::Pointer(b2_s_i)))
        {
            b1_s_i->remove_cond(P::Pointer(b2_s_i));
            b1_s_i->has_ratio_prior(false);
        }
        else
        {
            b2_s_i->remove_cond(P::Pointer(b1_s_i));
            b2_s_i->has_ratio_prior(false);
        }
    }

    Block_set* bset;
    if ((bset = dynamic_cast<Block_set*>(b1)))
    {
        bset->remove_inner_ctr_attach_conds(j1, b2->_size_p);
    }
    else if ((bset = dynamic_cast<Block_set*>(b2)))
    {
        bset->remove_inner_ctr_attach_conds(j2, b1->_size_p);
    }
}

/** Detaches block @em b2 from block @em b1. */
void Object::ctr_ctr_detach_shift_deps
(
    Abstract_block* b1, 
    size_t          j1, 
    Abstract_block* b2, 
    size_t          j2
)
{
    typedef Component::Parameter P;

    assert(j1 % 2 == 0);
    assert(j1 == j2-1);

    size_t i = j1 / 2;
    size_t j = j1;

    b2->detach_shift_deps(j, i);

    for (size_t i = 0; i < 3; i++)
    {
        P* b1_a_pos_ji = &(b1->_ctr_attach_pts_p[ j1 ][ i ]);
        P* b2_a_neg_ji = &(b2->_ctr_attach_pts_p[ j2 ][ i ]);

        P* b1_p_i = &(b1->_posn_p[ i ]);
        P* b2_p_i = &(b2->_posn_p[ i ]);

        b1_a_pos_ji->remove_shift_dep(P::Pos_dependent(b2_p_i));
        b2_a_neg_ji->remove_shift_dep(P::Pos_dependent(b1_p_i));

        for (size_t k = 0; k < 6; k += 2)
        {
            P* b1_a_pos_ki = &(b1->_ctr_attach_pts_p[ k+0 ][ i ]);
            P* b1_a_neg_ki = &(b1->_ctr_attach_pts_p[ k+1 ][ i ]);

            P* b2_a_pos_ki = &(b2->_ctr_attach_pts_p[ k+0 ][ i ]);
            P* b2_a_neg_ki = &(b2->_ctr_attach_pts_p[ k+1 ][ i ]);

            b1_a_pos_ji->remove_shift_dep(P::Pos_dependent(b2_a_pos_ki));
            b1_a_pos_ji->remove_shift_dep(P::Pos_dependent(b2_a_neg_ki));
            b2_a_neg_ji->remove_shift_dep(P::Pos_dependent(b1_a_pos_ki));
            b2_a_neg_ji->remove_shift_dep(P::Pos_dependent(b1_a_neg_ki));
        }
    }
}

/** Detaches block @em b2 from block @em b1. */
void Object::var_ctr_detach_shift_deps
(
    Abstract_block* b1, 
    size_t          j1, 
    Abstract_block* b2, 
    size_t          j2
)
{
    typedef Component::Parameter P;

    assert(j1 % 2 == 0);
    assert(j1 == j2-1);

    size_t i = j1 / 2;
    size_t j = j1;

    b2->detach_shift_deps(j, i);

    for (size_t i = 0; i < 3; i++)
    {
        P* b1_a_pos_ji = 0;
        P* b2_a_neg_ji = &(b2->_ctr_attach_pts_p[ j2 ][ i ]);

        P* b1_p_i = &(b1->_posn_p[ i ]);
        P* b2_p_i = &(b2->_posn_p[ i ]);

        if (i == j1/2)
        {
            b1_a_pos_ji = &(b1->_ctr_attach_pts_p[ j1 ][ i ]);
        }
        else
        {
            b1_a_pos_ji = &(b1->_var_attach_pts_p[ j1 ][ i ]);
        }

        b1_a_pos_ji->remove_shift_dep(P::Pos_dependent(b2_p_i));
        b2_a_neg_ji->remove_shift_dep(P::Pos_dependent(b1_p_i));

        for (size_t k = 0; k < 6; k += 2)
        {
            P* b1_a_pos_ki = &(b1->_ctr_attach_pts_p[ k+0 ][ i ]);
            P* b1_a_neg_ki = &(b1->_ctr_attach_pts_p[ k+1 ][ i ]);

            P* b2_a_pos_ki = &(b2->_ctr_attach_pts_p[ k+0 ][ i ]);
            P* b2_a_neg_ki = &(b2->_ctr_attach_pts_p[ k+1 ][ i ]);

            b1_a_pos_ji->remove_shift_dep(P::Pos_dependent(b2_a_pos_ki));
            b1_a_pos_ji->remove_shift_dep(P::Pos_dependent(b2_a_neg_ki));
            b2_a_neg_ji->remove_shift_dep(P::Pos_dependent(b1_a_pos_ki));
            b2_a_neg_ji->remove_shift_dep(P::Pos_dependent(b1_a_neg_ki));
        }
    }
}

/** Detaches block @em b2 from block @em b1. */
void Object::ctr_var_detach_shift_deps
(
    Abstract_block* b1, 
    size_t          j1, 
    Abstract_block* b2, 
    size_t          j2
)
{
    typedef Component::Parameter P;

    assert(j1 % 2 == 0);
    assert(j1 == j2-1);

    size_t i = j1 / 2;
    size_t j = j1;

    b2->detach_shift_deps(j, i);

    for (size_t i = 0; i < 3; i++)
    {
        P* b1_a_pos_ji = &(b1->_ctr_attach_pts_p[ j1 ][ i ]);
        P* b2_a_neg_ji = 0;

        P* b1_p_i = &(b1->_posn_p[ i ]);
        P* b2_p_i = &(b2->_posn_p[ i ]);

        if (i == j1/2)
        {
            b2_a_neg_ji = &(b2->_ctr_attach_pts_p[ j2 ][ i ]);
        }
        else
        {
            b2_a_neg_ji = &(b2->_var_attach_pts_p[ j2 ][ i ]);
        }

        b1_a_pos_ji->remove_shift_dep(P::Pos_dependent(b2_p_i));
        b2_a_neg_ji->remove_shift_dep(P::Pos_dependent(b1_p_i));

        for (size_t k = 0; k < 6; k += 2)
        {
            P* b1_a_pos_ki = &(b1->_ctr_attach_pts_p[ k+0 ][ i ]);
            P* b1_a_neg_ki = &(b1->_ctr_attach_pts_p[ k+1 ][ i ]);

            P* b2_a_pos_ki = &(b2->_ctr_attach_pts_p[ k+0 ][ i ]);
            P* b2_a_neg_ki = &(b2->_ctr_attach_pts_p[ k+1 ][ i ]);

            b1_a_pos_ji->remove_shift_dep(P::Pos_dependent(b2_a_pos_ki));
            b1_a_pos_ji->remove_shift_dep(P::Pos_dependent(b2_a_neg_ki));
            b2_a_neg_ji->remove_shift_dep(P::Pos_dependent(b1_a_pos_ki));
            b2_a_neg_ji->remove_shift_dep(P::Pos_dependent(b1_a_neg_ki));
        }
    }
}

/** Detaches block @em b2 from block @em b1. */
void Object::ctr_ctr_detach_scale_deps
(
    Abstract_block* b1, 
    size_t          j1, 
    Abstract_block* b2, 
    size_t          j2
) 
{
    typedef Component::Parameter P;

    assert(j1 % 2 == 0);
    assert(j1 == j2-1);

    size_t i = j1 / 2;
    size_t j = j1;

    b2->detach_scale_deps(j, i);

    for (size_t i = 0; i < 3; i++)
    {
        size_t j = i*2;

        P* b1_a_pos_ji = &(b1->_ctr_attach_pts_p[ j1 ][ i ]);
        P* b2_a_neg_ji = &(b2->_ctr_attach_pts_p[ j2 ][ i ]);

        P* b1_p_i = &(b1->_posn_p[ i ]);
        P* b2_p_i = &(b2->_posn_p[ i ]);

        if (j == j1)
        {
            b1_a_pos_ji->remove_scale_dep(P::Pos_dependent(b2_a_neg_ji));
            b2_a_neg_ji->remove_scale_dep(P::Pos_dependent(b1_a_pos_ji));
        }
        else
        {
            b1_a_pos_ji->remove_scale_dep(P::Pos_dependent(b2_p_i));
            b2_a_neg_ji->remove_scale_dep(P::Pos_dependent(b1_p_i));

            for (size_t k = 0; k < 6; k += 2)
            {
                P* b1_a_pos_ki = &(b1->_ctr_attach_pts_p[ k+0 ][ i ]);
                P* b1_a_neg_ki = &(b1->_ctr_attach_pts_p[ k+1 ][ i ]);

                P* b2_a_pos_ki = &(b2->_ctr_attach_pts_p[ k+0 ][ i ]);
                P* b2_a_neg_ki = &(b2->_ctr_attach_pts_p[ k+1 ][ i ]);

                b1_a_pos_ji->remove_scale_dep(P::Pos_dependent(b2_a_pos_ki));
                b1_a_pos_ji->remove_scale_dep(P::Pos_dependent(b2_a_neg_ki));
                b2_a_neg_ji->remove_scale_dep(P::Pos_dependent(b1_a_pos_ki));
                b2_a_neg_ji->remove_scale_dep(P::Pos_dependent(b1_a_neg_ki));
            }
        }
    }
}

void Object::var_ctr_detach_scale_deps
(
    Abstract_block* b1, 
    size_t          j1, 
    Abstract_block* b2, 
    size_t          j2
) 
{
    typedef Component::Parameter P;

    assert(j1 % 2 == 0);
    assert(j1 == j2-1);

    size_t i = j1 / 2;
    size_t j = j1;

    b2->detach_scale_deps(j, i);

    for (size_t i = 0; i < 3; i++)
    {
        size_t j = i*2;

        P* b1_a_pos_ji = 0;
        P* b2_a_neg_ji = &(b2->_ctr_attach_pts_p[ j2 ][ i ]);

        P* b2_p_i = &(b2->_posn_p[ i ]);

        if (j == j1)
        {
            b1_a_pos_ji = &(b1->_ctr_attach_pts_p[ j1 ][ i ]);

            b1_a_pos_ji->remove_scale_dep(P::Pos_dependent(b2_a_neg_ji));
            b2_a_neg_ji->remove_scale_dep(P::Pos_dependent(b1_a_pos_ji));
        }
        else
        {
            b1_a_pos_ji = &(b1->_var_attach_pts_p[ j1 ][ i ]);

            b1_a_pos_ji->remove_scale_dep(P::Pos_dependent(b2_p_i));
            b2_a_neg_ji->remove_scale_dep(P::Pos_dependent(b1_a_pos_ji));

            for (size_t k = 0; k < 6; k += 2)
            {
                P* b2_a_pos_ki = &(b2->_ctr_attach_pts_p[ k+0 ][ i ]);
                P* b2_a_neg_ki = &(b2->_ctr_attach_pts_p[ k+1 ][ i ]);

                b1_a_pos_ji->remove_scale_dep(P::Pos_dependent(b2_a_pos_ki));
                b1_a_pos_ji->remove_scale_dep(P::Pos_dependent(b2_a_neg_ki));
            }
        }
    }
}

/** Detaches block @em b2 to block @em b1. */
void Object::ctr_var_detach_scale_deps
(
    Abstract_block* b1, 
    size_t          j1, 
    Abstract_block* b2, 
    size_t          j2
) 
{
    typedef Component::Parameter P;

    assert(j1 % 2 == 0);
    assert(j1 == j2-1);

    size_t i = j1 / 2;
    size_t j = j1;

    b2->detach_scale_deps(j, i);

    for (size_t i = 0; i < 3; i++)
    {
        size_t j = i*2;

        P* b1_a_pos_ji = &(b1->_ctr_attach_pts_p[ j1 ][ i ]);
        P* b2_a_neg_ji = 0;

        P* b1_p_i = &(b1->_posn_p[ i ]);

        if (j == j1)
        {
            b2_a_neg_ji = &(b2->_ctr_attach_pts_p[ j2 ][ i ]);

            b1_a_pos_ji->remove_scale_dep(P::Pos_dependent(b2_a_neg_ji));
            b2_a_neg_ji->remove_scale_dep(P::Pos_dependent(b1_a_pos_ji));
        }
        else
        {
            b2_a_neg_ji = &(b2->_var_attach_pts_p[ j2 ][ i ]);

            b2_a_neg_ji->remove_scale_dep(P::Pos_dependent(b1_p_i));
            b1_a_pos_ji->remove_scale_dep(P::Pos_dependent(b2_a_neg_ji));

            for (size_t k = 0; k < 6; k += 2)
            {
                P* b1_a_pos_ki = &(b2->_ctr_attach_pts_p[ k+0 ][ i ]);
                P* b1_a_neg_ki = &(b2->_ctr_attach_pts_p[ k+1 ][ i ]);

                b2_a_neg_ji->remove_scale_dep(P::Pos_dependent(b1_a_pos_ki));
                b2_a_neg_ji->remove_scale_dep(P::Pos_dependent(b1_a_neg_ki));
            }
        }
    }
}

void Object::assert_attachment_consistency
(
    const Object*    obj, 
    const Component* c
)
{
    const Abstract_block* b;

    if ((b = dynamic_cast<const Abstract_block*>(c)))
    {
        assert_attachment_consistency(obj, b);
    }
    else
    {
        abort();
    }
}

void Object::assert_attachment_consistency
(
    const Object*         obj, 
    const Abstract_block* b
)
{
    using namespace jwsc;

    const size_t y = 1;
    const size_t neg_h = 2*y + 1;

    const Component* c = b;
    assert(obj->lowest_component() == c);

    Vector_d* blk_a_neg_h = 0;
    b->ctr_attach_pt_v(&blk_a_neg_h, neg_h);

    Vector_d* obj_p = 0;
    obj->posn_v(&obj_p);

    for (size_t i = 0; i < 3; i++)
    {
        assert(fabs(blk_a_neg_h->elts[ i ] - obj_p->elts[ i ]) < 1.0e-2);
    }

    free_vector_d(blk_a_neg_h);
    free_vector_d(obj_p);
}

void Object::assert_attachment_consistency
(
    const Component* c1, 
    const Component* c2
)
{
    const Abstract_block* b1;
    const Abstract_block* b2;

    if ((b1 = dynamic_cast<const Abstract_block*>(c1)) &&
        (b2 = dynamic_cast<const Abstract_block*>(c2)))
    {
        assert_attachment_consistency(b1, b2);
    }
    else
    {
        abort();
    }
}

void Object::assert_attachment_consistency
(
    const Abstract_block* b1, 
    const Abstract_block* b2
)
{
    using namespace jwsc;

    const Component* c1 = b1;
    const Component* c2 = b2;

    size_t i1, i2;
    Component::Attach_type t1, t2;

    for (size_t i = 0; i < 6; i += 2)
    {
        if (b1->attachment(i) == c2)
        {
            i1 = i+0;
            i2 = i+1;
            t1 = b1->attach_pt_type(i1);
            t2 = b2->attach_pt_type(i2);
            assert(b2->attachment(i2) == c1);
            break;
        }
        if (b2->attachment(i) == c1)
        {
            i1 = i+1;
            i2 = i+0;
            t1 = b1->attach_pt_type(i1);
            t2 = b2->attach_pt_type(i2);
            assert(b1->attachment(i1) == c2);
            break;
        }
    }

    Vector_d* a_pos_i = 0;
    Vector_d* a_neg_i = 0;

    b1->attach_pt_v(&a_pos_i, i1, t1);
    b2->attach_pt_v(&a_neg_i, i2, t2);

    for (size_t j = 0; j < 3; j++)
    {
        assert(fabs(a_pos_i->elts[ j ] - a_neg_i->elts[ j ]) < 1.0e-2);
    }

    free_vector_d(a_pos_i);
    free_vector_d(a_neg_i);
}

void Object::split_aligned_var_attach_pt
(
    Abstract_block* bs,
    Abstract_block* bi,
    uint32_t        bi_j
)
{
    using namespace jwsc;

    Vector_d* p  = 0;
    Vector_d* uv = 0;
    create_vector_d(&uv, 2);

    bs->var_attach_pt_v(&p, bi_j);
    transform_pt_from_world_to_object_coords(p);
    bs->transform_pt_from_object_to_component_coords(p);
    bs->transform_pt_from_block_to_face_coords(p, bi_j);

    uv->elts[0] = p->elts[0];
    uv->elts[1] = p->elts[1];
    if (fabs(p->elts[2]) > 1.0e-6)
    {
        std::cerr << __FILE__ << ":" << __LINE__ << " "
                  << p->elts[2] << std::endl;
        abort();
    }

    bi->set_var_attach_pt_values(uv, bi_j);

    free_vector_d(p);
    free_vector_d(uv);
}

void Object::split_copy_var_attach_pts
(
    Abstract_block* bs,
    Abstract_block* b1,
    uint32_t        b1_j,
    Abstract_block* b2,
    uint32_t        b2_j
)
{
    if (bs->has_var_attach_pt(b1_j))
    {
        split_aligned_var_attach_pt(bs, b2, b1_j);
    }
    if (bs->has_var_attach_pt(b2_j))
    {
        split_aligned_var_attach_pt(bs, b1, b2_j);
    }
}

void Object::split_block_posn
(
    jwsc::Vector_d** p1_out,
    jwsc::Vector_d** p2_out,
    Abstract_block*  bs,
    Abstract_block*  b1,
    uint32_t         b1_j,
    Abstract_block*  b2,
    uint32_t         b2_j
)
{
    using namespace jwsc;

    Vector_d* p1 = 0;
    Vector_d* p2 = 0;
    Vector_d* q  = 0;

    bs->ctr_attach_pt_v(&p1, b2_j);
    bs->ctr_attach_pt_v(&p2, b1_j);

    double r1 = b1->size(b1_j/2) / (b1->size(b1_j/2)+b2->size(b1_j/2));
    subtract_vectors_d(&q, p2, p1);
    multiply_vector_by_scalar_d(&q, q, r1*0.5);
    add_vectors_d(p1_out, p1, q);

    double r2 = 1.0 - r1;
    subtract_vectors_d(&q, p1, p2);
    multiply_vector_by_scalar_d(&q, q, r2*0.5);
    add_vectors_d(p2_out, p2, q);

    free_vector_d(q);
    free_vector_d(p1);
    free_vector_d(p2);
}

bool Object::swap_split_blocks
(
    Component*                   as,
    uint32_t                     as_k, 
    Abstract_block*              bs, 
    Abstract_block*              b1, 
    uint32_t                     b1_j, 
    const std::vector<uint32_t>& b1_a,
    Abstract_block*              b2, 
    uint32_t                     b2_j,
    const std::vector<uint32_t>& b2_a
)
{
    using namespace jwsc;

    if (as_k == b2_j)
    {
        return true;
    }
    if (as_k == b1_j)
    {
        return false;
    }

    uint32_t bs_k = (as_k % 2 == 0) ? as_k+1 : as_k-1;
    if (b2_a[ bs_k ])
    {
        return true;
    }
    return false;
}

void Object::split
(
    Component*             cs,
    Component*             c1, 
    uint32_t               c1_j, 
    Component::Attach_type c1_t,
    std::vector<uint32_t>  c1_a,
    Component*             c2, 
    uint32_t               c2_j,
    Component::Attach_type c2_t,
    std::vector<uint32_t>  c2_a
)
{
    Abstract_block* bs;
    Abstract_block* b1;
    Abstract_block* b2;

    if ((bs = dynamic_cast<Abstract_block*>(cs)) &&
        (b1 = dynamic_cast<Abstract_block*>(c1)) &&
        (b2 = dynamic_cast<Abstract_block*>(c2)))
    {
        split(bs, b1, c1_j, c1_t, c1_a, b2, c2_j, c2_t, c2_a);
    }
    else
    {
        abort();
    }
}

/** 
 * The components @em b1 and @em b2 must be in world coords. The merging
 * component @em bm must not be yet in the object, and must be in block
 * coordinates (centered at 0,0,0). The merged component is then the last index
 * in the object's component set.
 *
 * The split blocks are setup such that @em b2 conditions on @em b1.
 *
 * The size along b1_j/b2_j in @em bs must be the same as the sum of that in @em
 * b1 and @em b2.
 *
 * @warning Modifies the sizes of @em b1 and @em b2.
 */
void Object::split
(
    Abstract_block*        bs,
    Abstract_block*        b1, 
    uint32_t               b1_j, 
    Component::Attach_type b1_t,
    std::vector<uint32_t>  b1_a,
    Abstract_block*        b2, 
    uint32_t               b2_j,
    Component::Attach_type b2_t,
    std::vector<uint32_t>  b2_a
)
{
    using std::vector;
    using namespace jwsc;

    typedef Component::Attach_type Attach_t;
    static const Attach_t ctr_att = Component::CTR_ATTACH;
    static const Attach_t var_att = Component::VAR_ATTACH;

    assert(b1_j/2 == b2_j/2);
    assert(fabs(bs->size(b1_j/2)-(b1->size(b1_j/2)+b2->size(b1_j/2))) < 1.0e-8);

    Component* as = bs->conditional();
    uint32_t as_k;
    Attach_t as_t;
    if (as)
    {
        as_k = as->attachment(bs);
        as_t = as->has_var_attach_pt(as_k) ? var_att : ctr_att;
        if (swap_split_blocks(as, as_k, bs, b1, b1_j, b1_a, b2, b2_j, b2_a))
        {
            Abstract_block* b = b1; b1 = b2; b2 = b;
            uint32_t j = b1_j; b1_j = b2_j; b2_j = j;
            Attach_t b_t = b1_t; b1_t = b2_t; b2_t = b_t;
            vector<uint32_t> b_a = b1_a; b1_a = b2_a; b2_a = b_a;
        }
    }

    vector<Component*> cs(6, (Component*)0);
    vector<Attach_t> cs_t(6, ctr_att);
    vector<Attach_t> bs_t(6, ctr_att);
    Vector_d* b1_p = 0;
    Vector_d* b2_p = 0;

    split_copy_var_attach_pts(bs, b1, b1_j, b2, b2_j);
    split_block_posn(&b1_p, &b2_p, bs, b1, b1_j, b2, b2_j);

    for (uint32_t bs_k = 0; bs_k < 6; bs_k++)
    {
        Component* c = bs->attachment(bs_k);
        if (!c || c == as) continue;

        if (bs->has_var_attach_pt(bs_k))
            bs_t[ bs_k ] = var_att;

        uint32_t cs_k = (bs_k % 2 == 0) ? bs_k+1 : bs_k-1;
        cs[ bs_k ] = c;
        if (cs[ bs_k ]->has_var_attach_pt(cs_k)) 
            cs_t[ bs_k ] = var_att;

        detach_deps_and_conds(bs, bs_k, c, cs_k);
    }

    if (as)
    {
        uint32_t bs_k = (as_k % 2 == 0) ? as_k+1 : as_k-1;
        Attach_t bs_t = bs->has_var_attach_pt(bs_k) ? var_att : ctr_att;
        detach(as, as_k, bs, bs_k);
        b1->rotate(azimuth(), 0, 1, 0);
        b1->translate(b1_p->elts[0], b1_p->elts[1], b1_p->elts[2]);
        if (as_k != b1_j && as_k != b2_j)
        {
            copy_non_aligned_var_attach_pt(b1, bs_k, as, as_k, as_t);
        }
        attach_deps_and_conds(as, as_k, as_t, b1, bs_k, bs_t);
        add_component(b1);
    }
    else
    {
        position_detachment(bs);
        remove_component(bs);
        b1->rotate(azimuth(), 0, 1, 0);
        b1->translate(b1_p->elts[0], b1_p->elts[1], b1_p->elts[2]);
        add_component(b1);
    }

    attach(b1, b1_j, b1_t, b2, b2_j, b2_t);

    for (uint32_t bs_k = 0; bs_k < 6; bs_k++)
    {
        Component* c = cs[ bs_k ];
        if (!c) continue;

        uint32_t c_k = (bs_k % 2 == 0) ? bs_k+1 : bs_k-1;
        Attach_t c_t = cs_t[ bs_k ];

        if (b1_a[ bs_k ])
        {
            copy_non_aligned_var_attach_pt(b1, bs_k, c, c_k, c_t);
            attach_deps_and_conds(b1, bs_k, bs_t[ bs_k ], c, c_k, c_t);
        }
        else
        {
            assert(b2_a[ bs_k ]);
            copy_non_aligned_var_attach_pt(b2, bs_k, c, c_k, c_t);
            attach_deps_and_conds(b2, bs_k, bs_t[ bs_k ], c, c_k, c_t);
        }
    }

    update_lowest_component();

    uint32_t N_c = num_components();
    uint32_t n1 = component(b1);
    swap_component_indices(n1, N_c-2);
    uint32_t n2 = component(b2);
    swap_component_indices(n2, N_c-1);

    free_vector_d(b1_p);
    free_vector_d(b2_p);
}

void Object::align_blocks_to_merge
(
    Abstract_block* bm,
    Abstract_block* b1,
    uint32_t        b1_j,
    Abstract_block* b2,
    uint32_t        b2_j
)
{
    for (uint32_t i = 0; i < 3; i++)
    {
        if (i != (b1_j/2))
        {
            b1->size_shift_deps(i, bm->size(i));
            b2->size_shift_deps(i, bm->size(i));
        }
    }

    if (b1->has_var_attach_pt(b1_j))
    {
        assert(!b2->has_var_attach_pt(b2_j));
        b1->var_attach_pt_shift_deps(b1_j, 0, 0);
        b1->var_attach_pt_shift_deps(b1_j, 1, 0);
    }
    else if (b2->has_var_attach_pt(b2_j))
    {
        b2->var_attach_pt_shift_deps(b2_j, 0, 0);
        b2->var_attach_pt_shift_deps(b2_j, 1, 0);
    }
}

void Object::merge_aligned_var_attach_pt
(
    Abstract_block* bm,
    Abstract_block* bi,
    uint32_t        bi_j
)
{
    using namespace jwsc;

    Vector_d* p  = 0;
    Vector_d* uv = 0;
    create_vector_d(&uv, 2);

    bi->var_attach_pt_v(&p, bi_j);
    transform_pt_from_world_to_object_coords(p);
    bi->transform_pt_from_object_to_component_coords(p);
    bi->transform_pt_from_block_to_face_coords(p, bi_j);

    uv->elts[0] = p->elts[0];
    uv->elts[1] = p->elts[1];
    if (fabs(p->elts[2]) > 1.0e-6)
    {
        std::cerr << __FILE__ << ":" << __LINE__ << " "
                  << p->elts[2] << std::endl;
        abort();
    }

    bm->set_var_attach_pt_values(uv, bi_j);

    free_vector_d(p);
    free_vector_d(uv);
}

void Object::copy_non_aligned_var_attach_pt
(
    Component*             c1,
    uint32_t               c1_j,
    Component*             c2,
    uint32_t               c2_j,
    Component::Attach_type c2_t
)
{
    using namespace jwsc;

    Vector_d* p = 0;
    if (c2_t == Component::VAR_ATTACH)
    {
        c1->ctr_attach_pt_v(&p, c1_j);
        c2->var_attach_pt_v(p, c2_j);
    }
    else
    {
        c2->ctr_attach_pt_v(&p, c2_j);
        c1->var_attach_pt_v(p, c1_j);
    }
    free_vector_d(p);
}

void Object::merge_copy_var_attach_pts
(
    Abstract_block* bm,
    Abstract_block* b1,
    uint32_t        b1_j,
    Abstract_block* b2,
    uint32_t        b2_j
)
{
    for (uint32_t j = 0; j < 6; j++)
    {
        Component* c1 = b1->attachment(j);
        Component* c2 = b2->attachment(j);

        if (j == b1_j || j == b2_j)
            continue;

        if (c1 && c2)
        {
            abort();
        }
    }

    if (b1->has_var_attach_pt(b2_j))
    {
        merge_aligned_var_attach_pt(bm, b1, b2_j);
    }

    if (b2->has_var_attach_pt(b1_j))
    {
        merge_aligned_var_attach_pt(bm, b2, b1_j);
    }
}

void Object::merged_block_posn
(
    jwsc::Vector_d** p_out,
    Abstract_block*  b1,
    uint32_t         b1_j,
    Abstract_block*  b2,
    uint32_t         b2_j
)
{
    using namespace jwsc;

    Vector_d* p1 = 0;
    Vector_d* p2 = 0;

    b1->ctr_attach_pt_v(&p1, b2_j);
    b2->ctr_attach_pt_v(&p2, b1_j);

    add_vectors_d(p_out, p1, p2);
    multiply_vector_by_scalar_d(p_out, *p_out, 0.5);

    free_vector_d(p1);
    free_vector_d(p2);
}

void Object::merge
(
    Component* cm,
    Component* c1, 
    uint32_t   c1_j, 
    Component* c2, 
    uint32_t   c2_j
)
{
    Abstract_block* bm;
    Abstract_block* b1;
    Abstract_block* b2;

    if ((bm = dynamic_cast<Abstract_block*>(cm)) &&
        (b1 = dynamic_cast<Abstract_block*>(c1)) &&
        (b2 = dynamic_cast<Abstract_block*>(c2)))
    {
        merge(bm, b1, c1_j, b2, c2_j);
    }
    else
    {
        abort();
    }
}

/** 
 * The components @em b1 and @em b2 must be in world coords. The merging
 * component @em bm must not be yet in the object, and must be in block
 * coordinates (centered at 0,0,0). The merged component is then the last index
 * in the object's component set.
 *
 * The size along b1_j/b2_j in @em bm must be the same as the sum of that in @em
 * b1 and @em b2.
 *
 * @warning Modifies the sizes of @em b1 and @em b2.
 */
void Object::merge
(
    Abstract_block* bm,
    Abstract_block* b1, 
    uint32_t        b1_j, 
    Abstract_block* b2, 
    uint32_t        b2_j
)
{
    using std::vector;
    using namespace jwsc;

    typedef Component::Attach_type Attach_t;
    static const Attach_t ctr_att = Component::CTR_ATTACH;
    static const Attach_t var_att = Component::VAR_ATTACH;

    assert(b1_j/2 == b2_j/2);
    assert(fabs(bm->size(b1_j/2)-(b1->size(b1_j/2)+b2->size(b1_j/2))) < 1.0e-8);

    if (b1->conditions(b2))
    {
        Abstract_block* b = b1; b1 = b2; b2 = b;
        uint32_t j = b1_j; b1_j = b2_j; b2_j = j;
    }
    assert(b2->conditions(b1));

    Vector_d* bm_p = 0;

    align_blocks_to_merge(bm, b1, b1_j, b2, b2_j);
    merge_copy_var_attach_pts(bm, b1, b1_j, b2, b2_j);
    merged_block_posn(&bm_p, b1, b1_j, b2, b2_j);

    vector<Component*> c2(6, (Component*)0);
    vector<Attach_t> c2_t(6, ctr_att);
    vector<Attach_t> b2_t(6, ctr_att);

    for (uint32_t b2_k = 0; b2_k < 6; b2_k++)
    {
        c2[ b2_k ] = b2->attachment(b2_k);
        if (c2[ b2_k ] == (Component*)b1) 
            c2[ b2_k ] = 0;

        if (c2[ b2_k ])
        {
            uint32_t c2_k = (b2_k % 2 == 0) ? b2_k+1 : b2_k-1;
            if (c2[ b2_k ]->has_var_attach_pt(c2_k)) 
                c2_t[ b2_k ] = var_att;
            if (b2->has_var_attach_pt(b2_k))
                b2_t[ b2_k ] = var_att;
            detach_deps_and_conds(b2, b2_k, c2[ b2_k ], c2_k);
        }
    }

    detach(b1, b1_j, b2, b2_j);
    vector<Component*> c1(6, (Component*)0);
    vector<Attach_t> c1_t(6, ctr_att);
    vector<Attach_t> b1_t(6, ctr_att);
    Component* a1 = b1->conditional();

    for (uint32_t b1_k = 0; b1_k < 6; b1_k++)
    {
        c1[ b1_k ] = b1->attachment(b1_k);
        if (c1[ b1_k ] == a1)
            c1[ b1_k ] = 0;

        if (c1[ b1_k ])
        {
            uint32_t c1_k = (b1_k % 2 == 0) ? b1_k+1 : b1_k-1;
            if (c1[ b1_k ]->has_var_attach_pt(c1_k)) 
                c1_t[ b1_k ] = var_att;
            if (b1->has_var_attach_pt(b1_k))
                b1_t[ b1_k ] = var_att;
            detach_deps_and_conds(b1, b1_k, c1[ b1_k ], c1_k);
        }
    }

    if (a1)
    {
        uint32_t b1_k = b1->attachment(a1);
        Attach_t b1_t = b1->has_var_attach_pt(b1_k) ? var_att : ctr_att;
        uint32_t a1_k = (b1_k % 2 == 0) ? b1_k+1 : b1_k-1;
        Attach_t a1_t = a1->has_var_attach_pt(a1_k) ? var_att : ctr_att;
        detach(a1, a1_k, b1, b1_k);
        bm->rotate(azimuth(), 0, 1, 0);
        bm->translate(bm_p->elts[0], bm_p->elts[1], bm_p->elts[2]);
        if (a1_k != b1_j && a1_k != b2_j)
        {
            copy_non_aligned_var_attach_pt(bm, b1_k, a1, a1_k, a1_t);
        }
        attach_deps_and_conds(a1, a1_k, a1_t, bm, b1_k, b1_t);
        add_component(bm);
    }
    else
    {
        position_detachment(b1);
        remove_component(b1);
        bm->rotate(azimuth(), 0, 1, 0);
        bm->translate(bm_p->elts[0], bm_p->elts[1], bm_p->elts[2]);
        add_component(bm);
    }

    for (uint32_t bm_k = 0; bm_k < 6; bm_k++)
    {
        if (c1[ bm_k ])
        {
            uint32_t c1_k = (bm_k % 2 == 0) ? bm_k+1 : bm_k-1;
            copy_non_aligned_var_attach_pt(bm, bm_k, c1[ bm_k ], c1_k,
                    c1_t[ bm_k ]);
            attach_deps_and_conds(bm, bm_k, b1_t[ bm_k ], c1[ bm_k ], c1_k,
                    c1_t[ bm_k ]);
        }
        if (c2[ bm_k ])
        {
            uint32_t c2_k = (bm_k % 2 == 0) ? bm_k+1 : bm_k-1;
            copy_non_aligned_var_attach_pt(bm, bm_k, c2[ bm_k ], c2_k, 
                    c2_t[ bm_k ]);
            attach_deps_and_conds(bm, bm_k, b2_t[ bm_k ], c2[ bm_k ], c2_k,
                    c2_t[ bm_k ]);
        }
    }

    update_lowest_component();

    uint32_t N_c = num_components();
    uint32_t n = component(bm);
    swap_component_indices(n, N_c-1);

    free_vector_d(bm_p);
}


void Object::position_switched_block
(
    const Abstract_block*  b1, 
    Abstract_block*        b2
)
{
    using namespace jwsc;

    Object* obj = b1->object();
    assert(obj);

    Vector_d* b1_p = 0;
    Vector_d* b2_p = 0;
    b1->posn_v(&b1_p);
    b2->posn_v(&b2_p);

    double t_x = b2_p->elts[0];
    double t_y = b2_p->elts[1];
    double t_z = b2_p->elts[2];

    b2->translate(-t_x, -t_y, -t_z);
    b2->rotate(obj->azimuth(), 0, 1, 0);

    t_x = b1_p->elts[0];
    t_y = b1_p->elts[1];
    t_z = b1_p->elts[2];

    b2->translate(t_x, t_y, t_z);

    free_vector_d(b1_p);
    free_vector_d(b2_p);
}

/** 
 * The size of b2 must be the same as b1. The switched block maintains its
 * original ID.
 *
 * @warning The position of @em b2 is modified to match @em b1. The block @em b2
 * must not have any attachments.
 */
void Object::switch_block(Block_set* b1, Block* b2)
{
    for (uint32_t i = 0; i < 3; i++)
    {
        assert(fabs(b1->size(i) - b2->size(i)) < 1.0e-8);
    }
    position_switched_block(b1, b2);

    for (uint32_t j = 0; j < 6; j++)
    {
        Abstract_block* b;
        if ((b = dynamic_cast<Abstract_block*>(b1->attachment(j))))
        {
            b1->remove_inner_ctr_attach_conds(j, b->_size_p);
        }
        if (b1->has_var_attach_pt(j))
        {
            b2->enable_var_attach_pt(j);
        }
    }

    Block* b = b2->clone();
    b2->Abstract_block::operator=(*b1);
    b2->Model::operator=(*b);
    delete b;

    std::vector<Model*> new_models(1, this);
    std::vector<const Model*> old_models(1, this);
    uint32_t b1_i = component(b1);

    for (uint32_t i = 0; i < num_components(); i++)
    {
        old_models.push_back(component(i));
        if (i == b1_i)
        {
            _components[ i ] = b2;
        }
        new_models.push_back(component(i));
    }

    for (uint32_t i = 0; i < new_models.size(); i++)
    {
        new_models[ i ]->update_pointers(new_models, old_models);
    }

    b1->object(0);
}

/** 
 * The size of b2 must be the same as b1. The switched block maintains its
 * original ID.
 *
 * @warning The position of @em b2 is modified to match @em b1. The block @em b2
 * must not have any attachments.
 */
void Object::switch_block(Block* b1, Block_set* b2)
{
    for (uint32_t i = 0; i < 3; i++)
    {
        assert(fabs(b1->size(i) - b2->size(i)) < 1.0e-8);
    }
    position_switched_block(b1, b2);

    Block_set* b = b2->clone();
    b2->Abstract_block::operator=(*b1);
    b2->Model::operator=(*b);
    delete b;

    for (uint32_t j = 0; j < 6; j++)
    {
        Abstract_block* b;
        if ((b = dynamic_cast<Abstract_block*>(b1->attachment(j))))
        {
            b2->add_inner_ctr_attach_conds(j, b->_size_p);
        }
        if (b1->has_var_attach_pt(j))
        {
            b2->enable_var_attach_pt(j);
        }
    }

    std::vector<Model*> new_models(1, this);
    std::vector<const Model*> old_models(1, this);
    uint32_t b1_i = component(b1);

    for (uint32_t i = 0; i < num_components(); i++)
    {
        old_models.push_back(component(i));
        if (i == b1_i)
        {
            _components[ i ] = b2;
        }
        new_models.push_back(component(i));
    }

    for (uint32_t i = 0; i < new_models.size(); i++)
    {
        new_models[ i ]->update_pointers(new_models, old_models);
    }

    b1->object(0);
}

/** 
 * The size of b2 must be the same as b1. The switched block maintains its
 * original ID.
 *
 * @warning The position of @em b2 is modified to match @em b1. The block @em b2
 * must not have any attachments.
 */
void Object::switch_block(Block_set* b1, Block_set* b2)
{
    for (uint32_t i = 0; i < 3; i++)
    {
        assert(fabs(b1->size(i) - b2->size(i)) < 1.0e-8);
    }
    position_switched_block(b1, b2);

    Block_set* b = b2->clone();
    b2->Abstract_block::operator=(*b1);
    b2->Model::operator=(*b);
    delete b;

    for (uint32_t j = 0; j < 6; j++)
    {
        if (b1->has_var_attach_pt(j))
        {
            b2->enable_var_attach_pt(j);
        }
    }

    std::vector<Model*> new_models(1, this);
    std::vector<const Model*> old_models(1, this);
    uint32_t b1_i = component(b1);

    for (uint32_t i = 0; i < num_components(); i++)
    {
        old_models.push_back(component(i));
        if (i == b1_i)
        {
            _components[ i ] = b2;
        }
        new_models.push_back(component(i));
    }

    for (uint32_t i = 0; i < new_models.size(); i++)
    {
        new_models[ i ]->update_pointers(new_models, old_models);
    }

    b1->object(0);
}

/* ========================================================================== */
