/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Generic 3D object model.
 *
 * Currently composed of a collection of 3D Component.
 */


#ifndef OBJECT_H
#define OBJECT_H


#include <config.h>

#include <iostream>
#include <vector>

#include <jwsc/vector/vector.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/graphics/renderable.h>
#include <jwsc++/graphics/polygon.h>

#include "model.h"
#include "view.h"
#include "component.h"
#include "block.h"
#include "block_set.h"


/** 
 * @class Object
 *
 * @brief Model for a generic 3D object composed of Component.
 */
class Object : public jwscxx::graphics::Renderable,
               public jwscxx::graphics::Wire_renderable,
               public jwscxx::graphics::Wire_occlude_renderable,
               public jwscxx::graphics::Solid_renderable,
               public Model
{
    public:


/**
 * @class Object::Context
 *
 * @brief Context for Object model parameters.
 */
class Context : public Model::Context
{
    public:

        /** @brief Constructs a new default Object::Context. */
        Context();

        /** @brief Constructs a new Object::Context from another. */
        Context(const Object::Context& m);

        /** @brief Deletes this Object::Context. */
        virtual ~Context();

        /** @brief Copies a Object::Context into this one. */
        Object::Context& operator= (const Object::Context& ctx);

        /** @brief Clones this Object::Context. */
        Object::Context* clone() const;

        const Model::Parameter::Context& azimuth_c() const;

        void azimuth_c(const Model::Parameter::Context& azimuth);

        const Model::Parameter::Context& posn_c(size_t i) const;

        void posn_c(size_t i, const Model::Parameter::Context& posn);

    protected:

        /** @brief Rotation angle of this Object about its center. */
        Model::Parameter::Context _azimuth_c;

        /** @brief Position of this Object. */
        std::vector<Model::Parameter::Context> _posn_c;
};


    // Object
    public:

        /**
         * @name Standard
         *
         * Standard class members.
         * @{
         */

        /** @brief Constructs a new Object. */
        Object
        (
            const Object::Context*    obj_ctx, 
            const Block::Context*     block_ctx,
            const Block_set::Context* block_set_ctx
        );

        /** @brief Constructs a specific Object. */
        Object
        (
            double                    azimuth,
            double                    x,
            double                    y,
            double                    z,
            const Object::Context*    obj_ctx, 
            const Block::Context*     block_ctx,
            const Block_set::Context* block_set_ctx
        );

        /** @brief Constructs an Object by copying another. */
        Object(const Object& obj);

        /** @brief Deletes this Object. */
        virtual ~Object();

        /** @brief Copies an Object into this Object. */
        Object& operator= (const Object& obj);

        /** @brief Clones this Object. */
        Object* clone() const;

        /** @brief Reads this Object from an input stream. */
        void read(std::istream& in) 
            throw (jwscxx::base::IO_error, jwscxx::base::Arg_error);

        /** @brief Writes this Object to an output stream. */
        void write(std::ostream& out) const throw (jwscxx::base::IO_error);

        /** @brief Returns the model context. */
        const Object::Context* context() const;

        /*@}*/


        /**
         * @name Variable Parameters
         *
         * Variable model parameter members.
         * @{
         */

        /** @brief Returns the number of parameters. */
        uint32_t num_var_params() const;

        /** @brief Returns an instance parameter name. */
        const char* get_var_param_name(uint32_t d) const;

        /** @brief Returns an instance parameter name. */
        std::vector<const char*> get_var_param_names() const;

        /** @brief Returns an instance parameter value. */
        double get_var_param_value(uint32_t d) const;

        /** @brief Gets the instance parameter values. */
        void get_var_param_values(jwsc::Vector_d** p_out) const;

        /** @brief Sets the instance parameter values. */
        void set_var_param_values(const jwsc::Vector_d* p)
            throw (jwscxx::base::Arg_error);

        /** @brief Adds to the instance parameter values. */
        void add_var_param_values(const jwsc::Vector_d* p)
            throw (jwscxx::base::Arg_error);

        /** @brief Adds to the instance parameter values. */
        void add_var_param_values_shift_component_deps(const jwsc::Vector_d* p)
            throw (jwscxx::base::Arg_error);

        /** @brief Adds to the instance parameter values. */
        void add_var_param_values_scale_component_deps(const jwsc::Vector_d* p)
            throw (jwscxx::base::Arg_error);

        /** @brief Returns the parameter dynamics gradient steps as a point. */
        void get_var_param_values_dynamics_grad_eta(jwsc::Vector_d** p_out) 
            const;

        /** @brief Returns the parameter dynamics gradient steps as a point. */
        void get_var_param_values_dynamics_delta_t(jwsc::Vector_d** p_out) 
            const;

        /** @brief Prints variable parameter values. */
        void print_var_param_values(uint32_t iteration, std::ostream& out) 
            const;

        /** 
         * @brief Returns the category parameter statistics over an instance
         * parameter value.
         */
        void get_var_param_stats
        (
            double*  mean_out,
            double*  stdev_out,
            uint32_t d
        ) 
        const;

        /** 
         * @brief Returns the category parameter statistics over the instance
         * parameter values.
         */
        void get_var_param_stats
        (
            jwsc::Vector_d** means_out,
            jwsc::Vector_d** stdevs_out
        ) 
        const;

        /**
         * @brief Sets the category parameter statistics over the instance
         * parameter values.
         */
        void set_var_param_stats
        (
            const jwsc::Vector_d* means,
            const jwsc::Vector_d* stdevs
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Returns the parameter dynamics gradient steps. */
        void get_var_param_stats_dynamics_grad_eta
        (
            jwsc::Vector_d** means_out,
            jwsc::Vector_d** stdevs_out
        )
        const;

        /** @brief Returns the parameter dynamics gradient steps as a point. */
        void get_var_param_stats_dynamics_delta_t
        (
            jwsc::Vector_d** means_out,
            jwsc::Vector_d** stdevs_out
        )
        const;

        /** @brief Prints variable parameter statistics. */
        void print_var_param_stats(uint32_t iteration, std::ostream& out)
            const;

        /*@}*/


        /**
         * @name azimuth
         *
         * Rotation angle parameter members.
         * @{
         */

        /** @brief Returns the rotation angle of this Object. */
        const Object::Parameter* azimuth_p() const;

        /** @brief Returns the rotation angle of this Object. */
        double azimuth() const;

        /** @brief Sets the rotation angle of this Object. */
        void azimuth(double value);

        /** @brief Tests a rotation angle for this Object. */
        void test_azimuth(double value) const throw (jwscxx::base::Arg_error);

        /*@}*/


        /**
         * @name posn
         *
         * Position parameter members.
         * @{
         */

        /** @brief Returns a coord of the position of this Object. */
        const Component::Parameter* posn_p(size_t i) const;

        /** @brief Returns the position of this Object. */
        void posn_v(jwsc::Vector_d** p_out) const;

        /** @brief Returns coord of the position of this Object. */
        double posn(size_t i) const;

        /** @brief Sets a coord of the position of this Object. */
        void posn_shift_deps_v(const jwsc::Vector_d* p);

        void add_posn_shift_deps_v(const jwsc::Vector_d* p);

        /** @brief Sets a coord of the position of this Object. */
        void posn_shift_deps(size_t i, double posn);

        /** @brief Sets a coord of the position of this Object. */
        void posn_scale_deps(size_t i, double posn);

        /** @brief Tests a coord of the position for this Object. */
        void test_posn_shift_deps(size_t i, double posn)
            throw (jwscxx::base::Arg_error);

        /** @brief Tests a coord of the position for this Object. */
        void test_posn_scale_deps(size_t i, double posn)
            throw (jwscxx::base::Arg_error);

        /** @brief Asserts the consitency among components in this Object. */
        void assert_component_consistency();

        /** @brief Ensures the consistency among components in this Object. */
        void ensure_component_consistency();

        /*@}*/


        /**
         * @name Transform
         *
         * Transformation of internal representation members.
         * @{
         */

        /** 
         * @brief Transforms a point from this Object to world coordinates.
         */
        void transform_pt_from_object_to_world_coords(jwsc::Vector_d* p) const;

        /** 
         * @brief Transforms a point from world to this Object coordinates.
         */
        void transform_pt_from_world_to_object_coords(jwsc::Vector_d* p) const;

        void transform_from_world_to_object_coords();
        
        void transform_from_object_to_world_coords();

        /** 
         * @brief Returns whether this Object has been transformed into in
         * object coordinates.
         */
        bool in_object_coords() const;

        /** 
         * @brief Swaps the internal object width (x) and length (z) coordinates
         * of this Object and all its components.
         */
        void test_swap_object_coords() throw (jwscxx::base::Arg_error);

        /** 
         * @brief Swaps the internal object width (x) and length (z) coordinates
         * of this Object and all its components.
         */
        void swap_object_coords(bool pos_swap);

        /*@}*/


        /**
         * @name Likelihood
         *
         * Instance parameter likelihood members.
         * @{
         */

        /** @brief Calculates the log likelihood of the instance parameters. */
        double calc_var_param_log_likelihood() const;

        /** 
         * @brief Generates instance parameter values from the likelihood.
         */
        void sample_var_param_likelihood(jwsc::Vector_d** p_out);

        /*@}*/


        /**
         * @name Prior
         *
         * Category parameter prior probability members.
         * @{
         */

        /**
         * @brief Calculates the log prior probability of category parameters.
         */
        double calc_var_param_prior_log_prob() const;

        /** @brief Generates category parameters from the prior. */
        void sample_var_param_prior
        (
            jwsc::Vector_d** means_out,
            jwsc::Vector_d** stdevs_out
        ) 
        const;

        /*@}*/


        /**
         * @name Category Proposal
         *
         * Category proposal probability distribution members.
         * @{
         */

        void sample_var_param_cat_proposal
        (
            jwsc::Vector_d** means_out,
            jwsc::Vector_d** stdevs_out,
            const jwsc::Vector_d* means_in
        ) 
        const;

        /*@}*/


        /**
         * @name Proposal
         *
         * Proposal probability distribution members.
         * @{
         */

        /** 
         * @brief Calculates the proposal log probability of instance
         * parameters.
         */
        double calc_var_param_proposal_log_prob() const;

        /** 
         * @brief Calculates the Gaussian proposal log probability of instance
         * parameters.
         */
        double calc_var_param_gauss_proposal_log_prob() const;

        /** 
         * @brief Generates instance parameter values from the proposal
         * distribution.
         */
        void sample_var_param_proposal(jwsc::Vector_d** p_out) const;

        /**
         * @brief Generates instance parameter values from the Gaussian proposal
         * distribution.
         */
        void sample_var_param_gauss_proposal
        (
            jwsc::Vector_d**      p_out,
            const jwsc::Vector_d* means
        ) 
        const;

        /** 
         * @brief Returns the instance parameter Gaussian proposal sigmas as a
         * point.
         */
        void get_var_param_gauss_proposal_sigma(jwsc::Vector_d** p_out) const;

        /*@}*/


        /**
         * @name Pointers
         *
         * Pointer management members.
         * @{
         */

        /** 
         * @brief Serializes Model and Parameter pointers in this Object using
         * indices in the @em models collection.
         */
        void serialize(const std::vector<Model*>& models);

        /** 
         * @brief Deserializes Model and Parameter indices in this Object using
         * the @em models collection of pointers.
         */
        void deserialize
        (
            const std::vector<Model*>& models
        );

        /** 
         * @brief Updates the Model and Parameter pointers in this Object from
         * one set of models to another.
         */
        void update_pointers
        (
            const std::vector<Model*>&       new_models,
            const std::vector<const Model*>& old_models
        );

        /*@}*/


        /**
         * @name Graphical Representation
         *
         * Internal graphical representation methods.
         * @{
         */

        /** 
         * @brief Updates the graphical representation of this Object with its
         * current parameter values.
         *
         * This is a callback member function used by the Model::Parameter when
         * a value changes.
         */
        void update_graphical_rep();

        /*@}*/


        /**
         * @name Graphics
         *
         * Graphics rendering members.
         * @{
         */

        /** @brief Renders this Object with the GL. */
        void render() const;
        
        /** @brief Renders this Object using the component IDs as color. */
        void render_with_ids(const View* view) const;

        /** @brief Renders this Object with the GL as a wire-frame. */
        void wire_render() const;

        /** 
         * @brief Renders this Object with the GL as a wire-frame using its
         * IDs as color.
         */
        void wire_render_with_ids(const View* view) const;

        /** 
         * @brief Renders this Object with the GL as an occluded wire-frame 
         * into the depth buffer, to hide unseen lines.
         */
        void wire_occlude_render() const;

        /** @brief Renders this Object with the GL as a solid object. */
        void solid_render() const;

        /** 
         * @brief Renders this Object with the GL as a wire-frame using its
         * IDs as color.
         */
        void solid_render_with_ids(const View* view) const;

        /** @brief Projects this Object with the GL. */
        const std::vector<jwscxx::graphics::Polygon_d>& project();

        /*@}*/


        /**
         * @name Components
         *
         * Component management members.
         * @{
         */

        /** @brief Returns the number of components in this Object. */
        uint32_t num_components() const;

        /** @brief Returns the Object component at a particular index. */
        Component* component(size_t i) const throw (jwscxx::base::Arg_error);

        /** @brief Returns the Object component index. */
        size_t component(const Component* c) const 
            throw (jwscxx::base::Arg_error);

        /** @brief Returns the components in this Object. */
        const std::vector<Component*>& components() const;

        /** 
         * @brief Updates the position of this object and its lowest Component.
         */
        void update_lowest_component();

        /** @brief Swaps the components at the specified indices. */
        void swap_component_indices(uint32_t i, uint32_t j);

        /** @brief Returns the lowest component in this Object. */
        Component* lowest_component() const;

        static uint32_t sample_component_index
        (
            const std::vector<uint32_t>& indices
        )
        throw (jwscxx::base::Arg_error);

        /** @brief Selects a random index of a component. */
        uint32_t select_random_component() const 
            throw (jwscxx::base::Arg_error);

        /** @brief Selects a random index of a block in a Component set. */
        uint32_t select_random_block() const 
            throw (jwscxx::base::Arg_error);

        /** @brief Selects a random index of a block pair in a Component set. */
        uint32_t select_random_pair() const 
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Selects a random index of a block quartet in a Component set.
         */
        uint32_t select_random_quartet() const 
            throw (jwscxx::base::Arg_error);

        /*@}*/


        /**
         * @name Attachment
         *
         * Component attachment members.
         * @{
         */

        /** @brief Attaches a Component to this Object. */
        void attach(Component* c, Component::Attach_type t);

        /** @brief Attaches a Component to another already in this Object. */
        void attach
        (
            Component*             c1, 
            size_t                 i1, 
            Component::Attach_type t1, 
            Component*             c2, 
            size_t                 i2, 
            Component::Attach_type t2
        );

        /*@}*/


        /**
         * @name Detachment
         *
         * Component detachment members.
         * @{
         */

        /** @brief Detaches a Component from this Object. */
        Component* detach(Component* c);

        /** @brief Detaches a Component from another and from this Object. */
        Component* detach
        (
            Component* c1, 
            size_t     i1, 
            Component* c2, 
            size_t     i2
        );

        /*@}*/


        /**
         * @name Split
         *
         * Splits a component in two.
         * @{
         */

        void split
        (
            Component*             cs,
            Component*             c1, 
            uint32_t               c1_j, 
            Component::Attach_type c1_t,
            std::vector<uint32_t>  c1_a,
            Component*             c2, 
            uint32_t               c2_j,
            Component::Attach_type c2_t,
            std::vector<uint32_t>  c2_a
        );

        void split
        (
            Abstract_block*        bs,
            Abstract_block*        b1, 
            uint32_t               b1_j, 
            Component::Attach_type b1_t,
            std::vector<uint32_t>  b1_a,
            Abstract_block*        b2, 
            uint32_t               b2_j,
            Component::Attach_type b2_t,
            std::vector<uint32_t>  b2_a
        );

        /*@}*/


        /**
         * @name Merge
         *
         * Merges two attachments.
         * @{
         */

        void merge
        (
            Component* cm,
            Component* c1, 
            uint32_t   c1_j, 
            Component* c2, 
            uint32_t   c2_j
        );

        void merge
        (
            Abstract_block* bm,
            Abstract_block* b1, 
            uint32_t        b1_j, 
            Abstract_block* b2, 
            uint32_t        b2_j
        );

        /*@}*/


        /**
         * @name Switch
         *
         * Switches out one abstract block with another.
         * @{
         */

        /** @brief Switches the abstract block b1 with b2. */
        void switch_block(Block_set* b1, Block* b2);

        /** @brief Switches the abstract block b1 with b2. */
        void switch_block(Block* b1, Block_set* b2);

        /** @brief Switches the abstract block b1 with b2. */
        void switch_block(Block_set* b1, Block_set* b2);

        /*@}*/


        /**
         * @name Attachment Consistency
         *
         * Asserts the consistency between Object and its attached components.
         * @{
         */

        static void assert_attachment_consistency
        (
            const Object*    obj, 
            const Component* c
        );

        static void assert_attachment_consistency
        (
            const Object*         obj, 
            const Abstract_block* b
        );

        static void assert_attachment_consistency
        (
            const Component* c1, 
            const Component* c2
        );

        static void assert_attachment_consistency
        (
            const Abstract_block* b1, 
            const Abstract_block* b2
        );

        /*@}*/


    private:

        /**
         * @name azimuth
         *
         * Rotation angle parameter members.
         * @{
         */

        /** 
         * @brief Sets the rotation angle of this Object in object coordinates.
         */
        void azimuth_in_obj_coords(double value);

        /** 
         * @brief Tests a rotation angle for this Object in object coordinates.
         */
        void test_azimuth_in_obj_coords(double value) const 
            throw (jwscxx::base::Arg_error);

        /*@}*/


        /**
         * @name posn
         *
         * Position parameters methods.
         * @{
         */

        /** @brief Sets the position of this Object. */
        void posn_v(const jwsc::Vector_d* p);

        void position_object_below_component(const Component* c);

        static void position_attachment
        (
            const Object*          obj, 
            Component*             c,
            Component::Attach_type t
        );

        static void position_attachment
        (
            const Object*          obj, 
            Abstract_block*        b,
            Component::Attach_type t
        );

        static void position_attachment
        (
            const Component*       c1, 
            size_t                 i1,
            Component::Attach_type t1,
            Component*             c2,
            size_t                 i2,
            Component::Attach_type t2
        );

        static void position_attachment
        (
            const Abstract_block*  b1, 
            size_t                 i1,
            Component::Attach_type t1,
            Abstract_block*        b2,
            size_t                 i2,
            Component::Attach_type t2
        );

        static void position_detachment(Component* c);

        /*@}*/


        /**
         * @name Components
         *
         * Component management members.
         * @{
         */

        /** @brief Returns the next sequential component id. */
        size_t next_component_id(bool* looped);

        /** @brief Returns the next unused sequential component id. */
        size_t next_unused_component_id();

        /** @brief Removes a component ID from the list of used ones. */
        void remove_used_component_id(size_t id);

        /** @brief Sets the lowest component in this Object. */
        void lowest_component(Component* c);

        /** @brief Adds a Component into this Object. */
        void add_component(Component* c);

        /** @brief Removes a Component from this Object. */
        void remove_component(Component* c);

        double intersection_ratio
        (
            const Component* c1, 
            const Component* c2
        ) 
        const;

        double intersection_ratio
        (
            const Abstract_block* b1, 
            const Abstract_block* b2
        ) 
        const;

        uint32_t gen_var_param_conditional_likelihood_sample
        (
            Component* c, 
            uint32_t   n
        );

        /*@}*/


        /**
         * @name Transform
         *
         * Transformation of internal representation members.
         * @{
         */

        /** 
         * @brief Recursively swaps the width and length coordinates of a 
         * component and all its attachments.
         */
        static void swap_component_coords
        (
            Component* c_j1,
            Component* c_i1,
            uint32_t   i1_ign
        )
        throw (jwscxx::base::Arg_error);

        /*@}*/

        /**
         * @name Attachment
         *
         * Creates attachment dependencies between model parameters.
         * @{
         */

        /** @brief Attaches a Component to an Object. */
        static void attach_deps(Object* obj, Component* c);

        /** @brief Attaches a block shift dependencies to an Object. */
        static void attach_shift_deps(Object* obj, Abstract_block* b);

        /** @brief Attaches a block scale dependencies to an Object. */
        static void attach_scale_deps(Object* obj, Abstract_block* b);

        /** @brief Attaches a Component to another. */
        static void attach_deps_and_conds
        (
            Component*             c1, 
            size_t                 i1,
            Component::Attach_type t1,
            Component*             c2,
            size_t                 i2,
            Component::Attach_type t2
        );

        /** @brief Attaches a block center shift dependencies to another. */
        static void attach_deps
        (
            Abstract_block*        b1, 
            size_t                 j1, 
            Component::Attach_type t1,
            Abstract_block*        b2, 
            size_t                 j2,
            Component::Attach_type t2
        );

        /** @brief Attaches a block center shift dependencies to another. */
        static void ctr_ctr_attach_shift_deps
        (
            Abstract_block* b1, 
            size_t          j1, 
            Abstract_block* b2, 
            size_t          j2
        );

        /** @brief Attaches a block center shift dependencies to another. */
        static void var_ctr_attach_shift_deps
        (
            Abstract_block* b1, 
            size_t          j1, 
            Abstract_block* b2, 
            size_t          j2
        );

        /** @brief Attaches a block center shift dependencies to another. */
        static void ctr_var_attach_shift_deps
        (
            Abstract_block* b1, 
            size_t          j1, 
            Abstract_block* b2, 
            size_t          j2
        );

        /** @brief Attaches a block center scale dependencies to another. */
        static void ctr_ctr_attach_scale_deps
        (
            Abstract_block* b1, 
            size_t          j1, 
            Abstract_block* b2, 
            size_t          j2
        );

        /** @brief Attaches a block center scale dependencies to another. */
        static void var_ctr_attach_scale_deps
        (
            Abstract_block* b1, 
            size_t          j1, 
            Abstract_block* b2, 
            size_t          j2
        );

        /** @brief Attaches a block center scale dependencies to another. */
        static void ctr_var_attach_scale_deps
        (
            Abstract_block* b1, 
            size_t          j1, 
            Abstract_block* b2, 
            size_t          j2
        );

        static void attach_conds
        (
            Abstract_block* b1, 
            size_t          j1, 
            Abstract_block* b2, 
            size_t          j2
        );

        static void swap_conds
        (
            Component* c1, 
            size_t     j1, 
            Component* c2, 
            size_t     j2
        );

        /*@}*/


        /**
         * @name Detachment
         *
         * Removes attachment dependencies between model parameters.
         * @{
         */

        /** @brief Detaches a Component from an Object. */
        static void detach_deps(Object* obj, Component* c);

        /** @brief Detaches a block shift dependencies from an Object. */
        static void detach_shift_deps(Object* obj, Abstract_block* b);

        /** @brief Detaches a block scale dependencies from an Object. */
        static void detach_scale_deps(Object* obj, Abstract_block* b);

        /** @brief Detaches a Component from another. */
        static void detach_deps_and_conds
        (
            Component* c1, 
            size_t     i1, 
            Component* c2, 
            size_t     i2
        );

        /** @brief Detaches a block center shift dependencies from another. */
        static void detach_deps
        (
            Abstract_block* b1, 
            size_t          j1, 
            Abstract_block* b2, 
            size_t          j2
        );

        /** @brief Detaches a block center shift dependencies from another. */
        static void ctr_ctr_detach_shift_deps
        (
            Abstract_block* b1, 
            size_t          j1, 
            Abstract_block* b2, 
            size_t          j2
        );

        /** @brief Detaches a block center shift dependencies from another. */
        static void var_ctr_detach_shift_deps
        (
            Abstract_block* b1, 
            size_t j1, 
            Abstract_block* b2, 
            size_t j2
        );

        /** @brief Detaches a block center shift dependencies from another. */
        static void ctr_var_detach_shift_deps
        (
            Abstract_block* b1, 
            size_t          j1, 
            Abstract_block* b2, 
            size_t          j2
        );

        /** @brief Detaches a block center scale dependencies from another. */
        static void ctr_ctr_detach_scale_deps
        (
            Abstract_block* b1, 
            size_t          j1, 
            Abstract_block* b2, 
            size_t          j2
        );

        /** @brief Detaches a block center scale dependencies from another. */
        static void var_ctr_detach_scale_deps
        (
            Abstract_block* b1, 
            size_t          j1, 
            Abstract_block* b2, 
            size_t          j2
        );

        /** @brief Detaches a block center scale dependencies from another. */
        static void ctr_var_detach_scale_deps
        (
            Abstract_block* b1, 
            size_t          j1, 
            Abstract_block* b2, 
            size_t          j2
        );

        static void detach_conds
        (
            Abstract_block* b1, 
            size_t          j1, 
            Abstract_block* b2, 
            size_t          j2
        );

        /*@}*/


        /**
         * @name Split
         *
         * Splits a component into two.
         * @{
         */

        void split_aligned_var_attach_pt
        (
            Abstract_block* bs,
            Abstract_block* bi,
            uint32_t        bi_j
        );

        void split_copy_var_attach_pts
        (
            Abstract_block* bm,
            Abstract_block* b1,
            uint32_t        b1_j,
            Abstract_block* b2,
            uint32_t        b2_j
        );

        void split_block_posn
        (
            jwsc::Vector_d** p1_out,
            jwsc::Vector_d** p2_out,
            Abstract_block*  bs,
            Abstract_block*  b1,
            uint32_t         b1_j,
            Abstract_block*  b2,
            uint32_t         b2_j
        );

        bool swap_split_blocks
        (
            Component*      as,
            uint32_t        as_k, 
            Abstract_block* bs, 
            Abstract_block* b1, 
            uint32_t        b1_j, 
            const std::vector<uint32_t>& b1_a,
            Abstract_block* b2, 
            uint32_t        b2_j,
            const std::vector<uint32_t>& b2_a
        );

        /*@}*/


        /**
         * @name Merge
         *
         * Merges two attachments.
         * @{
         */

        void align_blocks_to_merge
        (
            Abstract_block* bm,
            Abstract_block* b1,
            uint32_t        b1_j,
            Abstract_block* b2,
            uint32_t        b2_j
        );

        void merge_aligned_var_attach_pt
        (
            Abstract_block* bm,
            Abstract_block* bi,
            uint32_t        bi_j
        );

        void copy_non_aligned_var_attach_pt
        (
            Component*             c1,
            uint32_t               c1_j,
            Component*             c2,
            uint32_t               c2_j,
            Component::Attach_type c2_t
        );

        void merge_copy_var_attach_pts
        (
            Abstract_block* bm,
            Abstract_block* b1,
            uint32_t        b1_j,
            Abstract_block* b2,
            uint32_t        b2_j
        );

        void merged_block_posn
        (
            jwsc::Vector_d** p_out,
            Abstract_block*  b1,
            uint32_t         b1_j,
            Abstract_block*  b2,
            uint32_t         b2_j
        );

        /*@}*/


        /**
         * @name Switch
         *
         * Switches out one abstract block with another.
         * @{
         */

        void position_switched_block
        (
            const Abstract_block*  b1, 
            Abstract_block*        b2
        );

        /*@}*/


    private:

        /** @brief Rotation angle of this Object. */
        Model::Parameter _azimuth_p;

        /** 
         * @brief Position x-coord of this Object.
         *
         * Contains three parameters indexed by x=0, y=1, z=2.
         */
        std::vector<Component::Parameter> _posn_p;

        /**
         * @brief Lowest component attached to this Object.
         *
         * There is only one attachment; it is defined by the position of this
         * object. If there are no components in this Object, the Pointer
         * contains a null ptr.
         */
        Component::Attachment _lowest_component;

        /** @brief Structure components in this Object. */
        std::vector<Component*> _components;

        /** @brief Storage for the projected faces of this Object. */
        std::vector<jwscxx::graphics::Polygon_d> _projection;

        /** @brief Block model context. */
        const Block::Context* _block_ctx;

        /** @brief Block_set model context. */
        const Block_set::Context* _block_set_ctx;

        /** @brief Position defining the object coord system in the world. */
        jwsc::Vector_d* _obj_coords_posn;

        /** @brief Angle defining the object coord system in the world. */
        double _obj_coords_azimuth;

        /** @brief Current coordinate frame this Object is in. */
        bool _obj_coords;

        /** @brief Next sequential component ID to use. */
        size_t _next_component_id;

        /** @brief Sorted set of current component ids. */
        std::vector<size_t> _used_component_ids;
};

#endif
