/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Functions to enable or disable different types of offscreen
 * rendering.
 */


#include <config.h>

#include <vector>

#include <GL/osmesa.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/graphics/offscreen.h>

#include "options.h"
#include "offscreen.h"


/** @brief Offscreen rendering buffer to use if not running interactively. */
static jwscxx::graphics::Offscreen_buffer* offscreen = 0;

/** @brief Offscreen rendering buffer to use per thread. */
static std::vector<jwscxx::graphics::Offscreen_buffer*> offscreen_threads;


void Offscreen::enable() throw (jwscxx::base::Exception)
{
    using namespace jwscxx::graphics;

    offscreen = new OSMesa32_offscreen_buffer(opts.view_width, opts.view_height,
            OSMESA_RGBA, 4, 4, 4);

    for (uint32_t i = 0; i < opts.num_threads; i++)
    {
        offscreen_threads.push_back(new OSMesa32_offscreen_buffer(
                    opts.reso_width, opts.reso_height, OSMESA_RGBA, 4, 4, 4));
    }

    activate();
}


bool Offscreen::enabled()
{
    return (offscreen != 0);
}


void Offscreen::disable() throw (jwscxx::base::Exception) 
{
    delete offscreen;

    for (uint32_t i = 0; i < offscreen_threads.size(); i++)
    {
        delete offscreen_threads[ i ];
    }
}


void Offscreen::activate() throw (jwscxx::base::Exception)
{
    offscreen->activate();
    init_gl();
}


void Offscreen::deactivate() throw (jwscxx::base::Exception)
{
    offscreen->deactivate();
}


void Offscreen::activate(uint32_t i) throw (jwscxx::base::Exception)
{
    if (i >= offscreen_threads.size())
    {
        throw jwscxx::base::Exception("Offscreen thread index too large");
    }
    offscreen_threads[ i ]->activate();
    init_gl();
}


void Offscreen::deactivate(uint32_t i) throw (jwscxx::base::Exception)
{
    if (i >= offscreen_threads.size())
    {
        throw jwscxx::base::Exception("Offscreen thread index too large");
    }
    offscreen_threads[ i ]->deactivate();
}
