/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Interactive sampler visualization.
 */


#include <config.h>

#include <sstream>
#include <cstdlib>
#include <cassert>

#include <inttypes.h>
#include <pthread.h>

#if defined FURNITURE_HAVE_OPENGL
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <X11/Xlib.h>
#elif defined FURNITURE_HAVE_OPENGL_FRAMEWORK
#include <OpenGL/OpenGL.h>
#endif

#if defined FURNITURE_HAVE_GLUT_FRAMEWORK
#include <GLUT/glut.h>
#elif defined FURNITURE_HAVE_GLUT
#include <GL/glut.h>
#endif

#include <jwsc/math/constants.h>
#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/image/image.h>
#include <jwsc/image/image_io.h>
#include <jwsc/image/image_util.h>

#include <jwsc++/base/exception.h>

#include "view.h"
#include "object.h"
#include "sampler.h"
#include "metropolis.h"
#include "dynamics.h"
#include "likelihood.h"
#include "options.h"
#include "offscreen.h"
#include "render.h"
#include "onscreen.h"


#if defined FURNITURE_INTERACTIVE


#if defined FURNITURE_HAVE_OPENGL
Display* display;
GLXDrawable drawable;
GLXContext context;
#elif defined FURNITURE_HAVE_OPENGL_FRAMEWORK
CGLContextObj context;
#endif

static pthread_t sampler_thread;

static pthread_mutex_t sampler_mux = PTHREAD_MUTEX_INITIALIZER;

static pthread_cond_t iter_to_run = PTHREAD_COND_INITIALIZER;

/** @brief Whether onscreen rendering with glut is currently enabled. */
static bool glut_enabled = false;

/** @brief Sampler run in the interactive environment. */
static Sampler* sampler = 0;

/** @brief Sampler parameters to use in the interactive environment. */
static Sampler::Parameters* params = 0;

/** @brief Filled quadric used for the world axis. */
static GLUquadric* quad_fill = 0;


/** @brief Types of interactive model changes to make. */
typedef enum
{
    CHANGE_VIEW_ALTITUDE,
    CHANGE_VIEW_FOCAL_LENGTH,
    CHANGE_VIEW_SCALE,
    CHANGE_VIEW_NEAR,
    CHANGE_VIEW_FAR,
    CHANGE_OBJ_AZIMUTH,
    CHANGE_OBJ_X_POSN,
    CHANGE_OBJ_Y_POSN,
    CHANGE_OBJ_Z_POSN,
    CHANGE_EDGE_NOISE_PROB
}
Model_change_type;


/**
 * @brief Interactive modes for the program.
 *
 * The modes are not mutually exclusive.
 */
struct Interactive_modes
{
    /** @brief Capture the current view as a 2D image. */
    bool capture_view;

    /** @brief Capture the current view as a 2D image and crops the padding. */
    bool capture_view_crop;

    /** @brief Draw the 3D world axis in the viewer. */
    bool render_world_axis;

    /** @brief Draw the input image on the focal plane. */
    bool render_image;

    /** @brief Draw the input image edge points on the focal plane. */
    bool render_edges;

    /** @brief Draw the MH proposal model. */
    bool render_wire_proposal;

    /** @brief Draw the model. */
    bool render_wire_model;

    /** @brief Draw the filled model. */
    bool render_filled_model;

    /** @brief Draw the shaded model. */
    bool render_shaded_model;

    /** @brief Draw the log likelihood surf point assignment visualization. */
    bool render_img_ll_surf_assign;

    /** @brief Draw the log likelihood edge point assignment visualization. */
    bool render_img_ll_edge_assign;

    /** @brief Draw the thresholded input input. */
    bool render_image_t;

    /** @brief Draw the edge map for the input image. */
    bool render_edge_map;

    /** @brief Renders a sphere where the world origin is. */
    bool render_origin;

    /** @brief Renders a small polygon on the ground plane under the object. */
    bool render_ground_plane;

    /** @brief Renders a sphere where the object center (position) is. */
    bool render_object_center;

    /** @brief Renders a border around where the image is located in the view.*/
    bool render_image_border;

    /** @brief Model component to interactively change. */
    Model_change_type change;
}
modes;


/** @brief Initialize the interactive modes. */
static void init_interactive_modes()
{
    modes.capture_view         = false;
    modes.capture_view_crop    = false;
    modes.render_world_axis    = false;
    modes.render_image         = false;
    modes.render_edges         = true;
    modes.render_wire_proposal = false;
    modes.render_wire_model    = true;
    modes.render_filled_model  = false;
    modes.render_shaded_model  = false;
    modes.render_img_ll_surf_assign = false;
    modes.render_img_ll_edge_assign = false;
    modes.render_image_t       = false;
    modes.render_edge_map      = false;
    modes.render_origin        = false;
    modes.render_ground_plane  = false;
    modes.render_object_center = false;
    modes.render_image_border  = true;
    modes.change               = CHANGE_VIEW_ALTITUDE;
}


static uint32_t ins_i = 0;
static uint32_t cat_c = 0;


/** @brief Renders the 3D world axis. */
static void render_world_axis()
{
    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();

    float axis_length  = 100.0f;
    float arrow_radius = axis_length / 10.0f;
    float arrow_height = axis_length / 5.0f;

    glColor3f(0.9f, 1.0f, 0.3f);

    // X
    glPushMatrix();
      glRotatef(90, 0, 1.0f, 0);
      gluCylinder(quad_fill, 1, 1, axis_length - 0.5f*arrow_height, 8, 1);
    glPopMatrix();
    glPushMatrix();
      glTranslatef(axis_length - arrow_height, 0, 0);
      glRotatef(90, 0, 1.0f, 0);
      gluCylinder(quad_fill, arrow_radius, 0, arrow_height, 16, 8);
    glPopMatrix();

    // Y
    glPushMatrix();
      glRotatef(-90, 1.0f, 0, 0);
      gluCylinder(quad_fill, 1, 1, axis_length - 0.5f*arrow_height, 8, 1);
    glPopMatrix();
    glPushMatrix();
      glTranslatef(0, axis_length - arrow_height, 0);
      glRotatef(-90, 1.0f, 0, 0);
      gluCylinder(quad_fill, arrow_radius, 0, arrow_height, 16, 8);
    glPopMatrix();

    // Z
    gluCylinder(quad_fill, 1, 1, axis_length - 0.5f*arrow_height, 8, 1);
    glPushMatrix();
      glTranslatef(0, 0, axis_length - arrow_height);
      gluCylinder(quad_fill, arrow_radius, 0, arrow_height, 16, 8);
    glPopMatrix();

    glPopMatrix();
}


/** @brief Renders the input image on the focal plane. */
static void render_image()
{
    const Sampler::Instance* ins = params->instance(cat_c, ins_i);
    Render::full_image(ins_i, ins->view);
}


static void render_img_ll_surf_assign()
{
    const Sampler::Instance* ins = params->instance(cat_c, ins_i);

    glDisable(GL_DEPTH_TEST);

    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();
    glLoadIdentity();

    uint32_t num_rows = ins->img_ll_surf_assign->num_rows;
    uint32_t num_cols = ins->img_ll_surf_assign->num_cols;

    float xtr_row = (num_rows % 2) ? 0.5 : 0.5;
    float xtr_col = (num_cols % 2) ? 0.5 : 0.5;

    for (uint32_t row = 0; row < num_rows; row++)
    {
        for (uint32_t col = 0; col < num_cols; col++)
        {
            uint32_t x = ins->img_ll_surf_assign->elts[ row ][ col ];

            switch (x)
            {
                case 1:  
                    // Magenta -- Surface
                    glColor3f(1, 0, 1);
                    break;
                case 2:  
                    // Blue -- Noisy surface
                    glColor3f(0, 0, 1);
                    break;
                case 3:  
                    // Red -- Missing surface
                    glColor3f(1, 0, 0);
                    break;
                default: 
                    // Black -- Background
                    glColor3f(0, 0, 0);
            }

            if (x > 0)
            {
                glBegin(GL_POINTS);
                glVertex3f((col - (num_cols/2.0f) + xtr_col), 
                          -(row - (num_rows/2.0f) + xtr_row), 
                          -ins->view->focal()*opts.resolution);
                glEnd();
            }
        }
    }

    glPopMatrix();

    glEnable(GL_DEPTH_TEST);
}


static void render_img_ll_edge_assign()
{
    const Sampler::Instance* ins = params->instance(cat_c, ins_i);

    glDisable(GL_DEPTH_TEST);

    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();
    glLoadIdentity();

    uint32_t num_rows = ins->img_ll_edge_assign->num_rows;
    uint32_t num_cols = ins->img_ll_edge_assign->num_cols;

    float xtr_row = (num_rows % 2) ? 0.5 : 0.5;
    float xtr_col = (num_cols % 2) ? 0.5 : 0.5;

    for (uint32_t row = 0; row < num_rows; row++)
    {
        for (uint32_t col = 0; col < num_cols; col++)
        {
            uint32_t x = ins->img_ll_edge_assign->elts[ row ][ col ];

            switch (x)
            {
                case 1:  
                    // Magenta -- Edge
                    glColor3f(1, 0, 1);
                    break;
                case 2:  
                    // Blue -- Noisy edge
                    glColor3f(0, 0, 1);
                    break;
                case 3:  
                    // Red -- Missing silhouette edge
                    glColor3f(1, 0, 0);
                    break;
                case 4:  
                    // Green -- Missing inner edge
                    glColor3f(0, 1, 0);
                    break;
                default: 
                    // Black -- Background
                    glColor3f(0, 0, 0);
            }

            if (x > 0)
            {
                glBegin(GL_POINTS);
                glVertex3f((col - (num_cols/2.0f) + xtr_col), 
                          -(row - (num_rows/2.0f) + xtr_row), 
                          -ins->view->focal()*opts.resolution);
                glEnd();
            }
        }
    }

    glPopMatrix();

    glEnable(GL_DEPTH_TEST);
}


static void render_image_t()
{
    const Sampler::Instance* ins = params->instance(cat_c, ins_i);

    glDisable(GL_DEPTH_TEST);

    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();
    glLoadIdentity();

    glClearColor(1, 1, 1, 0);
    glClear(GL_COLOR_BUFFER_BIT);

    uint32_t num_rows = ins->image_t->num_rows;
    uint32_t num_cols = ins->image_t->num_cols;

    float xtr_row = (num_rows % 2) ? 0.5 : 0.5;
    float xtr_col = (num_cols % 2) ? 0.5 : 0.5;

    for (uint32_t row = 0; row < num_rows; row++)
    {
        for (uint32_t col = 0; col < num_cols; col++)
        {
            float x = ins->image_t->elts[ row ][ col ];

            if (x < 1.0)
            {
                glColor3f(0.0, 0.0, 0.0);

                glBegin(GL_POINTS);
                glVertex3f((col - (num_cols/2.0f) + xtr_col), 
                          -(row - (num_rows/2.0f) + xtr_row), 
                          -ins->view->focal()*opts.resolution);
                glEnd();
            }
        }
    }

    glPopMatrix();

    glEnable(GL_DEPTH_TEST);
}


static void render_edge_map()
{
    const Sampler::Instance* ins = params->instance(cat_c, ins_i);
    const jwsc::Edge_set_f* edges = opts.edges[ins_i];

    glDisable(GL_DEPTH_TEST);

    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();
    glLoadIdentity();

    glClearColor(1, 1, 1, 0);
    glClear(GL_COLOR_BUFFER_BIT);

    uint32_t num_rows = opts.images_in[ ins_i ]->num_rows;
    uint32_t num_cols = opts.images_in[ ins_i ]->num_cols;

    uint32_t r = opts.images_in_row_pad[ ins_i ];
    uint32_t c = opts.images_in_col_pad[ ins_i ];

    float odd_row = ((opts.images_reso_max_num_rows - num_rows) % 2) ? 0 : 0.5;
    float odd_col = ((opts.images_reso_max_num_cols - num_cols) % 2) ? 0 : 0.5;

    glColor3f(0, 0, 0);
    glBegin(GL_POINTS);
    for (uint32_t e = 0; e < edges->num_edges; e++)
    {
        for (uint32_t p = 0; p < edges->num_pts[ e ]; p++)
        {
            uint32_t row = edges->pts[ e ][ p ].row - r;
            uint32_t col = edges->pts[ e ][ p ].col - c;

            float x =  (float)col + odd_col - ((float)num_cols / 2.0f);
            float y = -(float)row - odd_row + ((float)num_rows / 2.0f);
            float z = -ins->view->focal()*opts.resolution;

            glVertex3f(x, y, z);
        }
    }
    glEnd();

    glPopMatrix();

    glEnable(GL_DEPTH_TEST);
}


/** @brief Renders the input image edge points on the focal plane. */
static void render_edges()
{
    const Sampler::Instance* ins = params->instance(cat_c, ins_i);
    glColor3f(0, 1, 0);
    Render::edges(ins_i, ins->view);
}


static void render_origin()
{
    GLboolean prev_blend;
    GLint     prev_blend_src;
    GLint     prev_blend_dst;

    if ((prev_blend = glIsEnabled(GL_BLEND)))
    {
        glGetIntegerv(GL_BLEND_SRC, &prev_blend_src);
        glGetIntegerv(GL_BLEND_DST, &prev_blend_dst);
    }
    else
    {
        glEnable(GL_BLEND);
    }
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();

    float z = 0;
    float y = 0;
    float x = 0;

    glColor4f(0.0, 0.6, 0.0, 0.6);
    glTranslatef(x, y, z);
    gluSphere(quad_fill, 6, 20, 20);

    glPopMatrix();

    if (prev_blend)
    {
        glBlendFunc(prev_blend_src, prev_blend_dst);
    }
    else
    {
        glDisable(GL_BLEND);
    }
}


static void render_object_center()
{
    const Sampler::Instance* ins = params->instance(cat_c, ins_i);

    GLboolean prev_blend;
    GLint     prev_blend_src;
    GLint     prev_blend_dst;

    if ((prev_blend = glIsEnabled(GL_BLEND)))
    {
        glGetIntegerv(GL_BLEND_SRC, &prev_blend_src);
        glGetIntegerv(GL_BLEND_DST, &prev_blend_dst);
    }
    else
    {
        glEnable(GL_BLEND);
    }
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();

    float x = ins->obj->posn(0);
    float y = ins->obj->posn(1);
    float z = ins->obj->posn(2);

    glColor4f(0.1, 0.1, 1.0, 0.6);
    glTranslatef(x, y, z);
    gluSphere(quad_fill, 10, 20, 20);

    glPopMatrix();

    if (prev_blend)
    {
        glBlendFunc(prev_blend_src, prev_blend_dst);
    }
    else
    {
        glDisable(GL_BLEND);
    }
}


static void render_ground_plane()
{
    const Sampler::Instance* ins = params->instance(cat_c, ins_i);

    GLboolean prev_blend;
    GLint     prev_blend_src;
    GLint     prev_blend_dst;

    if ((prev_blend = glIsEnabled(GL_BLEND)))
    {
        glGetIntegerv(GL_BLEND_SRC, &prev_blend_src);
        glGetIntegerv(GL_BLEND_DST, &prev_blend_dst);
    }
    else
    {
        glEnable(GL_BLEND);
    }
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();

    float y_0 = 0;
    float z_0 = 0.5 * ins->view->get_half_focal_height();
    float x_0 = z_0 * ((float)opts.images_in_max_num_cols /
                       (float)opts.images_in_max_num_rows);

    float x = ins->obj->posn(0);
    float y = 0;
    float z = ins->obj->posn(2);

    glColor4f(1, 0.1, 0.1, 0.3);
    glBegin(GL_POLYGON);
    glVertex3f( x_0 + x, y_0 + y,  z_0 + z);
    glVertex3f(-x_0 + x, y_0 + y,  z_0 + z);
    glVertex3f(-x_0 + x, y_0 + y, -z_0 + z);
    glVertex3f( x_0 + x, y_0 + y, -z_0 + z);
    glEnd();

    glPopMatrix();

    if (prev_blend)
    {
        glBlendFunc(prev_blend_src, prev_blend_dst);
    }
    else
    {
        glDisable(GL_BLEND);
    }

}


static void render_image_border()
{
    const Sampler::Instance* ins = params->instance(cat_c, ins_i);

    glDisable(GL_DEPTH_TEST);

    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();
    glLoadIdentity();

    int num_rows = opts.images_in[ ins_i ]->num_rows;
    int num_cols = opts.images_in[ ins_i ]->num_cols;
    int row, col;

    float xtr_row = (num_rows % 2) ? 0.5 : 0.5;
    float xtr_col = (num_cols % 2) ? 0.5 : 0.5;

    glColor3f(0.0, 0.0, 0.0);

    glBegin(GL_LINE_LOOP);
    row = -1; col = -1;
    glVertex3f((col - (num_cols/2.0f) + xtr_col), 
            -(row - (num_rows/2.0f) + xtr_row), 
            -ins->view->focal()*opts.resolution);
    row = -1; col = num_cols;
    glVertex3f((col - (num_cols/2.0f) + xtr_col), 
            -(row - (num_rows/2.0f) + xtr_row), 
            -ins->view->focal()*opts.resolution);
    row = num_rows; col = num_cols;
    glVertex3f((col - (num_cols/2.0f) + xtr_col), 
            -(row - (num_rows/2.0f) + xtr_row), 
            -ins->view->focal()*opts.resolution);
    row = num_rows; col = -1;
    glVertex3f((col - (num_cols/2.0f) + xtr_col), 
            -(row - (num_rows/2.0f) + xtr_row), 
            -ins->view->focal()*opts.resolution);
    glEnd();

    glPopMatrix();

    glEnable(GL_DEPTH_TEST);
}


static void render_filled_model()
{
    const Sampler::Instance* ins = params->instance(cat_c, ins_i);
    glColor3f(0, 0, 0);
    Render::solid_object(ins->view, ins->obj);
}


static void render_shaded_model()
{
    const Sampler::Instance* ins = params->instance(cat_c, ins_i);

    static GLfloat ambient0  [] = {0.2f, 0.2f, 0.2f, 1.0f};
    static GLfloat diffuse0  [] = {0.8f, 0.8f, 0.8f, 1.0f};
    static GLfloat direction0[] = {0.0f, 0.0f, 0.0f, 1.0f};

    direction0[0] = ins->view->get_prp()->elts[0];
    direction0[1] = ins->view->get_prp()->elts[1];
    direction0[2] = ins->view->get_prp()->elts[2];

    glEnable(GL_LIGHT0);
    glLightfv(GL_LIGHT0, GL_AMBIENT, ambient0);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse0);
    glLightfv(GL_LIGHT0, GL_POSITION, direction0);

    static GLfloat amb_dif[4] = {0.7f, 0.6f, 0.5f, 0.2f};
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, amb_dif);

    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();

    ins->view->set_gl_projection();
    ins->view->set_gl_modelview();

    glEnable(GL_LIGHTING);
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, amb_dif);
    ins->obj->solid_render();
    glDisable(GL_LIGHTING);

    glClear(GL_DEPTH_BUFFER_BIT);
    glColor3f(0, 0, 0);
    ins->obj->render(); 

    glPopMatrix();
}


static void render_wire_model()
{
    const Sampler::Instance* ins = params->instance(cat_c, ins_i);

    if (params->member(cat_c, ins_i) > 0)
    {
        //glColor3ub((GLubyte)0xe6, (GLubyte)0xe6, (GLubyte)0x00);
        glColor3ub((GLubyte)0xEE, (GLubyte)0x00, (GLubyte)0x00);
    }
    else
    {
        glColor3ub((GLubyte)0xff, (GLubyte)0x8c, (GLubyte)0x00);
    }

    Render::wire_object(ins->view, ins->obj);
}


static void render_wire_proposal()
{
    const Sampler::Instance* ins = params->instance(cat_c, ins_i);

    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();

    ((Metropolis::Instance*)ins)->prop_view->set_gl_projection();
    ((Metropolis::Instance*)ins)->prop_view->set_gl_modelview();
    glColor3f(1, 0, 0);
    ((Metropolis::Instance*)ins)->prop_obj->render();

    glPopMatrix();
}


static void capture_view(uint32_t N)
{
    using namespace jwsc;

    char fname[256] = {0};
    snprintf(fname, 255, opts.capture_fmt, N);

    Image_f* img = 0;
    Error* err;

    View::capture_gl_view(&img);

    try
    {
        if ((err = write_image_f(img, fname)) != 0)
        {
            std::ostringstream ost(err->msg);
            free_error(err);
            throw jwscxx::base::IO_error(ost.str());
        }
    }
    catch (jwscxx::base::Exception e)
    { std::cerr << e.get_msg() << std::endl; }

    free_image_f(img);
}


static void capture_view_crop(uint32_t N)
{
    using namespace jwsc;

    char fname[256] = {0};
    snprintf(fname, 255, opts.capture_fmt, N);

    Image_f* img = 0;
    Error* err;

    uint32_t max_rows = opts.images_reso_max_num_rows;
    uint32_t max_cols = opts.images_reso_max_num_cols;
    uint32_t num_rows = opts.images_in[ ins_i ]->num_rows;
    uint32_t num_cols = opts.images_in[ ins_i ]->num_cols;
    uint32_t row      = opts.reso_model_pad + 0.5*(max_rows - num_rows);
    uint32_t col      = opts.reso_model_pad + 0.5*(max_cols - num_cols);

    View::capture_gl_view(&img);
    assert(crop_image_f(&img, img, row, col, num_rows, num_cols) == 0);

    try
    {
        if ((err = write_image_f(img, fname)) != 0)
        {
            std::ostringstream ost(err->msg);
            free_error(err);
            throw jwscxx::base::IO_error(ost.str());
        }
    }
    catch (jwscxx::base::Exception e)
    { std::cerr << e.get_msg() << std::endl; }

    free_image_f(img);
}


/** @brief Display callback for the GLUT library. */
static void display_glut()
{
    params->lock(cat_c, ins_i);

    Sampler::Instance* ins = params->instance(cat_c, ins_i);

    static uint32_t N = 1;

    glDrawBuffer(GL_BACK);
    glReadBuffer(GL_BACK);

    glClearColor(1, 1, 1, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    ins->view->set_gl_projection();
    ins->view->set_gl_modelview();

    if (modes.render_image)
    {
        render_image();
    }

    if (modes.render_image_t)
    {
        render_image_t();
    }

    if (modes.render_edge_map)
    {
        render_edge_map();
    }

    if (modes.render_edges)
    {
        render_edges();
    }

    if (modes.render_img_ll_surf_assign)
    {
        render_img_ll_surf_assign();
    }

    if (modes.render_img_ll_edge_assign)
    {
        render_img_ll_edge_assign();
    }

    if (modes.render_world_axis)
    {
        render_world_axis();
    }

    if (modes.render_origin)
    {
        render_origin();
    }

    if (modes.render_ground_plane)
    {
        render_ground_plane();
    }

    if (modes.render_object_center)
    {
        render_object_center();
    }

    if (modes.render_image_border)
    {
        render_image_border();
    }

    if (modes.render_shaded_model)
    {
        render_shaded_model();
    }

    if (modes.render_filled_model)
    {
        render_filled_model();
    }

    if (dynamic_cast<const Metropolis::Instance*>(ins) && 
            modes.render_wire_proposal)
    {
        render_wire_proposal();
    }

    if (modes.render_wire_model)
    {
        render_wire_model();
    }

    glFlush();

    glutSwapBuffers();

    if (modes.capture_view)
    {
        capture_view(N++);
        modes.capture_view = 0;
    }

    if (modes.capture_view_crop)
    {
        capture_view_crop(N++);
        modes.capture_view_crop = 0;
    }

    params->unlock(cat_c, ins_i);
}


/** @brief Reshape callback for the GLUT library. */
static void reshape_glut(int w, int h)
{
    float s = (float)h / (float)opts.reso_height;
    glPointSize(s+0.49999);
    glLineWidth(s);

    glClearColor(1, 1, 1, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glViewport(0, 0, (GLsizei)w, (GLsizei)h); 
}


static void* run_sampler(void* ignored)
{
    Offscreen::enable();
    Offscreen::activate();

    assert(pthread_mutex_lock(&sampler_mux) == 0);

    try { sampler->run(params, opts.iterations); }
    catch (jwscxx::base::Exception e)
    { 
        std::cerr << e.get_msg() << std::endl; 
        abort();
    }

    opts.iterations = 0;
    assert(pthread_cond_wait(&iter_to_run, &sampler_mux) == 0);

    while (opts.iterations)
    {
        try { sampler->run(params, opts.iterations); }
        catch (jwscxx::base::Exception e)
        { 
            std::cerr << e.get_msg() << std::endl; 
            abort();
        }

        opts.iterations = 0;
        assert(pthread_cond_wait(&iter_to_run, &sampler_mux) == 0);
    }

    return 0;
}


/** @brief Timer callback for the GLUT library. */
static void timer_glut(int ignored)
{
    glutPostRedisplay();
    glutTimerFunc(1000, timer_glut, 0);
}


static void run_sampler_menu_glut(int i);


/** @brief Keyboard callback for the GLUT library. */
static void keyboard_glut(unsigned char key, int a, int b)
{
    using namespace jwsc;

    uint32_t locked_cat_c = cat_c;
    uint32_t locked_ins_i = ins_i;

    params->lock(locked_cat_c, locked_ins_i);

    Offscreen::activate();

    Sampler::Category* cat = params->category(cat_c);
    Sampler::Instance* ins = params->instance(cat_c, ins_i);

    double x;

    View* view  = ins->view;
    Object* obj = ins->obj;

    Vector_f* v = 0;
    Matrix_f* R = 0;

    try
    {
    switch (key) {
        case 'k':
            switch (modes.change) {
                case CHANGE_VIEW_ALTITUDE:
                    x = view->zenith();
                    view->test_zenith(x + 0.01);
                    view->zenith(x + 0.01);
                    break;
                case CHANGE_VIEW_FOCAL_LENGTH:
                    x = view->focal();
                    view->test_focal(x + 10);
                    view->focal(x + 10);
                    break;
                case CHANGE_VIEW_SCALE:
                    x = view->scale();
                    view->test_scale(x + 0.01);
                    view->scale(x + 0.01);
                    break;
                case CHANGE_VIEW_NEAR:
                    x = view->get_clipping_near();
                    view->set_clipping(x+10, view->get_clipping_far());
                    break;
                case CHANGE_VIEW_FAR:
                    x = view->get_clipping_far();
                    view->set_clipping(view->get_clipping_near(), x+10);
                    break;
                case CHANGE_OBJ_AZIMUTH:
                    x = obj->azimuth();
                    obj->test_azimuth(x + 0.01);
                    obj->azimuth(x + 0.01);
                    break;
                case CHANGE_OBJ_X_POSN:
                    x = obj->posn(0);
                    obj->test_posn_shift_deps(0, x + 1);
                    obj->posn_shift_deps(0, x + 1);
                    break;
                case CHANGE_OBJ_Y_POSN:
                    x = obj->posn(1);
                    obj->test_posn_shift_deps(1, x + 1);
                    obj->posn_shift_deps(1, x + 1);
                    break;
                case CHANGE_OBJ_Z_POSN:
                    x = obj->posn(2);
                    obj->test_posn_shift_deps(2, x + 1);
                    obj->posn_shift_deps(2, x + 1);
                    break;
                case CHANGE_EDGE_NOISE_PROB:
                    if (opts.img_ll_edge_noise_prob + 1.0e-62 <= 1)
                    {
                        opts.img_ll_edge_noise_prob += 1.0e-62;
                        std::cerr << opts.img_ll_edge_noise_prob << std::endl;
                    }
                    break;
            }

            Likelihood::calc_neg_log_likelihood(&(ins->ll_potential),
                    &(ins->img_ll), &(ins->model_mat),
                    &(ins->img_ll_surf_assign), &(ins->img_ll_edge_assign),
                    ins->image_t, ins->edges, ins->img_ll_corr, ins->view,
                    ins->obj, opts.img_ll_bg_prob, opts.img_ll_surf_prob,
                    opts.img_ll_surf_noise_prob, opts.img_ll_surf_miss_prob,
                    opts.img_ll_edge_noise_prob,
                    opts.img_ll_edge_silho_miss_prob,
                    opts.img_ll_edge_inner_miss_prob);

            glutPostRedisplay();
            break;
        case 'j':
            switch (modes.change) {
                case CHANGE_VIEW_ALTITUDE:
                    x = view->zenith();
                    view->test_zenith(x - 0.01);
                    view->zenith(x - 0.01);
                    break;
                case CHANGE_VIEW_FOCAL_LENGTH:
                    x = view->focal();
                    view->test_focal(x - 10);
                    view->focal(x - 10);
                    break;
                case CHANGE_VIEW_SCALE:
                    x = view->scale();
                    view->test_scale(x - 0.01);
                    view->scale(x - 0.01);
                    break;
                case CHANGE_VIEW_NEAR:
                    x = view->get_clipping_near();
                    view->set_clipping(x-10, view->get_clipping_far());
                    break;
                case CHANGE_VIEW_FAR:
                    x = view->get_clipping_far();
                    view->set_clipping(view->get_clipping_near(), x-10);
                    break;
                case CHANGE_OBJ_AZIMUTH:
                    x = obj->azimuth();
                    obj->test_azimuth(x - 0.01);
                    obj->azimuth(x - 0.01);
                    break;
                case CHANGE_OBJ_X_POSN:
                    x = obj->posn(0);
                    obj->test_posn_shift_deps(0, x - 1);
                    obj->posn_shift_deps(0, x - 1);
                    break;
                case CHANGE_OBJ_Y_POSN:
                    x = obj->posn(1);
                    obj->test_posn_shift_deps(1, x - 1);
                    obj->posn_shift_deps(1, x - 1);
                    break;
                case CHANGE_OBJ_Z_POSN:
                    x = obj->posn(2);
                    obj->test_posn_shift_deps(2, x - 1);
                    obj->posn_shift_deps(2, x - 1);
                    break;
                case CHANGE_EDGE_NOISE_PROB:
                    if (opts.img_ll_edge_noise_prob - 1.0e-62 >= 0)
                    {
                        opts.img_ll_edge_noise_prob -= 1.0e-62;
                        std::cerr << params->img_ll_edge_noise_prob() 
                                  << std::endl;
                    }
                    break;
            }

            Likelihood::calc_neg_log_likelihood(&(ins->ll_potential),
                    &(ins->img_ll), &(ins->model_mat),
                    &(ins->img_ll_surf_assign), &(ins->img_ll_edge_assign),
                    ins->image_t, ins->edges, ins->img_ll_corr, ins->view,
                    ins->obj, opts.img_ll_bg_prob, opts.img_ll_surf_prob,
                    opts.img_ll_surf_noise_prob, opts.img_ll_surf_miss_prob,
                    opts.img_ll_edge_noise_prob,
                    opts.img_ll_edge_silho_miss_prob,
                    opts.img_ll_edge_inner_miss_prob);

            glutPostRedisplay();
            break;
        case 'i':
            ins->print_var_param_info(std::cout);
            ins->obj->assert_component_consistency();
            break;
        case 'I':
            cat->print_var_param_info(std::cout);
            break;
        case 'c':
            modes.capture_view = 1;
            glutPostRedisplay();
            break;
        case 'C':
            modes.capture_view_crop = 1;
            glutPostRedisplay();
            break;
        case 'n':
            if (ins_i+1 < params->num_instances())
            {
                ins_i++;
                glutPostRedisplay();
            }
            break;
        case 'p':
            if (ins_i > 0)
            {
                ins_i--;
                glutPostRedisplay();
            }
            break;
        case 'N':
            if (cat_c+1 < params->num_categories())
            {
                cat_c++;
                glutPostRedisplay();
            }
            break;
        case 'P':
            if (cat_c > 0)
            {
                cat_c--;
                glutPostRedisplay();
            }
            break;
        case 'w':
            write_results(params);
            break;
        case '1':
            run_sampler_menu_glut(1);
            break;
        case 'q':
            delete sampler;
            delete params;
            clean_up_furniture();
            if (jwsc::get_num_unhandled_errors() > 0)
            {
                std::cerr << "furniture: Unhandled errors: " 
                          << jwsc::get_num_unhandled_errors();
                exit(EXIT_FAILURE);
            }
            exit(EXIT_SUCCESS);
    }
    }
    catch (jwscxx::base::Exception e)
    {
        e.print_details();
    }

    free_vector_f(v);
    free_matrix_f(R);

    Offscreen::deactivate();
    Onscreen::activate();

    params->unlock(locked_cat_c, locked_ins_i);
}


/** @brief Mouse callback for the GLUT library. */
static void mouse_glut(int button, int state, int x, int y)
{
    ;
}


/** @brief Main menu callback for the GLUT library. */
static void main_menu_glut(int id)
{
    switch (id)
    {
        case 0:
            modes.capture_view = 1;
            glutPostRedisplay();
            break;
        case 1:
            write_results(params);
            break;
        case 2:
            exit(EXIT_SUCCESS);
    }
}


/** @brief Run the sampler submenu callback for the GLUT library. */
static void run_sampler_menu_glut(int i)
{
    if (pthread_mutex_trylock(&sampler_mux))
    {
        return;
    }

    opts.iterations = (uint32_t)i;
    assert(pthread_cond_signal(&iter_to_run) == 0);
    assert(pthread_mutex_unlock(&sampler_mux) == 0);
}


/** @brief Option submenu callback for the GLUT library. */
static void option_menu_glut(int id)
{
    switch (id)
    {
        case 0:
            modes.render_image = !modes.render_image;
            break;
        case 1:
            modes.render_edges = !modes.render_edges;
            break;
        case 2:
            modes.render_world_axis = !modes.render_world_axis;
            break;
        case 3:
            modes.render_wire_proposal = !modes.render_wire_proposal;
            break;
        case 4:
            modes.render_wire_model = !modes.render_wire_model;
            break;
        case 5:
            modes.render_filled_model = !modes.render_filled_model;
            break;
        case 6:
            modes.render_shaded_model = !modes.render_shaded_model;
            break;
        case 7:
            modes.render_img_ll_surf_assign = !modes.render_img_ll_surf_assign;
            break;
        case 8:
            modes.render_img_ll_edge_assign = !modes.render_img_ll_edge_assign;
            break;
        case 9:
            modes.render_image_t = !modes.render_image_t;
            break;
        case 10:
            modes.render_edge_map = !modes.render_edge_map;
            break;
        case 11:
            modes.render_origin = !modes.render_origin;
            break;
        case 12:
            modes.render_ground_plane = !modes.render_ground_plane;
            break;
        case 13:
            modes.render_object_center = !modes.render_object_center;
            break;
        case 14:
            modes.render_image_border = !modes.render_image_border;
            break;
    }
    glutPostRedisplay();
}


/** @brief Capture submenu callback for the GLUT library. */
static void capture_menu_glut(int id)
{
    switch (id)
    {
        case 0:
            modes.capture_view = !modes.capture_view;
            break;
        case 1:
            modes.capture_view_crop = !modes.capture_view_crop;
            break;
    }
}


static void img_ll_change_menu_glut(int id)
{
    switch (id)
    {
        case 0:
            modes.change = CHANGE_EDGE_NOISE_PROB;
            break;
    }
}


/** @brief Size change submenu callback for the GLUT library. */
static void obj_change_menu_glut(int id)
{
    switch (id)
    {
        case 0:
            modes.change = CHANGE_OBJ_AZIMUTH;
            break;
        case 1:
            modes.change = CHANGE_OBJ_X_POSN;
            break;
        case 2:
            modes.change = CHANGE_OBJ_Y_POSN;
            break;
        case 3:
            modes.change = CHANGE_OBJ_Z_POSN;
            break;
    }
}


/** @brief View change submenu callback for the GLUT library. */
static void view_change_menu_glut(int id)
{
    switch (id)
    {
        case 0:
            modes.change = CHANGE_VIEW_ALTITUDE;
            break;
        case 1:
            modes.change = CHANGE_VIEW_FOCAL_LENGTH;
            break;
        case 2:
            modes.change = CHANGE_VIEW_SCALE;
            break;
        case 3:
            modes.change = CHANGE_VIEW_NEAR;
            break;
        case 4:
            modes.change = CHANGE_VIEW_FAR;
            break;
    }
}


static void category_menu_glut(int c)
{
    cat_c = (uint32_t)c;
    glutPostRedisplay();
}


static void image_menu_glut(int i)
{
    ins_i = (uint32_t)i;
    glutPostRedisplay();
}


static void init_glut
(
    int          argc, 
    const char** argv, 
    const char*  title
)
{
    glutInit(&argc, (char**)argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_ACCUM
                                    | GLUT_ALPHA | GLUT_STENCIL);
    glutInitWindowSize(opts.reso_width, opts.reso_height);
    glutCreateWindow(title);
    glutDisplayFunc(display_glut); 
    glutReshapeFunc(reshape_glut);
    glutKeyboardFunc(keyboard_glut);
    glutMouseFunc(mouse_glut);
    glutTimerFunc(1000, timer_glut, 0);

    int run_sampler = glutCreateMenu(run_sampler_menu_glut);
    glutAddMenuEntry("1", 1);
    glutAddMenuEntry("10", 10);
    glutAddMenuEntry("100", 100);
    glutAddMenuEntry("1000", 1000);
    glutAddMenuEntry("10000", 10000);
    glutAddMenuEntry("100000", 100000);

    int img_ll_change = glutCreateMenu(img_ll_change_menu_glut);
    glutAddMenuEntry("Noise", 0);

    int obj_change = glutCreateMenu(obj_change_menu_glut);
    glutAddMenuEntry("Phi", 0);
    glutAddMenuEntry("X", 1);
    glutAddMenuEntry("Y", 2);
    glutAddMenuEntry("Z", 3);

    int view_change = glutCreateMenu(view_change_menu_glut);
    glutAddMenuEntry("Zenith", 0);
    glutAddMenuEntry("Focal Length", 1);
    glutAddMenuEntry("Scale", 2);
    glutAddMenuEntry("Near Clipping", 3);
    glutAddMenuEntry("Far Clipping", 4);

    int options = glutCreateMenu(option_menu_glut);
    glutAddMenuEntry("Image", 0);
    glutAddMenuEntry("Edge Points", 1);
    glutAddMenuEntry("Axis", 2);
    glutAddMenuEntry("Wire Proposal", 3);
    glutAddMenuEntry("Wire Model", 4);
    glutAddMenuEntry("Filled Model", 5);
    glutAddMenuEntry("Shaded Model", 6);
    glutAddMenuEntry("Surface LL Assign", 7);
    glutAddMenuEntry("Edge LL Assign", 8);
    glutAddMenuEntry("Threshed Image", 9);
    glutAddMenuEntry("Edge Map Image", 10);
    glutAddMenuEntry("Origin", 11);
    glutAddMenuEntry("Ground", 12);
    glutAddMenuEntry("Object Center", 13);
    glutAddMenuEntry("Image Border", 14);

    int capture = glutCreateMenu(capture_menu_glut);
    glutAddMenuEntry("Full", 0);
    glutAddMenuEntry("Cropped", 1);

    int image = glutCreateMenu(image_menu_glut);
    for (uint32_t i = 0; i < opts.images_in.size(); i++)
    {
        std::ostringstream ost; ost << i+1;
        glutAddMenuEntry(ost.str().c_str(), i);
    }

    int category = glutCreateMenu(category_menu_glut);
    for (uint32_t i = 0; i < opts.num_categories; i++)
    {
        std::ostringstream ost; ost << i+1;
        glutAddMenuEntry(ost.str().c_str(), i);
    }

    glutCreateMenu(main_menu_glut);
    glutAddSubMenu("Sampler", run_sampler);
    glutAddSubMenu("Image", image);
    glutAddSubMenu("Category", category);
    glutAddSubMenu("Image LL", img_ll_change);
    glutAddSubMenu("Object", obj_change);
    glutAddSubMenu("View", view_change);
    glutAddSubMenu("Options", options);
    glutAddSubMenu("Capture", capture);
    glutAddMenuEntry("Write Models", 1);
    glutAddMenuEntry("Quit", 2);

    glutAttachMenu(GLUT_RIGHT_BUTTON);
}


#endif


void Onscreen::enable
(
    int          argc, 
    const char** argv, 
    const char*  title
)
{
#if defined FURNITURE_INTERACTIVE
    init_interactive_modes();
    init_glut(argc, argv, title);
    init_gl();

#if defined FURNITURE_HAVE_OPENGL
    display = glXGetCurrentDisplay();
    drawable = glXGetCurrentDrawable();
    context = glXGetCurrentContext();
#elif defined FURNITURE_HAVE_OPENGL_FRAMEWORK
    context = CGLGetCurrentContext();
#endif

    glut_enabled = true;
#else
    // Should never happen.
    abort();
#endif
}


bool Onscreen::enabled()
{
#if defined FURNITURE_INTERACTIVE
    return glut_enabled;
#else
    // Should never happen.
    abort();
#endif
}


void Onscreen::activate() throw (jwscxx::base::Exception)
{
#if defined FURNITURE_INTERACTIVE
#if defined FURNITURE_HAVE_OPENGL
    assert(glXMakeCurrent(display, 0, 0) == True);
    assert(glXMakeCurrent(display, drawable, context) == True);
#elif defined FURNITURE_HAVE_OPENGL_FRAMEWORK
    CGLSetCurrentContext(context);
#endif
#else
    // Should never happen.
    abort();
#endif
}


void Onscreen::deactivate() throw (jwscxx::base::Exception)
{
#if defined FURNITURE_INTERACTIVE
#if defined FURNITURE_HAVE_OPENGL
    assert(glXMakeCurrent(display, None, 0) == True);
#endif
#else
    // Should never happen.
    abort();
#endif
}


/** This function probably does not return after it is called. */
void Onscreen::start
(
    Sampler*             s,
    Sampler::Parameters* p
)
{
#if defined FURNITURE_INTERACTIVE
    sampler = s;
    params = p;

    quad_fill = gluNewQuadric();
    gluQuadricDrawStyle(quad_fill, GLU_FILL);

    assert(pthread_create(&sampler_thread, 0, run_sampler, 0) == 0);

    Onscreen::activate();

    glutMainLoop();
#else
    // Should never happen.
    abort();
#endif
}
