/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Utility functions and globals used by all furniture related
 * programs.
 */


#ifndef FURNITURE_H
#define FURNITURE_H


#include <config.h>

#include <cstdlib>
#include <list>

#include <inttypes.h>

#if defined JWSCXX_HAVE_OPENGL
#include <GL/gl.h>
#include <GL/glu.h>
#elif defined JWSCXX_HAVE_OPENGL_FRAMEWORK
#include <OpenGL/OpenGL.h>
#endif

#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/image/image.h>
#include <jwsc/image/edge.h>

#include <jwsc++/config.h>
#include <jwsc++/base/option.h>
#include <jwsc++/base/exception.h>

#include "view.h"
#include "object.h"
#include "block.h"
#include "block_set.h"
#include "likelihood.h"
#include "sampler.h"


/** @brief Types of samplers supported. */
typedef enum
{
    METRO,
    LANGE,
    HYPER,
    STOCH,
    HYBRID
}
Sampler_type;


/** @brief Shared program option variables. */
struct Option_vars
{
    /** 
     * @brief Random seed.
     *
     * The random generator is initially seeded (srand) with zero.
     */
    uint32_t seed;

    /** @brief Number of threads to use for execution. */
    uint32_t num_threads;

    /** @brief Type of stochastic dynamics sampler to use. */
    Sampler_type sampler;

    /** @brief Directory to prefix all input (and some output) with. */
    const char* prefix_in_dir;

    /** @brief Directory to prefix all output with. */
    const char* prefix_out_dir;

    /** 
     * @brief Printf-formatted directory to write results to.
     *
     * The last directory should be one printf-formatted integer directory.
     * For example
     *
     * @code
     *    /tmp/prefix/category_dir_fmt/instance_dir_fmt
     * @endcode
     *
     * The results for each set of models fit to an image are prefix to an
     * indexed directory. The directory structure is created by the program.
     */
    const char* instance_dir_fmt;

    /** 
     * @brief Printf-formatted directory to write results to.
     *
     * The last directory should be one printf-formatted integer directory.
     * For example
     *
     * @code
     *    /tmp/prefix/category_dir_fmt/instance_dir_fmt
     * @endcode
     *
     * The results for each set of models fit to an image are prefix to an
     * indexed directory. The directory structure is created by the program.
     */
    const char* category_dir_fmt;

    /** @brief Edge points input file to use (instead of detection). */
    const char* categories_in_fname;

    /** @brief Edge points output file to use. */
    const char* categories_out_fname;

    uint32_t num_cat_samples;

    const char* cat_view_in_fname;

    const char* cat_object_in_fname;

    uint32_t cat_sample_img_width;

    uint32_t cat_sample_img_height;

    const char* cat_sample_img_out_fname;

    /** @brief Whether or not to use a pbuffer for offscreen rendering. */
    bool use_pbuffer;

    /** @brief Whether to run the sampler in interactive mode. */
    bool interactive;

    /** @brief Compress model files. */
    bool compress_models;

    /** @brief Compress information files. */
    bool compress_info;

    /** @brief Sampler information file name. */
    const char* sample_info_fname;

    /** 
     * @brief Printf-formatted file to write interactive captured view images
     * to.
     */
    const char* capture_fmt;

    /** @brief Number of iterations to run the sampler. */
    uint32_t iterations;

    /** @brief Resolution to down-sample the input image to. */
    float resolution;

    /** @brief Sigma for edge detection. */
    float edge_sigma;

    /** @brief Begin threshold for edge detection. */
    float edge_begin;

    /** @brief End threshold for edge detection. */
    float edge_end;

    /** @brief Padding for edge detection. */
    uint32_t edge_pad;

    /** @brief Edge points input file to use (instead of detection). */
    const char* edge_pts_in_fname;

    /** @brief Edge points output file to use. */
    const char* edge_pts_out_fname;

    /** @brief Edge map output image file to use. */
    const char* edge_map_fname;

    /** @brief Colored edge points output image file to use. */
    const char* edge_color_fname;

    /** @brief Padding to use around the rendered model scene. */
    uint32_t model_pad;

    /** 
     * @brief Padding to use around the rendered model scene after resolution.
     */
    uint32_t reso_model_pad;

    /** 
     * @brief Amount of row padding to use on each (resolution changed) image.
     */
    std::vector<uint32_t> images_in_row_pad;

    /** 
     * @brief Amount of column padding to use on each (resolution changed)
     * image.
     */
    std::vector<uint32_t> images_in_col_pad;

    /** @brief Image likelihood distance sigma. */
    double img_ll_dist_sigma;

    /** @brief Image likelihood angle sigma. */
    double img_ll_angle_sigma;

    /** @brief Image likelihood maximum distance to generate correspondences. */
    double img_ll_max_dist;

    /** @brief Image likelihood of a background point in the input image. */
    double img_ll_bg_prob;

    /** @brief Image likelihood of a surface point in the input image. */
    double img_ll_surf_prob;

    /** @brief Image likelihood of a noise surface point in the input image. */
    double img_ll_surf_noise_prob;

    /** 
     * @brief Image likelihood of a missing surface point in the model image.
     */
    double img_ll_surf_miss_prob;

    /** @brief Image likelihood of a noise edge point in the input image. */
    double img_ll_edge_noise_prob;

    /** 
     * @brief Image likelihood of a missing silhouette edge point in the model
     * image.
     */
    double img_ll_edge_silho_miss_prob;

    /** 
     * @brief Image likelihood of a missing inner edge point in the model image.
     */
    double img_ll_edge_inner_miss_prob;

    /** 
     * @brief Image likelihood number of discretized angles for correspondence.
     */
    uint32_t img_ll_num_angles;

    /** 
     * @brief Image likelihood number of discretized angles for correspondence. 
     */
    uint32_t img_ll_cache_size;

    /** 
     * @brief Image likelihood number of discretized angles for correspondence. 
     */
    uint32_t img_ll_cache_k;

    /** 
     * @brief Image likelihood number of discretized angles for correspondence.
     */
    float img_ll_cache_t;

    /** @brief View model should be randomly generated. */
    bool view_in_random;

    /** @brief View model file name to read from. */
    const char* view_in_fname;

    /** @brief View model file name to write to. */
    const char* view_out_fname;

    /** @brief Distance from the View to its near clipping plane. */
    float view_near;

    /** @brief Distance from the View to its far clipping plane. */
    float view_far;

    /** @brief Distance of the View from the origin. */
    float view_distance;

    /** @brief Best view model file name to read from. */
    const char* best_view_in_fname;

    /** @brief Best view model file name to write to. */
    const char* best_view_out_fname;

    /** @brief Object model should be randomly generated. */
    bool object_in_random;

    /** @brief Object model file name to read from. */
    const char* object_in_fname;

    /** @brief Object model file name to write to. */
    const char* object_out_fname;

    /** @brief Object model rendered as edges into an image. */
    const char* object_img_fname;

    /** @brief Best object model file name to read from. */
    const char* best_object_in_fname;

    /** @brief Best object model file name to write to. */
    const char* best_object_out_fname;

    /** @brief Best object model rendered as edges into an image. */
    const char* best_object_img_fname;

    /** @brief File name to read the sample mean vector to. */
    const char* sample_stats_in_fname;

    /** @brief File name to write the sample mean vector to. */
    const char* sample_stats_out_fname;

    /** @brief File name to write the samples to. */
    const char* samples_out_fname;

    /** 
     * @brief Number of eigenvalues to scale in the sample covariance read from
     * the sample stats in file.
     */
    uint32_t sample_cov_in_eig_k;

    /** 
     * @brief Scale for the k largest eigenvalues from the input sample
     * covariance matrix.
     */
    double sample_cov_in_eig_scale;

    /** @brief File to write information about the sample covariance matrix. */
    const char* sample_stats_info_fname;

    /** 
     * @brief Frequency of computing the sample covariance matrix and writing
     * information to the information file. 
     */
    uint32_t sample_stats_freq;

    /** @brief Amount of stochasticity in the dynamics. */
    float stoch_alpha;

    /** 
     * @brief Freqency to kick a partical and reset its momenta to a
     * non-stochastic value.
     */
    uint32_t stoch_kick;

    /** @brief Hybrid Monte Carlo number of leap-frog iterations. */
    uint32_t hybrid_l;

    /** @brief Hyperdynamics gamma parameter. */
    float hyper_gamma;

    /** @brief Hyperdynamics eta parameter. */
    float hyper_eta;

    /** @brief Hyperdynamics bias parameter. */
    float hyper_h;

    /** @brief Hyperdynamics scaling parameter. */
    float hyper_d;

    /** 
     * @brief For Metropolis-Hastings, number of Model_set to use for
     * computing the acceptance probability of a jump move.
     */
    uint32_t mh_jump_set_size;

    double mh_dynamics_prob;

    double mh_diffusion_prob;

    double mh_jump_prob;

    /** @brief Input image. */
    std::vector<jwsc::Image_f*> images_in;

    /** @brief Input image in full resolution. */
    std::vector<jwsc::Image_f*> full_images_in;

    /** @brief Input thresholded image. */
    std::vector<jwsc::Matrix_f*> images_t;

    /** @brief File name to write the thresholded image to. */
    const char* img_thresh_out_fname;

    /** @brief Number of input images. */
    uint32_t num_images;

    /** @brief Number of structure categories. */
    uint32_t num_categories;

    /** @brief Number of structure categories. */
    uint32_t min_num_members;

    /** 
     * @brief Max number of rows in the input images with no resolution change.
     */
    uint32_t images_in_max_num_rows;

    /** 
     * @brief Max number of cols in the input images with no resolution change.
     */
    uint32_t images_in_max_num_cols;

    /** 
     * @brief Max number of rows in the input images after resolution change.
     */
    uint32_t images_reso_max_num_rows;

    /** 
     * @brief Max number of cols in the input images after resolution change.
     */
    uint32_t images_reso_max_num_cols;

    /** @brief Model view width, including padding. */
    uint32_t view_width;

    /** @brief Model view height, including padding. */
    uint32_t view_height;

    /** @brief Model view width at specified resolution, including padding. */
    uint32_t reso_width;

    /** @brief Model view height at specified resolution, including padding. */
    uint32_t reso_height;

    /** @brief Edge set detected in the down-sampled image. */
    std::vector<jwsc::Edge_set_f*> edges;

    /** @brief Context for the parameters in the View. */
    View::Context* view_ctx;

    /** @brief Context for the parameters in the Object. */
    Object::Context* obj_ctx;

    /** @brief Context for the parameters in a block. */
    Block::Context* block_ctx;

    /** @brief Context for the parameters in a Block_set. */
    Block_set::Context* block_set_ctx;
};


extern Option_vars opts;


/** @brief Process the program options and initialize the @em opts variables. */
void process_furniture_options(int argc, const char** argv)
    throw (jwscxx::base::Exception);


/** 
 * @brief Process the @em furniture-cat program options and initialize the @em
 * opts variables. 
 */
void process_furniture_cat_options(int argc, const char** argv)
    throw (jwscxx::base::Exception);


/** @brief Creates and initializes the sampler and its parameters. */
void create_sampler_and_params
(
    Sampler**             sampler_out, 
    Sampler::Parameters** params_out
);


void create_rec_sampler_and_params
(
    Sampler**             sampler_out, 
    Sampler::Parameters** params_out
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);


void read_category(View** view_out, Object** obj_out)
    throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);


/** @brief Writes the model results stored in the sampler parameters. */
void write_results(Sampler::Parameters* params) throw (jwscxx::base::Exception);


/** @brief Frees allocated globals. */
void clean_up_furniture();


/** @brief Frees allocated globals. */
void clean_up_furniture_cat();


/** @brief Compresses a file using zlib. */
void compress(const char* fname) throw (jwscxx::base::Exception);

#endif
