/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Interactive sampler visualization.
 */


#include <config.h>

#include <cstdlib>
#include <cassert>

#include <inttypes.h>

#if defined FURNITURE_HAVE_OPENGL
#include <GL/gl.h>
#include <GL/glu.h>
#elif defined FURNITURE_HAVE_OPENGL_FRAMEWORK
#include <OpenGL/OpenGL.h>
#endif

#include <jwsc/image/edge.h>

#include "view.h"
#include "object.h"
#include "options.h"
#include "render.h"


/**
 * @param  view  Model view fit to the image.
 * @param  obj   Object to render.
 */
void Render::wire_object(const View* view, const Object* obj)
{
    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();

    view->set_gl_projection();
    view->set_gl_modelview();

    glClear(GL_DEPTH_BUFFER_BIT);
    obj->render();

    glPopMatrix();
}


/**
 * @param  view  Model view fit to the image.
 * @param  obj   Object to render.
 */
void Render::solid_object(const View* view, const Object* obj)
{
    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();

    view->set_gl_projection();
    view->set_gl_modelview();

    obj->solid_render();

    glPopMatrix();
}


/**
 * @param  i     Image (model) instance.
 * @param  view  Model view fit to the image.
 */
void Render::image(uint32_t i, const View* view)
{
    const jwsc::Image_f* img = opts.images_in[ i ];

    glDisable(GL_DEPTH_TEST);

    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();
    glLoadIdentity();

    glClearColor(1, 1, 1, 0);
    glClear(GL_COLOR_BUFFER_BIT);

    uint32_t num_rows = img->num_rows;
    uint32_t num_cols = img->num_cols;

    float xtr_row = (num_rows % 2) ? 0.5 : 0.5;
    float xtr_col = (num_cols % 2) ? 0.5 : 0.5;

    for (uint32_t row = 0; row < num_rows; row++)
    {
        for (uint32_t col = 0; col < num_cols; col++)
        {
            float r = img->pxls[ row ][ col ].r;
            float g = img->pxls[ row ][ col ].g;
            float b = img->pxls[ row ][ col ].b;

            glColor3f(r, g, b);

            glBegin(GL_POINTS);
            glVertex3f((col - (num_cols/2.0f) + xtr_col), 
                    -(row - (num_rows/2.0f) + xtr_row), 
                    -view->focal()*opts.resolution);
            glEnd();
        }
    }

    glPopMatrix();

    glEnable(GL_DEPTH_TEST);
}


/**
 * @param  i     Image (model) instance.
 * @param  view  Model view fit to the image.
 */
void Render::full_image(uint32_t i, const View* view)
{
    const jwsc::Image_f* img = opts.full_images_in[ i ];

    glDisable(GL_DEPTH_TEST);

    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();
    glLoadIdentity();

    glClearColor(1, 1, 1, 0);
    glClear(GL_COLOR_BUFFER_BIT);

    uint32_t num_rows = img->num_rows;
    uint32_t num_cols = img->num_cols;

    float xtr_row = (num_rows % 2) ? 0.5 : 0.5;
    float xtr_col = (num_cols % 2) ? 0.5 : 0.5;

    for (uint32_t row = 0; row < num_rows; row++)
    {
        for (uint32_t col = 0; col < num_cols; col++)
        {
            float r = img->pxls[ row ][ col ].r;
            float g = img->pxls[ row ][ col ].g;
            float b = img->pxls[ row ][ col ].b;

            glColor3f(r, g, b);

            glBegin(GL_POINTS);
            glVertex3f((col - (num_cols/2.0f) + xtr_col), 
                    -(row - (num_rows/2.0f) + xtr_row), 
                    -view->focal());
            glEnd();
        }
    }

    glPopMatrix();

    glEnable(GL_DEPTH_TEST);
}


/**
 * @param  i     Image (model) instance.
 * @param  view  Model view fit to the image.
 */
void Render::thresh_image(uint32_t i, const View* view)
{
    const jwsc::Matrix_f* img_t = opts.images_t[ i ];

    glDisable(GL_DEPTH_TEST);

    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();
    glLoadIdentity();

    glClearColor(1, 1, 1, 0);
    glClear(GL_COLOR_BUFFER_BIT);

    uint32_t num_rows = img_t->num_rows;
    uint32_t num_cols = img_t->num_cols;

    float xtr_row = (num_rows % 2) ? 0.5 : 0.5;
    float xtr_col = (num_cols % 2) ? 0.5 : 0.5;

    for (uint32_t row = 0; row < num_rows; row++)
    {
        for (uint32_t col = 0; col < num_cols; col++)
        {
            float r = img_t->elts[ row ][ col ];
            float g = img_t->elts[ row ][ col ];
            float b = img_t->elts[ row ][ col ];

            glColor3f(r, g, b);

            glBegin(GL_POINTS);
            glVertex3f((col - (num_cols/2.0f) + xtr_col), 
                    -(row - (num_rows/2.0f) + xtr_row), 
                    -view->focal()*opts.resolution);
            glEnd();
        }
    }

    glPopMatrix();

    glEnable(GL_DEPTH_TEST);
}


/**
 * @param  i     Image (model) instance.
 * @param  view  Model view fit to the image.
 */
void Render::edges(uint32_t i, const View* view)
{
    const jwsc::Edge_set_f* edges = opts.edges[ i ];

    glDisable(GL_DEPTH_TEST);

    glMatrixMode(GL_MODELVIEW);

    glPushMatrix();
    glLoadIdentity();

    uint32_t num_rows = opts.images_in[ i ]->num_rows;
    uint32_t num_cols = opts.images_in[ i ]->num_cols;

    uint32_t r = opts.images_in_row_pad[ i ];
    uint32_t c = opts.images_in_col_pad[ i ];

    float odd_row = ((opts.images_reso_max_num_rows - num_rows) % 2) ? 0 : 0.5;
    float odd_col = ((opts.images_reso_max_num_cols - num_cols) % 2) ? 0 : 0.5;

    glBegin(GL_POINTS);
    for (uint32_t e = 0; e < edges->num_edges; e++)
    {
        for (uint32_t p = 0; p < edges->num_pts[ e ]; p++)
        {
            uint32_t row = edges->pts[ e ][ p ].row - r;
            uint32_t col = edges->pts[ e ][ p ].col - c;

            float x =  (float)col + odd_col - ((float)num_cols / 2.0f);
            float y = -(float)row - odd_row + ((float)num_rows / 2.0f);
            float z = -view->focal()*opts.resolution;

            glVertex3f(x, y, z);
        }
    }
    glEnd();

    glPopMatrix();

    glEnable(GL_DEPTH_TEST);
}
