/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief MCMC sampler.
 */


#include <config.h>

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <list>
#include <cassert>

#include <inttypes.h>
#include <pthread.h>

#include <jwsc/math/constants.h>
#include <jwsc/vector/vector.h>
#include <jwsc/vector/vector_math.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/matrix/matrix_math.h>
#include <jwsc/prob/pdf.h>
#include <jwsc/stat/stat.h>
#include <jwsc/image/image.h>
#include <jwsc/image/edge.h>

#include <jwsc++/base/exception.h>

#include "view.h"
#include "object.h"
#include "block.h"
#include "block_set.h"
#include "likelihood.h"
#include "sampler.h"


extern void compress(const char* fname) throw (jwscxx::base::Exception);


/* --------------------------------  Sampler  ------------------------------- */

Sampler::~Sampler()
{
}

/* ========================================================================== */






/* ----------------------  Sampler::Thread_manager  ------------------------- */

Sampler::Thread_manager::Thread_manager(uint32_t thread)
{
    _thread = thread;
}

Sampler::Thread_manager::~Thread_manager() 
{ 
}

void Sampler::Thread_manager::thread(uint32_t thread)
    throw (jwscxx::base::Arg_error)
{
    if (thread >= _num_threads)
    {
        throw jwscxx::base::Arg_error("Thread index too large");
    }
    _thread = thread;
}

uint32_t Sampler::Thread_manager::thread() const
{
    return _thread;
}

uint32_t Sampler::Thread_manager::num_threads() const
{
    return _num_threads;
}

void Sampler::Thread_manager::num_threads(uint32_t n)
{
    _num_threads = n;
}

Sampler* Sampler::Thread_manager::sampler() const
{
    return _sampler;
}

Sampler::Parameters* Sampler::Thread_manager::params() const
{
    return _params;
}

void Sampler::Thread_manager::iterations(uint32_t i)
{
    _iterations = i;
}

uint32_t Sampler::Thread_manager::iterations() const
{
    return _iterations;
}

void Sampler::Thread_manager::toggle_params_lock()
{
    synchronize();

    if (last_thread())
    {
        params()->unlock();
        params()->lock();
        release();
    }
    else
    {
        wait();
    }

    desynchronize();
}

void Sampler::Thread_manager::lock_sampler_params()
{
    synchronize();

    if (last_thread())
    {
        params()->lock();
        release();
    }
    else
    {
        wait();
    }

    desynchronize();
}

void Sampler::Thread_manager::unlock_sampler_params()
{
    synchronize();

    if (last_thread())
    {
        params()->unlock();
        release();
    }
    else
    {
        wait();
    }

    desynchronize();
}

void Sampler::Thread_manager::synchronize()
{
    assert(pthread_mutex_lock(mux()) == 0);
}

bool Sampler::Thread_manager::last_thread() const
{
    return last();
}

void Sampler::Thread_manager::release()
{
    begin_waiting();
    assert(pthread_mutex_unlock(mux()) == 0);
}

void Sampler::Thread_manager::wait()
{
    begin_waiting();
    assert(pthread_mutex_unlock(mux()) == 0);

    while (busy_wait_1()) { }
}

void Sampler::Thread_manager::desynchronize()
{
    assert(pthread_mutex_lock(mux()) == 0);
    done_waiting();
    assert(pthread_mutex_unlock(mux()) == 0);

    while (busy_wait_2()) { }
}

bool Sampler::Thread_manager::last() const
{
    return ((_num_threads - _num_threads_waiting) == 1);
}

inline bool Sampler::Thread_manager::busy_wait_1() const
{
    return !_synchronized;
}

inline bool Sampler::Thread_manager::busy_wait_2() const
{
    return _synchronized;
}

void Sampler::Thread_manager::begin_waiting()
{
    assert(_num_threads > _num_threads_waiting);
    _num_threads_waiting++;

    if (_num_threads == _num_threads_waiting)
    {
        _synchronized = true;
    }
}

void Sampler::Thread_manager::done_waiting()
{
    assert(_num_threads_waiting > 0);
    _num_threads_waiting--;

    if (_num_threads_waiting == 0)
    {
        _synchronized = false;
    }
}

pthread_mutex_t* Sampler::Thread_manager::mux()
{
    return &_mux;
}

Sampler* Sampler::Thread_manager::_sampler = 0;

Sampler::Parameters* Sampler::Thread_manager::_params = 0;

uint32_t Sampler::Thread_manager::_iterations = 0;

uint32_t Sampler::Thread_manager::_num_threads = 1;

uint32_t Sampler::Thread_manager::_num_threads_waiting = 0;

bool Sampler::Thread_manager::_synchronized = false;

pthread_mutex_t Sampler::Thread_manager::_mux = PTHREAD_MUTEX_INITIALIZER;

/* ========================================================================== */








/* -------------------------  Sampler::Instance  ---------------------------- */

/**
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 * @throw  jwscxx::base::IO_error   Could not open info file.
 */
Sampler::Instance::Instance
(
    View*                       best_view,
    Object*                     best_obj,
    double                      best_img_ll,
    double                      best_potential,
    View*                       view,
    Object*                     obj,
    double                      img_ll,
    double                      potential,
    const jwsc::Matrix_f*       image_t,
    const jwsc::Edge_set_f*     edges,
    Likelihood::Correspondence* img_ll_corr,
    Likelihood::Cache*          img_ll_cache,
    jwsc::Matrix_u32*           img_ll_surf_assign,
    jwsc::Matrix_u32*           img_ll_edge_assign,
    jwsc::Matrix_u32*           model_mat,
    jwsc::Vector_d*             sample_mean_in,
    jwsc::Matrix_d*             sample_cov_in,
    const char*                 sample_info_fname,
    bool                        compress_sample_info,
    const char*                 sample_stats_info_fname
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
{
    this->best_view         = best_view;
    this->best_obj          = best_obj;
    this->best_img_ll       = best_img_ll;
    this->best_ll_potential = best_potential;
    this->view              = view;
    this->obj               = obj;
    this->img_ll            = img_ll;
    this->ll_potential      = potential;
    this->ll_kinetic        = 0;
    this->image_t           = image_t;
    this->edges             = edges;
    this->img_ll_corr       = img_ll_corr;
    this->img_ll_cache      = img_ll_cache;
    this->img_ll_surf_assign = img_ll_surf_assign;
    this->img_ll_edge_assign = img_ll_edge_assign;
    this->model_mat         = model_mat;
    this->sample_mean_in    = sample_mean_in;
    this->sample_cov_in     = sample_cov_in;
    this->num_samples       = 0;
    this->sample_mean       = 0;
    this->sample_cov        = 0;
    this->sample_cov_eigval = 0;
    this->sample_cov_eigvec = 0;
    this->sample_info_fname = sample_info_fname;
    this->compress_sample_info = compress_sample_info;

    sample_info.open(sample_info_fname);
    if (sample_info.fail())
    {
        std::ostringstream ost;
        ost << sample_info_fname << ": Could not open sample info file";
        throw jwscxx::base::IO_error(ost.str());
    }

    sample_stats_info.open(sample_stats_info_fname);
    if (sample_stats_info.fail())
    {
        std::ostringstream ost;
        ost << sample_stats_info_fname << 
            ": Could not open sample stats info file";
        throw jwscxx::base::IO_error(ost.str());
    }

    assert(pthread_mutex_init(&mux, 0) == 0);

    iteration = 0;
}


Sampler::Instance::~Instance()
{
    using namespace jwsc;

    delete best_view;
    delete best_obj;
    delete view;
    delete obj;
    delete img_ll_corr;
    delete img_ll_cache;

    free_matrix_u32(img_ll_surf_assign);
    free_matrix_u32(img_ll_edge_assign);
    free_matrix_u32(model_mat);

    free_vector_d(sample_mean_in);
    free_matrix_d(sample_cov_in);

    std::list<jwsc::Vector_d*>::iterator it;
    for (it = samples.begin(); it != samples.end(); it++)
    {
        jwsc::free_vector_d(*it);
    }

    free_vector_d(sample_mean);
    free_matrix_d(sample_cov);
    free_vector_d(sample_cov_eigval);
    free_matrix_d(sample_cov_eigvec);

    sample_info.close();
    sample_stats_info.close();

    if (compress_sample_info)
    {
        try { compress(sample_info_fname.c_str()); }
        catch (jwscxx::base::Exception e)
        { 
            std::cerr << "furniture: " << sample_info_fname << ": " 
                      << e.get_msg() << std::endl; 
        }
    }
}


uint32_t Sampler::Instance::num_var_params() const
{
    return view->num_var_params() + obj->num_var_params();
}


/** TODO need to be able to update sample stats when samples have varying
 * dimensions!!! */
void Sampler::Instance::update_sample_stats()
{
    using namespace jwsc;

    Vector_d* new_sample_mean = 0;
    Matrix_d* new_sample_cov  = 0;
    Vector_d* new_sample_cov_eigval = 0;
    Matrix_d* new_sample_cov_eigvec = 0;

    assert(num_samples > 0);

    // TODO !!!
    return;

    calc_sample_stats(&new_sample_mean, &new_sample_cov);
    assert(solve_real_symmetric_matrix_eigenproblem_d(
                &new_sample_cov_eigvec, &new_sample_cov_eigval, 
                new_sample_cov) == 0);

    uint32_t D = new_sample_cov_eigval->num_elts;

    for (uint32_t i = 0; i < D; i++)
    {
        sample_stats_info << new_sample_cov_eigval->elts[ i ] << " ";
    }

    double norm = 0;

    if (sample_cov)
    {
        calc_frobenius_norm_abs_diff_d(&norm, sample_cov, new_sample_cov);
    }
    sample_stats_info << norm << " ";

    if (sample_cov_eigvec)
    {
        calc_frobenius_norm_abs_diff_d(&norm, sample_cov_eigvec, 
                new_sample_cov_eigvec);
    }
    sample_stats_info << norm << " ";

    // Print out the three largest principle components of the covariance
    // matrix. TODO we assume the last (right) three columns are the principle
    // components.
    for (uint32_t j = D-3; j < D; j++)
    {
        for (uint32_t i = 0; i < D; i++)
        {
            sample_stats_info << new_sample_cov_eigvec->elts[ i ][ j ] << " ";
        }
    }

    sample_stats_info << std::endl;
    sample_stats_info.flush();

    free_vector_d(sample_mean);
    free_matrix_d(sample_cov);
    free_vector_d(sample_cov_eigval);
    free_matrix_d(sample_cov_eigvec);

    sample_mean = new_sample_mean;
    sample_cov  = new_sample_cov;
    sample_cov_eigval = new_sample_cov_eigval;
    sample_cov_eigvec = new_sample_cov_eigvec;
}


/** 
 * Unbiased MLE.
 *
 * @param  mean_out Sample mean.
 * @param  cov_out  Sample covariance matrix.
 */
void Sampler::Instance::calc_sample_stats
(
    jwsc::Vector_d** mean_out,
    jwsc::Matrix_d** cov_out
)
const throw (jwscxx::base::Result_error)
{
    using namespace jwsc;

    uint32_t N = num_samples;
    uint32_t D = num_var_params();
    Matrix_d* samples_mat = 0;

    if (N == 0)
    {
        free_vector_d(*mean_out); *mean_out = 0;
        free_matrix_d(*cov_out); *cov_out = 0;
        throw jwscxx::base::Result_error("No samples to compute stats");
    }

    create_matrix_d(&samples_mat, N, D);

    uint32_t n = 0;
    std::list<jwsc::Vector_d*>::const_iterator it;
    for (it = samples.begin(); it != samples.end(); it++)
    {
        for (uint32_t d = 0; d < D; d++)
        {
            samples_mat->elts[ n ][ d ] = (*it)->elts[ d ];
        }

        // NOTE: We are using the absolute value of the phi angle of the obj
        // because it is symmetric.
        //uint32_t p = view->num_var_params() + Object::p_phi;
        //samples_mat->elts[ n ][ p ] = fabs(samples_mat->elts[ n ][ p ]);

        n++;
    }

    mv_sample_stats_d(mean_out, cov_out, samples_mat);
    free_matrix_d(samples_mat);
}

void Sampler::Instance::print_var_param_info()
{
    print_var_param_info(sample_info);
}

void Sampler::Instance::print_var_param_info(std::ostream& out)
{
    out.precision(5);
    out.setf(std::ios_base::scientific, std::ios_base::floatfield);

    out << iteration << ',' << "Image Likelihood" << ','
        << img_ll << std::endl;

    out << iteration << ',' << "Likelihood Potential Energy" << ',' 
        << ll_potential << std::endl;

    out << iteration << ',' << "Likelihood Kinetic Energy" << ','
        << ll_kinetic << std::endl;

    out << iteration << ',' << "Total Likelihood Energy" << ','
        << ll_potential + ll_kinetic << std::endl;

    out << iteration << ',' << "Number of Components" << ','
        << obj->num_components() << std::endl;

    out.precision(4);
    out.setf(std::ios_base::fixed, std::ios_base::floatfield);

    view->print_var_param_values(iteration, out);
    obj->print_var_param_values(iteration, out);

    /*
    out << iteration << ',' << "Cache Hits" << ','
        << img_ll_cache->get_num_hits() << std::endl;

    out << iteration << ',' << "Percent Cache Hits" << ','
        << img_ll_cache->get_percent_hits() << std::endl;

    out << iteration << ',' << "Cache Size" << ','
        << img_ll_cache->get_cache_size() << std::endl;
        */

    out.flush();
}


/**
 * @param  p_out  Point to get the model parameters into.
 */
void Sampler::Instance::get_var_param_values(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    uint32_t D_1 = view->num_var_params();
    uint32_t D_2 = obj->num_var_params();
    uint32_t D   = num_var_params();

    Vector_d* p;
    create_vector_d(p_out, D);
    p = *p_out;

    Vector_d* q = 0;

    view->get_var_param_values(&q);
    assert(copy_vector_section_into_vector_d(p, 0, q, 0, D_1) == 0);

    obj->get_var_param_values(&q);
    assert(copy_vector_section_into_vector_d(p, D_1, q, 0, D_2) == 0);

    free_vector_d(q);
}


/**
 * @param  p  Point to set the model parameters from.
 *
 * @note Doesn't update the log likelihood or energy. They must be set
 * separately.
 */
void Sampler::Instance::set_var_param_values(const jwsc::Vector_d* p)
{
    using namespace jwsc;

    uint32_t D_1 = view->num_var_params();
    uint32_t D_2 = obj->num_var_params();

    Vector_d* q = 0;

    copy_vector_section_d(&q, p, 0, D_1);
    try { view->set_var_param_values(q); }
    catch (jwscxx::base::Arg_error e) { }

    copy_vector_section_d(&q, p, D_1, D_2);
    try { obj->set_var_param_values(q); }
    catch (jwscxx::base::Arg_error e) { }

    free_vector_d(q);
}


/**
 * Shifts the dependents of each component when modifying its parameters.
 *
 * @param  p  Point to add to the the parameters.
 *
 * @note Doesn't update the log likelihood or energy. They must be set
 * separately.
 */
void Sampler::Instance::add_var_param_values_shift_component_deps
(
    const jwsc::Vector_d* p
)
{
    using namespace jwsc;

    uint32_t D_1 = view->num_var_params();
    uint32_t D_2 = obj->num_var_params();

    Vector_d* q = 0;

    copy_vector_section_d(&q, p, 0, D_1);
    try { view->add_var_param_values(q); }
    catch (jwscxx::base::Arg_error e) { }

    copy_vector_section_d(&q, p, D_1, D_2);
    try { obj->add_var_param_values_shift_component_deps(q); }
    catch (jwscxx::base::Arg_error e) { }

    free_vector_d(q);
}


/**
 * Scales the dependents of each component when modifying its parameters.
 *
 * @param  p  Point to add to the the parameters.
 */
void Sampler::Instance::add_var_param_values_scale_component_deps
(
    const jwsc::Vector_d* p
)
{
    using namespace jwsc;

    uint32_t D_1 = view->num_var_params();
    uint32_t D_2 = obj->num_var_params();

    Vector_d* q = 0;

    copy_vector_section_d(&q, p, 0, D_1);
    try { view->add_var_param_values(q); }
    catch (jwscxx::base::Arg_error e) { }

    copy_vector_section_d(&q, p, D_1, D_2);
    try { obj->add_var_param_values_scale_component_deps(q); }
    catch (jwscxx::base::Arg_error e) { }

    free_vector_d(q);
}

/**
 * @param  delta_t_out   Stochastic dynamics delta_t values for the parameters.
 */
void Sampler::Instance::get_var_param_values_dynamics_delta_t
(
    jwsc::Vector_d** delta_t_out
) 
const
{
    using namespace jwsc;

    uint32_t D_1 = view->num_var_params();
    uint32_t D_2 = obj->num_var_params();
    uint32_t D   = num_var_params();

    Vector_d* delta_t;
    create_vector_d(delta_t_out, D);
    delta_t = *delta_t_out;

    Vector_d* q = 0;

    view->get_var_param_values_dynamics_delta_t(&q);
    assert(copy_vector_section_into_vector_d(delta_t, 0, q, 0, D_1) == 0);

    obj->get_var_param_values_dynamics_delta_t(&q);
    assert(copy_vector_section_into_vector_d(delta_t, D_1, q, 0, D_2) == 0);

    free_vector_d(q);
}

/**
 * @param  p_out  Point to get the model parameters into.
 */
void Sampler::Instance::get_var_param_stats(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    uint32_t D_1 = view->num_var_params();
    uint32_t D_2 = obj->num_var_params();
    uint32_t D   = num_var_params();

    Vector_d* p;
    create_vector_d(p_out, 2*D);
    p = *p_out;

    Vector_d* m_q = 0;
    Vector_d* s_q = 0;

    view->get_var_param_stats(&m_q, &s_q);
    assert(copy_vector_section_into_vector_d(p, 0, m_q, 0, D_1) == 0);
    assert(copy_vector_section_into_vector_d(p, D_1, s_q, 0, D_1) == 0);

    obj->get_var_param_stats(&m_q, &s_q);
    assert(copy_vector_section_into_vector_d(p, 2*D_1, m_q, 0, D_2) == 0);
    assert(copy_vector_section_into_vector_d(p, 2*D_1+D_2, s_q, 0, D_2) == 0);

    free_vector_d(m_q);
    free_vector_d(s_q);
}

/**
 * @param  p  Point to set the model parameters from.
 *
 * @note Doesn't update the log likelihood or energy. They must be set
 * separately.
 */
void Sampler::Instance::set_var_param_stats(const jwsc::Vector_d* p)
{
    using namespace jwsc;

    uint32_t D_1 = view->num_var_params();
    uint32_t D_2 = obj->num_var_params();

    Vector_d* m_q = 0;
    Vector_d* s_q = 0;

    copy_vector_section_d(&m_q, p, 0, D_1);
    copy_vector_section_d(&s_q, p, D_1, D_1);
    try { view->set_var_param_stats(m_q, s_q); }
    catch (jwscxx::base::Arg_error e) { }

    copy_vector_section_d(&m_q, p, 2*D_1, D_2);
    copy_vector_section_d(&s_q, p, 2*D_1+D_2, D_2);
    try { obj->set_var_param_stats(m_q, s_q); }
    catch (jwscxx::base::Arg_error e) { }

    free_vector_d(m_q);
    free_vector_d(s_q);
}

/**
 * @param  delta_t_out   Stochastic dynamics delta_t values for the parameters.
 */
void Sampler::Instance::get_var_param_stats_dynamics_delta_t
(
    jwsc::Vector_d** delta_t_out
) 
const
{
    using namespace jwsc;

    uint32_t D_1 = view->num_var_params();
    uint32_t D_2 = obj->num_var_params();
    uint32_t D   = num_var_params();

    Vector_d* delta_t;
    create_vector_d(delta_t_out, 2*D);
    delta_t = *delta_t_out;

    Vector_d* m_q = 0;
    Vector_d* s_q = 0;

    view->get_var_param_stats_dynamics_delta_t(&m_q, &s_q);
    copy_vector_section_into_vector_d(delta_t, 0, m_q, 0, D_1);
    copy_vector_section_into_vector_d(delta_t, D_1, s_q, 0, D_1);

    obj->get_var_param_stats_dynamics_delta_t(&m_q, &s_q);
    copy_vector_section_into_vector_d(delta_t, 2*D_1, m_q, 0, D_2);
    copy_vector_section_into_vector_d(delta_t, 2*D_1+D_2, s_q, 0, D_2);

    free_vector_d(m_q);
    free_vector_d(s_q);
}

/* -------------------------------------------------------------------------- */







/* -------------------------  Sampler::Category  ---------------------------- */

/**
 * @param  rep_ins            Representative instance for the category.
 * @param  lp_potential       Log prior potential energy.
 * @param  potential          Potential energy of the likelihood and prior.
 * @param  sample_info_fname  File to write sample information to.
 *
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 * @throw  jwscxx::base::IO_error   Could not open info file.
 */
Sampler::Category::Category
(
    const Instance* rep_ins,
    double          lp_potential,
    double          potential,
    const char*     sample_info_fname,
    bool            compress_sample_info
    )
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
{
    best_view    = rep_ins->best_view;
    best_obj     = rep_ins->best_obj;
    view         = rep_ins->view;
    obj          = rep_ins->obj;
    _iteration   = &(rep_ins->iteration);

    this->lp_potential = lp_potential;
    this->lp_kinetic   = 0;
    this->potential    = potential;
    this->kinetic      = 0;
    this->sample_info_fname = sample_info_fname;
    this->compress_sample_info = compress_sample_info;

    sample_info.open(sample_info_fname);
    if (sample_info.fail())
    {
        std::ostringstream ost;
        ost << sample_info_fname << ": Could not open sample info file";
        throw jwscxx::base::IO_error(ost.str());
    }
}

Sampler::Category::~Category()
{
    sample_info.close();

    if (compress_sample_info)
    {
        try { compress(sample_info_fname.c_str()); }
        catch (jwscxx::base::Exception e)
        { 
            std::cerr << "furniture: " << sample_info_fname << ": " 
                      << e.get_msg() << std::endl; 
        }
    }
}

size_t Sampler::Category::iteration() const
{
    return *_iteration;
}

uint32_t Sampler::Category::num_var_params() const
{
    return view->num_var_params() + obj->num_var_params();
}

void Sampler::Category::print_var_param_info()
{
    print_var_param_info(sample_info);
}

void Sampler::Category::print_var_param_info(std::ostream& out)
{
    double lp = 0;
    lp += view->calc_var_param_prior_log_prob();
    lp += obj->calc_var_param_prior_log_prob();

    out.precision(5);
    out.setf(std::ios_base::scientific, std::ios_base::floatfield);

    out << iteration() << ',' << "Prior Potential Energy" << ',' 
        << lp_potential << std::endl;

    out << iteration() << ',' << "Prior Kinetic Energy" << ','
        << lp_kinetic << std::endl;

    out << iteration() << ',' << "Total Prior Energy" << ','
        << lp_potential + lp_kinetic << std::endl;

    out << iteration() << ',' << "Potential Energy" << ',' 
        << potential << std::endl;

    out << iteration() << ',' << "Kinetic Energy" << ','
        << kinetic << std::endl;

    out << iteration() << ',' << "Total Energy" << ','
        << potential + kinetic << std::endl;

    out.precision(4);
    out.setf(std::ios_base::fixed, std::ios_base::floatfield);

    view->print_var_param_stats(iteration(), out);
    obj->print_var_param_stats(iteration(), out);

    out.flush();
}

/* -------------------------------------------------------------------------- */








/* ------------------------  Sampler::Parameters  --------------------------- */

Sampler::Parameters::Parameters
(
    const View::Context*      view_ctx,
    const Object::Context*    obj_ctx,
    const Block::Context*     block_ctx,
    const Block_set::Context* block_set_ctx,
    double                    img_ll_bg_prob,
    double                    img_ll_fg_prob,
    double                    img_ll_fg_noise_prob,
    double                    img_ll_fg_miss_prob,
    double                    img_ll_edge_noise_prob,
    double                    img_ll_edge_silho_miss_prob,
    double                    img_ll_edge_inner_miss_prob,
    uint32_t                  sample_stats_freq,
    uint32_t                  num_instances,
    uint32_t                  num_categories,
    uint32_t                  min_num_members,
    uint32_t                  num_threads
)
: _instances(num_categories, std::vector<Instance*>(num_instances, 0)),
  _categories(num_categories, 0)
{
    using namespace jwsc;

    assert(num_instances);
    assert(num_categories);

    _view_ctx          = view_ctx;
    _obj_ctx           = obj_ctx;
    _block_ctx         = block_ctx;
    _block_set_ctx     = block_set_ctx;
    _img_ll_bg_prob    = img_ll_bg_prob;
    _img_ll_fg_prob    = img_ll_fg_prob;
    _img_ll_fg_noise_prob  = img_ll_fg_noise_prob;
    _img_ll_fg_miss_prob   = img_ll_fg_miss_prob;
    _img_ll_edge_noise_prob = img_ll_edge_noise_prob;
    _img_ll_edge_silho_miss_prob = img_ll_edge_silho_miss_prob;
    _img_ll_edge_inner_miss_prob = img_ll_edge_inner_miss_prob;
    _sample_stats_freq = sample_stats_freq;
    _min_num_members   = min_num_members;
    _num_threads       = num_threads;

    assert(pthread_mutex_init(&_mux, 0) == 0);

    _cat_members = 0;
    create_zero_matrix_d(&_cat_members, num_categories, num_instances);
}


Sampler::Parameters::~Parameters()
{
    for (uint32_t c = 0; c < num_categories(); c++)
    {
        for (uint32_t i = 0; i < num_instances(); i++)
        {
            delete _instances[ c ][ i ];
        }
        delete _categories[ c ];
    }

    jwsc::free_matrix_d(_cat_members);
}


void Sampler::Parameters::lock()
{
    assert(pthread_mutex_lock(&_mux) == 0);

    for (uint32_t c = 0; c < num_categories(); c++)
    {
        for (uint32_t i = 0; i < num_instances(); i++)
        {
            lock(c, i);
        }
    }
}


void Sampler::Parameters::unlock()
{
    for (uint32_t c = 0; c < num_categories(); c++)
    {
        for (uint32_t i = 0; i < num_instances(); i++)
        {
            unlock(c, i);
        }
    }

    assert(pthread_mutex_unlock(&_mux) == 0);
}


void Sampler::Parameters::lock(uint32_t c, uint32_t i) 
{
    assert(pthread_mutex_lock(&(instance(c, i)->mux)) == 0);
}


void Sampler::Parameters::unlock(uint32_t c, uint32_t i) 
{
    assert(pthread_mutex_unlock(&(instance(c, i)->mux)) == 0);
}

/**
 * @param  c  Index of the category.
 * @param  i  Index of the Instance to get.
 *
 * @return A pointer to the indexed Instance.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Instance index.
 */
Sampler::Instance* Sampler::Parameters::instance
(
    uint32_t c, 
    uint32_t i
) 
const throw (jwscxx::base::Arg_error)
{
    if (i >= num_instances())
    {
        throw jwscxx::base::Arg_error("Model set index too large");
    }

    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    return _instances[ c ][ i ];
}

void Sampler::Parameters::instance
(
    Sampler::Instance* ins, 
    uint32_t           c,
    uint32_t           i
) 
throw (jwscxx::base::Arg_error)
{
    if (i >= num_instances())
    {
        throw jwscxx::base::Arg_error("Model instance index too large");
    }

    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    _instances[ c ][ i ] = ins;
}

/**
 * @param  c  Index of the category.
 *
 * @return A pointer to the indexed Category.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Category index.
 */
Sampler::Category* Sampler::Parameters::category(uint32_t c) const 
    throw (jwscxx::base::Arg_error)
{
    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    return _categories[ c ];
}

void Sampler::Parameters::category
(
    Sampler::Category* cat,
    uint32_t           c
) 
throw (jwscxx::base::Arg_error)
{
    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    _categories[ c ] = cat;
}

double Sampler::Parameters::member(uint32_t c, uint32_t i) const
    throw (jwscxx::base::Arg_error)
{
    if (i >= num_instances())
    {
        throw jwscxx::base::Arg_error("Model instance index too large");
    }

    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    return _cat_members->elts[ c ][ i ];
}


void Sampler::Parameters::member(double member, uint32_t c, uint32_t i)
    throw (jwscxx::base::Arg_error)
{
    if (i >= num_instances())
    {
        throw jwscxx::base::Arg_error("Model instance index too large");
    }

    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    _cat_members->elts[ c ][ i ] = member;
}


void Sampler::Parameters::member(const jwsc::Vector_d* member, uint32_t i)
    throw (jwscxx::base::Arg_error)
{
    if (i >= num_instances())
    {
        throw jwscxx::base::Arg_error("Model instance index too large");
    }

    if (member->num_elts > num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    for (uint32_t c = 0; c < num_categories(); c++)
    {
        this->member(member->elts[ c ], c, i);
    }
}


void Sampler::Parameters::member(jwsc::Vector_d** member_out, uint32_t i) const
    throw (jwscxx::base::Arg_error)
{
    if (i >= num_instances())
    {
        throw jwscxx::base::Arg_error("Model instance index too large");
    }

    create_vector_d(member_out, num_categories());

    for (uint32_t c = 0; c < num_categories(); c++)
    {
        (*member_out)->elts[ c ] = _cat_members->elts[ c ][ i ];
    }
}


uint32_t Sampler::Parameters::num_categories() const
{
    return _instances.size();
}


uint32_t Sampler::Parameters::num_instances() const
{
    return _instances[0].size();
}


uint32_t Sampler::Parameters::num_var_params(uint32_t c, uint32_t i) const
{
    return instance(c, i)->num_var_params();
}


uint32_t Sampler::Parameters::num_var_params(uint32_t c) const
{
    uint32_t N_i = num_instances();
    uint32_t D_i = 0;

    for (uint32_t i = 0; i < N_i; i++)
    {
        D_i += num_var_params(c, i);
    }

    assert(D_i == N_i*num_var_params(c, 0));

    return D_i;
}

/**
 * @param  i  Instance to sum the category parameters across.
 */
uint32_t Sampler::Parameters::num_cat_var_params(uint32_t i) const
{
    uint32_t N_c = num_categories();
    uint32_t D_c = 0;

    for (uint32_t c = 0; c < N_c; c++)
    {
        D_c += num_var_params(c, i);
    }

    return D_c;
}

void Sampler::Parameters::assign_random_members()
{
    using namespace jwsc;

    uint32_t N_c = num_categories();
    uint32_t N_i = num_instances();

    create_zero_matrix_d(&_cat_members, N_c, N_i);

    for (uint32_t i = 0; i < N_i; i++)
    {
        uint32_t u = (uint32_t)floor(sample_uniform_pdf_d(0, N_c));
        if (u == N_c) u--;

        _cat_members->elts[ u ][ i ] = 1;
    }

    Vector_u32* num_members = 0;
    create_zero_vector_u32(&num_members, N_c);

    for (uint32_t c = 0; c < N_c; c++)
    {
        for (uint32_t i = 0; i < N_i; i++)
        {
            if (member(c, i) > 0)
            {
                num_members->elts[ c ]++;
            }
        }
    }

    while (assign_random_member_to_small_category(num_members)) ;

    free_vector_u32(num_members);
}


void Sampler::Parameters::assign_members_to_one()
{
    using namespace jwsc;

    uint32_t N_c = num_categories();
    uint32_t N_i = num_instances();

    create_zero_matrix_d(&_cat_members, N_c, N_i);

    for (uint32_t i = 0; i < N_i; i++)
    {
        _cat_members->elts[ 0 ][ i ] = 1;
    }
}


void Sampler::Parameters::assign_min_potential_members()
{
    using namespace jwsc;

    // BEGIN HACK
    //
    // THIS FUNCTION IS NOT RE-ENTRANT!!!
    /*
    static uint32_t i = 0;

    i++;

    if ((i-1) % 500 != 0)
    {
        return;
    }
    */
    // END HACK

    uint32_t N_c = num_categories();
    uint32_t N_i = num_instances();

    if (N_c <= 1)
    {
        create_init_matrix_d(&_cat_members, N_c, N_i, 1.0);
        return;
    }

    Vector_u32* num_members = 0;
    create_zero_vector_u32(&num_members, N_c);

    create_zero_matrix_d(&_cat_members, N_c, N_i);

    for (uint32_t i = 0; i < N_i; i++)
    {
        uint32_t min_c = 0;
        _cat_members->elts[ min_c ][ i ] = 1;
        num_members->elts[ min_c ]++;

        for (uint32_t c = 1; c < N_c; c++)
        {
            if (instance(c, i)->ll_potential < instance(min_c, i)->ll_potential)
            {
                _cat_members->elts[ min_c ][ i ] = 0;
                num_members->elts[ min_c ]--;

                min_c = c;
                _cat_members->elts[ min_c ][ i ] = 1;
                num_members->elts[ min_c ]++;
            }
        }
    }

    //while (assign_random_member_to_small_category(num_members)) ;
    while (assign_min_variance_member_to_small_category(num_members)) ;

    for (uint32_t c = 0; c < N_c; c++)
    {
        assert(num_members->elts[ c ] >= min_num_members());
    }

    free_vector_u32(num_members);
}


const View::Context* Sampler::Parameters::view_ctx() const
{
    return _view_ctx;
}


const Object::Context* Sampler::Parameters::obj_ctx() const
{
    return _obj_ctx;
}


const Block::Context* Sampler::Parameters::block_ctx() const
{
    return _block_ctx;
}


const Block_set::Context* Sampler::Parameters::block_set_ctx() const
{
    return _block_set_ctx;
}


double Sampler::Parameters::img_ll_bg_prob() const
{
    return _img_ll_bg_prob;
}


double Sampler::Parameters::img_ll_fg_prob() const
{
    return _img_ll_fg_prob;
}


double Sampler::Parameters::img_ll_fg_noise_prob() const
{
    return _img_ll_fg_noise_prob;
}


double Sampler::Parameters::img_ll_fg_miss_prob() const
{
    return _img_ll_fg_miss_prob;
}


double Sampler::Parameters::img_ll_edge_noise_prob() const
{
    return _img_ll_edge_noise_prob;
}


double Sampler::Parameters::img_ll_edge_silho_miss_prob() const
{
    return _img_ll_edge_silho_miss_prob;
}


double Sampler::Parameters::img_ll_edge_inner_miss_prob() const
{
    return _img_ll_edge_inner_miss_prob;
}


uint32_t Sampler::Parameters::sample_stats_freq() const
{
    return _sample_stats_freq;
}


/**
 * @param  p_out  Point to get the parameters into.
 * @param  c      Index of the category.
 * @param  i      Index of the model instance to use.
 */
void Sampler::Parameters::get_var_param_values
(
    jwsc::Vector_d** p_out, 
    uint32_t         c, 
    uint32_t         i
) 
const
{
    using namespace jwsc;

    instance(c, i)->get_var_param_values(p_out);
}

/**
 * @param  p  Point to set the parameters with.
 * @param  c  Index of the category.
 * @param  i  Index of the model instance to use.
 */
void Sampler::Parameters::set_var_param_values
(
    const jwsc::Vector_d* p, 
    uint32_t              c, 
    uint32_t              i
)
{
    using namespace jwsc;

    bool cache_hit;

    Instance* ins = instance(c, i);
    ins->set_var_param_values(p);

    try {
        if (!(cache_hit = ins->img_ll_cache->query(
                        &(ins->ll_potential), ins->view, ins->obj)))
        {
            Likelihood::calc_neg_log_likelihood(&(ins->ll_potential),
                    &(ins->img_ll), &(ins->model_mat),
                    &(ins->img_ll_surf_assign), &(ins->img_ll_edge_assign),
                    ins->image_t, ins->edges, ins->img_ll_corr, ins->view,
                    ins->obj, img_ll_bg_prob(), img_ll_fg_prob(),
                    img_ll_fg_noise_prob(), img_ll_fg_miss_prob(),
                    img_ll_edge_noise_prob(), img_ll_edge_silho_miss_prob(),
                    img_ll_edge_inner_miss_prob());
            ins->img_ll_cache->insert(ins->ll_potential, ins->view, ins->obj);
        }
    } catch (jwscxx::base::Arg_error e) { abort(); }
}

/**
 * Shifts the dependents of each component when modifying its parameters.
 *
 * @param  p  Point to add to the the parameters.
 * @param  c  Index of the category.
 * @param  i  Index of the model set to use.
 */
void Sampler::Parameters::add_var_param_values_shift_component_deps
(
    const jwsc::Vector_d* p,
    uint32_t              c,
    uint32_t              i
)
{
    using namespace jwsc;

    bool cache_hit;

    Instance* ins = instance(c, i);
    ins->add_var_param_values_shift_component_deps(p);

    try {
        if (!(cache_hit = ins->img_ll_cache->query(
                        &(ins->ll_potential), ins->view, ins->obj)))
        {
            Likelihood::calc_neg_log_likelihood(&(ins->ll_potential),
                    &(ins->img_ll), &(ins->model_mat),
                    &(ins->img_ll_surf_assign), &(ins->img_ll_edge_assign),
                    ins->image_t, ins->edges, ins->img_ll_corr, ins->view,
                    ins->obj, img_ll_bg_prob(), img_ll_fg_prob(),
                    img_ll_fg_noise_prob(), img_ll_fg_miss_prob(),
                    img_ll_edge_noise_prob(), img_ll_edge_silho_miss_prob(),
                    img_ll_edge_inner_miss_prob());
            ins->img_ll_cache->insert(ins->ll_potential, ins->view, ins->obj);
        }
    } catch (jwscxx::base::Arg_error e) { abort(); }
}

/**
 * Scales the dependents of each component when modifying its parameters.
 *
 * @param  p  Point to add to the the parameters.
 * @param  c  Index of the category.
 * @param  i  Index of the model instance to use.
 */
void Sampler::Parameters::add_var_param_values_scale_component_deps
(
    const jwsc::Vector_d* p,
    uint32_t              c,
    uint32_t              i
)
{
    using namespace jwsc;

    bool cache_hit;

    Instance* ins = instance(c, i);
    ins->add_var_param_values_scale_component_deps(p);

    try {
        if (!(cache_hit = ins->img_ll_cache->query(
                        &(ins->ll_potential), ins->view, ins->obj)))
        {
            Likelihood::calc_neg_log_likelihood(&(ins->ll_potential),
                    &(ins->img_ll), &(ins->model_mat),
                    &(ins->img_ll_surf_assign), &(ins->img_ll_edge_assign),
                    ins->image_t, ins->edges, ins->img_ll_corr, ins->view,
                    ins->obj, img_ll_bg_prob(), img_ll_fg_prob(),
                    img_ll_fg_noise_prob(), img_ll_fg_miss_prob(),
                    img_ll_edge_noise_prob(), img_ll_edge_silho_miss_prob(),
                    img_ll_edge_inner_miss_prob());
            ins->img_ll_cache->insert(ins->ll_potential, ins->view, ins->obj);
        }
    } catch (jwscxx::base::Arg_error e) { abort(); }
}

/**
 * @param  p_out  Point to get the parameters into.
 * @param  c      Category index.
 */
void Sampler::Parameters::get_var_param_stats
(
    jwsc::Vector_d** p_out, 
    uint32_t         c
) 
const
{
    instance(c, 0)->get_var_param_stats(p_out);
}

/**
 * @param  p_out  Point to get the parameters into.
 * */
void Sampler::Parameters::get_var_param_stats(jwsc::Vector_d** p_out) const
{
    using namespace jwsc;

    uint32_t N_c = num_categories();
    uint32_t D   = 2*num_cat_var_params(0);
    uint32_t d   = 0;

    Vector_d* p;
    create_vector_d(p_out, D);
    p = *p_out;

    Vector_d* q = 0;

    for (uint32_t c = 0; c < N_c; c++)
    {
        get_var_param_stats(&q, c);
        copy_vector_section_into_vector_d(p, d, q, 0, q->num_elts);
        d += q->num_elts;
    }

    free_vector_d(q);

    assert(d == D);
}

/**
 * @param  p  Point to set the parameters with.
 * @param  c  Category index.
 */
void Sampler::Parameters::set_var_param_stats
(
    const jwsc::Vector_d* p, 
    uint32_t              c
) 
{
    Category* cat = category(c);
    Instance* ins = instance(c, 0);
    ins->set_var_param_stats(p);

    double view_lp = ins->view->calc_var_param_prior_log_prob();
    double obj_lp  = ins->obj->calc_var_param_prior_log_prob();
    cat->lp_potential = - (view_lp + obj_lp);
    cat->potential = cat->lp_potential;

    for (uint32_t i = 0; i < num_instances(); i++)
    {
        ins = instance(c, i);
        ins->set_var_param_stats(p);

        double view_ll = ins->view->calc_var_param_log_likelihood();
        double obj_ll  = ins->obj->calc_var_param_log_likelihood();
        ins->ll_potential = - (ins->img_ll + view_ll + obj_ll);
        cat->potential += ins->ll_potential;
    }
}

/**
 * @param  p  Point to set the parameters with.
 */
void Sampler::Parameters::set_var_param_stats(const jwsc::Vector_d* p) 
{
    using namespace jwsc;

    uint32_t D = 2*num_cat_var_params(0);
    uint32_t d = 0;

    Vector_d* q = 0;

    for (uint32_t c = 0; c < num_categories(); c++)
    {
        uint32_t D_c = 2*num_var_params(c, 0);
        copy_vector_section_d(&q, p, d, D_c);
        set_var_param_stats(q, c);
        d += D_c;
    }

    free_vector_d(q);

    assert(d == D);
}

/**
 * We assume the model instances are independent, both in the likelihood and
 * priors. So the energies across model instances is summed -- the potential
 * energy function is a negative log.
 */
double Sampler::Parameters::cat_lp_potential(uint32_t c) const
{
    return category(c)->lp_potential;
}

/**
 * @param  c  Index of the category.
 * @param  i  Index of the model instance to use.
 *
 * Hamiltonian -- sum of the potential and kinetic energies.
 */
double Sampler::Parameters::ins_ll_energy(uint32_t c, uint32_t i) const
{
    return ins_ll_potential(c, i) + ins_ll_kinetic(c, i);
}

/**
 * @param  c  Index of the category.
 * @param  i  Index of the model instance to use.
 *
 * We assume the model instances are independent, both in the likelihood and
 * priors. So the energies across model instances is summed -- the potential
 * energy function is a negative log.
 */
double Sampler::Parameters::ins_ll_potential(uint32_t c, uint32_t i) const
{
    return instance(c, i)->ll_potential;
}

/**
 * @param  c  Index of the category.
 * @param  i  Index of the model instance to use.
 *
 * We assume the model instances are independent, both in the likelihood and
 * priors. So the energies across model instances is summed -- the potential
 * energy function is a negative log.
 */
double Sampler::Parameters::best_ins_ll_potential(uint32_t c, uint32_t i) const
{
    return instance(c, i)->best_ll_potential;
}

/**
 * @param  c  Index of the category.
 * @param  i  Index of the model instance to use.
 */
double Sampler::Parameters::ins_ll_kinetic(uint32_t c, uint32_t i) const
{
    return instance(c, i)->ll_kinetic;
}

/**
 * @param  p  Model parameter point to calculate the kinetic energy of.
 * @param  c  Index of the category.
 * @param  i  Index of the model instance to use.
 */
void Sampler::Parameters::update_ins_ll_kinetic
(
    const jwsc::Vector_d* p, 
    uint32_t c,
    uint32_t i
)
{
    instance(c, i)->ll_kinetic = 0;
    calc_vector_dot_prod_d(&(instance(c, i)->ll_kinetic), p, p);
    instance(c, i)->ll_kinetic *= 0.5;
}

/**
 * @param  p  Model parameter point to calculate the kinetic energy of.
 * @param  c  Index of the category.
 */
void Sampler::Parameters::update_cat_lp_kinetic
(
    const jwsc::Vector_d* p, 
    uint32_t              c
)
{
    Category* cat = category(c);

    cat->lp_kinetic = 0;
    calc_vector_dot_prod_d(&(cat->lp_kinetic), p, p);
    cat->lp_kinetic *= 0.5;
}

/**
 * @param  grad_out  Gradient of the negative log likelihood and prior of the
 *                   model parameters.
 * @param  c         Index of the category.
 * @param  i         Index of the model instance to use.
 */
void Sampler::Parameters::grad_ins_potential
(
    jwsc::Vector_d** grad_out, 
    uint32_t         c,
    uint32_t         i
) 
const
{
    Instance* ins = instance(c, i);

    try 
    {
        Likelihood::calc_grad_neg_log_likelihood_and_prior_wrt_ins_params(
                grad_out, &(ins->model_mat), &(ins->img_ll_surf_assign),
                &(ins->img_ll_surf_assign), ins->ll_potential, ins->image_t,
                ins->edges, ins->img_ll_corr, ins->img_ll_cache, ins->view,
                ins->obj, img_ll_bg_prob(), img_ll_fg_prob(),
                img_ll_fg_noise_prob(), img_ll_fg_miss_prob(),
                img_ll_edge_noise_prob(), img_ll_edge_silho_miss_prob(),
                img_ll_edge_inner_miss_prob(), true);
    }
    catch (jwscxx::base::Arg_error e)
    {
        // Should never happen.
        abort();
    }
}

/**
 * @param  grad_out  Gradient of the negative log likelihood and prior of the
 *                   model parameters.
 * @param  c         Category index.
 */
void Sampler::Parameters::grad_cat_potential
(
    jwsc::Vector_d** grad_out, 
    uint32_t         c
) 
const
{
    Category* cat = category(c);

    std::vector<const View*> ins_views;
    std::vector<const Object*> ins_objs;

    for (uint32_t i = 0; i < num_instances(); i++)
    {
        ins_views.push_back(instance(c, i)->view);
        ins_objs.push_back(instance(c, i)->obj);
    }

    try 
    {
        Likelihood::calc_grad_neg_log_likelihood_and_prior_wrt_cat_params(
                grad_out, cat->lp_potential, cat->view, cat->obj, ins_views,
                ins_objs, true);
    }
    catch (jwscxx::base::Arg_error e)
    {
        // Should never happen.
        abort();
    }
}

/**
 * @param  grad_out  Gradient of the negative log likelihood and prior of the
 *                   model parameters.
 */
void Sampler::Parameters::grad_cat_potential(jwsc::Vector_d** grad_out) const
{
    using namespace jwsc;

    uint32_t D = 2*num_cat_var_params(0);
    uint32_t d = 0;

    create_vector_d(grad_out, D);
    Vector_d* grad = *grad_out;

    Vector_d* g = 0;

    for (uint32_t c = 0; c < num_categories(); c++)
    {
        grad_cat_potential(&g, c);
        copy_vector_section_into_vector_d(grad, d, g, 0, g->num_elts);
        d += g->num_elts;
    }

    free_vector_d(g);

    assert(d == D);
}

/**
 * @param  delta_t_out  Stochastic dynamics delta_t values for the parameters.
 * @param  c            Index of the category.
 * @param  i            Index of the model instance to use.
 */
void Sampler::Parameters::get_var_param_values_dynamics_delta_t
(
    jwsc::Vector_d** delta_t_out, 
    uint32_t         c,
    uint32_t         i
) 
const
{
    instance(c, i)->get_var_param_values_dynamics_delta_t(delta_t_out);
}

/**
 * @param  delta_t_out  Stochastic dynamics delta_t values for the parameters.
 * @param  c            Index of the category.
 */
void Sampler::Parameters::get_var_param_stats_dynamics_delta_t
(
    jwsc::Vector_d** delta_t_out, 
    uint32_t         c
) 
const
{
    instance(c, 0)->get_var_param_stats_dynamics_delta_t(delta_t_out);
}

/**
 * @param  delta_t_out  Stochastic dynamics delta_t values for the parameters.
 */
void Sampler::Parameters::get_var_param_stats_dynamics_delta_t
(
    jwsc::Vector_d** delta_t_out
) 
const
{
    using namespace jwsc;

    uint32_t N_c = num_categories();
    uint32_t D   = 2*num_cat_var_params(0);
    uint32_t d   = 0;

    Vector_d* delta_t;
    create_vector_d(delta_t_out, D);
    delta_t = *delta_t_out;

    Vector_d* q = 0;

    for (uint32_t c = 0; c < N_c; c++)
    {
        get_var_param_stats_dynamics_delta_t(&q, c);
        copy_vector_section_into_vector_d(delta_t, d, q, 0, q->num_elts);
        d += q->num_elts;
    }

    free_vector_d(q);

    assert(d == D);
}

/** @return Number of threads of execution to run. */
uint32_t Sampler::Parameters::num_threads() const
{
    return _num_threads;
}


uint32_t Sampler::Parameters::min_num_members() const
{
    return _min_num_members;
}


void Sampler::Parameters::add_sample(Instance* ins)
{
    jwsc::Vector_d* sample = 0;

    ins->get_var_param_values(&sample);
    ins->samples.push_back(sample);
    ins->num_samples++;

    if (sample_stats_freq() > 0 && 
            (ins->num_samples % sample_stats_freq()) == 0)
    {
        ins->update_sample_stats();
    }
}


void Sampler::Parameters::add_sample()
{
    for (uint32_t c = 0; c < num_categories(); c++)
    {
        for (uint32_t i = 0; i < num_instances(); i++)
        {
            add_sample(instance(c, i));
        }
    }
}


void Sampler::Parameters::clear_samples()
{
    for (uint32_t c = 0; c < num_categories(); c++)
    {
        for (uint32_t i = 0; i < num_instances(); i++)
        {
            Instance* ins = instance(c, i);

            std::list<jwsc::Vector_d*>::iterator it;
            for (it = ins->samples.begin(); it != ins->samples.end(); it++)
            {
                jwsc::free_vector_d(*it);
            }

            ins->samples.clear();
            ins->num_samples = 0;
        }
    }
}


bool Sampler::Parameters::assign_min_variance_member_to_small_category
(
    jwsc::Vector_u32* num_members
)
{
    using namespace jwsc;

    uint32_t N_c = num_categories();
    uint32_t N_i = num_instances();

    if (N_c <= 1)
        return false;

    for (uint32_t c = 0; c < N_c; c++)
    {
        if (num_members->elts[ c ] < min_num_members())
        {
            uint32_t min_i = N_i;
            double   min_s = -1;

            for (uint32_t i = 0; i < N_i; i++)
            {
                if (member(c, i) > 0)
                    continue;

                bool i_off_limits = false;
                for (uint32_t cc = 0; cc < N_c; cc++)
                {
                    if (member(cc, i) > 0 && 
                        num_members->elts[ cc ] <= min_num_members())
                    {
                        i_off_limits = true;
                        break;
                    }
                }
                if (i_off_limits)
                    continue;

                double x = 0;
                double x_sqr = 0;

                for (uint32_t cc = 0; cc < N_c; cc++)
                {
                    double p = instance(cc, i)->ll_potential;
                    x += p;
                    x_sqr += p*p;
                }

                double s = (1.0/N_c)*(x_sqr - (x*x)/N_c);
                //assert(s >= 0);

                if (min_s < 0 || s < min_s)
                {
                    min_s = s;
                    min_i = i;
                }
            }
            assert(min_i < N_i);

            for (uint32_t cc = 0; cc < N_c; cc++)
            {
                if (cc != c && member(cc, min_i) > 0)
                {
                    _cat_members->elts[ cc ][ min_i ] = 0;
                    num_members->elts[ cc ]--;

                    _cat_members->elts[ c ][ min_i ] = 1;
                    num_members->elts[ c ]++;

                    break;
                }
            }

            return true;
        }
    }

    return false;
}


bool Sampler::Parameters::assign_random_member_to_small_category
(
    jwsc::Vector_u32* num_members
)
{
    using namespace jwsc;

    uint32_t N_c = num_categories();

    if (N_c <= 1)
        return false;

    for (uint32_t c = 0; c < N_c; c++)
    {
        uint32_t i;

        if (num_members->elts[ c ] < min_num_members())
        {
            i = (uint32_t)floor(sample_uniform_pdf_d(0, num_instances()));
            if (i == num_instances()) i--;

            for (uint32_t cc = 0; cc < N_c; cc++)
            {
                if (cc != c && member(cc, i) > 0)
                {
                    _cat_members->elts[ cc ][ i ] = 0;
                    num_members->elts[ cc ]--;

                    _cat_members->elts[ c ][ i ] = 1;
                    num_members->elts[ c ]++;

                    break;
                }
            }

            return true;
        }
    }

    return false;
}

/* -------------------------------------------------------------------------- */
