/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief MCMC sampler.
 */


#ifndef SAMPLER_H
#define SAMPLER_H


#include <config.h>

#include <iostream>
#include <fstream>
#include <list>
#include <vector>

#include <inttypes.h>
#include <pthread.h>

#include <jwsc/base/time.h>
#include <jwsc/vector/vector.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/image/image.h>
#include <jwsc/image/edge.h>

#include <jwsc++/base/exception.h>

#include "view.h"
#include "object.h"
#include "block.h"
#include "block_set.h"
#include "likelihood.h"


/**
 * @class Sampler
 *
 * @brief Abstract sampler.
 */
class Sampler
{
    public:


/**
 * @class Sampler::Instance
 *
 * @brief Instance of models being inferred by the Sampler from data.
 *
 * This class is more or less a C structure with a helpful allocator and
 * deallocator. It's contained by the Sampler::Parameters class and used in a
 * matrix of Instance.
 */
class Instance
{
    public:

        /**
         * @name Standard
         *
         * Standard class members.
         * @{
         */

        /** @brief Constructs a new Instance for the Sampler. */
        Instance
        (
            View*                       best_view,
            Object*                     best_obj,
            double                      best_img_ll,
            double                      best_potential,
            View*                       view,
            Object*                     obj,
            double                      img_ll,
            double                      potential,
            const jwsc::Matrix_f*       image_t,
            const jwsc::Edge_set_f*     edges,
            Likelihood::Correspondence* img_ll_corr,
            Likelihood::Cache*          img_ll_cache,
            jwsc::Matrix_u32*           img_ll_surf_assign,
            jwsc::Matrix_u32*           img_ll_edge_assign,
            jwsc::Matrix_u32*           model_mat,
            jwsc::Vector_d*             sample_mean_in,
            jwsc::Matrix_d*             sample_cov_in,
            const char*                 sample_info_fname,
            bool                        compress_sample_info,
            const char*                 sample_stats_info_fname
        )
        throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);

        /** @brief Deletes this Instance. */
        virtual ~Instance();

        /*@}*/


        /**
         * @name Variable Parameters
         *
         * Variable model parameter members.
         * @{
         */

        /** @brief Number of parameters a sample has within this instance. */
        uint32_t num_var_params() const;

        /** @brief Gets a vector of model parameters. */
        void get_var_param_values(jwsc::Vector_d** p_out) const;

        /** @brief Sets the model parameters from a vector. */
        void set_var_param_values(const jwsc::Vector_d* p);

        /** 
         * @brief Adds a vector to the model parameters and shifts dependent
         * components' parameters.
         */
        void add_var_param_values_shift_component_deps(const jwsc::Vector_d* p);

        /** 
         * @brief Adds a vector to the model parameters and scales dependent 
         * components' parameters.
         */
        void add_var_param_values_scale_component_deps(const jwsc::Vector_d* p);

        /** 
         * @brief Returns a vector of stochastic dynamics delta_t from the
         * model parameters.
         */
        void get_var_param_values_dynamics_delta_t(jwsc::Vector_d** delta_t_out)
            const;

        /** @brief Gets a vector of model parameters. */
        void get_var_param_stats(jwsc::Vector_d** p_out) const;

        /** @brief Gets a vector of model parameters. */
        void set_var_param_stats(const jwsc::Vector_d* p);

        /** 
         * @brief Returns a vector of stochastic dynamics delta_t from the
         * model parameters.
         */
        void get_var_param_stats_dynamics_delta_t(jwsc::Vector_d** delta_t_out)
            const;

        /** @brief Prints variable parameter information. */
        void print_var_param_info();

        /** @brief Prints variable parameter information. */
        void print_var_param_info(std::ostream& out);

        /*@}*/


        /**
         * @name Sample Statistics
         *
         * Statistics over the generated samples.
         * @{
         */

        /** @brief Updates the set of accepted sample statistics. */
        void update_sample_stats();

        /** @brief Sample mean and covariance matrix. */
        void calc_sample_stats
        (
            jwsc::Vector_d** mean_out,
            jwsc::Matrix_d** cov_out
        )
        const throw (jwscxx::base::Result_error);

        /*@}*/


        /** @brief Best View. */
        View* best_view;

        /** @brief Best Object. */
        Object* best_obj;

        /** @brief Image log likelihood of the best models. */
        double best_img_ll;

        /** @brief Potential energy of the best models. */
        double best_ll_potential;

        /** @brief Most recently generated View sample. */
        View* view;

        /** @brief Most recently generated Object sample. */
        Object* obj;

        /** 
         * @brief Image log likelihood of the most recently generated models.
         */
        double img_ll;

        /** @brief Potential energy of the model likelihoods. */
        double ll_potential;

        /** @brief Kinetic energy of the model likelihoods. */
        double ll_kinetic;

        /** @brief Thresholded input images. */
        const jwsc::Matrix_f* image_t;

        /** 
         * @brief Edges points used in the likelihood correspondences and
         * calculation.
         */
        const jwsc::Edge_set_f* edges;

        /** 
         * @brief Image likelihood correspondence information between model
         * points and edge points in the data.
         */
        Likelihood::Correspondence* img_ll_corr;

        /** @brief Image likelihod value cache. */
        Likelihood::Cache* img_ll_cache;

        /** 
         * @brief Image likelihood pixel assignment for surface detections.
         *
         * Each element of the matrix specifies which type of generative
         * processes created the surface detection.
         *
         * @code
         *    0 - Background
         *    1 - Surface
         *    2 - Noisy surface
         *    3 - Missing surface
         * @endcode
         */
        jwsc::Matrix_u32* img_ll_surf_assign;

        /** 
         * @brief Image likelihood pixel assignment for edge detections.
         *
         * Each element of the matrix specifies which type of generative
         * processes created the edge detection.
         *
         * @code
         *    0 - Background
         *    1 - Edge
         *    2 - Noisy edge
         *    3 - Missing silhouette edge
         *    4 - Missing inner edge
         * @endcode
         */
        jwsc::Matrix_u32* img_ll_edge_assign;

        /** @brief Rendered model scene matrix for an Object under a View. */
        jwsc::Matrix_u32* model_mat;

        /** @brief Sample mean input from a previous run. */
        jwsc::Vector_d* sample_mean_in;

        /** @brief Sample covariance input from a previous run. */
        jwsc::Matrix_d* sample_cov_in;

        /** @brief Collection of generated samples. */
        std::list<jwsc::Vector_d*> samples;

        /** 
         * @brief Number of generated samples. 
         *
         * So we don't have to iterate over the list of samples to count how
         * many there are.
         */
        uint32_t num_samples;

        /** @brief Sample mean. */
        jwsc::Vector_d* sample_mean;

        /** @brief Sample covariance. */
        jwsc::Matrix_d* sample_cov;

        /** @brief Sample covariance eigenvalues. */
        jwsc::Vector_d* sample_cov_eigval;

        /** @brief Sample covariance eigenvectors. */
        jwsc::Matrix_d* sample_cov_eigvec;

        /** @brief File name eam for sample information. */
        std::string sample_info_fname;

        /** @brief Output stream for sample information. */
        std::ofstream sample_info;

        /** @brief Compress the sample information file. */
        bool compress_sample_info;

        /** @brief Output stream for sample statistics information. */
        std::ofstream sample_stats_info;

        /** 
         * @brief Mutex for parameter access between the furniture program and
         * its interactive interface.
         */
        pthread_mutex_t mux;

        /** @brief Iteration the sampler is on. */
        size_t iteration;
};


/**
 * @class Sampler::Category
 *
 * @brief Category of objects being inferred by the Sampler from data.
 *
 * This class is more or less a C structure with a helpful allocator and
 * deallocator. It's contained by the Sampler::Parameters class and used in a
 * vector of subcategories.
 */
class Category
{
    public:

        /**
         * @name Standard
         *
         * Standard class members.
         * @{
         */

        /** @brief Constructs a new Category for the Sampler. */
        Category
        (
            const Instance* ins,
            double          lp_potential,
            double          potential,
            const char*     sample_info_fname,
            bool            compress_sample_info
        )
        throw (jwscxx::base::Arg_error, jwscxx::base::IO_error);

        /** @brief Deletes this Instance. */
        virtual ~Category();

        /** @brief Returns the iteration the sampler is on. */
        size_t iteration() const;

        /*@}*/


        /**
         * @name Variable Parameters
         *
         * Variable model parameter members.
         * @{
         */

        /** @brief Number of parameters a sample has within this instance. */
        uint32_t num_var_params() const;

        /** @brief Prints variable parameter information. */
        void print_var_param_info();

        /** @brief Prints variable parameter information. */
        void print_var_param_info(std::ostream& out);

        /*@}*/


        /** @brief Best View. */
        const View* best_view;

        /** @brief Best Object. */
        const Object* best_obj;

        /** @brief Most recently generated View sample. */
        const View* view;

        /** @brief Most recently generated Object sample. */
        const Object* obj;

        /** @brief Potential energy of the model priors. */
        double lp_potential;

        /** @brief Kinetic energy of the model priors. */
        double lp_kinetic;

        /** @brief Potential energy of the models. */
        double potential;

        /** @brief Kinetic energy of the models. */
        double kinetic;

        /** @brief File name eam for sample information. */
        std::string sample_info_fname;

        /** @brief Output stream for sample information. */
        std::ofstream sample_info;

        /** @brief Compress the sample information file. */
        bool compress_sample_info;

        /** @brief Ellapsed time between sampler iterations. */
        jwsc::Timer timer;


    private:

        /** @brief Iteration the sampler is on. */
        const size_t* _iteration;
};


/**
 * @class Sampler::Parameters
 *
 * @brief Parameters for an MCMC sampler.
 */
class Parameters
{
    public:

        /**
         * @name Standard
         *
         * Standard class members.
         * @{
         */

        /** @brief Constructs the sampler Parameters. */
        Parameters
        (
            const View::Context*      view_ctx,
            const Object::Context*    obj_ctx,
            const Block::Context*     block_ctx,
            const Block_set::Context* block_set_ctx,
            double                    img_ll_bg_prob,
            double                    img_ll_fg_prob,
            double                    img_ll_fg_noise_prob,
            double                    img_ll_fg_miss_prob,
            double                    img_ll_edge_noise_prob,
            double                    img_ll_edge_silho_miss_prob,
            double                    img_ll_edge_inner_miss_prob,
            uint32_t                  sample_stats_freq,
            uint32_t                  num_instances,
            uint32_t                  num_categories,
            uint32_t                  min_num_members,
            uint32_t                  num_threads
        );

        /** @brief Deletes a Sampler_move_parameters. */
        virtual ~Parameters();

        /** 
         * @brief Returns the frequency (in number of accepted samples) to
         * estimate the sample statistics.
         */
        uint32_t sample_stats_freq() const;

        /*@}*/


        /**
         * @name Model instances
         *
         * Model instance members.
         * @{
         */

        /** @brief Returns the number of model instances in this Parameters. */
        uint32_t num_instances() const;

        /** 
         * @brief Returns a model instance from this set of sampler Parameters.
         */
        virtual Sampler::Instance* instance(uint32_t c, uint32_t i) const
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Assigns a model instance to this set of sampler Parameters.
         */
        void instance(Sampler::Instance* instance, uint32_t c, uint32_t i)
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Updates the model category parameter statistics using the
         * model instances.
         */
        void update_instance_stats();

        /*@}*/


        /**
         * @name Object category
         *
         * Object category members.
         * @{
         */

        /** @brief Returns the number of categories. */
        uint32_t num_categories() const;

        /** @brief Returns the number of category parameters. */
        uint32_t num_cat_var_params(uint32_t i) const;

        /** 
         * @brief Returns a category from this set of sampler Parameters.
         */
        virtual Sampler::Category* category(uint32_t c) const
            throw (jwscxx::base::Arg_error);

        /** 
         * @brief Assigns a category to this set of sampler Parameters.
         */
        void category(Sampler::Category* category, uint32_t c)
            throw (jwscxx::base::Arg_error);

        /** @brief Gets all category membership values for an instance. */
        void member(jwsc::Vector_d** member_out, uint32_t i) const
            throw (jwscxx::base::Arg_error);

        /** @brief Returns a category membership value for an instance. */
        double member(uint32_t c, uint32_t i) const 
            throw (jwscxx::base::Arg_error);

        /** @brief Sets the category member values of an instance. */
        void member(double member, uint32_t c, uint32_t i)
            throw (jwscxx::base::Arg_error);

        /** @brief Sets the category member values of an instance. */
        void member(const jwsc::Vector_d* member, uint32_t i)
            throw (jwscxx::base::Arg_error);

        /** @brief Returns the minimum number of members in a category. */
        uint32_t min_num_members() const;

        /** 
         * @brief Assigns random category membership values for each instance
         * to all categories.
         */
        void assign_random_members();

        /** @brief Assigns each instance to one random category. */
        void assign_members_to_one();

        /** 
         * @brief Assigns each instance to a category based on minimum potential
         * energy.
         */
        void assign_min_potential_members();

        /*@}*/


        /**
         * @name Variable Parameters
         *
         * Variable model parameter members.
         * @{
         */

        /** @brief Returns the total number of instance parameters. */
        uint32_t num_var_params(uint32_t c) const;

        /** @brief Returns the total number of instance parameters. */
        uint32_t num_var_params(uint32_t c, uint32_t i) const;

        /** @brief Gets a vector of model parameters. */
        void get_var_param_values
        (
            jwsc::Vector_d** p_out, 
            uint32_t         c, 
            uint32_t         i
        ) 
        const;

        /** @brief Sets the model parameters from a vector. */
        void set_var_param_values
        (
            const jwsc::Vector_d* p, 
            uint32_t              c, 
            uint32_t              i
        );

        /** 
         * @brief Adds a vector to the model parameters and shifts dependent
         * components' parameters.
         */
        void add_var_param_values_shift_component_deps
        (
            const jwsc::Vector_d* p, 
            uint32_t              c,
            uint32_t              i
        );

        /** 
         * @brief Adds a vector to the model parameters and scales dependent 
         * components' parameters.
         */
        void add_var_param_values_scale_component_deps
        (
             const jwsc::Vector_d* p,
             uint32_t              c,
             uint32_t              i
        );

        /** 
         * @brief Returns a vector of stochastic dynamics delta_t from the
         * model parameters.
         */
        void get_var_param_values_dynamics_delta_t
        (
            jwsc::Vector_d** delta_t_out, 
            uint32_t         c, 
            uint32_t         i
        ) 
        const;

        /** @brief Gets a vector of model parameters. */
        void get_var_param_stats(jwsc::Vector_d** p_out) const;

        /** @brief Gets a vector of model parameters. */
        void get_var_param_stats(jwsc::Vector_d** p_out, uint32_t c) const;

        /** @brief Sets a vector of model parameters. */
        void set_var_param_stats(const jwsc::Vector_d* p);

        /** @brief Sets a vector of model parameters. */
        void set_var_param_stats(const jwsc::Vector_d* p, uint32_t c);

        /** 
         * @brief Returns a vector of stochastic dynamics delta_t from the
         * model parameters.
         */
        void get_var_param_stats_dynamics_delta_t(jwsc::Vector_d** delta_t_out) 
            const;

        /** 
         * @brief Returns a vector of stochastic dynamics delta_t from the
         * model parameters.
         */
        void get_var_param_stats_dynamics_delta_t
        (
            jwsc::Vector_d** delta_t_out, 
            uint32_t         c
        ) 
        const;

        /*@}*/


        /**
         * @name Concurrency
         *
         * Concurrency management members.
         * @{
         */

        /** @brief Number of threads of execution to run. */
        uint32_t num_threads() const;

        /** @brief Locks this parameter set to stop external access. */
        void lock();

        /** @brief Locks this parameter set to stop external access. */
        void lock(uint32_t c, uint32_t i);

        /** @brief Unlocks this parameter set to allow external access. */
        void unlock();

        /** @brief Unlocks this parameter set to allow external access. */
        void unlock(uint32_t c, uint32_t i);

        /*@}*/


        /**
         * @name Energy
         *
         * Energy function members.
         * @{
         */

        /** @brief Returns the category Prior potential energy. */
        double cat_lp_potential(uint32_t c) const;

        /** @brief Returns the energy of the model in the parameters. */
        double ins_ll_energy(uint32_t c, uint32_t i) const;

        /** 
         * @brief Returns the potential energy of the model in the
         * parameters.
         */
        double ins_ll_potential(uint32_t c, uint32_t i) const;

        /** 
         * @brief Returns the potential energy of the best model in the
         * parameters. 
         */
        double best_ins_ll_potential(uint32_t c, uint32_t i) const;

        /** @brief Returns the kinetic energy of the model in the parameters. */
        double ins_ll_kinetic(uint32_t c, uint32_t i) const;

        /** 
         * @brief Returns the gradient of the potential energy function for
         * the model in the parameters.
         */
        void grad_ins_potential
        (
            jwsc::Vector_d** grad_out, 
            uint32_t         c, 
            uint32_t         i
        ) 
        const;

        /** 
         * @brief Returns the gradient of the potential energy function for
         * the model in the parameters.
         */
        void grad_cat_potential(jwsc::Vector_d** grad_out, uint32_t c) const;

        /** 
         * @brief Returns the gradient of the potential energy function for
         * the model in the parameters.
         */
        void grad_cat_potential(jwsc::Vector_d** grad_out) const;

        /*@}*/


        /**
         * @name Model Context
         *
         * Model context members.
         * @{
         */

        /** @brief Returns the View::Context for this Parameters. */
        const View::Context* view_ctx() const;

        /** @brief Returns the Object::Context for this Parameters. */
        const Object::Context* obj_ctx() const;

        /** @brief Returns the Block::Context for this Parameters. */
        const Block::Context* block_ctx() const;

        /** @brief Returns the Block_set::Context for this Parameters. */
        const Block_set::Context* block_set_ctx() const;

        /*@}*/


        /**
         * @name Likelihood
         *
         * Likelihood members.
         * @{
         */

        /** @brief Returns the image log likelihood background probability. */
        double img_ll_bg_prob() const;

        /** @brief Returns the image log likelihood background probability. */
        double img_ll_fg_prob() const;

        /** @brief Returns the image log likelihood noise probability. */
        double img_ll_fg_noise_prob() const;

        /** @brief Returns the image log likelihood phantom probability. */
        double img_ll_fg_miss_prob() const;

        /** @brief Returns the image log likelihood noise probability. */
        double img_ll_edge_noise_prob() const;

        /** @brief Returns the image log likelihood phantom probability. */
        double img_ll_edge_silho_miss_prob() const;

        /** @brief Returns the image log likelihood phantom probability. */
        double img_ll_edge_inner_miss_prob() const;

        /*@}*/


    protected:

        /**
         * @name Energy
         *
         * Energy function members.
         * @{
         */

        /** @brief Updates the kinetic energy of the model parameters. */
        void update_ins_ll_kinetic
        (
            const jwsc::Vector_d* p, 
            uint32_t              c, 
            uint32_t              i
        );

        /** @brief Updates the kinetic energy of the model parameters. */
        void update_cat_lp_kinetic(const jwsc::Vector_d* p, uint32_t c);

        /*@}*/


        /**
         * @name Samples
         *
         * Generated samples.
         * @{
         */

        /** @brief Adds the model instance parameters as a sample. */
        void add_sample(Sampler::Instance* ins);

        /** 
         * @brief Adds all the model instance parameters in this set as samples.
         */
        void add_sample();

        /** @brief Clears the generated samples. */
        void clear_samples();

        /*@}*/


        /**
         * @name Object category
         *
         * Object category members.
         * @{
         */

        bool assign_min_variance_member_to_small_category(jwsc::Vector_u32* N);

        bool assign_random_member_to_small_category(jwsc::Vector_u32* N);

        /*@}*/


        /** @brief View parameter context. */
        const View::Context* _view_ctx;

        /** @brief Object parameter context. */
        const Object::Context* _obj_ctx;

        /** @brief Block parameter context. */
        const Block::Context* _block_ctx;

        /** @brief Block_set parameter context. */
        const Block_set::Context* _block_set_ctx;

        /** 
         * @brief Probability of a background point in the image.
         *
         * Used in the image likelihood.
         */
        double _img_ll_bg_prob;

        /** 
         * @brief Probability of a foreground point in the image.
         *
         * Used in the image likelihood.
         */
        double _img_ll_fg_prob;

        /** 
         * @brief Probability of a foreground point begin noise in the image.
         *
         * Used in the image likelihood.
         */
        double _img_ll_fg_noise_prob;

        /** 
         * @brief Probability of a foreground point being a phantom model edge
         * point in the model image 
         *
         * Used in the image likelihood.
         */
        double _img_ll_fg_miss_prob;

        /** 
         * @brief Probability of a noise data edge point in the image.
         *
         * Used in the image likelihood.
         */
        double _img_ll_edge_noise_prob;

        /** 
         * @brief Probability of a phantom model edge point on the silhoutte of
         * the object in the image.
         * 
         * Used in the image likelihood.
         */
        double _img_ll_edge_silho_miss_prob;

        /** 
         * @brief Probability of a phantom model edge point not on the
         * silhouette of the object in the image.
         * 
         * Used in the image likelihood.
         */
        double _img_ll_edge_inner_miss_prob;

        /** 
         * @brief Frequency (in number of accepted samples) to estimate the
         * sample statistics.
         */
        uint32_t _sample_stats_freq;

        /** 
         * @brief Matrix of model instances inferred by the Sampler.
         *
         * Major index is the category and the minor is the instance.
         */
        std::vector< std::vector<Sampler::Instance*> > _instances;

        /**
         * @brief Vector of object (sub) categories inferred by the Sampler.
         *
         * Indexed by category.
         */
        std::vector<Sampler::Category*> _categories;

        /** 
         * @brief Category membership for each instance.
         *
         * Major index is the category and the minor is the instance. The matrix
         * values are in [0,1].
         */
        jwsc::Matrix_d* _cat_members;

        /** @brief Minimum number of members required per category. */
        uint32_t _min_num_members;

        /** @brief Number of threads of execution to run. */
        uint32_t _num_threads;

        /** 
         * @brief Mutex for parameter access between the furniture program and
         * its interactive interface.
         */
        pthread_mutex_t _mux;
};


/**
 * @class Sampler::Thread_manager
 *
 * Provides the thread synchronization to execute code by only one thread while
 * the others (busy) wait for it to finish. When any thread leaves the
 * synchronization block, the code is guarenteed to have been complete executed.
 *
 * The template for a synchronization block is as follows
 *
 * @code
 * synchronize();
 *
 * if (last_thread())
 * {
 *     // Code to be executed synchronously by only one thread goes here.
 *
 *     release();
 * }
 * else
 * {
 *     wait();
 * }
 *
 * if (thread() == t)
 * {
 *     // Code to be executed by the t^th thread goes here.
 * }
 *
 * // Code to be executed by all threads goes here.
 *
 * desynchronize();
 * @endcode
 */
class Thread_manager
{
    public:

        /** @brief Constructs a thread manager. */
        Thread_manager(uint32_t thread);

        /** @brief Deletes a thread manager. */
        virtual ~Thread_manager();

        void thread(uint32_t thread) throw (jwscxx::base::Arg_error);

        /** @brief Id of the current thread. */
        uint32_t thread() const;

        /** @brief Returns the number of executing threads. */
        uint32_t num_threads() const;

        /** @brief Sets the number of executing threads. */
        void num_threads(uint32_t n);

        /** @brief Sets the Metropolis sampler to use. */
        virtual void sampler(Sampler* sampler) 
            throw (jwscxx::base::Arg_error) = 0;

        /** @brief Returns the Metropolis sampler to use. */
        virtual Sampler* sampler() const;

        /** @brief Sets the Metropolis::Parameters to use. */
        virtual void params(Sampler::Parameters* params) 
            throw (jwscxx::base::Arg_error) = 0;

        /** @brief Returns the Metropolis::Parameters to use. */
        virtual Sampler::Parameters* params() const;

        /** @brief Sets the number of iterations to run the sampler. */
        void iterations(uint32_t i);

        /** @brief Number of iterations to execute the threaded sampler. */
        uint32_t iterations() const;

        /** @brief Unlocks and then re-locks the params. */
        void toggle_params_lock();

        /** @brief Locks the sampler parameters. */
        void lock_sampler_params();

        /** @brief Unlocks the sampler parameters. */
        void unlock_sampler_params();

        /** @brief Begins a synchronized thread block. */
        void synchronize();

        /** @brief Returns whether a thread is the last to enter. */
        bool last_thread() const;

        /** @brief Releases any waiting threads. */
        void release();

        /** @brief Causes a thread to wait for release. */
        void wait();

        /** @brief Frees all threads to run asynchronously. */
        void desynchronize();


    private:

        /** 
         * @brief Returns whether the current thread is the last one to
         * approach a synchronization block.
         *
         * @warning Must be protected by a mutex.
         */
        bool last() const;

        /** 
         * @brief Returns whether the current thread should keep busy waiting
         * for the other threads to approach a synchronization block.
         */
        bool busy_wait_1() const;

        /** 
         * @brief Returns whether the current thread should keep busy waiting
         * for the other threads to approach a synchronization block.
         */
        bool busy_wait_2() const;

        /** 
         * @brief Sets an indicator that the current thread is preparing to busy
         * wait for the other threads to approach a synchronization block.
         *
         * @warning Must be protected by a mutex.
         */
        void begin_waiting();

        /** 
         * @brief Sets an indicator that the current thread is done busy
         * waiting for the other threads to approach a synchronization block.
         *
         * @warning Must be protected by a mutex.
         */
        void done_waiting();

        /** @brief Returns a mutex for synchronization. */
        pthread_mutex_t* mux();



    protected:

        /** @brief Sampler executed by the threads. */
        static Sampler* _sampler;

        /** @brief Sampler parameters to use. */
        static Sampler::Parameters* _params;


    private:

        /** @brief Thread id. */
        uint32_t _thread;

        /** @brief Number of sampler iterations. */
        static uint32_t _iterations;

        /** @brief Number of threads. */
        static uint32_t _num_threads;

        /** @brief Number of threads busy waiting for synchronization. */
        static uint32_t _num_threads_waiting;

        /** @brief Whether the threads are currently synchronized. */
        static bool _synchronized;

        /** @brief Mutual exclusion data structure. */
        static pthread_mutex_t _mux;
};


    // Sampler
    public:

        virtual ~Sampler();

        /** @brief Runs the iterative MCMC sampler. */
        virtual void run(Sampler::Parameters* params, uint32_t iterations)
            throw (jwscxx::base::Exception) = 0;
};


#endif
