/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Model for a camera with a restricted view.
 */


#include <config.h>

#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cmath>
#include <cassert>

#include <jwsc/base/limits.h>
#include <jwsc/math/constants.h>
#include <jwsc/prob/pdf.h>
#include <jwsc/vector/vector.h>
#include <jwsc/vector/vector_math.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/base/readable.h>
#include <jwsc++/graphics/camera.h>

#include "model.h"
#include "view.h"


#define  ZENITH              0.6
#define  ZENITH_MIN          0.0
#define  ZENITH_MAX          1.57
#define  ZENITH_MU           0.5
#define  ZENITH_SIGMA        0.05
#define  ZENITH_LAMBDA       1.0
#define  ZENITH_ALPHA        1.0
#define  ZENITH_BETA         0.01
#define  ZENITH_DELTA_T      0.00005
#define  ZENITH_GRAD_ETA     0.005
#define  ZENITH_SIGMA_DELTA_T   0.0001
#define  ZENITH_SIGMA_GRAD_ETA  0.005

#define  FOCAL               1000.0
#define  FOCAL_MIN           1000.0
#define  FOCAL_MAX           1000.0
#define  FOCAL_MU            1000.0
#define  FOCAL_SIGMA         1.0
#define  FOCAL_LAMBDA        1.0
#define  FOCAL_ALPHA         1.0
#define  FOCAL_BETA          1.0
#define  FOCAL_DELTA_T       0.005
#define  FOCAL_GRAD_ETA      1.0
#define  FOCAL_SIGMA_DELTA_T   0.0001
#define  FOCAL_SIGMA_GRAD_ETA  0.5

#define  SCALE               1.0
#define  SCALE_MIN           0.1
#define  SCALE_MAX           4.0
#define  SCALE_MU            1.0
#define  SCALE_SIGMA         0.05
#define  SCALE_LAMBDA        1.0
#define  SCALE_ALPHA         1.0
#define  SCALE_BETA          0.01
#define  SCALE_DELTA_T       0.00005
#define  SCALE_GRAD_ETA      0.005
#define  SCALE_SIGMA_DELTA_T   0.0001
#define  SCALE_SIGMA_GRAD_ETA  0.005


/* -----------------------------  View::Context  ---------------------------- */

View::Context::Context() : Model::Context()
{
    typedef Model::Parameter::Context MPContext;
    typedef MPContext::Range          Range;
    typedef MPContext::Dynamics       Dynamics;
    typedef MPContext::Gaussian       Gaussian;
    typedef MPContext::Gaussian_gamma Gaussian_gamma;


    MPContext* c = &_zenith_c;
    c->range(Range(ZENITH_MIN, ZENITH_MAX));
    c->init_value(ZENITH);
    c->likelihood(Gaussian(ZENITH_MU, ZENITH_SIGMA));
    c->prior(Gaussian_gamma(ZENITH_MU, ZENITH_LAMBDA, ZENITH_ALPHA,
                 ZENITH_BETA));
    c->proposal(Gaussian(ZENITH_MU, ZENITH_SIGMA));
    c->gauss_proposal(Gaussian(ZENITH_MU, ZENITH_SIGMA));
    c->value_dynamics(Dynamics(ZENITH_DELTA_T, ZENITH_GRAD_ETA));
    c->mean_dynamics(Dynamics(ZENITH_DELTA_T, ZENITH_GRAD_ETA));
    c->stdev_dynamics(Dynamics(ZENITH_SIGMA_DELTA_T, ZENITH_SIGMA_GRAD_ETA));


    c = &_focal_c;
    c->range(Range(FOCAL_MIN, FOCAL_MAX));
    c->init_value(FOCAL);
    c->likelihood(Gaussian(FOCAL_MU, FOCAL_SIGMA));
    c->prior(Gaussian_gamma(FOCAL_MU, FOCAL_LAMBDA, FOCAL_ALPHA, FOCAL_BETA));
    c->proposal(Gaussian(FOCAL_MU, FOCAL_SIGMA));
    c->gauss_proposal(Gaussian(FOCAL_MU, FOCAL_SIGMA));
    c->value_dynamics(Dynamics(FOCAL_DELTA_T, FOCAL_GRAD_ETA));
    c->mean_dynamics(Dynamics(FOCAL_DELTA_T, FOCAL_GRAD_ETA));
    c->stdev_dynamics(Dynamics(FOCAL_SIGMA_DELTA_T, FOCAL_SIGMA_GRAD_ETA));


    c = &_scale_c;
    c->range(Range(SCALE_MIN, SCALE_MAX));
    c->init_value(SCALE);
    c->likelihood(Gaussian(SCALE_MU, SCALE_SIGMA));
    c->prior(Gaussian_gamma(SCALE_MU, SCALE_LAMBDA, SCALE_ALPHA, SCALE_BETA));
    c->proposal(Gaussian(SCALE_MU, SCALE_SIGMA));
    c->gauss_proposal(Gaussian(SCALE_MU, SCALE_SIGMA));
    c->value_dynamics(Dynamics(SCALE_DELTA_T, SCALE_GRAD_ETA));
    c->mean_dynamics(Dynamics(SCALE_DELTA_T, SCALE_GRAD_ETA));
    c->stdev_dynamics(Dynamics(SCALE_SIGMA_DELTA_T, SCALE_SIGMA_GRAD_ETA));
}

View::Context::Context(const View::Context& ctx)
    : Model::Context(ctx)
{
    *this = ctx;
}

View::Context::~Context()
{
}

View::Context& View::Context::operator= (const View::Context& ctx)
{
    Model::Context::operator=(ctx);

    _zenith_c = ctx._zenith_c;
    _focal_c  = ctx._focal_c;
    _scale_c  = ctx._scale_c;

    return *this;
}

View::Context* View::Context::clone() const
{
    return new View::Context(*this);
}

const Model::Parameter::Context& View::Context::zenith_c() const
{
    return _zenith_c;
}

void View::Context::zenith_c(const Model::Parameter::Context& zenith)
{
    _zenith_c = zenith;
}

const Model::Parameter::Context& View::Context::focal_c() const
{
    return _focal_c;
}

void View::Context::focal_c(const Model::Parameter::Context& focal)
{
    _focal_c = focal;
}

const Model::Parameter::Context& View::Context::scale_c() const
{
    return _scale_c;
}

void View::Context::scale_c(const Model::Parameter::Context& scale)
{
    _scale_c = scale;
}

/* ========================================================================== */







/* --------------------------------  View  ---------------------------------- */

View::View
(
    double               dist,
    double               width,
    double               height,
    double               near,
    double               far,
    const View::Context* ctx
)
: Camera_d(
        0, 0, 0, 
        0, 0, -ctx->focal_c().init_value(), 
        0, 1, 0, 
        60, near, far), 
  Model("View", ctx),
  _distance_p(0, "Distance", dist, 0, this),
  _width_p(1, "Width", width, 0, this),
  _height_p(2, "Height", height, 0, this),
  _zenith_p(3, "Zenith", ctx->zenith_c().init_value(),
          &(ctx->zenith_c()), this),
  _focal_p(4, "Focal Length", ctx->focal_c().init_value(),
          &(ctx->focal_c()),this),
  _scale_p(5, "Scale", 1, &(ctx->scale_c()), this)
{
    add_fixed_param(&_distance_p);
    add_fixed_param(&_width_p);
    add_fixed_param(&_height_p);

    add_var_param(&_zenith_p);
    add_var_param(&_focal_p);
    add_var_param(&_scale_p);

    rotate(zenith(), u);
    focal(ctx->focal_c().init_value());
    scale(ctx->scale_c().init_value());
    size(width, height);
    distance(dist);
}

/** @param  view  View to copy into this one. */
View::View(const View& view) 
    : Camera_d(view), 
      Model(view),
      _distance_p(view._distance_p),
      _width_p(view._width_p),
      _height_p(view._height_p),
      _zenith_p(view._zenith_p),
      _focal_p(view._focal_p),
      _scale_p(view._scale_p)
{
    add_fixed_param(&_distance_p);
    add_fixed_param(&_width_p);
    add_fixed_param(&_height_p);

    add_var_param(&_zenith_p);
    add_var_param(&_focal_p);
    add_var_param(&_scale_p);

    *this = view;
}

View::~View()
{
}

/** 
 * Performs a deep copy of this View.
 *
 * @param  view  View to copy into this one.
 *
 * @return A reference to this View.
 */
View& View::operator=(const View &view)
{
    Camera_d::operator=(view);
    Model::operator=(view);

    _distance_p = view._distance_p;
    _width_p    = view._width_p;
    _height_p   = view._height_p;
    _zenith_p   = view._zenith_p;
    _focal_p    = view._focal_p;
    _scale_p    = view._scale_p;

    return *this;
}

/** @return A new copy of this View. */
View* View::clone() const
{
    return new View(*this);
}

/** 
 * The file format is ASCII.
 *
 * @param  in  Input stream to read this View from.
 *
 * @throw  jwscxx::base::IO_error   Could not read from @em in.
 * @throw  jwscxx::base::Arg_error  Invalid arguments in file to read from.
 */
void View::read(std::istream& in) 
    throw (jwscxx::base::IO_error, jwscxx::base::Arg_error)
{
    Model::read(in);
    jwscxx::graphics::Camera_d::read(in);
}

/**
 * The file format is ASCII.
 *
 * @param  out  Output stream to write this View to.
 *
 * @throw  jwscxx::base::IO_error  Could not write to @em out.
 */
void View::write(std::ostream& out) const 
    throw (jwscxx::base::IO_error)
{
    Model::write(out);
    jwscxx::graphics::Camera_d::write(out);
}

const View::Context* View::context() const
{
    return dynamic_cast<const View::Context*>(Model::context());
}

/** 
 * If a Parameter value could not be set, it is left unchanged and an exception
 * is thrown, after an attempt to set the other parameters is made.
 */
void View::set_var_param_values(const jwsc::Vector_d* p)
    throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;

    uint32_t d = 0;
    uint32_t D = num_var_params();
    double   delta;
    double   epsilon = 1.0e-8;
    double   e;

    assert(D == p->num_elts);

    std::ostringstream ost;

    if (fabs(delta = p->elts[ d ] - zenith()) > 0)
    {
        e = (delta < 0) ? -epsilon : epsilon;
        try 
        { 
            test_zenith(zenith() + delta + e); 
            zenith(zenith() + delta);
        }
        catch (jwscxx::base::Arg_error e) 
        { ost << e.get_msg() << "  "; }
    }
    d++;

    if (fabs(delta = p->elts[ d ] - focal()) > 0)
    {
        e = (delta < 0) ? -epsilon : epsilon;
        try 
        { 
            test_focal(focal() + delta + e); 
            focal(focal() + delta);
        }
        catch (jwscxx::base::Arg_error e) 
        { ost << e.get_msg() << "  "; }
    }
    d++;

    if (fabs(delta = p->elts[ d ] - scale()) > 0)
    {
        e = (delta < 0) ? -epsilon : epsilon;
        try 
        { 
            test_scale(scale() + delta + e); 
            scale(scale() + delta);
        }
        catch (jwscxx::base::Arg_error e) 
        { ost << e.get_msg() << "  "; }
    }
    d++;

    assert(d == D);

    if (ost.str().size())
    {
        throw jwscxx::base::Arg_error(ost.str());
    }
}

double View::distance() const
{
    return _distance_p.value();
}

void View::distance(double distance)
{
    rotate(-zenith(), u);
    translate(0, 0, distance);
    rotate(zenith(), u);

    _distance_p.value(distance);

    update_graphical_rep();
}

void View::size(double* width_out, double* height_out) const
{
    *width_out = _width_p.value();
    *height_out = _height_p.value();
}

void View::size(double width, double height)
{
    double theta = std::atan((0.5f*height) / focal());
    double aperture = JWSC_RAD_TO_DEG * 2.0f * theta;
    set_aperture(aperture);

    _width_p.value(width);
    _height_p.value(height);

    update_graphical_rep();
}

const Model::Parameter* View::zenith_p() const
{
    return &_zenith_p;
}

double View::zenith() const
{
    return _zenith_p.value();
}

/**
 * @param  zenith  Angle in radians of this camera above the ground 
 *                   (x,z-plane).
 *
 * @warning Aborts on invalid arguments.
 */
void View::zenith(double zenith)
{
    try { _zenith_p.test_value(zenith); }
    catch (jwscxx::base::Arg_error e) { abort(); }

    rotate(zenith - this->zenith(), u);
    _zenith_p.value(zenith);

    update_graphical_rep();
}

/**
 * @param  zenith  Angle in radians of this camera above the ground 
 *                   (x,z-plane).
 *
 * @throw  jwscxx::base::Arg_error  Invalid zenith.
 */
void View::test_zenith(double zenith) throw (jwscxx::base::Arg_error)
{
    _zenith_p.test_value(zenith);
}

const Model::Parameter* View::focal_p() const
{
    return &_focal_p;
}

double View::focal() const
{
    return _focal_p.value();
}

/**
 * @param  focal  Focal length.
 *
 * @warning Aborts on invalid arguments.
 */
void View::focal(double focal)
{
    try { _focal_p.test_value(focal); }
    catch (jwscxx::base::Arg_error e) { abort(); }

    set_f_with_new_aperture_using_vrp(focal);
    _focal_p.value(focal);

    update_graphical_rep();
}

/**
 * @param  focal  Focal length.
 *
 * @throw  jwscxx::base::Arg_error  Invalid focal length.
 */
void View::test_focal(double focal) throw (jwscxx::base::Arg_error)
{
    _focal_p.test_value(focal);
}

const Model::Parameter* View::scale_p() const
{
    return &_scale_p;
}

double View::scale() const
{
    return _scale_p.value();
}

/**
 * @param  scale  World scaling.
 *
 * @warning Aborts on invalid arguments.
 */
void View::scale(double scale)
{
    try { _scale_p.test_value(scale); }
    catch (jwscxx::base::Arg_error e) { abort(); }

    _scale_p.value(scale);

    update_graphical_rep();
}

/**
 * @param  scale  World scaling.
 *
 * @throw  jwscxx::base::Arg_error  Invalid scale.
 */
void View::test_scale(double scale) throw (jwscxx::base::Arg_error)
{
    _scale_p.test_value(scale);
}

/**
 * Switches the GL matrix mode to GL_MODELVIEW, loads the identity matrix, and 
 * multiplies the camera rotation and translation matrices.
 */
void View::set_gl_modelview() const
{
    jwscxx::graphics::Camera_d::set_gl_modelview();

    glScalef(scale(), scale(), scale());
}

/* ========================================================================== */
