/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Camera model with a restricted view.
 */


#ifndef VIEW_H
#define VIEW_H


#include <config.h>

#include <iostream>
#include <cmath>

#include <jwsc/vector/vector.h>

#include <jwsc++/base/exception.h>
#include <jwsc++/graphics/camera.h>

#include "model.h"


/** 
 * @class View
 *
 * @brief Restricted view model for a camera. 
 *
 * The center of the camera is always a View::distance away from the origin and 
 * rotated a View::zenith angle off the ground plane. It is always looking at 
 * the world origin.
 */
class View : public jwscxx::graphics::Camera_d, 
             public Model
{
    public:


/**
 * @class View::Context
 *
 * @brief Context for View model parameters.
 */
class Context : public Model::Context
{
    public:

        /** @brief Constructs a new default View::Context. */
        Context();

        /** @brief Constructs a new View::Context from another. */
        Context(const View::Context& m);

        /** @brief Deletes this View::Context. */
        virtual ~Context();

        /** @brief Copies a Context into this one. */
        View::Context& operator= (const View::Context& ctx);

        /** @brief Clones this Context. */
        View::Context* clone() const;

        const Model::Parameter::Context& zenith_c() const;

        void zenith_c(const Model::Parameter::Context& zenith);

        const Model::Parameter::Context& focal_c() const;

        void focal_c(const Model::Parameter::Context& focal);

        const Model::Parameter::Context& scale_c() const;

        void scale_c(const Model::Parameter::Context& scale);

    protected:

        /** @brief Zenith angle of this View above the x,z plane. */
        Model::Parameter::Context _zenith_c;

        /** @brief Focal length of this View. */
        Model::Parameter::Context _focal_c;

        /** @brief Scale of the world. */
        Model::Parameter::Context _scale_c;
};


    // View
    public:

        /**
         * @name Standard
         *
         * Standard class methods.
         * @{
         */

        /** @brief Constructs a new View. */
        View
        (
            double               dist,
            double               width,
            double               height,
            double               near,
            double               far,
            const View::Context* ctx
        );

        /** @brief Constructs a View by copying another. */
        View(const View& view);

        /** @brief Deletes this View. */
        virtual ~View();
        
        /** @brief Copies a View into this one. */
        View& operator=(const View &view);

        /** @brief Clones this View. */
        View* clone() const;

        /** @brief Reads this View from an input stream. */
        void read(std::istream& in) throw (jwscxx::base::IO_error,
                jwscxx::base::Arg_error);

        /** @brief Writes this View to an output stream. */
        void write(std::ostream& out) const throw (jwscxx::base::IO_error);

        /** @brief Returns the model context. */
        const View::Context* context() const;

        /*@}*/


        /**
         * @name Variable Parameters
         *
         * Variable model parameter methods.
         * @{
         */

        /** @brief Sets the instance parameter values. */
        void set_var_param_values(const jwsc::Vector_d* p)
            throw (jwscxx::base::Arg_error);

        /*@}*/


        /**
         * @name distance
         *
         * Distance fixed parameter methods.
         * @{
         */

        /** @brief Returns the distance of the View from the origin. */
        double distance() const;

        /** @brief Sets the distance of the View from the origin. */
        void distance(double distance);

        /*@}*/


        /**
         * @name size
         *
         * View size fixed parameter methods.
         * @{
         */

        /**
         * @brief Sets the size of the View plane at the current focal length.
         */
        void size(double* width_out, double* height_out) const;

        /**
         * @brief Sets the size of the View plane at the current focal length.
         */
        void size(double width, double height);

        /*@}*/


        /**
         * @name zenith
         *
         * Zenith angle parameter methods.
         * @{
         */

        /** @brief Returns the zenith of this View. */
        const Model::Parameter* zenith_p() const;

        /** @brief Returns the zenith of this View. */
        double zenith() const;

        /** @brief Sets the zenith of this View. */
        void zenith(double zenith);

        /** @brief Tests a zenith for this View. */
        void test_zenith(double zenith) throw (jwscxx::base::Arg_error);

        /*@}*/


        /**
         * @name focal
         *
         * Focal length parameter methods.
         * @{
         */

        /** @brief Returns the focal length of this View. */
        const Model::Parameter* focal_p() const;

        /** @brief Returns the focal length of this View. */
        double focal() const;

        /** @brief Sets the focal length of this View. */
        void focal(double focal);

        /** @brief Tests a focal length for this View. */
        void test_focal(double focal) throw (jwscxx::base::Arg_error);

        /*@}*/


        /**
         * @name scale
         *
         * World scale parameter methods.
         * @{
         */

        /** @brief Returns the scale of this View. */
        const Model::Parameter* scale_p() const;

        /** @brief Returns the scale of this View. */
        double scale() const;

        /** @brief Sets the scale of this View. */
        void scale(double scale);

        /** @brief Tests a scale for this View. */
        void test_scale(double scale) throw (jwscxx::base::Arg_error);

        /*@}*/


        /**
         * @name Graphical Representation
         *
         * Internal graphical representation methods.
         * @{
         */

        /** 
         * @brief Updates the graphical representation of this View with its
         * current parameter values.
         *
         * This is a callback member function used by the Model::Parameter when
         * a value changes.
         */
        void update_graphical_representation();

        /*@}*/


        /**
         * @name Graphics
         *
         * Graphics rendering methods.
         * @{
         */

        /** @brief Sets the GL model view matrix with this View. */
        void set_gl_modelview() const;

        /*@}*/


    protected:

        /** @brief Distance of this camera view from the origin. */
        Model::Parameter _distance_p;

        /** @brief Width of this View image plane. */
        Model::Parameter _width_p;

        /** @brief Height of this View image plane. */
        Model::Parameter _height_p;

        /** @brief Zenith angle of this View above the x,z plane. */
        Model::Parameter _zenith_p;

        /** @brief Focal length of this View. */
        Model::Parameter _focal_p;

        /** @brief Scale of the world. */
        Model::Parameter _scale_p;
};


#endif
