/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief MCMC sampler.
 */


#include <config.h>

#include <iostream>
#include <iomanip>
#include <list>
#include <vector>
#include <algorithm>
#include <cassert>

#include <inttypes.h>
#include <pthread.h>

#if defined FURNITURE_HAVE_OPENGL
#include <GL/gl.h>
#include <GL/glu.h>
#elif defined FURNITURE_HAVE_OPENGL_FRAMEWORK
#include <OpenGL/OpenGL.h>
#endif

#include <jwsc/base/time.h>
#include <jwsc/base/limits.h>
#include <jwsc/math/constants.h>
#include <jwsc/vector/vector.h>
#include <jwsc/vector/vector_math.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/prob/pdf.h>
#include <jwsc/prob/mv_pdf.h>
#include <jwsc/image/edge.h>
#include <jwsc/stat/dynamics.h>

#include <jwsc++/base/exception.h>

#include "view.h"
#include "object.h"
#include "block.h"
#include "block_set.h"
#include "likelihood.h"
#include "sampler.h"
#include "offscreen.h"
#include "dynamics.h"
#include "metropolis.h"


#define  MAX_NUM_COMPONENTS  10


/* ----------------------------  Metropolis  -------------------------------- */

/**
 * @param  diff_prob      Probability of a stochastic dynamics move.
 * @param  jump_prob      Probability of a stochastic dynamics move.
 * @param  jump_set_size  Number of Instance to use when computing the
 *                        acceptance probability of a jump move. Set to zero
 *                        to use all model instances.
 */
Metropolis::Metropolis
(
    double   diff_prob,
    double   jump_prob,
    uint32_t jump_set_size
)
{
    _diff_prob     = diff_prob;
    _jump_prob     = jump_prob;
    _jump_set_size = jump_set_size;
}

Metropolis::~Metropolis()
{
    std::list<Metropolis::Move*>::iterator it;

    for (it = _diff_moves.begin(); it != _diff_moves.end(); it++)
    {
        delete *it;
    }
    for (it = _jump_moves.begin(); it != _jump_moves.end(); it++)
    {
        delete *it;
    }
}

uint32_t Metropolis::jump_set_size() const
{
    return _jump_set_size;
}

std::list<Metropolis::Move*> Metropolis::moves()
{
    std::list<Metropolis::Move*> moves(_diff_moves);
    moves.insert(moves.end(), _jump_moves.begin(), _jump_moves.end());
    return moves;
}

std::list<Metropolis::Move*> Metropolis::diffusion_moves()
{
    return std::list<Metropolis::Move*>(_diff_moves);
}

std::list<Metropolis::Move*> Metropolis::jump_moves()
{
    return std::list<Metropolis::Move*>(_jump_moves);
}

/** @param  move  Move to add to this Sampler. */
void Metropolis::add_diffusion(const Metropolis::Diffusion& move)
{
    _diff_moves.push_back(move.clone());
}

/** @param  moves  Balanced pair of jump moves to add to this Sampler. */
void Metropolis::add_jump(const Metropolis::Jump_pair& moves)
{
    _jump_moves.push_back(moves.first().clone());
    _jump_moves.push_back(moves.second().clone());
}

double Metropolis::diffusion_prob() const
{
    return _diff_prob;
}

double Metropolis::jump_prob() const
{
    return _jump_prob;
}

/**
 * @param  params      Parameters for the sampler.
 * @param  iterations  Number of iterations to run the sampler.
 *
 * @throw  jwscxx::base::Exception  An error occurred in the sampler.
 */
void Metropolis::run(Sampler::Parameters* params, uint32_t iterations)
    throw (jwscxx::base::Exception)
{
    Parameters* metropolis_params;
    assert((metropolis_params = dynamic_cast<Parameters*>(params)));

    run(metropolis_params, iterations);
}

/**
 * @param  params      Parameters for the sampler.
 * @param  iterations  Number of iterations to run the sampler.
 *
 * @throw  jwscxx::base::Exception  An error occurred in the sampler.
 */
void Metropolis::run(Metropolis::Parameters* params, uint32_t iterations)
    throw (jwscxx::base::Exception)
{
    typedef void* (*f)(void*);
    uint32_t N_t = params->num_threads();
    uint32_t N_i = params->num_instances();

    if (N_i < N_t)
    {
        N_t = N_i;
    }

    std::vector<Thread_manager> mgr(N_t);
    std::vector<pthread_t>      thr(N_t);
    std::vector<jwscxx::base::Exception*> ex(N_t, 0);

    for (uint32_t t = 0; t < N_t; t++)
    {
        mgr[ t ].thread(t);
        mgr[ t ].num_threads(N_t);
        mgr[ t ].sampler(this);
        mgr[ t ].params(params);
        mgr[ t ].iterations(iterations);
    }

    for (uint32_t t = 0; t < N_t; t++)
    {
        assert(pthread_create(&(thr[ t ]), 0, (f)run_thread_c, &(mgr[t])) == 0);
    }

    for (uint32_t t = 0; t < N_t; t++)
    {
        assert(pthread_join(thr[ t ], (void**)&(ex[ t ])) == 0);
        if (ex[ t ])
        {
            throw jwscxx::base::Exception(*(ex[ t ]));
        }
    }
}

void Metropolis::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    mgr->lock_sampler_params();

    uint32_t N_i = mgr->iterations();
    uint32_t N_c = mgr->params()->num_categories();
    uint32_t t   = mgr->thread();

    assert(Offscreen::enabled());
    Offscreen::activate(t);

    for (uint32_t i = 0; i < N_i; i++)
    {
        mgr->toggle_params_lock();
        mgr->select_sampler_moves();

        for (uint32_t c = 0; c < N_c; c++)
        {
            mgr->category(c);
            mgr->move(c)->run(mgr);
        }

        mgr->assign_min_potential_members();
        mgr->print_var_param_proposal_info();
    }

    assert(Offscreen::enabled());
    Offscreen::deactivate(t);

    mgr->unlock_sampler_params();
}

jwscxx::base::Exception* Metropolis::run_thread_c
(
    Metropolis::Thread_manager* mgr
)
{
    try
    {
        mgr->sampler()->run_thread(mgr);
    }
    catch (jwscxx::base::Exception e)
    {
        pthread_exit(new jwscxx::base::Exception(e));
    }

    return 0;
}

/* ========================================================================== */








/* ----------------------  Metropolis::Instance  ---------------------------- */

/**
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 * @throw  jwscxx::base::IO_error   Could not open info file.
 */
Metropolis::Instance::Instance
(
    View*                       best_view,
    Object*                     best_obj,
    double                      best_img_ll,
    double                      best_potential,
    View*                       view,
    Object*                     obj,
    double                      img_ll,
    double                      potential,
    const jwsc::Matrix_f*       image_t,
    const jwsc::Edge_set_f*     edges,
    Likelihood::Correspondence* img_ll_corr,
    Likelihood::Cache*          img_ll_cache,
    jwsc::Matrix_u32*           img_ll_surf_assign,
    jwsc::Matrix_u32*           img_ll_edge_assign,
    jwsc::Matrix_u32*           model_mat,
    jwsc::Vector_d*             sample_mean_in,
    jwsc::Matrix_d*             sample_cov_in,
    const char*                 sample_info_fname,
    bool                        compress_sample_info,
    const char*                 sample_stats_info_fname
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
: Sampler::Instance(best_view, best_obj, best_img_ll, best_potential, view,
        obj, img_ll, potential, image_t, edges, img_ll_corr, img_ll_cache,
        img_ll_surf_assign, img_ll_edge_assign, model_mat, sample_mean_in,
        sample_cov_in, sample_info_fname, compress_sample_info,
        sample_stats_info_fname)
{
    prop_view         = view->clone();
    prop_obj          = obj->clone();
    prop_img_ll       = img_ll;
    prop_ll_potential = ll_potential;
    prop_ll_kinetic   = ll_kinetic;
    accepted          = false;
}

Metropolis::Instance::~Instance()
{
    delete prop_view;
    delete prop_obj;
}

void Metropolis::Instance::print_var_param_proposal_info
(
    const char* move_name,
    double      membership
) 
{
    print_var_param_proposal_info(move_name, membership, sample_info);
}

void Metropolis::Instance::print_var_param_proposal_info
(
    const char*   move_name, 
    double        membership,
    std::ostream& out
)
{
    out << iteration << ',' << "Move" << ',' << move_name << ','
        << accepted << std::endl;

    out << iteration << ',' << "Membership" << ',' << membership << std::endl;

    out.precision(5);
    out.setf(std::ios_base::scientific, std::ios_base::floatfield);

    out << iteration << ',' << "Image Likelihood" << ','
        << prop_img_ll << std::endl;

    out << iteration << ',' << "Likelihood Potential Energy" << ',' 
        << prop_ll_potential << std::endl;

    out << iteration << ',' << "Likelihood Kinetic Energy" << ','
        << prop_ll_kinetic << std::endl;

    out << iteration << ',' << "Total Likelihood Energy" << ','
        << prop_ll_potential + prop_ll_kinetic << std::endl;

    out << iteration << ',' << "Number of Components" << ','
        << prop_obj->num_components() << std::endl;

    out.precision(4);
    out.setf(std::ios_base::fixed, std::ios_base::floatfield);

    prop_view->print_var_param_values(iteration, out);
    prop_obj->print_var_param_values(iteration, out);

    /*
    out << iteration << ',' << "Cache Hits" << ','
        << img_ll_cache->get_num_hits() << std::endl;

    out << iteration << ',' << "Percent Cache Hits" << ','
        << img_ll_cache->get_percent_hits() << std::endl;

    out << iteration << ',' << "Cache Size" << ','
        << img_ll_cache->get_cache_size() << std::endl;
        */

    out.flush();
}

/**
 * @param  p_out  Point to get the proposed model parameters into.
 */
void Metropolis::Instance::get_var_param_proposal_values
(
    jwsc::Vector_d** p_out
) 
const
{
    using namespace jwsc;

    uint32_t D_1 = prop_view->num_var_params();
    uint32_t D_2 = prop_obj->num_var_params();
    uint32_t D   = D_1 + D_2;

    Vector_d* p;
    create_vector_d(p_out, D);
    p = *p_out;

    Vector_d* q = 0;

    prop_view->get_var_param_values(&q);
    assert(copy_vector_section_into_vector_d(p, 0, q, 0, D_1) == 0);

    prop_obj->get_var_param_values(&q);
    assert(copy_vector_section_into_vector_d(p, D_1, q, 0, D_2) == 0);

    free_vector_d(q);
}

/**
 * Scales the dependents of each component when modifying its parameters.
 *
 * @param  p  Point to add to the the parameters.
 */
void Metropolis::Instance::add_var_param_proposal_values_scale_component_deps
(
    const jwsc::Vector_d* p
)
{
    using namespace jwsc;

    uint32_t D_1 = prop_view->num_var_params();
    uint32_t D_2 = prop_obj->num_var_params();

    Vector_d* q = 0;

    copy_vector_section_d(&q, p, 0, D_1);
    try { prop_view->add_var_param_values(q); }
    catch (jwscxx::base::Arg_error e) { }

    copy_vector_section_d(&q, p, D_1, D_2);
    try { prop_obj->add_var_param_values_scale_component_deps(q); }
    catch (jwscxx::base::Arg_error e) { }

    free_vector_d(q);
}

/**
 * @param  p_out  Point to get the model parameters into.
 */
void Metropolis::Instance::get_var_param_proposal_stats(jwsc::Vector_d** p_out) 
    const
{
    using namespace jwsc;

    uint32_t D_1 = prop_view->num_var_params();
    uint32_t D_2 = prop_obj->num_var_params();
    uint32_t D   = num_var_params();

    Vector_d* p;
    create_vector_d(p_out, 2*D);
    p = *p_out;

    Vector_d* m_q = 0;
    Vector_d* s_q = 0;

    prop_view->get_var_param_stats(&m_q, &s_q);
    assert(copy_vector_section_into_vector_d(p, 0, m_q, 0, D_1) == 0);
    assert(copy_vector_section_into_vector_d(p, D_1, s_q, 0, D_1) == 0);

    prop_obj->get_var_param_stats(&m_q, &s_q);
    assert(copy_vector_section_into_vector_d(p, 2*D_1, m_q, 0, D_2) == 0);
    assert(copy_vector_section_into_vector_d(p, 2*D_1+D_2, s_q, 0, D_2) == 0);

    free_vector_d(m_q);
    free_vector_d(s_q);
}

/**
 * @param  p  Point to set the model parameters from.
 *
 * @note Doesn't update the log likelihood or energy. They must be set
 * separately.
 */
void Metropolis::Instance::set_var_param_proposal_stats(const jwsc::Vector_d* p)
{
    using namespace jwsc;

    uint32_t D_1 = prop_view->num_var_params();
    uint32_t D_2 = prop_obj->num_var_params();

    Vector_d* m_q = 0;
    Vector_d* s_q = 0;

    copy_vector_section_d(&m_q, p, 0, D_1);
    copy_vector_section_d(&s_q, p, D_1, D_1);
    try { prop_view->set_var_param_stats(m_q, s_q); }
    catch (jwscxx::base::Arg_error e) { }

    copy_vector_section_d(&m_q, p, 2*D_1, D_2);
    copy_vector_section_d(&s_q, p, 2*D_1+D_2, D_2);
    try { prop_obj->set_var_param_stats(m_q, s_q); }
    catch (jwscxx::base::Arg_error e) { }

    free_vector_d(m_q);
    free_vector_d(s_q);
}

/* ========================================================================== */







/* ----------------------  Metropolis::Category  ---------------------------- */

/**
 * @param  rep_ins            Representative instance for the category.
 * @param  lp_potential       Log prior potential energy.
 * @param  potential          Potential energy of the likelihood and prior.
 * @param  sample_info_fname  File to write sample information to.
 *
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 * @throw  jwscxx::base::IO_error   Could not open info file.
 */
Metropolis::Category::Category
(
    const Metropolis::Instance* rep_ins,
    double                      lp_potential,
    double                      potential,
    const char*                 sample_info_fname,
    bool                        compress_sample_info
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
: Sampler::Category(rep_ins, lp_potential, potential, sample_info_fname,
        compress_sample_info)
{
    prop_view  = rep_ins->prop_view;
    prop_obj   = rep_ins->prop_obj;
    _accepted  = &(rep_ins->accepted);

    prop_lp_potential = lp_potential;
    prop_lp_kinetic   = lp_kinetic;
    prop_potential    = potential;
    prop_kinetic      = kinetic;
}

Metropolis::Category::~Category()
{
}

void Metropolis::Category::print_var_param_proposal_info(const char* move_name) 
{
    print_var_param_proposal_info(move_name, sample_info);
}

void Metropolis::Category::print_var_param_proposal_info
(
    const char*   move_name, 
    std::ostream& out
) 
{
    out << iteration() << ',' << "Move" << ',' << move_name << ','
        << accepted() << std::endl;

    out << iteration() << ',' << "Time" << ',' 
        << jwsc::user_time_s(&timer) << std::endl;

    out.precision(5);
    out.setf(std::ios_base::scientific, std::ios_base::floatfield);

    out << iteration() << ',' << "Prior Potential Energy" << ',' 
        << prop_lp_potential << std::endl;

    out << iteration() << ',' << "Prior Kinetic Energy" << ','
        << prop_lp_kinetic << std::endl;

    out << iteration() << ',' << "Total Prior Energy" << ','
        << prop_lp_potential + prop_lp_kinetic << std::endl;

    out << iteration() << ',' << "Potential Energy" << ',' 
        << prop_potential << std::endl;

    out << iteration() << ',' << "Kinetic Energy" << ','
        << prop_kinetic << std::endl;

    out << iteration() << ',' << "Total Energy" << ','
        << prop_potential + prop_kinetic << std::endl;

    out.precision(4);
    out.setf(std::ios_base::fixed, std::ios_base::floatfield);

    prop_view->print_var_param_stats(iteration(), out);
    prop_obj->print_var_param_stats(iteration(), out);

    out.flush();
}

bool Metropolis::Category::accepted() const
{
    return *_accepted;
}

/* -------------------------------------------------------------------------- */








/* ---------------------  Metropolis::Parameters  --------------------------- */

Metropolis::Parameters::Parameters
(
    const View::Context*      view_ctx,
    const Object::Context*    obj_ctx,
    const Block::Context*     block_ctx,
    const Block_set::Context* block_set_ctx,
    double                    img_ll_bg_prob,
    double                    img_ll_fg_prob,
    double                    img_ll_fg_noise_prob,
    double                    img_ll_fg_miss_prob,
    double                    img_ll_edge_noise_prob,
    double                    img_ll_edge_silho_miss_prob,
    double                    img_ll_edge_inner_miss_prob,
    uint32_t                  sample_stats_freq,
    uint32_t                  num_instances,
    uint32_t                  num_categories,
    uint32_t                  min_num_members,
    uint32_t                  num_threads
)
: Sampler::Parameters(view_ctx, obj_ctx, block_ctx, block_set_ctx,
        img_ll_bg_prob, img_ll_fg_prob, img_ll_fg_noise_prob,
        img_ll_fg_miss_prob, img_ll_edge_noise_prob,
        img_ll_edge_silho_miss_prob, img_ll_edge_inner_miss_prob,
        sample_stats_freq, num_instances, num_categories, min_num_members,
        num_threads)
{
}

Metropolis::Parameters::~Parameters()
{
}

/**
 * @param  c  Index of the category.
 * @param  i  Index of the Instance to get.
 *
 * @return A pointer to the indexed Instance.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Instance index.
 */
Metropolis::Instance* Metropolis::Parameters::instance
(
    uint32_t c, 
    uint32_t i
) 
const throw (jwscxx::base::Arg_error)
{
    if (i >= num_instances())
    {
        throw jwscxx::base::Arg_error("Model set index too large");
    }

    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    Metropolis::Instance* ins;
    assert((ins = dynamic_cast<Metropolis::Instance*>(_instances[ c ][ i ])));

    return ins;
}

/**
 * @param  c  Index of the category.
 *
 * @return A pointer to the indexed Category.
 *
 * @throw  jwscxx::base::Arg_error  Invalid Category index.
 */
Metropolis::Category* Metropolis::Parameters::category(uint32_t c) const 
    throw (jwscxx::base::Arg_error)
{
    if (c >= num_categories())
    {
        throw jwscxx::base::Arg_error("Category index too large");
    }

    Metropolis::Category* cat;
    assert((cat = dynamic_cast<Metropolis::Category*>(_categories[ c ])));

    return cat;
}

/**
 * @param  p_out  Point to get the parameters into.
 * @param  c      Index of the category.
 * @param  i      Index of the model instance to use.
 */
void Metropolis::Parameters::get_var_param_proposal_values
(
    jwsc::Vector_d** p_out, 
    uint32_t         c,
    uint32_t         i
) 
const
{
    instance(c, i)->get_var_param_proposal_values(p_out);
}

/**
 * Scales the dependents of each component when modifying its parameters.
 *
 * @param  p  Point to add to the the parameters.
 * @param  c  Index of the category.
 * @param  i  Index of the model instance to use.
 */
void Metropolis::Parameters::add_var_param_proposal_values_scale_component_deps
(
    const jwsc::Vector_d* p,
    uint32_t              c,
    uint32_t              i
)
{
    using namespace jwsc;

    bool cache_hit;

    Instance* ins = instance(c, i);
    ins->add_var_param_proposal_values_scale_component_deps(p);

    try {
        if (!(cache_hit = ins->img_ll_cache->query(&(ins->prop_ll_potential), 
                        ins->prop_view, ins->prop_obj)))
        {
            Likelihood::calc_neg_log_likelihood(
                    &(ins->prop_ll_potential), &(ins->prop_img_ll),
                    &(ins->model_mat), &(ins->img_ll_surf_assign),
                    &(ins->img_ll_edge_assign), ins->image_t, ins->edges,
                    ins->img_ll_corr, ins->prop_view, ins->prop_obj,
                    img_ll_bg_prob(), img_ll_fg_prob(), img_ll_fg_noise_prob(),
                    img_ll_fg_miss_prob(), img_ll_edge_noise_prob(),
                    img_ll_edge_silho_miss_prob(),
                    img_ll_edge_inner_miss_prob());
            ins->img_ll_cache->insert(ins->prop_ll_potential, ins->prop_view,
                    ins->prop_obj);
        }
    } catch (jwscxx::base::Arg_error e) { abort(); }
}

/**
 * @param  p_out  Point to get the parameters into.
 * @param  c      Index of the category.
 */
void Metropolis::Parameters::get_var_param_proposal_stats
(
    jwsc::Vector_d** p_out, 
    uint32_t         c
) 
const
{
    instance(c, 0)->get_var_param_proposal_stats(p_out);
}

/**
 * @param  p  Point to set the parameters with.
 * @param  c  Category index.
 */
void Metropolis::Parameters::set_var_param_proposal_stats
(
    const jwsc::Vector_d* p, 
    uint32_t              c
) 
{
    Category* cat = category(c);
    Instance* ins = instance(c, 0);
    ins->set_var_param_proposal_stats(p);

    double view_lp = ins->prop_view->calc_var_param_prior_log_prob();
    double obj_lp  = ins->prop_obj->calc_var_param_prior_log_prob();
    cat->prop_lp_potential = - (view_lp + obj_lp);
    cat->prop_potential = cat->prop_lp_potential;

    for (uint32_t i = 0; i < num_instances(); i++)
    {
        ins = instance(c, i);
        ins->set_var_param_proposal_stats(p);

        double view_ll = ins->prop_view->calc_var_param_log_likelihood();
        double obj_ll  = ins->prop_obj->calc_var_param_log_likelihood();
        ins->prop_ll_potential = - (ins->img_ll + view_ll + obj_ll);
        cat->prop_potential += ins->prop_ll_potential;
    }
}

/**
 * @param  grad_out  Gradient of the negative log likelihood and prior of the
 *                   proposed model parameters.
 * @param  c         Index of the category.
 * @param  i         Index of the model instance to use.
 */
void Metropolis::Parameters::proposal_grad_ins_potential
(
    jwsc::Vector_d** grad_out, 
    uint32_t         c,
    uint32_t         i
)
const
{
    Instance* ins = instance(c, i);

    try 
    {
        Likelihood::calc_grad_neg_log_likelihood_and_prior_wrt_ins_params(
                grad_out, &(ins->model_mat), &(ins->img_ll_surf_assign),
                &(ins->img_ll_edge_assign), ins->prop_ll_potential,
                ins->image_t, ins->edges, ins->img_ll_corr, ins->img_ll_cache,
                ins->prop_view, ins->prop_obj, img_ll_bg_prob(),
                img_ll_fg_prob(), img_ll_fg_noise_prob(), img_ll_fg_miss_prob(),
                img_ll_edge_noise_prob(), img_ll_edge_silho_miss_prob(),
                img_ll_edge_inner_miss_prob(), true);
    }
    catch (jwscxx::base::Arg_error e)
    {
        // Should never happen.
        abort();
    }
}

/**
 * @param  grad_out  Gradient of the negative log likelihood and prior of the
 *                   model parameters.
 * @param  c         Category index.
 */
void Metropolis::Parameters::proposal_grad_cat_potential
(
    jwsc::Vector_d** grad_out, 
    uint32_t         c
) 
const
{
    Category* cat = category(c);

    std::vector<const View*> ins_views;
    std::vector<const Object*> ins_objs;

    for (uint32_t i = 0; i < num_instances(); i++)
    {
        ins_views.push_back(instance(c, i)->prop_view);
        ins_objs.push_back(instance(c, i)->prop_obj);
    }

    try 
    {
        Likelihood::calc_grad_neg_log_likelihood_and_prior_wrt_cat_params(
                grad_out, cat->lp_potential, cat->prop_view, cat->prop_obj, 
                ins_views, ins_objs, true);
    }
    catch (jwscxx::base::Arg_error e)
    {
        // Should never happen.
        abort();
    }
}

/**
 * @param  q  Model parameter values.
 * @param  p  Model parameter momentum values.
 * @param  c  Index of the category.
 * @param  i  Index of the model instance to use.
 */
void Metropolis::Parameters::set_var_param_proposal_values_dynamics
(
    const jwsc::Vector_d* q,
    const jwsc::Vector_d* p,
    uint32_t              c,
    uint32_t              i
)
{
    using std::setw;
    using namespace jwsc;

    Vector_d* curr_q = 0;
    Vector_d* deltas = 0;

    get_var_param_proposal_values(&curr_q, c, i);
    subtract_vectors_d(&deltas, q, curr_q);

    add_var_param_proposal_values_scale_component_deps(deltas, c, i);

    update_proposal_ins_ll_kinetic(p, c, i);

    free_vector_d(curr_q);
    free_vector_d(deltas);
}

/**
 * @param  q  Model parameter values.
 * @param  p  Model parameter momentum values.
 * @param  c  Index of the category.
 */
void Metropolis::Parameters::set_var_param_proposal_stats_dynamics
(
    const jwsc::Vector_d* q,
    const jwsc::Vector_d* p,
    uint32_t              c
)
{
    set_var_param_proposal_stats(q, c);
    update_proposal_cat_lp_kinetic(p, c);
}

/** @brief Prints the variable parameter proposal information. */
void Metropolis::Parameters::print_var_param_proposal_info
(
    const std::vector<Metropolis::Move*>& moves
)
{
    uint32_t N_c = num_categories();
    uint32_t N_i = num_instances();

    for (uint32_t c = 0; c < N_c; c++)
    {
        Category* cat = category(c);
        cat->print_var_param_proposal_info(moves[ c ]->name());

        for (uint32_t i = 0; i < N_i; i++)
        {
            Instance* ins = instance(c, i);
            ins->print_var_param_proposal_info(moves[ c ]->name(),member(c, i));
        }
    }
}

void Metropolis::Parameters::update_proposal_ins_ll_kinetic
(
    const jwsc::Vector_d* p,
    uint32_t              c,
    uint32_t              i
)
{
    Instance* ins = instance(c, i);

    ins->prop_ll_kinetic = 0;
    calc_vector_dot_prod_d(&(ins->prop_ll_kinetic), p, p);
    ins->prop_ll_kinetic *= 0.5;
}

/**
 * @param  p  Model parameter point to calculate the kinetic energy of.
 * @param  c  Index of the category.
 */
void Metropolis::Parameters::update_proposal_cat_lp_kinetic
(
    const jwsc::Vector_d* p, 
    uint32_t              c
)
{
    Category* cat = category(c);

    cat->prop_lp_kinetic = 0;
    calc_vector_dot_prod_d(&(cat->prop_lp_kinetic), p, p);
    cat->prop_lp_kinetic *= 0.5;
}

void Metropolis::Parameters::init_cat_proposal(uint32_t c)
{ 
    Category* cat = category(c);

    cat->prop_lp_potential = cat->lp_potential;
    cat->prop_lp_kinetic   = cat->lp_kinetic;
    cat->prop_potential    = cat->potential;
    cat->prop_kinetic      = cat->kinetic;
}

void Metropolis::Parameters::init_ins_proposal(uint32_t c, uint32_t i)
{ 
    Instance* ins = instance(c, i);

    *(ins->prop_view)      = *(ins->view); 
    *(ins->prop_obj)       = *(ins->obj); 
    ins->prop_img_ll       = ins->img_ll; 
    ins->prop_ll_potential = ins->ll_potential; 
    ins->prop_ll_kinetic   = ins->ll_kinetic;
}

void Metropolis::Parameters::accept_cat_proposal(uint32_t c)
{
    Category* cat = category(c);

    cat->lp_potential = cat->prop_lp_potential;
    cat->lp_kinetic   = cat->prop_lp_kinetic;
    cat->potential    = cat->prop_potential;
    cat->kinetic      = cat->prop_kinetic;
}

void Metropolis::Parameters::accept_ins_proposal(uint32_t c, uint32_t i)
{
    Instance* ins = instance(c, i);

    *(ins->view)      = *(ins->prop_view);
    *(ins->obj)       = *(ins->prop_obj);
    ins->img_ll       = ins->prop_img_ll; 
    ins->ll_potential = ins->prop_ll_potential; 
    ins->ll_kinetic   = ins->prop_ll_kinetic; 

    if (ins->ll_potential < ins->best_ll_potential)
    {
        *(ins->best_view)      = *(ins->view);
        *(ins->best_obj)       = *(ins->obj);
        ins->best_img_ll       = ins->img_ll;
        ins->best_ll_potential = ins->ll_potential;
    }

    add_sample(ins);
}

void Metropolis::Parameters::update_cat_proposal_dependents(uint32_t c)
{
    Category* cat = category(c);

    double prop_view_lp = cat->prop_view->calc_var_param_prior_log_prob();
    double prop_obj_lp  = cat->prop_obj->calc_var_param_prior_log_prob();

    cat->prop_lp_potential = - (prop_view_lp + prop_obj_lp);
    cat->prop_potential = cat->prop_lp_potential;
    for (uint32_t i = 0; i < num_instances(); i++)
    {
        Instance* ins = instance(c, i);
        cat->prop_potential += ins->prop_ll_potential;
    }
}

/** Updates an instance that depends on the category parameters. */
void Metropolis::Parameters::update_ins_proposal_cat_dependents
(
    uint32_t c, 
    uint32_t i
)
{
    Instance* ins = instance(c, i);

    double prop_view_ll = ins->prop_view->calc_var_param_log_likelihood();
    double prop_obj_ll  = ins->prop_obj->calc_var_param_log_likelihood();

    ins->prop_ll_potential = - (ins->prop_img_ll + prop_view_ll + prop_obj_ll);
}

void Metropolis::Parameters::update_ins_proposal_dependents
(
    uint32_t c,
    uint32_t i
)
{
    Instance* ins = instance(c, i);

    // DEBUG
    ins->obj->ensure_component_consistency();
    ins->obj->assert_component_consistency();
    ins->prop_obj->ensure_component_consistency();
    ins->prop_obj->assert_component_consistency();

    Likelihood::calc_neg_log_likelihood(&(ins->prop_ll_potential),
            &(ins->prop_img_ll), &(ins->model_mat), &(ins->img_ll_surf_assign),
            &(ins->img_ll_edge_assign), ins->image_t, ins->edges,
            ins->img_ll_corr, ins->prop_view, ins->prop_obj, img_ll_bg_prob(),
            img_ll_fg_prob(), img_ll_fg_noise_prob(), img_ll_fg_miss_prob(),
            img_ll_edge_noise_prob(), img_ll_edge_silho_miss_prob(),
            img_ll_edge_inner_miss_prob());
}

/* ========================================================================== */






/* --------------------  Metropolis::Thread_manager  ------------------------ */

Metropolis::Thread_manager::Thread_manager(uint32_t thread)
    : Sampler::Thread_manager(thread)
{
    _category = 0;
}

Metropolis::Thread_manager::~Thread_manager() 
{ 
}

void Metropolis::Thread_manager::category(uint32_t c)
{
    _category = c;
}

uint32_t Metropolis::Thread_manager::category() const
{
    return _category;
}

void Metropolis::Thread_manager::sampler(Sampler* sampler)
    throw (jwscxx::base::Arg_error)
{
    if (!(_sampler = dynamic_cast<Metropolis*>(sampler)))
    {
        throw jwscxx::base::Arg_error(
                "Metropolis thread manager requires a metropolis sampler");
    }
}

Metropolis* Metropolis::Thread_manager::sampler() const
{
    return dynamic_cast<Metropolis*>(_sampler);
}

void Metropolis::Thread_manager::params
(
    Sampler::Parameters* params
)
throw (jwscxx::base::Arg_error)
{
    if (!(_params = dynamic_cast<Metropolis::Parameters*>(params)))
    {
        throw jwscxx::base::Arg_error(
                "Metropolis thread manager requires metropolis sampler params");
    }

    _moves.clear();
    _moves.insert(_moves.end(), this->params()->num_categories(), 0);
}

Metropolis::Parameters* Metropolis::Thread_manager::params() const
{
    return dynamic_cast<Metropolis::Parameters*>(_params);
}

void Metropolis::Thread_manager::move(Metropolis::Move* move, uint32_t c)
{
    try { _moves.at(c) = move; }
    catch (...) { abort(); }
}

Metropolis::Move* Metropolis::Thread_manager::move(uint32_t c)
{
    try { return _moves.at(c); }
    catch (...) { abort(); }
}

const std::vector<Metropolis::Move*>& Metropolis::Thread_manager::moves()
{
    return _moves;
}

void Metropolis::Thread_manager::select_sampler_moves()
    throw (jwscxx::base::Arg_error)
{
    using std::list;

    synchronize();

    if (last_thread())
    {
        uint32_t N_c = params()->num_categories();

        for (uint32_t c = 0; c < N_c; c++)
        {
            const Object* obj = params()->category(c)->obj;
            const std::vector<Component*>& components = obj->components();

            list<Metropolis::Move*>::iterator it;
            list<Metropolis::Move*> diff_moves = sampler()->diffusion_moves();
            for (it = diff_moves.begin(); it != diff_moves.end(); )
            {
                if ((*it)->has_supported_components(components))
                    it++;
                else
                    it = diff_moves.erase(it);
            }

            list<Metropolis::Move*> jump_moves = sampler()->jump_moves();
            for (it = jump_moves.begin(); it != jump_moves.end(); )
            {
                if ((*it)->has_supported_components(components))
                    it++;
                else
                    it = jump_moves.erase(it);
            }

            if (diff_moves.begin() == diff_moves.end() &&
                jump_moves.begin() == jump_moves.end())
            {
                throw jwscxx::base::Arg_error("No valid sampler moves");
            }

            double u = jwsc::sample_uniform_pdf_d(0, 1);
            double d = sampler()->diffusion_prob();
            double j = sampler()->jump_prob();
            list<Metropolis::Move*>* moves = &diff_moves;
            if (diff_moves.begin() == diff_moves.end() ||
                    (jump_moves.begin() != jump_moves.end() && u < (j/(d+j)))) 
            {
                moves = &jump_moves;
            }

            double p = 0;
            for (it = moves->begin(); it != moves->end(); it++)
            {
                p += (*it)->prob();
            }

            u = jwsc::sample_uniform_pdf_d(0, 1) * p;
            p = 0;

            for (it = moves->begin(); it != moves->end(); it++)
            {
                Metropolis::Move* m = *it;

                p += m->prob();
                if (u <= p)
                {
                    move(m, c);
                    break;
                }
            }
        }

        release();
    }
    else
    {
        wait();
    }

    desynchronize();
}

void Metropolis::Thread_manager::assign_min_potential_members()
{
    synchronize();

    if (last_thread())
    {
        params()->assign_min_potential_members();
        release();
    }
    else
    {
        wait();
    }

    desynchronize();
}

void Metropolis::Thread_manager::print_var_param_proposal_info()
{
    synchronize();

    if (last_thread())
    {
        params()->print_var_param_proposal_info(moves());
        release();
    }
    else
    {
        wait();
    }

    desynchronize();
}

/**
 * @param  p_out  Point to get the proposed model parameters into.
 * @param  mgr    Sampler thread manager.
 */
void Metropolis::Thread_manager::get_var_param_proposal_values_c
(
    jwsc::Vector_d**                  p_out, 
    const Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = mgr->params()->num_instances();
    uint32_t D_t = mgr->params()->num_var_params(c, t);
    uint32_t d   = 0;
    uint32_t D   = (N_i % N_t > t) ? ((N_i / N_t) + 1)*D_t : (N_i / N_t)*D_t;

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    Vector_d* p_i = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        mgr->params()->get_var_param_proposal_values(&p_i, c, i);
        copy_vector_section_into_vector_d(p, d, p_i, 0, D_t);
        d += D_t;
    }

    assert(d == D);
    free_vector_d(p_i);
}

/**
 * @param  p_out  Point to get the proposed model parameters into.
 * @param  mgr    Sampler thread manager.
 */
void Metropolis::Thread_manager::get_var_param_proposal_stats_c
(
    jwsc::Vector_d**                  p_out, 
    const Metropolis::Thread_manager* mgr
)
{
    uint32_t c = mgr->category();
    mgr->params()->get_var_param_stats(p_out, c);
}

/**
 * @param  p_out   Point to get the parameter stats and values into.
 * @param  mgr     Thread_manager to use.
 *
 * @return On success, 0 is returned. On error, a jwsc::Error is returned.
 */
void Metropolis::Thread_manager::get_var_param_proposal_stats_and_values_c
(
    jwsc::Vector_d**            p_out,
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->toggle_params_lock();

    uint32_t c    = mgr->category();
    uint32_t t    = mgr->thread();
    uint32_t N_t  = mgr->num_threads();
    uint32_t N_i  = mgr->params()->num_instances();
    uint32_t D_tv = mgr->params()->num_var_params(c, t);
    uint32_t D_ts = 2*D_tv;
    uint32_t d    = 0;
    uint32_t D    = (N_i % N_t > t) ? ((N_i / N_t) + 1)*D_tv : (N_i / N_t)*D_tv;

    if (t == 0)
    {
        D += D_ts;
    }

    create_vector_d(p_out, D);
    Vector_d* p = *p_out;
    Vector_d* p_i = 0;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    if (t == 0)
    {
        mgr->params()->get_var_param_proposal_stats(&p_i, c);
        copy_vector_section_into_vector_d(p, d, p_i, 0, D_ts);
        d += D_ts;
    }
    mgr->desynchronize();

    for (uint32_t i = t; i < N_i+t; i += N_t)
    {
        if (i < N_i)
        {
            mgr->params()->get_var_param_proposal_values(&p_i, c, i);
            copy_vector_section_into_vector_d(p, d, p_i, 0, D_tv);
            d += D_tv;
        }
        mgr->toggle_params_lock();
    }

    assert(get_num_unhandled_errors() == 0);
    assert(d == D);

    free_vector_d(p_i);
}

/**
 * @param  grad_out  Gradient of the negative log likelihood and prior of the
 *                   parameters in @em params
 * @param  mgr       Sampler thread manager.
 *
 * @return On success, 0 is returned. On error, a jwsc::Error is returned, 
 * @em *img_ll_out is freed and set to 0.
 */
jwsc::Error* 
Metropolis::Thread_manager::get_var_param_proposal_values_grad_ins_potential_c
(
    jwsc::Vector_d**            grad_out, 
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = mgr->params()->num_instances();
    uint32_t D_t = mgr->params()->num_var_params(c, t);
    uint32_t d   = 0;
    uint32_t D   = (N_i % N_t > t) ? ((N_i / N_t) + 1)*D_t : (N_i / N_t)*D_t;

    create_vector_d(grad_out, D);
    Vector_d* grad = *grad_out;
    Vector_d* grad_i = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        mgr->params()->proposal_grad_ins_potential(&grad_i, c, i);
        copy_vector_section_into_vector_d(grad, d, grad_i, 0, D_t);
        d += D_t;
    }

    assert(d == D);
    free_vector_d(grad_i);

    return 0;
}

/**
 * @param  grad_out  Gradient of the negative log likelihood and prior of the
 *                   parameters in @em params
 * @param  mgr       Sampler thread manager.
 *
 * @return On success, 0 is returned. On error, a jwsc::Error is returned, 
 * @em *img_ll_out is freed and set to 0.
 */
jwsc::Error* 
Metropolis::Thread_manager::get_var_param_proposal_stats_grad_cat_potential_c
(
    jwsc::Vector_d**            grad_out, 
    Metropolis::Thread_manager* mgr
)
{
    uint32_t c = mgr->category();
    mgr->params()->proposal_grad_cat_potential(grad_out, c);
    return 0;
}

/**
 * @param  grad_out  Gradient of the negative log likelihood and prior of the
 *                   parameters in @em params
 * @param  mgr       Thread_manager to use.
 *
 * @return On success, 0 is returned. On error, a jwsc::Error is returned.
 */
jwsc::Error* Metropolis::Thread_manager::get_var_param_proposal_stats_and_values_grad_potential_c
(
    jwsc::Vector_d**            grad_out,
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->toggle_params_lock();

    uint32_t c    = mgr->category();
    uint32_t t    = mgr->thread();
    uint32_t N_t  = mgr->num_threads();
    uint32_t N_i  = mgr->params()->num_instances();
    uint32_t D_tv = mgr->params()->num_var_params(c, t);
    uint32_t D_ts = 2*D_tv;
    uint32_t d    = 0;
    uint32_t D    = (N_i % N_t > t) ? ((N_i / N_t) + 1)*D_tv : (N_i / N_t)*D_tv;

    if (t == 0)
    {
        D += D_ts;
    }

    create_vector_d(grad_out, D);
    Vector_d* grad = *grad_out;
    Vector_d* grad_i = 0;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    if (t == 0)
    {
        mgr->params()->proposal_grad_cat_potential(&grad_i, c);
        copy_vector_section_into_vector_d(grad, d, grad_i, 0, D_ts);
        d += D_ts;
    }
    mgr->desynchronize();

    for (uint32_t i = t; i < N_i+t; i += N_t)
    {
        if (i < N_i)
        {
            mgr->params()->proposal_grad_ins_potential(&grad_i, c, i);
            copy_vector_section_into_vector_d(grad, d, grad_i, 0, D_tv);
            d += D_tv;
        }
        mgr->toggle_params_lock();
    }

    assert(get_num_unhandled_errors() == 0);
    assert(d == D);

    free_vector_d(grad_i);

    return 0;
}

/**
 * @param  q    Model parameter values.
 * @param  p    Model parameter momentum values.
 * @param  mgr  Thread_manager to use.
 *
 * @return On success, 0 is returned. On error, a jwsc::Error is returned.
 */
jwsc::Error* 
Metropolis::Thread_manager::set_var_param_proposal_values_dynamics_c
(
    const jwsc::Vector_d*       q,
    const jwsc::Vector_d*       p,
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = mgr->params()->num_instances();
    uint32_t D_t = mgr->params()->num_var_params(c, t);
    uint32_t d   = 0;
    uint32_t D   = (N_i % N_t > t) ? ((N_i / N_t) + 1)*D_t : (N_i / N_t)*D_t;

    Vector_d* q_i = 0;
    Vector_d* p_i = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        copy_vector_section_d(&q_i, q, d, D_t);
        copy_vector_section_d(&p_i, p, d, D_t);

        mgr->params()->set_var_param_proposal_values_dynamics(q_i, p_i, c, i);
        d += D_t;
    }

    assert(get_num_unhandled_errors() == 0);

    assert(d == D);
    free_vector_d(q_i);
    free_vector_d(p_i);

    return 0;
}

/**
 * @param  q    Model parameter values.
 * @param  p    Model parameter momentum values.
 * @param  mgr  Thread_manager to use.
 *
 * @return On success, 0 is returned. On error, a jwsc::Error is returned.
 */
jwsc::Error* 
Metropolis::Thread_manager::set_var_param_proposal_stats_dynamics_c
(
    const jwsc::Vector_d*       q,
    const jwsc::Vector_d*       p,
    Metropolis::Thread_manager* mgr
)
{
    uint32_t c = mgr->category();
    mgr->params()->set_var_param_proposal_stats_dynamics(q, p, c);
    return 0;
}

/**
 * @param  q    Model parameter values.
 * @param  p    Model parameter momentum values.
 * @param  mgr  Thread_manager to use.
 *
 * @return On success, 0 is returned. On error, a jwsc::Error is returned.
 */
jwsc::Error* 
Metropolis::Thread_manager::set_var_param_proposal_stats_and_values_dynamics_c
(
    const jwsc::Vector_d*       q,
    const jwsc::Vector_d*       p,
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->toggle_params_lock();

    uint32_t c    = mgr->category();
    uint32_t t    = mgr->thread();
    uint32_t N_t  = mgr->num_threads();
    uint32_t N_i  = mgr->params()->num_instances();
    uint32_t D_tv = mgr->params()->num_var_params(c, t);
    uint32_t D_ts = 2*D_tv;
    uint32_t d    = 0;
    uint32_t D    = (N_i % N_t > t) ? ((N_i / N_t) + 1)*D_tv : (N_i / N_t)*D_tv;

    Vector_d* q_i = 0;
    Vector_d* p_i = 0;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        for (uint32_t i = 0; i < N_i; i++)
        {
            Instance* ins = mgr->params()->instance(c, i);
            ins->accepted = true;
            ins->iteration++;
        }
        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    if (t == 0)
    {
        copy_vector_section_d(&q_i, q, d, D_ts);
        copy_vector_section_d(&p_i, p, d, D_ts);
        mgr->params()->set_var_param_proposal_stats_dynamics(q_i, p_i, c);
        d += D_ts;
        D += D_ts;
    }
    mgr->desynchronize();

    for (uint32_t i = t; i < N_i+t; i += N_t)
    {
        if (i < N_i)
        {
            copy_vector_section_d(&q_i, q, d, D_tv);
            copy_vector_section_d(&p_i, p, d, D_tv);
            mgr->params()->set_var_param_proposal_values_dynamics(q_i, p_i,c,i);
            d += D_tv;
        }
        mgr->toggle_params_lock();
    }

    mgr->synchronize();
    if (mgr->last_thread())
    {
        mgr->params()->accept_cat_proposal(c);
        mgr->params()->category(c)->print_var_param_proposal_info(
                mgr->move(c)->name());

        for (uint32_t i = 0; i < N_i; i++)
        {
            mgr->params()->accept_ins_proposal(c, i);
            mgr->params()->instance(c, i)->print_var_param_proposal_info(
                    mgr->move(c)->name(), mgr->params()->member(c, i));
        }
        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    mgr->desynchronize();

    assert(get_num_unhandled_errors() == 0);
    assert(d == D);

    free_vector_d(q_i);
    free_vector_d(p_i);

    return 0;
}

/**
 * @param  delta_t_out  Stochastic dynamics delta_t values for the parameters.
 * @param  mgr          Metropolis::Thread_manager to use.
 */
void Metropolis::Thread_manager::get_var_param_values_dynamics_delta_t_c
(
    jwsc::Vector_d**            delta_t_out, 
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = mgr->params()->num_instances();
    uint32_t D_t = mgr->params()->num_var_params(c, t);
    uint32_t d   = 0;
    uint32_t D   = (N_i % N_t > t) ? ((N_i / N_t) + 1)*D_t : (N_i / N_t)*D_t;

    create_vector_d(delta_t_out, D);
    Vector_d* delta_t = *delta_t_out;
    Vector_d* delta_t_i = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        mgr->params()->get_var_param_values_dynamics_delta_t(&delta_t_i, c, i);
        copy_vector_section_into_vector_d(delta_t, d, delta_t_i, 0, D_t);
        d += D_t;
    }

    assert(d == D);
    free_vector_d(delta_t_i);
}

/**
 * @param  delta_t_out  Stochastic dynamics delta_t values for the parameters.
 * @param  mgr          Metropolis::Thread_manager to use.
 */
void Metropolis::Thread_manager::get_var_param_stats_dynamics_delta_t_c
(
    jwsc::Vector_d**            delta_t_out, 
    Metropolis::Thread_manager* mgr
)
{
    uint32_t c = mgr->category();
    mgr->params()->get_var_param_stats_dynamics_delta_t(delta_t_out, c);
}

/**
 * @param  delta_t_out  Stochastic dynamics delta_t values for the parameters.
 * @param  mgr          Metropolis::Thread_manager to use.
 */
void Metropolis::Thread_manager::get_var_param_stats_and_values_dynamics_delta_t_c
(
    jwsc::Vector_d**            delta_t_out, 
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->toggle_params_lock();

    uint32_t c    = mgr->category();
    uint32_t t    = mgr->thread();
    uint32_t N_t  = mgr->num_threads();
    uint32_t N_i  = mgr->params()->num_instances();
    uint32_t D_tv = mgr->params()->num_var_params(c, t);
    uint32_t D_ts = 2*D_tv;
    uint32_t d    = 0;
    uint32_t D    = (N_i % N_t > t) ? ((N_i / N_t) + 1)*D_tv : (N_i / N_t)*D_tv;

    if (t == 0)
    {
        D += D_ts;
    }

    create_vector_d(delta_t_out, D);
    Vector_d* delta_t = *delta_t_out;
    Vector_d* delta_t_i = 0;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    if (t == 0)
    {
        mgr->params()->get_var_param_stats_dynamics_delta_t(&delta_t_i, c);
        copy_vector_section_into_vector_d(delta_t, d, delta_t_i, 0, D_ts);
        d += D_ts;
    }
    mgr->desynchronize();

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        mgr->params()->get_var_param_values_dynamics_delta_t(&delta_t_i, c, i);
        copy_vector_section_into_vector_d(delta_t, d, delta_t_i, 0, D_tv);
        d += D_tv;
    }

    assert(get_num_unhandled_errors() == 0);
    assert(d == D);

    free_vector_d(delta_t_i);
}

std::vector<Metropolis::Move*> Metropolis::Thread_manager::_moves;

/* ========================================================================== */






/* -----------------------  Metropolis::Move  ------------------------------- */

/**
 * @param  name  Name of the move.
 * @param  p     Probability of the move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Move::Move(const char* name, double p) 
    throw (jwscxx::base::Arg_error) 
    : _name(name)
{
    prob(p);
}

Metropolis::Move::~Move()
{
}

/**
 * @param  mgr  Thread manager for the move.
 */
void Metropolis::Move::run(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    init(mgr);
    run_thread(mgr);
    accept_or_reject(mgr);
    assert_params(mgr);
}

const char* Metropolis::Move::name() const 
{ 
    return _name.c_str();
}

double Metropolis::Move::prob() const 
{ 
    return _prob; 
}

/**
 * @param  p  New probability of the move.
 *
 * @throw  jwscxx::base::Arg_error  Value must be in [0,1].
 */
void Metropolis::Move::prob(double p) throw (jwscxx::base::Arg_error)
{
    _prob = p;
}

void Metropolis::Move::init(Metropolis::Thread_manager* mgr)
{
    if (mgr->thread() > 0)
        return;

    Category* cat = mgr->params()->category(mgr->category());
    zero_timer(&(cat->timer));
    start_timer(&(cat->timer));
}

void Metropolis::Move::assert_params(Metropolis::Thread_manager* mgr)
{
    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();
        uint32_t N_i = params->num_instances();
        uint32_t N_n = params->instance(c, 0)->obj->num_components();

        for (uint32_t n = 0; n < N_n; n++)
        {
            for (uint32_t i = 0; i < N_i; i++)
            {
                Instance* ins = params->instance(c, i);
                assert(N_n == ins->obj->num_components());

                Component* c1 = ins->obj->component(n);
                for (uint32_t m1 = 0; m1 < c1->num_ctr_attach_pts(); m1++)
                {
                    Component* c2;
                    if ((c2 = c1->attachment(m1)))
                    {
                        uint32_t m2;
                        try { m2 = c2->attachment(c1); }
                        catch (jwscxx::base::Arg_error e) { abort(); }
                        if (m1%2 > 0)
                            assert(m1 == m2+1);
                        else
                            assert(m1+1 == m2);
                    }
                }
            }
        }

        for (uint32_t n = 0; n < N_n; n++)
        {
            Instance* ins = params->instance(c, 0);
            Component* c1 = ins->obj->component(n);

            for (uint32_t m1 = 0; m1 < c1->num_ctr_attach_pts(); m1++)
            {
                bool c1_attach_0 = (c1->attachment(m1)) ? true : false;
                bool c1_attach_m;

                for (uint32_t i = 0; i < N_i; i++)
                {
                    ins = params->instance(c, i);
                    c1 = ins->obj->component(n);
                    c1_attach_m = (c1->attachment(m1)) ? true : false;
                    assert(c1_attach_0 == c1_attach_m);
                }
            }
        }

        for (uint32_t n = 0; n < N_n; n++)
        {
            Instance* ins = params->instance(c, 0);
            uint32_t id = ins->obj->component(n)->id();
            for (uint32_t i = 0; i < N_i; i++)
            {
                ins = params->instance(c, i);
                assert(id == ins->obj->component(n)->id());
            }
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Move::clip_param_val
(
    double* x, 
    const Model::Parameter::Context& ctx
) 
const
{
    static const double epsilon = 1.0e-6;

    if (*x < ctx.range().min()+epsilon)
    {
        *x = ctx.range().min()+epsilon;
    }
    if (*x > ctx.range().max()-epsilon)
    {
        *x = ctx.range().max()-epsilon;
    }
}

/* ========================================================================== */







/* ----------------------  Metropolos::Hybrid_dynamics  --------------------- */

Metropolis::Hybrid_dynamics::Hybrid_dynamics
(
    double      alpha,
    uint32_t    kick,
    uint32_t    leapfrog_i
) 
throw (jwscxx::base::Arg_error) 
{
    this->alpha(alpha);
    this->kick(kick);
    this->leapfrog_i(leapfrog_i);
}

double Metropolis::Hybrid_dynamics::alpha() const
{
    return _alpha;
}

void Metropolis::Hybrid_dynamics::alpha(double a) 
    throw (jwscxx::base::Arg_error)
{
    if (a < 0 || a > 1)
    {
        throw jwscxx::base::Arg_error("Alpha not in [0,1]");
    }

    _alpha = a;
}

uint32_t Metropolis::Hybrid_dynamics::kick() const
{
    return _kick;
}

void Metropolis::Hybrid_dynamics::kick(uint32_t k) 
    throw (jwscxx::base::Arg_error)
{
    _kick = k;
}

uint32_t Metropolis::Hybrid_dynamics::leapfrog_i() const
{
    return _leapfrog_i;
}

void Metropolis::Hybrid_dynamics::leapfrog_i(uint32_t i)
{
    _leapfrog_i = i;
}

/* ========================================================================== */






/* --------------  Metropolis::Hybrid_dynamics::Stochastic  ----------------- */

/**
 * @param  iterations  Number of iterations to run.
 * @param  alpha  Proportion of momentum to conserve per iteration. Must be in
 *                [0,1]. (1-alpha*alpha) scales gaussian noise added to the
 *                momentum per iteration.
 * @param  kick   Frequency to kick the momenta of the simulated particle. 
 *
 * @throw  jwscxx::base::Arg_error  Invalid arguments.
 */
Metropolis::Hybrid_dynamics::Stochastic::Stochastic
(
    uint32_t iterations,
    double   alpha, 
    uint32_t kick
)
throw (jwscxx::base::Arg_error)
    : Stochastic_dynamics(alpha, kick)
{
    _iterations = iterations;
}

uint32_t Metropolis::Hybrid_dynamics::Stochastic::iterations() const
{
    return _iterations;
}

void Metropolis::Hybrid_dynamics::Stochastic::iterations(uint32_t i) 
    throw (jwscxx::base::Arg_error)
{
    _iterations = i;
}

void Metropolis::Hybrid_dynamics::Stochastic::run_thread
(
    Metropolis::Thread_manager* mgr
) 
const throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    typedef void   (*f1)(Vector_d**, const void*);
    typedef Error* (*f2)(Vector_d**, void*);
    typedef Error* (*f3)(const Vector_d*, const Vector_d*, void*);

    Vector_d* delta_t = 0;
    mgr->get_var_param_stats_and_values_dynamics_delta_t_c(&delta_t, mgr);

    Error* err = stochastic_dynamics_2_d(iterations(), delta_t, alpha(), 
            kick(), mgr,
            (f1)(mgr->get_var_param_proposal_stats_and_values_c),
            (f2)(mgr->get_var_param_proposal_stats_and_values_grad_potential_c),
            (f3)(mgr->set_var_param_proposal_stats_and_values_dynamics_c));

    free_vector_d(delta_t);

    if (err)
    {
        throw jwscxx::base::Exception(err->msg);
    }
}

/* ========================================================================== */






/* ---------------------  Metropolis::Diffusion  ---------------------------- */

/**
 * @param  name  Name of the move.
 * @param  p     Probability of the move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 *
 * @note Use Metropolis::normalize_move_probs() to normalize all the move
 * probabilities so they sum to 1.
 */
Metropolis::Diffusion::Diffusion(const char* name, double p) 
    throw (jwscxx::base::Arg_error)
    : Move(name, p)
{
}

Metropolis::Diffusion::~Diffusion()
{
}

/* ========================================================================== */







/* ---------------------  Metropolis::Diffusion_ins  ------------------------ */

/**
 * @param  name  Name of the move.
 * @param  p     Probability of the move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 *
 * @note Use Metropolis::normalize_move_probs() to normalize all the move
 * probabilities so they sum to 1.
 */
Metropolis::Diffusion_ins::Diffusion_ins(const char* name, double p) 
    throw (jwscxx::base::Arg_error)
    : Diffusion(name, p)
{
}

Metropolis::Diffusion_ins::~Diffusion_ins()
{
}

void Metropolis::Diffusion_ins::init(Metropolis::Thread_manager* mgr)
{
    Move::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        uint32_t N_i = mgr->params()->num_instances();
        uint32_t c   = mgr->category();

        mgr->params()->init_cat_proposal(c);

        _log_alpha.clear();
        _log_alpha.insert(_log_alpha.end(), N_i, 0);

        _reject.clear();
        _reject.insert(_reject.end(), N_i, false);

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Diffusion_ins::accept_or_reject
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();

        uint32_t N_i = params->num_instances();
        uint32_t c   = mgr->category();

        Category* cat = params->category(c);
        bool cat_accept = false;

        for (uint32_t i = 0; i < N_i; i++)
        {
            Instance* ins = params->instance(c, i);

            double log_u = log(sample_uniform_pdf_d(JWSC_MIN_LOG_ARG, 1.0));
            ins->accepted = (log_u <= log_alpha(i) && !reject(i));
            //ins->accepted = !reject(i);

            if (ins->accepted)
            {
                cat->prop_potential -= ins->ll_potential;
                cat->prop_potential += ins->prop_ll_potential;
                params->accept_ins_proposal(c, i);
            }

            cat_accept = (cat_accept || ins->accepted);

            ins->iteration++;
        }

        if (cat_accept)
        {
            params->accept_cat_proposal(c);
        }

        stop_timer(&(cat->timer));

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

double Metropolis::Diffusion_ins::log_alpha(size_t i) const
{
    try { return _log_alpha.at(i); }
    catch (...) { abort(); }
}

void Metropolis::Diffusion_ins::log_alpha(double log_alpha, size_t i) 
{
    try { _log_alpha.at(i) = log_alpha; }
    catch (...) { abort(); }
}

double Metropolis::Diffusion_ins::reject(size_t i) const
{
    try { return _reject.at(i); }
    catch (...) { abort(); }
}

void Metropolis::Diffusion_ins::reject(bool reject, size_t i)
{
    try { _reject.at(i) = reject; }
    catch (...) { abort(); }
}

/* ========================================================================== */







/* ------------------------  Metropolis::Swap  ------------------------------ */

Metropolis::Swap::Swap(const char* name, double prob) 
    throw (jwscxx::base::Arg_error) : Diffusion_ins(name, prob)
{
}

Metropolis::Swap* Metropolis::Swap::clone() const
{
    return new Swap(*this);
}

bool Metropolis::Swap::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    return (components.size() > 0);
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::Swap::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    Vector_d* p = 0;
    Vector_d* q = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        Instance* ins = params->instance(c, i);

        params->init_ins_proposal(c, i);
        double u = jwsc::sample_uniform_pdf_d(0, 1);

        try 
        { 
            ins->prop_obj->test_swap_object_coords(); 
            ins->prop_obj->swap_object_coords((u < 0.5)); 
        }
        catch (jwscxx::base::Arg_error e) 
        { 
            reject(true, i);
        }

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
        assert(!std::isinf(la) && !std::isnan(la));
#endif

        log_alpha(la, i);
    }

    free_vector_d(p);
    free_vector_d(q);
}

/* ========================================================================== */







/* ------------------------  Metropolis::Gauss  ----------------------------- */

Metropolis::Gauss::Gauss(const char* name, double prob) 
    throw (jwscxx::base::Arg_error) : Diffusion_ins(name, prob)
{
}

Metropolis::Gauss* Metropolis::Gauss::clone() const
{
    return new Gauss(*this);
}

bool Metropolis::Gauss::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    return (components.size() > 0);
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::Gauss::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    Vector_d* p = 0;
    Vector_d* q = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        Instance* ins = params->instance(c, i);

        params->init_ins_proposal(c, i);

        ins->view->get_var_param_values(&p);
        ins->view->sample_var_param_gauss_proposal(&q, p);
        try { ins->prop_view->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }

        ins->obj->get_var_param_values(&p);
        ins->obj->sample_var_param_gauss_proposal(&q, p);
        try { ins->prop_obj->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_view->calc_var_param_log_likelihood();
        la -= ins->view->calc_var_param_log_likelihood();
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
        assert(!std::isinf(la) && !std::isnan(la));
#endif

        log_alpha(la, i);
    }

    free_vector_d(p);
    free_vector_d(q);
}

/* ========================================================================== */







/* ---------------------  Metropolis::MV_gauss_shift  ----------------------- */

Metropolis::MV_gauss_shift::MV_gauss_shift(const char* name, double prob) 
    throw (jwscxx::base::Arg_error) : Diffusion_ins(name, prob)
{
}

Metropolis::MV_gauss_shift::~MV_gauss_shift()
{
}

Metropolis::MV_gauss_shift* Metropolis::MV_gauss_shift::clone() const
{
    return new MV_gauss_shift(*this);
}

bool Metropolis::MV_gauss_shift::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    return (components.size() > 0);
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::MV_gauss_shift::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    Vector_d* mean  = 0;
    Matrix_d* cov   = 0;
    Vector_d* sigma = 0;
    Vector_d* p     = 0;
    Vector_d* q     = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        Instance* ins = params->instance(c, i);

        params->init_ins_proposal(c, i);

        uint32_t D_1 = ins->view->num_var_params();
        uint32_t D_2 = ins->obj->num_var_params();
        uint32_t D   = D_1 + D_2;

        create_zero_vector_d(&mean, D);

        if (ins->sample_cov_in)
        {
            copy_matrix_d(&cov, ins->sample_cov_in);
        }
        else
        {
            create_zero_matrix_d(&cov, D, D);
            uint32_t dd = 0;

            ins->view->get_var_param_gauss_proposal_sigma(&sigma);
            for (uint32_t d = 0; d < D_1; d++)
            {
                cov->elts[ dd ][ dd ] = sigma->elts[ d ]*sigma->elts[ d ]; 
                dd++;
            }

            ins->obj->get_var_param_gauss_proposal_sigma(&sigma);
            for (uint32_t d = 0; d < D_2; d++)
            {
                cov->elts[ dd ][ dd ] = sigma->elts[ d ]*sigma->elts[ d ];
                dd++;
            }
        }

        assert(sample_mv_gaussian_pdf_d(&p, mean, cov) == 0);

        assert(copy_vector_section_d(&q, p, 0, D_1) == 0);
        try { ins->prop_view->add_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }

        assert(copy_vector_section_d(&q, p, D_1, D_2) == 0);
        try { ins->prop_obj->add_var_param_values_shift_component_deps(q); }
        catch (jwscxx::base::Arg_error e) { }

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_view->calc_var_param_log_likelihood();
        la -= ins->view->calc_var_param_log_likelihood();
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
        assert(!std::isinf(la) && !std::isnan(la));
#endif

        log_alpha(la, i);
    }

    free_vector_d(mean);
    free_matrix_d(cov);
    free_vector_d(sigma);
    free_vector_d(p);
    free_vector_d(q);
}

/* ========================================================================== */







/* -----------------------  Metropolis::View_gauss  ------------------------- */

Metropolis::View_gauss::View_gauss(const char* name, double prob) 
    throw (jwscxx::base::Arg_error) : MV_gauss_shift(name, prob)
{
}

Metropolis::View_gauss* Metropolis::View_gauss::clone() const
{
    return new View_gauss(*this);
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::View_gauss::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    Vector_d* mean  = 0;
    Matrix_d* cov   = 0;
    Vector_d* sigma = 0;
    Vector_d* p     = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        Instance* ins = params->instance(c, i);

        params->init_ins_proposal(c, i);

        uint32_t D = ins->view->num_var_params();

        create_zero_vector_d(&mean, D);

        if (ins->sample_cov_in)
        {
            copy_matrix_d(&cov, ins->sample_cov_in);
        }
        else
        {
            create_zero_matrix_d(&cov, D, D);
            uint32_t dd = 0;

            ins->view->get_var_param_gauss_proposal_sigma(&sigma);
            for (uint32_t d = 0; d < D; d++)
            {
                cov->elts[ dd ][ dd ] = sigma->elts[ d ]*sigma->elts[ d ]; 
                dd++;
            }
        }

        assert(sample_mv_gaussian_pdf_d(&p, mean, cov) == 0);

        try { ins->prop_view->add_var_param_values(p); }
        catch (jwscxx::base::Arg_error e) { }

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_view->calc_var_param_log_likelihood();
        la -= ins->view->calc_var_param_log_likelihood();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
        assert(!std::isinf(la) && !std::isnan(la));
#endif

        log_alpha(la, i);
    }

    free_vector_d(mean);
    free_matrix_d(cov);
    free_vector_d(sigma);
    free_vector_d(p);
}

/* ========================================================================== */







/* -------------------  Metropolis::Object_gauss_shift  --------------------- */

Metropolis::Object_gauss_shift::Object_gauss_shift
(
    const char* name, 
    double prob
)
throw (jwscxx::base::Arg_error) : MV_gauss_shift(name, prob)
{
}

Metropolis::Object_gauss_shift* Metropolis::Object_gauss_shift::clone() const
{
    return new Object_gauss_shift(*this);
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::Object_gauss_shift::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    Vector_d* mean  = 0;
    Matrix_d* cov   = 0;
    Vector_d* sigma = 0;
    Vector_d* p     = 0;
    Vector_d* q     = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        Instance* ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        uint32_t D = ins->obj->Model::num_var_params();
        create_zero_vector_d(&mean, D);

        if (ins->sample_cov_in)
        {
            copy_matrix_d(&cov, ins->sample_cov_in);
        }
        else
        {
            create_zero_matrix_d(&cov, D, D);
            uint32_t dd = 0;

            ins->obj->Model::get_var_param_gauss_proposal_sigma(&sigma);
            for (uint32_t d = 0; d < D; d++)
            {
                cov->elts[ dd ][ dd ] = sigma->elts[ d ]*sigma->elts[ d ]; 
                dd++;
            }
        }

        assert(sample_mv_gaussian_pdf_d(&p, mean, cov) == 0);
        create_zero_vector_d(&q, ins->obj->num_var_params());

        assert(copy_vector_section_into_vector_d(q, 0, p, 0, D) == 0);
        try { ins->prop_obj->add_var_param_values_shift_component_deps(q); }
        catch (jwscxx::base::Arg_error e) { }

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
        assert(!std::isinf(la) && !std::isnan(la));
#endif

        log_alpha(la, i);
    }

    free_vector_d(mean);
    free_matrix_d(cov);
    free_vector_d(sigma);
    free_vector_d(p);
    free_vector_d(q);
}

/* ========================================================================== */







/* -----------------  Metropolis::Component_gauss_shift  -------------------- */

Metropolis::Component_gauss_shift::Component_gauss_shift
(
    const char* name, 
    double prob
)
throw (jwscxx::base::Arg_error) : MV_gauss_shift(name, prob)
{
}

Metropolis::Component_gauss_shift::~Component_gauss_shift()
{
}

Metropolis::Component_gauss_shift* Metropolis::Component_gauss_shift::clone() 
    const
{
    return new Component_gauss_shift(*this);
}

uint32_t Metropolis::Component_gauss_shift::comp_i()
{
    return _comp_i;
}

uint32_t Metropolis::Component_gauss_shift::comp_n()
{
    return _comp_n;
}

void Metropolis::Component_gauss_shift::init
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;
    using jwscxx::base::Arg_error;

    MV_gauss_shift::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        try { _comp_i = cat->obj->select_random_component(); }
        catch (Arg_error e) { abort(); }

        _comp_n = cat->obj->Model::num_var_params();
        for (uint32_t i = 0; i < _comp_i; i++)
        {
            _comp_n += cat->obj->component(i)->num_var_params();
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::Component_gauss_shift::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    Vector_d* mean  = 0;
    Matrix_d* cov   = 0;
    Vector_d* sigma = 0;
    Vector_d* p     = 0;
    Vector_d* q     = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        Instance* ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Component* comp = ins->prop_obj->component(comp_i());
        uint32_t D = comp->num_var_params();
        create_zero_vector_d(&mean, D);

        if (ins->sample_cov_in)
        {
            copy_matrix_d(&cov, ins->sample_cov_in);
        }
        else
        {
            create_zero_matrix_d(&cov, D, D);
            uint32_t dd = 0;

            comp->get_var_param_gauss_proposal_sigma(&sigma);
            for (uint32_t d = 0; d < D; d++)
            {
                cov->elts[ dd ][ dd ] = sigma->elts[ d ]*sigma->elts[ d ]; 
                dd++;
            }
        }

        assert(sample_mv_gaussian_pdf_d(&p, mean, cov) == 0);
        create_zero_vector_d(&q, ins->obj->num_var_params());

        assert(copy_vector_section_into_vector_d(q, comp_n(), p, 0, D) == 0);
        try { ins->prop_obj->add_var_param_values_shift_component_deps(q); }
        catch (jwscxx::base::Arg_error e) { }

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
        assert(!std::isinf(la) && !std::isnan(la));
#endif

        log_alpha(la, i);
    }

    free_vector_d(mean);
    free_matrix_d(cov);
    free_vector_d(sigma);
    free_vector_d(p);
    free_vector_d(q);
}

/* ========================================================================== */







/* --------------------  Metropolis::MV_gauss_scale  ------------------------ */

Metropolis::MV_gauss_scale::MV_gauss_scale(const char* name, double prob) 
    throw (jwscxx::base::Arg_error) : Diffusion_ins(name, prob)
{
}

Metropolis::MV_gauss_scale::~MV_gauss_scale()
{
}

Metropolis::MV_gauss_scale* Metropolis::MV_gauss_scale::clone() const
{
    return new MV_gauss_scale(*this);
}

bool Metropolis::MV_gauss_scale::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    return (components.size() > 0);
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::MV_gauss_scale::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    Vector_d* mean  = 0;
    Matrix_d* cov   = 0;
    Vector_d* sigma = 0;
    Vector_d* p     = 0;
    Vector_d* q     = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        Instance* ins = params->instance(c, i);

        params->init_ins_proposal(c, i);

        uint32_t D_1 = ins->view->num_var_params();
        uint32_t D_2 = ins->obj->num_var_params();
        uint32_t D   = D_1 + D_2;

        create_zero_vector_d(&mean, D);

        if (ins->sample_cov_in)
        {
            copy_matrix_d(&cov, ins->sample_cov_in);
        }
        else
        {
            create_zero_matrix_d(&cov, D, D);
            uint32_t dd = 0;

            ins->view->get_var_param_gauss_proposal_sigma(&sigma);
            for (uint32_t d = 0; d < D_1; d++)
            {
                cov->elts[ dd ][ dd ] = sigma->elts[ d ]*sigma->elts[ d ]; 
                dd++;
            }

            ins->obj->get_var_param_gauss_proposal_sigma(&sigma);
            for (uint32_t d = 0; d < D_2; d++)
            {
                cov->elts[ dd ][ dd ] = sigma->elts[ d ]*sigma->elts[ d ];
                dd++;
            }
        }

        assert(sample_mv_gaussian_pdf_d(&p, mean, cov) == 0);

        assert(copy_vector_section_d(&q, p, 0, D_1) == 0);
        try { ins->prop_view->add_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }

        assert(copy_vector_section_d(&q, p, D_1, D_2) == 0);
        try { ins->prop_obj->add_var_param_values_scale_component_deps(q); }
        catch (jwscxx::base::Arg_error e) { }

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_view->calc_var_param_log_likelihood();
        la -= ins->view->calc_var_param_log_likelihood();
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
        assert(!std::isinf(la) && !std::isnan(la));
#endif

        log_alpha(la, i);
    }

    free_vector_d(mean);
    free_matrix_d(cov);
    free_vector_d(sigma);
    free_vector_d(p);
    free_vector_d(q);
}

/* ========================================================================== */







/* ------------------  Metropolis::Object_gauss_scale  ---------------------- */

Metropolis::Object_gauss_scale::Object_gauss_scale
(
    const char* name, 
    double prob
)
throw (jwscxx::base::Arg_error) : MV_gauss_scale(name, prob)
{
}

Metropolis::Object_gauss_scale* Metropolis::Object_gauss_scale::clone() const
{
    return new Object_gauss_scale(*this);
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::Object_gauss_scale::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    Vector_d* mean  = 0;
    Matrix_d* cov   = 0;
    Vector_d* sigma = 0;
    Vector_d* p     = 0;
    Vector_d* q     = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        Instance* ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        uint32_t D = ins->obj->Model::num_var_params();
        create_zero_vector_d(&mean, D);

        if (ins->sample_cov_in)
        {
            copy_matrix_d(&cov, ins->sample_cov_in);
        }
        else
        {
            create_zero_matrix_d(&cov, D, D);
            uint32_t dd = 0;

            ins->obj->Model::get_var_param_gauss_proposal_sigma(&sigma);
            for (uint32_t d = 0; d < D; d++)
            {
                cov->elts[ dd ][ dd ] = sigma->elts[ d ]*sigma->elts[ d ]; 
                dd++;
            }
        }

        assert(sample_mv_gaussian_pdf_d(&p, mean, cov) == 0);
        create_zero_vector_d(&q, ins->obj->num_var_params());

        assert(copy_vector_section_into_vector_d(q, 0, p, 0, D) == 0);
        try { ins->prop_obj->add_var_param_values_scale_component_deps(q); }
        catch (jwscxx::base::Arg_error e) { }

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
        assert(!std::isinf(la) && !std::isnan(la));
#endif

        log_alpha(la, i);
    }

    free_vector_d(mean);
    free_matrix_d(cov);
    free_vector_d(sigma);
    free_vector_d(p);
    free_vector_d(q);
}

/* ========================================================================== */






/* -----------------  Metropolis::Component_gauss_scale  -------------------- */

Metropolis::Component_gauss_scale::Component_gauss_scale
(
    const char* name, 
    double prob
)
throw (jwscxx::base::Arg_error) : MV_gauss_scale(name, prob)
{
}

Metropolis::Component_gauss_scale* Metropolis::Component_gauss_scale::clone() 
    const
{
    return new Component_gauss_scale(*this);
}

uint32_t Metropolis::Component_gauss_scale::comp_i()
{
    return _comp_i;
}

uint32_t Metropolis::Component_gauss_scale::comp_n()
{
    return _comp_n;
}

void Metropolis::Component_gauss_scale::init
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;
    using jwscxx::base::Arg_error;

    MV_gauss_scale::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        try { _comp_i = cat->obj->select_random_component(); }
        catch (Arg_error e) { abort(); }

        _comp_n = cat->obj->Model::num_var_params();
        for (uint32_t i = 0; i < _comp_i; i++)
        {
            _comp_n += cat->obj->component(i)->num_var_params();
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::Component_gauss_scale::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    Vector_d* mean  = 0;
    Matrix_d* cov   = 0;
    Vector_d* sigma = 0;
    Vector_d* p     = 0;
    Vector_d* q     = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        Instance* ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Component* comp = ins->prop_obj->component(comp_i());
        uint32_t D = comp->num_var_params();
        create_zero_vector_d(&mean, D);

        if (ins->sample_cov_in)
        {
            copy_matrix_d(&cov, ins->sample_cov_in);
        }
        else
        {
            create_zero_matrix_d(&cov, D, D);
            uint32_t dd = 0;

            comp->get_var_param_gauss_proposal_sigma(&sigma);
            for (uint32_t d = 0; d < D; d++)
            {
                cov->elts[ dd ][ dd ] = sigma->elts[ d ]*sigma->elts[ d ]; 
                dd++;
            }
        }

        assert(sample_mv_gaussian_pdf_d(&p, mean, cov) == 0);
        create_zero_vector_d(&q, ins->obj->num_var_params());

        assert(copy_vector_section_into_vector_d(q, comp_n(), p, 0, D) == 0);
        try { ins->prop_obj->add_var_param_values_scale_component_deps(q); }
        catch (jwscxx::base::Arg_error e) { }

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
        assert(!std::isinf(la) && !std::isnan(la));
#endif

        log_alpha(la, i);
    }

    free_vector_d(mean);
    free_matrix_d(cov);
    free_vector_d(sigma);
    free_vector_d(p);
    free_vector_d(q);
}

/* ========================================================================== */







/* ----------------------  Metropolis::Cross  -------------------------- */

Metropolis::Cross::Cross(const char* name, double prob) 
    throw (jwscxx::base::Arg_error) : Component_gauss_shift(name, prob)
{
}

Metropolis::Cross* Metropolis::Cross::clone() const
{
    return new Cross(*this);
}

bool Metropolis::Cross::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Block_set::detachable_cross_blocks(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Cross::init
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;
    using jwscxx::base::Arg_error;

    Component_gauss_shift::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        const Object* obj = cat->obj;

        try { _comp_i = Block_set::detachable_cross_blocks(obj->components()); }
        catch (Arg_error e) { abort(); }

        Block_set* b = dynamic_cast<Block_set*>(obj->component(_comp_i));

        _comp_n = obj->Model::num_var_params();
        for (uint32_t i = 0; i < _comp_i; i++)
        {
            _comp_n += obj->component(i)->num_var_params();
        }
        _comp_n += b->num_var_param_cross_p();

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::Cross::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    Vector_d* mean  = 0;
    Matrix_d* cov   = 0;
    Vector_d* sigma = 0;
    Vector_d* p     = 0;
    Vector_d* q     = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        Instance* ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Block_set* b = dynamic_cast<Block_set*>(
                ins->prop_obj->component(comp_i()));
        uint32_t D = b->num_var_params();
        create_zero_vector_d(&mean, D);

        if (ins->sample_cov_in)
        {
            copy_matrix_d(&cov, ins->sample_cov_in);
        }
        else
        {
            create_zero_matrix_d(&cov, D, D);
            uint32_t dd = 0;

            b->get_var_param_gauss_proposal_sigma(&sigma);
            for (uint32_t d = 0; d < D; d++)
            {
                cov->elts[ dd ][ dd ] = sigma->elts[ d ]*sigma->elts[ d ]; 
                dd++;
            }
        }

        assert(sample_mv_gaussian_pdf_d(&p, mean, cov) == 0);
        create_zero_vector_d(&q, ins->obj->num_var_params());

        D = b->num_var_param_cross_p();
        uint32_t d = b->var_param_cross_i(0);

        assert(copy_vector_section_into_vector_d(q, comp_n(), p, d, D) == 0);
        try { ins->prop_obj->add_var_param_values_shift_component_deps(q); }
        catch (jwscxx::base::Arg_error e) { }

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
        assert(!std::isinf(la) && !std::isnan(la));
#endif

        log_alpha(la, i);
    }

    free_vector_d(mean);
    free_matrix_d(cov);
    free_vector_d(sigma);
    free_vector_d(p);
    free_vector_d(q);
}

/* ========================================================================== */







/* ---------------------  Metropolis::Recognize  ---------------------------- */

Metropolis::Recognize::Recognize(const char* name, double prob) 
    throw (jwscxx::base::Arg_error) : Diffusion_ins(name, prob)
{
}

Metropolis::Recognize::~Recognize()
{
}

Metropolis::Recognize* Metropolis::Recognize::clone() const
{
    return new Recognize(*this);
}

bool Metropolis::Recognize::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    return (components.size() > 0);
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::Recognize::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    Vector_d* q     = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        Instance* ins = params->instance(c, i);

        params->init_ins_proposal(c, i);

        ins->prop_view->sample_var_param_likelihood(&q);
        try { ins->prop_view->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }

        ins->prop_obj->sample_var_param_likelihood(&q);
        try { ins->prop_obj->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        //la += ins->prop_view->calc_var_param_log_likelihood();
        //la -= ins->view->calc_var_param_log_likelihood();
        //la += ins->prop_obj->calc_var_param_log_likelihood();
        //la -= ins->obj->calc_var_param_log_likelihood();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
        assert(!std::isinf(la) && !std::isnan(la));
#endif

        log_alpha(la, i);
    }

    free_vector_d(q);
}

/* ========================================================================== */







/* ------------------------  Metropolos::Dynamics_ins  ---------------------- */

Metropolis::Dynamics_ins::Dynamics_ins
(
    const char* name, 
    double      alpha, 
    uint32_t    kick, 
    uint32_t    leapfrog_i, 
    double      prob
) 
throw (jwscxx::base::Arg_error) 
    : Diffusion_ins(name, prob), Hybrid_dynamics(alpha, kick, leapfrog_i)
{
}

Metropolis::Dynamics_ins* Metropolis::Dynamics_ins::clone() const
{
    return new Dynamics_ins(*this);
}

bool Metropolis::Dynamics_ins::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    return (components.size() > 0);
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::Dynamics_ins::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    typedef void   (*f1)(Vector_d**, const void*);
    typedef Error* (*f2)(Vector_d**, void*);
    typedef Error* (*f3)(const Vector_d*, const Vector_d*, void*);

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    Vector_d* delta_t = 0;
    mgr->get_var_param_values_dynamics_delta_t_c(&delta_t, mgr);

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        params->init_ins_proposal(c, i);
    }

    // Not necessary because we always accept.
    //double lambda = (sample_uniform_pdf_d(0, 1) < 0.5) ? -1.0 : 1.0;
    //multiply_vector_by_scalar_d(&delta_t, delta_t, lambda);

    Error* err = stochastic_dynamics_2_d(leapfrog_i(), delta_t, 
            alpha(), kick(), mgr,
            (f1)(mgr->get_var_param_proposal_values_c),
            (f2)(mgr->get_var_param_proposal_values_grad_ins_potential_c), 
            (f3)(mgr->set_var_param_proposal_values_dynamics_c));

    if (err)
    {
        throw jwscxx::base::Exception(err->msg);
    }

    Vector_d* p = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        Instance* ins = params->instance(c, i);

        uint32_t M = ins->obj->num_components();
        if (M == 0)
        {
            reject(true, i);
            continue;
        }

        // Not necessary because the dynamics leaves the proposal updated.
        //params->update_ins_proposal_dependents(c, i);

        // We don't have any ideaswhat 'kinetic' is. It is better just to
        // assume the total energy has not changed and always accept.

        //double kinetic;
        //sample_standard_mv_gaussian_pdf_d(&p, M);
        //calc_vector_dot_prod_d(&kinetic, p, p);
        //kinetic *= 0.5;

        //double la = 0;
        //la -= ins->prop_ll_potential;
        //la += ins->ll_potential;
        //la -= ins->prop_ll_kinetic;
        //la += ins->ll_kinetic; 
        //log_alpha(la, c, i);


        // Always accept.
        log_alpha(0, i);
    }

    free_vector_d(delta_t);
    free_vector_d(p);
}

/* ========================================================================== */







/* ---------------------  Metropolis::Diffusion_cat  ------------------------ */

/**
 * @param  name  Name of the move.
 * @param  p     Probability of the move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 *
 * @note Use Metropolis::normalize_move_probs() to normalize all the move
 * probabilities so they sum to 1.
 */
Metropolis::Diffusion_cat::Diffusion_cat(const char* name, double p) 
    throw (jwscxx::base::Arg_error)
    : Diffusion(name, p)
{
}

Metropolis::Diffusion_cat::~Diffusion_cat()
{
}

Metropolis::Diffusion_cat* Metropolis::Diffusion_cat::clone() const
{
    return new Diffusion_cat(*this);
}

bool Metropolis::Diffusion_cat::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    return (components.size() > 0);
}

void Metropolis::Diffusion_cat::init(Metropolis::Thread_manager* mgr)
{
    Move::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        _log_alpha = 0;
        _reject = 0;

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::Diffusion_cat::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_i = params->num_instances();

    if (t > 0)
        return;

    Category* cat = params->category(c);
    params->init_cat_proposal(c);

    Vector_d* view_means  = 0;
    Vector_d* view_stdevs = 0;
    Vector_d* obj_means   = 0;
    Vector_d* obj_stdevs  = 0;

    cat->view->sample_var_param_prior(&view_means, &view_stdevs);
    cat->obj->sample_var_param_prior(&obj_means, &obj_stdevs);

    double cat_la = 0;

    for (uint32_t i = 0; i < N_i; i++)
    {
        Instance* ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        ins->prop_view->set_var_param_stats(view_means, view_stdevs);
        ins->prop_obj->set_var_param_stats(obj_means, obj_stdevs);

        params->update_ins_proposal_cat_dependents(c, i);

        double ins_la = 0;
        ins_la += ins->prop_view->calc_var_param_log_likelihood();
        ins_la -= ins->view->calc_var_param_log_likelihood();
        ins_la += ins->prop_obj->calc_var_param_log_likelihood();
        ins_la -= ins->obj->calc_var_param_log_likelihood();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
        assert(!std::isinf(ins_la) && !std::isnan(ins_la));
#endif

        cat_la += ins_la;
    }

    params->update_cat_proposal_dependents(c);

    cat_la += cat->prop_view->calc_var_param_prior_log_prob();
    cat_la -= cat->view->calc_var_param_prior_log_prob();
    cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
    cat_la -= cat->obj->calc_var_param_prior_log_prob();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

    log_alpha(cat_la);

    free_vector_d(view_means);
    free_vector_d(view_stdevs);
    free_vector_d(obj_means);
    free_vector_d(obj_stdevs);
}

void Metropolis::Diffusion_cat::accept_or_reject
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        uint32_t c = mgr->category();
        Parameters* params = mgr->params();

        double log_u = log(sample_uniform_pdf_d(JWSC_MIN_LOG_ARG, 1.0));
        bool accept = (log_u <= log_alpha() && !reject());
        //bool accept = !reject();

        if (accept)
        {
            params->accept_cat_proposal(c);
            uint32_t N_i = params->num_instances();
            for (uint32_t i = 0; i < N_i; i++)
            {
                params->accept_ins_proposal(c, i);
            }
        }

        for (uint32_t i = 0; i < params->num_instances(); i++)
        {
            Instance* ins = params->instance(c, i);
            ins->accepted = accept;
            ins->iteration++;
        }

        stop_timer(&(params->category(c)->timer));

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

double Metropolis::Diffusion_cat::log_alpha() const
{
    return _log_alpha;
}

void Metropolis::Diffusion_cat::log_alpha(double log_alpha)
{
    _log_alpha = log_alpha;
}

double Metropolis::Diffusion_cat::reject() const
{
    return _reject;
}

void Metropolis::Diffusion_cat::reject(bool reject)
{
    _reject = reject;
}

/* ========================================================================== */







/* ------------------------  Metropolis::View_cat  -------------------------- */

/**
 * @param  name  Name of the move.
 * @param  p     Probability of the move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 *
 * @note Use Metropolis::normalize_move_probs() to normalize all the move
 * probabilities so they sum to 1.
 */
Metropolis::View_cat::View_cat(const char* name, double p) 
    throw (jwscxx::base::Arg_error)
    : Diffusion_cat(name, p)
{
}

Metropolis::View_cat* Metropolis::View_cat::clone() const
{
    return new View_cat(*this);
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::View_cat::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_i = params->num_instances();

    if (t > 0)
        return;

    Category* cat = params->category(c);
    params->init_cat_proposal(c);

    Vector_d* means = 0;
    Vector_d* stdevs = 0;
    cat->view->sample_var_param_prior(&means, &stdevs);
    /*
    cat->view->get_var_param_stats(&means, &stdevs);

    uint32_t N_p = cat->view->num_var_params();
    for (uint32_t n_p = 0; n_p < N_p; n_p++)
    {
        double mx = means->elts[ n_p ];
        double sx = stdevs->elts[ n_p ];
        double me = 0.01;
        double se = 0.01;
        means->elts[ n_p ] = mx + sample_gaussian_pdf_d(0, me, -6*me, 6*me);
        stdevs->elts[ n_p ] = sx + sample_gaussian_pdf_d(0, se, -6*se, 6*se);
        if (stdevs->elts[ n_p ] < 0.01)
        {
            stdevs->elts[ n_p ] = 0.01;
        }
    }
    */

    double cat_la = 0;

    for (uint32_t i = 0; i < N_i; i++)
    {
        Instance* ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        ins->prop_view->set_var_param_stats(means, stdevs);

        params->update_ins_proposal_cat_dependents(c, i);

        double ins_la = 0;
        ins_la += ins->prop_view->calc_var_param_log_likelihood();
        ins_la -= ins->view->calc_var_param_log_likelihood();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
        assert(!std::isinf(ins_la) && !std::isnan(ins_la));
#endif

        cat_la += ins_la;
    }

    params->update_cat_proposal_dependents(c);

    cat_la += cat->prop_view->calc_var_param_prior_log_prob();
    cat_la -= cat->view->calc_var_param_prior_log_prob();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

    log_alpha(cat_la);

    free_vector_d(means);
    free_vector_d(stdevs);
}

/* ========================================================================== */







/* ----------------------  Metropolis::Object_cat  -------------------------- */

/**
 * @param  name  Name of the move.
 * @param  p     Probability of the move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 *
 * @note Use Metropolis::normalize_move_probs() to normalize all the move
 * probabilities so they sum to 1.
 */
Metropolis::Object_cat::Object_cat(const char* name, double p) 
    throw (jwscxx::base::Arg_error)
    : Diffusion_cat(name, p)
{
}

Metropolis::Object_cat* Metropolis::Object_cat::clone() const
{
    return new Object_cat(*this);
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::Object_cat::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_i = params->num_instances();

    if (t > 0)
        return;

    Category* cat = params->category(c);
    params->init_cat_proposal(c);

    Vector_d* means = 0;
    Vector_d* stdevs = 0;
    cat->obj->Model::sample_var_param_prior(&means, &stdevs);
    /*
    cat->obj->Model::get_var_param_stats(&means, &stdevs);

    uint32_t N_p = cat->obj->Model::num_var_params();
    uint32_t n_p = 0;
    double mx = means->elts[ n_p ];
    double sx = stdevs->elts[ n_p ];
    double me = 0.05;
    double se = 0.05;
    means->elts[ n_p ] = mx + sample_gaussian_pdf_d(0, me, -6*me, 6*me);
    stdevs->elts[ n_p ] = sx + sample_gaussian_pdf_d(0, se, -6*se, 6*se);
    if (stdevs->elts[ n_p ] < 0.01)
    {
        stdevs->elts[ n_p ] = 0.01;
    }

    for (n_p = 1; n_p < N_p; n_p++)
    {
        mx = means->elts[ n_p ];
        sx = stdevs->elts[ n_p ];
        me = 5.0;
        se = 5.0;
        means->elts[ n_p ] = mx + sample_gaussian_pdf_d(0, me, -6*me, 6*me);
        stdevs->elts[ n_p ] = sx + sample_gaussian_pdf_d(0, se, -6*se, 6*se);
        if (stdevs->elts[ n_p ] < 0.01)
        {
            stdevs->elts[ n_p ] = 0.01;
        }
    }
    */

    double cat_la = 0;

    for (uint32_t i = 0; i < N_i; i++)
    {
        Instance* ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        ins->prop_obj->Model::set_var_param_stats(means, stdevs);

        params->update_ins_proposal_cat_dependents(c, i);

        double ins_la = 0;
        ins_la += ins->prop_obj->calc_var_param_log_likelihood();
        ins_la -= ins->obj->calc_var_param_log_likelihood();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
        assert(!std::isinf(ins_la) && !std::isnan(ins_la));
#endif

        cat_la += ins_la;
    }

    params->update_cat_proposal_dependents(c);

    cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
    cat_la -= cat->obj->calc_var_param_prior_log_prob();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

    log_alpha(cat_la);

    free_vector_d(means);
    free_vector_d(stdevs);
}

/* ========================================================================== */







/* ---------------------  Metropolis::Component_cat  ------------------------ */

/**
 * @param  name  Name of the move.
 * @param  p     Probability of the move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 *
 * @note Use Metropolis::normalize_move_probs() to normalize all the move
 * probabilities so they sum to 1.
 */
Metropolis::Component_cat::Component_cat(const char* name, double p) 
    throw (jwscxx::base::Arg_error)
    : Diffusion_cat(name, p)
{
}

Metropolis::Component_cat* Metropolis::Component_cat::clone() const
{
    return new Component_cat(*this);
}

uint32_t Metropolis::Component_cat::comp_i()
{
    return _comp_i;
}

void Metropolis::Component_cat::init
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;
    using jwscxx::base::Arg_error;

    Diffusion_cat::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        try { _comp_i = cat->obj->select_random_component(); }
        catch (Arg_error e) { abort(); }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::Component_cat::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t t   = mgr->thread();
    uint32_t N_i = params->num_instances();

    if (t > 0)
        return;

    Category* cat = params->category(c);
    params->init_cat_proposal(c);

    Vector_d* means  = 0;
    Vector_d* stdevs = 0;
    Component* comp = cat->obj->component(comp_i());
    comp->sample_var_param_prior(&means, &stdevs);
    /*
    comp->get_var_param_stats(&means, &stdevs);

    uint32_t N_p = comp->num_var_params();
    for (uint32_t n_p = 0; n_p < N_p; n_p++)
    {
        bool   rp = comp->const_var_param(n_p)->has_ratio_prior();
        double mx = means->elts[ n_p ];
        double sx = stdevs->elts[ n_p ];
        double me = (rp) ? 0.05 : 5.0;
        double se = (rp) ? 0.05 : 5.0;
        means->elts[ n_p ] = mx + sample_gaussian_pdf_d(0, me, -6*me, 6*me);
        stdevs->elts[ n_p ] = sx + sample_gaussian_pdf_d(0, se, -6*se, 6*se);
        if (stdevs->elts[ n_p ] < 0.01)
        {
            stdevs->elts[ n_p ] = 0.01;
        }
    }
    */

    double cat_la = 0;

    for (uint32_t i = 0; i < N_i; i++)
    {
        Instance* ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        comp = ins->prop_obj->component(comp_i());
        comp->set_var_param_stats(means, stdevs);

        params->update_ins_proposal_cat_dependents(c, i);

        double ins_la = 0;
        ins_la += ins->prop_obj->calc_var_param_log_likelihood();
        ins_la -= ins->obj->calc_var_param_log_likelihood();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
        assert(!std::isinf(ins_la) && !std::isnan(ins_la));
#endif

        cat_la += ins_la;
    }

    params->update_cat_proposal_dependents(c);

    cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
    cat_la -= cat->obj->calc_var_param_prior_log_prob();

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
    assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

    log_alpha(cat_la);

    free_vector_d(means);
    free_vector_d(stdevs);
}

/* ========================================================================== */








/* ------------------------  Metropolos::Dynamics_cat  ---------------------- */

Metropolis::Dynamics_cat::Dynamics_cat
(
    const char* name, 
    double      alpha, 
    uint32_t    kick, 
    uint32_t    leapfrog_i, 
    double      prob
) 
throw (jwscxx::base::Arg_error) 
    : Diffusion_cat(name, prob),
      Hybrid_dynamics(alpha, kick, leapfrog_i),
      _dynamics(leapfrog_i, alpha, kick)
{
}

Metropolis::Dynamics_cat* Metropolis::Dynamics_cat::clone() const
{
    return new Dynamics_cat(*this);
}

bool Metropolis::Dynamics_cat::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    return (components.size() > 0);
}

/**
 * @param  mgr  Parameters for the move.
 */
void Metropolis::Dynamics_cat::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    typedef void   (*f1)(Vector_d**, const void*);
    typedef Error* (*f2)(Vector_d**, void*);
    typedef Error* (*f3)(const Vector_d*, const Vector_d*, void*);

    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t N_i = params->num_instances();

    mgr->synchronize();
    if (mgr->last_thread())
    {
        params->init_cat_proposal(c);
        for (uint32_t i = 0; i < N_i; i++)
        {
            params->init_ins_proposal(c, i);
        }
        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    mgr->desynchronize();

    _dynamics.run_thread(mgr);

    // We have no idea what the kinetic energy is for the current sample, so we
    // just accept it. In other words, we are assuming the total energy in the
    // state has remained the same.
    log_alpha(0);
}

/* ========================================================================== */







/* --------------------------  Metropolis::Jump  ---------------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Jump::Jump(const char* name, double p, double pair_p) 
    throw (jwscxx::base::Arg_error) : Move(name, p)
{
    _view_means  = 0;
    _view_stdevs = 0;
    _obj_means   = 0;
    _obj_stdevs  = 0;

    pair_prob(pair_p);
}

Metropolis::Jump::~Jump()
{
    using namespace jwsc;

    free_vector_d(_view_means);
    free_vector_d(_view_stdevs);
    free_vector_d(_obj_means);
    free_vector_d(_obj_stdevs);
}

/**
 * @param  mgr  Thread manager for the move.
 */
void Metropolis::Jump::run(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    init(mgr);
    run_thread(mgr);
    update_cat_log_alpha(mgr);
    accept_or_reject(mgr);
    assert_params(mgr);
}

double Metropolis::Jump::pair_prob() const 
{ 
    return _pair_prob; 
}

/**
 * @param  p  New probability of the move.
 *
 * @throw  jwscxx::base::Arg_error  Value must be in [0,1].
 */
void Metropolis::Jump::pair_prob(double p) throw (jwscxx::base::Arg_error)
{
    _pair_prob = p;
}

void Metropolis::Jump::init(Metropolis::Thread_manager* mgr)
{
    using namespace jwsc;

    Move::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();
        uint32_t N_i = params->num_instances();

        _ins_log_alpha.clear();
        _ins_log_alpha.insert(_ins_log_alpha.end(), N_i, 0);

        _cat_log_alpha = 0;
        _reject = false;
        _num_obj_components = params->instance(c, 0)->obj->num_components();

        Category* cat = params->category(c);
        cat->view->sample_var_param_prior(&_view_means, &_view_stdevs);
        cat->obj->sample_var_param_prior(&_obj_means, &_obj_stdevs);

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Jump::accept_or_reject(Metropolis::Thread_manager* mgr)
{
    using namespace jwsc;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        double move_1_log_prob = (prob() >= JWSC_MIN_LOG_ARG) ? log(prob())
            : log(JWSC_MIN_LOG_ARG);

        double move_2_log_prob = (pair_prob() >= JWSC_MIN_LOG_ARG) 
            ? log(pair_prob()) : log(JWSC_MIN_LOG_ARG);

        uint32_t N_i = params->num_instances();

        double log_alpha = 0;

        log_alpha += cat_log_alpha();
        log_alpha += sum_n_best_instance_log_alphas(mgr);

        log_alpha += move_2_log_prob;
        log_alpha -= move_1_log_prob;

        double log_u = log(sample_uniform_pdf_d(JWSC_MIN_LOG_ARG, 1.0));
        bool accept = (log_u <= log_alpha && !reject());
        //bool accept = !reject();

        if (accept)
        {
            params->accept_cat_proposal(c);
            for (uint32_t i = 0; i < N_i; i++)
            {
                params->accept_ins_proposal(c, i);
            }
        }

        for (uint32_t i = 0; i < N_i; i++)
        {
            Instance* ins = params->instance(c, i);
            ins->accepted = accept;
            ins->iteration++;
        }

        stop_timer(&(params->category(c)->timer));

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

double Metropolis::Jump::ins_log_alpha(size_t i) const
{
    try { return _ins_log_alpha.at(i); }
    catch (...) { abort(); }
}

void Metropolis::Jump::ins_log_alpha(double log_alpha, size_t i) 
{
    try { _ins_log_alpha.at(i) = log_alpha; }
    catch (...) { abort(); }
}

double Metropolis::Jump::cat_log_alpha() const
{
    return _cat_log_alpha;
}

void Metropolis::Jump::cat_log_alpha(double log_alpha)
{
    _cat_log_alpha = log_alpha;
}

double Metropolis::Jump::reject() const
{
    return _reject;
}

void Metropolis::Jump::reject(bool reject)
{
    _reject = reject;
}

uint32_t Metropolis::Jump::num_obj_components() const
{
    return _num_obj_components;
}

/** 
 * The thread manager is assumed to be synchronized.
 *
 * The value of @em n is the jump set size and is given by
 * Metropolis::_jump_set_size.
 */
double Metropolis::Jump::sum_n_best_instance_log_alphas
(
    Metropolis::Thread_manager* mgr
)
{
    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t N_i = params->num_instances();
    uint32_t N_j = mgr->sampler()->jump_set_size();

    std::vector<double> n_best;
    n_best.reserve(N_i);

    for (uint32_t i = 0; i < N_i; i++)
    {
        if (params->member(c, i) > 0)
        {
            n_best.push_back(ins_log_alpha(i));
        }
    }

    N_i = n_best.size();
    if (N_j == 0 || N_j > N_i)
    {
        N_j = N_i;
    }

    std::sort(n_best.begin(), n_best.end());

    double sum = 0;
    std::vector<double>::iterator it;
    for (it = n_best.begin()+(N_i - N_j); it != n_best.end(); it++)
    {
        sum += *it;
    }

    return sum;
}

const jwsc::Vector_d* Metropolis::Jump::view_means() const
{
    return _view_means;
}

const jwsc::Vector_d* Metropolis::Jump::view_stdevs() const
{
    return _view_stdevs;
}

const jwsc::Vector_d* Metropolis::Jump::obj_means() const
{
    return _obj_means;
}

const jwsc::Vector_d* Metropolis::Jump::obj_stdevs() const
{
    return _obj_stdevs;
}

/* ========================================================================== */







/* --------------------------  Metropolis::Birth  --------------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Birth::Birth(const char* name, double p, double pair_p) 
    throw (jwscxx::base::Arg_error) : Jump(name, p, pair_p)
{
    _means     = 0;
    _stdevs    = 0;
    _pt_means  = 0;
    _pt_stdevs = 0;
}

Metropolis::Birth::~Birth()
{
    using namespace jwsc;

    free_vector_d(_means);
    free_vector_d(_stdevs);
    free_vector_d(_pt_means);
    free_vector_d(_pt_stdevs);
}

void Metropolis::Birth::init
(
    Metropolis::Thread_manager* mgr
)
{
    Jump::init(mgr);
}

uint32_t Metropolis::Birth::attach_i()
{
    return _attach_i;
}

uint32_t Metropolis::Birth::attach_pt_i()
{
    return _attach_pt_i;
}

Component::Attach_type Metropolis::Birth::attach_pt_t()
{
    return _attach_pt_t;
}

const jwsc::Vector_d* Metropolis::Birth::means() const
{
    return _means;
}

const jwsc::Vector_d* Metropolis::Birth::stdevs() const
{
    return _stdevs;
}

const jwsc::Vector_d* Metropolis::Birth::pt_means() const
{
    return _pt_means;
}

const jwsc::Vector_d* Metropolis::Birth::pt_stdevs() const
{
    return _pt_stdevs;
}

void Metropolis::Birth::sample_var_param_stats
(
    Metropolis::Thread_manager* mgr,
    uint32_t                    i,
    const Abstract_block*       b1, 
    const Abstract_block*       b2
)
{
    if (mgr->thread() < i)
        return;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        b2->sample_var_param_prior(&_means, &_stdevs);

        uint32_t b1_j = attach_pt_i();
        uint32_t b2_j = (b1_j % 2) ? b1_j-1 : b1_j+1;

        if (attach_pt_t() == Component::VAR_ATTACH)
        {
            b1->sample_var_attach_pt_prior(&_pt_means, &_pt_stdevs, b1_j);
        }
        else
        {
            b2->sample_var_attach_pt_prior(&_pt_means, &_pt_stdevs, b2_j);
        }
        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    mgr->desynchronize();
}

/* ========================================================================== */







/* ---------------------  Metropolis::Block_birth  -------------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_birth::Block_birth(const char* name, double p, double pair_p) 
    throw (jwscxx::base::Arg_error) 
    : Birth(name, p, pair_p)
{
}

Metropolis::Block_birth* Metropolis::Block_birth::clone() const
{
    return new Block_birth(*this);
}

bool Metropolis::Block_birth::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    return (components.size() == 0);
}

void Metropolis::Block_birth::init
(
    Metropolis::Thread_manager* mgr
)
{
    Birth::init(mgr);

    mgr->synchronize();
    if (mgr->last_thread())
    {
        Block b(mgr->params()->block_ctx());
        b.sample_var_param_prior(&_means, &_stdevs);
        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    mgr->desynchronize();
}

void Metropolis::Block_birth::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    Vector_d* q = 0; 

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        View* prop_view = ins->prop_view;
        Object* prop_obj = ins->prop_obj;

        try { prop_view->set_var_param_stats(view_means(), view_stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        double view_q_lp = prop_view->calc_var_param_proposal_log_prob();
        prop_view->sample_var_param_proposal(&q);
        try { prop_view->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double view_prop_q_lp = prop_view->calc_var_param_proposal_log_prob();

        try { prop_obj->set_var_param_stats(obj_means(), obj_stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        double obj_q_lp = prop_obj->calc_var_param_proposal_log_prob();
        prop_obj->sample_var_param_proposal(&q);
        try { prop_obj->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double obj_prop_q_lp = prop_obj->calc_var_param_proposal_log_prob();

        Block* b = new Block(params->block_ctx());
        try { b->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        b->sample_var_param_likelihood(&q);
        try { b->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        prop_obj->attach(b, Component::CTR_ATTACH);

        double b_prop_q_lp = b->calc_var_param_log_likelihood();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_view->calc_var_param_log_likelihood();
        la -= ins->view->calc_var_param_log_likelihood();
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += view_q_lp;
        la -= view_prop_q_lp;
        la += obj_q_lp;
        la -= obj_prop_q_lp;
        la -= b_prop_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(q);
}

void Metropolis::Block_birth::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);
            const View* prop_view = cat->prop_view;
            const Object* prop_obj = cat->prop_obj;

            double view_q_lp = cat->view->calc_var_param_prior_log_prob();
            double view_prop_q_lp = prop_view->calc_var_param_prior_log_prob();

            double obj_q_lp = cat->obj->calc_var_param_prior_log_prob();
            double obj_prop_q_lp = prop_obj->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_view->calc_var_param_prior_log_prob();
            cat_la -= cat->view->calc_var_param_prior_log_prob();
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += view_q_lp;
            cat_la -= view_prop_q_lp;
            cat_la += obj_q_lp;
            cat_la -= obj_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/* ------------------  Metropolis::Block_to_block_birth  -------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_to_block_birth::Block_to_block_birth
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Birth(name, p, pair_p)
{
}

Metropolis::Block_to_block_birth* Metropolis::Block_to_block_birth::clone() 
    const
{
    return new Block_to_block_birth(*this);
}

bool Metropolis::Block_to_block_birth::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Block::attachable_block(components, false); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return (components.size() < MAX_NUM_COMPONENTS);
}

void Metropolis::Block_to_block_birth::init
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;
    using jwscxx::base::Arg_error;

    Birth::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _attach_i = Block::attachable_block(cat->obj->components(), 
                false); }
        catch (Arg_error e) { abort(); }

        const Block* b1 = dynamic_cast<Block*>(cat->obj->component(_attach_i));
        assert((b1->num_ctr_attach_pts() - b1->num_attachments()) > 0);

        try { _attach_pt_i = Block::attachable_block_attach_pt(b1, false); }
        catch (Arg_error e) { abort(); }

        _attach_pt_t = b1->select_random_attach_pt_type();

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Block_to_block_birth::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;
    typedef Abstract_block Ablock;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b1_i = attach_i();
    uint32_t b1_j = attach_pt_i();
    uint32_t b2_j = (b1_j % 2) ? b1_j-1 : b1_j+1;

    Component::Attach_type b1_t = attach_pt_t();
    Component::Attach_type b2_t = (b1_t == Component::CTR_ATTACH) 
        ? Component::VAR_ATTACH : Component::CTR_ATTACH;

    Vector_d* q  = 0; 
    Vector_d* uv = 0;
    Vector_d* p1 = 0;
    Vector_d* p2 = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Ablock* b1 = dynamic_cast<Ablock*>(prop_obj->component(b1_i));
        double b1_q_lp = b1->calc_var_param_proposal_log_likelihood();
        b1->posn_v(&p1);

        Block* b2 = new Block(params->block_ctx());
        prop_obj->attach(b1, b1_j, b1_t, b2, b2_j, b2_t);

        sample_var_param_stats(mgr, i, b1, b2);
        try { b2->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        b2->sample_var_param_likelihood(&q);
        try { b2->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }

        Ablock* bv = (b1_t == Component::VAR_ATTACH) ? b1 : b2;
        size_t bv_j = (b1_t == Component::VAR_ATTACH) ? b1_j : b2_j;

        try { bv->set_var_attach_pt_stats(pt_means(), pt_stdevs(), bv_j); }
        catch (jwscxx::base::Arg_error e) { abort(); }
        bv->sample_var_attach_pt_likelihood(&uv, bv_j);
        try { bv->set_var_attach_pt_values(uv, bv_j); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        b1->posn_v(&p2);
        subtract_vectors_d(&p1, p1, p2);
        prop_obj->add_posn_shift_deps_v(p1);

        double b1_prop_q_lp = b1->calc_var_param_proposal_log_likelihood();
        double b2_prop_q_lp = b2->calc_var_param_proposal_log_likelihood();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b1_q_lp;
        la -= b1_prop_q_lp;
        la -= b2_prop_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(p1);
    free_vector_d(p2);
    free_vector_d(uv);
    free_vector_d(q);
}

void Metropolis::Block_to_block_birth::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b1_i = attach_i();
            Block* b1 = dynamic_cast<Block*>(cat->obj->component(b1_i));
            double b1_q_lp = b1->calc_var_param_prior_log_prob();

            b1 = dynamic_cast<Block*>(cat->prop_obj->component(b1_i));
            double b1_prop_q_lp = b1->calc_var_param_prior_log_prob();

            uint32_t b2_i = cat->prop_obj->num_components()-1;
            Block* b2 = dynamic_cast<Block*>(cat->prop_obj->component(b2_i));
            double b2_prop_q_lp = b2->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b1_q_lp;
            cat_la -= b1_prop_q_lp;
            cat_la -= b2_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */








/* ---------------------  Metropolis::Pair_birth  --------------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Pair_birth::Pair_birth(const char* name, double p, double pair_p) 
    throw (jwscxx::base::Arg_error) 
    : Birth(name, p, pair_p)
{
}

Metropolis::Pair_birth* Metropolis::Pair_birth::clone() const
{
    return new Pair_birth(*this);
}

bool Metropolis::Pair_birth::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    return (components.size() == 0);
}

void Metropolis::Pair_birth::init(Metropolis::Thread_manager* mgr)
{
    Birth::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        _layout = Pair::select_random_layout();

        Pair b(_layout, mgr->params()->block_set_ctx());
        b.sample_var_param_prior(&_means, &_stdevs);

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Pair_birth::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    Vector_d* q = 0; 

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        View* prop_view = ins->prop_view;
        Object* prop_obj = ins->prop_obj;

        try { prop_view->set_var_param_stats(view_means(), view_stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        double view_q_lp = prop_view->calc_var_param_proposal_log_prob();
        prop_view->sample_var_param_proposal(&q);
        try { prop_view->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double view_prop_q_lp = prop_view->calc_var_param_proposal_log_prob();

        try { prop_obj->set_var_param_stats(obj_means(), obj_stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        double obj_q_lp = prop_obj->calc_var_param_proposal_log_prob();
        prop_obj->sample_var_param_proposal(&q);
        try { prop_obj->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double obj_prop_q_lp = prop_obj->calc_var_param_proposal_log_prob();

        Pair* b = new Pair(layout(), params->block_set_ctx());
        try { b->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        b->sample_var_param_likelihood(&q);
        try { b->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        prop_obj->attach(b, Component::CTR_ATTACH);

        double b_prop_q_lp = b->calc_var_param_log_likelihood();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_view->calc_var_param_log_likelihood();
        la -= ins->view->calc_var_param_log_likelihood();
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += view_q_lp;
        la -= view_prop_q_lp;
        la += obj_q_lp;
        la -= obj_prop_q_lp;
        la -= b_prop_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(q);
}

void Metropolis::Pair_birth::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);
            const View* prop_view = cat->prop_view;
            const Object* prop_obj = cat->prop_obj;

            double view_q_lp = cat->view->calc_var_param_prior_log_prob();
            double view_prop_q_lp = prop_view->calc_var_param_prior_log_prob();

            double obj_q_lp = cat->obj->calc_var_param_prior_log_prob();
            double obj_prop_q_lp = prop_obj->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_view->calc_var_param_prior_log_prob();
            cat_la -= cat->view->calc_var_param_prior_log_prob();
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += view_q_lp;
            cat_la -= view_prop_q_lp;
            cat_la += obj_q_lp;
            cat_la -= obj_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

Pair::Layout Metropolis::Pair_birth::layout()
{
    return _layout;
}

/* ========================================================================== */







/* ------------------  Metropolis::Pair_to_block_birth  --------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 * @param  height_only  Only allow height attachments to a Pair.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Pair_to_block_birth::Pair_to_block_birth
(
    const char* name, 
    double      p,
    double      pair_p,
    bool        height_only
) 
throw (jwscxx::base::Arg_error) 
    : Birth(name, p, pair_p)
{
    _height_only = height_only;
}

Metropolis::Pair_to_block_birth* Metropolis::Pair_to_block_birth::clone() const
{
    return new Pair_to_block_birth(*this);
}

bool Metropolis::Pair_to_block_birth::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Block::attachable_block(components, _height_only); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return (components.size() < MAX_NUM_COMPONENTS);
}

void Metropolis::Pair_to_block_birth::init(Metropolis::Thread_manager* mgr)
{
    Birth::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _attach_i = Block::attachable_block(cat->obj->components(),
                _height_only); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Block* b1 = dynamic_cast<Block*>(cat->obj->component(_attach_i));
        assert((b1->num_ctr_attach_pts() - b1->num_attachments()) > 0);

        try { _attach_pt_i = Block::attachable_block_attach_pt(b1,
                _height_only); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        _attach_pt_t = b1->select_random_attach_pt_type();
        _layout = Pair::select_random_layout();

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Pair_to_block_birth::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;
    typedef Abstract_block Ablock;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b1_i = attach_i();
    uint32_t b1_j = attach_pt_i();
    uint32_t b2_j = (b1_j % 2) ? b1_j-1 : b1_j+1;

    Component::Attach_type b1_t = attach_pt_t();
    Component::Attach_type b2_t = (b1_t == Component::CTR_ATTACH) 
        ? Component::VAR_ATTACH : Component::CTR_ATTACH;

    Pair::Layout b2_layout = layout();

    Vector_d* q  = 0; 
    Vector_d* uv = 0;
    Vector_d* p1 = 0;
    Vector_d* p2 = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Ablock* b1 = dynamic_cast<Ablock*>(prop_obj->component(b1_i));
        double b1_q_lp = b1->calc_var_param_proposal_log_likelihood();
        b1->posn_v(&p1);

        Pair* b2 = new Pair(b2_layout, params->block_set_ctx());
        prop_obj->attach(b1, b1_j, b1_t, b2, b2_j, b2_t);

        sample_var_param_stats(mgr, i, b1, b2);
        try { b2->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        b2->sample_var_param_likelihood(&q);
        try { b2->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }

        Ablock* bv = (b1_t == Component::VAR_ATTACH) ? b1 : b2;
        size_t bv_j = (b1_t == Component::VAR_ATTACH) ? b1_j : b2_j;

        try { bv->set_var_attach_pt_stats(pt_means(), pt_stdevs(), bv_j); }
        catch (jwscxx::base::Arg_error e) { abort(); }
        bv->sample_var_attach_pt_likelihood(&uv, bv_j);
        try { bv->set_var_attach_pt_values(uv, bv_j); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        b1->posn_v(&p2);
        subtract_vectors_d(&p1, p1, p2);
        prop_obj->add_posn_shift_deps_v(p1);

        double b1_prop_q_lp = b1->calc_var_param_proposal_log_likelihood();
        double b2_prop_q_lp = b2->calc_var_param_proposal_log_likelihood();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b1_q_lp;
        la -= b1_prop_q_lp;
        la -= b2_prop_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(p1);
    free_vector_d(p2);
    free_vector_d(uv);
    free_vector_d(q);
}

void Metropolis::Pair_to_block_birth::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b1_i = attach_i();
            Block* b1 = dynamic_cast<Block*>(cat->obj->component(b1_i));
            double b1_q_lp = b1->calc_var_param_prior_log_prob();

            b1 = dynamic_cast<Block*>(cat->prop_obj->component(b1_i));
            double b1_prop_q_lp = b1->calc_var_param_prior_log_prob();

            uint32_t b2_i = cat->prop_obj->num_components()-1;
            Pair* b2 = dynamic_cast<Pair*>(cat->prop_obj->component(b2_i));
            double b2_prop_q_lp = b2->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b1_q_lp;
            cat_la -= b1_prop_q_lp;
            cat_la -= b2_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

Pair::Layout Metropolis::Pair_to_block_birth::layout()
{
    return _layout;
}

/* ========================================================================== */







/* ------------------  Metropolis::Block_to_pair_birth  --------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 * @param  height_only  Only allow height attachments to a Pair.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_to_pair_birth::Block_to_pair_birth
(
    const char* name, 
    double      p,
    double      pair_p,
    bool        height_only
) 
throw (jwscxx::base::Arg_error) 
    : Birth(name, p, pair_p)
{
    _height_only = height_only;
}

Metropolis::Block_to_pair_birth* Metropolis::Block_to_pair_birth::clone() const
{
    return new Block_to_pair_birth(*this);
}

bool Metropolis::Block_to_pair_birth::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Pair::attachable_block_to_pair(components, _height_only); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return (components.size() < MAX_NUM_COMPONENTS);
}

void Metropolis::Block_to_pair_birth::init(Metropolis::Thread_manager* mgr)
{
    Birth::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _attach_i = Pair::attachable_block_to_pair(cat->obj->components(),
                _height_only); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Pair* b1 = dynamic_cast<Pair*>(cat->obj->component(_attach_i));
        assert((b1->num_ctr_attach_pts() - b1->num_attachments()) > 0);

        try { _attach_pt_i = Pair::attachable_block_to_pair_attach_pt(b1, 
                _height_only); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        _attach_pt_t = b1->select_random_attach_pt_type();

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Block_to_pair_birth::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;
    typedef Abstract_block Ablock;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b1_i = attach_i();
    uint32_t b1_j = attach_pt_i();
    uint32_t b2_j = (b1_j % 2) ? b1_j-1 : b1_j+1;

    Component::Attach_type b1_t = attach_pt_t();
    Component::Attach_type b2_t = (b1_t == Component::CTR_ATTACH) 
        ? Component::VAR_ATTACH : Component::CTR_ATTACH;

    Vector_d* q  = 0; 
    Vector_d* uv = 0;
    Vector_d* p1 = 0;
    Vector_d* p2 = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Ablock* b1 = dynamic_cast<Pair*>(prop_obj->component(b1_i));
        double b1_q_lp = b1->calc_var_param_proposal_log_likelihood();
        b1->posn_v(&p1);

        Block* b2 = new Block(params->block_ctx());
        prop_obj->attach(b1, b1_j, b1_t, b2, b2_j, b2_t);

        sample_var_param_stats(mgr, i, b1, b2);
        try { b2->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        b2->sample_var_param_likelihood(&q);
        try { b2->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }

        Ablock* bv = (b1_t == Component::VAR_ATTACH) ? b1 : b2;
        size_t bv_j = (b1_t == Component::VAR_ATTACH) ? b1_j : b2_j;

        try { bv->set_var_attach_pt_stats(pt_means(), pt_stdevs(), bv_j); }
        catch (jwscxx::base::Arg_error e) { abort(); }
        bv->sample_var_attach_pt_likelihood(&uv, bv_j);
        try { bv->set_var_attach_pt_values(uv, bv_j); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        b1->posn_v(&p2);
        subtract_vectors_d(&p1, p1, p2);
        prop_obj->add_posn_shift_deps_v(p1);

        double b1_prop_q_lp = b1->calc_var_param_proposal_log_likelihood();
        double b2_prop_q_lp = b2->calc_var_param_proposal_log_likelihood();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b1_q_lp;
        la -= b1_prop_q_lp;
        la -= b2_prop_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(p1);
    free_vector_d(p2);
    free_vector_d(uv);
    free_vector_d(q);
}

void Metropolis::Block_to_pair_birth::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b1_i = attach_i();
            Pair* b1 = dynamic_cast<Pair*>(cat->obj->component(b1_i));
            double b1_q_lp = b1->calc_var_param_prior_log_prob();

            b1 = dynamic_cast<Pair*>(cat->prop_obj->component(b1_i));
            double b1_prop_q_lp = b1->calc_var_param_prior_log_prob();

            uint32_t b2_i = cat->prop_obj->num_components()-1;
            Block* b2 = dynamic_cast<Block*>(cat->prop_obj->component(b2_i));
            double b2_prop_q_lp = b2->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b1_q_lp;
            cat_la -= b1_prop_q_lp;
            cat_la -= b2_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/* -------------------  Metropolis::Quartet_birth  -------------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Quartet_birth::Quartet_birth
(
    const char* name, 
    double      p, 
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Birth(name, p, pair_p)
{
}

Metropolis::Quartet_birth* Metropolis::Quartet_birth::clone() const
{
    return new Quartet_birth(*this);
}

bool Metropolis::Quartet_birth::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    return (components.size() == 0);
}

void Metropolis::Quartet_birth::init(Metropolis::Thread_manager* mgr)
{
    Birth::init(mgr);

    mgr->synchronize();
    if (mgr->last_thread())
    {
        Quartet b(mgr->params()->block_set_ctx());
        b.sample_var_param_prior(&_means, &_stdevs);
        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    mgr->desynchronize();
}

void Metropolis::Quartet_birth::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    Vector_d* q = 0; 

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        View* prop_view = ins->prop_view;
        Object* prop_obj = ins->prop_obj;

        try { prop_view->set_var_param_stats(view_means(), view_stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        double view_q_lp = prop_view->calc_var_param_proposal_log_prob();
        prop_view->sample_var_param_proposal(&q);
        try { prop_view->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double view_prop_q_lp = prop_view->calc_var_param_proposal_log_prob();

        try { prop_obj->set_var_param_stats(obj_means(), obj_stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        double obj_q_lp = prop_obj->calc_var_param_proposal_log_prob();
        prop_obj->sample_var_param_proposal(&q);
        try { prop_obj->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double obj_prop_q_lp = prop_obj->calc_var_param_proposal_log_prob();

        Quartet* b = new Quartet(params->block_set_ctx());
        try { b->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        b->sample_var_param_likelihood(&q);
        try { b->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        prop_obj->attach(b, Component::CTR_ATTACH);

        double b_prop_q_lp = b->calc_var_param_proposal_log_likelihood();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_view->calc_var_param_log_likelihood();
        la -= ins->view->calc_var_param_log_likelihood();
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += view_q_lp;
        la -= view_prop_q_lp;
        la += obj_q_lp;
        la -= obj_prop_q_lp;
        la -= b_prop_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(q);
}

void Metropolis::Quartet_birth::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);
            const View* prop_view = cat->prop_view;
            const Object* prop_obj = cat->prop_obj;

            double view_q_lp = cat->view->calc_var_param_prior_log_prob();
            double view_prop_q_lp = prop_view->calc_var_param_prior_log_prob();

            double obj_q_lp = cat->obj->calc_var_param_prior_log_prob();
            double obj_prop_q_lp = prop_obj->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_view->calc_var_param_prior_log_prob();
            cat_la -= cat->view->calc_var_param_prior_log_prob();
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += view_q_lp;
            cat_la -= view_prop_q_lp;
            cat_la += obj_q_lp;
            cat_la -= obj_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/* ----------------  Metropolis::Quartet_to_block_birth  -------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 * @param  height_only  Only allow height attachments to a Pair.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Quartet_to_block_birth::Quartet_to_block_birth
(
    const char* name, 
    double      p,
    double      pair_p,
    bool        height_only
) 
throw (jwscxx::base::Arg_error) 
    : Birth(name, p, pair_p)
{
    _height_only = height_only;
}

Metropolis::Quartet_to_block_birth* Metropolis::Quartet_to_block_birth::clone() 
    const
{
    return new Quartet_to_block_birth(*this);
}

bool Metropolis::Quartet_to_block_birth::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Block::attachable_block(components, _height_only); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return (components.size() < MAX_NUM_COMPONENTS);
}

void Metropolis::Quartet_to_block_birth::init(Metropolis::Thread_manager* mgr)
{
    Birth::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _attach_i = Block::attachable_block(cat->obj->components(),
                _height_only); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Block* b1 = dynamic_cast<Block*>(cat->obj->component(_attach_i));
        assert((b1->num_ctr_attach_pts() - b1->num_attachments()) > 0);

        try { _attach_pt_i = Block::attachable_block_attach_pt(b1,
                _height_only); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        _attach_pt_t = b1->select_random_attach_pt_type();

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Quartet_to_block_birth::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;
    typedef Abstract_block Ablock;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b1_i = attach_i();
    uint32_t b1_j = attach_pt_i();
    uint32_t b2_j = (b1_j % 2) ? b1_j-1 : b1_j+1;

    Component::Attach_type b1_t = attach_pt_t();
    Component::Attach_type b2_t = (b1_t == Component::CTR_ATTACH) 
        ? Component::VAR_ATTACH : Component::CTR_ATTACH;

    Vector_d* q  = 0; 
    Vector_d* uv = 0;
    Vector_d* p1 = 0;
    Vector_d* p2 = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Ablock* b1 = dynamic_cast<Ablock*>(prop_obj->component(b1_i));
        double b1_q_lp = b1->calc_var_param_proposal_log_likelihood();
        b1->posn_v(&p1);

        Quartet* b2 = new Quartet(params->block_set_ctx());
        prop_obj->attach(b1, b1_j, b1_t, b2, b2_j, b2_t);

        sample_var_param_stats(mgr, i, b1, b2);
        try { b2->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        b2->sample_var_param_likelihood(&q);
        try { b2->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }

        Ablock* bv = (b1_t == Component::VAR_ATTACH) ? b1 : b2;
        size_t bv_j = (b1_t == Component::VAR_ATTACH) ? b1_j : b2_j;

        try { bv->set_var_attach_pt_stats(pt_means(), pt_stdevs(), bv_j); }
        catch (jwscxx::base::Arg_error e) { abort(); }
        bv->sample_var_attach_pt_likelihood(&uv, bv_j);
        try { bv->set_var_attach_pt_values(uv, bv_j); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        b1->posn_v(&p2);
        subtract_vectors_d(&p1, p1, p2);
        prop_obj->add_posn_shift_deps_v(p1);

        double b1_prop_q_lp = b1->calc_var_param_proposal_log_likelihood();
        double b2_prop_q_lp = b2->calc_var_param_proposal_log_likelihood();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b1_q_lp;
        la -= b1_prop_q_lp;
        la -= b2_prop_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(p1);
    free_vector_d(p2);
    free_vector_d(uv);
    free_vector_d(q);
}

void Metropolis::Quartet_to_block_birth::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b1_i = attach_i();
            Block* b1 = dynamic_cast<Block*>(cat->obj->component(b1_i));
            double b1_q_lp = b1->calc_var_param_prior_log_prob();

            b1 = dynamic_cast<Block*>(cat->prop_obj->component(b1_i));
            double b1_prop_q_lp = b1->calc_var_param_prior_log_prob();

            uint32_t b2_i = cat->prop_obj->num_components()-1;
            Quartet* b2 =dynamic_cast<Quartet*>(cat->prop_obj->component(b2_i));
            double b2_prop_q_lp = b2->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b1_q_lp;
            cat_la -= b1_prop_q_lp;
            cat_la -= b2_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/* ----------------  Metropolis::Block_to_quartet_birth  -------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 * @param  height_only  Only allow height attachments to a Pair.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_to_quartet_birth::Block_to_quartet_birth
(
    const char* name, 
    double      p,
    double      pair_p,
    bool        height_only
) 
throw (jwscxx::base::Arg_error) 
    : Birth(name, p, pair_p)
{
    _height_only = height_only;
}

Metropolis::Block_to_quartet_birth* Metropolis::Block_to_quartet_birth::clone() 
    const
{
    return new Block_to_quartet_birth(*this);
}

bool Metropolis::Block_to_quartet_birth::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Quartet::attachable_block_to_quartet(components, _height_only); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return (components.size() < MAX_NUM_COMPONENTS);
}

void Metropolis::Block_to_quartet_birth::init(Metropolis::Thread_manager* mgr)
{
    Birth::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _attach_i = Quartet::attachable_block_to_quartet(
                cat->obj->components(), _height_only); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Quartet* b1 = dynamic_cast<Quartet*>(
                cat->obj->component(_attach_i));
        assert((b1->num_ctr_attach_pts() - b1->num_attachments()) > 0);

        try { _attach_pt_i = Quartet::attachable_block_to_quartet_attach_pt(b1, 
                _height_only); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        _attach_pt_t = b1->select_random_attach_pt_type();

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Block_to_quartet_birth::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;
    typedef Abstract_block Ablock;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b1_i = attach_i();
    uint32_t b1_j = attach_pt_i();
    uint32_t b2_j = (b1_j % 2) ? b1_j-1 : b1_j+1;

    Component::Attach_type b1_t = attach_pt_t();
    Component::Attach_type b2_t = (b1_t == Component::CTR_ATTACH) 
        ? Component::VAR_ATTACH : Component::CTR_ATTACH;

    Vector_d* q  = 0; 
    Vector_d* uv = 0;
    Vector_d* p1 = 0;
    Vector_d* p2 = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Ablock* b1 = dynamic_cast<Quartet*>(prop_obj->component(b1_i));
        double b1_q_lp = b1->calc_var_param_proposal_log_likelihood();
        b1->posn_v(&p1);

        Block* b2 = new Block(params->block_ctx());
        prop_obj->attach(b1, b1_j, b1_t, b2, b2_j, b2_t);

        sample_var_param_stats(mgr, i, b1, b2);
        try { b2->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        b2->sample_var_param_likelihood(&q);
        try { b2->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }

        Ablock* bv = (b1_t == Component::VAR_ATTACH) ? b1 : b2;
        size_t bv_j = (b1_t == Component::VAR_ATTACH) ? b1_j : b2_j;

        try { bv->set_var_attach_pt_stats(pt_means(), pt_stdevs(), bv_j); }
        catch (jwscxx::base::Arg_error e) { abort(); }
        bv->sample_var_attach_pt_likelihood(&uv, bv_j);
        try { bv->set_var_attach_pt_values(uv, bv_j); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        b1->posn_v(&p2);
        subtract_vectors_d(&p1, p1, p2);
        prop_obj->add_posn_shift_deps_v(p1);

        double b1_prop_q_lp = b1->calc_var_param_proposal_log_likelihood();
        double b2_prop_q_lp = b2->calc_var_param_proposal_log_likelihood();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b1_q_lp;
        la -= b1_prop_q_lp;
        la -= b2_prop_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(p1);
    free_vector_d(p2);
    free_vector_d(uv);
    free_vector_d(q);
}

void Metropolis::Block_to_quartet_birth::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b1_i = attach_i();
            Quartet* b1 = dynamic_cast<Quartet*>(cat->obj->component(b1_i));
            double b1_q_lp = b1->calc_var_param_prior_log_prob();

            b1 = dynamic_cast<Quartet*>(cat->prop_obj->component(b1_i));
            double b1_prop_q_lp = b1->calc_var_param_prior_log_prob();

            uint32_t b2_i = cat->prop_obj->num_components()-1;
            Block* b2 = dynamic_cast<Block*>(cat->prop_obj->component(b2_i));
            double b2_prop_q_lp = b2->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b1_q_lp;
            cat_la -= b1_prop_q_lp;
            cat_la -= b2_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/* --------------------  Metropolis::Cross_birth  --------------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Cross_birth::Cross_birth
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Birth(name, p, pair_p)
{
}

Metropolis::Cross_birth* Metropolis::Cross_birth::clone() const
{
    return new Cross_birth(*this);
}

bool Metropolis::Cross_birth::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Block_set::attachable_cross_blocks(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Cross_birth::sample_var_param_stats
(
    Metropolis::Thread_manager* mgr,
    uint32_t                    i,
    const Block_set*            b
)
{
    using namespace jwsc;

    if (mgr->thread() < i)
        return;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        Vector_d* b_m = 0;
        Vector_d* b_s = 0;

        b->get_var_param_stats(&b_m, &b_s);
        b->sample_var_param_prior(&_means, &_stdevs);
        assert(b_m->num_elts == _means->num_elts);

        uint32_t N = b->num_var_param_cross_p();
        uint32_t d = b->var_param_cross_i(0);

        for (uint32_t n = 0; n < N; n++)
        {
            b_m->elts[ d ] = _means->elts[ d ];
            b_s->elts[ d ] = _stdevs->elts[ d ];
            d++;
        }

        assert(d < b_m->num_elts && d < _means->num_elts);
        free_vector_d(b_m);
        free_vector_d(b_s);

        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    mgr->desynchronize();
}

void Metropolis::Cross_birth::init(Metropolis::Thread_manager* mgr)
{
    Birth::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _attach_i = Block_set::attachable_cross_blocks(
                cat->obj->components()); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        Block_set* b = dynamic_cast<Block_set*>(cat->obj->component(_attach_i));
        const Model::Parameter::Context* ctx = b->cross_p(0)->context();
        double m, s;
        ctx->sample_prior(&m, &s);
        _one_cross = (ctx->likelihood().sample(m, s, &(ctx->range())) <= 1.5);

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Cross_birth::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b_i = attach_i();
    Vector_d* q  = 0; 

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Block_set* b = dynamic_cast<Block_set*>(prop_obj->component(b_i));
        b->add_cross_blocks((one_cross()) ? 1 : 2);

        sample_var_param_stats(mgr, i, b);
        try { b->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        b->sample_var_param_cross_proposal(&q, one_cross());
        try { b->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }

        double b_q_lp = b->calc_var_param_cross_proposal_log_prob(q);

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la -= b_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(q);
}

void Metropolis::Cross_birth::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b_i = attach_i();
            Block_set* b = dynamic_cast<Block_set*>(cat->obj->component(b_i));
            double b_q_lp = b->calc_var_param_prior_log_prob();

            b = dynamic_cast<Block_set*>(cat->prop_obj->component(b_i));
            double b_prop_q_lp = b->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b_q_lp;
            cat_la -= b_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

bool Metropolis::Cross_birth::one_cross()
{
    return _one_cross;
}

/* ========================================================================== */







/* --------------------------  Metropolis::Death  --------------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Death::Death(const char* name, double p, double pair_p) 
    throw (jwscxx::base::Arg_error) : Jump(name, p, pair_p)
{
}

Metropolis::Death::~Death()
{
}

void Metropolis::Death::init(Metropolis::Thread_manager* mgr)
{
    Jump::init(mgr);
}

uint32_t Metropolis::Death::detach_i()
{
    return _detach_i;
}

uint32_t Metropolis::Death::detach_pt_i()
{
    return _detach_pt_i;
}

/** 
 * The thread manager is assumed to be synchronized.
 *
 * The value of @em n is the jump set size and is given by
 * Metropolis::_jump_set_size.
 */
double Metropolis::Death::sum_n_best_instance_log_alphas
(
    Metropolis::Thread_manager* mgr
)
{
    return Jump::sum_n_best_instance_log_alphas(mgr);
    /*
    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t N_i = params->num_instances();
    uint32_t N_j = mgr->sampler()->jump_set_size();

    std::vector<double> n_best;
    n_best.reserve(N_i);

    for (uint32_t i = 0; i < N_i; i++)
    {
        if (params->member(c, i) > 0)
        {
            n_best.push_back(ins_log_alpha(i));
        }
    }

    N_i = n_best.size();
    if (N_j == 0 || N_j > N_i)
    {
        N_j = N_i;
    }

    std::sort(n_best.begin(), n_best.end());

    double sum = 0;
    std::vector<double>::iterator it = n_best.begin();
    for (uint32_t j = 0; j < N_j; j++)
    {
        sum += *it;
        it++;
        assert(it != n_best.end());
    }

    return sum;
    */
}

/* ========================================================================== */







/* ---------------------  Metropolis::Block_death  -------------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_death::Block_death(const char* name, double p, double pair_p) 
    throw (jwscxx::base::Arg_error) 
    : Death(name, p, pair_p)
{
}

Metropolis::Block_death* Metropolis::Block_death::clone() const
{
    return new Block_death(*this);
}

bool Metropolis::Block_death::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    return (components.size() == 1 && dynamic_cast<Block*>(components[0]));
}

void Metropolis::Block_death::init(Metropolis::Thread_manager* mgr)
{
    using jwscxx::base::Arg_error;

    Death::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _detach_i = cat->obj->select_random_block(); } catch (jwscxx::base::Arg_error e) { abort(); }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Block_death::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    size_t b_i = detach_i();

    Vector_d* q = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        View* prop_view = ins->prop_view;
        Object* prop_obj = ins->prop_obj;

        Block* b = dynamic_cast<Block*>(ins->prop_obj->component(b_i));
        double b_q_lp = b->calc_var_param_proposal_log_likelihood();

        delete prop_obj->detach(b);

        try { prop_view->set_var_param_stats(view_means(), view_stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        double view_q_lp = prop_view->calc_var_param_proposal_log_prob();
        prop_view->sample_var_param_proposal(&q);
        try { prop_view->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double view_prop_q_lp = prop_view->calc_var_param_proposal_log_prob();

        try { prop_obj->set_var_param_stats(obj_means(), obj_stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        double obj_q_lp = prop_obj->calc_var_param_proposal_log_prob();
        prop_obj->sample_var_param_proposal(&q);
        try { prop_obj->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double obj_prop_q_lp = prop_obj->calc_var_param_proposal_log_prob();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_view->calc_var_param_log_likelihood();
        la -= ins->view->calc_var_param_log_likelihood();
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += view_q_lp;
        la -= view_prop_q_lp;
        la += obj_q_lp;
        la -= obj_prop_q_lp;
        la += b_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(q);
}

void Metropolis::Block_death::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);
            const View* prop_view = cat->prop_view;
            const Object* prop_obj = cat->prop_obj;

            double view_q_lp = cat->view->calc_var_param_prior_log_prob();
            double view_prop_q_lp = prop_view->calc_var_param_prior_log_prob();

            double obj_q_lp = cat->obj->calc_var_param_prior_log_prob();
            double obj_prop_q_lp = prop_obj->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_view->calc_var_param_prior_log_prob();
            cat_la -= cat->view->calc_var_param_prior_log_prob();
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += view_q_lp;
            cat_la -= view_prop_q_lp;
            cat_la += obj_q_lp;
            cat_la -= obj_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/* ------------------  Metropolis::Block_from_block_death  ------------------ */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_from_block_death::Block_from_block_death
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Death(name, p, pair_p)
{
}

Metropolis::Block_from_block_death* Metropolis::Block_from_block_death::clone() 
    const
{
    return new Block_from_block_death(*this);
}

bool Metropolis::Block_from_block_death::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Block::detachable_block(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Block_from_block_death::init(Metropolis::Thread_manager* mgr)
{
    using jwscxx::base::Arg_error;

    Death::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _detach_i = Block::detachable_block(cat->obj->components()); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Block* b = dynamic_cast<Block*>(cat->obj->component(_detach_i));

        try { _detach_pt_i = b->select_random_used_attach_pt(); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Block_from_block_death::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b2_i = detach_i();
    uint32_t b2_j = detach_pt_i();

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Block* b2 = dynamic_cast<Block*>(prop_obj->component(b2_i));
        Block* b1 = dynamic_cast<Block*>(b2->attachment(b2_j));
        uint32_t b1_j = b1->attachment(b2);
        double b1_q_lp = b1->calc_var_param_proposal_log_likelihood();
        double b2_q_lp = b2->calc_var_param_proposal_log_likelihood();
        if (b1->conditions(b2))
        {
            b1->scale_size_stats(b2, b2_j);
        }

        delete prop_obj->detach(b1, b1_j, b2, b2_j);

        double b1_prop_q_lp = b1->calc_var_param_proposal_log_likelihood();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b1_q_lp;
        la += b2_q_lp;
        la -= b1_prop_q_lp;

        ins_log_alpha(la, i);
    }
}

void Metropolis::Block_from_block_death::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b2_i = detach_i();
            uint32_t b2_j = detach_pt_i();
            Block* b2 = dynamic_cast<Block*>(cat->obj->component(b2_i));
            double b2_q_lp = b2->calc_var_param_prior_log_prob();

            Block* b1 = dynamic_cast<Block*>(b2->attachment(b2_j));
            uint32_t b1_i = cat->obj->component(b1);
            double b1_q_lp = b1->calc_var_param_prior_log_prob();

            b1_i = (b1_i > b2_i) ? b1_i-1 : b1_i;
            b1 = dynamic_cast<Block*>(cat->prop_obj->component(b1_i));
            double b1_prop_q_lp = b1->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b1_q_lp;
            cat_la += b2_q_lp;
            cat_la -= b1_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */








/* ---------------------  Metropolis::Pair_death  --------------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Pair_death::Pair_death(const char* name, double p, double pair_p) 
    throw (jwscxx::base::Arg_error) 
    : Death(name, p, pair_p)
{
}

Metropolis::Pair_death* Metropolis::Pair_death::clone() const
{
    return new Pair_death(*this);
}

bool Metropolis::Pair_death::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    return (components.size() == 1 && dynamic_cast<Pair*>(components[0]));
}

void Metropolis::Pair_death::init(Metropolis::Thread_manager* mgr)
{
    using jwscxx::base::Arg_error;

    Death::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _detach_i = cat->obj->select_random_pair(); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        _layout = Pair::select_random_layout();

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Pair_death::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    size_t b_i = detach_i();

    Vector_d* q = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        View* prop_view = ins->prop_view;
        Object* prop_obj = ins->prop_obj;

        Pair* b = dynamic_cast<Pair*>(ins->prop_obj->component(b_i));
        double b_q_lp = b->calc_var_param_proposal_log_likelihood();

        delete prop_obj->detach(b);

        try { prop_view->set_var_param_stats(view_means(), view_stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        double view_q_lp = prop_view->calc_var_param_proposal_log_prob();
        prop_view->sample_var_param_proposal(&q);
        try { prop_view->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double view_prop_q_lp = prop_view->calc_var_param_proposal_log_prob();

        try { prop_obj->set_var_param_stats(obj_means(), obj_stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        double obj_q_lp = prop_obj->calc_var_param_proposal_log_prob();
        prop_obj->sample_var_param_proposal(&q);
        try { prop_obj->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double obj_prop_q_lp = prop_obj->calc_var_param_proposal_log_prob();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_view->calc_var_param_log_likelihood();
        la -= ins->view->calc_var_param_log_likelihood();
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += view_q_lp;
        la -= view_prop_q_lp;
        la += obj_q_lp;
        la -= obj_prop_q_lp;
        la += b_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(q);
}

void Metropolis::Pair_death::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);
            const View* prop_view = cat->prop_view;
            const Object* prop_obj = cat->prop_obj;

            double view_q_lp = cat->view->calc_var_param_prior_log_prob();
            double view_prop_q_lp = prop_view->calc_var_param_prior_log_prob();

            double obj_q_lp = cat->obj->calc_var_param_prior_log_prob();
            double obj_prop_q_lp = prop_obj->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_view->calc_var_param_prior_log_prob();
            cat_la -= cat->view->calc_var_param_prior_log_prob();
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += view_q_lp;
            cat_la -= view_prop_q_lp;
            cat_la += obj_q_lp;
            cat_la -= obj_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

Pair::Layout Metropolis::Pair_death::layout()
{
    return _layout;
}

/* ========================================================================== */







/* ------------------  Metropolis::Pair_from_block_death  ------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Pair_from_block_death::Pair_from_block_death
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Death(name, p, pair_p)
{
}

Metropolis::Pair_from_block_death* Metropolis::Pair_from_block_death::clone() 
    const
{
    return new Pair_from_block_death(*this);
}

bool Metropolis::Pair_from_block_death::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Pair::detachable_pair_from_block(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Pair_from_block_death::init(Metropolis::Thread_manager* mgr)
{
    using jwscxx::base::Arg_error;

    Death::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _detach_i = Pair::detachable_pair_from_block(
                cat->obj->components()); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Pair* b = dynamic_cast<Pair*>(cat->obj->component(_detach_i));

        try { _detach_pt_i = b->select_random_used_attach_pt(); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        _layout = Pair::select_random_layout();

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Pair_from_block_death::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b2_i = detach_i();
    uint32_t b2_j = detach_pt_i();

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Pair* b2 = dynamic_cast<Pair*>(prop_obj->component(b2_i));
        Block* b1 = dynamic_cast<Block*>(b2->attachment(b2_j));
        uint32_t b1_j = b1->attachment(b2);
        double b1_q_lp = b1->calc_var_param_proposal_log_likelihood();
        double b2_q_lp = b2->calc_var_param_proposal_log_likelihood();
        if (b1->conditions(b2))
        {
            b1->scale_size_stats(b2, b2_j);
        }

        delete prop_obj->detach(b1, b1_j, b2, b2_j);

        double b1_prop_q_lp = b1->calc_var_param_proposal_log_likelihood();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b1_q_lp;
        la += b2_q_lp;
        la -= b1_prop_q_lp;

        ins_log_alpha(la, i);
    }
}

void Metropolis::Pair_from_block_death::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b2_i = detach_i();
            uint32_t b2_j = detach_pt_i();
            Pair* b2 = dynamic_cast<Pair*>(cat->obj->component(b2_i));
            double b2_q_lp = b2->calc_var_param_prior_log_prob();

            Block* b1 = dynamic_cast<Block*>(b2->attachment(b2_j));
            uint32_t b1_i = cat->obj->component(b1);
            double b1_q_lp = b1->calc_var_param_prior_log_prob();

            b1_i = (b1_i > b2_i) ? b1_i-1 : b1_i;
            b1 = dynamic_cast<Block*>(cat->prop_obj->component(b1_i));
            double b1_prop_q_lp = b1->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b1_q_lp;
            cat_la += b2_q_lp;
            cat_la -= b1_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

Pair::Layout Metropolis::Pair_from_block_death::layout()
{
    return _layout;
}

/* ========================================================================== */







/* ------------------  Metropolis::Block_from_pair_death  ------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_from_pair_death::Block_from_pair_death
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Death(name, p, pair_p)
{
}

Metropolis::Block_from_pair_death* Metropolis::Block_from_pair_death::clone() 
    const
{
    return new Block_from_pair_death(*this);
}

bool Metropolis::Block_from_pair_death::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Pair::detachable_block_from_pair(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Block_from_pair_death::init(Metropolis::Thread_manager* mgr)
{
    using jwscxx::base::Arg_error;

    Death::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _detach_i = Pair::detachable_block_from_pair(
                cat->obj->components()); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Block* b = dynamic_cast<Block*>(cat->obj->component(_detach_i));

        try { _detach_pt_i = b->select_random_used_attach_pt(); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Block_from_pair_death::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b2_i = detach_i();
    uint32_t b2_j = detach_pt_i();

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Block* b2 = dynamic_cast<Block*>(prop_obj->component(b2_i));
        Pair* b1 = dynamic_cast<Pair*>(b2->attachment(b2_j));
        uint32_t b1_j = b1->attachment(b2);
        double b1_q_lp = b1->calc_var_param_proposal_log_likelihood();
        double b2_q_lp = b2->calc_var_param_proposal_log_likelihood();
        if (b1->conditions(b2))
        {
            b1->scale_size_stats(b2, b2_j);
        }

        delete prop_obj->detach(b1, b1_j, b2, b2_j);

        double b1_prop_q_lp = b1->calc_var_param_proposal_log_likelihood();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b1_q_lp;
        la += b2_q_lp;
        la -= b1_prop_q_lp;

        ins_log_alpha(la, i);
    }
}

void Metropolis::Block_from_pair_death::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b2_i = detach_i();
            uint32_t b2_j = detach_pt_i();
            Block* b2 = dynamic_cast<Block*>(cat->obj->component(b2_i));
            double b2_q_lp = b2->calc_var_param_prior_log_prob();

            Pair* b1 = dynamic_cast<Pair*>(b2->attachment(b2_j));
            uint32_t b1_i = cat->obj->component(b1);
            double b1_q_lp = b1->calc_var_param_prior_log_prob();

            b1_i = (b1_i > b2_i) ? b1_i-1 : b1_i;
            b1 = dynamic_cast<Pair*>(cat->prop_obj->component(b1_i));
            double b1_prop_q_lp = b1->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b1_q_lp;
            cat_la += b2_q_lp;
            cat_la -= b1_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/* -------------------  Metropolis::Quartet_death  -------------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Quartet_death::Quartet_death
(
    const char* name, 
    double      p, 
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Death(name, p, pair_p)
{
}

Metropolis::Quartet_death* Metropolis::Quartet_death::clone() const
{
    return new Quartet_death(*this);
}

bool Metropolis::Quartet_death::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    return (components.size() == 1 && dynamic_cast<Quartet*>(components[0]));
}

void Metropolis::Quartet_death::init(Metropolis::Thread_manager* mgr)
{
    using jwscxx::base::Arg_error;

    Death::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _detach_i = cat->obj->select_random_quartet(); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Quartet_death::run_thread(Metropolis::Thread_manager* mgr)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    size_t b_i = detach_i();

    Vector_d* q = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        View* prop_view = ins->prop_view;
        Object* prop_obj = ins->prop_obj;

        Quartet* b = dynamic_cast<Quartet*>(ins->prop_obj->component(b_i));
        double b_q_lp = b->calc_var_param_proposal_log_likelihood();

        delete prop_obj->detach(b);

        try { prop_view->set_var_param_stats(view_means(), view_stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        double view_q_lp = prop_view->calc_var_param_proposal_log_prob();
        prop_view->sample_var_param_proposal(&q);
        try { prop_view->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double view_prop_q_lp = prop_view->calc_var_param_proposal_log_prob();

        try { prop_obj->set_var_param_stats(obj_means(), obj_stdevs()); }
        catch (jwscxx::base::Arg_error e) { }
        double obj_q_lp = prop_obj->calc_var_param_proposal_log_prob();
        prop_obj->sample_var_param_proposal(&q);
        try { prop_obj->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double obj_prop_q_lp = prop_obj->calc_var_param_proposal_log_prob();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_view->calc_var_param_log_likelihood();
        la -= ins->view->calc_var_param_log_likelihood();
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += view_q_lp;
        la -= view_prop_q_lp;
        la += obj_q_lp;
        la -= obj_prop_q_lp;
        la += b_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(q);
}

void Metropolis::Quartet_death::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);
            const View* prop_view = cat->prop_view;
            const Object* prop_obj = cat->prop_obj;

            double view_q_lp = cat->view->calc_var_param_prior_log_prob();
            double view_prop_q_lp = prop_view->calc_var_param_prior_log_prob();

            double obj_q_lp = cat->obj->calc_var_param_prior_log_prob();
            double obj_prop_q_lp = prop_obj->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_view->calc_var_param_prior_log_prob();
            cat_la -= cat->view->calc_var_param_prior_log_prob();
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += view_q_lp;
            cat_la -= view_prop_q_lp;
            cat_la += obj_q_lp;
            cat_la -= obj_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/*---------------  Metropolis::Quartet_from_block_death  -------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Quartet_from_block_death::Quartet_from_block_death
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Death(name, p, pair_p)
{
}

Metropolis::Quartet_from_block_death* 
Metropolis::Quartet_from_block_death::clone() const
{
    return new Quartet_from_block_death(*this);
}

bool Metropolis::Quartet_from_block_death::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Quartet::detachable_quartet_from_block(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Quartet_from_block_death::init(Metropolis::Thread_manager* mgr)
{
    using jwscxx::base::Arg_error;

    Death::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _detach_i = Quartet::detachable_quartet_from_block(
                cat->obj->components()); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Quartet* b = dynamic_cast<Quartet*>(
                cat->obj->component(_detach_i));

        try { _detach_pt_i = b->select_random_used_attach_pt(); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Quartet_from_block_death::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b2_i = detach_i();
    uint32_t b2_j = detach_pt_i();

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Quartet* b2 = dynamic_cast<Quartet*>(prop_obj->component(b2_i));
        Block* b1 = dynamic_cast<Block*>(b2->attachment(b2_j));
        uint32_t b1_j = b1->attachment(b2);
        double b1_q_lp = b1->calc_var_param_proposal_log_likelihood();
        double b2_q_lp = b2->calc_var_param_proposal_log_likelihood();
        if (b1->conditions(b2))
        {
            b1->scale_size_stats(b2, b2_j);
        }

        delete prop_obj->detach(b1, b1_j, b2, b2_j);

        double b1_prop_q_lp = b1->calc_var_param_proposal_log_likelihood();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b1_q_lp;
        la += b2_q_lp;
        la -= b1_prop_q_lp;

        ins_log_alpha(la, i);
    }
}

void Metropolis::Quartet_from_block_death::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b2_i = detach_i();
            uint32_t b2_j = detach_pt_i();
            Quartet* b2 = dynamic_cast<Quartet*>(cat->obj->component(b2_i));
            double b2_q_lp = b2->calc_var_param_prior_log_prob();

            Block* b1 = dynamic_cast<Block*>(b2->attachment(b2_j));
            uint32_t b1_i = cat->obj->component(b1);
            double b1_q_lp = b1->calc_var_param_prior_log_prob();

            b1_i = (b1_i > b2_i) ? b1_i-1 : b1_i;
            b1 = dynamic_cast<Block*>(cat->prop_obj->component(b1_i));
            double b1_prop_q_lp = b1->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b1_q_lp;
            cat_la += b2_q_lp;
            cat_la -= b1_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/*---------------  Metropolis::Block_from_quartet_death  -------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_from_quartet_death::Block_from_quartet_death
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Death(name, p, pair_p)
{
}

Metropolis::Block_from_quartet_death* 
Metropolis::Block_from_quartet_death::clone() const
{
    return new Block_from_quartet_death(*this);
}

bool Metropolis::Block_from_quartet_death::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Quartet::detachable_block_from_quartet(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Block_from_quartet_death::init(Metropolis::Thread_manager* mgr)
{
    using jwscxx::base::Arg_error;

    Death::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _detach_i = Quartet::detachable_block_from_quartet(
                cat->obj->components()); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Block* b = dynamic_cast<Block*>(cat->obj->component(_detach_i));

        try { _detach_pt_i = b->select_random_used_attach_pt(); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Block_from_quartet_death::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b2_i = detach_i();
    uint32_t b2_j = detach_pt_i();

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Block* b2 = dynamic_cast<Block*>(prop_obj->component(b2_i));
        Quartet* b1 = dynamic_cast<Quartet*>(b2->attachment(b2_j));
        uint32_t b1_j = b1->attachment(b2);
        double b1_q_lp = b1->calc_var_param_proposal_log_likelihood();
        double b2_q_lp = b2->calc_var_param_proposal_log_likelihood();
        if (b1->conditions(b2))
        {
            b1->scale_size_stats(b2, b2_j);
        }

        delete prop_obj->detach(b1, b1_j, b2, b2_j);

        double b1_prop_q_lp = b1->calc_var_param_proposal_log_likelihood();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b1_q_lp;
        la += b2_q_lp;
        la -= b1_prop_q_lp;

        ins_log_alpha(la, i);
    }
}

void Metropolis::Block_from_quartet_death::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b2_i = detach_i();
            uint32_t b2_j = detach_pt_i();
            Block* b2 = dynamic_cast<Block*>(cat->obj->component(b2_i));
            double b2_q_lp = b2->calc_var_param_prior_log_prob();

            Quartet* b1 = dynamic_cast<Quartet*>(b2->attachment(b2_j));
            uint32_t b1_i = cat->obj->component(b1);
            double b1_q_lp = b1->calc_var_param_prior_log_prob();

            b1_i = (b1_i > b2_i) ? b1_i-1 : b1_i;
            b1 = dynamic_cast<Quartet*>(cat->prop_obj->component(b1_i));
            double b1_prop_q_lp = b1->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b1_q_lp;
            cat_la += b2_q_lp;
            cat_la -= b1_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/* ----------------------  Metropolis::Cross_death  ------------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Cross_death::Cross_death
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Death(name, p, pair_p)
{
}

Metropolis::Cross_death* Metropolis::Cross_death::clone() const
{
    return new Cross_death(*this);
}

bool Metropolis::Cross_death::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Block_set::detachable_cross_blocks(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Cross_death::init(Metropolis::Thread_manager* mgr)
{
    using jwscxx::base::Arg_error;

    Death::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _detach_i = Block_set::detachable_cross_blocks(
                cat->obj->components()); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        Block_set* b = dynamic_cast<Block_set*>(cat->obj->component(_detach_i));
        const Model::Parameter::Context* ctx = b->cross_p(0)->context();
        double m, s;
        ctx->sample_prior(&m, &s);
        _one_cross = (ctx->likelihood().sample(m, s, &(ctx->range())) <= 1.5);

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Cross_death::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b_i = detach_i();

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Block_set* b = dynamic_cast<Block_set*>(prop_obj->component(b_i));

        double b_q_lp = b->calc_var_param_proposal_log_likelihood();

        uint32_t N = b->num_cross_blocks();
        b->remove_cross_blocks(one_cross() ? N-1 : N);

        double b_prop_q_lp = b->calc_var_param_proposal_log_likelihood();

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b_q_lp;
        la -= b_prop_q_lp;

        ins_log_alpha(la, i);
    }
}

void Metropolis::Cross_death::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b_i = detach_i();
            Block_set* b = dynamic_cast<Block_set*>(cat->obj->component(b_i));
            double b_q_lp = b->calc_var_param_prior_log_prob();

            b = dynamic_cast<Block_set*>(cat->prop_obj->component(b_i));
            double b_prop_q_lp = b->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b_q_lp;
            cat_la -= b_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

bool Metropolis::Cross_death::one_cross()
{
    return _one_cross;
}

/* ========================================================================== */







/* --------------------------  Metropolis::Split  --------------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Split::Split(const char* name, double p, double pair_p) 
    throw (jwscxx::base::Arg_error) : Jump(name, p, pair_p)
{
    _means_1   = 0;
    _stdevs_1  = 0;
    _means_2   = 0;
    _stdevs_2  = 0;
    _pt_means  = 0;
    _pt_stdevs = 0;
}

Metropolis::Split::~Split()
{
    using namespace jwsc;

    free_vector_d(_means_1);
    free_vector_d(_stdevs_1);
    free_vector_d(_means_2);
    free_vector_d(_stdevs_2);
    free_vector_d(_pt_means);
    free_vector_d(_pt_stdevs);
}

void Metropolis::Split::init
(
    Metropolis::Thread_manager* mgr
)
{
    Jump::init(mgr);
}

uint32_t Metropolis::Split::split_i()
{
    return _split_i;
}

uint32_t Metropolis::Split::split_pt_1_i()
{
    return _split_pt_1_i;
}

uint32_t Metropolis::Split::split_pt_2_i()
{
    return _split_pt_2_i;
}

Component::Attach_type Metropolis::Split::split_pt_1_t()
{
    return _split_pt_1_t;
}

Component::Attach_type Metropolis::Split::split_pt_2_t()
{
    return _split_pt_2_t;
}

double Metropolis::Split::split_1_r()
{
    return _split_1_r;
}

double Metropolis::Split::split_2_r()
{
    return _split_2_r;
}

const std::vector<uint32_t>& Metropolis::Split::split_1_a()
{
    return _split_1_a;
}

const std::vector<uint32_t>& Metropolis::Split::split_2_a()
{
    return _split_2_a;
}

uint32_t Metropolis::Split::c_posn_i()
{
    return _c_posn_i;
}

const jwsc::Vector_d* Metropolis::Split::means_1() const
{
    return _means_1;
}

const jwsc::Vector_d* Metropolis::Split::stdevs_1() const
{
    return _stdevs_1;
}

const jwsc::Vector_d* Metropolis::Split::means_2() const
{
    return _means_2;
}

const jwsc::Vector_d* Metropolis::Split::stdevs_2() const
{
    return _stdevs_2;
}

const jwsc::Vector_d* Metropolis::Split::pt_means() const
{
    return _pt_means;
}

const jwsc::Vector_d* Metropolis::Split::pt_stdevs() const
{
    return _pt_stdevs;
}

void Metropolis::Split::sample_var_param_stats
(
    Metropolis::Thread_manager* mgr,
    uint32_t                    i,
    const jwsc::Vector_d*       bs_m,
    const jwsc::Vector_d*       bs_s,
    const Abstract_block*       b1, 
    const Abstract_block*       b2
)
{
    using namespace jwsc;

    if (mgr->thread() < i)
        return;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        b1->sample_var_param_prior(&_means_1, &_stdevs_1);
        b2->sample_var_param_prior(&_means_2, &_stdevs_2);

        for (uint32_t whl = 0; whl < 3; whl++)
        {
            _stdevs_1->elts[ whl ] *= 10;
            _stdevs_2->elts[ whl ] *= 10;
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    mgr->desynchronize();
}

/* ========================================================================== */







/* ------------------  Metropolis::Block_to_block_split  -------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_to_block_split::Block_to_block_split
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Split(name, p, pair_p)
{
}

Metropolis::Block_to_block_split* Metropolis::Block_to_block_split::clone() 
    const
{
    return new Block_to_block_split(*this);
}

bool Metropolis::Block_to_block_split::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Block::splitable_block_to_block(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return (components.size() < MAX_NUM_COMPONENTS);
}

void Metropolis::Block_to_block_split::init
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;
    using jwscxx::base::Arg_error;
    typedef Component::Attach_type Attach_t;

    Split::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _split_i = cat->obj->select_random_block(); }
        catch (Arg_error e) { abort(); }

        const Block* b = dynamic_cast<Block*>(cat->obj->component(_split_i));

        try { _split_pt_1_i = b->select_random_attach_pt(); }
        catch (Arg_error e) { abort(); }
        _split_pt_2_i = (_split_pt_1_i % 2 == 0) 
            ? _split_pt_1_i+1 : _split_pt_1_i-1;

        _split_pt_1_t = b->select_random_attach_pt_type();
        _split_pt_2_t = (_split_pt_1_t == Component::CTR_ATTACH) 
            ? Component::VAR_ATTACH : Component::CTR_ATTACH;

        _c_posn_i = _split_i;
        if (cat->obj->num_components() > 1)
        {
            while (_c_posn_i == _split_i)
            {
                _c_posn_i = cat->obj->select_random_component();
            }
        }

        _split_1_r = sample_uniform_pdf_d(0.05,0.95); 
        _split_2_r = 1.0 - _split_1_r;

        _split_1_a.clear();
        _split_2_a.clear();
        _split_1_a.insert(_split_1_a.end(), 6, 0);
        _split_2_a.insert(_split_2_a.end(), 6, 0);
        for (uint32_t i = 0; i < 6; i++)
        {
            if (!b->attachment(i))
                continue;

            if (i == _split_pt_1_i)
            {
                _split_2_a[ i ] = 1;
            }
            else if (i == _split_pt_2_i)
            {
                _split_1_a[ i ] = 1;
            }
            else if (sample_uniform_pdf_d(0,1) < 0.5)
            {
                _split_1_a[ i ] = 1;
            }
            else
            {
                _split_2_a[ i ] = 1;
            }
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Block_to_block_split::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using std::vector;
    using namespace jwsc;

    typedef Abstract_block          Ablock;
    typedef Component::Attach_type  Attach_t;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t bs_i = split_i();
    uint32_t b1_j = split_pt_1_i();
    uint32_t b2_j = split_pt_2_i();
    Attach_t b1_t = Component::VAR_ATTACH;
    Attach_t b2_t = Component::CTR_ATTACH;
    double   b1_r = split_1_r();
    double   b2_r = split_2_r();
    const vector<uint32_t>& b1_a = split_1_a();
    const vector<uint32_t>& b2_a = split_2_a();

    Vector_d* q  = 0; 
    Vector_d* m  = 0; 
    Vector_d* uv = 0;
    Vector_d* p1 = 0;
    Vector_d* p2 = 0;
    Vector_d* bs_m = 0;
    Vector_d* bs_s = 0;
    double    sz1[3];
    double    sz2[3];

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Block* bs = dynamic_cast<Block*>(prop_obj->component(bs_i));
        bs->get_var_param_stats(&bs_m, &bs_s);

        for (uint32_t whl = 0; whl < 3; whl++)
        {
            if (whl == (b1_j/2))
            {
                sz1[ whl ] = b1_r*(bs->size(whl));
                sz2[ whl ] = b2_r*(bs->size(whl));
                clip_param_val(&(sz1[ whl ]), params->block_ctx()->size_c(whl));
                clip_param_val(&(sz2[ whl ]), params->block_ctx()->size_c(whl));
                bs->size_shift_deps(whl, sz1[ whl ]+sz2[ whl ]);
            }
            else
            {
                sz1[ whl ] = sz2[ whl ] = bs->size(whl);
            }
        }

        Block* b1 = new Block(sz1[0], sz1[1], sz1[2], params->block_ctx());
        Block* b2 = new Block(sz2[0], sz2[1], sz2[2], params->block_ctx());
        prop_obj->split(bs, b1, b1_j, b1_t, b1_a, b2, b2_j, b2_t, b2_a);
        delete bs;

        sample_var_param_stats(mgr, i, bs_m, bs_s, b1, b2);
        try { b1->set_var_param_stats(means_1(), stdevs_1()); }
        catch (jwscxx::base::Arg_error e) { }
        try { b2->set_var_param_stats(means_2(), stdevs_2()); }
        catch (jwscxx::base::Arg_error e) { }

        b2->posn_v(&p1);

        b1->get_var_param_values(&m);
        b1->sample_var_param_split_merge_proposal(&q, m, b1_j);
        try { b1->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double b_q_lp = b1->calc_var_param_split_merge_proposal_log_prob(
                q, m, b1_j);

        b2->posn_v(&p2);
        subtract_vectors_d(&p1, p1, p2);
        prop_obj->add_posn_shift_deps_v(p1);

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la -= b_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(p1);
    free_vector_d(p2);
    free_vector_d(uv);
    free_vector_d(m);
    free_vector_d(q);
    free_vector_d(bs_m);
    free_vector_d(bs_s);
}

void Metropolis::Block_to_block_split::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t bs_i = split_i();
            Block* bs = dynamic_cast<Block*>(cat->obj->component(bs_i));
            double bs_q_lp = bs->calc_var_param_prior_log_prob();

            uint32_t b1_i = cat->prop_obj->num_components() - 2;
            Block* b1 = dynamic_cast<Block*>(cat->prop_obj->component(b1_i));
            double b1_prop_q_lp = b1->calc_var_param_prior_log_prob();

            uint32_t b2_i = cat->prop_obj->num_components() - 1;
            Block* b2 = dynamic_cast<Block*>(cat->prop_obj->component(b2_i));
            double b2_prop_q_lp = b2->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += bs_q_lp;
            cat_la -= b1_prop_q_lp;
            cat_la -= b2_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/* ------------------  Metropolis::Pair_to_block_split  --------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 * @param  height_only  Only allow height attachments to a block set.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Pair_to_block_split::Pair_to_block_split
(
    const char* name, 
    double      p,
    double      pair_p,
    bool        height_only
) 
throw (jwscxx::base::Arg_error) 
    : Split(name, p, pair_p)
{
    _height_only = height_only;
}

Metropolis::Pair_to_block_split* Metropolis::Pair_to_block_split::clone() 
    const
{
    return new Pair_to_block_split(*this);
}

bool Metropolis::Pair_to_block_split::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Block_set::splitable_block_set_to_block(components, _height_only); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return (components.size() < MAX_NUM_COMPONENTS);
}

void Metropolis::Pair_to_block_split::init
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;
    using jwscxx::base::Arg_error;
    typedef Component::Attach_type Attach_t;

    Split::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _split_i = Block_set::splitable_block_set_to_block(
                cat->obj->components(), _height_only); }
        catch (Arg_error e) { abort(); }

        const Block* b = dynamic_cast<Block*>(cat->obj->component(_split_i));

        try { _split_pt_2_i = 
                 Block_set::splitable_block_set_to_block_attach_pt(b,
                         _height_only); }
        catch (Arg_error e) { abort(); }
        _split_pt_1_i = (_split_pt_2_i % 2 == 0) 
            ? _split_pt_2_i+1 : _split_pt_2_i-1;

        _split_pt_1_t = b->select_random_attach_pt_type();
        _split_pt_2_t = (_split_pt_1_t == Component::CTR_ATTACH) 
            ? Component::VAR_ATTACH : Component::CTR_ATTACH;

        _c_posn_i = _split_i;
        if (cat->obj->num_components() > 1)
        {
            while (_c_posn_i == _split_i)
            {
                _c_posn_i = cat->obj->select_random_component();
            }
        }

        _split_1_r = sample_uniform_pdf_d(0.05,0.95); 
        _split_2_r = 1.0 - _split_1_r;

        _split_1_a.clear();
        _split_2_a.clear();
        _split_1_a.insert(_split_1_a.end(), 6, 0);
        _split_2_a.insert(_split_2_a.end(), 6, 0);
        for (uint32_t i = 0; i < 6; i++)
        {
            if (!b->attachment(i))
                continue;

            bool not_a_bs = !(dynamic_cast<Block_set*>(b->attachment(i)));

            if (i == _split_pt_1_i)
            {
                _split_2_a[ i ] = 1;
            }
            else if (i == _split_pt_2_i)
            {
                _split_1_a[ i ] = 1;
            }
            else if (not_a_bs && sample_uniform_pdf_d(0,1) < 0.5)
            {
                _split_1_a[ i ] = 1;
            }
            else
            {
                _split_2_a[ i ] = 1;
            }

            if (_split_1_a[ i ])
            {
                assert(not_a_bs);
            }
        }

        _layout = Pair::select_random_layout();

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Pair_to_block_split::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using std::vector;
    using namespace jwsc;

    typedef Abstract_block          Ablock;
    typedef Component::Attach_type  Attach_t;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t bs_i = split_i();
    uint32_t b1_j = split_pt_1_i();
    uint32_t b2_j = split_pt_2_i();
    Attach_t b1_t = Component::VAR_ATTACH;
    Attach_t b2_t = Component::CTR_ATTACH;
    double   b1_r = split_1_r();
    double   b2_r = split_2_r();
    const vector<uint32_t>& b1_a = split_1_a();
    const vector<uint32_t>& b2_a = split_2_a();

    Vector_d* q  = 0; 
    Vector_d* m  = 0;
    Vector_d* uv = 0;
    Vector_d* p1 = 0;
    Vector_d* p2 = 0;
    Vector_d* bs_m = 0;
    Vector_d* bs_s = 0;
    double    sz1[4];
    double    sz2[3];

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Block* bs = dynamic_cast<Block*>(prop_obj->component(bs_i));
        bs->get_var_param_stats(&bs_m, &bs_s);

        for (uint32_t whl = 0; whl < 3; whl++)
        {
            if (whl == (b1_j/2))
            {
                sz1[ whl ] = b1_r*(bs->size(whl));
                sz2[ whl ] = b2_r*(bs->size(whl));
                clip_param_val(&(sz1[ whl ]), params->block_ctx()->size_c(whl));
                clip_param_val(&(sz2[ whl ]), params->block_ctx()->size_c(whl));
                bs->size_shift_deps(whl, sz1[ whl ]+sz2[ whl ]);
            }
            else
            {
                sz1[ whl ] = sz2[ whl ] = bs->size(whl);
            }
        }
        sz1[3] = params->block_set_ctx()->inner_size_c(0).sample_proposal();

        Pair* b1 = new Pair(sz1[0], sz1[1], sz1[2], sz1[3], layout(),
                params->block_set_ctx());
        Block* b2 = new Block(sz2[0], sz2[1], sz2[2], params->block_ctx());
        prop_obj->split(bs, b1, b1_j, b1_t, b1_a, b2, b2_j, b2_t, b2_a);
        delete bs;

        sample_var_param_stats(mgr, i, bs_m, bs_s, b1, b2);
        try { b1->set_var_param_stats(means_1(), stdevs_1()); }
        catch (jwscxx::base::Arg_error e) { }
        try { b2->set_var_param_stats(means_2(), stdevs_2()); }
        catch (jwscxx::base::Arg_error e) { }

        b2->posn_v(&p1);

        b1->get_var_param_values(&m);
        b1->sample_var_param_split_merge_proposal(&q, m, b1_j);
        try { b1->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double b_q_lp = b1->calc_var_param_split_merge_proposal_log_prob(
                q, m, b1_j);

        b2->posn_v(&p2);
        subtract_vectors_d(&p1, p1, p2);
        prop_obj->add_posn_shift_deps_v(p1);

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la -= b_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(p1);
    free_vector_d(p2);
    free_vector_d(uv);
    free_vector_d(m);
    free_vector_d(q);
    free_vector_d(bs_m);
    free_vector_d(bs_s);
}

void Metropolis::Pair_to_block_split::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;
    typedef Abstract_block Ablock;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t bs_i = split_i();
            Ablock* bs = dynamic_cast<Ablock*>(cat->obj->component(bs_i));
            double bs_q_lp = bs->calc_var_param_prior_log_prob();

            uint32_t b1_i = cat->prop_obj->num_components() - 2;
            Ablock* b1 = dynamic_cast<Ablock*>(cat->prop_obj->component(b1_i));
            double b1_prop_q_lp = b1->calc_var_param_prior_log_prob();

            uint32_t b2_i = cat->prop_obj->num_components() - 1;
            Ablock* b2 = dynamic_cast<Ablock*>(cat->prop_obj->component(b2_i));
            double b2_prop_q_lp = b2->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += bs_q_lp;
            cat_la -= b1_prop_q_lp;
            cat_la -= b2_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

Pair::Layout Metropolis::Pair_to_block_split::layout()
{
    return _layout;
}

/* ========================================================================== */







/* -----------------  Metropolis::Block_to_Pair_split  ---------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_to_pair_split::Block_to_pair_split
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Split(name, p, pair_p)
{
}

Metropolis::Block_to_pair_split* Metropolis::Block_to_pair_split::clone() 
    const
{
    return new Block_to_pair_split(*this);
}

bool Metropolis::Block_to_pair_split::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Pair::splitable_block_to_pair(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return (components.size() < MAX_NUM_COMPONENTS);
}

void Metropolis::Block_to_pair_split::init
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;
    using jwscxx::base::Arg_error;
    typedef Component::Attach_type Attach_t;

    Split::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _split_i = Pair::splitable_block_to_pair(cat->obj->components());}
        catch (Arg_error e) { abort(); }

        const Pair* b = dynamic_cast<Pair*>(cat->obj->component(_split_i));

        try { _split_pt_1_i = Pair::splitable_block_to_pair_attach_pt(b); }
        catch (Arg_error e) { abort(); }
        _split_pt_2_i = (_split_pt_1_i % 2 == 0) 
            ? _split_pt_1_i+1 : _split_pt_1_i-1;

        _split_pt_1_t = b->select_random_attach_pt_type();
        _split_pt_2_t = (_split_pt_1_t == Component::CTR_ATTACH) 
            ? Component::VAR_ATTACH : Component::CTR_ATTACH;

        _c_posn_i = _split_i;
        if (cat->obj->num_components() > 1)
        {
            while (_c_posn_i == _split_i)
            {
                _c_posn_i = cat->obj->select_random_component();
            }
        }

        _split_1_r = sample_uniform_pdf_d(0.05,0.95); 
        _split_2_r = 1.0 - _split_1_r;

        _split_1_a.clear();
        _split_2_a.clear();
        _split_1_a.insert(_split_1_a.end(), 6, 0);
        _split_2_a.insert(_split_2_a.end(), 6, 0);
        for (uint32_t i = 0; i < 6; i++)
        {
            if (!b->attachment(i))
                continue;

            if (i == _split_pt_1_i)
            {
                _split_2_a[ i ] = 1;
            }
            else if (i == _split_pt_2_i)
            {
                _split_1_a[ i ] = 1;
            }
            else if (sample_uniform_pdf_d(0,1) < 0.5)
            {
                _split_1_a[ i ] = 1;
            }
            else
            {
                _split_2_a[ i ] = 1;
            }
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Block_to_pair_split::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using std::vector;
    using namespace jwsc;

    typedef Abstract_block          Ablock;
    typedef Component::Attach_type  Attach_t;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t bs_i = split_i();
    uint32_t b1_j = split_pt_1_i();
    uint32_t b2_j = split_pt_2_i();
    Attach_t b1_t = Component::VAR_ATTACH;
    Attach_t b2_t = Component::CTR_ATTACH;
    double   b1_r = split_1_r();
    double   b2_r = split_2_r();
    const vector<uint32_t>& b1_a = split_1_a();
    const vector<uint32_t>& b2_a = split_2_a();

    Vector_d* q  = 0; 
    Vector_d* m  = 0;
    Vector_d* uv = 0;
    Vector_d* p1 = 0;
    Vector_d* p2 = 0;
    Vector_d* bs_m = 0;
    Vector_d* bs_s = 0;
    double    sz1[3];
    double    sz2[4];

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Pair* bs = dynamic_cast<Pair*>(prop_obj->component(bs_i));
        bs->get_var_param_stats(&bs_m, &bs_s);

        for (uint32_t whl = 0; whl < 3; whl++)
        {
            if (whl == (b1_j/2))
            {
                sz1[ whl ] = b1_r*(bs->size(whl));
                sz2[ whl ] = b2_r*(bs->size(whl));
                clip_param_val(&(sz1[ whl ]), params->block_ctx()->size_c(whl));
                clip_param_val(&(sz2[ whl ]), params->block_ctx()->size_c(whl));
                bs->size_shift_deps(whl, sz1[ whl ]+sz2[ whl ]);
            }
            else
            {
                sz1[ whl ] = sz2[ whl ] = bs->size(whl);
            }
        }
        sz2[3] = bs->inner_size(0);

        Block* b1 = new Block(sz1[0], sz1[1], sz1[2], params->block_ctx());
        Pair* b2 = new Pair(sz2[0], sz2[1], sz2[2], sz2[3], 
                //bs->cross(0), bs->cross(1), bs->cross(2), bs->cross(3), 
                bs->layout(), params->block_set_ctx());
        prop_obj->split(bs, b1, b1_j, b1_t, b1_a, b2, b2_j, b2_t, b2_a);
        delete bs;

        // TODO b2's cross block stats are re-sampled when they shouldn't be
        sample_var_param_stats(mgr, i, bs_m, bs_s, b1, b2);
        try { b1->set_var_param_stats(means_1(), stdevs_1()); }
        catch (jwscxx::base::Arg_error e) { }
        try { b2->set_var_param_stats(means_2(), stdevs_2()); }
        catch (jwscxx::base::Arg_error e) { }

        b2->posn_v(&p1);

        b1->get_var_param_values(&m);
        b1->sample_var_param_split_merge_proposal(&q, m, b1_j);
        try { b1->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double b_q_lp = b1->calc_var_param_split_merge_proposal_log_prob(
                q, m, b1_j);

        b2->posn_v(&p2);
        subtract_vectors_d(&p1, p1, p2);
        prop_obj->add_posn_shift_deps_v(p1);

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la -= b_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(p1);
    free_vector_d(p2);
    free_vector_d(uv);
    free_vector_d(m);
    free_vector_d(q);
    free_vector_d(bs_m);
    free_vector_d(bs_s);
}

void Metropolis::Block_to_pair_split::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;
    typedef Abstract_block Ablock;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t bs_i = split_i();
            Ablock* bs = dynamic_cast<Ablock*>(cat->obj->component(bs_i));
            double bs_q_lp = bs->calc_var_param_prior_log_prob();

            uint32_t b1_i = cat->prop_obj->num_components() - 2;
            Ablock* b1 = dynamic_cast<Ablock*>(cat->prop_obj->component(b1_i));
            double b1_prop_q_lp = b1->calc_var_param_prior_log_prob();

            uint32_t b2_i = cat->prop_obj->num_components() - 1;
            Ablock* b2 = dynamic_cast<Ablock*>(cat->prop_obj->component(b2_i));
            double b2_prop_q_lp = b2->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += bs_q_lp;
            cat_la -= b1_prop_q_lp;
            cat_la -= b2_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/* ----------------  Metropolis::Quartet_to_block_split  -------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 * @param  height_only  Only allow height attachments to a block set.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Quartet_to_block_split::Quartet_to_block_split
(
    const char* name, 
    double      p,
    double      pair_p,
    bool        height_only
) 
throw (jwscxx::base::Arg_error) 
    : Split(name, p, pair_p)
{
    _height_only = height_only;
}

Metropolis::Quartet_to_block_split* Metropolis::Quartet_to_block_split::clone() 
    const
{
    return new Quartet_to_block_split(*this);
}

bool Metropolis::Quartet_to_block_split::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Block_set::splitable_block_set_to_block(components, _height_only); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return (components.size() < MAX_NUM_COMPONENTS);
}

void Metropolis::Quartet_to_block_split::init
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;
    using jwscxx::base::Arg_error;
    typedef Component::Attach_type Attach_t;

    Split::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _split_i = Block_set::splitable_block_set_to_block(
                cat->obj->components(), _height_only); }
        catch (Arg_error e) { abort(); }

        const Block* b = dynamic_cast<Block*>(cat->obj->component(_split_i));

        try { _split_pt_2_i = 
                 Block_set::splitable_block_set_to_block_attach_pt(b,
                         _height_only); }
        catch (Arg_error e) { abort(); }
        _split_pt_1_i = (_split_pt_2_i % 2 == 0) 
            ? _split_pt_2_i+1 : _split_pt_2_i-1;

        _split_pt_1_t = b->select_random_attach_pt_type();
        _split_pt_2_t = (_split_pt_1_t == Component::CTR_ATTACH) 
            ? Component::VAR_ATTACH : Component::CTR_ATTACH;

        _c_posn_i = _split_i;
        if (cat->obj->num_components() > 1)
        {
            while (_c_posn_i == _split_i)
            {
                _c_posn_i = cat->obj->select_random_component();
            }
        }

        _split_1_r = sample_uniform_pdf_d(0.05,0.95); 
        _split_2_r = 1.0 - _split_1_r;

        _split_1_a.clear();
        _split_2_a.clear();
        _split_1_a.insert(_split_1_a.end(), 6, 0);
        _split_2_a.insert(_split_2_a.end(), 6, 0);
        for (uint32_t i = 0; i < 6; i++)
        {
            if (!b->attachment(i))
                continue;

            bool not_a_bs = !(dynamic_cast<Block_set*>(b->attachment(i)));

            if (i == _split_pt_1_i)
            {
                _split_2_a[ i ] = 1;
            }
            else if (i == _split_pt_2_i)
            {
                _split_1_a[ i ] = 1;
            }
            else if (not_a_bs && sample_uniform_pdf_d(0,1) < 0.5)
            {
                _split_1_a[ i ] = 1;
            }
            else
            {
                _split_2_a[ i ] = 1;
            }

            if (_split_1_a[ i ])
            {
                assert(not_a_bs);
            }
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Quartet_to_block_split::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using std::vector;
    using namespace jwsc;

    typedef Abstract_block          Ablock;
    typedef Component::Attach_type  Attach_t;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t bs_i = split_i();
    uint32_t b1_j = split_pt_1_i();
    uint32_t b2_j = split_pt_2_i();
    Attach_t b1_t = Component::VAR_ATTACH;
    Attach_t b2_t = Component::CTR_ATTACH;
    double   b1_r = split_1_r();
    double   b2_r = split_2_r();
    const vector<uint32_t>& b1_a = split_1_a();
    const vector<uint32_t>& b2_a = split_2_a();

    Vector_d* q  = 0; 
    Vector_d* m  = 0;
    Vector_d* uv = 0;
    Vector_d* p1 = 0;
    Vector_d* p2 = 0;
    Vector_d* bs_m = 0;
    Vector_d* bs_s = 0;
    double    sz1[5];
    double    sz2[3];

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Block* bs = dynamic_cast<Block*>(prop_obj->component(bs_i));
        bs->get_var_param_stats(&bs_m, &bs_s);

        for (uint32_t whl = 0; whl < 3; whl++)
        {
            if (whl == (b1_j/2))
            {
                sz1[ whl ] = b1_r*(bs->size(whl));
                sz2[ whl ] = b2_r*(bs->size(whl));
                clip_param_val(&(sz1[ whl ]), params->block_ctx()->size_c(whl));
                clip_param_val(&(sz2[ whl ]), params->block_ctx()->size_c(whl));
                bs->size_shift_deps(whl, sz1[ whl ]+sz2[ whl ]);
            }
            else
            {
                sz1[ whl ] = sz2[ whl ] = bs->size(whl);
            }
        }
        sz1[3] = params->block_set_ctx()->inner_size_c(0).sample_proposal();
        sz1[4] = params->block_set_ctx()->inner_size_c(1).sample_proposal();

        Quartet* b1 = new Quartet(sz1[0], sz1[1], sz1[2], sz1[3], sz1[4], 
                params->block_set_ctx());
        Block* b2 = new Block(sz2[0], sz2[1], sz2[2], params->block_ctx());
        prop_obj->split(bs, b1, b1_j, b1_t, b1_a, b2, b2_j, b2_t, b2_a);
        delete bs;

        sample_var_param_stats(mgr, i, bs_m, bs_s, b1, b2);
        try { b1->set_var_param_stats(means_1(), stdevs_1()); }
        catch (jwscxx::base::Arg_error e) { }
        try { b2->set_var_param_stats(means_2(), stdevs_2()); }
        catch (jwscxx::base::Arg_error e) { }

        b2->posn_v(&p1);

        b1->get_var_param_values(&m);
        b1->sample_var_param_split_merge_proposal(&q, m, b1_j);
        try { b1->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double b_q_lp = b1->calc_var_param_split_merge_proposal_log_prob(
                q, m, b1_j);

        b2->posn_v(&p2);
        subtract_vectors_d(&p1, p1, p2);
        prop_obj->add_posn_shift_deps_v(p1);

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la -= b_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(p1);
    free_vector_d(p2);
    free_vector_d(uv);
    free_vector_d(m);
    free_vector_d(q);
    free_vector_d(bs_m);
    free_vector_d(bs_s);
}

void Metropolis::Quartet_to_block_split::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;
    typedef Abstract_block Ablock;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t bs_i = split_i();
            Ablock* bs = dynamic_cast<Ablock*>(cat->obj->component(bs_i));
            double bs_q_lp = bs->calc_var_param_prior_log_prob();

            uint32_t b1_i = cat->prop_obj->num_components() - 2;
            Ablock* b1 = dynamic_cast<Ablock*>(cat->prop_obj->component(b1_i));
            double b1_prop_q_lp = b1->calc_var_param_prior_log_prob();

            uint32_t b2_i = cat->prop_obj->num_components() - 1;
            Ablock* b2 = dynamic_cast<Ablock*>(cat->prop_obj->component(b2_i));
            double b2_prop_q_lp = b2->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += bs_q_lp;
            cat_la -= b1_prop_q_lp;
            cat_la -= b2_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/* ----------------  Metropolis::Block_to_quartet_split  -------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_to_quartet_split::Block_to_quartet_split
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Split(name, p, pair_p)
{
}

Metropolis::Block_to_quartet_split* Metropolis::Block_to_quartet_split::clone() 
    const
{
    return new Block_to_quartet_split(*this);
}

bool Metropolis::Block_to_quartet_split::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Quartet::splitable_block_to_quartet(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return (components.size() < MAX_NUM_COMPONENTS);
}

void Metropolis::Block_to_quartet_split::init
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;
    using jwscxx::base::Arg_error;
    typedef Component::Attach_type Attach_t;

    Split::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _split_i = Quartet::splitable_block_to_quartet(
                cat->obj->components()); }
        catch (Arg_error e) { abort(); }

        const Quartet* b = dynamic_cast<Quartet*>(
                cat->obj->component(_split_i));

        try { _split_pt_1_i = Quartet::splitable_block_to_quartet_attach_pt(b);}
        catch (Arg_error e) { abort(); }
        _split_pt_2_i = (_split_pt_1_i % 2 == 0) 
            ? _split_pt_1_i+1 : _split_pt_1_i-1;

        _split_pt_1_t = b->select_random_attach_pt_type();
        _split_pt_2_t = (_split_pt_1_t == Component::CTR_ATTACH) 
            ? Component::VAR_ATTACH : Component::CTR_ATTACH;

        _c_posn_i = _split_i;
        if (cat->obj->num_components() > 1)
        {
            while (_c_posn_i == _split_i)
            {
                _c_posn_i = cat->obj->select_random_component();
            }
        }

        _split_1_r = sample_uniform_pdf_d(0.05,0.95); 
        _split_2_r = 1.0 - _split_1_r;

        _split_1_a.clear();
        _split_2_a.clear();
        _split_1_a.insert(_split_1_a.end(), 6, 0);
        _split_2_a.insert(_split_2_a.end(), 6, 0);
        for (uint32_t i = 0; i < 6; i++)
        {
            if (!b->attachment(i))
                continue;

            if (i == _split_pt_1_i)
            {
                _split_2_a[ i ] = 1;
            }
            else if (i == _split_pt_2_i)
            {
                _split_1_a[ i ] = 1;
            }
            else if (sample_uniform_pdf_d(0,1) < 0.5)
            {
                _split_1_a[ i ] = 1;
            }
            else
            {
                _split_2_a[ i ] = 1;
            }
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Block_to_quartet_split::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using std::vector;
    using namespace jwsc;

    typedef Abstract_block          Ablock;
    typedef Component::Attach_type  Attach_t;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t bs_i = split_i();
    uint32_t b1_j = split_pt_1_i();
    uint32_t b2_j = split_pt_2_i();
    Attach_t b1_t = Component::VAR_ATTACH;
    Attach_t b2_t = Component::CTR_ATTACH;
    double   b1_r = split_1_r();
    double   b2_r = split_2_r();
    const vector<uint32_t>& b1_a = split_1_a();
    const vector<uint32_t>& b2_a = split_2_a();

    Vector_d* q  = 0; 
    Vector_d* m  = 0;
    Vector_d* uv = 0;
    Vector_d* p1 = 0;
    Vector_d* p2 = 0;
    Vector_d* bs_m = 0;
    Vector_d* bs_s = 0;
    double    sz1[3];
    double    sz2[5];

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Quartet* bs = dynamic_cast<Quartet*>(prop_obj->component(bs_i));
        bs->get_var_param_stats(&bs_m, &bs_s);

        for (uint32_t whl = 0; whl < 3; whl++)
        {
            if (whl == (b1_j/2))
            {
                sz1[ whl ] = b1_r*(bs->size(whl));
                sz2[ whl ] = b2_r*(bs->size(whl));
                clip_param_val(&(sz1[ whl ]), params->block_ctx()->size_c(whl));
                clip_param_val(&(sz2[ whl ]), params->block_ctx()->size_c(whl));
                bs->size_shift_deps(whl, sz1[ whl ]+sz2[ whl ]);
            }
            else
            {
                sz1[ whl ] = sz2[ whl ] = bs->size(whl);
            }
        }
        sz2[3] = bs->inner_size(0);
        sz2[4] = bs->inner_size(1);

        Block* b1 = new Block(sz1[0], sz1[1], sz1[2], params->block_ctx());
        Quartet* b2 = new Quartet(sz2[0], sz2[1], sz2[2], sz2[3], sz2[4], 
                //bs->cross(0), bs->cross(1), bs->cross(2), bs->cross(3),
                params->block_set_ctx());
        prop_obj->split(bs, b1, b1_j, b1_t, b1_a, b2, b2_j, b2_t, b2_a);
        delete bs;

        sample_var_param_stats(mgr, i, bs_m, bs_s, b1, b2);
        try { b1->set_var_param_stats(means_1(), stdevs_1()); }
        catch (jwscxx::base::Arg_error e) { }
        try { b2->set_var_param_stats(means_2(), stdevs_2()); }
        catch (jwscxx::base::Arg_error e) { }

        b2->posn_v(&p1);

        b1->get_var_param_values(&m);
        b1->sample_var_param_split_merge_proposal(&q, m, b1_j);
        try { b1->set_var_param_values(q); }
        catch (jwscxx::base::Arg_error e) { }
        double b_q_lp = b1->calc_var_param_split_merge_proposal_log_prob(
                q, m, b1_j);

        b2->posn_v(&p2);
        subtract_vectors_d(&p1, p1, p2);
        prop_obj->add_posn_shift_deps_v(p1);

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la -= b_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(p1);
    free_vector_d(p2);
    free_vector_d(uv);
    free_vector_d(m);
    free_vector_d(q);
    free_vector_d(bs_m);
    free_vector_d(bs_s);
}

void Metropolis::Block_to_quartet_split::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;
    typedef Abstract_block Ablock;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t bs_i = split_i();
            Ablock* bs = dynamic_cast<Ablock*>(cat->obj->component(bs_i));
            double bs_q_lp = bs->calc_var_param_prior_log_prob();

            uint32_t b1_i = cat->prop_obj->num_components() - 2;
            Ablock* b1 = dynamic_cast<Ablock*>(cat->prop_obj->component(b1_i));
            double b1_prop_q_lp = b1->calc_var_param_prior_log_prob();

            uint32_t b2_i = cat->prop_obj->num_components() - 1;
            Ablock* b2 = dynamic_cast<Ablock*>(cat->prop_obj->component(b2_i));
            double b2_prop_q_lp = b2->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += bs_q_lp;
            cat_la -= b1_prop_q_lp;
            cat_la -= b2_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */








/* --------------------------  Metropolis::Merge  --------------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Merge::Merge(const char* name, double p, double pair_p) 
    throw (jwscxx::base::Arg_error) : Jump(name, p, pair_p)
{
    _means  = 0;
    _stdevs = 0;
}

Metropolis::Merge::~Merge()
{
    jwsc::free_vector_d(_means);
    jwsc::free_vector_d(_stdevs);
}

void Metropolis::Merge::init(Metropolis::Thread_manager* mgr)
{
    Jump::init(mgr);
}

uint32_t Metropolis::Merge::merge_1_i()
{
    return _merge_1_i;
}

uint32_t Metropolis::Merge::merge_2_i()
{
    return _merge_2_i;
}

uint32_t Metropolis::Merge::merge_pt_1_i()
{
    return _merge_pt_1_i;
}

uint32_t Metropolis::Merge::merge_pt_2_i()
{
    return _merge_pt_2_i;
}

uint32_t Metropolis::Merge::c_posn_i()
{
    return _c_posn_i;
}

const jwsc::Vector_d* Metropolis::Merge::means() const
{
    return _means;
}

const jwsc::Vector_d* Metropolis::Merge::stdevs() const
{
    return _stdevs;
}

void Metropolis::Merge::sample_var_param_stats
(
    Metropolis::Thread_manager* mgr,
    uint32_t                    i,
    const Abstract_block*       bm,
    const Abstract_block*       b1,
    const Abstract_block*       b2
)
{
    using namespace jwsc;

    if (mgr->thread() < i)
        return;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        bm->sample_var_param_prior(&_means, &_stdevs);

        for (uint32_t whl = 0; whl < 3; whl++)
        {
            _stdevs->elts[ whl ] *= 10;
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    mgr->desynchronize();
}

/** 
 * The thread manager is assumed to be synchronized.
 *
 * The value of @em n is the jump set size and is given by
 * Metropolis::_jump_set_size.
 */
double Metropolis::Merge::sum_n_best_instance_log_alphas
(
    Metropolis::Thread_manager* mgr
)
{
    return Jump::sum_n_best_instance_log_alphas(mgr);
    /*
    Parameters* params = mgr->params();

    uint32_t c   = mgr->category();
    uint32_t N_i = params->num_instances();
    uint32_t N_j = mgr->sampler()->jump_set_size();

    std::vector<double> n_best;
    n_best.reserve(N_i);

    for (uint32_t i = 0; i < N_i; i++)
    {
        if (params->member(c, i) > 0)
        {
            n_best.push_back(ins_log_alpha(i));
        }
    }

    N_i = n_best.size();
    if (N_j == 0 || N_j > N_i)
    {
        N_j = N_i;
    }

    std::sort(n_best.begin(), n_best.end());

    double sum = 0;
    std::vector<double>::iterator it = n_best.begin();
    for (uint32_t j = 0; j < N_j; j++)
    {
        sum += *it;
        it++;
        assert(it != n_best.end());
    }

    return sum;
    */
}

/* ========================================================================== */







/* ------------------  Metropolis::Block_from_block_merge  ------------------ */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_from_block_merge::Block_from_block_merge
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Merge(name, p, pair_p)
{
}

Metropolis::Block_from_block_merge* Metropolis::Block_from_block_merge::clone() 
    const
{
    return new Block_from_block_merge(*this);
}

bool Metropolis::Block_from_block_merge::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Block::mergeable_block_from_block(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Block_from_block_merge::init(Metropolis::Thread_manager* mgr)
{
    using jwscxx::base::Arg_error;

    Merge::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try 
        { _merge_1_i = Block::mergeable_block_from_block(
                cat->obj->components()); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Block* b1 = dynamic_cast<Block*>(cat->obj->component(_merge_1_i));

        try { _merge_pt_1_i = Block::mergeable_block_from_block_attach_pt(b1); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Block* b2 = dynamic_cast<Block*>(b1->attachment(_merge_pt_1_i));

        try { _merge_2_i = cat->obj->component(b2); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        try { _merge_pt_2_i = b2->attachment(b1); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        _c_posn_i = _merge_1_i;
        if (cat->obj->num_components() > 2)
        {
            while (_c_posn_i == _merge_1_i || _c_posn_i == _merge_2_i)
            {
                _c_posn_i = cat->obj->select_random_component();
            }
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Block_from_block_merge::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b1_i = merge_1_i();
    uint32_t b2_i = merge_2_i();
    uint32_t b1_j = merge_pt_1_i();
    uint32_t b2_j = merge_pt_2_i();
    uint32_t whl;

    Vector_d* q  = 0;
    Vector_d* m  = 0;
    Vector_d* p1 = 0;
    Vector_d* p2 = 0;
    double    sz[3];

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Block* b1 = dynamic_cast<Block*>(prop_obj->component(b1_i));
        Block* b2 = dynamic_cast<Block*>(prop_obj->component(b2_i));
        Block* bm = 0;

        for (whl = 0; whl < 3; whl++)
        {
            sz[ whl ] = (whl == (b1_j/2))
                ? b1->size(whl) + b2->size(whl) : b2->size(whl);
        }
        whl = b1_j/2;

        // Try to create a block of size sz. If it fails, resize b1 and b2 to
        // something that should work and try again. No, Virginia, there is no
        // Santa Claus, only ugly hacks.
        try { bm = new Block(sz[0], sz[1], sz[2], params->block_ctx()); }
        catch (jwscxx::base::Arg_error e) 
        { 
            sz[ whl ] = 0.5*b1->size(whl);
            try { b1->test_size_shift_deps(whl, sz[ whl ]); }
            catch (jwscxx::base::Arg_error e) 
            { sz[ whl ] = 0.5*b2->size(whl); }
            b1->size_shift_deps(whl, sz[ whl ]);
            b2->size_shift_deps(whl, sz[ whl ]);
            sz[ whl ] = b1->size(whl) + b2->size(whl);
            bm = new Block(sz[0], sz[1], sz[2], params->block_ctx());
        }

        b1->get_var_param_values(&m);
        b1->sample_var_param_split_merge_proposal(&q, m, b1_j);
        double b_q_lp = b1->calc_var_param_split_merge_proposal_log_prob(
                q, m, b1_j);

        prop_obj->merge(bm, b1, b1_j, b2, b2_j);
        bm->posn_v(&p1);

        sample_var_param_stats(mgr, i, bm, b1, b2);
        try { bm->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }

        double s = bm->size(whl) + (q->elts[ whl ] - b1->size(whl));
        clip_param_val(&s, params->block_ctx()->size_c(whl));
        try { bm->size_shift_deps(whl, s); }
        catch (jwscxx::base::Arg_error e) { }

        delete b1;
        delete b2;

        bm->posn_v(&p2);
        p1->elts[ whl ] = p2->elts[ whl ];
        subtract_vectors_d(&p1, p1, p2);
        prop_obj->add_posn_shift_deps_v(p1);

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(p1);
    free_vector_d(p2);
    free_vector_d(m);
    free_vector_d(q);
}

void Metropolis::Block_from_block_merge::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b1_i = merge_1_i();
            Block* b1 = dynamic_cast<Block*>(cat->obj->component(b1_i));
            double b1_q_lp = b1->calc_var_param_prior_log_prob();

            uint32_t b2_i = merge_2_i();
            Block* b2 = dynamic_cast<Block*>(cat->obj->component(b2_i));
            double b2_q_lp = b2->calc_var_param_prior_log_prob();

            uint32_t bm_i = num_obj_components() - 2;
            Block* bm = dynamic_cast<Block*>(cat->prop_obj->component(bm_i));
            double bm_prop_q_lp = bm->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b1_q_lp;
            cat_la += b2_q_lp;
            cat_la -= bm_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/* ------------------  Metropolis::Pair_from_block_merge  ------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Pair_from_block_merge::Pair_from_block_merge
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Merge(name, p, pair_p)
{
}

Metropolis::Pair_from_block_merge* 
Metropolis::Pair_from_block_merge::clone() const
{
    return new Pair_from_block_merge(*this);
}

bool Metropolis::Pair_from_block_merge::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Pair::mergeable_pair_from_block(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Pair_from_block_merge::init(Metropolis::Thread_manager* mgr)
{
    using jwscxx::base::Arg_error;

    Merge::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _merge_1_i = Pair::mergeable_pair_from_block(
                cat->obj->components()); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Pair* b1 = dynamic_cast<Pair*>(cat->obj->component(_merge_1_i));

        try { _merge_pt_1_i = Pair::mergeable_pair_from_block_attach_pt(b1); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Block* b2 = dynamic_cast<Block*>(b1->attachment(_merge_pt_1_i));

        try { _merge_2_i = cat->obj->component(b2); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        try { _merge_pt_2_i = b2->attachment(b1); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        _c_posn_i = _merge_1_i;
        if (cat->obj->num_components() > 2)
        {
            while (_c_posn_i == _merge_1_i || _c_posn_i == _merge_2_i)
            {
                _c_posn_i = cat->obj->select_random_component();
            }
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Pair_from_block_merge::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b1_i = merge_1_i();
    uint32_t b2_i = merge_2_i();
    uint32_t b1_j = merge_pt_1_i();
    uint32_t b2_j = merge_pt_2_i();
    uint32_t whl;

    Vector_d* q  = 0;
    Vector_d* m  = 0;
    Vector_d* p1 = 0;
    Vector_d* p2 = 0;
    double    sz[3];

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Pair* b1 = dynamic_cast<Pair*>(prop_obj->component(b1_i));
        Block* b2 = dynamic_cast<Block*>(prop_obj->component(b2_i));
        Block* bm = 0;

        for (whl = 0; whl < 3; whl++)
        {
            sz[ whl ] = (whl == (b1_j/2))
                ? b1->size(whl) + b2->size(whl) : b2->size(whl);
        }
        whl = b1_j/2;

        // Try to create a block of size sz. If it fails, resize b1 and b2 to
        // something that should work and try again. No, Virginia, there is no
        // Santa Claus, only ugly hacks.
        try { bm = new Block(sz[0], sz[1], sz[2], params->block_ctx()); }
        catch (jwscxx::base::Arg_error e) 
        { 
            sz[ whl ] = 0.5*b1->size(whl);
            try { b1->test_size_shift_deps(whl, sz[ whl ]); }
            catch (jwscxx::base::Arg_error e) 
            { sz[ whl ] = 0.5*b2->size(whl); }
            b1->size_shift_deps(whl, sz[ whl ]);
            b2->size_shift_deps(whl, sz[ whl ]);
            sz[ whl ] = b1->size(whl) + b2->size(whl);
            bm = new Block(sz[0], sz[1], sz[2], params->block_ctx());
        }

        b1->get_var_param_values(&m);
        b1->sample_var_param_split_merge_proposal(&q, m, b1_j);
        double b_q_lp = b1->calc_var_param_split_merge_proposal_log_prob(
                q, m, b1_j);

        prop_obj->merge(bm, b1, b1_j, b2, b2_j);
        bm->posn_v(&p1);

        sample_var_param_stats(mgr, i, bm, b1, b2);
        try { bm->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }

        double s = bm->size(whl) + (q->elts[ whl ] - b1->size(whl));
        clip_param_val(&s, params->block_ctx()->size_c(whl));
        try { bm->size_shift_deps(whl, s); }
        catch (jwscxx::base::Arg_error e) { }

        delete b1;
        delete b2;

        bm->posn_v(&p2);
        subtract_vectors_d(&p1, p1, p2);
        prop_obj->add_posn_shift_deps_v(p1);

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(q);
    free_vector_d(m);
    free_vector_d(p1);
    free_vector_d(p2);
}

void Metropolis::Pair_from_block_merge::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    typedef Abstract_block Ablock;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b1_i = merge_1_i();
            Ablock* b1 = dynamic_cast<Ablock*>(cat->obj->component(b1_i));
            double b1_q_lp = b1->calc_var_param_prior_log_prob();

            uint32_t b2_i = merge_2_i();
            Ablock* b2 = dynamic_cast<Ablock*>(cat->obj->component(b2_i));
            double b2_q_lp = b2->calc_var_param_prior_log_prob();

            uint32_t bm_i = num_obj_components() - 2;
            Ablock* bm = dynamic_cast<Ablock*>(cat->prop_obj->component(bm_i));
            double bm_prop_q_lp = bm->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b1_q_lp;
            cat_la += b2_q_lp;
            cat_la -= bm_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/* -----------------  Metropolis::Block_from_pair_merge  -------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 * @param  height_only  Only allow height attachments to a block set.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_from_pair_merge::Block_from_pair_merge
(
    const char* name, 
    double      p,
    double      pair_p,
    bool        height_only
) 
throw (jwscxx::base::Arg_error) 
    : Merge(name, p, pair_p)
{
    _height_only = height_only;
}

Metropolis::Block_from_pair_merge* 
Metropolis::Block_from_pair_merge::clone() const
{
    return new Block_from_pair_merge(*this);
}

bool Metropolis::Block_from_pair_merge::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Pair::mergeable_block_from_pair(components, _height_only); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Block_from_pair_merge::init(Metropolis::Thread_manager* mgr)
{
    using jwscxx::base::Arg_error;

    Merge::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _merge_1_i = Pair::mergeable_block_from_pair(
                cat->obj->components(), _height_only); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Block* b1 = dynamic_cast<Block*>(cat->obj->component(_merge_1_i));

        try { _merge_pt_1_i = Pair::mergeable_block_from_pair_attach_pt(b1, 
                _height_only); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Pair* b2 = dynamic_cast<Pair*>(b1->attachment(
                    _merge_pt_1_i));

        try { _merge_2_i = cat->obj->component(b2); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        try { _merge_pt_2_i = b2->attachment(b1); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        _c_posn_i = _merge_1_i;
        if (cat->obj->num_components() > 2)
        {
            while (_c_posn_i == _merge_1_i || _c_posn_i == _merge_2_i)
            {
                _c_posn_i = cat->obj->select_random_component();
            }
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Block_from_pair_merge::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b1_i = merge_1_i();
    uint32_t b2_i = merge_2_i();
    uint32_t b1_j = merge_pt_1_i();
    uint32_t b2_j = merge_pt_2_i();
    uint32_t whl;

    Vector_d* q  = 0;
    Vector_d* m  = 0;
    Vector_d* p1 = 0;
    Vector_d* p2 = 0;
    double    sz[4];

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Block* b1 = dynamic_cast<Block*>(prop_obj->component(b1_i));
        Pair* b2 = dynamic_cast<Pair*>(prop_obj->component(b2_i));
        Pair* bm = 0;

        for (whl = 0; whl < 3; whl++)
        {
            sz[ whl ] = (whl == (b1_j/2))
                ? b1->size(whl) + b2->size(whl) : b2->size(whl);
        }
        sz[3] = b2->inner_size(0);
        whl = b1_j/2;

        // Try to create a block of size sz. If it fails, resize b1 and b2 to
        // something that should work and try again. No, Virginia, there is no
        // Santa Claus, only ugly hacks.
        try { bm = new Pair(sz[0], sz[1], sz[2], sz[3], b2->layout(),
                params->block_set_ctx()); }
        catch (jwscxx::base::Arg_error e) 
        { 
            sz[ whl ] = 0.5*b1->size(whl);
            try { b1->test_size_shift_deps(whl, sz[ whl ]); }
            catch (jwscxx::base::Arg_error e) 
            { sz[ whl ] = 0.5*b2->size(whl); }
            b1->size_shift_deps(whl, sz[ whl ]);
            b2->size_shift_deps(whl, sz[ whl ]);
            sz[ whl ] = b1->size(whl) + b2->size(whl);
            bm = new Pair(sz[0], sz[1], sz[2], sz[3], b2->layout(),
                    params->block_set_ctx());
        }

        b1->get_var_param_values(&m);
        b1->sample_var_param_split_merge_proposal(&q, m, b1_j);
        double b_q_lp = b1->calc_var_param_split_merge_proposal_log_prob(
                q, m, b1_j);

        prop_obj->merge(bm, b1, b1_j, b2, b2_j);
        bm->posn_v(&p1);

        sample_var_param_stats(mgr, i, bm, b1, b2);
        try { bm->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }

        double s = bm->size(whl) + (q->elts[ whl ] - b1->size(whl));
        clip_param_val(&s, params->block_ctx()->size_c(whl));
        try { bm->size_shift_deps(whl, s); }
        catch (jwscxx::base::Arg_error e) { }

        delete b1;
        delete b2;

        bm->posn_v(&p2);
        subtract_vectors_d(&p1, p1, p2);
        prop_obj->add_posn_shift_deps_v(p1);

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(q);
    free_vector_d(m);
    free_vector_d(p1);
    free_vector_d(p2);
}

void Metropolis::Block_from_pair_merge::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    typedef Abstract_block Ablock;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b1_i = merge_1_i();
            Ablock* b1 = dynamic_cast<Ablock*>(cat->obj->component(b1_i));
            double b1_q_lp = b1->calc_var_param_prior_log_prob();

            uint32_t b2_i = merge_2_i();
            Ablock* b2 = dynamic_cast<Ablock*>(cat->obj->component(b2_i));
            double b2_q_lp = b2->calc_var_param_prior_log_prob();

            uint32_t bm_i = num_obj_components() - 2;
            Ablock* bm = dynamic_cast<Ablock*>(cat->prop_obj->component(bm_i));
            double bm_prop_q_lp = bm->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b1_q_lp;
            cat_la += b2_q_lp;
            cat_la -= bm_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/* ----------------  Metropolis::Quartet_from_block_merge  ------------------ */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Quartet_from_block_merge::Quartet_from_block_merge
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Merge(name, p, pair_p)
{
}

Metropolis::Quartet_from_block_merge* 
Metropolis::Quartet_from_block_merge::clone() const
{
    return new Quartet_from_block_merge(*this);
}

bool Metropolis::Quartet_from_block_merge::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Quartet::mergeable_quartet_from_block(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Quartet_from_block_merge::init(Metropolis::Thread_manager* mgr)
{
    using jwscxx::base::Arg_error;

    Merge::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _merge_1_i = Quartet::mergeable_quartet_from_block(
                cat->obj->components()); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Quartet* b1 = dynamic_cast<Quartet*>(cat->obj->component(
                    _merge_1_i));

        try { _merge_pt_1_i = 
            Quartet::mergeable_quartet_from_block_attach_pt(b1); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Block* b2 = dynamic_cast<Block*>(b1->attachment(_merge_pt_1_i));

        try { _merge_2_i = cat->obj->component(b2); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        try { _merge_pt_2_i = b2->attachment(b1); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        _c_posn_i = _merge_1_i;
        if (cat->obj->num_components() > 2)
        {
            while (_c_posn_i == _merge_1_i || _c_posn_i == _merge_2_i)
            {
                _c_posn_i = cat->obj->select_random_component();
            }
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Quartet_from_block_merge::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b1_i = merge_1_i();
    uint32_t b2_i = merge_2_i();
    uint32_t b1_j = merge_pt_1_i();
    uint32_t b2_j = merge_pt_2_i();
    uint32_t whl;

    Vector_d* q  = 0;
    Vector_d* m  = 0;
    Vector_d* p1 = 0;
    Vector_d* p2 = 0;
    double    sz[3];

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Quartet* b1 = dynamic_cast<Quartet*>(prop_obj->component(b1_i));
        Block* b2 = dynamic_cast<Block*>(prop_obj->component(b2_i));
        Block* bm = 0;

        for (whl = 0; whl < 3; whl++)
        {
            sz[ whl ] = (whl == (b1_j/2))
                ? b1->size(whl) + b2->size(whl) : b2->size(whl);
        }
        whl = b1_j/2;

        // Try to create a block of size sz. If it fails, resize b1 and b2 to
        // something that should work and try again. No, Virginia, there is no
        // Santa Claus, only ugly hacks.
        try { bm = new Block(sz[0], sz[1], sz[2], params->block_ctx()); }
        catch (jwscxx::base::Arg_error e) 
        { 
            sz[ whl ] = 0.5*b1->size(whl);
            try { b1->test_size_shift_deps(whl, sz[ whl ]); }
            catch (jwscxx::base::Arg_error e) 
            { sz[ whl ] = 0.5*b2->size(whl); }
            b1->size_shift_deps(whl, sz[ whl ]);
            b2->size_shift_deps(whl, sz[ whl ]);
            sz[ whl ] = b1->size(whl) + b2->size(whl);
            bm = new Block(sz[0], sz[1], sz[2], params->block_ctx());
        }

        b1->get_var_param_values(&m);
        b1->sample_var_param_split_merge_proposal(&q, m, b1_j);
        double b_q_lp = b1->calc_var_param_split_merge_proposal_log_prob(
                q, m, b1_j);

        prop_obj->merge(bm, b1, b1_j, b2, b2_j);
        bm->posn_v(&p1);

        sample_var_param_stats(mgr, i, bm, b1, b2);
        try { bm->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }

        double s = bm->size(whl) + (q->elts[ whl ] - b1->size(whl));
        clip_param_val(&s, params->block_ctx()->size_c(whl));
        try { bm->size_shift_deps(whl, s); }
        catch (jwscxx::base::Arg_error e) { }

        delete b1;
        delete b2;

        bm->posn_v(&p2);
        subtract_vectors_d(&p1, p1, p2);
        prop_obj->add_posn_shift_deps_v(p1);

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(q);
    free_vector_d(m);
    free_vector_d(p1);
    free_vector_d(p2);
}

void Metropolis::Quartet_from_block_merge::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    typedef Abstract_block Ablock;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b1_i = merge_1_i();
            Ablock* b1 = dynamic_cast<Ablock*>(cat->obj->component(b1_i));
            double b1_q_lp = b1->calc_var_param_prior_log_prob();

            uint32_t b2_i = merge_2_i();
            Ablock* b2 = dynamic_cast<Ablock*>(cat->obj->component(b2_i));
            double b2_q_lp = b2->calc_var_param_prior_log_prob();

            uint32_t bm_i = num_obj_components() - 2;
            Ablock* bm = dynamic_cast<Ablock*>(cat->prop_obj->component(bm_i));
            double bm_prop_q_lp = bm->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b1_q_lp;
            cat_la += b2_q_lp;
            cat_la -= bm_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */







/* ----------------  Metropolis::Block_from_quartet_merge  ------------------ */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 * @param  height_only  Only allow height attachments to a block set.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_from_quartet_merge::Block_from_quartet_merge
(
    const char* name, 
    double      p,
    double      pair_p,
    bool        height_only
) 
throw (jwscxx::base::Arg_error) 
    : Merge(name, p, pair_p)
{
    _height_only = height_only;
}

Metropolis::Block_from_quartet_merge* 
Metropolis::Block_from_quartet_merge::clone() const
{
    return new Block_from_quartet_merge(*this);
}

bool Metropolis::Block_from_quartet_merge::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Quartet::mergeable_block_from_quartet(components, _height_only); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Block_from_quartet_merge::init(Metropolis::Thread_manager* mgr)
{
    using jwscxx::base::Arg_error;

    Merge::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _merge_1_i = Quartet::mergeable_block_from_quartet(
                cat->obj->components(), _height_only); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Block* b1 = dynamic_cast<Block*>(cat->obj->component(_merge_1_i));

        try { _merge_pt_1_i = 
            Quartet::mergeable_block_from_quartet_attach_pt(b1, _height_only); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        const Quartet* b2 = dynamic_cast<Quartet*>(b1->attachment(
                    _merge_pt_1_i));

        try { _merge_2_i = cat->obj->component(b2); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        try { _merge_pt_2_i = b2->attachment(b1); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        _c_posn_i = _merge_1_i;
        if (cat->obj->num_components() > 2)
        {
            while (_c_posn_i == _merge_1_i || _c_posn_i == _merge_2_i)
            {
                _c_posn_i = cat->obj->select_random_component();
            }
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Block_from_quartet_merge::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t b1_i = merge_1_i();
    uint32_t b2_i = merge_2_i();
    uint32_t b1_j = merge_pt_1_i();
    uint32_t b2_j = merge_pt_2_i();
    uint32_t whl;

    Vector_d* q  = 0;
    Vector_d* m  = 0;
    Vector_d* p1 = 0;
    Vector_d* p2 = 0;
    double    sz[5];

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Block* b1 = dynamic_cast<Block*>(prop_obj->component(b1_i));
        Quartet* b2 = dynamic_cast<Quartet*>(prop_obj->component(b2_i));
        Quartet* bm = 0;

        for (whl = 0; whl < 3; whl++)
        {
            sz[ whl ] = (whl == (b1_j/2))
                ? b1->size(whl) + b2->size(whl) : b2->size(whl);
        }
        sz[3] = b2->inner_size(0);
        sz[4] = b2->inner_size(1);
        whl = b1_j/2;

        // Try to create a block of size sz. If it fails, resize b1 and b2 to
        // something that should work and try again. No, Virginia, there is no
        // Santa Claus, only ugly hacks.
        try { bm = new Quartet(sz[0], sz[1], sz[2], sz[3], sz[4],
                    params->block_set_ctx()); }
        catch (jwscxx::base::Arg_error e) 
        { 
            sz[ whl ] = 0.5*b1->size(whl);
            try { b1->test_size_shift_deps(whl, sz[ whl ]); }
            catch (jwscxx::base::Arg_error e) 
            { sz[ whl ] = 0.5*b2->size(whl); }
            b1->size_shift_deps(whl, sz[ whl ]);
            b2->size_shift_deps(whl, sz[ whl ]);
            sz[ whl ] = b1->size(whl) + b2->size(whl);
            bm = new Quartet(sz[0], sz[1], sz[2], sz[3], sz[4],
                    params->block_set_ctx()); 
        }

        b1->get_var_param_values(&m);
        b1->sample_var_param_split_merge_proposal(&q, m, b1_j);
        double b_q_lp = b1->calc_var_param_split_merge_proposal_log_prob(
                q, m, b1_j);

        prop_obj->merge(bm, b1, b1_j, b2, b2_j);
        bm->posn_v(&p1);

        sample_var_param_stats(mgr, i, bm, b1, b2);
        try { bm->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }

        double s = bm->size(whl) + (q->elts[ whl ] - b1->size(whl));
        clip_param_val(&s, params->block_ctx()->size_c(whl));
        try { bm->size_shift_deps(whl, s); }
        catch (jwscxx::base::Arg_error e) { }

        delete b1;
        delete b2;

        bm->posn_v(&p2);
        subtract_vectors_d(&p1, p1, p2);
        prop_obj->add_posn_shift_deps_v(p1);

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la += b_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(q);
    free_vector_d(m);
    free_vector_d(p1);
    free_vector_d(p2);
}

void Metropolis::Block_from_quartet_merge::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    typedef Abstract_block Ablock;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            Category* cat = params->category(c);

            uint32_t b1_i = merge_1_i();
            Ablock* b1 = dynamic_cast<Ablock*>(cat->obj->component(b1_i));
            double b1_q_lp = b1->calc_var_param_prior_log_prob();

            uint32_t b2_i = merge_2_i();
            Ablock* b2 = dynamic_cast<Ablock*>(cat->obj->component(b2_i));
            double b2_q_lp = b2->calc_var_param_prior_log_prob();

            uint32_t bm_i = num_obj_components() - 2;
            Ablock* bm = dynamic_cast<Ablock*>(cat->prop_obj->component(bm_i));
            double bm_prop_q_lp = bm->calc_var_param_prior_log_prob();

            params->update_cat_proposal_dependents(c);

            double cat_la = 0;
            cat_la += cat->prop_obj->calc_var_param_prior_log_prob();
            cat_la -= cat->obj->calc_var_param_prior_log_prob();
            cat_la += b1_q_lp;
            cat_la += b2_q_lp;
            cat_la -= bm_prop_q_lp;

#if defined FURNITURE_HAVE_ISINF && defined FURNITURE_HAVE_ISNAN
            assert(!std::isinf(cat_la) && !std::isnan(cat_la));
#endif

            cat_log_alpha(cat_la);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

/* ========================================================================== */








/* --------------------------  Metropolis::Switch  -------------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Switch::Switch(const char* name, double p, double pair_p) 
    throw (jwscxx::base::Arg_error) : Jump(name, p, pair_p)
{
    _means  = 0;
    _stdevs = 0;
}

Metropolis::Switch::~Switch()
{
    jwsc::free_vector_d(_means);
    jwsc::free_vector_d(_stdevs);
}

void Metropolis::Switch::init(Metropolis::Thread_manager* mgr)
{
    Jump::init(mgr);
}

void Metropolis::Switch::update_cat_log_alpha
(
    Metropolis::Thread_manager* mgr
)
{
    using namespace jwsc;
    typedef Abstract_block Ablock;

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        if (!reject())
        {
            params->update_cat_proposal_dependents(c);
            cat_log_alpha(0);
        }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

uint32_t Metropolis::Switch::switch_i()
{
    return _switch_i;
}

const jwsc::Vector_d* Metropolis::Switch::means() const
{
    return _means;
}

const jwsc::Vector_d* Metropolis::Switch::stdevs() const
{
    return _stdevs;
}

void Metropolis::Switch::sample_var_param_stats
(
    Metropolis::Thread_manager* mgr,
    uint32_t                    i,
    const Abstract_block*       b1,
    const Abstract_block*       b2
)
{
    using namespace jwsc;

    if (mgr->thread() < i)
        return;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        Vector_d* b1_m = 0;
        Vector_d* b1_s = 0;

        b1->get_var_param_stats(&b1_m, &b1_s);
        b2->sample_var_param_prior(&_means, &_stdevs);

        for (uint32_t whl = 0; whl < 3; whl++)
        {
            _means->elts[ whl ] = b1_m->elts[ whl ];
            _stdevs->elts[ whl ] = b1_s->elts[ whl ];
        }

        free_vector_d(b1_m);
        free_vector_d(b1_s);

        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    mgr->desynchronize();
}

/* ========================================================================== */







/* ----------------  Metropolis::Pair_to_block_switch  ---------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Pair_to_block_switch::Pair_to_block_switch
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Switch(name, p, pair_p)
{
}

Metropolis::Pair_to_block_switch* 
Metropolis::Pair_to_block_switch::clone() const
{
    return new Pair_to_block_switch(*this);
}

bool Metropolis::Pair_to_block_switch::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Pair::switchable_pair_to_block(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Pair_to_block_switch::init
(
    Metropolis::Thread_manager* mgr
)
{
    Switch::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _switch_i = Pair::switchable_pair_to_block(
                cat->obj->components()); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Pair_to_block_switch::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t  bq_i = switch_i();
    Vector_d* bq_v = 0;
    Vector_d* bb_v = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Pair* bq = dynamic_cast<Pair*>(prop_obj->component(bq_i));
        Block* bb = new Block(params->block_ctx());

        sample_var_param_stats(mgr, i, bq, bb);
        try { bb->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }

        bq->get_var_param_values(&bq_v);
        bb->sample_var_param_switch_proposal(&bb_v, bq_v);
        try { bb->set_var_param_values(bb_v); }
        catch (jwscxx::base::Arg_error e) { abort(); }
        double bb_prop_q_lp = bb->calc_var_param_switch_proposal_log_prob(bb_v);

        bq->remove_cross_blocks(bq->num_cross_blocks());
        prop_obj->switch_block(bq, bb);
        delete bq;

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la -= bb_prop_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(bq_v);
    free_vector_d(bb_v);
}

/* ========================================================================== */







/* ----------------  Metropolis::Block_to_pair_switch  ---------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 * @param  height_only  Only allow height attachments to a block set.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_to_pair_switch::Block_to_pair_switch
(
    const char* name, 
    double      p,
    double      pair_p,
    bool        height_only
) 
throw (jwscxx::base::Arg_error) 
    : Switch(name, p, pair_p)
{
    _height_only = height_only;
}

Metropolis::Block_to_pair_switch* 
Metropolis::Block_to_pair_switch::clone() const
{
    return new Block_to_pair_switch(*this);
}

bool Metropolis::Block_to_pair_switch::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Pair::switchable_block_to_pair(components, _height_only); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

Pair::Layout Metropolis::Block_to_pair_switch::layout()
{
    return _layout;
}

void Metropolis::Block_to_pair_switch::init
(
    Metropolis::Thread_manager* mgr
)
{
    Switch::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _switch_i = Pair::switchable_block_to_pair(
                cat->obj->components(), _height_only); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        _layout = Pair::select_random_layout();

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Block_to_pair_switch::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t  bb_i = switch_i();
    Vector_d* bb_v = 0;
    Vector_d* bq_v = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Block* bb = dynamic_cast<Block*>(prop_obj->component(bb_i));
        Pair* bq = new Pair(layout(), params->block_set_ctx());

        sample_var_param_stats(mgr, i, bb, bq);
        try { bq->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }

        bb->get_var_param_values(&bb_v);
        bq->Block_set::sample_var_param_switch_proposal(&bq_v, bb_v);
        try { bq->set_var_param_values(bq_v); }
        catch (jwscxx::base::Arg_error e) { abort(); }
        double bq_prop_q_lp = bq->calc_var_param_switch_proposal_log_prob(bq_v);

        prop_obj->switch_block(bb, bq);
        delete bb;

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la -= bq_prop_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(bb_v);
    free_vector_d(bq_v);
}

/* ========================================================================== */







/* ----------------  Metropolis::Quartet_to_block_switch  ------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Quartet_to_block_switch::Quartet_to_block_switch
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Switch(name, p, pair_p)
{
}

Metropolis::Quartet_to_block_switch* 
Metropolis::Quartet_to_block_switch::clone() const
{
    return new Quartet_to_block_switch(*this);
}

bool Metropolis::Quartet_to_block_switch::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Quartet::switchable_quartet_to_block(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Quartet_to_block_switch::init
(
    Metropolis::Thread_manager* mgr
)
{
    Switch::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _switch_i = Quartet::switchable_quartet_to_block(
                cat->obj->components()); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Quartet_to_block_switch::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t  bq_i = switch_i();
    Vector_d* bq_v = 0;
    Vector_d* bb_v = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Quartet* bq = dynamic_cast<Quartet*>(prop_obj->component(bq_i));
        Block* bb = new Block(params->block_ctx());

        sample_var_param_stats(mgr, i, bq, bb);
        try { bb->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }

        bq->get_var_param_values(&bq_v);
        bb->sample_var_param_switch_proposal(&bb_v, bq_v);
        try { bb->set_var_param_values(bb_v); }
        catch (jwscxx::base::Arg_error e) { abort(); }
        double bb_prop_q_lp = bb->calc_var_param_switch_proposal_log_prob(bb_v);

        bq->remove_cross_blocks(bq->num_cross_blocks());
        prop_obj->switch_block(bq, bb);
        delete bq;

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la -= bb_prop_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(bq_v);
    free_vector_d(bb_v);
}

/* ========================================================================== */







/* ----------------  Metropolis::Block_to_quartet_switch  ------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 * @param  height_only  Only allow height attachments to a block set.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_to_quartet_switch::Block_to_quartet_switch
(
    const char* name, 
    double      p,
    double      pair_p,
    bool        height_only
) 
throw (jwscxx::base::Arg_error) 
    : Switch(name, p, pair_p)
{
    _height_only = height_only;
}

Metropolis::Block_to_quartet_switch* 
Metropolis::Block_to_quartet_switch::clone() const
{
    return new Block_to_quartet_switch(*this);
}

bool Metropolis::Block_to_quartet_switch::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Quartet::switchable_block_to_quartet(components, _height_only); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Block_to_quartet_switch::init
(
    Metropolis::Thread_manager* mgr
)
{
    Switch::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _switch_i = Quartet::switchable_block_to_quartet(
                cat->obj->components(), _height_only); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Block_to_quartet_switch::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t  bb_i = switch_i();
    Vector_d* bb_v = 0;
    Vector_d* bq_v = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Block* bb = dynamic_cast<Block*>(prop_obj->component(bb_i));
        Quartet* bq = new Quartet(params->block_set_ctx());

        sample_var_param_stats(mgr, i, bb, bq);
        try { bq->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }

        bb->get_var_param_values(&bb_v);
        bq->Block_set::sample_var_param_switch_proposal(&bq_v, bb_v);
        try { bq->set_var_param_values(bq_v); }
        catch (jwscxx::base::Arg_error e) { abort(); }
        double bq_prop_q_lp = bq->calc_var_param_switch_proposal_log_prob(bq_v);

        prop_obj->switch_block(bb, bq);
        delete bb;

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la -= bq_prop_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(bb_v);
    free_vector_d(bq_v);
}

/* ========================================================================== */







/* ----------------  Metropolis::Quartet_to_pair_switch  -------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Quartet_to_pair_switch::Quartet_to_pair_switch
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Switch(name, p, pair_p)
{
}

Metropolis::Quartet_to_pair_switch* 
Metropolis::Quartet_to_pair_switch::clone() const
{
    return new Quartet_to_pair_switch(*this);
}

bool Metropolis::Quartet_to_pair_switch::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Quartet::switchable_quartet_to_pair(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

Pair::Layout Metropolis::Quartet_to_pair_switch::layout()
{
    return _layout;
}

void Metropolis::Quartet_to_pair_switch::init
(
    Metropolis::Thread_manager* mgr
)
{
    Switch::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _switch_i = Quartet::switchable_quartet_to_pair(
                cat->obj->components()); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        _layout = Pair::select_random_layout();

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Quartet_to_pair_switch::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t  bq_i = switch_i();
    Vector_d* bp_v = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Quartet* bq = dynamic_cast<Quartet*>(prop_obj->component(bq_i));
        Pair* bp = new Pair(bq->size(0), bq->size(1), bq->size(2),
                bq->inner_size(0), bq->cross(0), bq->cross(1), bq->cross(2),
                bq->cross(3), layout(), params->block_set_ctx());

        sample_var_param_stats(mgr, i, bq, bp);
        try { bp->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }

        bp->sample_var_param_switch_proposal(&bp_v, bq);
        try { bp->set_var_param_values(bp_v); }
        catch (jwscxx::base::Arg_error e) { abort(); }
        double bp_prop_q_lp = bp->calc_var_param_switch_proposal_log_prob(bp_v);

        prop_obj->switch_block(bq, bp);
        delete bq;

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la -= bp_prop_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(bp_v);
}

void Metropolis::Quartet_to_pair_switch::sample_var_param_stats
(
    Metropolis::Thread_manager* mgr,
    uint32_t                    i,
    const Quartet*              b1,
    const Pair*                 b2
)
{
    using namespace jwsc;

    Switch::sample_var_param_stats(mgr, i, b1, b2);

    if (mgr->thread() < i)
        return;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        uint32_t d1 = b1->var_param_inner_size_i(0);
        uint32_t d2 = b2->var_param_inner_size_i(0);

        Vector_d* b1_m = 0;
        Vector_d* b1_s = 0;
        b1->get_var_param_stats(&b1_m, &b1_s);

        assert(b1_m->num_elts > d1+1);
        assert(b1_s->num_elts > d1+1);
        assert(_means->num_elts > d2);
        assert(_stdevs->num_elts > d2);

        if (layout() == Pair::WIDTH)
        {
            _means->elts[ d2 ] = b1_m->elts[ d1+0 ];
            _stdevs->elts[ d2 ] = b1_s->elts[ d1+0 ];
        }
        else
        {
            _means->elts[ d2 ] = b1_m->elts[ d1+1 ];
            _stdevs->elts[ d2 ] = b1_s->elts[ d1+1 ];
        }

        uint32_t N = b1->num_var_param_cross_p();
        if (N > 0)
        {
            d1 = b1->var_param_cross_i(0);
            d2 = b2->var_param_cross_i(0);

            assert(b1_m->num_elts >= d1+N);
            assert(_means->num_elts >= d2+N);

            for (uint32_t n = 0; n < N; n++)
            {
                _means->elts[ d2+n ] = b1_m->elts[ d1+n ];
                _stdevs->elts[ d2+n ] = b1_s->elts[ d1+n ];
            }
        }

        free_vector_d(b1_m);
        free_vector_d(b1_s);

        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    mgr->desynchronize();
}

/* ========================================================================== */







/* -----------------  Metropolis::Pair_to_quartet_switch  ------------------- */

/**
 * @param  name    Name of the balanced move.
 * @param  p       Probability of the balanced move.
 * @param  pair_p  Probability of the paired, balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Pair_to_quartet_switch::Pair_to_quartet_switch
(
    const char* name, 
    double      p,
    double      pair_p
) 
throw (jwscxx::base::Arg_error) 
    : Switch(name, p, pair_p)
{
}

Metropolis::Pair_to_quartet_switch* 
Metropolis::Pair_to_quartet_switch::clone() const
{
    return new Pair_to_quartet_switch(*this);
}

bool Metropolis::Pair_to_quartet_switch::has_supported_components
(
    const std::vector<Component*>& components
) 
const
{
    try { Pair::switchable_pair_to_quartet(components); }
    catch (jwscxx::base::Arg_error e) { return false; }
    return true;
}

void Metropolis::Pair_to_quartet_switch::init
(
    Metropolis::Thread_manager* mgr
)
{
    Switch::init(mgr);

    mgr->synchronize();

    if (mgr->last_thread())
    {
        Parameters* params = mgr->params();
        uint32_t c = mgr->category();

        Category* cat = params->category(c);
        params->init_cat_proposal(c);

        try { _switch_i = Pair::switchable_pair_to_quartet(
                cat->obj->components()); }
        catch (jwscxx::base::Arg_error e) { abort(); }

        mgr->release();
    }
    else
    {
        mgr->wait();
    }

    mgr->desynchronize();
}

void Metropolis::Pair_to_quartet_switch::run_thread
(
    Metropolis::Thread_manager* mgr
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    uint32_t c = mgr->category();

    Parameters* params = mgr->params();
    Instance* ins = params->instance(c, mgr->thread());

    uint32_t t   = mgr->thread();
    uint32_t N_t = mgr->num_threads();
    uint32_t N_i = params->num_instances();

    uint32_t  bp_i = switch_i();
    Vector_d* bq_v = 0;

    for (uint32_t i = t; i < N_i; i += N_t)
    {
        ins = params->instance(c, i);
        params->init_ins_proposal(c, i);

        Object* prop_obj = ins->prop_obj;
        Pair* bp = dynamic_cast<Pair*>(prop_obj->component(bp_i));
        Quartet* bq = new Quartet(bp->size(0), bp->size(1), bp->size(2),
                bp->inner_size(0), bp->inner_size(0), bp->cross(0),
                bp->cross(1), bp->cross(2), bp->cross(3),
                params->block_set_ctx());

        sample_var_param_stats(mgr, i, bp, bq);
        try { bq->set_var_param_stats(means(), stdevs()); }
        catch (jwscxx::base::Arg_error e) { }

        bq->sample_var_param_switch_proposal(&bq_v, bp);
        try { bq->set_var_param_values(bq_v); }
        catch (jwscxx::base::Arg_error e) { abort(); }
        double bq_prop_q_lp = bq->calc_var_param_switch_proposal_log_prob(bq_v);

        prop_obj->switch_block(bp, bq);
        delete bp;

        params->update_ins_proposal_dependents(c, i);

        double la = 0;
        la += ins->prop_img_ll;
        la -= ins->img_ll;
        la += ins->prop_obj->calc_var_param_log_likelihood();
        la -= ins->obj->calc_var_param_log_likelihood();
        la -= bq_prop_q_lp;

        ins_log_alpha(la, i);
    }

    free_vector_d(bq_v);
}

void Metropolis::Pair_to_quartet_switch::sample_var_param_stats
(
    Metropolis::Thread_manager* mgr,
    uint32_t                    i,
    const Pair*                 b1,
    const Quartet*              b2
)
{
    using namespace jwsc;

    Switch::sample_var_param_stats(mgr, i, b1, b2);

    if (mgr->thread() < i)
        return;

    mgr->synchronize();
    if (mgr->last_thread())
    {
        uint32_t d1 = b1->var_param_inner_size_i(0);
        uint32_t d2 = b2->var_param_inner_size_i(0);

        Vector_d* b1_m = 0;
        Vector_d* b1_s = 0;
        b1->get_var_param_stats(&b1_m, &b1_s);

        assert(b1_m->num_elts > d1);
        assert(b1_s->num_elts > d1);
        assert(_means->num_elts > d2+1);
        assert(_stdevs->num_elts > d2+1);

        if (b1->layout() == Pair::WIDTH)
        {
            _means->elts[ d2+0 ] = b1_m->elts[ d1 ];
            _stdevs->elts[ d2+0 ] = b1_s->elts[ d1 ];
        }
        else
        {
            _means->elts[ d2+1 ] = b1_m->elts[ d1 ];
            _stdevs->elts[ d2+1 ] = b1_s->elts[ d1 ];
        }

        uint32_t N = b1->num_var_param_cross_p();
        if (N > 0)
        {
            d1 = b1->var_param_cross_i(0);
            d2 = b2->var_param_cross_i(0);

            assert(b1_m->num_elts >= d1+N);
            assert(_means->num_elts >= d2+N);

            for (uint32_t n = 0; n < N; n++)
            {
                _means->elts[ d2+n ] = b1_m->elts[ d1+n ];
                _stdevs->elts[ d2+n ] = b1_s->elts[ d1+n ];
            }
        }

        free_vector_d(b1_m);
        free_vector_d(b1_s);

        mgr->release();
    }
    else
    {
        mgr->wait();
    }
    mgr->desynchronize();
}

/* ========================================================================== */








/* -----------------------  Metropolis::Jump_pair  -------------------------- */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 */
Metropolis::Jump_pair::Jump_pair
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : _name_1(name_1), _name_2(name_2)
{
    _prob_1 = p_1;
    _prob_2 = p_2;
}

Metropolis::Jump_pair::~Jump_pair()
{
}

/* ========================================================================== */







/* -----------------------  Metropolis::Birth_death  ------------------------ */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Birth_death::Birth_death
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Jump_pair(name_1, name_2, p_1, p_2)
{
}

Metropolis::Birth_death::~Birth_death()
{
}

/* ========================================================================== */







/* --------------------  Metropolis::Block_birth_death  --------------------- */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_birth_death::Block_birth_death
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Birth_death(name_1, name_2, p_1, p_2), 
      _first(name_1, p_1, p_2),
      _second(name_2, p_2, p_1)
{
}

const Metropolis::Block_birth& Metropolis::Block_birth_death::first() const
{
    return _first;
}

const Metropolis::Block_death& Metropolis::Block_birth_death::second() const
{
    return _second;
}

/* ========================================================================== */







/* -----------------  Metropolis::Block_block_birth_death  ------------------ */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_block_birth_death::Block_block_birth_death
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Birth_death(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2),
      _second(name_2, p_2, p_1)
{
}

const Metropolis::Block_to_block_birth& 
Metropolis::Block_block_birth_death::first() const
{
    return _first;
}

const Metropolis::Block_from_block_death& 
Metropolis::Block_block_birth_death::second() const
{
    return _second;
}

/* ========================================================================== */







/* --------------------  Metropolis::Pair_birth_death  ---------------------- */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Pair_birth_death::Pair_birth_death
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Birth_death(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2),
      _second(name_2, p_2, p_1)
{
}

const Metropolis::Pair_birth& Metropolis::Pair_birth_death::first() const
{
    return _first;
}

const Metropolis::Pair_death& Metropolis::Pair_birth_death::second() const
{
    return _second;
}

/* ========================================================================== */







/* -----------------  Metropolis::Pair_block_birth_death  ------------------- */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Pair_block_birth_death::Pair_block_birth_death
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Birth_death(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2, false),
      _second(name_2, p_2, p_1)
{
}

const Metropolis::Pair_to_block_birth& 
Metropolis::Pair_block_birth_death::first() const
{
    return _first;
}

const Metropolis::Pair_from_block_death& 
Metropolis::Pair_block_birth_death::second() const
{
    return _second;
}

/* ========================================================================== */







/* -----------------  Metropolis::Block_pair_birth_death  ------------------- */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_pair_birth_death::Block_pair_birth_death
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Birth_death(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2, false),
      _second(name_2, p_2, p_1)
{
}

const Metropolis::Block_to_pair_birth& 
Metropolis::Block_pair_birth_death::first() const
{
    return _first;
}

const Metropolis::Block_from_pair_death& 
Metropolis::Block_pair_birth_death::second() const
{
    return _second;
}

/* ========================================================================== */







/* --------------------  Metropolis::Quartet_birth_death  ------------------- */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Quartet_birth_death::Quartet_birth_death
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Birth_death(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2),
      _second(name_2, p_2, p_1)
{
}

const Metropolis::Quartet_birth& Metropolis::Quartet_birth_death::first() const
{
    return _first;
}

const Metropolis::Quartet_death& Metropolis::Quartet_birth_death::second() const
{
    return _second;
}

/* ========================================================================== */







/* -----------------  Metropolis::Quartet_block_birth_death  ---------------- */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Quartet_block_birth_death::Quartet_block_birth_death
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Birth_death(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2, false),
      _second(name_2, p_2, p_1)
{
}

const Metropolis::Quartet_to_block_birth& 
Metropolis::Quartet_block_birth_death::first() const
{
    return _first;
}

const Metropolis::Quartet_from_block_death& 
Metropolis::Quartet_block_birth_death::second() const
{
    return _second;
}

/* ========================================================================== */







/* -----------------  Metropolis::Block_quartet_birth_death  ---------------- */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_quartet_birth_death::Block_quartet_birth_death
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Birth_death(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2, false),
      _second(name_2, p_2, p_1)
{
}

const Metropolis::Block_to_quartet_birth& 
Metropolis::Block_quartet_birth_death::first() const
{
    return _first;
}

const Metropolis::Block_from_quartet_death& 
Metropolis::Block_quartet_birth_death::second() const
{
    return _second;
}

/* ========================================================================== */







/* -------------------  Metropolis::Cross_birth_death  ---------------------- */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Cross_birth_death::Cross_birth_death
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Birth_death(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2),
      _second(name_2, p_2, p_1)
{
}

const Metropolis::Cross_birth& 
Metropolis::Cross_birth_death::first() const
{
    return _first;
}

const Metropolis::Cross_death& 
Metropolis::Cross_birth_death::second() const
{
    return _second;
}

/* ========================================================================== */







/* -----------------------  Metropolis::Split_merge  ------------------------ */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Split_merge::Split_merge
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Jump_pair(name_1, name_2, p_1, p_2)
{
}

Metropolis::Split_merge::~Split_merge()
{
}

/* ========================================================================== */







/* -----------------  Metropolis::Block_block_split_merge  ------------------ */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_block_split_merge::Block_block_split_merge
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Split_merge(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2),
      _second(name_2, p_2, p_1)
{
}

const Metropolis::Block_to_block_split& 
Metropolis::Block_block_split_merge::first() const
{
    return _first;
}

const Metropolis::Block_from_block_merge& 
Metropolis::Block_block_split_merge::second() const
{
    return _second;
}

/* ========================================================================== */







/* -----------------  Metropolis::Pair_block_split_merge  ------------------- */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Pair_block_split_merge::Pair_block_split_merge
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Split_merge(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2, false),
      _second(name_2, p_2, p_1)
{
}

const Metropolis::Pair_to_block_split& 
Metropolis::Pair_block_split_merge::first() const
{
    return _first;
}

const Metropolis::Pair_from_block_merge& 
Metropolis::Pair_block_split_merge::second() const
{
    return _second;
}

/* ========================================================================== */







/* ------------------  Metropolis::Block_pair_split_merge  ------------------ */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_pair_split_merge::Block_pair_split_merge
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Split_merge(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2),
      _second(name_2, p_2, p_1, false)
{
}

const Metropolis::Block_to_pair_split& 
Metropolis::Block_pair_split_merge::first() const
{
    return _first;
}

const Metropolis::Block_from_pair_merge& 
Metropolis::Block_pair_split_merge::second() const
{
    return _second;
}

/* ========================================================================== */







/* ---------------  Metropolis::Quartet_block_split_merge  ------------------ */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Quartet_block_split_merge::Quartet_block_split_merge
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Split_merge(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2, false),
      _second(name_2, p_2, p_1)
{
}

const Metropolis::Quartet_to_block_split& 
Metropolis::Quartet_block_split_merge::first() const
{
    return _first;
}

const Metropolis::Quartet_from_block_merge& 
Metropolis::Quartet_block_split_merge::second() const
{
    return _second;
}

/* ========================================================================== */







/* -----------------  Metropolis::Block_quartet_split_merge  ---------------- */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_quartet_split_merge::Block_quartet_split_merge
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Split_merge(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2),
      _second(name_2, p_2, p_1, false)
{
}

const Metropolis::Block_to_quartet_split& 
Metropolis::Block_quartet_split_merge::first() const
{
    return _first;
}

const Metropolis::Block_from_quartet_merge& 
Metropolis::Block_quartet_split_merge::second() const
{
    return _second;
}

/* ========================================================================== */







/* ---------------------  Metropolis::Switch_pair  -------------------------- */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Switch_pair::Switch_pair
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Jump_pair(name_1, name_2, p_1, p_2)
{
}

Metropolis::Switch_pair::~Switch_pair()
{
}

/* ========================================================================== */







/* -------------------  Metropolis::Block_pair_switch  ---------------------- */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_pair_switch::Block_pair_switch
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Switch_pair(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2, false),
      _second(name_2, p_2, p_1)
{
}

const Metropolis::Block_to_pair_switch& 
Metropolis::Block_pair_switch::first() const
{
    return _first;
}

const Metropolis::Pair_to_block_switch& 
Metropolis::Block_pair_switch::second() const
{
    return _second;
}

/* ========================================================================== */







/* -------------------  Metropolis::Block_quartet_switch  ------------------- */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Block_quartet_switch::Block_quartet_switch
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Switch_pair(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2, false),
      _second(name_2, p_2, p_1)
{
}

const Metropolis::Block_to_quartet_switch& 
Metropolis::Block_quartet_switch::first() const
{
    return _first;
}

const Metropolis::Quartet_to_block_switch& 
Metropolis::Block_quartet_switch::second() const
{
    return _second;
}

/* ========================================================================== */







/* -------------------  Metropolis::Pair_quartet_switch  -------------------- */

/**
 * @param  name_1  Name of the 1st balanced move.
 * @param  name_2  Name of the 2nd balanced move.
 * @param  p_1     Probability of the 1st balanced move.
 * @param  p_2     Probability of the 2nd balanced move.
 *
 * @throw  jwscxx::base::Arg_error  Probability must be in [0,1].
 */
Metropolis::Pair_quartet_switch::Pair_quartet_switch
(
    const char* name_1, 
    const char* name_2,
    double      p_1,
    double      p_2
) 
throw (jwscxx::base::Arg_error)
    : Switch_pair(name_1, name_2, p_1, p_2),
      _first(name_1, p_1, p_2),
      _second(name_2, p_2, p_1)
{
}

const Metropolis::Pair_to_quartet_switch& 
Metropolis::Pair_quartet_switch::first() const
{
    return _first;
}

const Metropolis::Quartet_to_pair_switch& 
Metropolis::Pair_quartet_switch::second() const
{
    return _second;
}

/* ========================================================================== */
