/**
 * This work is licensed under a Creative Commons 
 * Attribution-Noncommercial-Share Alike 3.0 United States License.
 * 
 *    http://creativecommons.org/licenses/by-nc-sa/3.0/us/
 * 
 * You are free:
 * 
 *    to Share - to copy, distribute, display, and perform the work
 *    to Remix - to make derivative works
 * 
 * Under the following conditions:
 * 
 *    Attribution. You must attribute the work in the manner specified by the
 *    author or licensor (but not in any way that suggests that they endorse you
 *    or your use of the work).
 * 
 *    Noncommercial. You may not use this work for commercial purposes.
 * 
 *    Share Alike. If you alter, transform, or build upon this work, you may
 *    distribute the resulting work only under the same or similar license to
 *    this one.
 * 
 * For any reuse or distribution, you must make clear to others the license
 * terms of this work. The best way to do this is with a link to this web page.
 * 
 * Any of the above conditions can be waived if you get permission from the
 * copyright holder.
 * 
 * Apart from the remix rights granted under this license, nothing in this
 * license impairs or restricts the author's moral rights.
 */


/**
 * @file
 *
 * @author Joseph Schlecht
 *
 * @brief Utility functions and globals used by all furniture related
 * programs.
 */


#include <config.h>

#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>
#include <cstdlib>
#include <cstdio>
#include <cerrno>
#include <cstring>
#include <cassert>
#include <vector>
#include <list>

#include <inttypes.h>
#include <unistd.h>

#if defined JWSCXX_HAVE_OPENGL
#include <GL/gl.h>
#include <GL/glu.h>
#elif defined JWSCXX_HAVE_OPENGL_FRAMEWORK
#include <OpenGL/OpenGL.h>
#endif

#if defined FURNITURE_HAVE_ZLIB
#include <zlib.h>
#endif

#if defined FURNITURE_HAVE_OPENMPI
#include <mpi.h>
#endif

#include <jwsc/base/memory.h>
#include <jwsc/vector/vector.h>
#include <jwsc/vector/vector_io.h>
#include <jwsc/matrix/matrix.h>
#include <jwsc/matrix/matrix_math.h>
#include <jwsc/matrix/matrix_io.h>
#include <jwsc/matrix/matrix_util.h>
#include <jwsc/image/image.h>
#include <jwsc/image/image_io.h>
#include <jwsc/image/image_util.h>
#include <jwsc/image/edge.h>

#include <jwsc++/config.h>
#include <jwsc++/base/option.h>
#include <jwsc++/base/exception.h>

#include "view.h"
#include "object.h"
#include "block.h"
#include "block_set.h"
#include "sampler.h"
#include "metropolis.h"
#include "dynamics.h"
#include "likelihood.h"
#include "render.h"
#include "options.h"


#define  SEED                    0
#define  NUM_THREADS             1
#define  NUM_CATEGORIES          1
#define  MIN_NUM_MEMBERS         1
#define  SAMPLER                 METRO
#define  PREFIX_DIR              "."
#define  INSTANCE_DIR_FMT        "%04u"
#define  CATEGORY_DIR_FMT        "%02u"
#define  CATEGORIES_IN_FNAME     0
#define  CATEGORIES_OUT_FNAME    0
#define  USE_PBUFFER             true
#define  INTERACTIVE             false
#define  COMPRESS_MODELS         false
#define  COMPRESS_INFO           false
#define  SAMPLE_INFO_FNAME       "/dev/stdout"
#define  CAPTURE_FMT             "capture-%04u.tiff"
#define  ITERATIONS              0
#define  RESOLUTION              1.0f
#define  EDGE_SIGMA              1.0f
#define  EDGE_BEGIN              0.05f
#define  EDGE_END                0.025f
#define  EDGE_PAD                100
#define  EDGE_PTS_IN_FNAME       0
#define  EDGE_PTS_OUT_FNAME      0
#define  EDGE_MAP_FNAME          0
#define  EDGE_COLOR_FNAME        0
#define  IMG_THRESH_OUT_FNAME    0
#define  MODEL_PAD               100
#define  IMG_LL_DIST_SIGMA       10.0
#define  IMG_LL_ANGLE_SIGMA      0.01
#define  IMG_LL_MAX_DIST         25
#define  IMG_LL_BG_PROB          1
#define  IMG_LL_SURF_PROB        1
#define  IMG_LL_SURF_NOISE_PROB  1.0e-8
#define  IMG_LL_SURF_MISS_PROB   1.0e-8
#define  IMG_LL_EDGE_NOISE_PROB  1.0e-64
#define  IMG_LL_EDGE_SILHO_MISS_PROB  1.0e-64
#define  IMG_LL_EDGE_INNER_MISS_PROB  1.0e-64
#define  IMG_LL_NUM_ANGLES       12
#define  IMG_LL_CACHE_SIZE       0
#define  IMG_LL_CACHE_K          4
#define  IMG_LL_CACHE_T          1.0e-8
#define  BEST_VIEW_IN_FNAME      0
#define  BEST_VIEW_OUT_FNAME     0
#define  VIEW_IN_RANDOM          false
#define  VIEW_IN_FNAME           0
#define  VIEW_OUT_FNAME          "view.model"
#define  VIEW_NEAR               100.0
#define  VIEW_FAR                10000.0
#define  VIEW_DISTANCE           900.0
#define  BEST_OBJECT_IN_FNAME    0
#define  BEST_OBJECT_OUT_FNAME   0
#define  OBJECT_IN_RANDOM        false
#define  OBJECT_IN_FNAME         0
#define  OBJECT_OUT_FNAME        "object.model"
#define  OBJECT_IMG_FNAME        0
#define  BEST_OBJECT_IMG_FNAME   0
#define  SAMPLE_STATS_IN_FNAME   0
#define  SAMPLE_STATS_OUT_FNAME  0
#define  SAMPLES_OUT_FNAME       0
#define  SAMPLE_COV_IN_EIG_K     3
#define  SAMPLE_COV_IN_EIG_SCALE 10
#define  SAMPLE_STATS_INFO_FNAME "/dev/null"
#define  SAMPLE_STATS_FREQ       100
#define  STOCH_ALPHA             0.9f
#define  STOCH_KICK              0
#define  HYBRID_L                10
#define  HYPER_GAMMA             1.0e-2f
#define  HYPER_ETA               1.0e-2f
#define  HYPER_H                 5000
#define  HYPER_D                 0.75
#define  MH_JUMP_SET_SIZE        0
#define  MH_DYNAMICS_PROB        1.0
#define  MH_DIFFUSION_PROB       1.0
#define  MH_JUMP_PROB            1.0
#define  CAT_VIEW_IN_FNAME       "view.model"
#define  CAT_OBJECT_IN_FNAME     "object.model"
#define  NUM_CAT_SAMPLES         1000
#define  CAT_SAMPLE_IMG_WIDTH    200
#define  CAT_SAMPLE_IMG_HEIGHT   200
#define  CAT_SAMPLE_IMG_OUT_FNAME "cat-samples.tiff"


extern void print_usage();


/** @brief Program option variable values. */
Option_vars opts;


/** @brief Program options. */
static jwscxx::base::Options* options = 0;


static uint32_t skip_csv_fields(const char** str_in_out, uint32_t N) 
    throw (jwscxx::base::Arg_error)
{
    uint32_t n;
    const char* str = *str_in_out;

    for (n = 0; n < N; n++)
    {
        while (*str != ',' && *str != 0)
            str++;

        if (*str == 0)
            break;

        str++;
    }

    *str_in_out = str;

    return n;
}

/** @brief Processes the 'help' program options. */
static void process_help_opt()
{   
    print_usage();
    options->print(std::cerr, 30);
    exit(EXIT_SUCCESS);
}

/** @brief Processes the 'version' program options. */
static void process_version_opt()
{
    std::cout << FURNITURE_PACKAGE_STRING << "\n";
    exit(EXIT_SUCCESS);
}

/** @brief Processes the 'sampler' program options. */
static void process_sampler_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'sampler' requires an argument");
    }
    if (strncmp(arg, "metro", 5) == 0)
    {
        opts.sampler = METRO;
    }
    else if (strncmp(arg, "lange", 5) == 0)
    {
        opts.sampler = LANGE;
    }
    else if (strncmp(arg, "hyper", 5) == 0)
    {
        opts.sampler = HYPER;
    }
    else if (strncmp(arg, "stoch", 5) == 0)
    {
        opts.sampler = STOCH;
    }
    else if (strncmp(arg, "hybrid", 6) == 0)
    {
        opts.sampler = HYBRID;
    }
    else
    {
        throw Arg_error("Option 'sampler' must be one of {metro, stoch, hybrid, lange, hyper}");
    }
}

/** @brief Processes the 'pbuffer' program options. */
static void process_pbuffer_opt() throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

#if defined JWSCXX_HAVE_GLX_PBUFFER
    opts.use_pbuffer = true;
#else
    opts.use_pbuffer = false;
    throw Arg_error("Option 'pbuffer': System does not have pbuffers");
#endif
}

/** @brief Processes the 'interactive' program options. */
static void process_interactive_opt() throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

#if defined FURNITURE_INTERACTIVE
    opts.interactive = true;
#else
    opts.interactive = false;
    throw Arg_error("Option 'interactive': System does not support interactive mode");
#endif
}
 
/** @brief Processes the 'pixmap' program options. */
static void process_pixmap_opt()
{
    opts.use_pbuffer = false;
}

/** @brief Processes the 'compress-models' program options. */
static void process_compress_models_opt() throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

#if defined FURNITURE_HAVE_ZLIB
    opts.compress_models = true;
#else
    opts.compress_models = false;
    throw Arg_error("Option 'compress-models': System does not have zlib");
#endif
}

/** @brief Processes the 'compress-info' program options. */
static void process_compress_info_opt() throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

#if defined FURNITURE_HAVE_ZLIB
    opts.compress_info = true;
#else
    opts.compress_info = false;
    throw Arg_error("Option 'compress-info': System does not have zlib");
#endif
}

/** @brief Processes the 'prefix-in-dir' program options. */
static void process_prefix_in_dir_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'prefix-in-dir' requires an argument");
    }
    opts.prefix_in_dir = arg;
}

/** @brief Processes the 'prefix-out-dir' program options. */
static void process_prefix_out_dir_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'prefix-out-dir' requires an argument");
    }
    opts.prefix_out_dir = arg;
}

/** @brief Processes the 'instance-dir-fmt' program options. */
static void process_instance_dir_fmt_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'instance-dir-fmt' requires an argument");
    }
    opts.instance_dir_fmt = arg;
}

/** @brief Processes the 'category-dir-fmt' program options. */
static void process_category_dir_fmt_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'category-dir-fmt' requires an argument");
    }
    opts.category_dir_fmt = arg;
}

/** @brief Processes the 'categories-in' program options. */
static void process_categories_in_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'categories-in' requires an argument");
    }
    opts.categories_in_fname = arg;
}

/** @brief Processes the 'categories-out' program options. */
static void process_categories_out_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'categories-out' requires an argument");
    }
    opts.categories_out_fname = arg;
}

/** @brief Processes the 'options' program options. */
static void process_options_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'options' requires an argument");
    }
    try
    {
        options->process(arg);
    }
    catch (jwscxx::base::IO_error e)
    {
        throw Arg_error(e.get_msg());
    }
}

/** @brief Processes the 'capture' program options. */
static void process_capture_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'capture' requires an argument");
    }
    opts.capture_fmt = arg;
}

/** @brief Processes the 'img-ll-dist-sigma' program options. */
static void process_img_ll_dist_sigma_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'img-ll-dist-sigma' requires an argument");
    }
    if (sscanf(arg, "%lf", &(opts.img_ll_dist_sigma)) != 1 || 
            opts.img_ll_dist_sigma < 1.0e-8)
    {
        throw Arg_error("Option 'img-ll-dist-sigma' must be > 0");
    }

    opts.img_ll_dist_sigma *= opts.resolution;
}

/** @brief Processes the 'img-ll-angle-sigma' program options. */
static void process_img_ll_angle_sigma_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'img-ll-angle-sigma' requires an argument");
    }
    if (sscanf(arg, "%lf", &(opts.img_ll_angle_sigma)) != 1 || 
            opts.img_ll_angle_sigma < 1.0e-8)
    {
        throw Arg_error("Option 'img-ll-angle-sigma' must be > 0");
    }
}

/** @brief Processes the 'img-ll-max-dist' program options. */
static void process_img_ll_max_dist_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'img-ll-max-dist' requires an argument");
    }
    if (sscanf(arg, "%lf", &(opts.img_ll_max_dist)) != 1 || 
            opts.img_ll_max_dist < 1.0e-8)
    {
        throw Arg_error("Option 'img-ll-max-dist' must be > 0");
    }

    opts.img_ll_max_dist *= opts.resolution;
}

/** @brief Processes the 'img-ll-bg-prob' program options. */
static void process_img_ll_bg_prob_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'img-ll-bg-prob' requires an argument");
    }
    if (sscanf(arg, "%lf", &(opts.img_ll_bg_prob)) != 1 || opts.img_ll_bg_prob < 0)
    {
        throw Arg_error("Option 'img-ll-bg-prob' must be >= 0");
    }
}

/** @brief Processes the 'img-ll-surf-prob' program options. */
static void process_img_ll_surf_prob_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'img-ll-surf-prob' requires an argument");
    }
    if (sscanf(arg, "%lf", &(opts.img_ll_surf_prob)) != 1 || opts.img_ll_surf_prob < 0)
    {
        throw Arg_error("Option 'img-ll-surf-prob' must be >= 0");
    }
}

/** @brief Processes the 'img-ll-surf-noise-prob' program options. */
static void process_img_ll_surf_noise_prob_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'img-ll-surf-noise-prob' requires an argument");
    }
    if (sscanf(arg, "%lf", &(opts.img_ll_surf_noise_prob)) != 1 || 
            opts.img_ll_surf_noise_prob < 0)
    {
        throw Arg_error("Option 'img-ll-surf-noise-prob' must be >= 0");
    }
}

/** @brief Processes the 'img-ll-surf-miss-prob' program options. */
static void process_img_ll_surf_miss_prob_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'img-ll-surf-miss-prob' requires an argument");
    }
    if (sscanf(arg, "%lf", &(opts.img_ll_surf_miss_prob)) != 1 || 
            opts.img_ll_surf_miss_prob < 0)
    {
        throw Arg_error("Option 'img-ll-surf-miss-prob' must be >= 0");
    }
}

/** @brief Processes the 'img-ll-edge-noise-prob' program options. */
static void process_img_ll_edge_noise_prob_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'img-ll-edge-noise-prob' requires an argument");
    }
    if (sscanf(arg, "%lf", &(opts.img_ll_edge_noise_prob)) != 1 || 
            opts.img_ll_edge_noise_prob < 0)
    {
        throw Arg_error("Option 'img-ll-edge-noise-prob' must be >= 0");
    }
}

/** @brief Processes the 'img-ll-edge-silho-miss-prob' program options. */
static void process_img_ll_edge_silho_miss_prob_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'img-ll-edge-silho-miss-prob' requires an argument");
    }
    if (sscanf(arg, "%lf", &(opts.img_ll_edge_silho_miss_prob)) != 1 || 
            opts.img_ll_edge_silho_miss_prob < 0)
    {
        throw Arg_error("Option 'img-ll-edge-silho-miss-prob' must be >= 0");
    }
}

/** @brief Processes the 'img-ll-edge-inner-miss-prob' program options. */
static void process_img_ll_edge_inner_miss_prob_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'img-ll-edge-inner-miss-prob' requires an argument");
    }
    if (sscanf(arg, "%lf", &(opts.img_ll_edge_inner_miss_prob)) != 1 || 
            opts.img_ll_edge_inner_miss_prob < 0)
    {
        throw Arg_error("Option 'img-ll-edge-inner-miss-prob' must be >= 0");
    }
}

/** @brief Processes the 'img-ll-num-angles' program options. */
static void process_img_ll_num_angles_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'img-ll-num-angles' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.img_ll_num_angles)) != 1)
    {
        throw Arg_error("Option 'img-ll-num-angles' must be > 0");
    }
}

/** @brief Processes the 'img-ll-cache-size' program options. */
static void process_img_ll_cache_size_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'img-ll-cache-size' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.img_ll_cache_size)) != 1)
    {
        throw Arg_error("Option 'img-ll-cache-size' must be > 0");
    }
}

/** @brief Processes the 'img-ll-cache-k' program options. */
static void process_img_ll_cache_k_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'img-ll-cache-k' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.img_ll_cache_k)) != 1)
    {
        throw Arg_error("Option 'img-ll-cache-k' must be > 0");
    }
}

/** @brief Processes the 'img-ll-cache-t' program options. */
static void process_img_ll_cache_t_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'img-ll-cache-t' requires an argument");
    }
    if (sscanf(arg, "%f", &(opts.img_ll_cache_t)) != 1 || opts.img_ll_cache_t < 0)
    {
        throw Arg_error("Option 'img-ll-cache-t' must be >= 0");
    }
}

/** @brief Processes the 'resolution' program options. */
static void process_resolution_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'resolution' requires an argument");
    }
    if (sscanf(arg, "%f", &(opts.resolution)) != 1 || opts.resolution <= 0 || 
            opts.resolution > 1)
    {
        throw Arg_error("Option 'resolution' must be in (0,1]");
    }

    opts.reso_model_pad = (uint32_t)ceil(opts.resolution*opts.model_pad);
    opts.img_ll_max_dist *= opts.resolution;
    opts.img_ll_dist_sigma *= opts.resolution;
}

/** @brief Processes the 'model-pad' program options. */
static void process_model_pad_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'model-pad' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.model_pad)) != 1)
    {
        throw Arg_error("Option 'model_pad' requires an argument");
    }

    opts.reso_model_pad = (uint32_t)ceil(opts.resolution*opts.model_pad);
}

/** @brief Processes the 'seed' program options. */
static void process_seed_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'seed' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.seed)) != 1)
    {
        throw Arg_error("Option 'seed' must be a non-negative integer");
    }

    srand(opts.seed);
}

/** @brief Processes the 'num-threads' program options. */
static void process_num_threads_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'num-threads' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.num_threads)) != 1 || opts.num_threads < 1)
    {
        throw Arg_error("Option 'num-threads' must > 0");
    }
}

/** @brief Processes the 'num-categories' program options. */
static void process_num_categories_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'num-categories' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.num_categories)) != 1 || opts.num_categories < 1)
    {
        throw Arg_error("Option 'num-categories' must > 0");
    }
}

/** @brief Processes the 'min-num-members' program options. */
static void process_min_num_members_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'min-num-memberss' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.min_num_members)) != 1)
    {
        throw Arg_error("Option 'min-num-members' requires an argument");
    }
}

/** @brief Processes the 'sample-info' program options. */
static void process_sample_info_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'sample-info' requires an argument");
    }
    opts.sample_info_fname = arg;
}

/** @brief Processes the 'iterations' program options. */
static void process_iterations_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'iterations' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.iterations)) != 1)
    {
        throw Arg_error("Option 'iterations' must be a non-negative integer");
    }
}

/** @brief Processes the 'edges-in' program options. */
static void process_edges_in_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'edges-in' requires an argument");
    }
    opts.edge_pts_in_fname = arg;
}

/** @brief Processes the 'edges-out' program options. */
static void process_edges_out_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'edges-out' requires an argument");
    }
    opts.edge_pts_out_fname = arg;
}

/** @brief Processes the 'edge-map' program options. */
static void process_edge_map_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'edge-map' requires an argument");
    }
    opts.edge_map_fname = arg;
}

/** @brief Processes the 'edge-color' program options. */
static void process_edge_color_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'edge-color' requires an argument");
    }
    opts.edge_color_fname = arg;
}

/** @brief Processes the 'edge-sigma' program options. */
static void process_edge_sigma_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'edge-sigma' requires an argument");
    }
    if (sscanf(arg, "%f", &(opts.edge_sigma)) != 1 || opts.edge_sigma < 1.0e-8)
    {
        throw Arg_error("Option 'edge-sigma' must be > 0");
    }
}

/** @brief Processes the 'edge-begin' program options. */
static void process_edge_begin_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'edge-begin' requires an argument");
    }
    if (sscanf(arg, "%f", &(opts.edge_begin)) != 1)
    {
        throw Arg_error("Option 'edge-begin' requires an argument");
    }
}

/** @brief Processes the 'edge-end' program options. */
static void process_edge_end_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'edge-end' requires an argument");
    }
    if (sscanf(arg, "%f", &(opts.edge_end)) != 1)
    {
        throw Arg_error("Option 'edge-end' requires an argument");
    }
}

/** @brief Processes the 'edge-pad' program options. */
static void process_edge_pad_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'edge-pad' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.edge_pad)) != 1)
    {
        throw Arg_error("Option 'edge-pad' requires an argument");
    }
}

/** @brief Processes the 'img-thresh-out' program options. */
static void process_img_thresh_out_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'img-thresh-out' requires an argument");
    }
    opts.img_thresh_out_fname = arg;
}

/** @brief Processes the 'view-in-random' program options. */
static void process_view_in_random_opt()
{
    opts.view_in_random = true;
}
 
/** @brief Processes the 'view-in' program options. */
static void process_view_in_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'view-in' requires an argument");
    }
    opts.view_in_fname = arg;
}

/** @brief Processes the 'view-out' program options. */
static void process_view_out_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'view-out' requires an argument");
    }
    opts.view_out_fname = arg;
}

/** @brief Processes the 'best-view-in' program options. */
static void process_best_view_in_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'best-view-in' requires an argument");
    }
    opts.best_view_in_fname = arg;
}

/** @brief Processes the 'best-view-out' program options. */
static void process_best_view_out_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'best-view-out' requires an argument");
    }
    opts.best_view_out_fname = arg;
}

/** @brief Processes the 'view-distance' program options. */
static void process_view_distance_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'view-distance' requires an argument");
    }
    if (sscanf(arg, "%f", &(opts.view_distance)) != 1)
    {
        throw Arg_error("Option 'view-distance' requires an argument");
    }
}

/** @brief Processes the 'view-clipping' program options. */
static void process_view_clipping_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'view-clipping' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &(opts.view_near), &(opts.view_far)) != 2)
    {
        throw Arg_error("Option 'view-clipping' has format near,far");
    }
}

/** @brief Processes the 'view-zenith' program options. */
static void process_view_zenith_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float value;

    if (!arg)
    {
        throw Arg_error("Option 'view-zenith' requires an argument");
    }
    if (sscanf(arg, "%f", &value) != 1)
    {
        throw Arg_error("Option 'view-zenith' requires an argument");
    }
    try
    {
        Model::Parameter::Context zenith(opts.view_ctx->zenith_c());

        zenith.init_value(value);
        opts.view_ctx->zenith_c(zenith);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-zenith': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-f' program options. */
static void process_view_f_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float value;

    if (!arg)
    {
        throw Arg_error("Option 'view-f' requires an argument");
    }
    if (sscanf(arg, "%f", &value) != 1)
    {
        throw Arg_error("Option 'view-f' requires an argument");
    }
    try
    {
        Model::Parameter::Context focal(opts.view_ctx->focal_c());

        focal.init_value(value);
        opts.view_ctx->focal_c(focal);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-f': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-scale' program options. */
static void process_view_scale_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float value;

    if (!arg)
    {
        throw Arg_error("Option 'view-scale' requires an argument");
    }
    if (sscanf(arg, "%f", &value) != 1)
    {
        throw Arg_error("Option 'view-scale' requires an argument");
    }
    try
    {
        Model::Parameter::Context scale(opts.view_ctx->scale_c());

        scale.init_value(value);
        opts.view_ctx->scale_c(scale);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-scale': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-zenith-range' program options. */
static void process_view_zenith_range_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float min, max;

    if (!arg)
    {
        throw Arg_error("Option 'view-zenith-range' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &min, &max) != 2)
    {
        throw Arg_error("Option 'view-zenith-range' has format min,max");
    }
    try
    {
        Model::Parameter::Context zenith(opts.view_ctx->zenith_c());
        Model::Parameter::Context::Range range(min, max);

        zenith.range(range);
        opts.view_ctx->zenith_c(zenith);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-zenith-range': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-f-range' program options. */
static void process_view_f_range_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float min, max;

    if (!arg)
    {
        throw Arg_error("Option 'view-f-range' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &min, &max) != 2)
    {
        throw Arg_error("Option 'view-f-range' has format min,max");
    }
    try
    {
        Model::Parameter::Context focal(opts.view_ctx->focal_c());
        Model::Parameter::Context::Range range(min, max);

        focal.range(range);
        opts.view_ctx->focal_c(focal);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-f-range': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-scale-range' program options. */
static void process_view_scale_range_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float min, max;

    if (!arg)
    {
        throw Arg_error("Option 'view-scale-range' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &min, &max) != 2)
    {
        throw Arg_error("Option 'view-scale-range' has format min,max");
    }
    try
    {
        Model::Parameter::Context scale(opts.view_ctx->scale_c());
        Model::Parameter::Context::Range range(min, max);

        scale.range(range);
        opts.view_ctx->scale_c(scale);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-scale-range': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-zenith-likelihood' program options. */
static void process_view_zenith_likelihood_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'view-zenith-likelihood' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'view-zenith-likelihood' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context zenith(opts.view_ctx->zenith_c());
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        zenith.likelihood(gauss);
        opts.view_ctx->zenith_c(zenith);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-zenith-likelihood': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-f-likelihood' program options. */
static void process_view_f_likelihood_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'view-f-likelihood' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'view-f-likelihood' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context focal(opts.view_ctx->focal_c());
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        focal.likelihood(gauss);
        opts.view_ctx->focal_c(focal);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-f-likelihood': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-scale-likelihood' program options. */
static void process_view_scale_likelihood_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'view-scale-likelihood' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'view-scale-likelihood' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context scale(opts.view_ctx->scale_c());
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        scale.likelihood(gauss);
        opts.view_ctx->scale_c(scale);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-scale-likelihood': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-zenith-proposal' program options. */
static void process_view_zenith_proposal_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'view-zenith-proposal' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'view-zenith-proposal' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context zenith(opts.view_ctx->zenith_c());
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        zenith.proposal(gauss);
        zenith.gauss_proposal(gauss);
        opts.view_ctx->zenith_c(zenith);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-zenith-proposal': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-f-proposal' program options. */
static void process_view_f_proposal_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'view-f-proposal' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'view-f-proposal' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context focal(opts.view_ctx->focal_c());
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        focal.proposal(gauss);
        focal.gauss_proposal(gauss);
        opts.view_ctx->focal_c(focal);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-f-proposal': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-scale-proposal' program options. */
static void process_view_scale_proposal_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'view-scale-proposal' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'view-scale-proposal' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context scale(opts.view_ctx->scale_c());
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        scale.proposal(gauss);
        scale.gauss_proposal(gauss);
        opts.view_ctx->scale_c(scale);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-scale-proposal': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-zenith-dynamics' program options. */
static void process_view_zenith_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'view-zenith-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'view-zenith-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        Model::Parameter::Context zenith(opts.view_ctx->zenith_c());
        Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

        zenith.value_dynamics(dynamics);
        opts.view_ctx->zenith_c(zenith);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-zenith-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-zenith-mean-dynamics' program options. */
static void process_view_zenith_mean_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'view-zenith-mean-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'view-zenith-mean-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        Model::Parameter::Context zenith(opts.view_ctx->zenith_c());
        Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

        zenith.mean_dynamics(dynamics);
        opts.view_ctx->zenith_c(zenith);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-zenith-mean-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-zenith-stdev-dynamics' program options. */
static void process_view_zenith_stdev_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'view-zenith-stdev-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'view-zenith-stdev-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        Model::Parameter::Context zenith(opts.view_ctx->zenith_c());
        Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

        zenith.stdev_dynamics(dynamics);
        opts.view_ctx->zenith_c(zenith);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-zenith-stdev-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-f-dynamics' program options. */
static void process_view_f_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'view-f-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'view-f-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        Model::Parameter::Context focal(opts.view_ctx->focal_c());
        Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

        focal.value_dynamics(dynamics);
        opts.view_ctx->focal_c(focal);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-f-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-f-mean-dynamics' program options. */
static void process_view_f_mean_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'view-f-mean-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'view-f-mean-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        Model::Parameter::Context focal(opts.view_ctx->focal_c());
        Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

        focal.mean_dynamics(dynamics);
        opts.view_ctx->focal_c(focal);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-f-mean-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-f-stdev-dynamics' program options. */
static void process_view_f_stdev_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'view-f-stdev-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'view-f-stdev-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        Model::Parameter::Context focal(opts.view_ctx->focal_c());
        Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

        focal.stdev_dynamics(dynamics);
        opts.view_ctx->focal_c(focal);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-f-stdev-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-scale-dynamics' program options. */
static void process_view_scale_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'view-scale-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'view-scale-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        Model::Parameter::Context scale(opts.view_ctx->scale_c());
        Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

        scale.value_dynamics(dynamics);
        opts.view_ctx->scale_c(scale);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-scale-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-scale-mean-dynamics' program options. */
static void process_view_scale_mean_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'view-scale-mean-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'view-scale-mean-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        Model::Parameter::Context scale(opts.view_ctx->scale_c());
        Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

        scale.mean_dynamics(dynamics);
        opts.view_ctx->scale_c(scale);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-scale-mean-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'view-scale-stdev-dynamics' program options. */
static void process_view_scale_stdev_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'view-scale-stdev-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'view-scale-stdev-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        Model::Parameter::Context scale(opts.view_ctx->scale_c());
        Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

        scale.stdev_dynamics(dynamics);
        opts.view_ctx->scale_c(scale);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'view-scale-stdev-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-in-random' program options. */
static void process_object_in_random_opt()
{
    opts.object_in_random = true;
}
 
/** @brief Processes the 'object-in' program options. */
static void process_object_in_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'object-in' requires an argument");
    }
    opts.object_in_fname = arg;
}

/** @brief Processes the 'object-out' program options. */
static void process_object_out_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'object-out' requires an argument");
    }
    opts.object_out_fname = arg;
}

/** @brief Processes the 'best-object-in' program options. */
static void process_best_object_in_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'best-object-in' requires an argument");
    }
    opts.best_object_in_fname = arg;
}

/** @brief Processes the 'best-object-out' program options. */
static void process_best_object_out_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'best-object-out' requires an argument");
    }
    opts.best_object_out_fname = arg;
}
 
/** @brief Processes the 'object-azimuth' program options. */
static void process_object_azimuth_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float value;

    if (!arg)
    {
        throw Arg_error("Option 'object-azimuth' requires an argument");
    }
    if (sscanf(arg, "%f", &value) != 1)
    {
        throw Arg_error("Option 'object-azimuth' requires an argument");
    }
    try
    {
        Model::Parameter::Context azimuth(opts.obj_ctx->azimuth_c());

        azimuth.init_value(value);
        opts.obj_ctx->azimuth_c(azimuth);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-azimuth': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-x-posn' program options. */
static void process_object_x_posn_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float value;

    if (!arg)
    {
        throw Arg_error("Option 'object-x-posn' requires an argument");
    }
    if (sscanf(arg, "%f", &value) != 1)
    {
        throw Arg_error("Option 'object-x-posn' requires an argument");
    }
    try
    {
        Model::Parameter::Context x_posn(opts.obj_ctx->posn_c(0));

        x_posn.init_value(value);
        opts.obj_ctx->posn_c(0, x_posn);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-x-posn': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-y-posn' program options. */
static void process_object_y_posn_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float value;

    if (!arg)
    {
        throw Arg_error("Option 'object-y-posn' requires an argument");
    }
    if (sscanf(arg, "%f", &value) != 1)
    {
        throw Arg_error("Option 'object-y-posn' requires an argument");
    }
    try
    {
        Model::Parameter::Context y_posn(opts.obj_ctx->posn_c(1));

        y_posn.init_value(value);
        opts.obj_ctx->posn_c(1, y_posn);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-y-posn': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-z-posn' program options. */
static void process_object_z_posn_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float value;

    if (!arg)
    {
        throw Arg_error("Option 'object-z-posn' requires an argument");
    }
    if (sscanf(arg, "%f", &value) != 1)
    {
        throw Arg_error("Option 'object-z-posn' requires an argument");
    }
    try
    {
        Model::Parameter::Context z_posn(opts.obj_ctx->posn_c(2));

        z_posn.init_value(value);
        opts.obj_ctx->posn_c(2, z_posn);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-z-posn': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-azimuth-range' program options. */
static void process_object_azimuth_range_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float min, max;

    if (!arg)
    {
        throw Arg_error("Option 'object-azimuth-range' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &min, &max) != 2)
    {
        throw Arg_error("Option 'object-azimuth-range' has format min,max");
    }
    try
    {
        Model::Parameter::Context azimuth(opts.obj_ctx->azimuth_c());
        Model::Parameter::Context::Range range(min, max);

        azimuth.range(range);
        opts.obj_ctx->azimuth_c(azimuth);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-azimuth-range': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-x-posn-range' program options. */
static void process_object_x_posn_range_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float min, max;

    if (!arg)
    {
        throw Arg_error("Option 'object-x-posn-range' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &min, &max) != 2)
    {
        throw Arg_error("Option 'object-x-posn-range' has format min,max");
    }
    try
    {
        Model::Parameter::Context x_posn(opts.obj_ctx->posn_c(0));
        Model::Parameter::Context::Range range(min, max);

        x_posn.range(range);
        opts.obj_ctx->posn_c(0, x_posn);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-x-posn-range': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-y-posn-range' program options. */
static void process_object_y_posn_range_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float min, max;

    if (!arg)
    {
        throw Arg_error("Option 'object-y-posn-range' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &min, &max) != 2)
    {
        throw Arg_error("Option 'object-y-posn-range' has format min,max");
    }
    try
    {
        Model::Parameter::Context y_posn(opts.obj_ctx->posn_c(1));
        Model::Parameter::Context::Range range(min, max);

        y_posn.range(range);
        opts.obj_ctx->posn_c(1, y_posn);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-y-posn-range': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-z-posn-range' program options. */
static void process_object_z_posn_range_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float min, max;

    if (!arg)
    {
        throw Arg_error("Option 'object-z-posn-range' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &min, &max) != 2)
    {
        throw Arg_error("Option 'object-z-posn-range' has format min,max");
    }
    try
    {
        Model::Parameter::Context z_posn(opts.obj_ctx->posn_c(2));
        Model::Parameter::Context::Range range(min, max);

        z_posn.range(range);
        opts.obj_ctx->posn_c(2, z_posn);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-z-posn-range': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-x-posn-likelihood' program options. */
static void process_object_x_posn_likelihood_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'object-x-posn-likelihood' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'object-x-posn-likelihood' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context x_posn(opts.obj_ctx->posn_c(0));
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        x_posn.likelihood(gauss);
        opts.obj_ctx->posn_c(0, x_posn);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-x-posn-likelihood': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-y-posn-likelihood' program options. */
static void process_object_y_posn_likelihood_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'object-y-posn-likelihood' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'object-y-posn-likelihood' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context y_posn(opts.obj_ctx->posn_c(1));
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        y_posn.likelihood(gauss);
        opts.obj_ctx->posn_c(1, y_posn);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-y-posn-likelihood': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-z-posn-likelihood' program options. */
static void process_object_z_posn_likelihood_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'object-z-posn-likelihood' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'object-z-posn-likelihood' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context z_posn(opts.obj_ctx->posn_c(2));
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        z_posn.likelihood(gauss);
        opts.obj_ctx->posn_c(2, z_posn);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-z-posn-likelihood': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-azimuth-proposal' program options. */
static void process_object_azimuth_proposal_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'object-azimuth-proposal' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'object-azimuth-proposal' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context azimuth(opts.obj_ctx->azimuth_c());
        Model::Parameter::Context::Uniform uni;
        Model::Parameter::Context::Circular_gaussian gauss(mu, sigma);

        azimuth.proposal(uni);
        azimuth.gauss_proposal(gauss);
        opts.obj_ctx->azimuth_c(azimuth);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-azimuth-proposal': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-x-posn-proposal' program options. */
static void process_object_x_posn_proposal_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'object-x-posn-proposal' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'object-x-posn-proposal' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context x_posn(opts.obj_ctx->posn_c(0));
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        x_posn.proposal(gauss);
        x_posn.gauss_proposal(gauss);
        opts.obj_ctx->posn_c(0, x_posn);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-x-posn-proposal': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-y-posn-proposal' program options. */
static void process_object_y_posn_proposal_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'object-y-posn-proposal' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'object-y-posn-proposal' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context y_posn(opts.obj_ctx->posn_c(1));
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        y_posn.proposal(gauss);
        y_posn.gauss_proposal(gauss);
        opts.obj_ctx->posn_c(1, y_posn);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-y-posn-proposal': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-z-posn-proposal' program options. */
static void process_object_z_posn_proposal_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'object-z-posn-proposal' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'object-z-posn-proposal' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context z_posn(opts.obj_ctx->posn_c(2));
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        z_posn.proposal(gauss);
        z_posn.gauss_proposal(gauss);
        opts.obj_ctx->posn_c(2, z_posn);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-z-posn-proposal': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-azimuth-dynamics' program options. */
static void process_object_azimuth_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'object-azimuth-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'object-azimuth-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        Model::Parameter::Context azimuth(opts.obj_ctx->azimuth_c());
        Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

        azimuth.value_dynamics(dynamics);
        opts.obj_ctx->azimuth_c(azimuth);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-azimuth-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-azimuth-mean-dynamics' program options. */
static void process_object_azimuth_mean_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'object-azimuth-mean-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'object-azimuth-mean-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        Model::Parameter::Context azimuth(opts.obj_ctx->azimuth_c());
        Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

        azimuth.mean_dynamics(dynamics);
        opts.obj_ctx->azimuth_c(azimuth);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-azimuth-mean-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-azimuth-stdev-dynamics' program options. */
static void process_object_azimuth_stdev_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'object-azimuth-stdev-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'object-azimuth-stdev-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        Model::Parameter::Context azimuth(opts.obj_ctx->azimuth_c());
        Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

        azimuth.stdev_dynamics(dynamics);
        opts.obj_ctx->azimuth_c(azimuth);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-azimuth-stdev-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-posn-dynamics' program options. */
static void process_object_posn_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'object-posn-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'object-posn-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        for (uint32_t i = 0; i < 3; i++)
        {
            Model::Parameter::Context posn(opts.obj_ctx->posn_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            posn.value_dynamics(dynamics);
            opts.obj_ctx->posn_c(i, posn);
        }

        for (uint32_t i = 0; i < 2; i++)
        {
            Model::Parameter::Context posn(opts.block_ctx->var_attach_pt_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            posn.value_dynamics(dynamics);
            opts.block_ctx->var_attach_pt_c(i, posn);
        }

        for (uint32_t i = 0; i < 2; i++)
        {
            Model::Parameter::Context posn(
                    opts.block_set_ctx->var_attach_pt_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            posn.value_dynamics(dynamics);
            opts.block_set_ctx->var_attach_pt_c(i, posn);
        }
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-posn-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-posn-mean-dynamics' program options. */
static void process_object_posn_mean_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'object-posn-mean-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'object-posn-mean-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        for (uint32_t i = 0; i < 3; i++)
        {
            Model::Parameter::Context posn(opts.obj_ctx->posn_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            posn.mean_dynamics(dynamics);
            opts.obj_ctx->posn_c(i, posn);
        }

        for (uint32_t i = 0; i < 2; i++)
        {
            Model::Parameter::Context posn(opts.block_ctx->var_attach_pt_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            posn.mean_dynamics(dynamics);
            opts.block_ctx->var_attach_pt_c(i, posn);
        }

        for (uint32_t i = 0; i < 2; i++)
        {
            Model::Parameter::Context posn(
                    opts.block_set_ctx->var_attach_pt_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            posn.mean_dynamics(dynamics);
            opts.block_set_ctx->var_attach_pt_c(i, posn);
        }
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-posn-mean-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-posn-stdev-dynamics' program options. */
static void process_object_posn_stdev_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'object-posn-stdev-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'object-posn-stdev-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        for (uint32_t i = 0; i < 3; i++)
        {
            Model::Parameter::Context posn(opts.obj_ctx->posn_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            posn.stdev_dynamics(dynamics);
            opts.obj_ctx->posn_c(i, posn);
        }

        for (uint32_t i = 0; i < 2; i++)
        {
            Model::Parameter::Context posn(opts.block_ctx->var_attach_pt_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            posn.stdev_dynamics(dynamics);
            opts.block_ctx->var_attach_pt_c(i, posn);
        }

        for (uint32_t i = 0; i < 2; i++)
        {
            Model::Parameter::Context posn(
                    opts.block_set_ctx->var_attach_pt_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            posn.stdev_dynamics(dynamics);
            opts.block_set_ctx->var_attach_pt_c(i, posn);
        }
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-posn-stdev-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-block-w' program options. */
static void process_object_block_w_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float value;

    if (!arg)
    {
        throw Arg_error("Option 'object-block-w' requires an argument");
    }
    if (sscanf(arg, "%f", &value) != 1)
    {
        throw Arg_error("Option 'object-block-w' requires an argument");
    }
    try
    {
        Model::Parameter::Context width(opts.block_ctx->size_c(0));

        width.init_value(value);
        opts.block_ctx->size_c(0, width);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-block-w': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-block-h' program options. */
static void process_object_block_h_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float value;

    if (!arg)
    {
        throw Arg_error("Option 'object-block-h' requires an argument");
    }
    if (sscanf(arg, "%f", &value) != 1)
    {
        throw Arg_error("Option 'object-block-h' requires an argument");
    }
    try
    {
        Model::Parameter::Context height(opts.block_ctx->size_c(1));

        height.init_value(value);
        opts.block_ctx->size_c(1, height);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-block-h': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-block-l' program options. */
static void process_object_block_l_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float value;

    if (!arg)
    {
        throw Arg_error("Option 'object-block-l' requires an argument");
    }
    if (sscanf(arg, "%f", &value) != 1)
    {
        throw Arg_error("Option 'object-block-l' requires an argument");
    }
    try
    {
        Model::Parameter::Context length(opts.block_ctx->size_c(2));

        length.init_value(value);
        opts.block_ctx->size_c(2, length);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-block-l': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-block-w-range' program options. */
static void process_object_block_w_range_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float min, max;

    if (!arg)
    {
        throw Arg_error("Option 'object-block-w-range' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &min, &max) != 2)
    {
        throw Arg_error("Option 'object-block-w-range' has format min,max");
    }
    try
    {
        Model::Parameter::Context width(opts.block_ctx->size_c(0));
        Model::Parameter::Context::Range range(min, max);

        width.range(range);
        opts.block_ctx->size_c(0, width);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-block-w-range': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-block-h-range' program options. */
static void process_object_block_h_range_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float min, max;

    if (!arg)
    {
        throw Arg_error("Option 'object-block-h-range' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &min, &max) != 2)
    {
        throw Arg_error("Option 'object-block-h-range' has format min,max");
    }
    try
    {
        Model::Parameter::Context height(opts.block_ctx->size_c(1));
        Model::Parameter::Context::Range range(min, max);

        height.range(range);
        opts.block_ctx->size_c(1, height);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-block-h-range': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-block-l-range' program options. */
static void process_object_block_l_range_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float min, max;

    if (!arg)
    {
        throw Arg_error("Option 'object-block-l-range' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &min, &max) != 2)
    {
        throw Arg_error("Option 'object-block-l-range' has format min,max");
    }
    try
    {
        Model::Parameter::Context length(opts.block_ctx->size_c(2));
        Model::Parameter::Context::Range range(min, max);

        length.range(range);
        opts.block_ctx->size_c(2, length);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-block-l-range': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-block-w-likelihood' program options. */
static void process_object_block_w_likelihood_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'object-block-w-likelihood' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'object-block-w-likelihood' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context width(opts.block_ctx->size_c(0));
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        width.likelihood(gauss);
        opts.block_ctx->size_c(0, width);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-block-w-likelihood': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-block-h-likelihood' program options. */
static void process_object_block_h_likelihood_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'object-block-h-likelihood' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'object-block-h-likelihood' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context height(opts.block_ctx->size_c(1));
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        height.likelihood(gauss);
        opts.block_ctx->size_c(1, height);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-block-h-likelihood': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-block-l-likelihood' program options. */
static void process_object_block_l_likelihood_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'object-block-l-likelihood' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'object-block-l-likelihood' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context length(opts.block_ctx->size_c(2));
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        length.likelihood(gauss);
        opts.block_ctx->size_c(2, length);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-block-l-likelihood': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-block-size-prior' program options. */
static void process_object_block_size_prior_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;
    using jwscxx::base::Arg_error;

    uint32_t N;
    float    w_mu, h_mu, l_mu;
    float    lambda, alpha, beta;

    Vector_d* pi = 0;
    Matrix_d* mu = 0;

    std::ostringstream ost;
    const char* parg    = arg;
    const char* fmt_err = "Format is N,w_mu1,h_mu1,l_mu1,...,w_muN,h_muN,l_muN,lambda,alpha,beta";

    try {

    if (!arg)
    {
        throw Arg_error(fmt_err);
    }

    if (sscanf(parg, "%u,", &N) != 1 || N < 1 || skip_csv_fields(&parg, 1) != 1)
    {
        throw Arg_error(fmt_err);
    }

    create_init_vector_d(&pi, N, 1.0/N);
    create_matrix_d(&mu, N, 3);

    for (size_t n = 0; n < N; n++)
    {
        if (sscanf(parg, "%f,%f,%f,", &w_mu, &h_mu, &l_mu) != 3 || 
                skip_csv_fields(&parg, 3) != 3)
        {
            throw Arg_error(fmt_err);
        }
        mu->elts[ n ][0] = w_mu;
        mu->elts[ n ][1] = h_mu;
        mu->elts[ n ][2] = l_mu;
    }

    if (sscanf(parg, "%f,%f,%f", &lambda, &alpha, &beta) != 3)
    {
        throw Arg_error(fmt_err);
    }

    Component::Paramgroup::Context size(opts.block_ctx->size_c());
    Component::Paramgroup::Context::Iso_gauss_gamma gg(pi, mu, lambda,
            alpha, beta);

    size.prior(gg);
    opts.block_ctx->size_c(size);

    } catch (Arg_error e)
    { ost << "Option 'object-block-size-prior': " << e.get_msg(); }

    free_vector_d(pi);
    free_matrix_d(mu);

    if (ost.str().size())
    {
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-block-attach-prior' program options. */
static void process_object_block_attach_prior_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;
    using jwscxx::base::Arg_error;

    float lambda, alpha, beta;

    std::ostringstream ost;
    const char* fmt_err = "Format is lambda,alpha,beta";

    try {

    if (!arg)
    {
        throw Arg_error(fmt_err);
    }

    if (sscanf(arg, "%f,%f,%f", &lambda, &alpha, &beta) != 3)
    {
        throw Arg_error(fmt_err);
    }

    Component::Paramgroup::Context attach(opts.block_ctx->var_attach_pts_c());
    Component::Paramgroup::Context::UV_gauss_gamma gg(lambda, alpha, beta);

    attach.prior(gg);
    opts.block_ctx->var_attach_pts_c(attach);

    } catch (Arg_error e)
    { 
        ost << "Option 'object-block-attach-prior': " << e.get_msg(); 
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-block-w-proposal' program options. */
static void process_object_block_w_proposal_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'object-block-w-proposal' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'object-block-w-proposal' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context width(opts.block_ctx->size_c(0));
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        width.proposal(gauss);
        width.gauss_proposal(gauss);
        opts.block_ctx->size_c(0, width);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-block-w-proposal': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-block-h-proposal' program options. */
static void process_object_block_h_proposal_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'object-block-h-proposal' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'object-block-h-proposal' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context height(opts.block_ctx->size_c(1));
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        height.proposal(gauss);
        height.gauss_proposal(gauss);
        opts.block_ctx->size_c(1, height);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-block-h-proposal': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-block-l-proposal' program options. */
static void process_object_block_l_proposal_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float mu, sigma;

    if (!arg)
    {
        throw Arg_error("Option 'object-block-l-proposal' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &mu, &sigma) != 2)
    {
        throw Arg_error("Option 'object-block-l-proposal' has format mu,sigma");
    }
    try
    {
        Model::Parameter::Context length(opts.block_ctx->size_c(2));
        Model::Parameter::Context::Gaussian gauss(mu, sigma);

        length.proposal(gauss);
        length.gauss_proposal(gauss);
        opts.block_ctx->size_c(2, length);
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-block-l-proposal': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-size-dynamics' program options. */
static void process_object_size_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'object-size-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'object-size-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        for (uint32_t i = 0; i < 3; i++)
        {
            Model::Parameter::Context size(opts.block_ctx->size_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            size.value_dynamics(dynamics);
            opts.block_ctx->size_c(i, size);
        }

        for (uint32_t i = 0; i < 3; i++)
        {
            Model::Parameter::Context size(opts.block_set_ctx->size_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            size.value_dynamics(dynamics);
            opts.block_set_ctx->size_c(i, size);
        }

        for (uint32_t i = 0; i < 2; i++)
        {
            Model::Parameter::Context size(opts.block_set_ctx->inner_size_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            size.value_dynamics(dynamics);
            opts.block_set_ctx->inner_size_c(i, size);
        }
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-size-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-size-mean-dynamics' program options. */
static void process_object_size_mean_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'object-size-mean-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'object-size-mean-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        for (uint32_t i = 0; i < 3; i++)
        {
            Model::Parameter::Context size(opts.block_ctx->size_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            size.mean_dynamics(dynamics);
            opts.block_ctx->size_c(i, size);
        }

        for (uint32_t i = 0; i < 3; i++)
        {
            Model::Parameter::Context size(opts.block_set_ctx->size_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            size.mean_dynamics(dynamics);
            opts.block_set_ctx->size_c(i, size);
        }

        for (uint32_t i = 0; i < 2; i++)
        {
            Model::Parameter::Context size(opts.block_set_ctx->inner_size_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            size.mean_dynamics(dynamics);
            opts.block_set_ctx->inner_size_c(i, size);
        }
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-size-mean-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-size-stdev-dynamics' program options. */
static void process_object_size_stdev_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'object-size-stdev-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'object-size-stdev-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        for (uint32_t i = 0; i < 3; i++)
        {
            Model::Parameter::Context size(opts.block_ctx->size_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            size.stdev_dynamics(dynamics);
            opts.block_ctx->size_c(i, size);
        }

        for (uint32_t i = 0; i < 3; i++)
        {
            Model::Parameter::Context size(opts.block_set_ctx->size_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            size.stdev_dynamics(dynamics);
            opts.block_set_ctx->size_c(i, size);
        }

        for (uint32_t i = 0; i < 2; i++)
        {
            Model::Parameter::Context size(opts.block_set_ctx->inner_size_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            size.stdev_dynamics(dynamics);
            opts.block_set_ctx->inner_size_c(i, size);
        }
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-size-stdev-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-ratio-size-mean-dynamics' program options. */
static void process_object_ratio_size_mean_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'object-ratio-size-mean-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'object-ratio-size-mean-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        for (uint32_t i = 0; i < 3; i++)
        {
            Model::Parameter::Context size(opts.block_ctx->size_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            size.ratio_mean_dynamics(dynamics);
            opts.block_ctx->size_c(i, size);
        }

        for (uint32_t i = 0; i < 3; i++)
        {
            Model::Parameter::Context size(opts.block_set_ctx->size_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            size.ratio_mean_dynamics(dynamics);
            opts.block_set_ctx->size_c(i, size);
        }

        for (uint32_t i = 0; i < 2; i++)
        {
            Model::Parameter::Context size(opts.block_set_ctx->inner_size_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            size.ratio_mean_dynamics(dynamics);
            opts.block_set_ctx->inner_size_c(i, size);
        }
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-ratio-size-mean-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-ratio-size-stdev-dynamics' program options. */
static void process_object_ratio_size_stdev_dynamics_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    float delta_t, grad_eta;

    if (!arg)
    {
        throw Arg_error("Option 'object-ratio-size-stdev-dynamics' requires an argument");
    }
    if (sscanf(arg, "%f,%f", &delta_t, &grad_eta) != 2)
    {
        throw Arg_error("Option 'object-ratio-size-stdev-dynamics' has format delta_t,grad_eta");
    }
    try
    {
        for (uint32_t i = 0; i < 3; i++)
        {
            Model::Parameter::Context size(opts.block_ctx->size_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            size.ratio_stdev_dynamics(dynamics);
            opts.block_ctx->size_c(i, size);
        }

        for (uint32_t i = 0; i < 3; i++)
        {
            Model::Parameter::Context size(opts.block_set_ctx->size_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            size.ratio_stdev_dynamics(dynamics);
            opts.block_set_ctx->size_c(i, size);
        }

        for (uint32_t i = 0; i < 2; i++)
        {
            Model::Parameter::Context size(opts.block_set_ctx->inner_size_c(i));
            Model::Parameter::Context::Dynamics dynamics(delta_t, grad_eta);

            size.ratio_stdev_dynamics(dynamics);
            opts.block_set_ctx->inner_size_c(i, size);
        }
    }
    catch (jwscxx::base::Arg_error e)
    {
        std::ostringstream ost;
        ost << "Option 'object-ratio-size-stdev-dynamics': " << e.get_msg();
        throw Arg_error(ost.str());
    }
}

/** @brief Processes the 'object-img' program options. */
static void process_object_img_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'object-img' requires an argument");
    }
    opts.object_img_fname = arg;
}

/** @brief Processes the 'best-object-img' program options. */
static void process_best_object_img_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'best-object-img' requires an argument");
    }
    opts.best_object_img_fname = arg;
}

/** @brief Processes the 'sample-stats-in' program options. */
static void process_sample_stats_in_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'sample-stats-in' requires an argument");
    }
    opts.sample_stats_in_fname = arg;
}

/** @brief Processes the 'sample-stats-out' program options. */
static void process_sample_stats_out_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'sample-stats-out' requires an argument");
    }
    opts.sample_stats_out_fname = arg;
}

/** @brief Processes the 'samples-out' program options. */
static void process_samples_out_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'samples-out' requires an argument");
    }
    opts.samples_out_fname = arg;
}

/** @brief Processes the 'sample-cov-in-eig-k' program options. */
static void process_sample_cov_in_eig_k_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'sample-cov-in-eig-k' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.sample_cov_in_eig_k)) != 1)
    {
        throw Arg_error("Option 'sample-cov-in-eig-k' requires an argument");
    }
}

/** @brief Processes the 'sample-cov-in-eig-scale' program options. */
static void process_sample_cov_in_eig_scale_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'sample-cov-in-eig-scale' requires an argument");
    }
    if (sscanf(arg, "%lf", &(opts.sample_cov_in_eig_scale)) != 1 || 
            opts.sample_cov_in_eig_scale <= 0)
    {
        throw Arg_error("Option 'samples-in-eig-scale' must be > 0");
    }
}

/** @brief Processes the 'sample-stats-info' program options. */
static void process_sample_stats_info_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'sample-stats-info' requires an argument");
    }
    opts.sample_stats_info_fname = arg;
}

/** @brief Processes the 'sample-stats-freq' program options. */
static void process_sample_stats_freq_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'sample-stats-freq' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.sample_stats_freq)) != 1)
    {
        throw Arg_error("Option 'samples-stats-freq' requires an argument");
    }
}

/** @brief Processes the 'stoch-alpha' program options. */
static void process_stoch_alpha_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'stoch-alpha' requires an argument");
    }
    if (sscanf(arg, "%f", &(opts.stoch_alpha)) != 1 || opts.stoch_alpha < 0 || 
            opts.stoch_alpha > 1)
    {
        throw Arg_error("Option 'stoch-alpha' must be in [0,1]");
    }
}

/** @brief Processes the 'stoch-kick' program options. */
static void process_stoch_kick_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'stoch-kick' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.stoch_kick)) != 1)
    {
        throw Arg_error("Option 'stoch-kick' requires an argument");
    }
}

/** @brief Processes the 'hybrid-l' program options. */
static void process_hybrid_l_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'hybrid-l' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.hybrid_l)) != 1 || opts.hybrid_l < 1)
    {
        throw Arg_error("Option 'hybrid-l' must be > 0");
    }
}

/** @brief Processes the 'hyper-gamma' program options. */
static void process_hyper_gamma_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'hyper-gamma' requires an argument");
    }
    if (sscanf(arg, "%f", &(opts.hyper_gamma)) != 1 || 
            opts.hyper_gamma < 1.0e-16)
    {
        throw Arg_error("Option 'hyper-gamma' must be > 0");
    }
}

/** @brief Processes the 'hyper-eta' program options. */
static void process_hyper_eta_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'hyper-eta' requires an argument");
    }
    if (sscanf(arg, "%f", &(opts.hyper_eta)) != 1 || opts.hyper_eta < 1.0e-16)
    {
        throw Arg_error("Option 'hyper-eta' must be > 0");
    }
}

/** @brief Processes the 'hyper-h' program options. */
static void process_hyper_h_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'hyper-h' requires an argument");
    }
    if (sscanf(arg, "%f", &(opts.hyper_h)) != 1 || opts.hyper_h < 0)
    {
        throw Arg_error("Option 'hyper-h' must be > 0");
    }
}

/** @brief Processes the 'hyper-d' program options. */
static void process_hyper_d_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'hyper-d' requires an argument");
    }
    if (sscanf(arg, "%f", &(opts.hyper_d)) != 1 || opts.hyper_d < 0)
    {
        throw Arg_error("Option 'hyper-d' must be > 0");
    }
}

/** @brief Processes the 'mh-jump-set-size' program options. */
static void process_mh_jump_set_size_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'mh-jump-set-size' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.mh_jump_set_size)) != 1)
    {
        throw Arg_error("Option 'mh-jump-set-size' requires an argument");
    }
}

/** @brief Processes the 'mh-dynamics-prob' program options. */
static void process_mh_dynamics_prob_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'mh-dynamics-prob' requires an argument");
    }
    if (sscanf(arg, "%lf", &(opts.mh_dynamics_prob)) != 1 || 
            opts.mh_dynamics_prob < 0 || opts.mh_dynamics_prob > 1)
    {
        throw Arg_error("Option 'mh-dynamics-prob' must be in [0,1]");
    }
}

/** @brief Processes the 'mh-diffusion-prob' program options. */
static void process_mh_diffusion_prob_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'mh-diffusion-prob' requires an argument");
    }
    if (sscanf(arg, "%lf", &(opts.mh_diffusion_prob)) != 1 || 
            opts.mh_diffusion_prob < 0 || opts.mh_diffusion_prob > 1)
    {
        throw Arg_error("Option 'mh-diffusion-prob' must be in [0,1]");
    }
}

/** @brief Processes the 'mh-jump-prob' program options. */
static void process_mh_jump_prob_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'mh-jump-prob' requires an argument");
    }
    if (sscanf(arg, "%lf", &(opts.mh_jump_prob)) != 1 || 
            opts.mh_jump_prob < 0 || opts.mh_jump_prob > 1)
    {
        throw Arg_error("Option 'mh-diffusion-prob' must be in [0,1]");
    }
}

/** @brief Processes the 'num-images' program options. */
static void process_num_images_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'num-images' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.num_images)) != 1 || 
            opts.num_images == 0)
    {
        throw Arg_error("Option 'num-images' must be > 0");
    }
}

/** @brief Processes the 'cat-view-in' program options. */
static void process_cat_view_in_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'cat-view-in' requires an argument");
    }
    opts.cat_view_in_fname = arg;
}

/** @brief Processes the 'cat-object-in' program options. */
static void process_cat_object_in_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'cat-object-in' requires an argument");
    }
    opts.cat_object_in_fname = arg;
}

/** @brief Processes the 'num-cat-samples' program options. */
static void process_num_cat_samples_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'num-cat-samples' requires an argument");
    }
    if (sscanf(arg, "%u", &(opts.num_cat_samples)) == 0)
    {
        throw Arg_error("Option 'num-cat-samples' requires an argument");
    }
}

/** @brief Processes the 'cat-sample-img-size' program options. */
static void process_cat_sample_img_size_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'cat-sample-img-size' requires an argument");
    }
    if (sscanf(arg, "%u,%u", &(opts.cat_sample_img_width),
                &(opts.cat_sample_img_height)) !=2)
    {
        throw Arg_error("Option 'cat-sample-img-size' has format width,height");
    }
}

/** @brief Processes the 'cat-sample-img-out' program options. */
static void process_cat_sample_img_out_opt(const char* arg) 
throw (jwscxx::base::Arg_error)
{
    using jwscxx::base::Arg_error;

    if (!arg)
    {
        throw Arg_error("Option 'cat-sample-img-out' requires an argument");
    }
    opts.cat_sample_img_out_fname = arg;
}



/** @brief Initializes the opts values. */
static void init_opts()
{
    opts.seed = SEED;
    opts.num_threads = NUM_THREADS;
    opts.num_categories = NUM_CATEGORIES;
    opts.min_num_members = MIN_NUM_MEMBERS;
    opts.sampler = SAMPLER;
    opts.prefix_in_dir = PREFIX_DIR;
    opts.prefix_out_dir = PREFIX_DIR;
    opts.instance_dir_fmt = INSTANCE_DIR_FMT;
    opts.category_dir_fmt = CATEGORY_DIR_FMT;
    opts.categories_in_fname = CATEGORIES_IN_FNAME;
    opts.categories_out_fname = CATEGORIES_OUT_FNAME;

#if defined JWSCXX_HAVE_GLX_PBUFFER
    opts.use_pbuffer = USE_PBUFFER;
#else
    opts.use_pbuffer = false;
#endif

#if defined FURNITURE_INTERACTIVE
    opts.interactive = INTERACTIVE;
#else
    opts.interactive = false;
#endif

#if defined FURNITURE_HAVE_ZLIB
    opts.compress_models = COMPRESS_MODELS;
    opts.compress_info = COMPRESS_INFO;
#else
    opts.compress_models = false;
    opts.compress_info = false;
#endif

    opts.sample_info_fname = SAMPLE_INFO_FNAME;
    opts.capture_fmt = CAPTURE_FMT;
    opts.iterations = ITERATIONS;
    opts.resolution = RESOLUTION;
    opts.edge_sigma = EDGE_SIGMA;
    opts.edge_begin = EDGE_BEGIN;
    opts.edge_end = EDGE_END;
    opts.edge_pad = EDGE_PAD;
    opts.edge_pts_in_fname = EDGE_PTS_IN_FNAME;
    opts.edge_pts_out_fname = EDGE_PTS_OUT_FNAME;
    opts.edge_map_fname = EDGE_MAP_FNAME;
    opts.edge_color_fname = EDGE_COLOR_FNAME;
    opts.img_thresh_out_fname = IMG_THRESH_OUT_FNAME;
    opts.model_pad = MODEL_PAD;
    opts.img_ll_dist_sigma = IMG_LL_DIST_SIGMA;
    opts.img_ll_angle_sigma = IMG_LL_ANGLE_SIGMA;
    opts.img_ll_max_dist = IMG_LL_MAX_DIST;
    opts.img_ll_bg_prob = IMG_LL_BG_PROB;
    opts.img_ll_surf_prob = IMG_LL_SURF_PROB;
    opts.img_ll_surf_noise_prob = IMG_LL_SURF_NOISE_PROB;
    opts.img_ll_surf_miss_prob = IMG_LL_SURF_MISS_PROB;
    opts.img_ll_edge_noise_prob = IMG_LL_EDGE_NOISE_PROB;
    opts.img_ll_edge_silho_miss_prob = IMG_LL_EDGE_SILHO_MISS_PROB;
    opts.img_ll_edge_inner_miss_prob = IMG_LL_EDGE_INNER_MISS_PROB;
    opts.img_ll_num_angles = IMG_LL_NUM_ANGLES;
    opts.img_ll_cache_size = IMG_LL_CACHE_SIZE;
    opts.img_ll_cache_k = IMG_LL_CACHE_K;
    opts.img_ll_cache_t = IMG_LL_CACHE_T;
    opts.view_in_random = VIEW_IN_RANDOM;
    opts.view_in_fname = VIEW_IN_FNAME;
    opts.view_out_fname = VIEW_OUT_FNAME;
    opts.view_near = VIEW_NEAR;
    opts.view_far = VIEW_FAR;
    opts.view_distance = VIEW_DISTANCE;
    opts.best_view_in_fname = BEST_VIEW_IN_FNAME;
    opts.best_view_out_fname = BEST_VIEW_OUT_FNAME;
    opts.object_in_random = OBJECT_IN_RANDOM;
    opts.object_in_fname = OBJECT_IN_FNAME;
    opts.object_out_fname = OBJECT_OUT_FNAME;
    opts.best_object_in_fname = BEST_OBJECT_IN_FNAME;
    opts.best_object_out_fname = BEST_OBJECT_OUT_FNAME;
    opts.object_img_fname = OBJECT_IMG_FNAME;
    opts.best_object_img_fname = BEST_OBJECT_IMG_FNAME;
    opts.sample_stats_in_fname = SAMPLE_STATS_IN_FNAME;
    opts.sample_stats_out_fname = SAMPLE_STATS_OUT_FNAME;
    opts.samples_out_fname = SAMPLES_OUT_FNAME;
    opts.sample_cov_in_eig_k = SAMPLE_COV_IN_EIG_K;
    opts.sample_cov_in_eig_scale = SAMPLE_COV_IN_EIG_SCALE;
    opts.sample_stats_info_fname = SAMPLE_STATS_INFO_FNAME;
    opts.sample_stats_freq = SAMPLE_STATS_FREQ;
    opts.stoch_alpha = STOCH_ALPHA;
    opts.stoch_kick = STOCH_KICK;
    opts.hybrid_l = HYBRID_L;
    opts.hyper_gamma = HYPER_GAMMA;
    opts.hyper_eta = HYPER_ETA;
    opts.hyper_h = HYPER_H;
    opts.hyper_d = HYPER_D;
    opts.mh_jump_set_size = MH_JUMP_SET_SIZE;
    opts.mh_dynamics_prob = MH_DYNAMICS_PROB;
    opts.mh_diffusion_prob = MH_DIFFUSION_PROB;
    opts.mh_jump_prob = MH_JUMP_PROB;
    opts.num_images = 0;
    opts.images_in_max_num_rows = 0;
    opts.images_in_max_num_cols = 0;
    opts.images_reso_max_num_rows = 0;
    opts.images_reso_max_num_cols = 0;
    opts.view_width = 0;
    opts.view_height = 0;
    opts.reso_width = 0;
    opts.reso_height = 0;
    opts.view_ctx = 0;
    opts.obj_ctx = 0;
    opts.block_ctx = 0;
    opts.block_set_ctx = 0;
    opts.cat_view_in_fname = CAT_VIEW_IN_FNAME;
    opts.cat_object_in_fname = CAT_OBJECT_IN_FNAME;
    opts.num_cat_samples = NUM_CAT_SAMPLES;
    opts.cat_sample_img_width = CAT_SAMPLE_IMG_WIDTH;
    opts.cat_sample_img_height = CAT_SAMPLE_IMG_HEIGHT;
    opts.cat_sample_img_out_fname = CAT_SAMPLE_IMG_OUT_FNAME;
}


/** @brief Initialize the options. */
static void init_furniture_options()
{
    using jwscxx::base::Option_no_arg;
    using jwscxx::base::Option_with_arg;

    char        s_name;
    const char* l_name; 
    const char* desc;

    void (*fnoarg)(void);
    void (*farg)(const char*) throw (jwscxx::base::Arg_error);

    opts.view_ctx      = new View::Context();
    opts.obj_ctx       = new Object::Context();
    opts.block_ctx     = new Block::Context();
    opts.block_set_ctx = new Block_set::Context();
    options            = new jwscxx::base::Options();

    s_name = 'h';
    l_name = "help";
    desc   = "Program usage";
    fnoarg = process_help_opt;
    options->add(new Option_no_arg(s_name, l_name, desc, fnoarg));

    s_name = 'v';
    l_name = "version";
    desc   = "Program version.";
    fnoarg = process_version_opt;
    options->add(new Option_no_arg(s_name, l_name, desc, fnoarg));

    s_name = 0;
    l_name = "options";
    desc   = "Process a file of program options.";
    farg   = process_options_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "seed";
    desc   = "Random seed to use in srand(). Must be a non-negative integer.";
    farg   = process_seed_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "num-threads";
    desc   = "Number of threads of execution to use.";
    farg   = process_num_threads_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "num-categories";
    desc   = "Number of structure categories.";
    farg   = process_num_categories_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "min-num-members";
    desc   = "Minimum number of members a category must have.";
    farg   = process_min_num_members_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 'i';
    l_name = "interactive";
    desc   = "Visualize the program in an interactive windowing environmant.";
    fnoarg = process_interactive_opt;
    options->add(new Option_no_arg(s_name, l_name, desc, fnoarg));

    s_name = 0;
    l_name = "sampler";
    desc   = "Type of sampler to use. Must be one of {metro, stoch, hybrid, lange, hyper}.";
    farg   = process_sampler_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "num-images";
    desc   = "Number of input images to read in using the printf-formated input image name. Must be > 0.";
    farg   = process_num_images_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "prefix-in-dir";
    desc   = "Directory to prefix all input files with.";
    farg   = process_prefix_in_dir_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "prefix-out-dir";
    desc   = "Directory to prefix all output files with. The directory structure is created by the program if necessary.";
    farg   = process_prefix_out_dir_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "instance-dir-fmt";
    desc   = "Printf-formatted prefix for all input and output files per image. For example '%04d'. The integer is for sequential model input and output files. The directory structure is created by the program.";
    farg   = process_instance_dir_fmt_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "category-dir-fmt";
    desc   = "Printf-formatted prefix for all input and output files per category. For example '%02d'. The integer is for sequential model input and output files. The directory structure is created by the program.";
    farg   = process_category_dir_fmt_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "categories-in";
    desc   = "File to write the category membership info to for each image.";
    farg   = process_categories_in_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "categories-out";
    desc   = "File to read the category membership info from for each image.";
    farg   = process_categories_out_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "capture";
    desc   = "Printf-formatted file to write interactive captured view images to.";
    farg   = process_capture_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "sample-info";
    desc   = "File to print sampler information to. Stdout by default.";
    farg   = process_sample_info_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "compress-models";
    desc   = "Compress the model files. Requires zlib.";
    fnoarg = process_compress_models_opt;
    //options->add(new Option_no_arg(s_name, l_name, desc, fnoarg));

    s_name = 0;
    l_name = "compress-info";
    desc   = "Compress the sampler information files. Requires zlib.";
    fnoarg = process_compress_info_opt;
    options->add(new Option_no_arg(s_name, l_name, desc, fnoarg));

    s_name = 0;
    l_name = "iterations";
    desc   = "Number of iterations to run the sampler.";
    farg   = process_iterations_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "resolution";
    desc   = "Fraction of the image to down sample to. Must be in (0,1].";
    farg   = process_resolution_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "pbuffer";
    desc   = "Use pbuffer rendering.";
    fnoarg = process_pbuffer_opt;
    options->add(new Option_no_arg(s_name, l_name, desc, fnoarg));

    s_name = 0;
    l_name = "pixmap";
    desc   = "Use pixmap rendering.";
    fnoarg = process_pixmap_opt;
    options->add(new Option_no_arg(s_name, l_name, desc, fnoarg));

    s_name = 0;
    l_name = "edges-in";
    desc   = "File name to read the edge points for the image from. If not specified, the edges in the image are detected.";
    farg   = process_edges_in_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "edges-out";
    desc   = "File name to write the edge points for the image to.";
    farg   = process_edges_out_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "edge-map";
    desc   = "File name to write the edge point image map for the input image to.";
    farg   = process_edge_map_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "edge-color";
    desc   = "File name to write the colored edge points for the input image to.";
    farg   = process_edge_color_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "edge-sigma";
    desc   = "Sigma for edge detection. Must be > 0.";
    farg   = process_edge_sigma_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "edge-begin";
    desc   = "Begin threshold for edge detection.";
    farg   = process_edge_begin_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "edge-end";
    desc   = "End threshold for edge detection.";
    farg   = process_edge_end_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "edge-pad";
    desc   = "Padding for edge detection.";
    farg   = process_edge_pad_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "img-thresh-out";
    desc   = "File to write the thresholded images to.";
    farg   = process_img_thresh_out_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "img-ll-dist-sigma";
    desc   = "Image log likelihood sigma for the distance between edge points and model points along the model point normal. Must be > 0.";
    farg   = process_img_ll_dist_sigma_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "img-ll-angle-sigma";
    desc   = "Image log likelihood sigma for angles between edge point gradient and model point normal. Must be > 0.";
    farg   = process_img_ll_angle_sigma_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "img-ll-max-dist";
    desc   = "Image log likelihood maximum distance between an edge point and a model point to be considered for correspondence. Must be > 0.";
    farg   = process_img_ll_max_dist_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "img-ll-bg-prob";
    desc   = "Probability for a pixel to be in the background. Must be >= 0.";
    farg   = process_img_ll_bg_prob_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "img-ll-surf-prob";
    desc   = "Probability for a pixel to be in the foreground. Must be >= 0.";
    farg   = process_img_ll_surf_prob_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "img-ll-surf-noise-prob";
    desc   = "Probability of a pixel to be in the foreground, but be noise. Must be >= 0.";
    farg   = process_img_ll_surf_noise_prob_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "img-ll-surf-miss-prob";
    desc   = "Probability of a pixel to be missing from the foreground. Must be >= 0.";
    farg   = process_img_ll_surf_miss_prob_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "img-ll-edge-noise-prob";
    desc   = "Probability of a pixel to be an edge point, but be noise. Must be >= 0.";
    farg   = process_img_ll_edge_noise_prob_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "img-ll-edge-silho-miss-prob";
    desc   = "Probability of a missing silhoutte edge point detection. Must be >= 0.";
    farg   = process_img_ll_edge_silho_miss_prob_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "img-ll-edge-inner-miss-prob";
    desc   = "Probability of a missing inner edge point detection. Must be >= 0.";
    farg   = process_img_ll_edge_inner_miss_prob_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "img-ll-num-angles";
    desc   = "Image log likelihood number of angles to precompute distance and probabilities for.";
    farg   = process_img_ll_num_angles_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "img-ll-cache-size";
    desc   = "Number of elements to allow in the cache.";
    farg   = process_img_ll_cache_size_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "img-ll-cache-k";
    desc   = "Number of cache elements needed with ll_cache_t distance to have a hit.";
    farg   = process_img_ll_cache_k_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "img-ll-cache-t";
    desc   = "Distance a set of ll_cache_k elements have to be from a query to be considered a cache hit.";
    farg   = process_img_ll_cache_t_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "model-pad";
    desc   = "Padding for the model scene image.";
    farg   = process_model_pad_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-in-random";
    desc   = "Generate view model randomly.";
    fnoarg = process_view_in_random_opt;
    options->add(new Option_no_arg(s_name, l_name, desc, fnoarg));

    s_name = 0;
    l_name = "view-in";
    desc   = "View model input file to read from.";
    farg   = process_view_in_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-out";
    desc   = "View model output file to write to.";
    farg   = process_view_out_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "best-view-in";
    desc   = "Best view model input file to read from.";
    farg   = process_best_view_in_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "best-view-out";
    desc   = "Best view model output file to write to.";
    farg   = process_best_view_out_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-distance";
    desc   = "Distance of a view model.";
    farg   = process_view_distance_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-clipping";
    desc   = "Distance to the near and far clipping planes of a view model.";
    farg   = process_view_clipping_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-zenith";
    desc   = "Zenith of a view model.";
    farg   = process_view_zenith_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-f";
    desc   = "Focal length of a view model.";
    farg   = process_view_f_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-scale";
    desc   = "Scale of a view model.";
    farg   = process_view_scale_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-zenith-range";
    desc   = "Zenith of a view model. Has argument format min,max.";
    farg   = process_view_zenith_range_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-f-range";
    desc   = "Focal length of a view model. Has argument format min,max.";
    farg   = process_view_f_range_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-scale-range";
    desc   = "Scale of a view model. Has argument format min,max.";
    farg   = process_view_scale_range_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-zenith-likelihood";
    desc   = "Zenith of a view model. Has argument format mu,sigma.";
    farg   = process_view_zenith_likelihood_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-f-likelihood";
    desc   = "Focal length of a view model. Has argument format mu,sigma.";
    farg   = process_view_f_likelihood_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-scale-likelihood";
    desc   = "Scale of a view model. Has argument format mu,sigma.";
    farg   = process_view_scale_likelihood_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-zenith-proposal";
    desc   = "Zenith of a view model. Has argument format mu,sigma.";
    farg   = process_view_zenith_proposal_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-f-proposal";
    desc   = "Focal length of a view model. Has argument format mu,sigma.";
    farg   = process_view_f_proposal_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-scale-proposal";
    desc   = "Scale of a view model. Has argument format mu,sigma.";
    farg   = process_view_scale_proposal_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-zenith-dynamics";
    desc   = "Zenith of a view model. Has argument format delta_t,grad_eta.";
    farg   = process_view_zenith_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-zenith-mean-dynamics";
    desc   = "Zenith of a view model. Has argument format delta_t,grad_eta.";
    farg   = process_view_zenith_mean_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-zenith-stdev-dynamics";
    desc   = "Zenith of a view model. Has argument format delta_t,grad_eta.";
    farg   = process_view_zenith_stdev_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-f-dynamics";
    desc   = "Focal length of a view model. Has argument format delta_t,grad_eta.";
    farg   = process_view_f_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-f-mean-dynamics";
    desc   = "Focal length of a view model. Has argument format delta_t,grad_eta.";
    farg   = process_view_f_mean_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-f-stdev-dynamics";
    desc   = "Focal length of a view model. Has argument format delta_t,grad_eta.";
    farg   = process_view_f_stdev_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-scale-dynamics";
    desc   = "View scale. Has argument format delta_t,grad_eta.";
    farg   = process_view_scale_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-scale-mean-dynamics";
    desc   = "View scale. Has argument format delta_t,grad_eta.";
    farg   = process_view_scale_mean_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "view-scale-stdev-dynamics";
    desc   = "View scale. Has argument format delta_t,grad_eta.";
    farg   = process_view_scale_stdev_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-in-random";
    desc   = "Generate object model randomly.";
    fnoarg = process_object_in_random_opt;
    options->add(new Option_no_arg(s_name, l_name, desc, fnoarg));

    s_name = 0;
    l_name = "object-in";
    desc   = "Object model input file to read from.";
    farg   = process_object_in_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-out";
    desc   = "Object model output file to write to.";
    farg   = process_object_out_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "best-object-in";
    desc   = "Best object model input file to read from.";
    farg   = process_best_object_in_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "best-object-out";
    desc   = "Best object model output file to write to.";
    farg   = process_best_object_out_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-img";
    desc   = "Object model rendered as edges into an image.";
    farg   = process_object_img_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "best-object-img";
    desc   = "Best object model rendered as edges into an image.";
    farg   = process_best_object_img_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-azimuth";
    desc   = "Object rotation angle.";
    farg   = process_object_azimuth_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-x-posn";
    desc   = "Object position x-corrd.";
    farg   = process_object_x_posn_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-y-posn";
    desc   = "Object position y-coord.";
    farg   = process_object_y_posn_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-z-posn";
    desc   = "Object position z-coord.";
    farg   = process_object_z_posn_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-azimuth-range";
    desc   = "Object rotation angle. Has argument format min,max.";
    farg   = process_object_azimuth_range_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-x-posn-range";
    desc   = "Object position x-corrd. Has argument format min,max.";
    farg   = process_object_x_posn_range_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-y-posn-range";
    desc   = "Object position y-coord. Has argument format min,max.";
    farg   = process_object_y_posn_range_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-z-posn-range";
    desc   = "Object position z-coord. Has argument format min,max.";
    farg   = process_object_z_posn_range_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-x-posn-likelihood";
    desc   = "Object position x-coord. Has argument format mu,sigma.";
    farg   = process_object_x_posn_likelihood_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-y-posn-likelihood";
    desc   = "Object position y-coord. Has argument format mu,sigma.";
    farg   = process_object_y_posn_likelihood_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-z-posn-likelihood";
    desc   = "Object position z-coord. Has argument format mu,sigma.";
    farg   = process_object_z_posn_likelihood_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-azimuth-proposal";
    desc   = "Object rotation angle. Has argument format mu,sigma.";
    farg   = process_object_azimuth_proposal_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-x-posn-proposal";
    desc   = "Object position x-coord. Has argument format mu,sigma.";
    farg   = process_object_x_posn_proposal_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-y-posn-proposal";
    desc   = "Object position y-coord. Has argument format mu,sigma.";
    farg   = process_object_y_posn_proposal_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-z-posn-proposal";
    desc   = "Object position z-coord. Has argument format mu,sigma.";
    farg   = process_object_z_posn_proposal_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-azimuth-dynamics";
    desc   = "Object rotation angle. Has argument format delta_t,grad_eta.";
    farg   = process_object_azimuth_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-azimuth-mean-dynamics";
    desc   = "Object rotation angle. Has argument format delta_t,grad_eta.";
    farg   = process_object_azimuth_mean_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-azimuth-stdev-dynamics";
    desc   = "Object rotation angle. Has argument format delta_t,grad_eta.";
    farg   = process_object_azimuth_stdev_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-posn-dynamics";
    desc   = "Object position dynamics parameters. Has argument format delta_t,grad_eta.";
    farg   = process_object_posn_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-posn-mean-dynamics";
    desc   = "Object position dynamics position. Has argument format delta_t,grad_eta.";
    farg   = process_object_posn_mean_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-posn-stdev-dynamics";
    desc   = "Object position dynamics position. Has argument format delta_t,grad_eta.";
    farg   = process_object_posn_stdev_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-block-w";
    desc   = "Object block width.";
    farg   = process_object_block_w_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-block-h";
    desc   = "Object block height.";
    farg   = process_object_block_h_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-block-l";
    desc   = "Object block length.";
    farg   = process_object_block_l_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-block-w-range";
    desc   = "Object block width. Has argument format min,max.";
    farg   = process_object_block_w_range_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-block-h-range";
    desc   = "Object block height. Has argument format min,max.";
    farg   = process_object_block_h_range_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-block-l-range";
    desc   = "Object block length. Has argument format min,max.";
    farg   = process_object_block_l_range_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-block-w-likelihood";
    desc   = "Object block width. Has argument format mu,sigma.";
    farg   = process_object_block_w_likelihood_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-block-h-likelihood";
    desc   = "Object block height. Has argument format mu,sigma.";
    farg   = process_object_block_h_likelihood_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-block-l-likelihood";
    desc   = "Object block length. Has argument format mu,sigma.";
    farg   = process_object_block_l_likelihood_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-block-size-prior";
    desc   = "Object block size joint prior density. The function is a mixture of Gaussians over the mean and Gamma over the precision. Has argument format N,pi1,w_mu1,h_mu1,l_mu1,...,piN,w_muN,h_muN,l_muN,lambda,alpha,beta, where N is the number of mixture component. The weights, pi, are normalized to sum to 1.";
    farg   = process_object_block_size_prior_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-block-attach-prior";
    desc   = "Object block variable attachment point joint prior density. The function is a Gaussian-Gamma distribution with zero mean. Has argument format lambda,alpha,beta. All must be > 0.";
    farg   = process_object_block_attach_prior_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-block-w-proposal";
    desc   = "Object block width. Has argument format mu,sigma.";
    farg   = process_object_block_w_proposal_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-block-h-proposal";
    desc   = "Object block height. Has argument format mu,sigma.";
    farg   = process_object_block_h_proposal_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-block-l-proposal";
    desc   = "Object block length. Has argument format mu,sigma.";
    farg   = process_object_block_l_proposal_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-size-dynamics";
    desc   = "Object component size dynamics parameters. Has argument format delta_t,grad_eta.";
    farg   = process_object_size_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-size-mean-dynamics";
    desc   = "Object component size dynamics parameters. Has argument format delta_t,grad_eta.";
    farg   = process_object_size_mean_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-size-stdev-dynamics";
    desc   = "Object component size dynamics parameters. Has argument format delta_t,grad_eta.";
    farg   = process_object_size_stdev_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-ratio-size-mean-dynamics";
    desc   = "Object component size dynamics parameters. Has argument format delta_t,grad_eta.";
    farg   = process_object_ratio_size_mean_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "object-ratio-size-stdev-dynamics";
    desc   = "Object component size dynamics parameters. Has argument format delta_t,grad_eta.";
    farg   = process_object_ratio_size_stdev_dynamics_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "sample-stats-in";
    desc   = "File name to read the sample statistics from.";
    farg   = process_sample_stats_in_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "sample-stats-out";
    desc   = "File name to write the sample statistics to.";
    farg   = process_sample_stats_out_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "samples-out";
    desc   = "File name to write the generated samples to.";
    farg   = process_samples_out_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "sample-cov-in-eig-k";
    desc   = "Number of largest eigenvalues to scale in the input sample covariance matrix.";
    farg   = process_sample_cov_in_eig_k_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "sample-cov-in-eig-scale";
    desc   = "Scale to use for the k largest eigenvalues from the input sample covariance matrix.";
    farg   = process_sample_cov_in_eig_scale_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "sample-stats-info";
    desc   = "File to write sample covariance matrix information to.";
    farg   = process_sample_stats_info_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "sample-stats-freq";
    desc   = "Freqency of estimating the sample covariance matrix.";
    farg   = process_sample_stats_freq_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "stoch-alpha";
    desc   = "Controls the stochastic step size in the dynamics. A value close to zero yields completely random step direction; close to one follows the dynamics almost exactly. Must be in [0,1].";
    farg   = process_stoch_alpha_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "stoch-kick";
    desc   = "Freqency (in iterations) to kick the particle by resetting its momenta to a non-stochastic value (alpha=1).";
    farg   = process_stoch_kick_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "hybrid-l";
    desc   = "Number of leap-frog iterations in the Hybrid Monte Carlo sampler. Must be > 0.";
    farg   = process_hybrid_l_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "hyper-gamma";
    desc   = "Largest size of step to take during gradient descent in hyperdynamics. Must be > 0.";
    farg   = process_hyper_gamma_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "hyper-eta";
    desc   = "Size of finite differences step in hyperdynamics minimization. Must be > 0.";
    farg   = process_hyper_eta_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "hyper-h";
    desc   = "Hyperdynamics h parameter. Must be > 0.";
    farg   = process_hyper_h_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "hyper-d";
    desc   = "Hyperdynamics d parameter. Must be > 0.";
    farg   = process_hyper_d_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "mh-jump-set-size";
    desc   = "Number of model instances in Metropolis-Hastings jump moves to use for computing the acceptance probability. Set to zero to use all of them.";
    farg   = process_mh_jump_set_size_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "mh-dynamics-prob";
    desc   = "Probability of a stochastic dynamics move in the Metropolis sampler. Must be in [0,1].";
    farg   = process_mh_dynamics_prob_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "mh-diffusion-prob";
    desc   = "Probability of a diffusion move in the Metropolis sampler. Must be in [0,1].";
    farg   = process_mh_diffusion_prob_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "mh-jump-prob";
    desc   = "Probability of a jump move in the Metropolis sampler. Must be in [0,1].";
    farg   = process_mh_jump_prob_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));
}


/** @brief Initialize the options. */
static void init_furniture_cat_options()
{
    using jwscxx::base::Option_no_arg;
    using jwscxx::base::Option_with_arg;

    char        s_name;
    const char* l_name; 
    const char* desc;

    void (*fnoarg)(void);
    void (*farg)(const char*) throw (jwscxx::base::Arg_error);

    opts.view_ctx      = new View::Context();
    opts.obj_ctx       = new Object::Context();
    opts.block_ctx     = new Block::Context();
    opts.block_set_ctx = new Block_set::Context();
    options            = new jwscxx::base::Options();

    s_name = 'h';
    l_name = "help";
    desc   = "Program usage";
    fnoarg = process_help_opt;
    options->add(new Option_no_arg(s_name, l_name, desc, fnoarg));

    s_name = 'v';
    l_name = "version";
    desc   = "Program version.";
    fnoarg = process_version_opt;
    options->add(new Option_no_arg(s_name, l_name, desc, fnoarg));

    s_name = 0;
    l_name = "options";
    desc   = "Process a file of program options.";
    farg   = process_options_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "seed";
    desc   = "Random seed to use in srand(). Must be a non-negative integer.";
    farg   = process_seed_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "pbuffer";
    desc   = "Use pbuffer rendering.";
    fnoarg = process_pbuffer_opt;
    options->add(new Option_no_arg(s_name, l_name, desc, fnoarg));

    s_name = 0;
    l_name = "pixmap";
    desc   = "Use pixmap rendering.";
    fnoarg = process_pixmap_opt;
    options->add(new Option_no_arg(s_name, l_name, desc, fnoarg));

    s_name = 0;
    l_name = "cat-view-in";
    desc   = "View model constaining the category parameters to use.";
    farg   = process_cat_view_in_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "cat-object-in";
    desc   = "Object model constaining the category parameters to use.";
    farg   = process_cat_object_in_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "num-cat-samples";
    desc   = "Number of samples to draw from the category.";
    farg   = process_num_cat_samples_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "cat-sample-img-size";
    desc   = "Size of the category sample image to generate. Has format width,height.";
    farg   = process_cat_sample_img_size_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));

    s_name = 0;
    l_name = "cat-sample-img-out";
    desc   = "Name of the image to render the category sampled models to.";
    farg   = process_cat_sample_img_out_opt;
    options->add(new Option_with_arg(s_name, l_name, desc, farg));
}


static bool is_fname_special(const char* fname)
{
    if (strncmp(fname, "/dev/null", 9) == 0 ||
        strncmp(fname, "/dev/stdin", 10) == 0 ||
        strncmp(fname, "/dev/stdout", 11) == 0 ||
        strncmp(fname, "/dev/stderr", 11) == 0)
    {
        return true;
    }

    return false;
}


static const char* printf_int_fmt
(
    const char*  fmt, 
    uint32_t     i
)
{
    static char chars[1024] = {0};

    snprintf(chars, 1024, fmt, i);

    return chars;
}


static void printf_category_out
(
    std::string& fname_out, 
    uint32_t     c, 
    const char*  fname_in
)
{
    fname_out.clear();

    if (is_fname_special(fname_in))
    {
        fname_out.assign(fname_in);
        return;
    }

    fname_out.assign(opts.prefix_out_dir); fname_out += '/';
#if defined FURNITURE_HAVE_OPENMPI
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    fname_out += printf_int_fmt("%02u", rank+1); fname_out += '/';
#endif
    fname_out += printf_int_fmt(opts.category_dir_fmt, c+1); fname_out += '/';

    std::string script("mkdir -p ");
    script += fname_out;

    if (system(script.c_str()) < 0)
    {
        std::ostringstream ost;
        ost << fname_out << ": Could not create directory";
        throw jwscxx::base::Arg_error(ost.str());
    }

    fname_out += fname_in;
}


static void printf_category_instance_in
(
    std::string& fname_out,
    uint32_t     c, 
    uint32_t     i,
    const char*  fname_in
)
{
    fname_out.clear();

    if (is_fname_special(fname_in))
    {
        fname_out.assign(fname_in);
        return;
    }

    fname_out.assign(opts.prefix_in_dir); fname_out += '/';
#if defined FURNITURE_HAVE_OPENMPI
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    fname_out += printf_int_fmt("%02u", rank+1); fname_out += '/';
#endif
    fname_out += printf_int_fmt(opts.category_dir_fmt, c+1); fname_out += '/';
    fname_out += printf_int_fmt(opts.instance_dir_fmt, i+1); fname_out += '/';
    fname_out += fname_in;
}


static void printf_category_instance_out
(
    std::string& fname_out,
    uint32_t     c, 
    uint32_t     i,
    const char*  fname_in
)
{
    fname_out.clear();

    if (is_fname_special(fname_in))
    {
        fname_out.assign(fname_in);
        return;
    }

    fname_out.assign(opts.prefix_out_dir); fname_out += '/';
#if defined FURNITURE_HAVE_OPENMPI
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    fname_out += printf_int_fmt("%02u", rank+1); fname_out += '/';
#endif
    fname_out += printf_int_fmt(opts.category_dir_fmt, c+1); fname_out += '/';
    fname_out += printf_int_fmt(opts.instance_dir_fmt, i+1); fname_out += '/';

    std::string script("mkdir -p ");
    script += fname_out;

    if (system(script.c_str()) < 0)
    {
        std::ostringstream ost;
        ost << fname_out << ": Could not create directory";
        throw jwscxx::base::Arg_error(ost.str());
    }

    fname_out += fname_in;
}


/** @brief Reads the input images. */
static void read_images(int argc, const char** argv, int argi)
    throw (jwscxx::base::Arg_error)
{
    using namespace jwsc;
    using jwscxx::base::Arg_error;

    std::string fname;
    Error* err;
    std::ostringstream ost;

    if ((argc - argi) < 1)
    {
        throw Arg_error("No input images");
    }

    opts.images_in_max_num_rows = 0;
    opts.images_in_max_num_cols = 0;

    if (opts.num_images)
    {
        for (uint32_t i = 0; i < opts.num_images; i++)
        {
            fname.assign(printf_int_fmt(argv[ argi ], i+1));
            Image_f* image = 0;

            if ((err = read_image_f(&image, fname.c_str())) != 0)
            {
                ost << fname << ": " << err->msg;
                free_error(err);
                throw Arg_error(ost.str());
            }

            opts.images_in.push_back(image);
            if (image->num_rows > opts.images_in_max_num_rows)
            {
                opts.images_in_max_num_rows = image->num_rows;
            }
            if (image->num_cols > opts.images_in_max_num_cols)
            {
                opts.images_in_max_num_cols = image->num_cols;
            }

            Image_f* full_image = 0;
            copy_image_f(&full_image, image);
            opts.full_images_in.push_back(full_image);
        }
    }
    else
    {
        opts.num_images = argc - argi;

        for (uint32_t i = 0; i < opts.num_images; i++)
        {
            fname = argv[ argi + i ];
            Image_f* image = 0;

            if ((err = read_image_f(&image, fname.c_str())) != 0)
            {
                ost << fname << ": " << err->msg;
                free_error(err);
                throw Arg_error(ost.str());
            }

            opts.images_in.push_back(image);
            if (image->num_rows > opts.images_in_max_num_rows)
            {
                opts.images_in_max_num_rows = image->num_rows;
            }
            if (image->num_cols > opts.images_in_max_num_cols)
            {
                opts.images_in_max_num_cols = image->num_cols;
            }

            Image_f* full_image = 0;
            copy_image_f(&full_image, image);
            opts.full_images_in.push_back(full_image);
        }
    }

    if (opts.min_num_members*opts.num_categories > opts.num_images)
    {
        throw Arg_error("Cluster membership requirements too large");
    }

    opts.view_width  = opts.images_in_max_num_cols + 2*opts.model_pad;
    opts.view_height = opts.images_in_max_num_rows + 2*opts.model_pad;
}


static void threshold_images()
{
    using namespace jwsc;

    for (uint32_t i = 0; i < opts.num_images; i++)
    {
        Matrix_f* mat = 0;
        create_matrix_from_image_f(&mat, opts.images_in[ i ], 
                0.3, 0.59, 0.11);
        dynamically_threshold_matrix_f(&mat, mat);
        opts.images_t.push_back(mat);
    }
}


/** @brief Writes the thresholded image. */
static void write_threshold_images
(
    Sampler::Parameters* params, 
    uint32_t             c, 
    uint32_t             i
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
{
    using namespace jwsc;

    std::string fname;
    //Error* err;

    if (opts.img_thresh_out_fname)
    {
        jwsc::Image_f* model_img = 0;
        Sampler::Instance* ins = params->instance(c, i);

        glEnable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glClearColor(1, 1, 1, 0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        ins->view->set_gl_projection();
        ins->view->set_gl_modelview();

        if (!opts.interactive)
        {
            int w = opts.view_width;
            int h = opts.view_height;
            float s = (float)h / (float)opts.reso_height;
            glPointSize(s+0.49999);
            glLineWidth(s);
            glViewport(0, 0, (GLsizei)w, (GLsizei)h); 
        }

        glColor3f(0, 0, 0);
        Render::thresh_image(i, ins->view);
        glFlush();

        View::capture_gl_view(&model_img);
        printf_category_instance_out(fname, c, i, opts.img_thresh_out_fname);

        jwsc::write_image_f(model_img, fname.c_str());
        jwsc::free_image_f(model_img);

        /*
        Image_f* img = 0;

        for (uint32_t i = 0; i < opts.num_images; i++)
        {
            create_image_from_matrix_f(&img, opts.images_t[ i ]);

            printf_category_instance_out(fname, 0, i, 
                    opts.img_thresh_out_fname);
            if ((err = write_image_f(img, fname.c_str())) != 0)
            {
                free_image_f(img);
                std::ostringstream ost;
                ost << fname << ": " << err->msg;
                free_error(err);
                throw jwscxx::base::IO_error(ost.str());
            }

        }

        free_image_f(img);
        */
    }
}


/** @brief Reads or detects the image edges. */
static void read_or_detect_edges()
    throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
{
    using namespace jwsc;
    using jwscxx::base::IO_error;
    using jwscxx::base::Arg_error;

    std::string fname;
    Error* err;

    opts.images_reso_max_num_rows = 0;
    opts.images_reso_max_num_cols = 0;

    if (opts.edge_pts_in_fname)
    {
        if (opts.resolution < 1.0f)
        {
            throw Arg_error("Cannot read in edge points and change resolution");
        }

        for (uint32_t i = 0; i < opts.num_images; i++)
        {
            printf_category_instance_in(fname, 0, i, opts.edge_pts_in_fname);
            Edge_set_f* edge_set = 0;

            if ((err = read_edge_set_f(&edge_set, fname.c_str())) != 0)
            {
                std::ostringstream ost;
                ost << fname << ": " << err->msg;
                free_error(err);
                throw IO_error(ost.str());
            }

            opts.edges.push_back(edge_set);
            if (opts.images_in[ i ]->num_rows > opts.images_reso_max_num_rows)
            {
                opts.images_reso_max_num_rows = opts.images_in[ i ]->num_rows;
            }
            if (opts.images_in[ i ]->num_cols > opts.images_reso_max_num_cols)
            {
                opts.images_reso_max_num_cols = opts.images_in[ i ]->num_cols;
            }
        }
    }
    else
    {
        Matrix_f* mat = 0;

        for (uint32_t i = 0; i < opts.num_images; i++)
        {
            Edge_set_f* edge_set = 0;

            down_sample_image_f(&(opts.images_in[ i ]), opts.images_in[ i ],
                    opts.resolution, opts.resolution);
            create_matrix_from_image_f(&mat, opts.images_in[ i ], 
                    0.3, 0.59, 0.11);
            detect_matrix_edge_set_f(&edge_set, mat, opts.edge_sigma, 
                    opts.edge_begin, opts.edge_end, opts.edge_pad);

            opts.edges.push_back(edge_set);
            if (opts.images_in[ i ]->num_rows > opts.images_reso_max_num_rows)
            {
                opts.images_reso_max_num_rows = opts.images_in[ i ]->num_rows;
            }
            if (opts.images_in[ i ]->num_cols > opts.images_reso_max_num_cols)
            {
                opts.images_reso_max_num_cols = opts.images_in[ i ]->num_cols;
            }
        }

        free_matrix_f(mat);
    }

    opts.reso_width  = opts.images_reso_max_num_cols + 2*opts.reso_model_pad;
    opts.reso_height = opts.images_reso_max_num_rows + 2*opts.reso_model_pad;
}


/** @brief Writes the edges. */
static void write_edges() 
    throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
{
    using namespace jwsc;

    std::string fname;
    Error* err;

    if (opts.edge_pts_out_fname)
    {
        for (uint32_t i = 0; i < opts.num_images; i++)
        {
            printf_category_instance_out(fname, 0, i, opts.edge_pts_out_fname);
            if ((err = write_edge_set_f(opts.edges[ i ], fname.c_str())) != 0)
            {
                std::ostringstream ost;
                ost << fname << ": " << err->msg;
                free_error(err);
                throw jwscxx::base::IO_error(ost.str());
            }
        }
    }
}


/** @brief Writes the edges. */
static void write_edge_map(Sampler::Parameters* params, uint32_t c, uint32_t i) 
    throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
{
    using namespace jwsc;

    std::string fname;
    //Error* err;

    if (opts.edge_map_fname)
    {
        jwsc::Image_f* model_img = 0;
        Sampler::Instance* ins = params->instance(c, i);

        glEnable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glClearColor(1, 1, 1, 0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        ins->view->set_gl_projection();
        ins->view->set_gl_modelview();

        if (!opts.interactive)
        {
            int w = opts.view_width;
            int h = opts.view_height;
            float s = (float)h / (float)opts.reso_height;
            glPointSize(s+0.49999);
            glLineWidth(s);
            glViewport(0, 0, (GLsizei)w, (GLsizei)h); 
        }

        glColor3f(0, 0, 0);
        Render::edges(i, ins->view);
        glFlush();

        View::capture_gl_view(&model_img);
        printf_category_instance_out(fname, c, i, opts.edge_map_fname);

        jwsc::write_image_f(model_img, fname.c_str());
        jwsc::free_image_f(model_img);

        /*
        Matrix_f* mat = 0;
        Image_f* img = 0;

        for (uint32_t i = 0; i < opts.num_images; i++)
        {
            create_init_matrix_f(&mat, opts.images_in[ i ]->num_rows, 
                    opts.images_in[ i ]->num_cols, 1.0);

            for (uint32_t e = 0; e < opts.edges[ i ]->num_edges; e++)
            {
                for (uint32_t p = 0; p < opts.edges[ i ]->num_pts[ e ]; p++)
                {
                    mat->elts[ opts.edges[ i ]->pts[ e ][ p ].row ]
                             [ opts.edges[ i ]->pts[ e ][ p ].col ] = 0.0f;
                }
            }

            create_image_from_matrix_f(&img, mat);

            printf_category_instance_out(fname, 0, i, opts.edge_map_fname);
            if ((err = write_image_f(img, fname.c_str())) != 0)
            {
                free_image_f(img);
                std::ostringstream ost;
                ost << fname << ": " << err->msg;
                free_error(err);
                throw jwscxx::base::IO_error(ost.str());
            }

        }

        free_matrix_f(mat);
        free_image_f(img);
        */
    }
}


/** @brief Writes the edges colored into an image. */
static void write_edge_color() 
    throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
{
    using namespace jwsc;

    std::string fname;
    Error* err;

    if (opts.edge_color_fname)
    {
        Image_f* img = 0;

        for (uint32_t i = 0; i < opts.num_images; i++)
        {
            randomly_color_edge_set_f(&img, opts.images_in[ i ], 
                    opts.edges[ i ]);

            printf_category_instance_out(fname, 0, i, opts.edge_color_fname);
            if ((err = write_image_f(img, fname.c_str())) != 0)
            {
                free_image_f(img);
                std::ostringstream ost;
                ost << fname << ": " << err->msg;
                free_error(err);
                throw jwscxx::base::IO_error(ost.str());
            }
        }

        free_image_f(img);
    }
}


/** @brief Applies the model padding to the edges. */
static void pad() throw (jwscxx::base::IO_error)
{
    using namespace jwsc;

    Matrix_f* m = 0;

    // Add the resolution updated model padding to the edge points
    for (uint32_t i = 0; i < opts.num_images; i++)
    {
        uint32_t num_rows = opts.images_in[ i ]->num_rows;
        uint32_t num_cols = opts.images_in[ i ]->num_cols;

        uint32_t r = (uint32_t)floor(0.5*(opts.images_reso_max_num_rows -
                    num_rows));
        uint32_t c = (uint32_t)floor(0.5*(opts.images_reso_max_num_cols -
                    num_cols));

        opts.images_in_row_pad.push_back(opts.reso_model_pad + r);
        opts.images_in_col_pad.push_back(opts.reso_model_pad + c);

        copy_matrix_f(&m, opts.images_t[ i ]);
        create_init_matrix_f(&(opts.images_t[ i ]), opts.reso_height,
                opts.reso_width, 1.0);
        copy_matrix_block_into_matrix_f(opts.images_t[ i ], 
                opts.images_in_row_pad[i], opts.images_in_col_pad[i],
                m, 0, 0, num_rows, num_cols);

        for (uint32_t e = 0; e < opts.edges[ i ]->num_edges; e++)
        {
            for (uint32_t p = 0; p < opts.edges[ i ]->num_pts[ e ]; p++)
            {
                opts.edges[ i ]->pts[ e ][ p ].row += opts.images_in_row_pad[i];
                opts.edges[ i ]->pts[ e ][ p ].col += opts.images_in_col_pad[i];
            }
        }
    }

    free_matrix_f(m);
}


/** @brief Create and initize the sampler moves. */
static void add_metropolis_moves(Metropolis* sampler) 
    throw (jwscxx::base::Arg_error)
{
    double s = opts.mh_dynamics_prob;
    double d = opts.mh_diffusion_prob;
    double j = opts.mh_jump_prob;

    sampler->add_diffusion(Metropolis::Dynamics_cat("Dynamics",
                opts.stoch_alpha, opts.stoch_kick, opts.hybrid_l, s));

    sampler->add_diffusion(Metropolis::Swap("Swap", d));
    sampler->add_diffusion(Metropolis::MV_gauss_shift("MVShift", d));
    sampler->add_diffusion(Metropolis::View_gauss("View", d));
    sampler->add_diffusion(Metropolis::Object_gauss_shift("ObjShift", d));
    sampler->add_diffusion(Metropolis::Component_gauss_shift("CmpShift", d));
    sampler->add_diffusion(Metropolis::MV_gauss_scale("MVScale", d));
    sampler->add_diffusion(Metropolis::Object_gauss_scale("ObjScale", d));
    sampler->add_diffusion(Metropolis::Component_gauss_scale("CmpScale", d));
    sampler->add_diffusion(Metropolis::Diffusion_cat("Category", d));
    sampler->add_diffusion(Metropolis::View_cat("ViewCat", d));
    sampler->add_diffusion(Metropolis::Object_cat("ObjCat", d));
    sampler->add_diffusion(Metropolis::Component_cat("CmpCat", d));

    sampler->add_jump(Metropolis::Block_birth_death("BlkBirth", 
                "BlkDeath", j, j));
    sampler->add_jump(Metropolis::Block_block_birth_death("BlkBirth", 
                "BlkDeath", j, j));
    sampler->add_jump(Metropolis::Block_pair_birth_death("BlkBirth", 
                "BlkDeath", j, j));
    sampler->add_jump(Metropolis::Block_quartet_birth_death("BlkBirth", 
                "BlkDeath", j, j));

    sampler->add_jump(Metropolis::Block_block_split_merge("BlkSplit", 
                "BlkMerge", j, j));
    sampler->add_jump(Metropolis::Block_pair_split_merge("BlkSplit", 
                "BlkMerge", j, j));
    sampler->add_jump(Metropolis::Pair_block_split_merge("PairSplit", 
                "PairMerge", j, j));
    sampler->add_jump(Metropolis::Block_quartet_split_merge("BlkSplit", 
                "BlkMerge", j, j));
    sampler->add_jump(Metropolis::Quartet_block_split_merge("QrtSplit", 
                "QrtMerge", j, j));

    sampler->add_jump(Metropolis::Quartet_birth_death("QrtBirth",
            "QrtDeath", j, j));
    sampler->add_jump(Metropolis::Quartet_block_birth_death("QrtBirth",
            "QrtDeath", j, j));

    sampler->add_jump(Metropolis::Pair_birth_death("PairBirth",
            "PairDeath", j, j));
    sampler->add_jump(Metropolis::Pair_block_birth_death("PairBirth",
            "PairDeath", j, j));

    sampler->add_jump(Metropolis::Block_pair_switch("BlkToPair", 
                "PairToBlk", j, j));
    sampler->add_jump(Metropolis::Block_quartet_switch("BlkToQrt", 
                "QrtToBlk", j, j));
    sampler->add_jump(Metropolis::Pair_quartet_switch("PairToQrt", 
                "QrtToPair", j, j));

    sampler->add_diffusion(Metropolis::Cross("Cross", d));
    sampler->add_jump(Metropolis::Cross_birth_death("CrossBirth",
            "CrossDeath", j, j));
}


/** @brief Create and initize the recognition sampler moves. */
static void add_rec_metropolis_moves(Metropolis* sampler) 
    throw (jwscxx::base::Arg_error)
{
    //double s = opts.mh_dynamics_prob;
    double d = opts.mh_diffusion_prob;

    /*
    sampler->add_diffusion(Metropolis::Dynamics_cat("Dynamics",
                opts.stoch_alpha, opts.stoch_kick, opts.hybrid_l, s));
    sampler->add_diffusion(Metropolis::Dynamics_cat("DynShort",
                opts.stoch_alpha, opts.stoch_kick, opts.hybrid_l/5, 10*s));
                */
    
    //sampler->add_diffusion(Metropolis::Swap("Swap", d));
    sampler->add_diffusion(Metropolis::MV_gauss_shift("MVShift", d));
    sampler->add_diffusion(Metropolis::View_gauss("View", d));
    sampler->add_diffusion(Metropolis::Object_gauss_shift("ObjShift", d));
    sampler->add_diffusion(Metropolis::Component_gauss_shift("CmpShift", d));
    sampler->add_diffusion(Metropolis::MV_gauss_scale("MVScale", d));
    sampler->add_diffusion(Metropolis::Object_gauss_scale("ObjScale", d));
    sampler->add_diffusion(Metropolis::Component_gauss_scale("CmpScale", d));
    /*
    sampler->add_diffusion(Metropolis::Diffusion_cat("Category", d));
    sampler->add_diffusion(Metropolis::View_cat("ViewCat", d));
    sampler->add_diffusion(Metropolis::Object_cat("ObjCat", d));
    sampler->add_diffusion(Metropolis::Component_cat("CmpCat", d));
    */
    //sampler->add_diffusion(Metropolis::Recognize("Recog", 10*d));
    sampler->add_diffusion(Metropolis::Recognize("Recog", d));
}


static void scale_largest_sample_cov_in_eigenvalues(jwsc::Matrix_d* cov)
{
    using namespace jwsc;

    if ((fabs(opts.sample_cov_in_eig_scale - 1.0) < 1.0e-8) ||
         opts.sample_cov_in_eig_k == 0)
    {
        return;
    }

    Matrix_d* eigvec      = 0;
    Vector_d* eigval      = 0;
    Matrix_d* eigval_diag = 0;

    assert(solve_real_symmetric_matrix_eigenproblem_d(&eigvec, &eigval, 
                cov) == 0);

    uint32_t D = eigval->num_elts;
    uint32_t d, dd;
    uint32_t n;
    double   v;

    assert(D >= opts.sample_cov_in_eig_k);

    for (d = 0; d < D; d++)
    {
        v = eigval->elts[ d ];
        n = 0;
        for (dd = 0; dd < D; dd++)
        {
            if (v <= eigval->elts[ dd ])
            {
                n++;
            }
        }

        if (n == opts.sample_cov_in_eig_k)
            break;
    }
    for (d = 0; d < D; d++)
    {
        if (v <= eigval->elts[ d ])
        {
            eigval->elts[ d ] *= opts.sample_cov_in_eig_scale;
        }
    }

    create_diagonal_matrix_d(&eigval_diag, eigval);
    multiply_matrices_d(&cov, eigvec, eigval_diag);
    transpose_matrix_d(&eigvec, eigvec);
    multiply_matrices_d(&cov, cov, eigvec);

    free_vector_d(eigval);
    free_matrix_d(eigvec);
    free_matrix_d(eigval_diag);
}


/** @brief Reads sample-estimated statistics for the posterior. */
static void read_sample_stats
(
    jwsc::Vector_d** mean_out, 
    jwsc::Matrix_d** cov_out,
    uint32_t         c,
    uint32_t         i
) 
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
{
    using namespace jwsc;
    using jwscxx::base::IO_error;

    if (!opts.sample_stats_in_fname)
        return;

    FILE*  fp;
    Error* err;

    std::string fname;
    printf_category_instance_in(fname, c, i, opts.sample_stats_in_fname);

    if (!(fp = fopen(fname.c_str(), "r")))
    {
        std::ostringstream ost;
        ost << fname << ": Could not open file";
        throw IO_error(ost.str());
    }

    if ((err = read_vector_with_header_fp_d(mean_out, fp)) ||
        (err = read_matrix_with_header_fp_d(cov_out, fp)))
    {
        IO_error ex = IO_error(err->msg);
        free_error(err);
        throw ex;
    }

    fclose(fp);

    scale_largest_sample_cov_in_eigenvalues(*cov_out);
}


/** @brief Initializes a Instance from a file or randomly. */
static Sampler::Instance* read_or_sample_instance
(
    View*                       view, 
    Object*                     obj,
    Likelihood::Correspondence* ll_corr,
    uint32_t                    c,
    uint32_t                    i
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;
    using jwscxx::base::Readable;

    std::string fname_1;
    std::string fname_2;

    View*   best_view = view->clone();
    Object* best_obj  = obj->clone();

    Vector_d* means  = 0;
    Vector_d* stdevs = 0;

    if (opts.view_in_fname)
    {
        printf_category_instance_in(fname_1, c, i, opts.view_in_fname);
        ((Model*)view)->Readable::read(fname_1.c_str());
    }
    else if (opts.view_in_random)
    {
        view->sample_var_param_prior(&means, &stdevs);
        try { view->set_var_param_stats(means, stdevs); }
        catch (jwscxx::base::Arg_error e) { }

        view->zenith(view->zenith_p()->context()->sample_likelihood());
        view->focal(view->focal_p()->context()->sample_likelihood());
        view->scale(view->scale_p()->context()->sample_likelihood());
    }

    view->size(opts.images_in_max_num_cols + 2*opts.model_pad,
               opts.images_in_max_num_rows + 2*opts.model_pad);

    if (opts.best_view_in_fname)
    {
        printf_category_instance_in(fname_1, c, i, opts.best_view_in_fname);
        ((Model*)best_view)->Readable::read(fname_1.c_str());
    }
    else
    {
        *best_view = *view;
    }

    best_view->size(opts.images_in_max_num_cols + 2*opts.model_pad,
                    opts.images_in_max_num_rows + 2*opts.model_pad);

    if (opts.object_in_fname)
    {
        printf_category_instance_in(fname_1, c, i, opts.object_in_fname);
        obj->Readable::read(fname_1.c_str());
    }
    else if (opts.object_in_random)
    {
        obj->sample_var_param_prior(&means, &stdevs);
        try { obj->set_var_param_stats(means, stdevs); }
        catch (jwscxx::base::Arg_error e) { }

        obj->azimuth(obj->azimuth_p()->context()->sample_likelihood());
        obj->posn_shift_deps(0, obj->posn_p(0)->context()->sample_likelihood());
        obj->posn_shift_deps(1, obj->posn_p(1)->context()->sample_likelihood());
        obj->posn_shift_deps(2, obj->posn_p(2)->context()->sample_likelihood());
    }

    if (opts.best_object_in_fname)
    {
        printf_category_instance_in(fname_1, c, i, opts.best_object_in_fname);
        best_obj->Readable::read(fname_1.c_str());
    }
    else
    {
        *best_obj = *obj;
    }

    free_vector_d(means);
    free_vector_d(stdevs);

    Likelihood::Cache* ll_cache = new Likelihood::Cache(opts.img_ll_cache_size,
            opts.img_ll_cache_k, opts.img_ll_cache_t);

    Matrix_u32* surf_ll_assign = 0;
    Matrix_u32* edge_ll_assign = 0;
    Matrix_u32* model_mat = 0;
    Vector_d* sample_mean_in = 0;
    Matrix_d* sample_cov_in = 0;

    double best_potential, best_img_ll;
    double potential, img_ll;

    Likelihood::calc_neg_log_likelihood(&best_potential, &best_img_ll,
            &model_mat, &surf_ll_assign, &edge_ll_assign, opts.images_t[ i ],
            opts.edges[ i ], ll_corr, best_view, best_obj, opts.img_ll_bg_prob,
            opts.img_ll_surf_prob, opts.img_ll_surf_noise_prob,
            opts.img_ll_surf_miss_prob, opts.img_ll_edge_noise_prob,
            opts.img_ll_edge_silho_miss_prob, opts.img_ll_edge_inner_miss_prob);

    Likelihood::calc_neg_log_likelihood(&potential, &img_ll,
            &model_mat, &surf_ll_assign, &edge_ll_assign, opts.images_t[ i ],
            opts.edges[ i ], ll_corr, view, obj, opts.img_ll_bg_prob,
            opts.img_ll_surf_prob, opts.img_ll_surf_noise_prob,
            opts.img_ll_surf_miss_prob, opts.img_ll_edge_noise_prob,
            opts.img_ll_edge_silho_miss_prob, opts.img_ll_edge_inner_miss_prob);

    read_sample_stats(&sample_mean_in, &sample_cov_in, c, i);

    printf_category_instance_out(fname_1, c, i, opts.sample_info_fname);
    printf_category_instance_out(fname_2, c, i, opts.sample_stats_info_fname);

    Sampler::Instance* ins;

    switch (opts.sampler)
    {
        case METRO:
            ins = new Metropolis::Instance(best_view, best_obj,
                    best_img_ll, best_potential, view, obj, img_ll, potential,
                    opts.images_t[ i ], opts.edges[ i ], ll_corr, ll_cache,
                    surf_ll_assign, edge_ll_assign, model_mat, sample_mean_in,
                    sample_cov_in, fname_1.c_str(), opts.compress_info,
                    fname_2.c_str());
            break;
        case STOCH:
            ins = new Stochastic_dynamics::Instance(best_view, best_obj,
                    best_img_ll, best_potential, view, obj, img_ll, potential,
                    opts.images_t[ i ], opts.edges[ i ], ll_corr, ll_cache,
                    surf_ll_assign, edge_ll_assign, model_mat, sample_mean_in,
                    sample_cov_in, fname_1.c_str(), opts.compress_info, 
                    fname_2.c_str());
            break;
        case HYBRID:
            ins = new Hybrid_MC::Instance(best_view, best_obj,
                    best_img_ll, best_potential, view, obj, img_ll, potential,
                    opts.images_t[ i ], opts.edges[ i ], ll_corr, ll_cache,
                    surf_ll_assign, edge_ll_assign, model_mat, sample_mean_in,
                    sample_cov_in, fname_1.c_str(), opts.compress_info,
                    fname_2.c_str());
            break;
        case LANGE:
            ins = new Langevin::Instance(best_view, best_obj, best_img_ll,
                    best_potential, view, obj, img_ll, potential, 
                    opts.images_t[ i ], opts.edges[ i ], ll_corr, ll_cache,
                    surf_ll_assign, edge_ll_assign, model_mat, sample_mean_in,
                    sample_cov_in, fname_1.c_str(), opts.compress_info,
                    fname_2.c_str());
            break;
        case HYPER:
            ins = new Hyperdynamics::Instance(best_view, best_obj, 
                    best_img_ll, best_potential, view, obj, img_ll, potential,
                    opts.images_t[ i ], opts.edges[ i ], ll_corr, ll_cache,
                    surf_ll_assign, edge_ll_assign, model_mat, sample_mean_in,
                    sample_cov_in, fname_1.c_str(), opts.compress_info,
                    fname_2.c_str());
            break;
    }

    return ins;
}


static Sampler::Instance* sample_instance
(
    View*                       view, 
    Object*                     obj,
    std::vector<View*>&         cat_views, 
    std::vector<Object*>&       cat_objs,
    Likelihood::Correspondence* ll_corr,
    uint32_t                    c,
    uint32_t                    i
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;
    using jwscxx::base::Readable;

    std::string fname_1;
    std::string fname_2;

    View*   best_view = view->clone();
    Object* best_obj  = obj->clone();

    Vector_d* p = 0;
    Vector_d* means  = 0;
    Vector_d* stdevs = 0;

    *view = *(cat_views[ c ]);
    view->sample_var_param_likelihood(&p);
    try { view->set_var_param_values(p); }
    catch (jwscxx::base::Arg_error e) { }

    view->size(opts.images_in_max_num_cols + 2*opts.model_pad,
               opts.images_in_max_num_rows + 2*opts.model_pad);

    *best_view = *view;
    best_view->size(opts.images_in_max_num_cols + 2*opts.model_pad,
                    opts.images_in_max_num_rows + 2*opts.model_pad);

    *obj = *(cat_objs[ c ]);
    obj->sample_var_param_likelihood(&p);
    try { obj->set_var_param_values(p); }
    catch (jwscxx::base::Arg_error e) { }
    *best_obj = *obj;

    free_vector_d(p);
    free_vector_d(means);
    free_vector_d(stdevs);

    Likelihood::Cache* ll_cache = new Likelihood::Cache(opts.img_ll_cache_size,
            opts.img_ll_cache_k, opts.img_ll_cache_t);

    Matrix_u32* surf_ll_assign = 0;
    Matrix_u32* edge_ll_assign = 0;
    Matrix_u32* model_mat = 0;
    Vector_d* sample_mean_in = 0;
    Matrix_d* sample_cov_in = 0;

    double best_potential, best_img_ll;
    double potential, img_ll;

    Likelihood::calc_neg_log_likelihood(&best_potential, &best_img_ll,
            &model_mat, &surf_ll_assign, &edge_ll_assign, opts.images_t[ i ],
            opts.edges[ i ], ll_corr, best_view, best_obj, opts.img_ll_bg_prob,
            opts.img_ll_surf_prob, opts.img_ll_surf_noise_prob,
            opts.img_ll_surf_miss_prob, opts.img_ll_edge_noise_prob,
            opts.img_ll_edge_silho_miss_prob, opts.img_ll_edge_inner_miss_prob);

    Likelihood::calc_neg_log_likelihood(&potential, &img_ll,
            &model_mat, &surf_ll_assign, &edge_ll_assign, opts.images_t[ i ],
            opts.edges[ i ], ll_corr, view, obj, opts.img_ll_bg_prob,
            opts.img_ll_surf_prob, opts.img_ll_surf_noise_prob,
            opts.img_ll_surf_miss_prob, opts.img_ll_edge_noise_prob,
            opts.img_ll_edge_silho_miss_prob, opts.img_ll_edge_inner_miss_prob);

    read_sample_stats(&sample_mean_in, &sample_cov_in, c, i);

    printf_category_instance_out(fname_1, c, i, opts.sample_info_fname);
    printf_category_instance_out(fname_2, c, i, opts.sample_stats_info_fname);

    Sampler::Instance* ins;

    switch (opts.sampler)
    {
        case METRO:
            ins = new Metropolis::Instance(best_view, best_obj,
                    best_img_ll, best_potential, view, obj, img_ll, potential,
                    opts.images_t[ i ], opts.edges[ i ], ll_corr, ll_cache,
                    surf_ll_assign, edge_ll_assign, model_mat, sample_mean_in,
                    sample_cov_in, fname_1.c_str(), opts.compress_info,
                    fname_2.c_str());
            break;
        case STOCH:
            ins = new Stochastic_dynamics::Instance(best_view, best_obj,
                    best_img_ll, best_potential, view, obj, img_ll, potential,
                    opts.images_t[ i ], opts.edges[ i ], ll_corr, ll_cache,
                    surf_ll_assign, edge_ll_assign, model_mat, sample_mean_in,
                    sample_cov_in, fname_1.c_str(), opts.compress_info,
                    fname_2.c_str());
            break;
        case HYBRID:
            ins = new Hybrid_MC::Instance(best_view, best_obj,
                    best_img_ll, best_potential, view, obj, img_ll, potential,
                    opts.images_t[ i ], opts.edges[ i ], ll_corr, ll_cache,
                    surf_ll_assign, edge_ll_assign, model_mat, sample_mean_in,
                    sample_cov_in, fname_1.c_str(), opts.compress_info,
                    fname_2.c_str());
            break;
        case LANGE:
            ins = new Langevin::Instance(best_view, best_obj, best_img_ll,
                    best_potential, view, obj, img_ll, potential, 
                    opts.images_t[ i ], opts.edges[ i ], ll_corr, ll_cache,
                    surf_ll_assign, edge_ll_assign, model_mat, sample_mean_in,
                    sample_cov_in, fname_1.c_str(), opts.compress_info,
                    fname_2.c_str());
            break;
        case HYPER:
            ins = new Hyperdynamics::Instance(best_view, best_obj, 
                    best_img_ll, best_potential, view, obj, img_ll, potential,
                    opts.images_t[ i ], opts.edges[ i ], ll_corr, ll_cache,
                    surf_ll_assign, edge_ll_assign, model_mat, sample_mean_in,
                    sample_cov_in, fname_1.c_str(), opts.compress_info,
                    fname_2.c_str());
            break;
    }

    return ins;
}


/** @brief Initializes Instance from a file or randomly. */
static void read_or_sample_instances(Sampler::Parameters* params)
    throw (jwscxx::base::Exception)
{
    View* view;
    Object* obj;
    Likelihood::Correspondence* ll_corr;
    Sampler::Instance* ins;

    for (uint32_t i = 0; i < opts.num_images; i++)
    {
        for (uint32_t c = 0; c < opts.num_categories; c++)
        {
            view = new View(opts.view_distance, opts.view_width,
                    opts.view_height, opts.view_near, opts.view_far,
                    opts.view_ctx);

            obj = new Object(opts.obj_ctx, opts.block_ctx, opts.block_set_ctx);

            ll_corr = new Likelihood::Correspondence(opts.edges[ i ],
                    opts.reso_height, opts.reso_width, opts.img_ll_num_angles,
                    opts.img_ll_angle_sigma, opts.img_ll_dist_sigma,
                    opts.img_ll_max_dist);

            ins = read_or_sample_instance(view, obj, ll_corr, c, i);
            params->instance(ins, c, i);
        }
    }
}


/** @brief Initializes Instance from a file or randomly. */
static void sample_instances
(
    Sampler::Parameters*  params,
    std::vector<View*>&   cat_views, 
    std::vector<Object*>& cat_objs
)
throw (jwscxx::base::Exception)
{
    View* view;
    Object* obj;
    Likelihood::Correspondence* ll_corr;
    Sampler::Instance* ins;

    for (uint32_t c = 0; c < opts.num_categories; c++)
    {
        for (uint32_t i = 0; i < opts.num_images; i++)
        {
            view = new View(opts.view_distance, opts.view_width,
                    opts.view_height, opts.view_near, opts.view_far,
                    opts.view_ctx);

            obj = new Object(opts.obj_ctx, opts.block_ctx, opts.block_set_ctx);

            ll_corr = new Likelihood::Correspondence(opts.edges[ i ],
                    opts.reso_height, opts.reso_width, opts.img_ll_num_angles,
                    opts.img_ll_angle_sigma, opts.img_ll_dist_sigma,
                    opts.img_ll_max_dist);

            ins = sample_instance(view, obj, cat_views, cat_objs, ll_corr, c,i);
            params->instance(ins, c, i);
        }
    }
}


static void read_or_assign_category_members(Sampler::Parameters* params)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    if (!opts.categories_in_fname)
    {
        //params->assign_random_members();
        params->assign_members_to_one();
        return;
    }

    std::string fname;
    Error* e;

    uint32_t N_i = params->num_instances();
    uint32_t N_c = params->num_categories();

    Vector_d* membership = 0;

    for (uint32_t i = 0; i < N_i; i++)
    {
        for (uint32_t c = 0; c < N_c; c++)
        {
            printf_category_instance_in(fname, c, i, opts.categories_in_fname);
            if ((e = read_vector_d(&membership, fname.c_str())))
            {
                std::ostringstream ost;
                ost << fname << ": " << e->msg;
                throw jwscxx::base::IO_error(ost.str());
            }
            params->member(membership->elts[0], c, i);
        }
    }

    free_vector_d(membership);
}


static void write_category_members(Sampler::Parameters* params)
    throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    if (!opts.categories_out_fname)
        return;

    std::string fname;
    Error* e;

    uint32_t N_i = params->num_instances();
    uint32_t N_c = params->num_categories();

    Vector_d* membership = 0;
    create_vector_d(&membership, 1);

    for (uint32_t i = 0; i < N_i; i++)
    {
        for (uint32_t c = 0; c < N_c; c++)
        {
            membership->elts[0] = params->member(c, i);

            printf_category_instance_out(fname, c, i,opts.categories_out_fname);
            if ((e = write_vector_d(membership, fname.c_str())))
            {
                std::ostringstream ost;
                ost << fname << ": " << e->msg;
                throw jwscxx::base::IO_error(ost.str());
            }
        }
    }

    free_vector_d(membership);
}


/** @brief Creates a model category. */
static Sampler::Category* create_category
(
    Sampler::Parameters* params,
    uint32_t             c
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;

    Sampler::Category* cat;
    Sampler::Instance* rep_ins = params->instance(c, 0);

    double view_lp = rep_ins->view->calc_var_param_prior_log_prob();
    double obj_lp  = rep_ins->obj->calc_var_param_prior_log_prob();
    double lp_potential = - (view_lp + obj_lp);

    double potential = lp_potential;
    for (uint32_t i = 0; i < params->num_instances(); i++)
    {
        Sampler::Instance* ins = params->instance(c, i);
        potential += ins->ll_potential;
    }

    std::string fname;
    printf_category_out(fname, c, opts.sample_info_fname);

    switch (opts.sampler)
    {
        case METRO:
            cat = new Metropolis::Category(
                    dynamic_cast<Metropolis::Instance*>(rep_ins), 
                    lp_potential, potential, fname.c_str(), opts.compress_info);
            break;
        case STOCH:
            cat = new Stochastic_dynamics::Category(
                    dynamic_cast<Stochastic_dynamics::Instance*>(rep_ins), 
                    lp_potential, potential, fname.c_str(), opts.compress_info);
            break;
        case HYBRID:
            cat = new Hybrid_MC::Category(
                    dynamic_cast<Hybrid_MC::Instance*>(rep_ins),
                    lp_potential, potential, fname.c_str(), opts.compress_info);
            break;
        case LANGE:
            cat = new Langevin::Category(
                    dynamic_cast<Langevin::Instance*>(rep_ins),
                    lp_potential, potential, fname.c_str(), opts.compress_info);
            break;
        case HYPER:
            cat = new Hyperdynamics::Category(
                    dynamic_cast<Hyperdynamics::Instance*>(rep_ins),
                    lp_potential, potential, fname.c_str(), opts.compress_info);
            break;
    }

    return cat;
}


/** @brief Creates the model categories. */
static void create_categories(Sampler::Parameters* params)
    throw (jwscxx::base::Exception)
{
    Sampler::Category* cat;

    uint32_t N_c = params->num_categories();

    for (uint32_t c = 0; c < N_c; c++)
    {
        cat = create_category(params, c);
        params->category(cat, c);
    }
}


/** @brief Writes sample-estimated statistics for the posterior. */
static void write_instance_sample_stats
(
    const Sampler::Parameters* params,
    uint32_t                   c, 
    uint32_t                   i
)
throw (jwscxx::base::Exception)
{
    using namespace jwsc;
    using jwscxx::base::IO_error;

    if (!opts.sample_stats_out_fname)
        return;

    Vector_d* mean = 0;
    Matrix_d* cov  = 0;

    try
    {
        params->instance(c, i)->calc_sample_stats(&mean, &cov);
    }
    catch (jwscxx::base::Result_error e)
    {
        return;
    }

    std::string fname;
    printf_category_instance_out(fname, c, i, opts.sample_stats_out_fname);

    FILE* fp;
    if (!(fp = fopen(fname.c_str(), "w")))
    {
        std::ostringstream ost;
        ost << fname << ": Could not open file";
        throw IO_error(ost.str());
    }

    Error* err;
    if ((err = write_vector_with_header_fp_d(mean, fp)) ||
        (err = write_matrix_with_header_fp_d(cov, fp)))
    {
        IO_error ex = IO_error(err->msg);
        free_error(err);
        throw ex;
    }

    free_vector_d(mean);
    free_matrix_d(cov);

    fclose(fp);
}


/** @brief Writes the generated samples from the posterior. */
static void write_instance_samples
(
    const Sampler::Parameters* params,
    uint32_t                   c, 
    uint32_t                   i
) 
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
{
    using namespace jwsc;
    using jwscxx::base::IO_error;

    uint32_t N = params->instance(c, i)->samples.size();

    if (!opts.samples_out_fname || N == 0)
        return;

    uint32_t D = params->instance(c, i)->samples.front()->num_elts;

    Matrix_d* samples_mat = 0;
    create_matrix_d(&samples_mat, N, D);

    uint32_t n = 0;
    std::list<jwsc::Vector_d*>::const_iterator it;
    for (it = params->instance(c, i)->samples.begin(); 
            it != params->instance(c, i)->samples.end(); it++)
    {
        for (uint32_t d = 0; d < D; d++)
        {
            samples_mat->elts[ n ][ d ] = (*it)->elts[ d ];
        }
        n++;
    }

    std::string fname;
    printf_category_instance_out(fname, c, i, opts.samples_out_fname);

    Error* err;
    if ((err = write_matrix_d(samples_mat, fname.c_str())))
    {
        IO_error ex = IO_error(err->msg);
        free_error(err);
        throw ex;
    }

    free_matrix_d(samples_mat);
}


/** @brief Writes a Instance to files. */
static void write_instance_models
(
    Sampler::Parameters* params,
    uint32_t             c,
    uint32_t             i
) 
throw (jwscxx::base::Exception)
{
    using jwscxx::base::Writeable;

    std::string fname;

    if (opts.best_view_out_fname)
    {
        printf_category_instance_out(fname, c, i, opts.best_view_out_fname);
        ((Model*)(params->instance(c, i)->best_view))->Writeable::write(
                fname.c_str());
    }
    if (opts.best_object_out_fname)
    {
        printf_category_instance_out(fname, c, i, opts.best_object_out_fname);
        params->instance(c, i)->best_obj->Writeable::write(fname.c_str());
    }
    if (opts.view_out_fname)
    {
        printf_category_instance_out(fname, c, i, opts.view_out_fname);
        ((Model*)(params->instance(c, i)->view))->Writeable::write(fname.c_str());
    }
    if (opts.object_out_fname)
    {
        printf_category_instance_out(fname, c, i, opts.object_out_fname);
        params->instance(c, i)->obj->Writeable::write(fname.c_str());
    }
    if (opts.object_img_fname)
    {
        jwsc::Image_f* model_img = 0;
        Sampler::Instance* ins = params->instance(c, i);

        glEnable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glClearColor(1, 1, 1, 0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        ins->view->set_gl_projection();
        ins->view->set_gl_modelview();

        if (!opts.interactive)
        {
            int w = opts.view_width;
            int h = opts.view_height;
            float s = (float)h / (float)opts.reso_height;
            glPointSize(s+0.49999);
            glLineWidth(s);
            glViewport(0, 0, (GLsizei)w, (GLsizei)h); 
        }
        Render::full_image(i, ins->view);

        glColor3f(0, 1, 0);
        Render::edges(i, ins->view);
        glColor3f(1, 0, 0);
        Render::wire_object(ins->view, ins->obj);
        glFlush();

        View::capture_gl_view(&model_img);
        printf_category_instance_out(fname, c, i, opts.object_img_fname);

        jwsc::write_image_f(model_img, fname.c_str());
        jwsc::free_image_f(model_img);
    }
    if (opts.best_object_img_fname)
    {
        jwsc::Image_f* model_img = 0;
        Sampler::Instance* ins = params->instance(c, i);

        glEnable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);
        glClearColor(1, 1, 1, 0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        ins->view->set_gl_projection();
        ins->view->set_gl_modelview();

        if (!opts.interactive)
        {
            int w = opts.view_width;
            int h = opts.view_height;
            float s = (float)h / (float)opts.reso_height;
            glPointSize(s+0.49999);
            glLineWidth(s);
            glViewport(0, 0, (GLsizei)w, (GLsizei)h); 
        }
        Render::full_image(i, ins->best_view);

        glColor3f(0, 1, 0);
        Render::edges(i, ins->best_view);
        glColor3f(1, 0, 0);
        Render::wire_object(ins->best_view, ins->best_obj);
        glFlush();

        View::capture_gl_view(&model_img);
        printf_category_instance_out(fname, c, i, opts.best_object_img_fname);

        jwsc::write_image_f(model_img, fname.c_str());
        jwsc::free_image_f(model_img);
    }
}


/**
 * @param  argc  From the main function.
 * @param  argv  From the main function.
 */
void process_furniture_options(int argc, const char** argv)
    throw (jwscxx::base::Exception)
{
    init_opts();
    init_furniture_options();

    int argi = argc - options->process(argc, argv) + 1;
    read_images(argc, argv, argi);
    read_or_detect_edges();
    write_edges();
    write_edge_color();
    threshold_images();
    pad();
}


/**
 * @param  argc  From the main function.
 * @param  argv  From the main function.
 */
void process_furniture_cat_options(int argc, const char** argv)
    throw (jwscxx::base::Exception)
{
    init_opts();
    init_furniture_cat_options();
    options->process(argc, argv);

    opts.view_width  = opts.cat_sample_img_width;
    opts.view_height = opts.cat_sample_img_height;
    opts.reso_width  = opts.cat_sample_img_width;
    opts.reso_height = opts.cat_sample_img_height;
}


static void alloc_sampler_and_params
(
    Sampler**             sampler_out, 
    Sampler::Parameters** params_out
)
{
    assert(*sampler_out == 0);
    assert(*params_out == 0);

    Sampler* sampler;
    Sampler::Parameters* params;

    switch (opts.sampler)
    {
        case METRO:
            sampler = new Metropolis(opts.mh_diffusion_prob, opts.mh_jump_prob,
                    opts.mh_jump_set_size);
            params = new Metropolis::Parameters(opts.view_ctx, opts.obj_ctx,
                    opts.block_ctx, opts.block_set_ctx, opts.img_ll_bg_prob,
                    opts.img_ll_surf_prob, opts.img_ll_surf_noise_prob,
                    opts.img_ll_surf_miss_prob, opts.img_ll_edge_noise_prob,
                    opts.img_ll_edge_silho_miss_prob,
                    opts.img_ll_edge_inner_miss_prob, opts.sample_stats_freq,
                    opts.num_images, opts.num_categories, opts.min_num_members,
                    opts.num_threads);
            break;
        case STOCH:
            sampler = new Stochastic_dynamics(opts.stoch_alpha, 
                    opts.stoch_kick);
            params = new Stochastic_dynamics::Parameters(opts.view_ctx,
                    opts.obj_ctx, opts.block_ctx, opts.block_set_ctx,
                    opts.img_ll_bg_prob, opts.img_ll_surf_prob,
                    opts.img_ll_surf_noise_prob, opts.img_ll_surf_miss_prob,
                    opts.img_ll_edge_noise_prob,
                    opts.img_ll_edge_silho_miss_prob,
                    opts.img_ll_edge_inner_miss_prob, opts.sample_stats_freq,
                    opts.num_images, opts.num_categories, opts.min_num_members,
                    opts.num_threads);
            break;
        case HYBRID:
            sampler = new Hybrid_MC(opts.stoch_alpha, opts.stoch_kick, 
                    opts.hybrid_l);
            params = new Hybrid_MC::Parameters(opts.view_ctx, opts.obj_ctx,
                    opts.block_ctx, opts.block_set_ctx, opts.img_ll_bg_prob,
                    opts.img_ll_surf_prob, opts.img_ll_surf_noise_prob,
                    opts.img_ll_surf_miss_prob, opts.img_ll_edge_noise_prob,
                    opts.img_ll_edge_silho_miss_prob,
                    opts.img_ll_edge_inner_miss_prob, opts.sample_stats_freq,
                    opts.num_images, opts.num_categories, opts.min_num_members,
                    opts.num_threads);
            break;
        case LANGE:
            sampler = new Langevin();
            params = new Langevin::Parameters(opts.view_ctx, opts.obj_ctx,
                    opts.block_ctx, opts.block_set_ctx, opts.img_ll_bg_prob,
                    opts.img_ll_surf_prob, opts.img_ll_surf_noise_prob,
                    opts.img_ll_surf_miss_prob, opts.img_ll_edge_noise_prob,
                    opts.img_ll_edge_silho_miss_prob,
                    opts.img_ll_edge_inner_miss_prob, opts.sample_stats_freq,
                    opts.num_images, opts.num_categories, opts.min_num_members,
                    opts.num_threads);
            break;
        case HYPER:
            sampler = new Hyperdynamics(opts.hyper_gamma, opts.hyper_eta,
                    opts.hyper_h, opts.hyper_d);
            params = new Hyperdynamics::Parameters(opts.view_ctx, opts.obj_ctx,
                    opts.block_ctx, opts.block_set_ctx, opts.img_ll_bg_prob,
                    opts.img_ll_surf_prob, opts.img_ll_surf_noise_prob,
                    opts.img_ll_surf_miss_prob, opts.img_ll_edge_noise_prob,
                    opts.img_ll_edge_silho_miss_prob,
                    opts.img_ll_edge_inner_miss_prob, opts.sample_stats_freq,
                    opts.num_images, opts.num_categories, opts.min_num_members,
                    opts.num_threads);
            break;
    }

    *sampler_out = sampler;
    *params_out = params;
}


static void read_categories
(
    std::vector<View*>& cat_views, 
    std::vector<Object*>& cat_objs
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
{
    using jwscxx::base::Readable;

    View* view;
    Object* obj;
    std::string fname;

    cat_views.reserve(opts.num_categories);
    cat_objs.reserve(opts.num_categories);

    for (uint32_t c = 0; c < opts.num_categories; c++)
    {
        view = new View(opts.view_distance, opts.view_width,
                opts.view_height, opts.view_near, opts.view_far,
                opts.view_ctx);
        printf_category_instance_in(fname, c, 0, opts.view_in_fname);
        ((Model*)view)->Readable::read(fname.c_str());
        view->size(opts.images_in_max_num_cols + 2*opts.model_pad,
                   opts.images_in_max_num_rows + 2*opts.model_pad);

        obj = new Object(opts.obj_ctx, opts.block_ctx, opts.block_set_ctx);
        printf_category_instance_in(fname, c, 0, opts.object_in_fname);
        obj->Readable::read(fname.c_str());

        cat_views.push_back(view);
        cat_objs.push_back(obj);
    }
}


void read_category(View** view_out, Object** obj_out) 
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
{
    using jwscxx::base::Readable;

    assert(*view_out == 0);
    assert(*obj_out == 0);

    *view_out = new View(opts.view_distance, opts.view_width,
            opts.view_height, opts.view_near, opts.view_far,
            opts.view_ctx);
    View* view = *view_out;
    ((Model*)view)->Readable::read(opts.cat_view_in_fname);
    view->size(opts.cat_sample_img_width, opts.cat_sample_img_height);

    *obj_out = new Object(opts.obj_ctx, opts.block_ctx, opts.block_set_ctx);
    (*obj_out)->Readable::read(opts.cat_object_in_fname);
}


/** 
 * Depending on the program options, the sampler parameters may be initialized
 * randomly or read from file.
 */
void create_sampler_and_params
(
    Sampler**             sampler_out, 
    Sampler::Parameters** params_out
)
{
    alloc_sampler_and_params(sampler_out, params_out);
    read_or_sample_instances(*params_out);
    create_categories(*params_out);
    read_or_assign_category_members(*params_out);

    if (opts.sampler == METRO)
    {
        add_metropolis_moves(dynamic_cast<Metropolis*>(*sampler_out));
    }
}


/** 
 * Depending on the program options, the sampler parameters may be initialized
 * randomly or read from file.
 */
void create_rec_sampler_and_params
(
    Sampler**             sampler_out, 
    Sampler::Parameters** params_out
)
throw (jwscxx::base::Arg_error, jwscxx::base::IO_error)
{
    std::vector<View*> cat_views;
    std::vector<Object*> cat_objs;

    alloc_sampler_and_params(sampler_out, params_out);
    read_categories(cat_views, cat_objs);
    sample_instances(*params_out, cat_views, cat_objs);
    create_categories(*params_out);
    (*params_out)->assign_random_members();

    for (uint32_t i = 0; i < cat_views.size(); i++)
    {
        delete cat_views[ i ];
        delete cat_objs[ i ];
    }

    if (opts.sampler == METRO)
    {
        add_rec_metropolis_moves(dynamic_cast<Metropolis*>(*sampler_out));
    }
}


void write_results(Sampler::Parameters* params) throw (jwscxx::base::Exception)
{
    write_category_members(params);

    uint32_t N_i = params->num_instances();
    uint32_t N_c = params->num_categories();

    for (uint32_t i = 0; i < N_i; i++)
    {
        for (uint32_t c = 0; c < N_c; c++)
        {
            write_instance_sample_stats(params, c, i);
            write_instance_samples(params, c, i);
            write_instance_models(params, c, i);
            write_edge_map(params, c, i);
            write_threshold_images(params, c, i);
        }
    }
}


void clean_up_furniture()
{
    delete options;
    delete opts.view_ctx;
    delete opts.obj_ctx;
    delete opts.block_ctx;
    delete opts.block_set_ctx;

    for (size_t i = 0; i < opts.images_in.size(); i++)
    {
        free_image_f(opts.images_in[ i ]);
        free_image_f(opts.full_images_in[ i ]);
    }
    for (size_t i = 0; i < opts.images_t.size(); i++)
    {
        free_matrix_f(opts.images_t[ i ]);
    }
    for (size_t i = 0; i < opts.edges.size(); i++)
    {
        free_edge_set_f(opts.edges[ i ]);
    }
}


void clean_up_furniture_cat()
{
    delete options;
    delete opts.view_ctx;
    delete opts.obj_ctx;
    delete opts.block_ctx;
    delete opts.block_set_ctx;
}


void compress(const char* fname) throw (jwscxx::base::Exception)
{
#if defined FURNITURE_HAVE_ZLIB
    using jwscxx::base::IO_error;

    static const uint32_t buf_len = 1048576;
    uint8_t buf[ buf_len ];

    FILE* fp = fopen(fname, "r");
    if (!fp)
    {
        throw IO_error(strerror(errno));
    }

    std::string gzfname(fname);
    gzfname += ".gz";
    gzFile gzfp = gzopen(gzfname.c_str(), "w");
    if (!gzfp)
    {
        throw IO_error("Could not open compressed file");
    }

    int32_t N = fread(buf, sizeof(uint8_t), buf_len, fp);
    if (ferror(fp))
    {
        throw IO_error(strerror(errno));
    }

    while (N > 0)
    {
        if (gzwrite(gzfp, buf, N) != N)
        {
            throw IO_error("Could not write to compressed file");
        }

        N = fread(buf, sizeof(char), buf_len, fp);
        if (ferror(fp))
        {
            throw IO_error(strerror(errno));
        }
    }
    assert(feof(fp));

    if (fclose(fp) != 0)
    {
        throw IO_error(strerror(errno));
    }

    if (gzclose(gzfp) != 0)
    {
        throw IO_error("Could not close compressed file");
    }

    if (unlink(fname) != 0)
    {
        throw IO_error(strerror(errno));
    }
#else
    throw jwscxx::base::Exception("No compression support");
#endif
}
